#include "avoidObs/inc/PassagewayTree.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /*************** Class PassagewayNode *************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PassagewayNode::PassagewayNode() : Node<PassagewayWP*>()
{
  this->val_ = NULL;
}

/**
 *  \brief copy constructor
 */
PassagewayNode::PassagewayNode(const PassagewayNode& n) : Node<PassagewayWP*>(n)
{
  this->val_ = n.val()->copy();
}

// /**
//  *  \brief init constructor (from a passageway graph and system constraints)
//  *  \param pgraph : the input passageway graph
//  *  \param sys : the input system constraints
//  */
// PassagewayNode::PassagewayNode(const GraphPassageway& pgraph, const System& sys) : Node<PassagewayST*>()
// {
//   this->buildFromPGraph(pgraph, sys);
// }

/**
 *  \brief destructor
 */
PassagewayNode::~PassagewayNode()
{
  if(this->val_ != NULL) delete this->val_;
}


/**************** Utility function ****************/

/**
 *  \brief return a vector containing the state-times (for a differential drive system) that constitute the best branch
 *  \return a vector containing the state-times (for a differential drive system) that constitute the best branch
 */
vector<PassagewayWP>* PassagewayNode::bestBranch_PassagewayWP()
{
  double minT = 0;
  float maxS = 0;
  Node<PassagewayWP*>* bestNode = NULL;

  // get back the leaves of the tree
  vector<Node<PassagewayWP*>*> vecLeaves;
  this->getLeavesNodes(vecLeaves);

  // check if the vector of leaves is not empty
  if(vecLeaves.size() == 0) return NULL;

  // search for the node minimizing its time value and corresponding to the goal
  int N = vecLeaves.size();
  for(int i=0; i<N; i++)
    {
      float curS = vecLeaves.at(i)->val()->s;
      float curT = vecLeaves.at(i)->val()->t;

      if(i == 0 || curS > maxS || (curS == maxS && curT < minT))
	{
	  maxS = curS;
	  minT = curT;
	  bestNode = vecLeaves.at(i);
	}
    }

  // get back the branch of the tree going to the best found leaf
  std::list<PassagewayWP*> list_vals;
  bestNode->getVecValuesFromRoot(list_vals);

  // push all the values of the list into a resulting vector
  vector<PassagewayWP>* vec_res = new vector<PassagewayWP>;
  for(std::list<PassagewayWP*>::iterator it = list_vals.begin(); it != list_vals.end(); it++)
    {
      vec_res->push_back(**it);
    }
  return vec_res;
}
    
/**
 *  \brief return a vector containing the state-times (for a differential drive system) that constitute the best branch
 *  \return a vector containing the state-times (for a differential drive system) that constitute the best branch
 */
vector<DD2StateT>* PassagewayNode::bestBranch_DD2StateT()
{
  vector<PassagewayWP>* branch_pwp = this->bestBranch_PassagewayWP();

  // check if the best branch has been found
  if(branch_pwp == NULL) return NULL;

  // create the vector of differential drive states
  vector<DD2StateT>* branch_res = new vector<DD2StateT>;
  
  int N = branch_pwp->size();
  for(int i=0; i<N; i++)
    {
      branch_res->push_back(branch_pwp->at(i).toDD2StateT());
    }

  delete branch_pwp;
  return branch_res;
}


// /************ Passageway tree building ************/

// /*
//  *  \brief build a passageway tree from a a passageway graph and system constraints
//  *  \param pgraph : the input passageway graph
//  *  \param sys : the input system constraints
//  */
// void PassagewayNode::buildFromPGraph(const GraphPassageway& pgraph, const System& sys)
// {
// }
