#ifndef MPC_COLLISIONCHECK_HPP
#define MPC_COLLISIONCHECK_HPP

#include "motion/inc/PCTraj.hpp"
#include "motion/inc/PSTraj.hpp"
#include "system/inc/System.hpp"
#include "environment/inc/DEM.hpp"
#include "environment/inc/BoolObsMap.hpp"

#include <vector>

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>


namespace mpc
{

                                    /**************************************************/
                                    /************** Class CollisionCheck **************/
                                    /**************************************************/

  /** \class CollisionCheck
   *  \brief Collision checking methods
   */
  class CollisionCheck
  {
  public : 
    static double dt_traj_sampling_;               /**< trajectory sampling time */
    static bool unknown_is_obs_;                   /**< true if unknown cells of a grid map are considered as obstacles */

  public : 
    /**
     *  \brief check for collision between a trajectory and an obstacle map
     *  \param map : the obstacle map
     *  \param traj : the followed state trajectory
     *  \return INFINITY if no collisions exists between both trajectory and obstacle map
     *  \return the minimal time of collision if a collision may occur
     */
    static double checkCol(const BoolObsMap& map, const PSTraj& traj);

    /**
     *  \brief check for collision between a trajectory and an obstacle map
     *  \param map : the obstacle map
     *  \param traj : the followed state trajectory
     *  \return INFINITY if no collisions exists between both trajectory and obstacle map
     *  \return the minimal time of collision if a collision may occur
     */
    static double checkCol(const BoolObsMap& map, const STraj& traj);

    /**
     *  \brief check for collision all along a trajectory
     *  \param map : the obstacle map
     *  \param traj : the followed state trajectory
     *  \param vec_col : resulting vector containing the status (collision=true or free=false) of each configuration along the discrete path
     */
    static void checkCol(const BoolObsMap& map, const STraj& straj, vector<bool>& vec_col);
  };
};

#endif // MPC_COLLISIONCHECK_HPP
