#ifndef VEC_HPP
#define VEC_HPP

#include <vector>
// #include <QList>

class Vec
{
protected : 
  std::vector <double>* tab_;
  bool autoDelete_;

public : 
  /********** Constructors and destructors **********/
  Vec();                                              // default constructor
  Vec(const Vec& v);                                  // copy constructor
  Vec(int n);                                         // init construtor (init a vector of n elements)
  virtual ~Vec();                                     // destructor


  /******************* Accessors ********************/
  int size() const;                                   // return the size of the vector
  double at(int i);                                   // return the i-th element of the vector

  /*************** Utility functions ****************/
  void clear();                                       // delete all the elements of the vector
  bool equals(const Vec& v) const;                    // return true if v and the current vector are equals
  bool tabNull() const;                               // return true if tab_ is null
  void display() const;                                     // display the current vector

  /******************* Operators ********************/
  double& operator[](int i);                          // operator[] : access to the i-th element of the vector
  void operator+=(const Vec& v);                      // operator+= : add vector v to the current vector
  void operator=(const Vec& v);                       // operator=  : assign the vector v to the current vector

  /***************** Math functions *****************/
//   static Vec barycentre(const QList<Vec>& ls);        // return the barycentre of a list of vectors
  static Vec barycentre(const std::vector<Vec>& ls);  // return the barycentre of a table of vectors

  double norm() const;                                // return the norm of the current vector
  double squareNorm() const;                          // return the square of the norm of the current vector
  Vec normalized() const;                             // return a normalized vector from the current vector

public : 
  /************** Protected accessors ***************/
  double getVal(int i) const;                         // access to the i-th element of the vector
  void setVal(int i,double v);                        // set the i-th element of the vector


  /******************* Operators ********************/
Vec operator+(const Vec& v2) const;                   // operator+  : add two vectors
Vec operator-(const Vec& v2) const;                   // operator-  : sub two vectors
bool operator==(const Vec& v2) const;                 // operator== : return true if both vectors are equals
Vec operator*(double d) const;                        // operator*  : multiply a vector by a scalar
double operator*(Vec& v2) const;                      // operator*  : scalar product between two vectors
Vec operator/(double d) const;                        // operator/  : divide a vector by a scalar

};

#endif // VEC_HPP


#ifndef VEC2D_HPP
#define VEC2D_HPP

class Vec2D : public Vec
{
public : 
  /********** Constructors and destructors **********/
  Vec2D();                                            // default constructor
  Vec2D(double x,double y);                           // init constructor
  Vec2D(const Vec2D& v);                              // copy constructor
  virtual ~Vec2D();                                   // destructor
};

#endif // VEC2D_HPP


#ifndef VEC3D_HPP
#define VEC3D_HPP

class Vec3D : public Vec
{
public : 
  /********** Constructors and destructors **********/
  Vec3D();                                            // default constructor
  Vec3D(double x,double y,double z);                  // init constructor
  Vec3D(const Vec3D& v);                              // copy constructor
  virtual ~Vec3D();                                   // destructor

  /******************* Operators ********************/
//   Vec3D operator*(const Vec3D& v2) const;             // operator*  : dot product (produit scalaire)
  Vec3D operator^(const Vec3D& v2) const;             // operator^  : cross product (produit vectoriel)

//   /***************** Math functions *****************/
//   Vec3D normalized() const;                           // return a normalized vector from the current vector
};

#endif // VEC3D_HPP


#ifndef VEC5D_HPP
#define VEC5D_HPP

class Vec5D : public Vec
{
public : 
  /********** Constructors and destructors **********/
  Vec5D();                                            // default constructor
  Vec5D(const Vec5D& v);                              // copy constructor
  virtual ~Vec5D();                                   // destructor
};

#endif // VEC5D_HPP


#ifndef VEC7D_HPP
#define VEC7D_HPP

class Vec7D : public Vec
{
public : 
  /********** Constructors and destructors **********/
  Vec7D();                                            // default constructor
  Vec7D(const Vec7D& v);                              // copy constructor
  virtual ~Vec7D();                                   // destructor
};

#endif // VEC7D_HPP
