#include "Object.hpp"
using namespace mpc;

#include <cstdlib>

                                    /**************************************************/
                                    /****************** Class Object ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Object::Object()
{
//   this->pos_    = NULL;
  this->bshape_ = NULL;
}

/**
 *  \brief copy constructor
 *  \param c : the object to copy
 */
Object::Object(const Object& o)
{
  // the position cannot be copie here
}

/**
 *  \brief destructor
 */
Object::~Object()
{
//   if(this->pos_) delete this->pos_;
  if(this->bshape_) delete this->bshape_;
}


/******************** Accessors *******************/

/**
 *  \brief setter : set the bounding shape of the object
 *  \param bs : the new bounding shape of the object
 */
void Object::setBShape(const BoundingShape& bs)
{ 
  switch(bs.type())
    {
    case UNDEFINED_BSHAPE : 
      break;
    case SPHERE_BSHAPE : 
      this->bshape_ = new BDisk(*((BDisk*)(&bs)));
      break;
    case POLYGON_BSHAPE : 
      this->bshape_ = new BPolygon(*((BPolygon*)(&bs)));
      break;
//     case BOX_BSHAPE : 
//       this->bshape_ = new BBox(*((BBox*)(&bs)));
//       break;
    default : 
      break;
    }
  
}


                                    /**************************************************/
                                    /***************** Class Object2D *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Object2D::Object2D() : Object()
{
}

/**
 *  \brief init constructor
 *  \param bs : the bounding shape of the object
 */
Object2D::Object2D(const BoundingShape& bs)
{
  // copy the bounding shape
  switch(bs.type())
    {
    case UNDEFINED_BSHAPE : 
      this->bshape_ = NULL;
      break;
    case POINT_BSHAPE : 
      this->bshape_ = new PointShape(*((PointShape*)(&bs)));
      break;
    case SPHERE_BSHAPE :
      this->bshape_ = new BDisk(*((BDisk*)(&bs)));
      break;
    case POLYGON_BSHAPE : 
      this->bshape_ = new BPolygon(*((BPolygon*)(&bs)));
      break;
    case BOX_BSHAPE : 
      this->bshape_ = new BBox(*((BBox*)(&bs)));
      break;
    default : 
      this->bshape_ = NULL;
      break;
    }
}

/**
 *  \brief copy constructor
 *  \param o : the object to copy
 */
Object2D::Object2D(const Object2D& o) : Object()
{
//   // copy the position
//   this->pos_ = new PositionXYTh((PositionXYTh)(o.pos()));

  if(bshape_ != NULL)
    {
      // copy the bounding shape
      switch(this->bshape().type())
	{
	case UNDEFINED_BSHAPE : 
	  this->bshape_ = NULL;
	  break;
	case POINT_BSHAPE : 
	  this->bshape_ = new PointShape(*((PointShape*)(o.bshape_)));
	  break;
	case SPHERE_BSHAPE : 
	  this->bshape_ = new BDisk(*((BDisk*)(o.bshape_)));
	  break;
	case POLYGON_BSHAPE : 
	  this->bshape_ = new BPolygon(*((BPolygon*)(o.bshape_)));
	  break;
	case BOX_BSHAPE : 
	  this->bshape_ = new BBox(*((BBox*)(o.bshape_)));
	  break;
	default : 
	  this->bshape_ = NULL;
	  break;
	}
    }
  else
    {
      //cout << "bshape null :-(" << endl;
    }
}

/**
 *  \brief destructor
 */
Object2D::~Object2D()
{
  //if(this->bshape_ != NULL) delete this->bshape_;
}


/**************** Motion functions ****************/

/**
 *  \brief translate the system component
 *  \param p : vector of the translation
 */
void Object2D::translate(const PositionXY& p)
{
  this->pos().translate(p);
}

/**
 *  \brief rotate the system component around p about an angle a
 *  \param p : center of the rotation
 *  \param a : angle of the rotation
 */
void Object2D::rotate(const PositionXY& p,const double& a)
{
  this->pos().rotate(p,a);
}
