#include "PositionTime.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class PositionXYT ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PositionXYT::PositionXYT()
{
  this->coord_ = vector<double>(3);
}

/**
 *  \brief copy constructor
 *  \param p : the position to copy
 */
PositionXYT::PositionXYT(const PositionXYT& p) : PositionXY(p)
{
}

/**
 *  \brief init constructor
 *  \param x : the first position coord
 *  \param y : the second position coord
 */
PositionXYT::PositionXYT(double x,double y,double t)
{
  this->coord_ = vector<double>(3);
  this->x() = x;
  this->y() = y;
  this->t() = t;
}

/**
 *  \brief destructor
 */
PositionXYT::~PositionXYT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : assignation
 *  \param p2 : the position-time to assign to the current one
 */
void PositionXYT::operator=(const PositionXYT& p2)
{
  this->x() = p2.x();
  this->y() = p2.y();
  this->t() = p2.t();
}

/**
 *  \brief operator + : sumation
 *  \param p2 : the second vector to add
 *  \return the sum of both positions
 */
PositionXYT PositionXYT::operator+(const PositionXYT& p2) const
{
  PositionXYT pres;
  pres.x() = this->x() + p2.x();
  pres.y() = this->y() + p2.y();
  pres.t() = this->t() + p2.t();
  return pres;
}
  
/**
 *  \brief operator + : substraction
 *  \param p2 : the second position to substract
 *  \return the substraction of both positions
 */
PositionXYT PositionXYT::operator-(const PositionXYT& p2) const
{
  PositionXYT pres;
  pres.x() = this->x() - p2.x();
  pres.y() = this->y() - p2.y();
  pres.t() = this->t() - p2.t();
  return pres;
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current position and the scalar
 */
PositionXYT PositionXYT::operator*(double d) const
{
  PositionXYT pres;
  pres.x() = this->x() * d;
  pres.y() = this->y() * d;
  pres.t() = this->t() * d;
  return pres;
}

/**
 *  \brief operator / : division by a scalar
 *  \param d : the scalar to divide with
 *  \return the division result
 */
PositionXYT PositionXYT::operator/(double d) const
{
  PositionXYT pres;
  if(d != 0)
    {
      pres.x() = this->x() / d;
      pres.y() = this->y() / d;
      pres.t() = this->t() / d;
    }
  return pres;
}

/**
 *  \brief operator ^ : cross product
 *  \param p2 : the second vector
 *  \return the cross product of both vectors
 */
PositionXYT PositionXYT::operator^(const PositionXYT& p2) const
{
  return PositionXYT(this->y()*p2.t() - this->t()*p2.y(),
		     this->t()*p2.x() - this->x()*p2.t(),
		     this->x()*p2.y() - this->y()*p2.x());
}


                                    /**************************************************/
                                    /*************** Class PositionXYYhT **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PositionXYThT::PositionXYThT()
{
  this->coord_ = vector<double>(4);
}

/**
 *  \brief copy constructor
 *  \param p : the position to copy
 */
PositionXYThT::PositionXYThT(const PositionXYThT& p) : PositionXYTh(p)
{
}

/**
 *  \brief init constructor
 *  \param x : the first position coord
 *  \param y : the second position coord
 */
PositionXYThT::PositionXYThT(double x,double y,double th,double t)
{
  this->coord_ = vector<double>(4);
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
PositionXYThT::~PositionXYThT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : assignation
 *  \param p2 : the position-time to assign to the current one
 */
void PositionXYThT::operator=(const PositionXYThT& p2)
{
  this->x()  = p2.x();
  this->y()  = p2.y();
  this->th() = p2.th();
  this->t()  = p2.t();
}
