#include "State.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /******************* Class State ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
State::State() : Vector()
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
State::State(const State& s) : Vector(s)
{
}

/**
 *  \brief init constructor
 *  \param n : dimension of the state
 */
State::State(int n) : Vector(n)
{
}

/**
 *  \brief destructor
 */
State::~State()
{
}


                                    /**************************************************/
                                    /****************** Class PMState *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PMState::PMState() : State(4)
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
PMState::PMState(const PMState& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param sx : speed sx
 *  \param sy : speed sy
 */
PMState::PMState(double x,double y,double sx,double sy) : State(4)
{
  this->x()  = x;
  this->y()  = y;
  this->sx() = sx;
  this->sy() = sy;
}

/**
 *  \brief destructor
 */
PMState::~PMState()
{
}

/**
 *  \brief n constructor
 *  \param n : the size of the state vector
 */
PMState::PMState(int n) : State(n)
{
}


                                    /**************************************************/
                                    /***************** Class PM3DState ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PM3DState::PM3DState() : State(6)
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
PM3DState::PM3DState(const PM3DState& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param z : position z
 *  \param sx : speed sx
 *  \param sy : speed sy
 *  \param sz : speed sz
 */
PM3DState::PM3DState(double x,double y,double z,double sx,double sy,double sz) : State(6)
{
  this->x()  = x;
  this->y()  = y;
  this->z()  = z;
  this->sx() = sx;
  this->sy() = sy;
  this->sz() = sz;
}

/**
 *  \brief destructor
 */
PM3DState::~PM3DState()
{
}

/**
 *  \brief n constructor
 *  \param n : the size of the state vector
 */
PM3DState::PM3DState(int n) : State(n)
{
}


                                    /**************************************************/
                                    /****************** Class CLState *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CLState::CLState() : State(5)
{ 
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
CLState::CLState(const CLState& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param ph : steering angle
 *  \param s : speed s
 */
CLState::CLState(double x,double y,double th,double ph,double s) : State(5)
{
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->ph() = ph;
  this->s()  = s;
}

/**
 *  \brief destructor
 */
CLState::~CLState()
{
}

/**
 *  \brief n constructor
 *  \param n : the size of the state vector
 */
CLState::CLState(int n) : State(n)
{
}


                                    /**************************************************/
                                    /****************** Class SSState *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SSState::SSState() : State(6)
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
SSState::SSState(const SSState& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param sx : speed sx
 *  \param sy : speed sy
 *  \param om : angular velocity
 */
SSState::SSState(double x,double y,double th,double sx,double sy,double om) : State(6)
{
  this->x()  = x;
  this->y()  = y;
  this->th() = th;
  this->sx() = sx;
  this->sy() = sy;
  this->om() = om;
}

/**
 *  \brief destructor
 */
SSState::~SSState()
{
}

/**
 *  \brief n constructor
 *  \param n : the size of the state vector
 */
SSState::SSState(int n) : State(n)
{
}


                                    /**************************************************/
                                    /****************** Class DDState *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DDState::DDState() : State(5)
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
DDState::DDState(const DDState& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param omr : right wheel velocity
 *  \param oml : left wheel velocity
 */
DDState::DDState(double x,double y,double th,double omr,double oml) : State(5)
{
  this->x()   = x;
  this->y()   = y;
  this->th()  = th;
  this->omr() = omr;
  this->oml() = oml;
}

/**
 *  \brief destructor
 */
DDState::~DDState()
{
}

/**
 *  \brief n constructor
 *  \param n : the size of the state vector
 */
DDState::DDState(int n) : State(n)
{
}


                                    /**************************************************/
                                    /****************** Class DD2State *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2State::DD2State() : State(5)
{
}

/**
 *  \brief copy constructor
 *  \param p : the control to copy
 */
DD2State::DD2State(const DD2State& s) : State(s)
{
}

/**
 *  \brief init constructor
 *  \param x : position x
 *  \param y : position y
 *  \param th : heading th
 *  \param s : linear velocity
 *  \param om : angular velocity
 */
DD2State::DD2State(double x,double y,double th,double s,double om) : State(5)
{
  this->x()   = x;
  this->y()   = y;
  this->th()  = th;
  this->s()   = s;
  this->om()  = om;
}

/**
 *  \brief destructor
 */
DD2State::~DD2State()
{
}

/**
 *  \brief n constructor
 *  \param n : the size of the state vector
 */
DD2State::DD2State(int n) : State(n)
{
}


//                                     /**************************************************/
//                                     /***************** Class ManaState ****************/
//                                     /**************************************************/

// /********** Constructors and destructors **********/

// /**
//  *  \brief default constructor
//  */
// ManaState::ManaState() : WCState()
// {
// }

// /**
//  *  \brief copy constructor
//  *  \param p : the control to copy
//  */
// ManaState::ManaState(const ManaState& s) : WCState(s)
// {
// }

// /**
//  *  \brief init constructor
//  *  \param x : position x
//  *  \param y : position y
//  *  \param th : heading th
//  *  \param v : linear velocity
//  *  \param w : angular velocity
//  */
// ManaState::ManaState(double x,double y,double th,double v,double w) : WCState(x,y,th,v,w)
// {
// }

// /**
//  *  \brief destructor
//  */
// ManaState::~ManaState()
// {
// }
