#include "Vec.hpp"

#include <cmath>
#include <iostream>
using namespace std;

double eeps = 1e-10;

using namespace std;

/****************************************************************************************
 ************************************ class Vec *****************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
Vec::Vec()
{
  tab_ = new vector <double>;
}

// copy constructor
Vec::Vec(const Vec& v)
{
  if(v.tab_ == NULL)
    {
      return;
    }
  tab_ = new vector <double>;
  for(int i=0; i<v.size(); i++)
    {
      //(*(this->tab_))[i] = (*(v.tab_))[i];
      this->tab_->push_back((*(v.tab_))[i]);
    }
}

// init construtor (init a vector of n elements)
Vec::Vec(int n)
{
  tab_ = new vector <double>;
  for(int i=0; i<n; i++)
    {
      this->tab_->push_back(0);
    }
}

// destructor
Vec::~Vec()
{
  if(this->tab_ != NULL)
    {
      if(this->tab_->size() != 0) this->tab_->clear();
      delete this->tab_;
      this->tab_ = NULL;
    }
}


/******************* Accessors ********************/

// return the size of the vector
int Vec::size() const
{
  if(this->tabNull())
    {
      return -1;
    }
  else
    {
      return this->tab_->size();
    }
}

// return the i-th element of the vector
double Vec::at(int i)
{
//   if(!this->tabNull())
//     {
      return (*(this->tab_))[i];
//     }
//   else 
//     {
//       return 0;
//     }
}


/*************** Utility functions ****************/

// delete all the elements of the vector
void Vec::clear()
{
  if(!this->tabNull())
    {
      this->tab_->clear();
    }
}

// return true if v and the current vector are equals
bool Vec::equals(const Vec& v) const
{
  if(this->tabNull() || v.tabNull())
    {
      return false;
    }
  else if(this->size() != v.size())
    {
      return false;
    }
  else
    {
      int i=0;
      bool trouve = false;
      while(i<this->size() && !trouve)
	{
	  if(this->getVal(i) >= v.getVal(i) - eeps && this->getVal(i) <= v.getVal(i) + eeps)
	    i++;
	  else
	    trouve = true;
	}
      if(trouve) return false;
      else return true;
    }
}

// return true if tab_ is null
bool Vec::tabNull() const
{
  return (this->tab_ == NULL);
}

// display the current vector
void Vec::display() const 
{
  if(this->tabNull())
    {
      cout << "NULL ";
    }
  else
    {
      cout << "(";
      for(int i=0; i<this->size()-1; i++)
	{
	  cout << this->getVal(i);
	  cout << ",\t";
	}
      if(this->size() != 0) cout << this->getVal(this->size()-1);
      cout << ") ";
    }
}


/******************* Operators ********************/

// operator[] : access to the i-th element of the vector
double& Vec::operator[](int i)
{
//   if(!(this->tabNull()))
//     {
      return (*(this->tab_))[i];
//     }
//   else 
//     {
//       double *a = new double;
//       *a = 0;
//       return *a;
//     }
}

// operator+= : add vector v to the current vector
void Vec::operator+=(const Vec& v)
{
  if(this->tabNull() || v.tabNull())
    {
      return;
    }
  else if(this->size() != v.size())
    {
      return;
    }
  else
    {
      for(int i=0; i<this->size(); i++)
	{
	  (*(this->tab_))[i] += v.getVal(i);
	}
    }
}

// operator+  : add two vectors
Vec Vec::operator+(const Vec& v2) const
{
  if(this->tabNull() || v2.tabNull())
    {
      return Vec();
    }
  else if(this->size() != v2.size())
    {
      return Vec();
    }
  else
    {
      Vec v(this->size());
      for(int i=0; i<v.size(); i++)
	{
	  v[i] = this->getVal(i) + v2.getVal(i);
	}
      return v;
    }
}

// operator-  : sub two vectors
Vec Vec::operator-(const Vec& v2) const
{
  if(this->tabNull() || v2.tabNull())
    {
      return Vec();
    }
  else if(this->size() != v2.size())
    {
      return Vec();
    }
  else
    {
      Vec v(this->size());
      for(int i=0; i<v.size(); i++)
	{
	  v[i] = this->getVal(i) - v2.getVal(i);
	}
      return v;
    }
}

// operator=  : assign the vector v to the current vector
void Vec::operator=(const Vec& v)
{
  if(v.tabNull())
    {
      return;
    }
  if(this->tabNull())
    {
      tab_ = new vector <double>;
    }
  if(this->size() == 0)
    {
      for(int i=0; i<v.size(); i++)
	{
	  this->tab_->push_back(0);
	}
    }
  if(this->size() != v.size())
    {
      return;
    }
  else
    {
      for(int i=0; i<this->size(); i++)
	{
	  (*(this->tab_))[i] = v.getVal(i);
	}
      return;
    }
}

// operator== : return true if both vectors are equals
bool Vec::operator==(const Vec& v2) const
{
  return this->equals(v2);
}

// operator*  : multiply a vector by a scalar
Vec Vec::operator*(double d) const
{
  if(this->tabNull())
    {
      return Vec();
    }
  else
    {
      Vec res(this->size());
      for(int i=0; i<res.size(); i++)
	{
	  res[i] = this->getVal(i) * d;
	}
      return res;
    }
}

// operator*  : scalar product between two vectors
double Vec::operator*(Vec& v2) const
{
  // check if the size of both vector is the same
  if(this->size() == v2.size())
    {
      // compute the scalar product
      double res = 0;
      for(int i=0; i<this->size(); i++)
	{
	  res += (this->getVal(i) * v2.getVal(i));
	}
      return res;
    }
  else
    {
      return 0;
    }
}

// operator/  : divide a vector by a scalar
Vec Vec::operator/(double d) const
{
  if(this->tabNull())
    {
      return Vec();
    }
  else
    {
      Vec res(this->size());
      for(int i=0; i<res.size(); i++)
	{
	  res[i] = this->getVal(i) / d;
	}
      return res;
    }
}


/***************** Math functions *****************/

// // return the barycentre of a list of vectors
// Vec Vec::barycentre(const QList<Vec>& ls)
// {
//   if(ls.count() == 0)
//     {
//       return Vec();
//     }
//   if((ls[0]).tabNull())
//     {
//       return Vec();
//     }
//   Vec res((ls[0]).size());
  
//   for(int i=0; i<ls.count(); i++)
//     {
//       if((ls[i]).tabNull() || (ls[i]).size() != res.size())
// 	return Vec();
//       else
// 	res += ls[i];
//     }
//   return res / ((double)ls.count());
// }

// return the barycentre of a table of vectors
Vec Vec::barycentre(const vector<Vec>& ls)
{
  if(ls.size() == 0)
    {
      return Vec();
    }
  if((ls[0]).tabNull())
    {
      return Vec();
    }
  Vec res((ls[0]).size());
  
  for(unsigned int i=0; i<ls.size(); i++)
    {
      if((ls[i]).tabNull() || (ls[i]).size() != res.size())
	return Vec();
      else
	res += ls[i];
    }
  return res / ((double)ls.size());
}

// return the norm of the current vector
double Vec::norm() const
{
  return sqrt(squareNorm());
}

// return the square of the norm of the current vector
double Vec::squareNorm() const
{
  if(this->tabNull())
    {
      return 0;
    }
  else
    {
      double res = 0;
      for(int i=0; i<this->size(); i++)
	{
	  res += (this->getVal(i) * this->getVal(i));
	}
      return res;
    }
}

// return a normalized vector from the current vector
Vec Vec::normalized() const
{
  // get back the norm of the vector
  double norm = this->norm();

  // check if the norm is not null
  if(norm == 0)
    {
      return Vec(this->size());
    }
  else
    {
      Vec vres;
      for(int i=0; i<this->size(); i++)
	{
	  vres.tab_->push_back(this->getVal(i) / norm);
	}
      return vres;
    }
}


/************** Protected accessors ***************/

// access to the i-th element of the vector
double Vec::getVal(int i) const
{
//   if(!(this->tabNull()))
//     {
      return (*(this->tab_))[i];
//     }
//   else 
//     {
//       return 0;
//     }
}

// set the i-th element of the vector
void Vec::setVal(int i,double v)
{
  if(this->tabNull() || i >= this->size())
    {
      return;
    }
  else
    {
      (*(this->tab_))[i] = v;
    }
}


/****************************************************************************************
 *********************************** class Vec2D ****************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
Vec2D::Vec2D() : Vec(2)
{
}

// init constructor
Vec2D::Vec2D(double x,double y)
{
  this->tab_->push_back(x);
  this->tab_->push_back(y);
}

// copy constructor
Vec2D::Vec2D(const Vec2D& v) : Vec(v)
{
}

// destructor
Vec2D::~Vec2D()
{
  this->clear();
//   if(tab_ != NULL) delete(tab_);
}


/****************************************************************************************
 *********************************** class Vec3D ****************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
Vec3D::Vec3D() : Vec(3)
{
}

// init constructor
Vec3D::Vec3D(double x,double y,double z) : Vec(3)
{
  this->setVal(0,x);
  this->setVal(1,y);
  this->setVal(2,z);
}

// copy constructor
Vec3D::Vec3D(const Vec3D& v) : Vec(v)
{
}

// destructor
Vec3D::~Vec3D()
{
  this->clear();
  //if(tab_ != NULL) delete(tab_);
}


/******************* Operators ********************/

// // operator*  : dot product (produit scalaire)
// Vec3D Vec3D::operator*(const Vec3D& v2) const
// {
// }

// operator^  : cross product (produit vectoriel)
Vec3D Vec3D::operator^(const Vec3D& v2) const
{
  return Vec3D(this->getVal(1)*v2.getVal(2) - this->getVal(2)*v2.getVal(1),
	       this->getVal(2)*v2.getVal(0) - this->getVal(0)*v2.getVal(2),
	       this->getVal(0)*v2.getVal(1) - this->getVal(1)*v2.getVal(0));
}

/***************** Math functions *****************/

// // return a normalized vector from the current vector
// Vec3D Vec3D::normalized() const
// {
//   // get back the norm of the vector
//   double norm = this->norm();

//   // check if the norm is not null
//   if(norm == 0)
//     {
//       return Vec3D();
//     }
//   else
//     {
//       Vec3D vres;
//       for(int i=0; i<this->size(); i++)
// 	{
// 	  vres.tab_->push_back(this->getVal(i) / norm);
// 	}
//       return vres;
//     }
// }


/****************************************************************************************
 *********************************** class Vec5D ****************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
Vec5D::Vec5D() : Vec(5)
{
}

// copy constructor
Vec5D::Vec5D(const Vec5D& v) : Vec(v)
{
}

// destructor
Vec5D::~Vec5D()
{
  this->clear();
  if(tab_ != NULL) delete(tab_);
}


/****************************************************************************************
 *********************************** class Vec7D ****************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
Vec7D::Vec7D() : Vec(7)
{
}

// copy constructor
Vec7D::Vec7D(const Vec7D& v) : Vec(v)
{
}

// destructor
Vec7D::~Vec7D()
{
  this->clear();
  if(tab_ != NULL) delete(tab_);
}
