#include "controller/inc/SimpleController.hpp"
using namespace mpc;

#include <stdlib.h>

                                    /**************************************************/
                                    /************* Class SimpleController *************/
                                    /**************************************************/

System* system_;                                       /**< the system used */
double dt_;                                            /**< period at which is called the controller */

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 *  Only used by subclasses
 */
SimpleController::SimpleController()
{
  this->sys_ = NULL;
  this->dt_ = 0.5;
  this->trajgen_ = NULL;
  this->debug_ = 0;
  this->ctraj_ = NULL;
  this->straj_ = NULL;
}

/**
 *  \brief init constructor
 *  \param sys : the robotic system handled
 *  \param dt : the controller period
 *  \param debug : the debugging level
 */
SimpleController::SimpleController(System* sys,double dt, int debug)
{
  this->sys_ = sys;
  this->dt_ = dt;
  this->trajgen_ = NULL;
  this->debug_ = debug;
  this->ctraj_ = NULL;
  this->straj_ = NULL;
}

/**
 *  \brief destructor
 */
SimpleController::~SimpleController()
{
  boost::mutex::scoped_lock lock(mutexTraj_);
  if(ctraj_ != NULL) delete ctraj_;
  if(straj_ != NULL) delete straj_;
}


/*************** Getters and Setters **************/

/**
 *  \brief return the controller period
 *  \return the controller period
 */
double SimpleController::dt() const
{
  return this->dt_;
}

/**
 *  \brief set the controller period
 *  \param dt : the new controller period
 */
void SimpleController::setDT(double dt)
{
  this->dt_ = dt;
}


/*************** Control Computation **************/

/**
 *  \brief Initialize the standard parameters of the trajectory generator
 *  \param dbg : the trajectory generator debugging mode
 */
void SimpleController::initTrajGen(int dbg)
{
  if(trajgen_ != NULL)
    {
      /// common trajgen options
      trajgen_->setTcStatus(TIME_FREE,0,999999999.9);
      
      // init options of the trajectory generation
      trajgen_->setMaxNbStep(20);                 // default = 20
      trajgen_->setMaxResidual(0.05);             // default = 0.05
      
      // init the debug level mode
      trajgen_->setDBG(dbg);                      // default = 0
    }
}
