#ifndef DDDEFORMATION_HPP
#define DDDEFORMATION_HPP

#include "STDeformation.hpp"
#include "DDTrajGen.hpp"
//#include "DiffDriveSys.hpp"

#include <vector>
#include <list>


class DDDeformation : public STDeformation
{
public : 
  std::vector <Trajectory*> vecTraj_;
  std::list <Vec_DP*>* listCtrl_;
  std::list <int>* feaCtrl_;

  DiffDriveSys* diffdrive_;

  bool parallelExec;
  unsigned int nbThread;

protected : 
  
public : 
  /************** Constructors and destructors *************/
  
  DDDeformation();                                           // default constructor
  DDDeformation(Trajectory *traj,Environment *env,Robot *rb,int** typeS,StateR*** tns,StateR** tfi,StateR** tfe); 
                                                             // init constructor
  virtual ~DDDeformation();                                  // destructor

  /****************** Getters and Setters ******************/
  inline bool isParallelExec() const { return parallelExec; }
  inline void setParallelExec(bool p) { parallelExec = p; }
  inline int getNbThread() const { return nbThread; }
  inline void setNbThread(int nbt) { nbThread = nbt; }

  /*** Computing of the forces applied on the trajectory ***/

  DDStateT computeDDFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int *fea);
    // computing of the internal forces
  void computeDDFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int *fea, DDStateT& fres);
    // computing of the internal forces
  DDStateT computeDDF6Ext(const StateR& cur, Environment* env, double time, bool last);
    // compute the total external forces
  DDStateT computeDDFExt(const StateR& cur, Environment* env, double time, bool last);
    // compute the total external forces
  DDStateT computeDDStFExt(const StateR& cur, Environment* env, bool last);
    // compute forces exerted by static obstacles
  DDStateT computeDDForces(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, double time, bool last);// compute the total forces

  /********** Connectivity and Collision checking **********/

  double checkConnectivity(Trajectory* traj, double curTime);
    // check the connectivity of the trajectory
//   virtual void checkNHConnectivity(int i, const StateR& oldCur, const StateR& prev, const StateR& cur, StateR& force);
//     // check the connectivity between two nodes of the trajectory
  void adjustTime(int curStep);
    // adjust the time of the nodes of the trajectory to maintain its consistency
  virtual double checkCollision(STrajectory* traj,Environment* env,double curTime);
    // check if the still-to-be executed part of the trajectory is in collision with the environment
    // return the first time from the current time where a collision occurs. return -1 if there is no collision

  /************** Spatio-temporal deformation **************/
  
  virtual void initDeform(int curStep);                      // init the deformation and check the temporal space between two consecutive states
  virtual int finishDeform(double time);                     // finish the deformation
  virtual int deform(double time, int curStep);              // deform the current trajectory

  /*************** Feedback control methods ****************/

  void initListCtrl(std::list <Vec_DP*>* listCtrl);          // init the list of feedback controls

  /************ Trajectory refinement functions ************/

  DDStateT getStateTraj(double t) const;                      // return the state of the deformed trajectory at time t
  void displayTraj() const;                                  // display the refined trajectory in an openGL viewer
  
};

#endif // STDDDEFORMATION_HPP
