#include "DDDeformation.hpp"

/************** Constructors and destructors *************/

// default constructor
DDDeformation::DDDeformation()
{
  this->listCtrl_ = NULL;
  this->feaCtrl_ = NULL;
  this->parallelExec = false;
  this->nbThread = 5;
}

// init constructor
DDDeformation::DDDeformation(Trajectory *traj,Environment *env,Robot *rb,int** typeS,StateR*** tns,StateR** tfi,StateR** tfe) : STDeformation(traj,env,rb,typeS,tns,tfi,tfe)
{
  this->listCtrl_ = NULL;
  this->feaCtrl_ = NULL;
  this->parallelExec = false;
  this->nbThread = 5;
}

// destructor
DDDeformation::~DDDeformation()
{
}


/*** Computing of the forces applied on the trajectory ***/

// computing of the internal forces
DDStateT DDDeformation::computeDDFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int *fea)
{
//   cout << endl << "###############################################" << endl;

//   cout << "prev : "; prev.display(); cout << endl;
//   cout << "next : "; next.display(); cout << endl;

  DDTrajGen carModel;
  
  Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th(); 
                x0[3] = (prev.omr() + prev.oml())/2.0; x0[4] = (prev.omr() - prev.oml())/(2.0*rb_->getLength()); x0[5] = 0.0;

  Vec_DP xf(6); xf[0] = next.x(); xf[1] = next.y(); xf[2] = next.th(); 
                xf[3] = (next.omr() + next.oml())/2.0; xf[4] = (next.omr() - next.oml())/(2.0*rb_->getLength()); xf[5] = next.t() - prev.t();
  // check the temporal consistency between both nodes
  double dx = next.x() - prev.x();
  double dy = next.y() - prev.y();
  double dt = next.t() - prev.t();

  if(dt < 0 || (rb_->getWVMax())*(rb_->getWVMax())*dt*dt < dx*dx+dy*dy)
    {
      cout << "@dt : " << dt;
      dt = sqrt(dx*dx+dy*dy) / rb_->getWVMax();
      xf[5] = dt;
      cout << "  dx : " << dx << "  dy : " << dy << "  vmax : " << rb_->getWVMax() << "  xf[5] : " << xf[5] << endl;
    }

  Vec_DP constraints(3); 
  constraints[0] = rb_->getWAMax();
  constraints[1] = rb_->getWVMax();
  constraints[2] = rb_->getLength();

//   cout << "iparam : ";
//   if(*param == NULL) cout << "nul" << endl;
//   else cout << "(" << (**param)[0] << "," << (**param)[1] << "," << (**param)[2] << "," << (**param)[3] << "," << 
//     (**param)[4] << "," << (**param)[5] << "," << (**param)[6] << ")" << endl;

  int res = carModel.ComputeTrajectory(x0,xf,constraints,param);

//   cout << "fparam : ";
//   if(*param == NULL) cout << "nul" << endl;
//   else cout << "(" << (**param)[0] << "," << (**param)[1] << "," << (**param)[2] << "," << (**param)[3] << "," << 
//     (**param)[4] << "," << (**param)[5] << "," << (**param)[6] << ")" << endl;
  
  double tSur2 = 0.5*dt;

//   cout << "residual : " << carModel.finalResidual_ << endl;

  // check if the trajectory generator has found a solution
  if(res == 0 || (carModel.finalResidual_ < 2.0 && carModel.finalErrTH_ < 0.5))
    {
      if(res == 0) 
	{
	  cout << "a near solution has been found" << endl;
	  *fea = T_FEA;
	}
      else
	{
	  cout << "a far solution has been found" << endl;
	  *fea = T_FAR;
	}

//       cout << "tSur2 : " << tSur2 << endl;

      cout << "mid : (" << carModel.funcX(tSur2) << "," << carModel.funcY(tSur2) << "," << carModel.funcTheta(tSur2) << 
	"," << carModel.funcWRVel(tSur2) << "," << carModel.funcWLVel(tSur2) << "," << prev.t() + tSur2 << ") : ("
	   << carModel.funcVel(tSur2) << "," << carModel.funcAVel(tSur2) << ")" << endl;

      cout << "final : (" << carModel.funcX(dt) << "," << carModel.funcY(dt) << "," << carModel.funcTheta(dt) << 
	"," << carModel.funcWRVel(dt) << "," << carModel.funcWLVel(dt) << "," << prev.t() + dt << ") : ("
	   << carModel.funcVel(dt) << "," << carModel.funcAVel(dt) << ")" << endl;


//       cout << "fctWRVel : " << endl;
//       carModel.fctWRVel_->display();
//       cout << "fctWLVel : " << endl;
//       carModel.fctWLVel_->display();

//       cout << "carModel.fctWRVel_->getVal(tSur2) : " << carModel.fctWRVel_->getVal(tSur2) << endl;;
//       cout << "carModel.fctWLVel_->getVal(tSur2) : " << carModel.fctWLVel_->getVal(tSur2) << endl;;

      DDStateT res(carModel.funcX(tSur2) - cur.x(),
		   carModel.funcY(tSur2) - cur.y(),
		   carModel.funcTheta(tSur2) - cur.th(),
		   carModel.funcWRVel(tSur2) - cur.omr(),
		   carModel.funcWLVel(tSur2) - cur.oml(),
		   prev.t() + tSur2 - cur.t());
      return res;
    }
  else
    {
      *fea = T_UNFEA;
      
      // destroy the bad parameters vector
      *param = NULL;

      // get back the velocity of both wheels on the previous state
      double prevWLVel = prev.oml();
      double prevWRVel = prev.omr();

      // compute a deceleration control
      double wlacc = 0;
      if(dt != 0) wlacc = -prevWLVel / dt;
      if(wlacc < -0.3*rb_->getWAMax()) wlacc = -0.3*rb_->getWAMax();
      else if(wlacc > 0.3*rb_->getWAMax()) wlacc = 0.3*rb_->getWAMax();
      double wracc = 0;
      if(dt != 0) wracc = -prevWRVel / dt;
      if(wracc < -0.3*rb_->getWAMax()) wracc = -0.3*rb_->getWAMax();
      else if(wracc > 0.3*rb_->getWAMax()) wracc = 0.3*rb_->getWAMax();
      
      cout << "dec control...";
      cout << "  wracc : " << wracc;
      cout << "  wlacc : " << wlacc << endl;

      // compute the state reached by the deceleration control
      //ModelFInt carModel2;
      DDTrajGen ddtrajgen;
      ddtrajgen.setInitialGuess(wracc,0.0,0.0,wlacc,0.0,0.0,dt);

      Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th(); 
      x0[3] = (prev.omr() + prev.oml())/2.0; x0[4] = (prev.omr() - prev.oml())/(2.0*rb_->getLength()); x0[5] = 0.0;
      Vec_DP xf(6); xf[5] = dt;
      Vec_DP constraints(3); 
      constraints[0] = rb_->getWAMax(); 
      constraints[1] = rb_->getWVMax(); 
      constraints[2] = rb_->getLength();
      
      ddtrajgen.setInitialParam(x0,xf,constraints);
      ddtrajgen.ForwardSolution();
      
      const GLfloat redMaterial[]={1.,0.,0.,1.};
      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
      glPointSize(30.0); 
      
      DDStateT res(ddtrajgen.funcX(tSur2) - cur.x(),
		  ddtrajgen.funcY(tSur2) - cur.y(),
		  ddtrajgen.funcTheta(tSur2) - cur.th(),
		  ddtrajgen.funcWRVel(tSur2) - cur.omr(),
		  ddtrajgen.funcWLVel(tSur2) - cur.oml(),
		  prev.t() + tSur2 - cur.t());
      
      cout << "no solution has been found" << endl;
      return res;
    }
}

// computing of the internal forces
void DDDeformation::computeDDFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int *fea, DDStateT& fres)
{
  fres = computeDDFInt(prev,cur,next,last,param,fea);
}

// compute the total external forces
DDStateT DDDeformation::computeDDF6Ext(const StateR& cur, Environment* env, double time, bool last)
{
  // verify if the current state is not too far from the current time to be under external forces
  if(cur.t() - time > (*reliabilityTime))
    {
      DDStateT force;
      if(fabs(force.t()) > 10.0) 
	{ cout << "force0 : "; force.display(); cout << endl; }
      return force;
    }

  // init needed variables
  double Dmax = *spMed;
  double Tmax = *tmpMed;
  double MDmax = Dmax * (*alphaMax);
  double MTmax = Tmax * (*betaMax);
  
  // compute spatial coefficient Alpha (for the metric)
  double Alpha = *alphaMin + ((*reliabilityTime) - cur.t() + time) * ((*alphaMax - *alphaMin) / (*reliabilityTime));
  
  // compute spatial coefficient Alpha (for the metric)
  double Beta = *betaMax - (cur.t() - time) * ((*betaMax - *betaMin) / (*reliabilityTime));

  // search for the nearest obstacle
  int nearestObs = -1;
  double sqMminDist = 0;
  DDStateT nearestState; 
  
  if(this->modeDeform == SPATONLY_DEFORM)
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  // get back the current obstacle
	  ObsDisk2D* curObs = (ObsDisk2D*) this->env_->getObs(i);
	  
	  double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();
	  
	  // get back the last position of the obstacle
	  ConfigR* configObs = NULL;
	  configObs = curObs->getLastCfg();
	  
	  if(configObs != NULL)
	    {
	      Vec3D curRbNearObs(configObs->x() - cur.x(), configObs->y() - cur.y(), 0);

	      double sqMminDistObs = (Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1]);
	      
	      // check if this obstacle is the nearest obstacle from the current robot's control point
	      if(nearestObs == -1 || sqMminDistObs < sqMminDist)
		{
		  sqMminDist = sqMminDistObs;
		  nearestObs = i;
		  nearestState = DDStateT(configObs->x(),configObs->y(),0,0,0,cur.t());
		}
	    }

	  // deletions
	  if(configObs != NULL) delete configObs;
	}
    }
  else
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  // get back the current obstacle
	  ObsDisk2D* curObs = (ObsDisk2D*) this->env_->getObs(i);
	  
	  double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();  
	  //double MDTmax = sqrt(Alpha*Alpha*(Dmax + radObs + rb_->getRadius())*(Dmax + radObs + rb_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  double sqMDTmax = (Alpha*Alpha*(Dmax + radObs + rb_->getRadius())*(Dmax + radObs + rb_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  //cout << "MDTmax : " << MDTmax << endl;
	  
	  // get back the last position of the obstacle
	  ConfigR* configObs = NULL;
	  configObs = curObs->getLastCfg();
	  
	  // compute the coordinates of the obstacle's last control point in the metric frame
	  Vec3D MconfigObs(Alpha * configObs->x(), Alpha * configObs->y(), Beta * curObs->getLastTime());
	  
	  // compute the coordinates of the robot's control point in the metric frame
	  Vec3D MconfigRb(Alpha * cur.x(), Alpha * cur.y(), Beta * cur.t());
	  
	  // compute the robot's speed vector in the metric frame
	  Vec3D MspeedObs(Alpha * curObs->getLastSpdX(), Alpha * curObs->getLastSpdY(), Beta);
	  
	  // compute the vector configRobot - configObs with the metric
	  Vec3D MORi (MconfigRb[0] - MconfigObs[0], MconfigRb[1] - MconfigObs[1], MconfigRb[2] - MconfigObs[2]);
	  
	  // compute the norm of MspeedRb
	  double MnormSObs = sqrt(MspeedObs[0]*MspeedObs[0] + MspeedObs[1]*MspeedObs[1] + MspeedObs[2]*MspeedObs[2]);
	  
	  // compute the normalized vector NMspeedRb
	  Vec3D NMspeedObs;
	  if(MnormSObs != 0)
	    {
	      NMspeedObs[0] = MspeedObs[0]/MnormSObs;
	      NMspeedObs[1] = MspeedObs[1]/MnormSObs;
	      NMspeedObs[2] = MspeedObs[2]/MnormSObs;
	    }
	  
	  // compute the scalar product between MORi and nSObs
	  double pdScal = MORi[0] * NMspeedObs[0] + MORi[1] * NMspeedObs[1] + MORi[2] * NMspeedObs[2];
	  
	  //compute the vector OP (obstacle's nearest control point - obstacle's init control point) in the metric frame
	  Vec3D NOP(pdScal * NMspeedObs[0], pdScal * NMspeedObs[1], pdScal * NMspeedObs[2]);
	  
	  // compute the vector OP in the normal frame
	  Vec3D OP(NOP[0]/Alpha, NOP[1]/Alpha, NOP[2]/Beta);
	  
	  // compute the obstacle's nearest position
	  DDStateT nearestStateCurObs(configObs->x() + OP[0], configObs->y() + OP[1], 0, 0, 0, curObs->getLastTime() + OP[2]);
	  
	  // compute the vector nearestStateCurObs - curRb
	  //Vec3D curRbNearObs(OP[0] - cur.x(), OP[1] - cur.y(), OP[2] - cur.t());
	  Vec3D curRbNearObs(nearestStateCurObs.x() - cur.x(), nearestStateCurObs.y() - cur.y(), nearestStateCurObs.t() - cur.t());

	  double sqMminDistObs = (Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1] + Beta*Beta*curRbNearObs[2]*curRbNearObs[2]);
	  
	  // check if this obstacle is the nearest obstacle from the current robot's control point
	  if((sqMminDistObs < sqMDTmax) && (nearestObs == -1 || sqMminDistObs < sqMminDist))
	    {
	      sqMminDist = sqMminDistObs;
	      nearestObs = i;
	      nearestState = nearestStateCurObs;
	    }
	  
	  // deletions
	  if(configObs != NULL) delete configObs;
	}
    }
  
  // compute the force to apply to the current state
  if(nearestObs != -1)
    {
      // add the nearest state to the corresponding display table entry
      if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
      (*tableNearState)[I] = new StateR5D(nearestState);

      // compute the coefficient due to the reliability time
      double sigma = sqrt(-0.5 * (*reliabilityTime) * (*reliabilityTime) / log(*epsilonRT));
      double coefRT = exp(-0.5 * (time - cur.t())/sigma * (time - cur.t())/sigma );
      //double coefRT = exp((time - cur.t())*(time - cur.t())*log(*epsilonRT)/((*reliabilityTime)*(*reliabilityTime)));

      // compute the direction of the force
      double dx = (cur.x() - nearestState.x());
      double dy = (cur.y() - nearestState.y());
      double dt = (cur.t() - nearestState.t());
      double sqdd = dx*dx + dy*dy;
      double length = sqrt(dx*dx + dy*dy + dt*dt);
      double Ndd = 0;

      DDStateT dir;
      if(length != 0 && !last)
 	{
	  dir = DDStateT(dx/length,dy/length,0,0,0,dt/length);
	  //if(dd != 0) Ndd = dd/length;
 	}
      else if(length != 0 && last)
 	{
 	  //return (searchNearTempFS(cur, env) * coefRT);
 	  dir = DDStateT(0,0,0,0,0,dt/length);
 	}
      else
 	dir = DDStateT();
      
      // check if this state is in collision
      double radObs = ((ObsDisk2D*)(this->env_->getObs(nearestObs)))->getRadius();
      double radRb = rb_->getRadius();
      DDStateT forceNoCol;
      if(!last && sqdd < (radObs + radRb)*(radObs + radRb))
	{
	  double dd = sqrt(sqdd);
	  double ddx = (dx * (radObs +radRb) / dd) - dx;
	  double ddy = (dy * (radObs +radRb) / dd) - dy;
	  //forceNoCol = StateR5D(ddx,ddy,0,0,0);
	  forceNoCol = DDStateT();
	}

      // compute the base external force
      double val = (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));
      DDStateT baseExtForce;
      if(val > 0) baseExtForce = dir * (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));

      if(this->modeDeform == SPATONLY_DEFORM)
	{
	  DDStateT force(baseExtForce.x() + forceNoCol.x(),
			 baseExtForce.y() + forceNoCol.y(),
			 0,0,0,0);
	  force = force * (coefRT);
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force1 : "; force.display(); cout << endl; }
	  return force;
	}
      else if(this->modeDeform == TEMPONLY_DEFORM)
	{
	  DDStateT force(0,
			 0,
			 0,
			 0,
			 0,
			 baseExtForce.t() + forceNoCol.t());
	  force = force * (coefRT);
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force2 : "; force.display(); cout << endl; }
	  return force;
	}
      else
	{
	  // compute the force to apply
	  DDStateT force(baseExtForce.x() + forceNoCol.x(),
			 baseExtForce.y() + forceNoCol.y(),
			 0,
			 0,
			 0,
			 baseExtForce.t() + forceNoCol.t());
	  force = force * (coefRT);
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force3 : "; force.display(); cout << endl; }
	  return force;
	}
    }
  else
    {
      if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
      (*tableNearState)[I] = NULL;
      
      DDStateT force;
      return force;
    }
}

// compute the total external forces
DDStateT DDDeformation::computeDDFExt(const StateR& cur, Environment* env, double time, bool last)
{
//   StateR5D cur5D(cur.x(), cur.y(), cur.t(), 0, 0);
//   StateR5D fExt5D = this->computeFExt(cur5D, env, time, last);
//   return DDStateT(fExt5D.x(), fExt5D.y(), 0, 0, 0, fExt5D.t());

  return this->computeDDF6Ext(cur,env,time,last);
}

// compute forces exerted by static obstacles
DDStateT DDDeformation::computeDDStFExt(const StateR& cur, Environment* env, bool last)
{
  StateR5D cur5D(cur.x(), cur.y(), cur.t(), 0, 0);
  StateR5D fExt5D = this->computeStFExt(cur5D, env, last);
  return DDStateT(fExt5D.x(), fExt5D.y(), 0, 0, 0, fExt5D.t());
}

// compute the total forces
DDStateT DDDeformation::computeDDForces(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, double time, bool last)
{
  DDStateT vRes;
  return vRes;
}


  /********** Connectivity and Collision checking **********/

// check the connectivity of the trajectory
double DDDeformation::checkConnectivity(Trajectory* traj, double curTime)
{
  bool trouve = false;
  int i=0;
  double dt = -1;
  list <Vec_DP*>::iterator it = listCtrl_->begin();
  while(it != listCtrl_->end() && i < traj->getSize() && !trouve)
    {
      // get back the time of the current state
      double time = ((DDStateT*)traj->at(i))->t();

      // test if the current state has not been already visited
      if(time > curTime)
	{
	  // test if the connectivity has been lost on the current state
	  if(*it == NULL)
	    {
	      trouve = true;
	      dt = time - curTime;
	    }
	  else
	    {
	      i++;
	      it++;
	    }
	}
      else
	{
	  i++;
	  it++;
	}
    }
  return dt;
}

// // check the connectivity between two nodes of the trajectory
// void DDDeformation::checkNHConnectivity(int i, const StateR& oldCur, const StateR& prev, const StateR& cur, StateR& force)
// {
// }

// adjust the time of the nodes of the trajectory to maintain its consistency
void DDDeformation::adjustTime(int curStep)
{
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      if(i != 0 && i > curStep)
	{
	  // compute the distance from the previous node
	  DDStateT* prev = (DDStateT*)((*(this->traj_))[i-1]);
	  DDStateT* cur = (DDStateT*)((*(this->traj_))[i]);
	  
	  double dx = cur->x() - prev->x();
	  double dy = cur->y() - prev->y();
	  double dist = sqrt(dx*dx + dy*dy);
	  double dt = cur->t() - prev->t();

	  //double difDir = computeMinDiffAngle(normalizedAngle(prev->th()),normalizedAngle(cur->th()));

	  //double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
	  double coefDist = 1.0;

	  double cdist = coefDist*dist;

	  double vprev = 0.5*(prev->omr() + prev->omr());
	  double vcur = 0.5*(cur->omr() + cur->omr());

	  double vmoy = 0.5*(vprev + vcur);

	  double newDT = cdist / vmoy;

	  if(dt < 0.5 * newDT)
	    cur->setT(prev->t() + 0.5*newDT);
	}
    }
}

// check if the still-to-be executed part of the trajectory is in collision with the environment
// return the first time from the current time where a collision occurs. return -1 if there is no collision
double DDDeformation::checkCollision(STrajectory* traj,Environment* env,double curTime)
{
}


/************** Spatio-temporal deformation **************/

// init the deformation and check the temporal space between two consecutive states
void DDDeformation::initDeform(int curStep)
{
  double maxTime = 2*this->timeStep_;
  double maxDist = 4*this->posStep_;
  double maxDir  = 0.2*M_PI;

  // check if the list of feedback controls must be initialized
  if(listCtrl_ == NULL)
    {
      listCtrl_ = new list <Vec_DP*>(traj_->getSize() - 1);
      list <Vec_DP*>::iterator it = listCtrl_->begin();
      for(it = listCtrl_->begin(); it != listCtrl_->end(); it++)
	*it = NULL;

      feaCtrl_ = new list <int>(traj_->getSize() - 1);
      list <int>::iterator it2 = feaCtrl_->begin();
      for(it2 = feaCtrl_->begin(); it2 != feaCtrl_->end(); it2++)
	*it2 = 0;
    }

  // clear the tables of forces, states types and nearest states
  if(*typeState != NULL) delete [] (*typeState);
  if(*tableNearState != NULL)delete [] (*tableNearState);
  if(*tableFInt != NULL)delete [] (*tableFInt);
  if(*tableFExt != NULL)delete [] (*tableFExt);
  
//   // check if the space between two consecutive states of the trajectory is too large
//   // and if the orientation difference between them is not too great too
//   std::list <Vec_DP*>::iterator it =  listCtrl_->begin();
//   for(int j=0; j<curStep; j++) it++;

//   //cout << "listCtrl_->size() : " << listCtrl_->size() << "traj_->getSize() : " << traj_->getSize() << endl;

//   int i=curStep;
//   while(i < this->traj_->getSize() - 1 && it != listCtrl_->end())
//     {
//       // get back the current and the next states
//       StateR* curS  = (*this->traj_)[i];
//       StateR* nextS = (*this->traj_)[i+1];

//       double dx = nextS->x() - curS->x();
//       double dy = nextS->y() - curS->y();
//       double sqDist = dx*dx + dy*dy;
      
//       double dTH = nextS->th() - curS->th();

//       // check if the spatial or temporal space is too large
//       if(nextS->t() - curS->t() > maxTime || sqDist > maxDist)
// 	{
// 	  // get back the current control parameters
// 	  Vec_DP* curParam = *it;

// 	  if(curParam != NULL)
// 	    {
// 	      // compute the state reached by the current control
// 	      DDTrajGen carModel2;
// 	      carModel2.setInitialGuess((*curParam)[0],(*curParam)[1],(*curParam)[2],(*curParam)[3],
// 					(*curParam)[4],(*curParam)[5],0.5*(*curParam)[6]);
// 	      Vec_DP x0(6); x0[0] = curS->x(); x0[1] = curS->y(); x0[2] = curS->th(); 
// 	                                      x0[3] = (curS->omr() + curS->oml())/2.0; x0[4] = (curS->omr() - curS->oml())/(2.0*rb_->getLength()); x0[5] = 0.0;
// 	      Vec_DP xf(6); xf[0] = curS->x(); xf[1] = curS->y(); xf[2] = curS->th(); 
// 	                                      xf[3] = (curS->omr() + curS->oml())/2.0; xf[4] = (curS->omr() - curS->oml())/(2.0*rb_->getLength()); xf[5] = 0.0;
// 	      Vec_DP constraints(3); 
// 	      constraints[0] = rb_->getWAMax();
// 	      constraints[1] = rb_->getWVMax();
// 	      constraints[2] = rb_->getLength();

// 	      carModel2.setInitialParam(x0,xf,constraints);
// 	      carModel2.ForwardSolution();
	      
// 	      double x = carModel2.vecState_xsim[0];
// 	      double y = carModel2.vecState_xsim[1];
// 	      double th = carModel2.vecState_xsim[2];
// 	      double v = carModel2.vecState_xsim[3];
// 	      double om = carModel2.vecState_xsim[4];
// 	      double omr = v + rb_->getLength() * om;
// 	      double oml = v - rb_->getLength() * om;
// 	      double t = curS->t() + 0.5*(*curParam)[6];
	      
// 	      StateR* itdS = new StateS6D(x,y,th,omr,oml,t);
// 	      // insert it in the trajectory
// 	      this->traj_->insert(i+1, itdS);
	      
// 	      // change the time of the control from the current state
// 	      (*curParam)[6] *= 0.5;
	      
// 	      // insert the same control for the new state
// 	      Vec_DP* newParam = new Vec_DP(*curParam);
// 	      listCtrl_->insert(it,newParam);
// 	    }
// 	}

//       i++;
//       it++;
//     }

  // reinit the tables of states types and nearest states
  *typeState = new int [this->traj_->getSize()];
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      (*typeState)[i] = -1;
    }
  *tableNearState = (StateR**)(new DDStateT* [this->traj_->getSize()]);
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      (*tableNearState)[i] = NULL;
    }
  
  *tableFInt = new DDStateT [this->traj_->getSize()];
  
  *tableFExt = new DDStateT [this->traj_->getSize()];
}

// finish the deformation
int DDDeformation::finishDeform(double time)
{
  int res = 0;

  // check if a collision may occur during the prediction horizon
  double dt = this->checkCollision(this->traj_, this->env_, time);
  if(dt != -1 && dt < 0.3*(*predHorizon_))
    {
      res = -2;
    }
  else
    {
      // check the connectivity of the trajectory
      double dt2 = this->checkConnectivity(this->traj_, time);
      if(dt != -1 && dt < 0.2*(*predHorizon_))
	{
	  res = -3;
	}
    }
  
  this->oldTime_ = time;

  return res;
}

// deform the current trajectory
int DDDeformation::deform(double time, int curStep)
{  
//   cout << "DDDeformation::deform" << endl;

  // init the deformation
  this->initDeform(curStep);
  
  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      return -1;
    }
  else
    {
      // save the old trajectory
      DDTrajectory oldTraj(*((DDTrajectory*)(this->traj_)));
      
      // apply the repulsive forces from the obstacles
      for(int i=1; i<oldTraj.getSize(); i++)
	{
	  I = i;

	  if(i > curStep)
	    {
	      if((oldTraj)[i]->t() < time + (*reliabilityTime))
		{
		  bool last;
		  if(i == oldTraj.getSize()-1) last = true;
		  else last = false;

		  // compute the external forces due to dynamic obstacles
		  DDStateT fExtDyn = this->computeDDFExt(*((oldTraj)[i]), this->env_, time, last);

		  // compute the external forces due to static obstacles
		  DDStateT fExtSt = this->computeDDStFExt(*((oldTraj)[i]), this->env_, last);		  

		  // compute the total external forces
		  DDStateT forces;
		  forces = (fExtDyn*(*K_EXT)) + (fExtSt*(*K_STEXT));
		  forces = forces * (lastDeformDuration_ *2.5);

		  (*tableFExt)[i] = forces;

		  // check the connectivity of the curve
		  //this->checkConnectivity(i, *(oldTraj[i]), *((*(this->traj_))[i-1]), *((*(this->traj_))[i]), forces);

		  // apply the forces
		  ((*(this->traj_))[i])->setX( ((oldTraj)[i])->x() + (forces.x()) );
		  ((*(this->traj_))[i])->setY( ((oldTraj)[i])->y() + (forces.y()) );
		  ((*(this->traj_))[i])->setTH( ((oldTraj)[i])->th() + (forces.th()) );
 		  ((*(this->traj_))[i])->setOMR( ((oldTraj)[i])->omr() + (forces.omr()) );
 		  ((*(this->traj_))[i])->setOML( ((oldTraj)[i])->oml() + (forces.oml()) );
		  ((*(this->traj_))[i])->setT( ((oldTraj)[i])->t() + (forces.t()) );

		}
	    }
	}

      // adjust the time of each node
      this->adjustTime(curStep);
      
      // apply the internal forces
      if(parallelExec)
	{
// 	  // save the intermediate trajectory
// 	  Trajectory6D itdTraj(*((Trajectory6D*)(this->traj_)));

// 	  list <Vec_DP*>::iterator it = listCtrl_->begin();
	  
// 	  int i=1;
// 	  while(i<itdTraj.getSize()-1)
// 	    {
// 	      int nbThreadUsed = 0;
// 	      boost::thread (**tabThread);
// 	      tabThread = new boost::thread* [nbThread];
	      
// 	      StateS6D* tabRes = new StateS6D[nbThread];
	      
// 	      int lastI = -1;
	      
// 	      cout << "search threads to create from " << i << endl;
	      
// 	      while(i<itdTraj.getSize()-1 && nbThreadUsed != this->nbThread)
// 		{
// 		  it++;
		  
// 		  if(i > curStep)
// 		    {
// 		      if(itdTraj[i]->t() < time + (*reliabilityTime))
// 			{
// 			  lastI = i;
			  
// 			  bool last;
// 			  if(i == itdTraj.getSize()-1) last = true;
// 			  else last = false;
			  
// 			  // create the thread allowing to compute the trajectory generation process
// 			  boost::thread* th = new boost::thread(boost::bind(&STNHDeformation::computeNHFInt,this,
// 										  boost::cref(*(itdTraj[i-1])), 
// 										  boost::cref(*(itdTraj[i])), 
// 										  boost::cref(*(itdTraj[i+1])), 
// 										  last, &(*it), 
// 										  boost::ref(tabRes[nbThreadUsed])) );
// 			  cout << "thread created" << endl;
// 			  tabThread[nbThreadUsed] = th;
// 			  //StateS6D forces = this->computeNHFInt(*(itdTraj[i-1]), *(itdTraj[i]), *(itdTraj[i+1]), last, &(*it));
// 			  //cout << "thread put in the table" << endl;
// 			  //(tabThread[nbThreadUsed])->join();

// 			  //cout << "force : "; (tabRes[nbThreadUsed]).display(); cout << endl;

// 			  nbThreadUsed++;
// 			}
// 		    }
		  
// 		  i++;
// 		}

// 	      cout << "nbThreadUsed : " << nbThreadUsed << " at node : " << lastI << endl; 

// 	      // get back the results of parallel computations and apply the forces on the trajectory
// 	      for(int j=0; j<nbThreadUsed; j++)
// 		{
// 		  (tabThread[j])->join();

// 		  int ind = lastI - nbThreadUsed + j + 1;

// 		  // get back the force computed
// 		  StateS6D forces = tabRes[j];

// 		  // process that allows to reduce the force applied on the last states of the trajectory
// 		  if(ind > traj_->getSize() - 11) forces = forces*(0.1*(traj_->getSize()-i));
		  
// 		  cout << "ind : " << ind << "  "; forces.display(); cout << endl;

// 		  // update the table of internal forces
// 		  (*tableFInt)[ind] = forces;
		  
// 		  // compute the coefficient due to the reliability time
// 		  double coefRT = exp(((time - (itdTraj[i])->t())*(time - (itdTraj[i])->t())*log(*epsilonRT))/((*reliabilityTime)*(*reliabilityTime)));
		  
// 		  ((*(this->traj_))[ind])->setX( (itdTraj[ind])->x() + coefRT*(forces.x()) );
// 		  ((*(this->traj_))[ind])->setY( (itdTraj[ind])->y() + coefRT*(forces.y()) );
// 		  ((*(this->traj_))[ind])->setTH( (itdTraj[ind])->th() + coefRT*(forces.th()) );
// 		  ((*(this->traj_))[ind])->setS( (itdTraj[ind])->s() + coefRT*(forces.s()) );
// 		  ((*(this->traj_))[ind])->setPH( (itdTraj[ind])->ph() + coefRT*(forces.ph()) );
// 		  ((*(this->traj_))[ind])->setT( (itdTraj[ind])->t() + coefRT*(forces.t()) );
// 		}
	      
// 	      cout << "Forces applied" << endl;

// 	      // clear the allocated tables
// 	      delete tabThread;
// 	      cout << "tabThread deleted" << endl;
// 	      //delete tabRes;

// 	      cout << "tabRes deleted" << endl;
// 	    }
	}

      else
	{
	  list <Vec_DP*>::iterator it = listCtrl_->begin();
	  list <int>::iterator it2 = feaCtrl_->begin();
	  
	  for(int i=1; i<this->traj_->getSize()-1; i++)
	    {
	      it++;
	      it2++;

	      if(i > curStep)
		{
		  if((*traj_)[i]->t() < time + (*reliabilityTime))
		    {		  
		      bool last;
		      if(i == oldTraj.getSize()-1) last = true;
		      else last = false;
		      
		      DDStateT forces = this->computeDDFInt(*((*(this->traj_))[i-1]), *((*(this->traj_))[i]), *((*(this->traj_))[i+1]), last, &(*it), &(*it2));
// 		      if(i > traj_->getSize() - 11) forces = forces*(0.1*(traj_->getSize()-i));
		      
		      // update the table of internal forces
		      (*tableFInt)[i] = forces;
		      
		      // compute the coefficient due to the reliability time
		      double coefRT = exp(((time - ((*(this->traj_))[i])->t())*(time - ((*(this->traj_))[i])->t())*log(*epsilonRT))/((*reliabilityTime)*(*reliabilityTime)));
		      
		      ((*(this->traj_))[i])->setX( ((*(this->traj_))[i])->x() + coefRT*(forces.x()) );
		      ((*(this->traj_))[i])->setY( ((*(this->traj_))[i])->y() + coefRT*(forces.y()) );
		      ((*(this->traj_))[i])->setTH( ((*(this->traj_))[i])->th() + coefRT*(forces.th()) );
		      ((*(this->traj_))[i])->setOMR( ((*(this->traj_))[i])->omr() + coefRT*(forces.omr()) );
		      ((*(this->traj_))[i])->setOML( ((*(this->traj_))[i])->oml() + coefRT*(forces.oml()) );
		      ((*(this->traj_))[i])->setT( ((*(this->traj_))[i])->t() + coefRT*(forces.t()) );
		    }
		}
	    }
	}

    }

//   cout << "DDDeformation::deformation finished" << endl;

  // finish the deformation
  return this->finishDeform(time);
}


/*************** Feedback control methods ****************/

// init the list of feedback controls
void DDDeformation::initListCtrl(std::list <Vec_DP*>* listCtrl)
{
  this->listCtrl_ = listCtrl;
}


/************ Trajectory refinement functions ************/

// return the state of the deformed trajectory at time t
DDStateT DDDeformation::getStateTraj(double t) const
{
  // check if the time is defined over the trajectory
  if(t<0)
    {
      cout << "DDDeformation::getStateTraj : Warning. TIME UNDIFINED!" << endl;
      return DDStateT();
    }
  else if(t>this->traj_->at(traj_->getSize()-1)->t())
    {
      return *((DDStateT*)(this->traj_->at(traj_->getSize()-1)));
    }

  bool trouve = false;
  int i=1; 
  while(!trouve && i < this->traj_->getSize())
    {
      if(this->traj_->at(i)->t() > t)
	{
	  trouve = true;
	  i--;
	}
      else
	{
	  i++;
	}
    }

//   cout << "traj_->getSize() : " << traj_->getSize() << endl;
//   cout << "i found: " << i << endl;

  // get back the corresponding state
  DDStateT* cur = (DDStateT*) (this->traj_->at(i));
  DDStateT* next = (DDStateT*) (this->traj_->at(i+1));

//   // get back the control corresponding to the node
//   list <Vec_DP*>::iterator it = listCtrl_->begin();
//   for(int j=0; j<i; j++) it++;

//   if(*it == NULL)
//     {
      // no control is defined here
      double coef = (t - cur->t()) / (next->t() - cur->t());
      return DDStateT(cur->x()*(1.0-coef)   + next->x()*coef,
		     cur->y()*(1.0-coef)   + next->y()*coef,
		     cur->th()*(1.0-coef)  + next->th()*coef,
		     cur->omr()*(1.0-coef) + next->omr()*coef,
		     cur->oml()*(1.0-coef) + next->oml()*coef,
		     cur->t()*(1.0-coef)   + next->t()*coef);
//     }
//   else
//     {
//       // initialize the forward method
//       Vec_DP x0(6); x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
//       x0[3] = (cur->omr() + cur->oml())/2.0; x0[4] = (cur->omr() - cur->oml())/(2.0*rb_->getLength()); x0[5] = 0.0;
      
//       Vec_DP xf(6); xf[0] = next->x(); xf[1] = next->y(); xf[2] = next->th(); 
//       x0[3] = (next->omr() + next->oml())/2.0; x0[4] = (next->omr() - next->oml())/(2.0*rb_->getLength()); xf[5] = next->t() - cur->t();
      
//       Vec_DP constraints(3); 
//       constraints[0] = rb_->getWAMax();
//       constraints[1] = rb_->getWVMax();
//       constraints[2] = rb_->getLength();
      
//       DDTrajGen carModel;
//       carModel.setInitialGuess((**it)[0],(**it)[1],(**it)[2],(**it)[3],(**it)[4],(**it)[5],(**it)[6]);
//       carModel.setInitialParam(x0,xf,constraints);
//       carModel.ForwardSolution();
      
//       double x = carModel.vecState_xsim[0];
//       double y = carModel.vecState_xsim[1];
//       double th = carModel.vecState_xsim[2];
//       double v = carModel.vecState_xsim[3];
//       double om = carModel.vecState_xsim[4]; cout << "v : " << v << "  om : " << om << endl;
//       double omr = carModel.funcWRVel(next->t() - cur->t()); //v + rb_->getLength() * om;
//       double oml = carModel.funcWLVel(next->t() - cur->t()); //v - rb_->getLength() * om;
      
//       return DDStateT(x,y,th,omr,oml,t);
//     }
}

// display the refined trajectory in an openGL viewer
void DDDeformation::displayTraj() const
{
  const GLfloat blueMaterial[]={0.0,0.0,1.0,1.};
  const GLfloat greenMaterial[]={0.0,1.0,0.0,1.};
  const GLfloat redMaterial[]={1.0,0.0,0.0,1.};
  const GLfloat blackMaterial[]={0.0,0.0,0.0,1.};
  const GLfloat yellowMaterial[]={1.0,1.0,0.0,1.};

  glLineWidth(3.0);

  for(int i=0; i<traj_->getSize()-2; i++)
    {
      // get back the corresponding state
      DDStateT* cur = (DDStateT*) (this->traj_->at(i));
      DDStateT* next = (DDStateT*) (this->traj_->at(i+1));
      DDStateT* nnext = NULL;
      if(i<traj_->getSize()-2) nnext = (DDStateT*) (this->traj_->at(i+2));
      else nnext = (DDStateT*) (this->traj_->at(i+1));
      
      // get back the control corresponding to the node
      list <Vec_DP*>::iterator it = listCtrl_->begin();
      for(int j=0; j<=i; j++) it++;
      list <int>::iterator it2 = feaCtrl_->begin();
      for(int j=0; j<=i && j != traj_->getSize()-1; j++) it2++;

      // select the color of the trajectory to display
      switch(*it2)
	{
	case T_FEA : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  break;
	case T_FAR : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
	  break;
	case T_UNFEA : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  break;
	default : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);
	  break;
	}
      
      glPushMatrix();
      glTranslated(cur->x(),
		   cur->y(),
		   cur->t());
      gluSphere(gluNewQuadric(), 1.0, 5, 5);
      glPopMatrix();
      
      glBegin(GL_LINE_STRIP);

      // define the sampling of subtrajectories
      int nb = 20;

      if(*it == NULL)
	{
	  // no control is defined here
	  for(int k=0; k<=nb; k++)
	    {
	      double t = cur->t() + (((double)k)/((double)nb))*(nnext->t() - cur->t());
	      double coef = (t - cur->t()) / (nnext->t() - cur->t());
	      
	      double x = cur->x()*(1.0-coef)   + nnext->x()*coef;
	      double y = cur->y()*(1.0-coef)   + nnext->y()*coef;
	      
	      glVertex3f(x,y,t);
	  
	      if(k == nb) {
		glEnd();
// 		cout << "trajCtrl[" << i << "] : (" << x << "," << y << ",,,," << t << ")" << endl;
	      }
	    }
	}
      else
	{
	  // a control is defined here
	  double x0[6]; x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
	  x0[3] = (cur->omr() + cur->oml())/2.0; x0[4] = (cur->omr() - cur->oml())/(2.0*rb_->getLength()); x0[5] = 0.0;
	      
	  double constraints[3];
	  constraints[0] = rb_->getWAMax();
	  constraints[1] = rb_->getWVMax();
	  constraints[2] = rb_->getLength();
	  
	  double parameters[7];
	  parameters[0] = (**it)[0];
	  parameters[1] = (**it)[1];
	  parameters[2] = (**it)[2];
	  parameters[3] = (**it)[3];
	  parameters[4] = (**it)[4];
	  parameters[5] = (**it)[5];
	  parameters[6] = (**it)[6];
	  
	  DDTrajGen trajgen;
	  trajgen.ComputeFromIG(x0,constraints,parameters);
	  
	  for(int k=0; k<=nb; k++)
	    {
	      double curDT = (((double)k)/((double)nb))*(nnext->t() - cur->t());
	      double t = cur->t() + curDT;
	      double rstatet[6];

	      trajgen.getStateTReachedAt(rstatet,curDT);

	      double x = rstatet[0];
	      double y = rstatet[1];
	      double th = rstatet[2];
	      double wrvel = rstatet[3];
	      double wlvel = rstatet[4];

	      glVertex3f(x,y,t);
	      
	      if(k == nb) 
		{
		  glEnd();

// 		  cout << i << " :: ";
// 		  cout << "cur : "; cur->display(); cout << " :: ";
//  		  cout << "ctrl : (" << (**it)[0] << "," << (**it)[1] << "," << (**it)[2] << "," << (**it)[3] 
//  		       << "," << (**it)[4] << "," << (**it)[5] << "," << (**it)[6] << ") -> ";
//  		  cout << "sreached : (" << x << "," << y << "," << th << "," << wrvel << ","<< wlvel << "," << t << ")" << endl;

		  if(i%10 == 0)
		    {
		      glMaterialfv(GL_FRONT, GL_AMBIENT, yellowMaterial);
		      glMaterialfv(GL_FRONT, GL_DIFFUSE, yellowMaterial);
		    }
		  else
		    {
		      glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
		      glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);
		    }
		  glPushMatrix();
		  glTranslated(x,y,t);
		  gluSphere(gluNewQuadric(), 1.0, 5, 5);
		  glPopMatrix();
		}
	    }
	}
    }

//       int nb = 50;
//       for(int k=0; k<nb; k++)
// 	{
// 	  double t = cur->t() + (((double)k)/((double)nb))*(nnext->t() - cur->t());
// 	  //cout << "  t : " << t;

// 	  if(*it == NULL)
// 	    {
// 	      // no control is defined here
// 	      for(int k=0; k<=nb; k++)
// 		{
// 		  double t = cur->t() + (((double)k)/((double)nb))*(nnext->t() - cur->t());
// 		  double coef = (t - cur->t()) / (nnext->t() - cur->t());
		  
// 		  double x = cur->x()*(1.0-coef)   + nnext->x()*coef;
// 		  double y = cur->y()*(1.0-coef)   + nnext->y()*coef;
		  
// 		  glVertex3f(x,y,t);
		  
// 		  if(k == nb) {
// 		    glEnd();
// 		  }
// 		}
// 	    }
// 	  else
// 	    {
// 	      // initialize the forward method
// 	      Vec_DP x0(6); x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
// 	      x0[3] = (cur->omr() + cur->oml())/2.0; x0[4] = (cur->omr() - cur->oml())/(2.0*rb_->getLength()); x0[5] = 0.0;
	      
// 	      Vec_DP xf(6); xf[0] = nnext->x(); xf[1] = nnext->y(); xf[2] = nnext->th(); 
// 	      x0[3] = (nnext->omr() + nnext->oml())/2.0; x0[4] = (nnext->omr() - nnext->oml())/(2.0*rb_->getLength()); xf[5] = nnext->t() - cur->t();
	      
// 	      Vec_DP constraints(3);
// 	      constraints[0] = rb_->getWAMax();
// 	      constraints[1] = rb_->getWVMax();
// 	      constraints[2] = rb_->getLength();
	      
// 	      DDTrajGen carModel;
// 	      carModel.setInitialGuess((**it)[0],(**it)[1],(**it)[2],(**it)[3],(**it)[4],(**it)[5],(**it)[6]);
// 	      carModel.setInitialParam(x0,xf,constraints);
// 	      carModel.ForwardSolution();
	      
// 	      double x = carModel.funcX(t - cur->t()); //carModel.vecState_xsim[0];
// 	      double y = carModel.funcY(t - cur->t()); //carModel.vecState_xsim[1];
	      
// 	      glVertex3f(x,y,t);
// 	    }
// 	}
//       glEnd();
//     }
}
