#include "STDeformation.hpp"
#include "ReachabilityCone2D.hpp"

#include "gnuplot_i.hpp"

                             /*********************************************************/
                             /************** Constructors and destructors *************/
                             /*********************************************************/

// default constructor
STDeformation::STDeformation()
{
}

// init constructor
STDeformation::STDeformation(STrajectory *traj, Environment *env, Robot *rb, int** typeS)
{
  this->traj_      = traj;
  this->env_       = env;
  this->rb_        = rb;
  this->typeState  = typeS;
  this->oldTime_   = 0;
  this->timeStep_  = 0.5;
  this->posStep_   = 1.5;
  this->modeDeform = SPATIOTEMP_DEFORM;
  this->lastDeformDuration_ = 0.0;
}

// init constructor
STDeformation::STDeformation(STrajectory *traj,Environment *env,Robot *rb,int** typeS, StateR*** tns,StateR** tfi,StateR** tfe)
{
  this->traj_          = traj;
  this->env_           = env;
  this->rb_            = rb;
  this->typeState      = typeS;
  this->tableNearState = (StateR***)tns;
  this->tableFInt      = (StateR**)tfi;
  this->tableFExt      = (StateR**)tfe;
  this->oldTime_       = 0;
  this->timeStep_      = 0.5;
  this->posStep_       = 1.5;
  this->modeDeform     = SPATIOTEMP_DEFORM;
  this->lastDeformDuration_ = 0.0;
}

// destructor
STDeformation::~STDeformation()
{
}


                             /*********************************************************/
                             /*********************** Accessors ***********************/
                             /*********************************************************/

// return the current trajectory
STrajectory* STDeformation::getTraj() const
{
  return this->traj_;
}

// set the current trajectory
void STDeformation::setTraj(STrajectory* traj)
{
  this->traj_ = traj;
}

// return the current environment
Environment* STDeformation::getEnv() const
{
  return this->env_;
}

// set the current environment
void STDeformation::setEnv(Environment* env)
{
  this->env_ = env;
}

// return the current robot
Robot* STDeformation::getRobot() const
{
  return this->rb_;
}

// set the current robot
void STDeformation::setRobot(Robot* rb)
{
  this->rb_ = rb;
}

// set the deformation mode
void STDeformation::setModeDeform(modeDeform_t mode)
{
  this->modeDeform = mode;
}

// set the coefficients of internal and external forces and the balance between external spatial and temporal forces
void STDeformation::setForcesCoef(double* kint, double* kext, double *kstext, double* bal)
{
  this->K_INT     = kint;
  this->K_EXT     = kext;
  this->K_STEXT   = kstext;
  this->K_SPT_TMP = bal;
}

// set the coefficients of internal and external forces and the balance between external spatial and temporal forces
void STDeformation::setForcesCoef(double* kint, double* kext, double *kstext, double* amin, double* amax, double* bmin, double* bmax)
{
  this->K_INT     = kint;
  this->K_EXT     = kext;
  this->K_STEXT   = kstext;
  this->alphaMin  = amin;
  this->alphaMax  = amax;
  this->betaMin   = bmin;
  this->betaMax   = bmax;
}

// // set the distance from the robot until which forces take effects (dynamic obstacles)
// void STDeformation::setMaxEffectDist(double* mad)
// {
//   this->maxEffectDist = mad;
// }

// // set the distance from the robot until which forces take effects (static obstacles)
// void STDeformation::setMaxStEffectDist(double* mad)
// {
//   this->maxStEffectDist = mad;
// }

// set the spatial and temporal distances from the robot until which forces take effects
void STDeformation::setMaxEffectDist(double* spMed,double* tmpMed, double* staticMed)
{
  this->spMed           = spMed;
  this->tmpMed          = tmpMed;
  this->maxStEffectDist = staticMed;
}

// set the reliability time and the associated epsilon value
void STDeformation::setReliability(double* time, double* eps)
{
  this->reliabilityTime = time;
  this->epsilonRT = eps;
}

// set the prediction horizon, the reliability time and the associated epsilon value
void STDeformation::setReliability(double* predH, double* time, double* eps)
{
  this->predHorizon_ = predH;
  this->reliabilityTime = time;
  this->epsilonRT = eps;
}

// set if the temporal consistency is used or not
void STDeformation::setTemporalConst(bool* b)
{
  this->useTemporalConsistency = b;
}

// set the time step
void STDeformation::setTimeStep(double ts)
{
  this->timeStep_ = ts;
}

// set the position step
void STDeformation::setPosStep(double ps)
{
  this->posStep_ = ps;
}


                             /*********************************************************/
                             /*** Computing of the forces applied on the trajectory ***/
                             /*********************************************************/

// computing of the internal forces
StateR5D STDeformation::computeFInt(const StateR& prev, const StateR& cur, const StateR& next)
{
  //ReachabilityCone2D rc2D(this->rb_);
  //State5D s5Dprev(prev);
  //State5D s5Dnext(next);
  //rc2D.computeFromToV2(s5Dprev,s5Dnext);
  //State5D sBary2 = rc2D.barycentre();
  //State5D sBary2 = rc2D.baryReachableDisk(s5Dprev, s5Dnext, (prev.t() + next.t()) / 2);

  State5D sBary((prev.x() + next.x())/2, (prev.y() + next.y())/2, (prev.t() + next.t())/2, (prev.sx() + next.sx())/2, (prev.sy() + next.sy())/2);

  StateR5D sRes = StateR5D(sBary - cur) * (*K_INT);
  //StateR5D sRes = StateR5D(sBary2 - cur) * K_INT;
  return sRes;
}

StateR5D STDeformation::computeF1Ext(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, bool last)
{
  double TTC0 = 5;
  double k_t = 0.5;
  double k_s = 0.8;

  double robotRadius = this->rb_->getRadius();
  double maxSpd = this->rb_->getVMax();

  if(env != NULL)
    {
      StateR5D sRes;

      double TTCmin = 100000.0;
      int indMin = -1;
      double dIM1min = 100000.0;
      double dIP1min = 100000.0;

      // searching of the lesser time to collision
      for(int i=0; i<env->getNbObs(); i++)
	{
 
	  // get back the radius of the obstacle
	  double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();

	  // get back the i-1 th future configuration of the obstacle
	  ConfigR* cfgObsIM1 = env->getObs(i)->getFutureCfg(prev.t());

	  // get back the i-1 th future configuration of the obstacle
	  ConfigR* cfgObsI = env->getObs(i)->getFutureCfg(cur.t());

	  // get back the i-1 th future configuration of the obstacle
	  ConfigR* cfgObsIP1 = env->getObs(i)->getFutureCfg(next.t());

	  //verify that all computed configurations are not null
	  if(cfgObsIM1 != NULL && cfgObsI != NULL && cfgObsIP1 != NULL)
	    {

	      // computing of the distance from the previous state to the obstacle
	      double dxIM1 = cfgObsIM1->x() - prev.x();
	      double dyIM1 = cfgObsIM1->y() - prev.y();
	      double dIM1 = sqrt(dxIM1*dxIM1 + dyIM1*dyIM1) - radObs - robotRadius;
	      
	      // computing of the distance from the current state to the obstacle
	      double dxI = cfgObsI->x() - cur.x();
	      double dyI = cfgObsI->y() - cur.y();
	      double dI = sqrt(dxI*dxI + dyI*dyI) - radObs - robotRadius;
	      
	      // computing of the distance from the previous state to the obstacle
	      double dxIP1 = cfgObsIP1->x() - next.x();
	      double dyIP1 = cfgObsIP1->y() - next.y();
	      double dIP1 = sqrt(dxIP1*dxIP1 + dyIP1*dyIP1) - radObs - robotRadius;
	      
	      // verify that dIM1 and dIP1 are not equals
	      if(dIM1 != dIP1)
		{
		  // compute the speed of collision
		  double spdCol = (dIP1 - dIM1)/(next.t() - prev.t());
		  
		  // compute the time-to-collision
		  double TTC = dI/spdCol;
		  
		  // check if the TTC is minimal
		  if(fabs(TTC) < fabs(TTCmin))
		    {
		      TTCmin = TTC;
		      indMin = i;
		      dIM1min = dIM1;
		      dIP1min = dIP1;
		    }
		}
	    }
	      
	  // delete the future configurations
	  if(cfgObsIM1 != NULL) delete cfgObsIM1;
	  if(cfgObsI != NULL) delete cfgObsI;
	  if(cfgObsIP1 != NULL) delete cfgObsIP1;
	}

      // treatment of the last state
      if(last)
	{
	  if(fabs(TTCmin) < 0.5*fabs(TTC0))
	    {
	      double F_t = 0.5;
	      sRes.setT(F_t);
	      return sRes;
	    }
	}

      // check if the minimal TTC is lesser than TTC0
      if(fabs(TTCmin) < fabs(TTC0))
	{
	  // compute the force to apply
	  double F_t;
	  if(dIP1min > dIM1min)
	    {
	      // compute the speed coeficient
	      double coefSpdX;
	      if(cur.sx() > k_s * maxSpd)
		coefSpdX = 1.0 - (cur.sx() - k_s*maxSpd)/((1-k_s)*maxSpd);
	      else
		coefSpdX = 1.0;

	      //if(coefSpdX < k_t-1) coefSpdX = k_t-1;
	      if(coefSpdX < 0) coefSpdX = 0;
	      
	      double coefSpdY;
	      if(cur.sy() > k_s * maxSpd)
		coefSpdY = 1.0 - (cur.sy() - k_s*maxSpd)/((1-k_s)*maxSpd);
	      else
		coefSpdY = 1.0;

	      //if(coefSpdY < 0.5*k_t-1) coefSpdY = 0.5*k_t-1;
	      if(coefSpdY < 0) coefSpdY = 0;

	      F_t = (TTC0 - fabs(TTCmin))/TTC0 * k_t * (prev.t() - cur.t()) * coefSpdX * coefSpdY;
	    }
	  else
	    {
	      // compute the speed coeficient
// 	      double coefSpdX;
// 	      if(cur.sx() < k_s * -maxSpd)
// 		coefSpdX = 1.0 - (cur.sx() - k_s*maxSpd)/((1-k_s)*maxSpd);
// 	      else
// 		coefSpdX = 1.0;

// 	      if(coefSpdX < k_t-1) coefSpdX = k_t-1;

// 	      double coefSpdY;
// 	      if(cur.sy() < k_s * -maxSpd)
// 		coefSpdY = 1.0 - (cur.sy() - k_s*maxSpd)/((1-k_s)*maxSpd);
// 	      else
// 		coefSpdY = 1.0;

// 	      if(coefSpdY < k_t-1) coefSpdY = k_t-1;

	      F_t = (TTC0 - fabs(TTCmin))/TTC0 * k_t * (next.t() - cur.t()); // * coefSpdX * coefSpdY;
	    }

	  sRes.setT(F_t);
	}

      return sRes;
    }
  else
    {
      return StateR5D();
    }
}

StateR5D STDeformation::computeF2Ext(const StateR& cur, Environment* env)
{ 
  double dMin = 60.0f;

  double robotRadius = this->rb_->getRadius();
  
  if(env != NULL)
    {
      StateR5D sRes;
      for(int i=0; i<env->getNbObs(); i++)
	{
	  ConfigR* cfgObs = env->getObs(i)->getFutureCfg(cur.t());

	  if(cfgObs != NULL)
	    {
	      StateR5D sCur;
	      ConfigR2D* curCfg = (ConfigR2D*) (cur.getCfg());
	      ConfigR2D vObsState(curCfg->x() - cfgObs->x(), curCfg->y() - cfgObs->y());
	      //vObsState = cur.getCfg() - *cfgObs;

	      //double vOSsqn = vObsState.squareNorm();
	      //double vOSn = sqrt(vOSsqn);

	      double vOSn = sqrt(vObsState.x()*vObsState.x() + vObsState.y()*vObsState.y());

	      ConfigR2D vObsStateN(vObsState.x()/vOSn, vObsState.y()/vOSn);

	      double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();

	      if(vOSn < dMin + radObs + robotRadius)
		{
		  // double sigma = sqrt(-0.5 * reliabilityTime * reliabilityTime / log(epsilonRT));

		  double dist = (dMin - vOSn + radObs + robotRadius); // * exp(-0.5 * (time - cur.t())/sigma * (time - cur.t())/sigma );
		  if(dist > 0.5f) dist = 0.5f;

		  if(dist < 0) cout << "BBBBBBBBBBBBBBBBBBBBBBBBBBB" << endl;

// 		  if(vOSn < radObs + robotRadius)
// 		    cout << "Interieur : " << endl;

		  //cout << "obs : " << i << "     dist : " << dist << endl;
		  if(vOSn == 0) cout << "c'est la fin du monde!!!" << endl;
		  //sCur = StateR5D(vObsState.x() * dist / abs(vOSn),vObsState.y() * dist / abs(vOSn),0,0,0);
		  sCur = StateR5D(vObsStateN.x() * dist, vObsStateN.y() * dist, 0, 0, 0);
		}
	      
	      sRes += sCur;
	      delete curCfg;
	      delete cfgObs;
	    }
	  else
	    {
	      cout << "AAAAAAAAAAAAAAAAAAAAA" << endl;
	    }
	}

      return sRes;
    }
  else
    {
      return StateR5D();
    }
}

// compute the external spatial force due to the obstacles
StateR5D STDeformation::computeF3Ext(const StateR& cur, Environment* env)
{ 
  double dMin = 100.0f;
  
  double robotRadius = this->rb_->getRadius();
  
  StateR5D sRes;

  if(env != NULL)
    {
      int indMin = 0;
      double min = -1000;
      ConfigR2D vObsStateMin;
      
      // search the nearest obstacle
      for(int i=0; i<env->getNbObs(); i++)
	{
	  ConfigR* cfgObs = env->getObs(i)->getFutureCfg(cur.t());

	  if(cfgObs != NULL)
	    {
	      ConfigR2D* curCfg = (ConfigR2D*) cur.getCfg();
	      ConfigR2D vObsState = *curCfg - *cfgObs;

	      double vOSn = vObsState.norm();
	      double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();
	      if(vOSn - radObs - robotRadius < dMin)
		{
		  if(i == 0 || vOSn - radObs - robotRadius < min)
		    {
		      indMin = i;
		      min = vOSn - radObs - robotRadius;
		      vObsStateMin = vObsState/vOSn;
		    }
		}
	      delete curCfg;
	      delete cfgObs;
	    }
	}

      if(min != -1000)
	{
	  // compute the force from the nearest obstacle
	  double distNearObs = min;
	  double normForce = K_EXT_SPT * 0.5 * (dMin - distNearObs); // * (dMin - distNearObs); // dMin - distNearObs;
	  // cout << "distNearObs : " << distNearObs << "    normForce : " << normForce << endl;
	  sRes = StateR5D(vObsStateMin.x() * normForce, vObsStateMin.y() * normForce, 0, 0, 0);
	}
    }
  return sRes;
}

// compute the external temporary force due to the obstacles
StateR5D STDeformation::computeF4Ext(const StateR& cur, Environment* env)
{ 
  //cout << "0" << endl;

  // double Tmin  = 5.0;
  double Tmax  = 10.0;
  double Tstep = 0.5;

  //cout << "1" << endl;

  if(env != NULL)
    {
      //cout << "2" << endl;

      // testing if the current state is in collision with an other obstacle
      bool trouveCol = false;
      int i=0;

      //cout << "3" << endl;
      
      while(i<env->getNbObs() && !trouveCol)
	{
	  //cout << "31" << endl;

	  // get back the radius of the i-th obstacle
	  double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();

	  //cout << "32" << endl;

	  // get back the future configuration of the i-th obstacle
	  ConfigR* cfgObsI = env->getObs(i)->getFutureCfg(cur.t());

	  if(cfgObsI != NULL)
	    {
	      //cout << "33" << endl;
	      
	      // compute the (square) distance between the current config and the i-th obstacle
	      double dx = cur.x() - cfgObsI->x();
	      double dy = cur.y() - cfgObsI->y();
	      double sqDist = dx*dx + dy*dy;
	      
	      //cout << "34" << endl;
	      
	      // test if the current config is in collision with the i-th obstacle
	      if(sqDist < radObs*radObs)
		{
		  trouveCol = true;
		}
	      else
		{
		  i++;
		}
	      
	      //cout << "35" << endl;
	      
	      // delete the config of the obstacle
	      if(cfgObsI != NULL) delete cfgObsI;
	      
	      //cout << "36" << endl;
	    }
	  else
	    {
	      i++;
	    }
	}

      //cout << "4" << endl;

      if(trouveCol)
	{
	  // case where the current state is already in collision : 
	  // search for the nearest temporary free space
	  
	  //cout << "5" << endl;

	  StateR5D sRes;

	  bool trouveFS = false;
	  double dt = 0;
	  double dtMinFS = -1;

	  while(!trouveFS)
	    {
	      //cout << "51" << endl;

	      dt += Tstep;

	      // check if a shift of time dt in the past could ensure non-collision
	      bool trouveColPast = false;
	      int i=0;
	      double distMinPast = -1;
	      
	      while(!trouveColPast && i < env->getNbObs())
		{
		  //cout << "52" << endl;

		  // get back the radius of the i-th obstacle
		  double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();

		  // get back the past temporary-shifted configuration of the obstacle
		  ConfigR* pastCfg = NULL;
		  if(cur.t() - dt > 0)
		    {
		      pastCfg = env->getObs(i)->getFutureCfg(cur.t() - dt);
		    }

		  //cout << "53" << endl;

		  if(pastCfg != NULL)
		    {
		      // compute the distance from the current state position to the past obstacle position
		      double pastDx = pastCfg->x() - cur.x();
		      double pastDy = pastCfg->y() - cur.y();
		      double pastSqDist = pastDx*pastDx + pastDy*pastDy;
		      
		      // test if the current state is in collision with the past temporary-shifted obstacle position
		      if(pastSqDist < radObs*radObs)
			{
			  trouveColPast = true;
			}
		      else
			{
			  if(distMinPast == -1) distMinPast = pastSqDist;
			  else
			    {
			      if(pastSqDist < distMinPast) distMinPast = pastSqDist;
			      i++;
			    }
			}

		      delete pastCfg;
		    }
		  else i++;
		}

	      //cout << "54" << endl;

	      // test if all obstacles are free of collision from this shift
	      if(!trouveColPast)
		{
		  dtMinFS = -dt;
		  trouveFS = true;
		}

	      //cout << "55" << endl;


	      // check if a shift of time dt in the future could ensure non-collision
	      bool trouveColFuture = false;
	      int j=0;
	      double distMinFuture = -1;

	      while(!trouveColFuture && j < env->getNbObs())
		{
		  //cout << "56" << endl;

		  // get back the radius of the i-th obstacle
		  double radObs = ((ObsDisk2D*)env->getObs(j))->getRadius();

		  //cout << "561" << endl;

		  // get back the past temporary-shifted configuration of the obstacle
		  ConfigR* futureCfg = NULL;
		  futureCfg = env->getObs(j)->getFutureCfg(cur.t() + dt);

		  //cout << "57" << endl;

		  if(futureCfg != NULL)
		    {
		      // compute the distance from the current state position to the future obstacle position
		      double futureDx = futureCfg->x() - cur.x();
		      double futureDy = futureCfg->y() - cur.y();
		      double futureSqDist = futureDx*futureDx + futureDy*futureDy;
		      
		      // test if the current state is in collision with the past temporary-shifted obstacle position
		      if(futureSqDist < radObs*radObs)
			{
			  trouveColFuture = true;
			}
		      else
			{
			  if(distMinFuture == -1) distMinFuture = futureSqDist;
			  else
			    {
			      if(futureSqDist < distMinFuture) distMinFuture = futureSqDist;
			      j++;
			    }
			}

		      delete futureCfg;
		    }
		  else j++;
		}

	      //cout << "58" << endl;

	      // test if all obstacles are free of collision from this shift
	      if(!trouveColFuture && (trouveColPast || distMinFuture > distMinPast))
		{
		  dtMinFS = dt;
		  trouveFS = true;
		}
	    }

	  //cout << "59" << endl;

	  // computing of the resulting force
	  double sign;
	  if(dtMinFS > 0) sign = 1.0;
	  else sign = -1.0;
	  double Fres = 10.0 * K_EXT_TMP * sign;

	  sRes.setT(Fres);

	  return sRes;
	}

      else
	{
	  // case where the current state is not in collision : 
	  // search for the nearest obstacle (temporary distance)
	
	  //cout << "6" << endl;

	  double curDtMin = 0.0;
	  int indMin = -1;

	  StateR5D sRes;
  
	  for(int i=0; i<env->getNbObs(); i++)
	    {
	      bool finSNO = false;
	      bool trouveNO = false;
	      double dt = 0;

	      // get back the radius of the i-th obstacle
	      double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();
	      
	      //cout << "61" << endl;

	      while(!finSNO && !trouveNO)
		{
		  dt += Tstep;

		  if(dt > Tmax || (indMin != -1 && dt > fabs(curDtMin)))
		    {
		      // stop the search if the maximum time has been reached
		      finSNO = true;
		    }
		  else
		    {
		      // test if the current state is in collision with the both temporary-shifted obstacle positions
		      
		      bool pastCol = false;
		      bool futureCol = false;
		      double pastSqDist = 0;
		      double futureSqDist = 0;

		      // get back the past temporary-shifted configuration of the obstacle
		      ConfigR* pastCfg = NULL;
		      if(cur.t() - dt > 0)
			{
			  pastCfg = env->getObs(i)->getFutureCfg(cur.t() - dt);
			}

		      if(pastCfg != NULL)
			{
			  // compute the distance from the current state position to the past obstacle position
			  double pastDx = pastCfg->x() - cur.x();
			  double pastDy = pastCfg->y() - cur.y();
			  pastSqDist = pastDx*pastDx + pastDy*pastDy;

			  // test if the current state is in collision with the past temporary-shifted obstacle position
			  if(pastSqDist < radObs*radObs)
			    {
			      pastCol = true;
			      trouveNO = true;
			      indMin = i;
			      curDtMin = -dt;
			    }
			  delete pastCfg;
			}

		      // get back the future temporary-shifted configuration of the obstacle
		      ConfigR* futureCfg = NULL;
		      futureCfg = env->getObs(i)->getFutureCfg(cur.t() + dt);

		      if(futureCfg != NULL)
			{
			  // compute the distance from the current state position to the future obstacle position
			  double futureDx = futureCfg->x() - cur.x();
			  double futureDy = futureCfg->y() - cur.y();
			  futureSqDist = futureDx*futureDx + futureDy*futureDy;

			  if(futureSqDist < radObs*radObs && (!pastCol || futureSqDist < pastSqDist))
			    {
			      futureCol = true;
			      trouveNO = true;
			      indMin = i;
			      curDtMin = dt;
			    }
			  delete futureCfg;
			}
		      
		    }
		}
	    }

	  //cout << "62" << endl;

	  // computing of the resulting force
	  if(indMin != -1)
	    {
	      double sign;
	      if(curDtMin > 0) sign = 1.0;
	      else sign = -1.0;
	      double Fres = ((Tmax - fabs(curDtMin))*(Tmax - fabs(curDtMin))) / (Tmax * Tmax) * K_EXT_TMP * -sign;

	      sRes.setT(Fres);
	    }

	  //cout << "63" << endl;

	  return sRes;
	}
   
    }
  else
    {
      //cout << "7" << endl;

      return StateR5D();
    }
}

// search the nearest temporary free space
StateR5D STDeformation::searchNearTempFS(const StateR& cur, Environment* env)
{
  StateR5D sRes;

  double Tmax = 1.5; // 3.0;
  double Tstep = 0.3; // 0.5
  
  bool trouveFS = false;
  double dt = 0;
  double dtMinFS = -1;
  
  while(!trouveFS)
    { 
      dt += Tstep;
      
      // check if a shift of time dt in the past could ensure non-collision
      bool trouveColPast = false;
      int i=0;
      double distMinPast = -1;
      
      while(!trouveColPast && i < env->getNbObs())
	{ 
	  // get back the radius of the i-th obstacle
	  double radObs = ((ObsDisk2D*)env->getObs(i))->getRadius();
	  
	  // get back the past temporary-shifted configuration of the obstacle
	  ConfigR* pastCfg = NULL;
	  if(cur.t() - dt > 0)
	    {
	      pastCfg = env->getObs(i)->getFutureCfg(cur.t() - dt);
	    }
	   
	  if(pastCfg != NULL)
	    {
	      // compute the distance from the current state position to the past obstacle position
	      double pastDx = pastCfg->x() - cur.x();
	      double pastDy = pastCfg->y() - cur.y();
	      double pastSqDist = pastDx*pastDx + pastDy*pastDy;
	      
	      // test if the current state is in collision with the past temporary-shifted obstacle position
	      if(pastSqDist < radObs*radObs)
		{
		  trouveColPast = true;
		}
	      else
		{
		  if(distMinPast == -1) distMinPast = pastSqDist;
		  else
		    {
		      if(pastSqDist < distMinPast) distMinPast = pastSqDist;
		      i++;
		    }
		}
	      
	      delete pastCfg;
		    }
	  else i++;
	}
      
      // test if all obstacles are free of collision from this shift
      if(!trouveColPast)
	{
	  dtMinFS = -dt;
	  trouveFS = true;
	}
      
      // check if a shift of time dt in the future could ensure non-collision
      bool trouveColFuture = false;
      int j=0;
      double distMinFuture = -1;
      
      while(!trouveColFuture && j < env->getNbObs())
	{ 
	  // get back the radius of the i-th obstacle
	  double radObs = ((ObsDisk2D*)env->getObs(j))->getRadius();
	  
	  // get back the past temporary-shifted configuration of the obstacle
	  ConfigR* futureCfg = NULL;
	  futureCfg = env->getObs(j)->getFutureCfg(cur.t() + dt);
	  
	  if(futureCfg != NULL)
	    {
	      // compute the distance from the current state position to the future obstacle position
	      double futureDx = futureCfg->x() - cur.x();
	      double futureDy = futureCfg->y() - cur.y();
	      double futureSqDist = futureDx*futureDx + futureDy*futureDy;
	      
	      // test if the current state is in collision with the past temporary-shifted obstacle position
	      if(futureSqDist < radObs*radObs)
		{
		  trouveColFuture = true;
		}
	      else
		{
		  if(distMinFuture == -1) distMinFuture = futureSqDist;
		  else
		    {
		      if(futureSqDist < distMinFuture) distMinFuture = futureSqDist;
		      j++;
		    }
		}
	      
	      delete futureCfg;
	    }
	  else j++;
	}
      
      // test if all obstacles are free of collision from this shift
      if(!trouveColFuture && (trouveColPast || distMinFuture > distMinPast))
	{
	  dtMinFS = dt;
	  trouveFS = true;
	}
    }

  // compute the direction of the resulting force
  StateR5D force;
  if(dtMinFS > 0) force = StateR5D(0,0,(1 + dtMinFS / Tmax) * (*K_EXT),0,0);
  else force = StateR5D(0,0,(-1 + dtMinFS / Tmax) * (*K_EXT),0,0);
  
  return force;
}

// // compute the external spatial and temporary force due to the obstacles
// StateR5D STDeformation::computeF5Ext(const StateR& cur, Environment* env, double time, bool last)
// {
//   double Tstep = 0.5;

//   double dSTmax = (*maxEffectDist);
//   double vMin = this->rb_->getVMax() * 0.20 * (*K_SPT_TMP);

//   double Tmax  = dSTmax / vMin;
//   if(Tmax > (*reliabilityTime)) Tmax = (*reliabilityTime);

//   double thresTime = 4.0;
//   double Mmax = 50.0;

//   // verify that the environment is not NULL
//   if(this->env_ == NULL)
//     {
// #if DEBUG >= 1
//       cout << "DeformIfc::computeF5Ext : Warning. The environment is null." << endl;
// #endif
//       return StateR5D();
//     }
  
//   // search the nearest obstacle from the current state
//   int nearestObs = -1;
//   double minDist = 0;
//   double minM = 0;
//   StateR5D nearestState;
  
//   for(int i=0; i<this->env_->getNbObs(); i++)
//     {
//       // get back the radius of the i-th obstacle
//       double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();

//       // get back the configuration of the i-th obstacle at the current time
//       ConfigR* curCfgObs = this->env_->getObs(i)->getFutureCfg(cur.t());
      
//       bool col = false;
//       if(curCfgObs != NULL)
// 	{
// 	  // compute the distance from the obstacle to the current state
// 	  double curDX = cur.x() - curCfgObs->x();
// 	  double curDY = cur.y() - curCfgObs->y();
// 	  double curSqDist = curDX*curDX + curDY*curDY;
	  
// 	  // test if the obstacle is in collision with the current state 
// 	  if(curSqDist < radObs*radObs)
// 	    {
// 	      if(last)
// 		{
// 		  delete curCfgObs;
// 		  return searchNearTempFS(cur, env);
// 		}

// 	      if((nearestObs == -1) || sqrt(curSqDist) - radObs < minDist)
// 		{
// 		  nearestObs = i;
// 		  minDist = sqrt(curSqDist) - radObs;
// 		  minM = 0;
// 		  nearestState = StateR5D(curCfgObs->x(), curCfgObs->y(), cur.t(), 0, 0);
// 		  col = true;
// 		}
// 	    }
// 	  delete curCfgObs;
// 	}

//       //if(col && last) cout << "col" << endl;

//       if(!col)
// 	{
// 	  // compute for each time step the spatio-temporal distance between the current state and the obstacle
// 	  for(double curTime=cur.t() - Tmax; curTime <= cur.t() + Tmax; curTime += Tstep)
// 	    {
// 	      // verify that the current time curTime is not smaller than 0
// 	      if(curTime >= 0)
// 		{
// 		  // get back the configuration of the i-th obstacle at time curTime
// 		  ConfigR* cfgObs = this->env_->getObs(i)->getFutureCfg(curTime);
		  
// 		  // get back the configuration of the i-th obstacle at time curTime + Tstep
// 		  ConfigR* cfgObs2 = this->env_->getObs(i)->getFutureCfg(curTime + Tstep);
		  
// 		  // verify that cfgObs and cfgObs2 are not null
// 		  if(cfgObs != NULL && cfgObs2 != NULL)
// 		    { 
// 		      // compute the speed of the i-th obstacle at time curTime
// 		      Speed2D spdObs((cfgObs2->x() - cfgObs->x()) / Tstep, (cfgObs2->y() - cfgObs->y()) / Tstep);
		      
// 		      // compute the metric
// 		      double M = 0;
// 		      double sqNormSpdObs = spdObs.sx() * spdObs.sx() + spdObs.sy() * spdObs.sy();
// 		      double sqNormSpdRb = cur.sx() * cur.sx() + cur.sy() * cur.sy();
// 		      if(sqNormSpdObs >= sqNormSpdRb)
// 			{
// 			  M = sqrt(sqNormSpdObs) + vMin;
// 			}
// 		      else
// 			{
// 			  M = sqrt(sqNormSpdRb) + vMin;
// 			}

// 		      if(fabs(cur.t() - time) < thresTime)
// 			{
// 			  M = Mmax + ((1.0 * (*K_SPT_TMP)) - Mmax) * fabs(cur.t() - time) / thresTime;
// 			}
// 		      else
// 			{
// 			  M = 1.0 * (*K_SPT_TMP);
// 			}
		      
// 		      // compute the radius of the robot crossed by the vector curCfg - cfgObs
		      
// 		      // compute the normalized vector (p_i,t_i) - (p_j^obs,t_j^obs)
// 		      Speed2D v2((cur.x() - cfgObs->x()) / (cur.t() - curTime), (cur.y() - cfgObs->y()) / (cur.t() - curTime));
		      
// 		      // compute the square norm of this vector
// 		      double sqNormV2 = v2.sx()*v2.sx() + v2.sy()*v2.sy();
		      
// 		      // compute the time t_exit to go out of the obstacle
// 		      double t_exit = (radObs + this->rb_->getRadius()) / sqrt(spdObs.sx()*spdObs.sx() + spdObs.sy()*spdObs.sy()
// 		      			+ v2.sx()*v2.sx() + v2.sy()*v2.sy()
// 		      			- 2.0*spdObs.sx()*v2.sx() - 2.0*spdObs.sy()*v2.sy());
		      
// 		      // compute the radius of the robot crossed
// 		      double radObsCrossed = sqrt(M*M + sqNormV2) * t_exit;
// 		      //double radObsCrossed = radObs;
// // 		      if(radObsCrossed < 1)
// // 			{
// // 			  cout << "radObsCrossed : " << radObsCrossed << "\t";
// // 			  cout << "v_obs : (" << spdObs.sx() << "," << spdObs.sy()
// // 			}
		      
// 		      // compute the distance from the obstacle to the current state
// 		      double dx = cur.x() - cfgObs->x();
// 		      double dy = cur.y() - cfgObs->y();
// 		      double dt = M * (cur.t() - curTime);
// 		      double dist = sqrt(dx*dx + dy*dy + dt*dt) - radObsCrossed;
		      
// 		      // save the current obstacle state if its the nearest
// 		      if((nearestObs == -1 && dist < dSTmax) || dist < minDist)
// 			{
// 			  nearestObs = i;
// 			  minDist = dist;
// 			  nearestState = StateR5D(cfgObs->x(), cfgObs->y(), curTime, spdObs.sx(), spdObs.sy());
// 			  minM = M;
// 			}
// 		    }
// 		  if(cfgObs != NULL) delete cfgObs;
// 		  if(cfgObs2 != NULL) delete cfgObs2;
// 		}
// 	    }
// 	}
//     }

//   // compute the force to apply to the current state
//   if(nearestObs != -1)
//     {
// //       cout << "dist = " << minDist << "\t";

//       if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
//       (*tableNearState)[I] = new StateR5D(nearestState);

//       // compute the coefficient due to the reliability time
//       double sigma = sqrt(-0.5 * (*reliabilityTime) * (*reliabilityTime) / log(*epsilonRT));
//       double coefRT = exp(-0.5 * (time - cur.t())/sigma * (time - cur.t())/sigma );

//       // compute the direction of the force
//       double dx = cur.x() - nearestState.x();
//       double dy = cur.y() - nearestState.y();
//       double dt = minM * (cur.t() - nearestState.t());
//       double dsx = 0;
//       double dsy = 0;
//       double length = sqrt(dx*dx + dy*dy + dt*dt);
//       //cout << "length : " << length << "\t";

//       StateR5D dir;
//       if(length != 0 && !last)
// 	{
//  	  if(minM != 0) dir = StateR5D(dx/length,dy/length,dt/(minM * length),dsx,dsy);
//  	  else dir = StateR5D(dx/length,dy/length,dt/length,dsx,dsy);

// // 	  cout << "dir : (" << dir.x() << "," << dir.y() << "," << dir.t() << ")\t";
// 	}
//       else if(length != 0 && last)
// 	{
// 	  //return (searchNearTempFS(cur, env) * coefRT);
// 	  dir = StateR5D(0,0,dt/(minM * length),dsx,dsy);
// 	  cout << "length : " << length << endl;
// // 	  cout << "[LAST] dir : (" << dir.x() << "," << dir.y() << "," << dir.t() << ")\t";
// 	}
//       else
// 	dir = StateR5D();

//       // compute the force to apply
//       StateR5D force = dir * ((dSTmax - minDist) * (dSTmax - minDist) / (dSTmax * dSTmax) * (*K_EXT)) * coefRT;
// //       cout << "force : (" << force.x() << "," << force.y() << "," << force.t() << ")" << endl;
// //       cout << "norm : " << sqrt(force.x() * force.x() + force.y() * force.y() + force.t() * force.t()) << endl;
//       return force;
//     }
//   else
//     {
//       if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
//       (*tableNearState)[I] = NULL;

//       if(last)
// 	{
// 	  if(*useTemporalConsistency)
// 	    {
// 	    StateR5D force;
// 	    force.setT(-0.5 * (*K_EXT));
// 	    return force;
// 	  }
// 	  else
// 	    {
// 	      return StateR5D();
// 	    }
// 	}
//       else
// 	return StateR5D();
//     }
// }

// compute the external spatial and temporary force due to the obstacles
StateR5D STDeformation::computeF6Ext(const StateR& cur, Environment* env, double time, bool last)
{
  // verify if the current state is not too far from the current time to be under external forces
  if(cur.t() - time > (*reliabilityTime))
    {
      StateR5D force;
      if(fabs(force.t()) > 10.0) 
	{ cout << "force0 : "; force.display(); cout << endl; }
      return force;
    }

  // init needed variables
  double Dmax = *spMed;
  double Tmax = *tmpMed;
  double MDmax = Dmax * (*alphaMax);
  double MTmax = Tmax * (*betaMax);
//   double MDTmax;
//   if(MDmax > MTmax) MDTmax = MDmax;
//   else MDTmax = MTmax;

//   MDTmax = ((*reliabilityTime) - cur.t() + time)/(*reliabilityTime) * MDmax + (cur.t() - time)/(*reliabilityTime) * MTmax;
  
  // compute spatial coefficient Alpha (for the metric)
  double Alpha = *alphaMin + ((*reliabilityTime) - cur.t() + time) * ((*alphaMax - *alphaMin) / (*reliabilityTime));
  //double Alpha = *alphaMax;
  
  // compute spatial coefficient Alpha (for the metric)
  double Beta = *betaMax - (cur.t() - time) * ((*betaMax - *betaMin) / (*reliabilityTime));
  //double Beta = *betaMin;

  // search for the nearest obstacle
  int nearestObs = -1;
  double sqMminDist = 0;
  StateR5D nearestState; 
  
  if(this->modeDeform == SPATONLY_DEFORM)
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  // get back the current obstacle
	  ObsDisk2D* curObs = (ObsDisk2D*) this->env_->getObs(i);
	  
	  double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();
	  
	  // get back the last position of the obstacle
	  ConfigR* configObs = NULL;
	  configObs = curObs->getLastCfg();
	  
	  if(configObs != NULL)
	    {
	      Vec3D curRbNearObs(configObs->x() - cur.x(), configObs->y() - cur.y(), 0);

	      double sqMminDistObs = (Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1]);
	      
	      // check if this obstacle is the nearest obstacle from the current robot's control point
	      if(nearestObs == -1 || sqMminDistObs < sqMminDist)
		{
		  sqMminDist = sqMminDistObs;
		  nearestObs = i;
		  nearestState = StateR5D(configObs->x(),configObs->y(),cur.t(),0,0);
		}
	    }

	  // deletions
	  if(configObs != NULL) delete configObs;
	}
    }
  else
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  // get back the current obstacle
	  ObsDisk2D* curObs = (ObsDisk2D*) this->env_->getObs(i);
	  
	  double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();  
	  //double MDTmax = sqrt(Alpha*Alpha*(Dmax + radObs + rb_->getRadius())*(Dmax + radObs + rb_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  double sqMDTmax = (Alpha*Alpha*(Dmax + radObs + rb_->getRadius())*(Dmax + radObs + rb_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  //cout << "MDTmax : " << MDTmax << endl;
	  
	  // get back the last position of the obstacle
	  ConfigR* configObs = NULL;
	  configObs = curObs->getLastCfg();
	  
	  // compute the coordinates of the obstacle's last control point in the metric frame
	  Vec3D MconfigObs(Alpha * configObs->x(), Alpha * configObs->y(), Beta * curObs->getLastTime());
	  
	  // compute the coordinates of the robot's control point in the metric frame
	  Vec3D MconfigRb(Alpha * cur.x(), Alpha * cur.y(), Beta * cur.t());
	  
	  // compute the robot's speed vector in the metric frame
	  Vec3D MspeedObs(Alpha * curObs->getLastSpdX(), Alpha * curObs->getLastSpdY(), Beta);
	  
	  // compute the vector configRobot - configObs with the metric
	  Vec3D MORi (MconfigRb[0] - MconfigObs[0], MconfigRb[1] - MconfigObs[1], MconfigRb[2] - MconfigObs[2]);
	  
	  // compute the norm of MspeedRb
	  double MnormSObs = sqrt(MspeedObs[0]*MspeedObs[0] + MspeedObs[1]*MspeedObs[1] + MspeedObs[2]*MspeedObs[2]);
	  
	  // compute the normalized vector NMspeedRb
	  Vec3D NMspeedObs;
	  if(MnormSObs != 0)
	    {
	      NMspeedObs[0] = MspeedObs[0]/MnormSObs;
	      NMspeedObs[1] = MspeedObs[1]/MnormSObs;
	      NMspeedObs[2] = MspeedObs[2]/MnormSObs;
	    }
	  
	  // compute the scalar product between MORi and nSObs
	  double pdScal = MORi[0] * NMspeedObs[0] + MORi[1] * NMspeedObs[1] + MORi[2] * NMspeedObs[2];
	  
	  //compute the vector OP (obstacle's nearest control point - obstacle's init control point) in the metric frame
	  Vec3D NOP(pdScal * NMspeedObs[0], pdScal * NMspeedObs[1], pdScal * NMspeedObs[2]);
	  
	  // compute the vector OP in the normal frame
	  Vec3D OP(NOP[0]/Alpha, NOP[1]/Alpha, NOP[2]/Beta);
	  
	  // compute the obstacle's nearest position
	  StateR5D nearestStateCurObs(configObs->x() + OP[0], configObs->y() + OP[1], curObs->getLastTime() + OP[2], 0, 0);
	  
	  // compute the vector nearestStateCurObs - curRb
	  //Vec3D curRbNearObs(OP[0] - cur.x(), OP[1] - cur.y(), OP[2] - cur.t());
	  Vec3D curRbNearObs(nearestStateCurObs.x() - cur.x(), nearestStateCurObs.y() - cur.y(), nearestStateCurObs.t() - cur.t());
	  //cout << "curRbNearObs : (" << curRbNearObs[0] << "," << curRbNearObs[1] << "," << curRbNearObs[2] << ")" << endl;
	  
	  // compute the norm of curRbNearObs ie. the minimal distance from the current robot's control point to the current obs
	  //double MminDistObs = sqrt(Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1] + Beta*Beta*curRbNearObs[2]*curRbNearObs[2]);
	  double sqMminDistObs = (Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1] + Beta*Beta*curRbNearObs[2]*curRbNearObs[2]);
	  //cout << "MminDistObs : " << MminDistObs << "  Alpha : " << Alpha << "  Beta : " << Beta << "  cur.t() - time : " << cur.t() - time << "  reliabilityTime : " << (*reliabilityTime) << endl; 
	  
	  // check if this obstacle is the nearest obstacle from the current robot's control point
	  if((sqMminDistObs < sqMDTmax) && (nearestObs == -1 || sqMminDistObs < sqMminDist))
	    {
	      sqMminDist = sqMminDistObs;
	      nearestObs = i;
	      nearestState = nearestStateCurObs;
	    }
	  
	  // deletions
	  if(configObs != NULL) delete configObs;
	}
    }
  
  // compute the force to apply to the current state
  if(nearestObs != -1)
    {
      // add the nearest state to the corresponding display table entry
      if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
      (*tableNearState)[I] = new StateR5D(nearestState);

      // compute the coefficient due to the reliability time
      double sigma = sqrt(-0.5 * (*reliabilityTime) * (*reliabilityTime) / log(*epsilonRT));
      double coefRT = exp(-0.5 * (time - cur.t())/sigma * (time - cur.t())/sigma );
      //double coefRT = exp((time - cur.t())*(time - cur.t())*log(*epsilonRT)/((*reliabilityTime)*(*reliabilityTime)));

      // compute the direction of the force
      double dx = (cur.x() - nearestState.x());
      double dy = (cur.y() - nearestState.y());
      double dt = (cur.t() - nearestState.t());
      double dsx = 0;
      double dsy = 0;
      //double dd = sqrt(dx*dx + dy*dy);
      double sqdd = dx*dx + dy*dy;
      double length = sqrt(dx*dx + dy*dy + dt*dt);
      double Ndd = 0;

      StateR5D dir;
      if(length != 0 && !last)
 	{
// 	  if(Alpha == 0) dir = StateR5D(dx/(length * Alpha),dy/(length * Alpha),0,dsx,dsy);
// 	  else if(Beta == 0) dir = StateR5D(0,0,dt/(length * Beta),dsx,dsy);
	  dir = StateR5D(dx/length,dy/length,dt/length,dsx,dsy);
	  //if(dd != 0) Ndd = dd/length;
 	}
      else if(length != 0 && last)
 	{
 	  //return (searchNearTempFS(cur, env) * coefRT);
 	  dir = StateR5D(0,0,dt/length,dsx,dsy);
 	}
      else
 	dir = StateR5D();
      
      // check if this state is in collision
      double radObs = ((ObsDisk2D*)(this->env_->getObs(nearestObs)))->getRadius();
      double radRb = rb_->getRadius();
      StateR5D forceNoCol;
      if(!last && sqdd < (radObs + radRb)*(radObs + radRb))
	{
	  double dd = sqrt(sqdd);
	  double ddx = (dx * (radObs +radRb) / dd) - dx;
	  double ddy = (dy * (radObs +radRb) / dd) - dy;
	  //forceNoCol = StateR5D(ddx,ddy,0,0,0);
	  forceNoCol = StateR5D();
	}

      // compute the base external force
      double val = (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));
      StateR5D baseExtForce;
      if(val > 0) baseExtForce = dir * (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));

      if(this->modeDeform == SPATONLY_DEFORM)
	{
	  StateR5D force(baseExtForce.x() + forceNoCol.x(),
			 baseExtForce.y() + forceNoCol.y(),
			 0,
			 baseExtForce.sx() + forceNoCol.sx(),
			 baseExtForce.sy() + forceNoCol.sy());
	  force = force * (coefRT * (*K_EXT));
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force1 : "; force.display(); cout << endl; }
	  return force;
	}
      else if(this->modeDeform == TEMPONLY_DEFORM)
	{
	  StateR5D force(0,
			 0,
			 baseExtForce.t() + forceNoCol.t(),
			 baseExtForce.sx() + forceNoCol.sx(),
			 baseExtForce.sy() + forceNoCol.sy());
	  force = force * (coefRT * (*K_EXT));
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force2 : "; force.display(); cout << endl; }
	  return force;
	}
      else
	{
	  // compute the force to apply
	  StateR5D force(baseExtForce.x() + forceNoCol.x(),
			 baseExtForce.y() + forceNoCol.y(),
			 baseExtForce.t() + forceNoCol.t(),
			 baseExtForce.sx() + forceNoCol.sx(),
			 baseExtForce.sy() + forceNoCol.sy());
	  force = force * (coefRT * (*K_EXT));
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force3 : "; force.display(); cout << endl; }
	  return force;
	}
    }
  else
    {
      if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
      (*tableNearState)[I] = NULL;
      
      if(last && (this->modeDeform != SPATONLY_DEFORM))
 	{
 	  if(*useTemporalConsistency)
 	    {
	      StateR5D force;
	      force.setT(-0.5 * (*K_EXT));
	      if(fabs(force.t()) > 10.0) 
		{ cout << "force4 : "; force.display(); cout << endl; }
	      
	      return force;
	    }
 	  else
 	    {
	      StateR5D force;
	      if(fabs(force.t()) > 10.0) 
		{ cout << "force5 : "; force.display(); cout << endl; }
 	      return force;
 	    }
 	}
      else
	{
	  StateR5D force;
	  if(fabs(force.t()) > 10.0) 
	    { cout << "force6 : "; force.display(); cout << endl; }
	  return force;
	}
    }
}


// compute the total external forces
StateR5D STDeformation::computeFExt(const StateR& cur, Environment* env, double time, bool last)
{
  StateR5D sRes = this->computeF6Ext(cur,env,time,last);
  return sRes;
  //StateR5D vFSExt = computeStFExt(cur,env,last);
  //return (sRes + vFSExt);
}

// compute the external forces due to static obstacles
StateR5D STDeformation::computeStFExt(const StateR& cur, Environment* env, bool last)
{
  if(last)
    {
      return StateR5D();
    }
  else
    {
      // get back the maximal effective distance
      //double Dmax = *spMed * 0.8 + this->rb_->getRadius();
      double Dmax = (*maxStEffectDist) + this->rb_->getRadius();
      
      // check if the environment is not null
      if(this->env_ == NULL || this->env_->getNbStObs() <= 0)
	{
	  return StateR5D();
	}

      int nbObs = this->env_->getNbStObs();
      double minDist = -1;
      ConfigR2D nearest;
      ConfigR2D centerNearest;
      bool collision = false;

      // search for the nearest static obstacle from cur
      for(int i=0; i<nbObs; i++)
	{
	  // compute the nearest position of the current obstacle to cur
	  double dist;
	  ConfigR2D near;
	  ConfigR2D curCfg(cur.x(),cur.y());
	  bool inside = this->env_->getStObs(i)->nearPosFrom(&curCfg, near, dist);

	  // check if the current obstacle is the nearest obstacle from cur
	  if(inside || (!collision && dist < minDist) || (minDist == -1 && dist < Dmax))
	    {
	      minDist = dist;
	      nearest = near;
	      centerNearest = ConfigR2D(this->env_->getStObs(i)->getInitCfg()->x(),
					this->env_->getStObs(i)->getInitCfg()->y());
	      if(inside) collision = true;
	    }
	}

      // check if a force must be applied
      if(minDist != -1)
	{
	  // compute the direction in which the force must be applied
	  Vec2D vecDir(0.0,0.0);
	  if(collision)
	    {
	      vecDir.setVal(0, nearest.x() - cur.x());
	      vecDir.setVal(1, nearest.y() - cur.y());
	    }
	  else
	    {
	      vecDir.setVal(0, cur.x() - nearest.x());
	      vecDir.setVal(1, cur.y() - nearest.y());
	    }
	  Vec vecNDir = vecDir.normalized();	  
	  StateR5D dir(vecNDir.getVal(0),vecNDir.getVal(1),0.0,0.0,0.0);

	  // compute a small force coming from the center of the obstacle
	  Vec2D vecForceCenter(cur.x() - centerNearest.x(),
			       cur.y() - centerNearest.y());
	  Vec vecNForceCenter = vecForceCenter.normalized();
	  StateR5D forceCenter(0.3*vecNForceCenter.getVal(0),
			       0.3*vecNForceCenter.getVal(1),
			       0.0,
			       0.0,
			       0.0);

	  if(collision)
	    {
	      StateR5D force = ((dir * ((this->rb_->getRadius() + fabs(minDist))/Dmax + 1.0)) + forceCenter) * (*K_EXT);
	      return force;
	    }
	  else
	    {
	      StateR5D force = ((dir * (1.0 - minDist/Dmax)) + forceCenter) * (*K_EXT);// * 0.25;
	      return force;
	    }
	}
      else return StateR5D();
    }
}

// compute the total forces
StateR5D STDeformation::computeForces(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, double time,bool last)
{
  StateR5D vFInt = computeFInt(prev,cur,next);
  StateR5D vFExt = computeFExt(cur,env,time,last);

  StateR5D vRes = StateR5D(vFInt + vFExt);
  return vRes;
}

// return the reliability coefficient to apply on the forces 
// knowing the difference dt between the current time and the 
// time of the node on which we apply the force
double STDeformation::getCoefReliability(double dt) const
{
  double uncerHorizon = (*reliabilityTime) - (*predHorizon_);
  
  if(dt <= (*predHorizon_))
    {
      return 1.0;
    }
  else if(dt <= (*reliabilityTime))
    {
      double uncerTime = dt - (*predHorizon_);
      return exp(uncerTime*uncerTime*log(*epsilonRT)/((uncerHorizon)*(uncerHorizon)));
    }
  else return 0.0;
}


                             /*********************************************************/
                             /********** Connectivity and Collision checking **********/
                             /*********************************************************/


double STDeformation::adjustTime(STrajectory& oldTraj, STrajectory& newTraj, int i, int j)
{
  if(j == 0)
    {
      // no adjustment to the first state
      return 0;
    }
  else
    {
      double oldTIm1 = (oldTraj[i-1]).t();
      double newTIm1 = (newTraj[j-1]).t();

      double oldPIm1x = (oldTraj[i-1]).x();
      double oldPIm1y = (oldTraj[i-1]).y();
      double oldPIx = (oldTraj[i]).x();
      double oldPIy = (oldTraj[i]).y();
      double oldDX = oldPIx - oldPIm1x;
      double oldDY = oldPIy - oldPIm1y;
      double oldDist = sqrt(oldDX*oldDX + oldDY*oldDY);

      double newPIm1x = (newTraj[j-1]).x();
      double newPIm1y = (newTraj[j-1]).y();
      double newPIx = (newTraj[j]).x();
      double newPIy = (newTraj[j]).y();
      double newDX = newPIx - newPIm1x;
      double newDY = newPIy - newPIm1y;
      double newDist = sqrt(newDX*newDX + newDY*newDY);

      double oldVIm1x = (oldTraj[i-1]).sx();
      double oldVIm1y = (oldTraj[i-1]).sy();
      double oldVIx = (oldTraj[i]).sx();
      double oldVIy = (oldTraj[i]).sy();
      double oldVMoy = (sqrt(oldVIm1x*oldVIm1x + oldVIm1y*oldVIm1y) + sqrt(oldVIx*oldVIx + oldVIy*oldVIy))/2;

      if(oldVMoy == 0)
	{
	  cout << "DeformIfc::adjustTime : Warning. oldVMoy = 0" << endl;
	  return 0;
	}

      double dtRes = (newTIm1 - oldTIm1 + (newDist - oldDist)/(oldVMoy));
      //cout << "newTIm1 : " << newTIm1 << "   oldTIm1 : " << oldTIm1 << "   newDist : " << newDist << "   oldDist : " << oldDist << "   oldVMoy : " << oldVMoy << "   oldPIm1x : " << oldPIm1x << "   oldPIm1y : " << oldPIm1y << endl;
      return dtRes;
    }
}

// check the connectivity between two nodes of the trajectory
void STDeformation::checkConnectivity(int i, const StateR& oldCur, const StateR& prev, const StateR& cur, StateR& force)
{
  double maxSpd = this->rb_->getVMax();
  double maxAcc = this->rb_->getAMax();

  // check the temporal consistency of the curve
  double oldDX = oldCur.x() - prev.x();
  double oldDY = oldCur.y() - prev.y();
  bool noMoreConnec1 = false;
  bool noMoreConnec2 = false;
  bool noMoreConnec11 = false;
  bool noMoreConnec22 = false;
  
  if(oldCur.t() + force.t() < prev.t() + oldDX / this->rb_->getVMax())
    {
      (*typeState)[i] = -2;
      noMoreConnec1 = true;
      if(useTemporalConsistency) force.setT(prev.t() + fabs(oldDX) / this->rb_->getVMax() - oldCur.t());
    }
  if(oldCur.t() + force.t() < prev.t() + oldDY / this->rb_->getVMax())
    {
      (*typeState)[i] = -2;
      noMoreConnec2 = true;
      if(useTemporalConsistency) force.setT(prev.t() + fabs(oldDY) / this->rb_->getVMax() - oldCur.t());
    }
  if(!noMoreConnec1 && !noMoreConnec2)
    {
      (*typeState)[i] = 0;
    }
  
  if(fabs(force.t()) > 2.0)
    {
      cout << "prev.t() : " << prev.t() << endl;
      cout << "oldCur.t() : " << oldCur.t() << endl;
      cout << "oldDX : " << oldDX << endl;
      cout << "oldDY : " << oldDY << endl;
      cout << "this->rb_->getVMax() : " << this->rb_->getVMax() << endl;
    }

  // check the connectivity of the curve
  double newDX = oldCur.x() + force.x() - prev.x();
  double newDY = oldCur.y() + force.y() - prev.y();
  double DMAX = maxSpd * (oldCur.t() + force.t() - prev.t());
  bool noMoreConnec3 = false;
  bool noMoreConnec4 = false;
  bool noMoreConnec5 = false;
  bool noMoreConnec6 = false;
  
  if(DMAX < 0) DMAX = -DMAX;
  
  if( newDX > DMAX )
    {
      (*typeState)[i] = -2;
      noMoreConnec3 = true;
    }
  else if ( newDX < -DMAX )
    {
      (*typeState)[i] = -2;
      noMoreConnec4 = true;
    }
  if( newDY > DMAX ) 
    {
      (*typeState)[i] = -2;
      noMoreConnec5 = true;
    }
  else if ( newDY < -DMAX ) 
    {
      (*typeState)[i] = -2;
      noMoreConnec6 = true;
    }
  if(!noMoreConnec1 && !noMoreConnec2 && !noMoreConnec3 && !noMoreConnec4 && !noMoreConnec5 && !noMoreConnec6)
    {
      (*typeState)[i] = 0;
    }
}

// check if the still-to-be executed part of the trajectory is in collision with the environment
// return the first time from the current time where a collision occurs 
// return -1 if there is no collision
double STDeformation::checkCollision(STrajectory* traj,Environment* env,double curTime)
{
  double res = -1;

  // check if the trajectory is not null
  if(traj == NULL)
    return -1;
  
  // check if the environment is not null
  if(env == NULL)
    return -1;
  
  bool trouve = false;
  int i=0;

  while(!trouve && i<traj->getSize())
    {
      // get back the time of the current state
      double x = ((PMSTraj*)this)->at(i).x(); //((StateS6D*)traj->at(i))->x();
      double y = ((PMSTraj*)this)->at(i).y(); //((StateS6D*)traj->at(i))->y();
      double time = ((PMSTraj*)this)->at(i).t(); //((StateS6D*)traj->at(i))->t();

      // test if the current state has not been already visited
      if(time > curTime)
	{
	  // test if a static obstacle obstructs the path between the previous and the current state
	  if(env->getNbStObs() != 0)
	    {
	      int j=0;
	      while(!trouve && j<env->getNbStObs())
		{
		  // conpute two configurations from the previous and the current state
		  ConfigR2D prevCfg(((StateS6D*)traj->at(i-1))->x(),((StateS6D*)traj->at(i-1))->y());
		  ConfigR2D curCfg(((StateS6D*)traj->at(i))->x(),((StateS6D*)traj->at(i))->y());

		  // test if the current static obstacle obstructs the path
		  if(env->getStObs(j)->obstruct(&prevCfg,&curCfg))
		    {
		      trouve = true;
		      res = time - curTime;
		    }
		  else j++;
		}
	    }
	  
	  // test if a dynamic obstacle obstructs the path between the previous and the current state
	  if(env->getNbObs() != 0)
	    {
	      int j=0;
	      while(!trouve && j<env->getNbObs())
		{
		  // get back the configuration of the obstacle at time "time"
		  ConfigR* cfgObs = env->getObs(j)->getFutureCfg(time);

		  if(cfgObs != NULL)
		    {
		      // compute the distance between the current state and the current obstacle
		      double dx = x - cfgObs->x();
		      double dy = y - cfgObs->y();
		      double sqDist = dx*dx + dy*dy;
		      
		      double minDist = (((ObsDisk2D*)env->getObs(j))->getRadius()) + this->rb_->getRadius();
		      
		      if(sqDist < minDist*minDist)
			{
			  trouve = true;
			  //res = time - curTime;
			  res = 2.0*(time - curTime); // hack to take in account collsion with dynamic obstacle
			                              // as late as possible
			}
		      
		      // delete the configuration of the obstacle
		      delete cfgObs;
		    }
		  j++;
		}
	    }
	}
      i++;
    }

  return res;
}


                             /*********************************************************/
                             /************** Spatio-temporal deformation **************/
                             /*********************************************************/

// init the deformation and check the temporal space between two consecutive states
void STDeformation::initDeform()
{
  double maxTime = 2*this->timeStep_;
  double maxDist = 2*this->posStep_;

  // clear the tables of forces, states types and nearest states
  if(*typeState != NULL) delete [] (*typeState);
  if(*tableNearState != NULL)delete [] (*tableNearState);
  if(*tableFInt != NULL)delete [] (*tableFInt);
  if(*tableFExt != NULL)delete [] (*tableFExt);

  // check if the space between two consecutive states of the trajectory are too large
  int i=0;
  while(i < this->traj_->getSize() - 1)
    {
      // get back the current and the next states
      StateR* curS  = (*this->traj_)[i];
      StateR* nextS = (*this->traj_)[i+1];

      double dx = nextS->x() - curS->x();
      double dy = nextS->y() - curS->y();
      double sqDist = dx*dx + dy*dy;
      
      // check if the temporal space is too large
      if(nextS->t() - curS->t() > maxTime || sqDist > maxDist)
	{
	  // add a new state between both current and next
	  StateR5D* midS = new StateR5D((curS->x() + nextS->x())/2, (curS->y() + nextS->y())/2, (curS->t() + nextS->t())/2, (curS->sx() + nextS->sx())/2, (curS->sy() + nextS->sy())/2);

	  // insert it in the trajectory
	  this->traj_->insert(i+1, midS);
	}

      i++;
    }

  // reinit the tables of states types and nearest states
  *typeState = new int [this->traj_->getSize()];
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      (*typeState)[i] = -1;
    }
  *tableNearState = (StateR**)(new StateR5D* [this->traj_->getSize()]);
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      (*tableNearState)[i] = NULL;
    }
  
  *tableFInt = new StateR5D [this->traj_->getSize()];
  
  *tableFExt = new StateR5D [this->traj_->getSize()];
}

// finish the deformation
int STDeformation::finishDeform(double time)
{
  this->oldTime_ = time;

  return 0;
}

// deform the current trajectory
int STDeformation::deform(double time, int curStep)
{
  double maxSpd = this->rb_->getVMax();
  double elapsedTime = time - this->oldTime_;
  
  // init the deformation
  this->initDeform();

  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      return -1;
    }
  else
    {
      //lVec.clear();

      // save the old trajectory
      PMSTraj oldTraj(*((PMSTraj*)(this->traj_)));

      // apply the repulsive forces from the obstacles
      for(int i=1; i<oldTraj.getSize(); i++)
	{
	  I = i;

	  if(i > curStep)
	    {
	      bool last;
	      if(i == oldTraj.getSize()-1) last = true;
	      else last = false;
	      
	      StateR5D fExtDyn;
	      StateR5D fExtDyn2;
	      
	      if((oldTraj)[i]->t() < time + (*reliabilityTime))
		{
		  // compute the external forces
		  fExtDyn2 = this->computeFExt(*((oldTraj)[i]), this->env_, time, last); // * elapsedTime;
		  (*tableFExt)[i] = fExtDyn2;

		  fExtDyn = fExtDyn2;

		  // check the connectivity of the curve
		  this->checkConnectivity(i, *(oldTraj[i]), *((*(this->traj_))[i-1]), *((*(this->traj_))[i]), fExtDyn);
		}

	      StateR5D fExtSt = this->computeStFExt(*((oldTraj)[i]), this->env_, last);

	      StateR5D forces = (fExtDyn) + (fExtSt);

	      if(fabs(forces.t()) > 2.0) 
		{
		  cout << "i : " << i << "  fExtDyn : "; fExtDyn.display(); cout << "  fExtDyn2 : "; fExtDyn2.display(); cout << "  forces : "; forces.display(); cout << "  forces.t() : " << forces.t() << endl;
		}

	      // apply the forces
	      ((*(this->traj_))[i])->setX( ((oldTraj)[i])->x() + (forces.x()) );
	      ((*(this->traj_))[i])->setY( ((oldTraj)[i])->y() + (forces.y()) );
	      ((*(this->traj_))[i])->setT( ((oldTraj)[i])->t() + (forces.t()) );
	      ((*(this->traj_))[i])->setSX( ((oldTraj)[i])->sx() + (forces.sx()) );
	      ((*(this->traj_))[i])->setSY( ((oldTraj)[i])->sy() + (forces.sy()) );
	    }
	}

      // adjust the speed of each state
      for(int i=1; i<this->traj_->getSize()-1; i++)
	{
	  // get back the i-1 th i th and i+1 th states
	  StateR* prev = ((*(this->traj_))[i-1]);
	  StateR* cur = ((*(this->traj_))[i]);
	  StateR* next = ((*(this->traj_))[i+1]);

	  // compute the distance between the i-1 th and i+1 th configurations
	  double dx = next->x() - prev->x();
	  double dy = next->y() - prev->y();

	  // compute the time between the i-1 th and i+1 th times
	  double dt = next->t() - prev->t();

	  // compute the new speeds of the current state
	  double sx = dx/dt;
	  double sy = dy/dt;

	  // set the speeds of the current state
	  cur->setSX(sx);
	  cur->setSY(sy);
	}

      // apply the internal forces
      for(int i=1; i<this->traj_->getSize()-1; i++)
	{
	  if(i > curStep)
	    {
	      StateR5D forces = this->computeFInt(*((*(this->traj_))[i-1]), *((*(this->traj_))[i]), *((*(this->traj_))[i+1]));
	      (*tableFInt)[i] = forces;
	      //StateR5D forces = this->computeFInt(*(oldTraj[i-1]), *(oldTraj[i]), *(oldTraj[i+1]));

	      ((*(this->traj_))[i])->setX( ((*(this->traj_))[i])->x() + 0.95*(forces.x()) );
	      ((*(this->traj_))[i])->setY( ((*(this->traj_))[i])->y() + 0.95*(forces.y()) );
	      ((*(this->traj_))[i])->setT( ((*(this->traj_))[i])->t() + 0.95*(forces.t()) );
	      ((*(this->traj_))[i])->setSX( ((*(this->traj_))[i])->sx() + 0.95*(forces.sx()) );
	      ((*(this->traj_))[i])->setSY( ((*(this->traj_))[i])->sy() + 0.95*(forces.sy()) );

// 	      if((dispForces && i == indDispForces) || (i != this->traj_->getSize()-1 && i == curStep + 1) || (dispAllForces && i%5 == 0))
// 		{
// 		  FXVec3d v1(((*(this->traj_))[i])->x(), ((*(this->traj_))[i])->y(), ((*(this->traj_))[i])->t());
// 		  FXVec3d v2(((*(this->traj_))[i])->x() + 40 * forces.x(), ((*(this->traj_))[i])->y() + 40 * forces.y(), ((*(this->traj_))[i])->t() + 40 * forces.t());
// 		  lVec.append(v1);
// 		  lVec.append(v2);
// 		}
	      
	      //cout << "oldTraj[" << i << "] = "; (oldTraj[i])->display(); cout << "\t";
	      //cout << "traj[" << i << "] = "; ((*(this->traj_))[i])->display(); cout << endl;
	    }
	}

      // test if some states are in collision
      if(this->env_ != NULL && this->env_->getNbObs() != 0)
	{
	  for(int i=1; i<this->traj_->getSize()-1; i++)
	    {
	      StateR* curS = (*(this->traj_))[i];

	      if((*(this->traj_))[i]->t() > time)
		{
		  bool trouveCol = false;
		  int j=0;

		  while(!trouveCol && j < this->env_->getNbObs())
		    {
		      // get back the previewed configuration of the obstacle
		      ConfigR* cfgObs = this->env_->getObs(j)->getFutureCfg(curS->t());

		      if(cfgObs != NULL)
			{
			  // get back the radius of the obstacle
			  double radObs = ((ObsDisk2D*)(this->env_->getObs(j)))->getRadius();
			  
			  // compute the distance from the current state to the j-th obstacle
			  double dx = curS->x() - cfgObs->x();
			  double dy = curS->y() - cfgObs->y();
			  double sqDist = dx*dx + dy*dy;
			  
			  // test if the current state is in collision with the j-th obstacle
			  if(sqDist < radObs*radObs)
			    {
			      trouveCol = true;
			    }
			  else j++;

			  delete cfgObs;
			}
		      else j++;
		    }
		  
		  // test if a collision has been found
		  if(trouveCol)
		    {
		      (*typeState)[i] = -1;
		    }
		}
	    }
	}

      //if(this->path_ != NULL) delete this->path_;
      //this->path_ = this->traj_->getPath();
    }

  // finish the deformation
  return this->finishDeform(time);
}


                             /*********************************************************/
                             /******************* Display functions *******************/
                             /*********************************************************/

// Display the predictional horizon profile
void STDeformation::displayPredHorizonProfile(double dt) const
{
  //double predHorizon = 3.0;
  double uncerHorizon = (*reliabilityTime) - (*predHorizon_);

  // initialize both time and reliability vectors
  vector <double> vecTime;
  vector <double> vecReliability;
  
  // compute the number of points to display
  int nbP = ((int)((1.2 * (*reliabilityTime)) / dt)) + 1; 

  cout << "prediction horizon : " << (*predHorizon_) << endl;
  cout << "reliabilityTime : " << (*reliabilityTime) << endl;

  // compute all points
  for(int i=0; i<nbP; i++)
    {
      double time = i*dt;
      
      // compute the coefficient due to the reliability time
      double coefRT = this->getCoefReliability(time);
      
      // add the new values in respective vectors
      vecTime.push_back(time);
      vecReliability.push_back(coefRT);
    }

  // display the graph in a gnuplot window
  Gnuplot gCoefRT;
  gCoefRT.reset_plot();
  gCoefRT.set_style("lines").plot_xy(vecTime,vecReliability,"Reliability Coef of the Previsional Model");

  wait_for_key();
}
