#include "STNHDeformation.hpp"

#include "SmoothTrajGen.hpp"
#include "HermiteTrajGen.hpp"
#include "MathFunc.hpp"
#include "ControlR.hpp"
#include "Utils.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#define NBRANDCTRL 300
#define fieldOfView 0.25*M_PI


/********************************************************************************************
 ************************************** class ParamCtrl *************************************
 ********************************************************************************************/

/************** Constructors and destructors *************/

// default constructor
ParamCtrl::ParamCtrl()
{
  this->B  = 0;
  this->C  = 0;
  this->D  = 0;
  this->B2 = 0;
  this->C2 = 0;
  this->D2 = 0;  
}

// copy constructor
ParamCtrl::ParamCtrl(const ParamCtrl& p)
{
  this->B  = p.B;
  this->C  = p.C;
  this->D  = p.D;
  this->B2 = p.B2;
  this->C2 = p.C2;
  this->D2 = p.D2;
}

// init constructor
ParamCtrl::ParamCtrl(double b,double c,double d,double b2,double c2,double d2)
{
  this->B  = b;
  this->C  = c;
  this->D  = d;
  this->B2 = b2;
  this->C2 = c2;
  this->D2 = d2;
}

// destructor
ParamCtrl::~ParamCtrl()
{
  // nothing to do
}


/********************************************************************************************
 *********************************** class STNHDeformation **********************************
 ********************************************************************************************/

/************** Constructors and destructors *************/

// default constructor
STNHDeformation::STNHDeformation() : STDeformation()
{
  this->listCtrl_ = NULL;
  this->feaCtrl_ = NULL;
  this->parallelExec = false;
  this->nbThread = 5;
  this->usePreviousParam_ = true;
  this->useResampling_ = true;
}

// init constructor
STNHDeformation::STNHDeformation(Trajectory *traj,Environment *env,Robot *rb,int** typeS,StateR*** tns,StateR** tfi,StateR** tfe) : STDeformation(traj,env,rb,typeS,tns,tfi,tfe)
{
  this->listCtrl_ = NULL;
  this->feaCtrl_ = NULL;
  this->parallelExec = false;
  this->nbThread = 5;
  this->usePreviousParam_ = true;
  this->useResampling_ = true;
}

// destructor
STNHDeformation::~STNHDeformation()
{
}

  
/*** Computing of the forces applied on the trajectory ***/

// computing of the internal forces
StateS6D STNHDeformation::computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last, Vec_DP** param, int *fea)
{
  CLTrajGen carModel;
  carModel.L_ = rb_->getLength();

//   cout << "carModel cree" << endl;

  Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th(); 
                x0[3] = prev.s()*tan(prev.ph())/rb_->getLength(); x0[4] = prev.s(); x0[5] = 0.0;
  Vec_DP xf(6); xf[0] = next.x(); xf[1] = next.y(); xf[2] = next.th(); 
                xf[3] = next.s()*tan(next.ph())/rb_->getLength(); xf[4] = next.s(); xf[5] = next.t() - prev.t();
  // check the temporal consistency between both nodes
  double dx = next.x() - prev.x();
  double dy = next.y() - prev.y();
  double dt = next.t() - prev.t();
  if(dt < 0)// || rb_->getVMax()*rb_->getVMax()*dt*dt < dx*dx+dy*dy)
    {
      cout << "@dt : " << dt;
      xf[5] = sqrt(dx*dx+dy*dy) / rb_->getVMax();
      cout << "  dx : " << dx << "  dy : " << dy << "  vmax : " << rb_->getVMax() << "  xf[5] : " << xf[5] << endl;
    }

  Vec_DP constraints(6); 
  constraints[0] = 0.5;
  constraints[1] = rb_->getAMax();
  constraints[2] = rb_->getVMax();
  constraints[3] = rb_->getLength();
  constraints[4] = rb_->getMaxSteering();
  constraints[5] = rb_->getMaxSteerSpd();

//   cout << "rb_->getLength() : " << rb_->getLength() << endl;
//   if(*param != NULL)
//     cout << "ictrl : (" << (**param)[0] << "," << (**param)[1] << "," << (**param)[2] << "," << (**param)[3] << "," 
// 	 << (**param)[4] << "," << (**param)[5] << "," << (**param)[6] << ")" << endl;
//   else cout << "ictrl null" << endl;

//   cout << "contraintes initialisees" << endl;

//   if(*param != NULL) cout << "param non nul avant" << "\t";
//   else cout << "param nul avant" << "\t";
  int res = carModel.ComputeTrajectory(x0,xf,constraints,param);
//   if(*param != NULL) cout << "param non nul apres" << endl;
//   else cout << "param nul apres" << endl;

  //wait_for_key();

//   cout << "trajectoire generee" << endl;

  double tSur2 = 0.5*(next.t() - prev.t());

  // check if the trajectory generator has found a solution
  if(res == 0 || (carModel.finalResidual_ < 20.0 && carModel.finalErrTH_ < 0.5))
    {
      if(res == 0) 
	{
	  *fea = T_FEA;
// 	  cout << "a near solution has been found (residual : " << carModel.finalResidual_ << ")" << endl;
	}
      else
	{
	  *fea = T_FAR;
// 	  cout << "a far solution has been found (residual : " << carModel.finalResidual_ << ")" << endl;
	}

//       cout << "fctrl : (" << (**param)[0] << "," << (**param)[1] << "," << (**param)[2] << "," << (**param)[3] << "," 
// 	   << (**param)[4] << "," << (**param)[5] << "," << (**param)[6] << ")" << endl;

//       cout << "midPoint : (" << carModel.funcX(tSur2) << "," << carModel.funcY(tSur2) << "," << 
// 	carModel.funcTheta(tSur2) << "," << carModel.funcVel(tSur2) << "," << carModel.funcAVel(tSur2) << ")" << endl;
//       cout << "finalPoint : (" << carModel.funcX(dt) << "," << carModel.funcY(dt) << "," << 
// 	carModel.funcTheta(dt) << "," << carModel.funcVel(dt) << "," << carModel.funcAVel(dt) << ")" << endl;

      double x = carModel.funcX(tSur2);
      double y = carModel.funcY(tSur2); 
      double th = carModel.funcTheta(tSur2);
      double v = carModel.funcVel(tSur2);
      double om = carModel.funcAVel(tSur2);
      double ph;
      if(v != 0.0) ph = atan(rb_->getLength() * om / v);
      else
	{
	  cout << "STNHDeformation::computeNHFInt : Warning. vres = 0. Cannot compute ph..." << endl;
	  exit(1);
	}

      StateS6D res(x - cur.x(),
		   y - cur.y(),
		   th - cur.th(),
		   v - cur.s(),
		   ph - cur.ph(),
		   prev.t() + tSur2 - cur.t());

//       cout << "avant retour" << endl;
      return res;
    }
  else
    {
      *fea = T_UNFEA;
//       cout << "no solution has been found (residual : " << carModel.finalResidual_ << ")" << endl;
      
      // destroy the bad parameters vector
      *param = NULL;
      
      // create a deceleration control
      ControlTimeAAS decCtrl(-0.5 * rb_->getAMax(),0.0,tSur2);
      
      // compute the state reached by the deceleration control
      //ModelFInt carModel2;
      CLTrajGen carModel2;
      carModel2.L_ = rb_->getLength();
      carModel2.setInitialGuess(decCtrl.getAS(),0.0,0.0,decCtrl.getAcc(),0.0,0.0,decCtrl.getTime());
      Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th(); 
                    x0[3] = prev.s()*tan(prev.ph())/rb_->getLength(); x0[4] = prev.s(); x0[5] = 0.0;
      Vec_DP xf(6); xf[0] = prev.x(); xf[1] = prev.y(); xf[2] = prev.th(); 
                    xf[3] = prev.s()*tan(prev.ph())/rb_->getLength(); xf[4] = prev.s(); xf[5] = 0.0;
      Vec_DP constraints(6); constraints[0] = 0.5; constraints[1] = rb_->getAMax(); constraints[2] = rb_->getVMax(); 
      constraints[3] = rb_->getLength(); constraints[4] = rb_->getMaxSteering(); constraints[5] = rb_->getMaxSteerSpd();
      
      carModel2.setInitialParam(x0,xf,constraints);
      carModel2.ForwardSolution();
      
//       const GLfloat redMaterial[]={1.,0.,0.,1.};
//       glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
//       glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
//       carModel.toGL(prev.t());
      
      //carModel2.displayProfiles();
      
      double x = carModel2.vecState_xsim[0];
      double y = carModel2.vecState_xsim[1];
      double th = carModel2.vecState_xsim[2];
      double v = carModel2.vecState_xsim[4];
      double om = carModel2.vecState_xsim[3];
      double ph = atan2(rb_->getLength()*om,v);
      double t = prev.t() + decCtrl.getTime();
      
      const GLfloat redMaterial[]={1.,0.,0.,1.};
      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
      glPointSize(30.0); 
            
      StateS6D sres(x - cur.x(),
		    y - cur.y(),
		    th - cur.th(),
		    v - cur.s(),
		    ph - cur.ph(),
		    t - cur.t());
      return sres;
    }
}

// computing of the internal forces
void STNHDeformation::computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int* fea, StateS6D& fres)
{
//   cout << "entree ds computeNHFInt" << endl;
//   cout << "prev : "; prev.display(); cout << endl;
//   cout << "cur : "; cur.display(); cout << endl;
//   cout << "next : "; next.display(); cout << endl;
  fres = this->computeNHFInt(prev,cur,next,last,param,fea);
}

// // computing of the internal forces
// StateS6D STNHDeformation::computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next, const StateR& next5, int &seg5, Trajectory6D* traj, int I)
// {
// //   cout << "///////////////////////////////////// I : " << I << " /////////////////////////////////////" << endl;

//   bool disconnected1     = false;
//   disco_spacc_t discoSA1 = NO_DISCO_SPACC;
//   disco_curv_t discoC1   = NO_DISCO_CURV;

//   // compute the left rotation center from n_{-}
//   Vec2D lrc = ((StateS6D)prev).maxLeftRotationCenter(rb_->getMaxCurvature());
//   // compute the distance between lrc and next
//   double dxLrcNext = lrc.getVal(0) - next.x();
//   double dyLrcNext = lrc.getVal(1) - next.y();
//   double sqDistLrcNext = dxLrcNext*dxLrcNext + dyLrcNext*dyLrcNext;

//   // compute the right rotation center from n_{-}
//   Vec2D rrc = ((StateS6D)prev).maxRightRotationCenter(rb_->getMaxCurvature());
//   // compute the distance between lrc and next
//   double dxRrcNext = rrc.getVal(0) - next.x();
//   double dyRrcNext = rrc.getVal(1) - next.y();
//   double sqDistRrcNext = dxRrcNext*dxRrcNext + dyRrcNext*dyRrcNext;

//   // compute the minimal curvature radius
//   double minCurvRad = 1.0 / rb_->getMaxCurvature();

//   // compute the distance n_{i} and n_{+}
//   double dxPrevNext = next.x() - prev.x();
//   double dyPrevNext = next.y() - prev.y();
//   double sqDistPrevNext = sqrt(dxPrevNext*dxPrevNext + dyPrevNext*dyPrevNext);

//   // compute the minimal and maximal length travelable by prev during a time dt
//   double dt = next.t() - prev.t();
//   double maxTravLength = MathFunc::maxTravDist(prev.s(),rb_->getVMax(),rb_->getAMax(),dt);
//   double minTravLength = MathFunc::minTravDist(prev.s(),0,-rb_->getAMax(),dt);

//   // test all straightforward cases where n_{i} and n_{+} could be disconnected
//   if(sqDistLrcNext < minCurvRad*minCurvRad)
//     {
//       disconnected1 = true;
//       //leftTurnDisc = true;
//       discoC1 = DISCO_LEFT_CURV;
// //       cout << "Straight forward disco left curv   DistLrcNext : " << sqrt(sqDistLrcNext) << "   minCurvRad : " << minCurvRad << endl;
//     }
//   else if(sqDistRrcNext < minCurvRad*minCurvRad)
//     {
//       disconnected1 = true;
//       //rightTurnDisc = true;
//       discoC1 = DISCO_RIGHT_CURV;
// //       cout << "Straight forward disco right curv" << endl;
//     }

// //   if(sqDistPrevNext < minTravLength*minTravLength)
// //     {
// //       disconnected1 = true;
// //       minSpeedAccDisc = true;
// //     }
// //   else
// //   if(sqDistPrevNext > maxTravLength*maxTravLength)
// //     {
// //       disconnected1 = true;
// //       //maxSpeedAccDisc = true;
// //       discoSA1 = DISCO_MAX_SPACC;
// //     }
  
//   if(!disconnected1)
//     {
//       // try to compute a hermite interpolation between prev and next
//       StateS6D forceH1 = this->computeNHFIntHe(prev,cur,next,discoSA1,discoC1,traj,I);

// //       if(discoC1 == DISCO_LEFT_CURV) cout << "Hermite1 disco left curv" << endl;
// //       if(discoC1 == DISCO_RIGHT_CURV) cout << "Hermite1 disco right curv" << endl;

//       if(discoSA1 == NO_DISCO_SPACC && discoC1 == NO_DISCO_CURV)
// 	{
// // 	  double distForce = sqrt(forceH1.x()*forceH1.x() + forceH1.y()*forceH1.y());
// // 	  if(distForce > 100.0)
// // 	    {
// // 	      cout << "I : " << I << "  HermiteTraj1 -> distForce : " << distForce << endl;
// // 	      cout << "  --> prev : "; prev.display(); cout << endl;
// // 	      cout << "  --> next : "; next.display(); cout << endl;
// // 	    }
// 	  return forceH1;
// 	}
//       else 
// 	{
// 	  disconnected1 = true;
// 	}
//     }

//   // if we cant reach the goal configuration, ensure to reach at least the goal position
//   if(disconnected1 && I == traj->getSize() -2)
//     {
//       // get back previous position
//       Vec2D prevPos(traj->at(traj->getSize()-3)->x(),traj->at(traj->getSize()-3)->y());
      
//       // get back last position
//       Vec2D lastPos(traj->at(traj->getSize()-1)->x(),traj->at(traj->getSize()-1)->y());

//       // compute the vector \overrightarrow{prevPos lastPos}
//       Vec2D pplp(lastPos.getVal(0) - prevPos.getVal(0), lastPos.getVal(1) - prevPos.getVal(1));
//       // compute the associated direction
//       double dirPPLP = atan2(pplp.getVal(1),pplp.getVal(0));

//       // compute the difference between dirPPLP and prevDir
//       double diffAngle = computeMinDiffAngle(dirPPLP, traj->at(traj->getSize()-3)->th());

//       // compute the symetric direction from the previous dir to the vector \overrightarrow{prevPos lastPos}
//       double newLastDir = normalizedAngle(dirPPLP + diffAngle);

//       // set the last orientation
//       traj->at(traj->getSize()-1)->setTH(newLastDir);

// //       cout << "prevDir : " << traj->at(traj->getSize()-3)->th();
// //       cout << "vecDir : " << dirPPLP << endl;
// //       cout << "newLastDir : " << newLastDir << endl;

//       // try to compute a new Hermite trajectory 
//       disco_spacc_t discoSA2 = NO_DISCO_SPACC;
//       disco_curv_t discoC2   = NO_DISCO_CURV;

//       StateS6D forceH2 = this->computeNHFIntHe(prev,cur,*(traj->at(traj->getSize()-1)),discoSA2,discoC2,traj,I);
      
//       if(discoSA2 == NO_DISCO_SPACC && discoC2 == NO_DISCO_CURV)
// 	{
// 	  return forceH2;
// 	}
//       else
// 	{
// 	  return StateS6D();
// 	}
//     }

//   bool disconnected5     = false;
//   disco_spacc_t discoSA5 = NO_DISCO_SPACC;
//   disco_curv_t discoC5   = NO_DISCO_CURV;
//   if(disconnected1)
//     {
//       disconnected5 = true;
//       bool trouve = false;
//       int i=I+5;
      
//       while(!trouve && i<traj->getSize())
// 	{
// 	  bool isInFOV = isInTheFieldOfView(Vec2D(prev.x(),prev.y()),
// 					    prev.th(),
// 					    Vec2D(traj->at(i)->x(),traj->at(i)->y()),
// 					    traj->at(i)->th(),
// 					    fieldOfView);

// 	  if(isInFOV) trouve = true;
// 	  else i++;
// 	}

//       if(!trouve) disconnected5 = true;
//       else
// 	{
// 	  StateS6D nextTry = *(traj->at(i));
	  
// 	  // try to compute a hermite interpolation between prev and nextTry
// 	  StateS6D forceH5 = this->computeNHFIntHe(prev,cur,nextTry,discoSA5,discoC5,traj,I);
// 	  if(discoSA5 == NO_DISCO_SPACC && discoC5 == NO_DISCO_CURV)
// 	    {
// 	      seg5 = i-I-1;

// // 	      double distForce = sqrt(forceH5.x()*forceH5.x() + forceH5.y()*forceH5.y());
// // 	      if(distForce > 100.0)
// // 		{
// // 		  cout << "I : " << I << "  HermiteTraj5 -> distForce : " << distForce << endl;
// // 		  cout << "  --> prev : "; prev.display(); cout << endl;
// // 		  cout << "  --> nextTry : " << i-I-1 << " : "; next.display(); cout << endl;
// // 		}
// 	      return forceH5;
// 	    }
// 	  else disconnected5 = true;
// 	}
//     }

//   // case where n_{i} and n_{+} are disconnected : 
//   // we use the integration by Runge-Kutta
  
//   if(disconnected5)
// //    if(disconnected1)
//     {
      
//       //StateS6D forceRK = this->computeNHFIntRK(prev, cur, next, discoSA1, discoC1);
//       StateS6D forceRK;

// //       double distForce = sqrt(forceRK.x()*forceRK.x() + forceRK.y()*forceRK.y());
// //       if(distForce > 100.0)
// // 	{
// // 	  cout << "I : " << I << "  RungeKuttaTraj -> distForce : " << distForce << endl;
// // 	  cout << "  --> prev : "; prev.display(); cout << endl;
// // 	  cout << "  --> next : "; next.display(); cout << endl;
// // 	}
//       return forceRK;
//     }
//   else
//     {
//       cout << "Unknow case" << endl;
//     }
// }

// // computing of the internal forces by hermite interpolation
// StateS6D STNHDeformation::computeNHFIntHe(const StateR& prev, const StateR& cur, const StateR& next, disco_spacc_t& discoSA, disco_curv_t& discoC, Trajectory6D* traj,int I)
// {
//   StateS6D *ptrPrev = new StateS6D(prev);
//   StateS6D *ptrNext = new StateS6D(next);

//   //SmoothTrajGen htg(this->rb_, NULL, ptrPrev, ptrNext);
//   HermiteTrajGen htg(this->rb_, NULL, ptrPrev, ptrNext);
//   int resHTG = htg.computeTraj();

//   delete ptrPrev;
//   delete ptrNext;

//   switch(resHTG)
//     {
//     case NO_TRAJ_FEA : 
//       {
// 	cout << "NO_TRAJ_FEA" << endl;

// 	// compute an intermediate state between prev and next
// 	StateS6D sres;
// 	bool resHItdS = HermiteTrajGen::hermiteIntmdState(((StateS6D)prev),((StateS6D)next),sres);
// 	if(resHItdS)
// 	  {
// 	    cout << "NO_TRAJ_FEA1" << endl;
// 	    // insert the node sres in the trajectory
// 	    //traj->insert(I+1, new StateS6D(sres));
// 	    cout << "NO_TRAJ_FEA2" << endl;

// 	    // try to compute a hermite interpolation between prev and the new intermediate state
// 	    return this->computeNHFIntHe(prev, cur, sres, discoSA, discoC, traj, I);
// 	  }
// 	else
// 	  {
// 	    cout << "NO_TRAJ_FEA3" << endl;
// 	    // search on which side the turn has disconnected
// 	    ConfigR3D prevCfg(prev.x(),prev.y(),prev.th());
// 	    Vec2D nextPos(next.x(),next.y());
// 	    leftright_t side = prevCfg.wichSideToTurn(nextPos);
// 	    cout << "NO_TRAJ_FEA4" << endl;

// 	    // conclude on which side the turn has disconnected
// 	    if(side == LEFT) discoC = DISCO_LEFT_CURV; //leftTurnDisc = true;
// 	    else discoC = DISCO_RIGHT_CURV; //rightTurnDisc = true;
// 	    discoSA = NO_DISCO_SPACC;
// 	    cout << "NO_TRAJ_FEA5" << endl;

// 	    return StateS6D();	    
// 	  }
// 	break;
//       }
//     case CURV_OVR : 
//     case DCURV_OVR : 
//       {
// 	//disconnected = true;
// 	// search on which side the turn has disconnected
// 	ConfigR3D prevCfg(prev.x(),prev.y(),prev.th());
// 	Vec2D nextPos(next.x(),next.y());
// 	leftright_t side = prevCfg.wichSideToTurn(nextPos);

// 	// conclude on which side the turn has disconnected
// 	if(side == LEFT) discoC = DISCO_LEFT_CURV; //leftTurnDisc = true;
// 	else discoC = DISCO_RIGHT_CURV; //rightTurnDisc = true;
// 	discoSA = NO_DISCO_SPACC;

// 	return StateS6D();
// 	break;
//       }
//     case TRAJ_FEA : 
//       {
// 	// get back the trajectory
// 	Trajectory* traj = htg.getTraj();
// 	//vecTraj->append(new Trajectory6D(*((Trajectory6D*)traj)));

// 	discoC = NO_DISCO_CURV;
// 	discoSA = NO_DISCO_SPACC;

// 	// compute the middle state of the trajectory
// 	StateR* midState = traj->getMiddleState();

// 	StateS6D force(midState->x() - cur.x(),
// 		       midState->y() - cur.y(),
// 		       midState->th() - cur.th(),
// 		       midState->s() - cur.s(),
// 		       midState->ph() - cur.ph(),
// 		       midState->t() - cur.t());
	
// 	// check if the force is not to great
// 	// compute the distance between prev and next
// 	double dx = next.x() - prev.x();
// 	double dy = next.y() - prev.y();
// 	double sqDistPN = dx*dx + dy*dy;
	
// 	// compute the distance traveled by the force
// 	double sqDistForce = force.x()*force.x() + force.y()*force.y();

// 	if(sqDistForce < 1.0 * sqDistPN)
// 	  return force;
// 	else
// 	  {
// 	    cout << "too great force applied   DistForce : " << sqrt(sqDistForce) << "   DistPN : " << sqrt(sqDistPN) << endl;
// 	    if(midState->ph() > 0) discoC = DISCO_LEFT_CURV;
// 	    else discoC = DISCO_RIGHT_CURV;
// 	    return StateS6D();
// 	  }
// 	break;
//       }
//     }  
// }

// // computing of the internal forces by Runge-Kutta
// StateS6D STNHDeformation::computeNHFIntRK(const StateR& prev, const StateR& cur, const StateR& next, disco_spacc_t& discoSA, disco_curv_t& discoC)
// {
//   double midTime = 0.5*(next.t() - prev.t());
  
//   // choose the control to apply in function of discontinuities

// //   double ctAcc;
// //   if(discoSA == NO_DISCO_SPACC) ctAcc = 0;
// //   else if(discoSA == DISCO_MIN_SPACC) ctAcc = -rb_->getAMax();
// //   else ctAcc = rb_->getAMax();

// //   double ctDC;
// //   if(discoC == NO_DISCO_CURV) ctDC = 0;
// //   else if(discoC == DISCO_LEFT_CURV) ctDC = rb_->getMaxCurvSpd();
// //   else ctDC = -rb_->getMaxCurvSpd();

// //   ControlTimeR ctrlT(ctAcc,ctDC,midTime);

//   int nbCtrl = NBRANDCTRL;
//   // allocate a table for states computed
//   StateS6D* tabS = new StateS6D [nbCtrl];
//   double sqMinDist = 0;
//   int near = -1;
// //   ControlTimeR nearCtrl;
//   Robot rb(*rb_);
//   //rb.display();

//   // compute several random controls
//   for(int i=0; i<nbCtrl; i++)
//     {
//       ControlTimeR randCtrl = ControlTimeR::randControl(prev,rb,0.5);
//       // compute the state reached by the previous state under the control choosen
//       tabS[i] = randCtrl.stateReachedByV2(((StateS6D)prev));

//       double dx = tabS[i].x() - cur.x();
//       double dy = tabS[i].y() - cur.y();
//       double sqDist = dx*dx + dy*dy;
//       if(near == -1 || sqDist < sqMinDist)
// 	{
// 	  near = i;
// 	  sqMinDist = sqDist;
// // 	  nearCtrl = randCtrl;
// 	}
//     }
  
//   StateS6D nearState = tabS[near];
// //   if(fabs(nearState.x() - cur.x()) > 10.0 || fabs(nearState.y() - cur.y()) > 10.0)
// //     {
// //       cout << "Control : ("; nearCtrl.display(); cout << endl;
// //       cout << "prev : "; prev.display(); cout << endl;
// //       cout << "nearState : "; nearState.display(); cout << endl;
// //       cout << "cur : "; cur.display(); cout << endl;
// //     }

//   //delete tabS;

//   // compute the resulting force to apply
//   return StateS6D(nearState.x() - cur.x(),
// 		  nearState.y() - cur.y(),
// 		  nearState.th() - cur.th(),
// 		  nearState.s() - cur.s(),
// 		  nearState.ph() - cur.ph(),
// 		  nearState.t() - cur.t());
// }


// // computing of the internal forces
// StateS6D STNHDeformation::computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next)
// {
//   return StateS6D((prev.x() + next.x())*0.5 - cur.x(),
// 		  (prev.y() + next.y())*0.5 - cur.y(),
// 		  (prev.th() + next.th())*0.5 - cur.th(),
// 		  (prev.s() + next.s())*0.5 - cur.s(),
// 		  (prev.kp() + next.kp())*0.5 - cur.kp(),
// 		  (prev.t() + next.t())*0.5 - cur.t()) * (*K_INT);

//   NHReachabilityCone nhrc(this->rb_);
//   ConfigR3D prevCfg(prev.x(), prev.y(), prev.th());
//   ConfigR3D nextCfg(next.x(), next.y(), next.th());
//   int resCptgRC = nhrc.computeTDFromTo(prevCfg,nextCfg);

//   // compute the internal force
//   StateS6D forceInt;
//   switch(resCptgRC)
//     {
//     case -6 : 
//       break;
//     case -5 : 
//       break;
//     case -4 : 
//       break;
//     case -3 : 
//       break;
//     case -2 : 
//       break;
//     case -1 : 
//       break;
//     case 0 : 
//       break;
//     case 2 : 
//       {
// 	// case where the nhrc has been correctly computed
// 	ConfigR3D cBary = nhrc.getBaryCfg();
// 	forceInt = StateS6D(cur.x() - cBary.x(),
// 			    cur.y() - cBary.y(), 
// 			    cur.th() - cBary.th(),
// 			    0,
// 			    0,
// 			    cur.t() - 0.5*(next.t() - prev.t())
// 			    ) * (*K_INT);
// 	break;
//       }
//     default : 
//       // this case should not appear
//       exit(-1);
//       break;
//     }
//   return forceInt;
// }

// compute the external spatial and temporary force due to the obstacles
StateS6D STNHDeformation::computeNHF6Ext(const StateR& cur, Environment* env, double time, bool last)
{
  // verify if the current state is not too far from the current time to be under external forces
  if(cur.t() - time > (*reliabilityTime))
    {
      StateS6D force;
      if(fabs(force.t()) > 10.0) 
	{ cout << "force0 : "; force.display(); cout << endl; }
      return force;
    }
  
  // init needed variables
  double Dmax = *spMed;
  double Tmax = *tmpMed;
  double MDmax = Dmax * (*alphaMax);
  double MTmax = Tmax * (*betaMax);
  
  // compute spatial coefficient Alpha (for the metric)
  double Alpha = *alphaMin + ((*reliabilityTime) - cur.t() + time) * ((*alphaMax - *alphaMin) / (*reliabilityTime));
  
  // compute spatial coefficient Alpha (for the metric)
  double Beta = *betaMax - (cur.t() - time) * ((*betaMax - *betaMin) / (*reliabilityTime));
  //double Beta = *betaMin + (cur.t() - time) * ((*betaMax - *betaMin) / (*reliabilityTime));

  // search for the nearest obstacle
  int nearestObs = -1;
  double sqMminDist = 0;
  StateS6D nearestState; 
  
  if(this->modeDeform == SPATONLY_DEFORM)
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  // get back the current obstacle
	  ObsDisk2D* curObs = (ObsDisk2D*) this->env_->getObs(i);
	  
	  double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();
	  
	  // get back the last position of the obstacle
	  ConfigR* configObs = NULL;
	  configObs = curObs->getLastCfg();
	  
	  if(configObs != NULL)
	    {
	      Vec3D curRbNearObs(configObs->x() - cur.x(), configObs->y() - cur.y(), 0);

	      double sqMminDistObs = (Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1]);
	      
	      // check if this obstacle is the nearest obstacle from the current robot's control point
	      if(nearestObs == -1 || sqMminDistObs < sqMminDist)
		{
		  sqMminDist = sqMminDistObs;
		  nearestObs = i;
		  nearestState = StateS6D(configObs->x(),configObs->y(),0,0,0,cur.t());
		}
	    }

	  // deletions
	  if(configObs != NULL) delete configObs;
	}
    }
  else
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  // get back the current obstacle
	  ObsDisk2D* curObs = (ObsDisk2D*) this->env_->getObs(i);
	  
	  double radObs = ((ObsDisk2D*)(this->env_->getObs(i)))->getRadius();  
	  //double MDTmax = sqrt(Alpha*Alpha*(Dmax + radObs + rb_->getRadius())*(Dmax + radObs + rb_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  double sqMDTmax = (Alpha*Alpha*(Dmax + radObs + rb_->getRadius())*(Dmax + radObs + rb_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  //cout << "MDTmax : " << MDTmax << endl;
	  
	  // get back the last position of the obstacle
	  ConfigR* configObs = NULL;
	  configObs = curObs->getLastCfg();
	  
	  // compute the coordinates of the obstacle's last control point in the metric frame
	  Vec3D MconfigObs(Alpha * configObs->x(), Alpha * configObs->y(), Beta * curObs->getLastTime());
	  
	  // compute the coordinates of the robot's control point in the metric frame
	  Vec3D MconfigRb(Alpha * cur.x(), Alpha * cur.y(), Beta * cur.t());
	  
	  // compute the robot's speed vector in the metric frame
	  Vec3D MspeedObs(Alpha * curObs->getLastSpdX(), Alpha * curObs->getLastSpdY(), Beta);
	  
	  // compute the vector configRobot - configObs with the metric
	  Vec3D MORi (MconfigRb[0] - MconfigObs[0], MconfigRb[1] - MconfigObs[1], MconfigRb[2] - MconfigObs[2]);
	  
	  // compute the norm of MspeedRb
	  double MnormSObs = sqrt(MspeedObs[0]*MspeedObs[0] + MspeedObs[1]*MspeedObs[1] + MspeedObs[2]*MspeedObs[2]);
	  
	  // compute the normalized vector NMspeedRb
	  Vec3D NMspeedObs;
	  if(MnormSObs != 0)
	    {
	      NMspeedObs[0] = MspeedObs[0]/MnormSObs;
	      NMspeedObs[1] = MspeedObs[1]/MnormSObs;
	      NMspeedObs[2] = MspeedObs[2]/MnormSObs;
	    }
	  
	  // compute the scalar product between MORi and nSObs
	  double pdScal = MORi[0] * NMspeedObs[0] + MORi[1] * NMspeedObs[1] + MORi[2] * NMspeedObs[2];
	  
	  //compute the vector OP (obstacle's nearest control point - obstacle's init control point) in the metric frame
	  Vec3D NOP(pdScal * NMspeedObs[0], pdScal * NMspeedObs[1], pdScal * NMspeedObs[2]);
	  
	  // compute the vector OP in the normal frame
	  Vec3D OP(NOP[0]/Alpha, NOP[1]/Alpha, NOP[2]/Beta);
	  
	  // compute the obstacle's nearest position
	  StateS6D nearestStateCurObs(configObs->x() + OP[0], configObs->y() + OP[1], 0, 0, 0, curObs->getLastTime() + OP[2]);
	  
	  // compute the vector nearestStateCurObs - curRb
	  //Vec3D curRbNearObs(OP[0] - cur.x(), OP[1] - cur.y(), OP[2] - cur.t());
	  Vec3D curRbNearObs(nearestStateCurObs.x() - cur.x(), nearestStateCurObs.y() - cur.y(), nearestStateCurObs.t() - cur.t());

	  double sqMminDistObs = (Alpha*Alpha*curRbNearObs[0]*curRbNearObs[0] + Alpha*Alpha*curRbNearObs[1]*curRbNearObs[1] + Beta*Beta*curRbNearObs[2]*curRbNearObs[2]);
	  
	  // check if this obstacle is the nearest obstacle from the current robot's control point
	  if((sqMminDistObs < sqMDTmax) && (nearestObs == -1 || sqMminDistObs < sqMminDist))
	    {
	      sqMminDist = sqMminDistObs;
	      nearestObs = i;
	      nearestState = nearestStateCurObs;
	    }
	  
	  // deletions
	  if(configObs != NULL) delete configObs;
	}
    }
  
  // compute the force to apply to the current state
  if(nearestObs != -1)
    {
      // add the nearest state to the corresponding display table entry
      if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
      (*tableNearState)[I] = new StateS6D(nearestState);

      // compute the coefficient due to the reliability time
      double sigma = sqrt(-0.5 * (*reliabilityTime) * (*reliabilityTime) / log(*epsilonRT));
      double coefRT = exp(-0.5 * (time - cur.t())/sigma * (time - cur.t())/sigma );
      //double coefRT = exp((time - cur.t())*(time - cur.t())*log(*epsilonRT)/((*reliabilityTime)*(*reliabilityTime)));

      // compute the direction of the force
      double dx = (cur.x() - nearestState.x());
      double dy = (cur.y() - nearestState.y());
      double dt = (cur.t() - nearestState.t());
      double sqdd = dx*dx + dy*dy;
      double length = sqrt(dx*dx + dy*dy + dt*dt);
      double Ndd = 0;

      StateS6D dir;
      if(length != 0 && !last)
 	{
	  dir = StateS6D(dx/length,dy/length,0,0,0,dt/length);
	  //if(dd != 0) Ndd = dd/length;
 	}
      else if(length != 0 && last)
 	{
 	  //return (searchNearTempFS(cur, env) * coefRT);
 	  dir = StateS6D(0,0,0,0,0,dt/length);
 	}
      else
 	dir = StateS6D();
      
      // check if this state is in collision
      double radObs = ((ObsDisk2D*)(this->env_->getObs(nearestObs)))->getRadius();
      double radRb = rb_->getRadius();
      StateS6D forceNoCol;
      if(!last && sqdd < (radObs + radRb)*(radObs + radRb))
	{
	  double dd = sqrt(sqdd);
	  double ddx = (dx * (radObs +radRb) / dd) - dx;
	  double ddy = (dy * (radObs +radRb) / dd) - dy;
	  //forceNoCol = StateR5D(ddx,ddy,0,0,0);
	  forceNoCol = StateS6D();
	}

      // compute the base external force
//       double val = (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));
//       StateS6D baseExtForce;
//       if(val > 0) baseExtForce = dir * (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));

      double val = (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));
      StateS6D baseExtForce;
      if(val > 0) baseExtForce = StateS6D(dir.x() * val,
					  dir.y() * val,
					  dir.th() * val,
					  dir.s() * val,
					  dir.ph() * val,
					  dir.t() * val);

      if(this->modeDeform == SPATONLY_DEFORM)
	{
	  StateS6D force(baseExtForce.x() + forceNoCol.x(),
			 baseExtForce.y() + forceNoCol.y(),
			 0,0,0,0);
	  //force = force * (coefRT);
	  force = StateS6D(force.x() * coefRT,
			   force.y() * coefRT,
			   force.th() * coefRT,
			   force.s() * coefRT,
			   force.ph() * coefRT,
			   force.t() * coefRT);

	  if(fabs(force.t()) > 10.0) 
	    { cout << "force1 : "; force.display(); cout << endl; }
	  return force;
	}
      else if(this->modeDeform == TEMPONLY_DEFORM)
	{
	  StateS6D force(0,
			 0,
			 0,
			 0,
			 0,
			 baseExtForce.t() + forceNoCol.t());
	  //force = force * (coefRT);
	  force = StateS6D(force.x() * coefRT,
			   force.y() * coefRT,
			   force.th() * coefRT,
			   force.s() * coefRT,
			   force.ph() * coefRT,
			   force.t() * coefRT);

	  if(fabs(force.t()) > 10.0) 
	    { cout << "force2 : "; force.display(); cout << endl; }
	  return force;
	}
      else
	{
	  // compute the force to apply
	  StateS6D force(baseExtForce.x() + forceNoCol.x(),
			 baseExtForce.y() + forceNoCol.y(),
			 0,
			 0,
			 0,
			 baseExtForce.t() + forceNoCol.t());
	  //force = force * (coefRT);
	  force = StateS6D(force.x() * coefRT,
			   force.y() * coefRT,
			   force.th() * coefRT,
			   force.s() * coefRT,
			   force.ph() * coefRT,
			   force.t() * coefRT);

	  if(fabs(force.t()) > 10.0) 
	    { cout << "force3 : "; force.display(); cout << endl; }
	  return force;
	}
    }
  else
    {
      if((*tableNearState)[I] != NULL) delete (*tableNearState)[I];
      (*tableNearState)[I] = NULL;
      
      StateS6D force;
      return force;
    }
}

// compute the total external forces
StateS6D STNHDeformation::computeNHFExt(const StateR& cur, Environment* env, double time, bool last)
{
//   StateR5D cur5D(cur.x(), cur.y(), cur.t(), 0, 0);
//   StateR5D fExt5D = this->computeFExt(cur5D, env, time, last);
//   return StateS6D(fExt5D.x(), fExt5D.y(), 0, 0, 0, fExt5D.t());

  return this->computeNHF6Ext(cur,env,time,last);
}

// compute forces exerted by static obstacles
StateS6D STNHDeformation::computeNHStFExt(const StateR& cur, Environment* env, bool last)
{
  StateR5D cur5D(cur.x(), cur.y(), cur.t(), 0, 0);
  StateR5D fExt5D = this->computeStFExt(cur5D, env, last);
  return StateS6D(fExt5D.x(), fExt5D.y(), 0, 0, 0, fExt5D.t());
}

// compute the total forces
StateS6D STNHDeformation::computeNHForces(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, double time, bool last)
{
  bool seg5 = false;
  StateS6D vFInt; // = computeNHFInt(prev,cur,next,next,seg5);
  StateS6D vFExt = computeNHFExt(cur,env,time,last);

  //StateS6D vRes = vFInt + vFExt;
  StateS6D vRes(vFInt.x() + vFExt.x(),
		vFInt.y() + vFExt.y(),
		vFInt.th() + vFExt.th(),
		vFInt.s() + vFExt.s(),
		vFInt.ph() + vFExt.ph(),
		vFInt.t() + vFExt.t());
  return vRes;
}


/********** Connectivity and Collision checking **********/

// check the connectivity of the trajectory
double STNHDeformation::checkConnectivity(Trajectory* traj, double curTime)
{
  bool trouve = false;
  int i=0;
  double dt = -1;
  list <Vec_DP*>::iterator it = listCtrl_->begin();
  while(it != listCtrl_->end() && i < traj->getSize() && !trouve)
    {
      // get back the time of the current state
      double time = ((StateS6D*)traj->at(i))->t();

      // test if the current state has not been already visited
      if(time > curTime)
	{
	  // test if the connectivity has been lost on the current state
	  if(*it == NULL)
	    {
	      trouve = true;
	      dt = time - curTime;
	    }
	  else
	    {
	      i++;
	      it++;
	    }
	}
      else
	{
	  i++;
	  it++;
	}
    }
  return dt;
}

// check the connectivity between two nodes of the trajectory
void STNHDeformation::checkNHConnectivity(int i, const StateR& oldCur, const StateR& prev, const StateR& cur, StateR& force)
{
  double maxSpd = this->rb_->getVMax();

  // check the temporal consistency of the curve
  double oldDX = oldCur.x() - prev.x();
  double oldDY = oldCur.y() - prev.y();
  bool noMoreConnec1 = false;
  bool noMoreConnec2 = false;
  
  if(oldCur.t() + force.t() < prev.t() + oldDX / this->rb_->getVMax())
    {
      (*typeState)[i] = -2;
      noMoreConnec1 = true;
      if(useTemporalConsistency) force.setT(prev.t() + fabs(oldDX) / this->rb_->getVMax() - oldCur.t());
    }
  if(oldCur.t() + force.t() < prev.t() + oldDY / this->rb_->getVMax())
    {
      (*typeState)[i] = -2;
      noMoreConnec2 = true;
      if(useTemporalConsistency) force.setT(prev.t() + fabs(oldDY) / this->rb_->getVMax() - oldCur.t());
    }
  if(!noMoreConnec1 && !noMoreConnec2)
    {
      (*typeState)[i] = 0;
    }
  
  // check the connectivity of the curve
  double newDX = oldCur.x() + force.x() - prev.x();
  double newDY = oldCur.y() + force.y() - prev.y();
  double DMAX = maxSpd * (oldCur.t() + force.t() - prev.t());
  bool noMoreConnec3 = false;
  bool noMoreConnec4 = false;
  bool noMoreConnec5 = false;
  bool noMoreConnec6 = false;
  
  if(DMAX < 0) DMAX = -DMAX;
  
  if( newDX > DMAX )
    {
      (*typeState)[i] = -2;
      noMoreConnec3 = true;
    }
  else if ( newDX < -DMAX )
    {
      (*typeState)[i] = -2;
      noMoreConnec4 = true;
    }
  if( newDY > DMAX ) 
    {
      (*typeState)[i] = -2;
      noMoreConnec5 = true;
    }
  else if ( newDY < -DMAX ) 
    {
      (*typeState)[i] = -2;
      noMoreConnec6 = true;
    }
  if(!noMoreConnec1 && !noMoreConnec2 && !noMoreConnec3 && !noMoreConnec4 && !noMoreConnec5 && !noMoreConnec6)
    {
      (*typeState)[i] = 0;
    }
}

// adjust the time of the nodes of the trajectory to maintain its consistency
void STNHDeformation::adjustTime(int curStep)
{
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      if(i != 0 && i > curStep)
	{
	  // compute the distance from the previous node
	  StateS6D* prev = (StateS6D*)((*(this->traj_))[i-1]);
	  StateS6D* cur = (StateS6D*)((*(this->traj_))[i]);
	  
	  double dx = cur->x() - prev->x();
	  double dy = cur->y() - prev->y();
	  double dist = sqrt(dx*dx + dy*dy);
	  double dt = cur->t() - prev->t();

	  double difDir = computeMinDiffAngle(normalizedAngle(prev->th()),normalizedAngle(cur->th()));

	  double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);

	  double cdist = coefDist*dist;

	  double vmoy = 0.5*(prev->s() + cur->s());

	  double newDT = cdist / vmoy;

	  if(dt < 0.5 * newDT)
	    cur->setT(prev->t() + 0.5*newDT);
	}
    }
}

// check if the still-to-be executed part of the trajectory is in collision with the environment
// return the first time from the current time where a collision occurs. return -1 if there is no collision
double STNHDeformation::checkCollision(STrajectory* traj,Environment* env,double curTime)
{
}


/*********************** Resampling **********************/

// resample the trajectory
void STNHDeformation::resample(int curStep)
{
  cout << "resampling..." << endl;

  double minTime = 0.5*this->timeStep_;
  double minDist = 0.5*this->posStep_;

  double maxTime = 2*this->timeStep_;
  double maxDist = 2*this->posStep_;
  double maxDir  = 0.2*M_PI;

  // check if the space between two consecutive states of the trajectory is too large
  // and if the orientation difference between them is not too great too
  std::list <Vec_DP*>::iterator it =  listCtrl_->begin();
  for(int j=0; j<curStep; j++) it++;
  
  int i=curStep;
  while(i < this->traj_->getSize() - 1 && it != listCtrl_->end())
    {
      // get back the current and the next states
      StateR* curS  = (*this->traj_)[i];
      StateR* nextS = (*this->traj_)[i+1];
      
      double dx = nextS->x() - curS->x();
      double dy = nextS->y() - curS->y();
      double sqDist = dx*dx + dy*dy;
      
      double dTH = nextS->th() - curS->th();
      
      bool tooShort = false;

      // check if the spatial or temporal space is too short
      if((i < this->traj_->getSize() - 2))
	{
	  StateR* nnextS = (*this->traj_)[i+2];
	  double dx2 = nextS->x() - curS->x();
	  double dy2 = nextS->y() - curS->y();
	  double sqDist2 = dx2*dx2 + dy2*dy2;
	  
	  double dt2 = nnextS->t() - curS->t();
	  
	  if(dt2 < minTime || sqDist2 < minDist*minDist)
	    {
	      this->deleteStateFromTraj(i+1,i+2);
	      tooShort = true;
	    }
	}
	
      // check if the spatial or temporal space is too large
      if((!tooShort) && (nextS->t() - curS->t() > maxTime || sqDist > maxDist*maxDist))
	{
// 	  cout << "curS : "; curS->display(); cout << endl;
// 	  cout << "nextS : "; nextS->display(); cout << endl;
// 	  cout << "nextS->t()-curS->t() : " << nextS->t()-curS->t() << "   maxTime : " << maxTime << endl;
// 	  cout << "sqDist : " << sqDist << "   maxDist : " << maxDist << endl;
	  
	  // handle the particular case of the subtrajectory reaching the final state
	  if(i == this->traj_->getSize() - 2)
	    {
	      
	      StateS6D s0 = *((StateS6D*)curS);
	      StateS6D sf = *((StateS6D*)nextS);
	      
	      // try to generate a trajectory between both states
	      StateS6D sres;
	      Vec_DP* param = NULL;
	      int res = this->computeItdStateT_Strict(s0,sf,sres,&param);
	      
	      // check if a trajectory has been found
	      if(res == 0)
		{
		  // a trajectory has been found : 
		  // add a new node before the final state
		  this->insertStateInTraj(sres,i+1,&param,T_FEA);
		  
		  cout << "A node has been added before the final node" << endl;
		}
	      else
		{
		  // no trajectory has been found
		  // try to compute an alternative trajectory
		  
		  // reinit the parameters
		  param = NULL;
		  
		  // compute the distance between p0 and pf
		  double dist = sqrt(sqDist);
		  
		  // compute an average velocity between v0 and vf
		  double vmoy = 0.5 * (s0.s() + sf.s());
		  
		  // compute the time to travel dist with the velocity vmoy
		  double tmoy;
		  if(vmoy <= 0.0)
		    {
		      // the average time cannot be computed
		      cout << "STNHDeformation::initDeform : Warning. The average time cannot be computed" << endl;
		      return;
		    }
		  else
		    {
		      tmoy = dist / vmoy;
		    }
		  
		  double oldDT = sf.t() - s0.t();
		  sf.setT(s0.t() + 1.2*tmoy);
		  
		  int res2 = this->computeItdStateT_Strict(s0,sf,sres,&param);
		  
		  if(res2 == 0)
		    {
		      // a trajectory has been found : 
		      // add a new node before the final state
		      this->insertStateInTraj(sres,i+1,&param,T_FEA);
		      
		      // change the time of the final state
		      nextS->setT(sf.t());
		      
		      cout << "A node has been added before the final node (alternative solution)" << endl;
		    }
		  else
		    {
		      cout << "STNHDeformation::initDeform : Warning. No solution has been found to add a node before the final state..." << endl;
		    }
		}
	      
	    }
	  else // general case
	    {
	      // get back the current control parameters
	      Vec_DP* curParam = *it;
	      
	      if(curParam != NULL)
		{
		  // compute the state reached by the current control
		  //ModelFInt carModel2;
		  CLTrajGen carModel2;
		  carModel2.L_ = rb_->getLength();
		  // 	      carModel2.setInitialGuess((*curParam)[0],(*curParam)[1],(*curParam)[2],(*curParam)[3],
		  // 					(*curParam)[4],(*curParam)[5],0.5*(*curParam)[6]);
		  carModel2.setInitialGuess((*curParam)[0],(*curParam)[1],(*curParam)[2],(*curParam)[3],
					    (*curParam)[4],(*curParam)[5],0.5*(nextS->t()-curS->t()));
		  
		  Vec_DP x0(6); x0[0] = curS->x(); x0[1] = curS->y(); x0[2] = curS->th(); 
		  x0[3] = curS->s()*tan(curS->ph())/rb_->getLength(); x0[4] = curS->s(); x0[5] = 0.0;
		  Vec_DP xf(6); xf[0] = curS->x(); xf[1] = curS->y(); xf[2] = curS->th(); 
		  xf[3] = curS->s()*tan(curS->ph())/rb_->getLength(); xf[4] = curS->s(); xf[5] = 0.0;
		  Vec_DP constraints(6); constraints[0] = 0.5; constraints[1] = rb_->getAMax(); 
		  constraints[2] = rb_->getVMax(); constraints[3] = rb_->getLength(); 
		  constraints[4] = rb_->getMaxSteering(); constraints[5] = rb_->getMaxSteerSpd();
		  
		  carModel2.setInitialParam(x0,xf,constraints);
		  carModel2.ForwardSolution();
		  
		  double x = carModel2.vecState_xsim[0];
		  double y = carModel2.vecState_xsim[1];
		  double th = carModel2.vecState_xsim[2];
		  double v = carModel2.vecState_xsim[4];
		  double om = carModel2.vecState_xsim[3];
		  double ph = atan2(rb_->getLength()*om,v);
		  double t = curS->t() + 0.5*(nextS->t()-curS->t());
		  
		  //StateR* itdS = new StateS6D(x,y,th,v,ph,t);
		  StateS6D itdS(x,y,th,v,ph,t);
		  
		  // change the time of the control from the current state
		  (*curParam)[6] = 0.5*(nextS->t()-curS->t());
		  // copy the control to insert it for the new state
		  Vec_DP* newParam = new Vec_DP(*curParam);
		  
		  // 	      cout << "itdS : "; itdS.display(); cout << endl;
		  // 	      cout << "newParam : (" << (*newParam)[0] << "," << (*newParam)[1] << "," << (*newParam)[2] << "," << 
		  // 		(*newParam)[3] << "," << (*newParam)[4] << "," << (*newParam)[5] << "," << (*newParam)[6] << ")" << endl;
		  
		  // insert it in the trajectory
		  this->insertStateInTraj(itdS,i+1,&newParam,T_FEA);
		  
		  i++;
		  it++;
		  //this->traj_->insert(i+1, itdS);
		  //listCtrl_->insert(it,newParam);
		  
		  cout << "itd state added" << endl;
		}
	    }
	}
      i++;
      it++;
    }
}

/************** Spatio-temporal deformation **************/

// init the deformation and check the temporal space between two consecutive states
void STNHDeformation::initDeform(int curStep)
{
  // initialize the robotic system's constraints
  this->initConstraints();

  // check if the list of feedback controls must be initialized
  if(listCtrl_ == NULL)
    {
      listCtrl_ = new list <Vec_DP*>(traj_->getSize() - 1);
      list <Vec_DP*>::iterator it = listCtrl_->begin();
      for(it = listCtrl_->begin(); it != listCtrl_->end(); it++)
	*it = NULL;

      feaCtrl_ = new list <int>(traj_->getSize() - 1);
      list <int>::iterator it2 = feaCtrl_->begin();
      for(it2 = feaCtrl_->begin(); it2 != feaCtrl_->end(); it2++)
	*it2 = 0;
    }

  // clear the tables of forces, states types and nearest states
  if(*typeState != NULL) delete [] (*typeState);
  if(*tableNearState != NULL)delete [] (*tableNearState);
  if(*tableFInt != NULL)delete [] (*tableFInt);
  if(*tableFExt != NULL)delete [] (*tableFExt);

  if(useResampling_)
    {
      this->resample(curStep);
    }
  
  // reinit the tables of states types and nearest states
  *typeState = new int [this->traj_->getSize()];
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      (*typeState)[i] = -1;
    }
  *tableNearState = (StateR**)(new StateS6D* [this->traj_->getSize()]);
  for(int i=0; i<this->traj_->getSize(); i++)
    {
      (*tableNearState)[i] = NULL;
    }
  
  *tableFInt = new StateS6D [this->traj_->getSize()];
  
  *tableFExt = new StateS6D [this->traj_->getSize()];
}

// finish the deformation
int STNHDeformation::finishDeform(double time, int curStep)
{
  this->removeLoop(curStep);

  int res = 0;

  // check if a collision may occur during the prediction horizon
  double dt = this->checkCollision(this->traj_, this->env_, time);
  if(dt != -1 && dt < 0.3*(*predHorizon_))
    {
      res = -2;
    }
  else
    {
      // check the connectivity of the trajectory
      double dt2 = this->checkConnectivity(this->traj_, time);
      if(dt != -1 && dt < 0.2*(*predHorizon_))
	{
	  res = -3;
	}
    }
  
  this->oldTime_ = time;

  return res;
}

// deform the current trajectory
int STNHDeformation::deform(double time, int curStep)
{
  double maxSpd = this->rb_->getVMax();
  double elapsedTime = time - this->oldTime_;

  // clear the intermediate trajectory (for debug mode)
  itdTraj_.clear();
  itdTraj2_.clear();

  // init the deformation
  this->initDeform(curStep);
  
  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      return -1;
    }
  else
    {
      // save the old trajectory
      Trajectory6D oldTraj(*((Trajectory6D*)(this->traj_)));
      
      // apply the repulsive forces from the obstacles
      for(int i=1; i<this->traj_->getSize(); i++)
	{
	  I = i;

	  if(i > curStep)
	    {
	      if((*(this->traj_))[i]->t() < time + (*reliabilityTime))
		{
		  bool last;
		  if(i == this->traj_->getSize()-1) last = true;
		  else last = false;
		  
		  // compute the external forces due to dynamic obstacles
		  StateS6D fExtDyn = this->computeNHFExt(*((oldTraj)[i]), this->env_, time, last); // * elapsedTime;
		  
		  // compute the external forces due to static obstacles
		  StateS6D fExtSt = this->computeNHStFExt(*((oldTraj)[i]), this->env_, last);		  
		  
		  // compute the total external forces
		  StateS6D forces;
		  //if(i > traj_->getSize() - 21) forces = (fExtDyn + fExtSt)*(0.05*(traj_->getSize()-i));
		  //else 
// 		  forces = (fExtDyn*(*K_EXT)) + (fExtSt*(*K_STEXT));
//  		  forces = forces * (lastDeformDuration_ *2.5);

		  forces = StateS6D(((fExtDyn.x()*(*K_EXT)) + (fExtSt.x()*(*K_STEXT))) * (lastDeformDuration_ *2.5),
				    ((fExtDyn.y()*(*K_EXT)) + (fExtSt.y()*(*K_STEXT))) * (lastDeformDuration_ *2.5),
				    ((fExtDyn.th()*(*K_EXT)) + (fExtSt.th()*(*K_STEXT))) * (lastDeformDuration_ *2.5),
				    ((fExtDyn.s()*(*K_EXT)) + (fExtSt.s()*(*K_STEXT))) * (lastDeformDuration_ *2.5),
				    ((fExtDyn.ph()*(*K_EXT)) + (fExtSt.ph()*(*K_STEXT))) * (lastDeformDuration_ *2.5),
				    ((fExtDyn.t()*(*K_EXT)) + (fExtSt.t()*(*K_STEXT))) * (lastDeformDuration_ *2.5));
		  
		  (*tableFExt)[i] = forces;
		  
		  // check the connectivity of the curve
		  //this->checkConnectivity(i, *(oldTraj[i]), *((*(this->traj_))[i-1]), *((*(this->traj_))[i]), forces);
		  
		  // apply the forces
		  ((*(this->traj_))[i])->setX( ((oldTraj)[i])->x() + (forces.x()) );
		  ((*(this->traj_))[i])->setY( ((oldTraj)[i])->y() + (forces.y()) );
		  ((*(this->traj_))[i])->setTH( ((oldTraj)[i])->th() + (forces.th()) );
 		  ((*(this->traj_))[i])->setS( ((oldTraj)[i])->s() + (forces.s()) );
 		  ((*(this->traj_))[i])->setPH( ((oldTraj)[i])->ph() + (forces.ph()) );
		  ((*(this->traj_))[i])->setT( ((oldTraj)[i])->t() + (forces.t()) );
		  
		  itdTraj_.push_back(StateS6D(this->traj_->at(i)->x(),
					      this->traj_->at(i)->y(),
					      this->traj_->at(i)->th(),
					      this->traj_->at(i)->s(),
					      this->traj_->at(i)->ph(),
					      this->traj_->at(i)->t()));
		}
	    }
	}

      // adjust the time of each node
      this->adjustTime(curStep);

//       // adjust the speed and the curvature of each state
//       for(int i=1; i<this->traj_->getSize()-1; i++)
// 	{
// 	  // get back the i-1 th i th and i+1 th states
// 	  StateR* prev = ((*(this->traj_))[i-1]);
// 	  StateR* cur = ((*(this->traj_))[i]);
// 	  StateR* next = ((*(this->traj_))[i+1]);
	  
// 	  // compute the distance between the i-1 th and i+1 th configurations
// 	  double dx = next->x() - prev->x();
// 	  double dy = next->y() - prev->y();
// 	  double dpos = sqrt(dx*dx + dy*dy);
	  
// 	  // compute the direction change between the i-1 th and i+1 th configurations
// 	  double dth = next->th() - prev->th();
	  
// 	  // compute the time between the i-1 th and i+1 th times
// 	  double dt = next->t() - prev->t();
	  
// 	  // compute the new speed of the current state
// 	  double s;
// 	  if(dt > 0) s = dpos/dt;
// 	  else s = 0;
// 	  cur->setS(s);

// 	  // compute the curvature of the current state
// 	  double ph;
// 	  if(s!=0) ph = dth / s;
// 	  else ph = 0;
// 	  cur->setPH(ph);
// 	}
      
      // apply the internal forces
      if(parallelExec)
	{
	  // save the intermediate trajectory
	  Trajectory6D itdTraj(*((Trajectory6D*)(this->traj_)));

	  list <Vec_DP*>::iterator it = listCtrl_->begin();
	  list <int>::iterator it2 = feaCtrl_->begin();

	  int i=1;
	  while(i<itdTraj.getSize()-1)
	    {
	      int nbThreadUsed = 0;
	      boost::thread (**tabThread);
	      tabThread = new boost::thread* [nbThread];
	      
	      StateS6D* tabRes = new StateS6D[nbThread];
	      
	      int lastI = -1;
	      
	      cout << "search threads to create from " << i << endl;
	      
	      while(i<itdTraj.getSize()-1 && nbThreadUsed != this->nbThread)
		{
		  it++;
		  it2++;
		  
		  if(i > curStep)
		    {
		      if(itdTraj[i]->t() < time + (*reliabilityTime))
			{
			  lastI = i;
			  
			  bool last;
			  if(i == itdTraj.getSize()-1) last = true;
			  else last = false;
			  
			  // create the thread allowing to compute the trajectory generation process
			  boost::thread* th;
			                          // = new boost::thread(boost::bind(&STNHDeformation::computeNHFInt,this,
// 										  boost::cref(*(itdTraj[i-1])), 
// 										  boost::cref(*(itdTraj[i])), 
// 										  boost::cref(*(itdTraj[i+1])), 
// 									          last, &(*it), &(*it2),
// 										  boost::ref(tabRes[nbThreadUsed])) );
			  cout << "thread created" << endl;
			  tabThread[nbThreadUsed] = th;
			  //StateS6D forces = this->computeNHFInt(*(itdTraj[i-1]), *(itdTraj[i]), *(itdTraj[i+1]), last, &(*it));
			  //cout << "thread put in the table" << endl;
			  //(tabThread[nbThreadUsed])->join();

			  //cout << "force : "; (tabRes[nbThreadUsed]).display(); cout << endl;

			  nbThreadUsed++;
			}
		    }
		  
		  i++;
		}

	      cout << "nbThreadUsed : " << nbThreadUsed << " at node : " << lastI << endl; 

	      // get back the results of parallel computations and apply the forces on the trajectory
	      for(int j=0; j<nbThreadUsed; j++)
		{
		  (tabThread[j])->join();

		  int ind = lastI - nbThreadUsed + j + 1;

		  // get back the force computed
		  StateS6D forces = tabRes[j];

		  // process that allows to reduce the force applied on the last states of the trajectory
		  if(ind > traj_->getSize() - 11)
		    {
		      //forces = forces*(0.1*(traj_->getSize()-i));
		      
		      forces = StateS6D(forces.x() * (0.1*(traj_->getSize()-i)),
				       forces.y() * (0.1*(traj_->getSize()-i)),
				       forces.th() * (0.1*(traj_->getSize()-i)),
				       forces.s() * (0.1*(traj_->getSize()-i)),
				       forces.ph() * (0.1*(traj_->getSize()-i)),
				       forces.t() * (0.1*(traj_->getSize()-i)));
		    }
		  
		  cout << "ind : " << ind << "  "; forces.display(); cout << endl;

		  // update the table of internal forces
		  (*tableFInt)[ind] = forces;
		  
		  // compute the coefficient due to the reliability time
		  double coefRT = exp(((time - (itdTraj[i])->t())*(time - (itdTraj[i])->t())*log(*epsilonRT))/((*reliabilityTime)*(*reliabilityTime)));
		  
		  ((*(this->traj_))[ind])->setX( (itdTraj[ind])->x() + coefRT*(forces.x()) );
		  ((*(this->traj_))[ind])->setY( (itdTraj[ind])->y() + coefRT*(forces.y()) );
		  ((*(this->traj_))[ind])->setTH( (itdTraj[ind])->th() + coefRT*(forces.th()) );
		  ((*(this->traj_))[ind])->setS( (itdTraj[ind])->s() + coefRT*(forces.s()) );
		  ((*(this->traj_))[ind])->setPH( (itdTraj[ind])->ph() + coefRT*(forces.ph()) );
		  ((*(this->traj_))[ind])->setT( (itdTraj[ind])->t() + coefRT*(forces.t()) );
		}
	      
	      cout << "Forces applied" << endl;

	      // clear the allocated tables
	      delete tabThread;
	      cout << "tabThread deleted" << endl;
	      //delete tabRes;

	      cout << "tabRes deleted" << endl;
	    }
	}

      else
	{
	  // save the current trajectory to display it
	  for(int i=0; i<this->traj_->getSize(); i++)
	    {
	      StateS6D stmp = *(this->traj_->at(i));
	      itdTraj2_.push_back(stmp);
// 	      cout << "itdTraj2[" << i << "] : "; stmp.display(); cout << endl;
	    }
	  
	  list <Vec_DP*>::iterator it = listCtrl_->begin();
	  list <int>::iterator it2 = feaCtrl_->begin();
	  
	  for(int i=1; i<this->traj_->getSize()-1; i++)
	    {
	      it++;
	      it2++;

	      if(!usePreviousParam_) (*it) = NULL;
	      
	      if(i > curStep)
		{
		  if((*traj_)[i]->t() < time + (*reliabilityTime))
		    {		  
		      bool last;
		      if(i == oldTraj.getSize()-1) last = true;
		      else last = false;

// 		      cout << "////////////////////// deb //////////////////////" << endl;
// 		      cout << i-1 << " : "; this->traj_->at(i-1)->display(); cout << endl;
// 		      cout << i << " : "; this->traj_->at(i)->display(); cout << endl;
// 		      cout << i+1 << " : "; this->traj_->at(i+1)->display(); cout << endl;

		      StateS6D forces = this->computeNHFInt(*((*(this->traj_))[i-1]), *((*(this->traj_))[i]), *((*(this->traj_))[i+1]), last, &(*it), &(*it2));
		      if(i > traj_->getSize() - 11)
		      {
			//forces = forces*(0.1*(traj_->getSize()-i));
			
			forces = StateS6D(forces.x() * (0.1*(traj_->getSize()-i)),
					  forces.y() * (0.1*(traj_->getSize()-i)),
					  forces.th() * (0.1*(traj_->getSize()-i)),
					  forces.s() * (0.1*(traj_->getSize()-i)),
					  forces.ph() * (0.1*(traj_->getSize()-i)),
					  forces.t() * (0.1*(traj_->getSize()-i)));
		      }

// 		      cout << "////////////////////// fin //////////////////////" << endl;
		      
		      // update the table of internal forces
		      (*tableFInt)[i] = forces;
		      
		      // compute the coefficient due to the reliability time
		      double coefRT = exp(((time - ((*(this->traj_))[i])->t())*(time - ((*(this->traj_))[i])->t())*log(*epsilonRT))/((*reliabilityTime)*(*reliabilityTime)));
		      
// 		      cout << "coefRT : " << coefRT << endl;

		      ((*(this->traj_))[i])->setX( ((*(this->traj_))[i])->x() + coefRT*(forces.x()) );
		      ((*(this->traj_))[i])->setY( ((*(this->traj_))[i])->y() + coefRT*(forces.y()) );
		      ((*(this->traj_))[i])->setTH( ((*(this->traj_))[i])->th() + coefRT*(forces.th()) );
		      ((*(this->traj_))[i])->setS( ((*(this->traj_))[i])->s() + coefRT*(forces.s()) );
		      ((*(this->traj_))[i])->setPH( ((*(this->traj_))[i])->ph() + coefRT*(forces.ph()) );
		      ((*(this->traj_))[i])->setT( ((*(this->traj_))[i])->t() + coefRT*(forces.t()) );
		    }
		}
	    }
	}

    }

  cout << "deformation finished" << endl;

  // finish the deformation
  return this->finishDeform(time,curStep);
}


/*************** Feedback control methods ****************/

// init the list of feedback controls
void STNHDeformation::initListCtrl(std::list <Vec_DP*>* listCtrl)
{
  this->listCtrl_ = listCtrl;
}


/***************** Trajectory computation ****************/

// initialize the trajectory generator and compute a trajectory from prev to next
// return the trajectory generator return status
int STNHDeformation::genTraj(const StateS6D& prev,const StateS6D& next,CLTrajGen& carModel,Vec_DP** param,double& dt) const
{
  carModel.L_ = rb_->getLength();

  Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th();
                x0[3] = prev.s()*tan(prev.ph())/rb_->getLength(); x0[4] = prev.s(); x0[5] = 0.0;
  Vec_DP xf(6); xf[0] = next.x(); xf[1] = next.y(); xf[2] = next.th(); 
                xf[3] = next.s()*tan(next.ph())/rb_->getLength(); xf[4] = next.s(); xf[5] = next.t() - prev.t();

  // check the temporal consistency between both nodes
  double dx = next.x() - prev.x();
  double dy = next.y() - prev.y();
  dt = next.t() - prev.t();
  if(dt < 0)
    {
      dt = (dx*dx+dy*dy) / rb_->getVMax();
      xf[5] = dt;
    }

  int res = carModel.ComputeTrajectory(x0,xf,constraints_,param);
  return res;
}

// compute the intermediate node to add between prev and next 
// strict case : compute the node only if the convergence has been ensured
// return status : 0 -> convergence has been ensured
//                 1 -> convergence has not been ensured
int STNHDeformation::computeItdStateT_Strict(const StateS6D& prev,const StateS6D& next,StateS6D& res,Vec_DP** param)
{
  CLTrajGen carModel;
  double dt = 0.0;
  
  //TrajGen::setDBG(2);
  int resTG = this->genTraj(prev,next,carModel,param,dt);
  //TrajGen::setDBG(0);

//   cout << "residual : " << carModel.finalResidual_ << endl;

  // check the result status of the trajectory generation process
  if(resTG == 0)
    {
      // we compute the intermediate time at which the mid state must be evaluated
      double tSur2 = 0.5*dt;
      
      // we compute the intermediate state of the trajectory
      double x = carModel.funcX(tSur2);
      double y = carModel.funcY(tSur2);
      double th = carModel.funcTheta(tSur2);
      double v = carModel.funcVel(tSur2);
      double om = carModel.funcAVel(tSur2);
      double ph;
      if(v != 0.0) ph = atan(rb_->getLength() * om / v);
      else
	{
	  cout << "STNHDeformation::computeNHFInt : Warning. vres = 0. Cannot compute ph..." << endl;
	  exit(1);
	}
      
      res = StateS6D(x,y,th,v,ph,prev.t() + tSur2);
      
      return 0;
    }
  else
    {
      // the convergence has not been ensured : 
      // we notice it and let the result state unchanged
      return 1;
    }
}

// compute the intermediate node to add between prev and next 
// average case : compute a intermediate node even if the convergence hasn't been ensured
// return status : 0 -> near convergence has been ensured
//                 1 -> far convergence has been ensured
//                 2 -> convergence hasn't been ensured
int STNHDeformation::computeItdStateT_Avg(const StateS6D& prev,const StateS6D& next,StateS6D& res,Vec_DP** param)
{
  CLTrajGen carModel;
  double dt = 0.0;
  int resTG = this->genTraj(prev,next,carModel,param,dt);
  
  // we compute the intermediate time at which the mid state must be evaluated
  double tSur2 = 0.5*dt;

  // check if the trajectory generator has found a solution
  if(resTG == 0 || (carModel.finalResidual_ < 20.0 && carModel.finalErrTH_ < 0.5))
    {
      // we compute the intermediate state of the trajectory
      double x = carModel.funcX(tSur2);
      double y = carModel.funcY(tSur2); 
      double th = carModel.funcTheta(tSur2);
      double v = carModel.funcVel(tSur2);
      double om = carModel.funcAVel(tSur2);
      double ph;
      if(v != 0.0) ph = atan(rb_->getLength() * om / v);
      else
	{
	  cout << "STNHDeformation::computeNHFInt : Warning. vres = 0. Cannot compute ph..." << endl;
	  exit(1);
	}

      res = StateS6D(x,y,th,v,ph,prev.t() + tSur2);

      if(resTG == 0) 
	return 0;
      else
	return 1;
    }
  else
    {
      // delete the deprecated input parameters
      *param = NULL;

      // create a deceleration control
      ControlTimeAAS decCtrl(-0.5 * rb_->getAMax(),0.0,tSur2);

      // compute the state reached by the deceleration control
      CLTrajGen carModel2;
      carModel2.L_ = rb_->getLength();
      carModel2.setInitialGuess(decCtrl.getAS(),0.0,0.0,decCtrl.getAcc(),0.0,0.0,decCtrl.getTime());
      Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th(); 
                    x0[3] = prev.s()*tan(prev.ph())/rb_->getLength(); x0[4] = prev.s(); x0[5] = 0.0;
      Vec_DP xf(6); xf[0] = prev.x(); xf[1] = prev.y(); xf[2] = prev.th(); 
                    xf[3] = prev.s()*tan(prev.ph())/rb_->getLength(); xf[4] = prev.s(); xf[5] = 0.0;

      carModel2.setInitialParam(x0,xf,constraints_);
      carModel2.ForwardSolution();

      double x = carModel2.vecState_xsim[0];
      double y = carModel2.vecState_xsim[1];
      double th = carModel2.vecState_xsim[2];
      double v = carModel2.vecState_xsim[4];
      double om = carModel2.vecState_xsim[3];
      double ph = atan2(rb_->getLength()*om,v);
      
      const GLfloat redMaterial[]={1.,0.,0.,1.};
      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
      glPointSize(30.0);
            
      res = StateS6D(x,y,th,v,ph,prev.t() + tSur2);
      
      return 2;
    }
}

/**************** Trajectory modifications ***************/

// return the state of the deformed trajectory at time t
StateS6D STNHDeformation::getStateTraj(double t) const
{
  // check if the time is defined over the trajectory
  if(t<0)
    {
      cout << "STNHDeformation::getStateTraj : Warning. TTTTTTIIIIIIMMMMMMMEEEE UUUUNNNNDDDDIIIFFFIIINNNEEEDDD!!!!!!" << endl;
      return StateS6D();
    }
  else if(t>this->traj_->at(traj_->getSize()-1)->t())
    {
      return *((StateS6D*)(this->traj_->at(traj_->getSize()-1)));
    }

  bool trouve = false;
  int i=1; 
  while(!trouve && i < this->traj_->getSize())
    {
      if(this->traj_->at(i)->t() > t)
	{
	  trouve = true;
	  i--;
	}
      else
	{
	  i++;
	}
    }

//   cout << "traj_->getSize() : " << traj_->getSize() << endl;
//   cout << "i found: " << i << endl;

  // get back the corresponding state
  StateS6D* cur = (StateS6D*) (this->traj_->at(i));
  StateS6D* next = (StateS6D*) (this->traj_->at(i+1));

//   // get back the control corresponding to the node
//   list <Vec_DP*>::iterator it = listCtrl_->begin();
//   for(int j=0; j<i; j++) it++;

//   if(*it == NULL)
//     {
      // no control is defined here
      double coef = (t - cur->t()) / (next->t() - cur->t());
      return StateS6D(cur->x()*(1.0-coef)   + next->x()*coef,
		      cur->y()*(1.0-coef)   + next->y()*coef,
		      normalizedAngle(cur->th() + coef * computeMinDiffAngle(next->th(),cur->th())),
// 		      cur->th()*(1.0-coef)  + next->th()*coef,
		      cur->s()*(1.0-coef) + next->s()*coef,
		      cur->ph()*(1.0-coef) + next->ph()*coef,
		      cur->t()*(1.0-coef)   + next->t()*coef);
//     }
//   else
//     {
//       // initialize the forward method
//       Vec_DP x0(6); x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
//       x0[3] = (cur->omr() + cur->oml())/2.0; x0[4] = (cur->omr() - cur->oml())/(2.0*rb_->getLength()); x0[5] = 0.0;
      
//       Vec_DP xf(6); xf[0] = next->x(); xf[1] = next->y(); xf[2] = next->th(); 
//       x0[3] = (next->omr() + next->oml())/2.0; x0[4] = (next->omr() - next->oml())/(2.0*rb_->getLength()); xf[5] = next->t() - cur->t();
      
//       Vec_DP constraints(3); 
//       constraints[0] = rb_->getWAMax();
//       constraints[1] = rb_->getWVMax();
//       constraints[2] = rb_->getLength();
      
//       DDTrajGen carModel;
//       carModel.setInitialGuess((**it)[0],(**it)[1],(**it)[2],(**it)[3],(**it)[4],(**it)[5],(**it)[6]);
//       carModel.setInitialParam(x0,xf,constraints);
//       carModel.ForwardSolution();
      
//       double x = carModel.vecState_xsim[0];
//       double y = carModel.vecState_xsim[1];
//       double th = carModel.vecState_xsim[2];
//       double v = carModel.vecState_xsim[3];
//       double om = carModel.vecState_xsim[4]; cout << "v : " << v << "  om : " << om << endl;
//       double omr = carModel.funcWRVel(next->t() - cur->t()); //v + rb_->getLength() * om;
//       double oml = carModel.funcWLVel(next->t() - cur->t()); //v - rb_->getLength() * om;
      
//       return DDState(x,y,th,omr,oml,t);
//     }
}

// display the refined trajectory in an openGL viewer
void STNHDeformation::displayTraj() const
{
  const GLfloat blueMaterial[]={0.0,0.0,1.0,1.};
  const GLfloat greenMaterial[]={0.0,1.0,0.0,1.};
  const GLfloat redMaterial[]={1.0,0.0,0.0,1.};
  const GLfloat blackMaterial[]={0.0,0.0,0.0,1.};
  const GLfloat yellowMaterial[]={1.0,1.0,0.0,1.};

  glLineWidth(3.0);
  
//   list <Vec_DP*>::iterator it = listCtrl_->begin();
//   list <int>::iterator it2 = feaCtrl_->begin();

  for(int i=0; i<traj_->getSize()-2; i++)
    {
//       cout << "i : " << i << " :: ";

      // get back the corresponding state
      StateS6D* cur = (StateS6D*) (this->traj_->at(i));
      StateS6D* next = (StateS6D*) (this->traj_->at(i+1));
      StateS6D* nnext = NULL;
      if(i<traj_->getSize()-2) nnext = (StateS6D*) (this->traj_->at(i+2));
      else nnext = (StateS6D*) (this->traj_->at(i+1));
      
//       cout << "cur : "; cur->display(); cout << " :: ";

      // get back the control corresponding to the node
      list <Vec_DP*>::iterator it = listCtrl_->begin();
      for(int j=0; j<=i && j != traj_->getSize()-1; j++) it++;
      list <int>::iterator it2 = feaCtrl_->begin();
      for(int j=0; j<=i && j != traj_->getSize()-1; j++) it2++;

//       it++;
//       it2++;

      // select the color of the trajectory to display
      switch(*it2)
	{
	case T_FEA : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  break;
	case T_FAR : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
	  break;
	case T_UNFEA : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  break;
	default : 
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);
	  break;
	}
      
      glPushMatrix();
      glTranslated(cur->x(),
		   cur->y(),
		   cur->t());
      gluSphere(gluNewQuadric(), 1.0, 5, 5);
      glPopMatrix();

      glBegin(GL_LINE_STRIP);

      // define the sampling of subtrajectories
      int nb = 20;

      if(*it == NULL)
	{
	  // no control is defined here
	  for(int k=0; k<=nb; k++)
	    {
	      double t = cur->t() + (((double)k)/((double)nb))*(nnext->t() - cur->t());
	      double coef = (t - cur->t()) / (nnext->t() - cur->t());
	      
	      double x = cur->x()*(1.0-coef)   + nnext->x()*coef;
	      double y = cur->y()*(1.0-coef)   + nnext->y()*coef;
	      
	      glVertex3f(x,y,t);
	  
	      if(k == nb) {
		glEnd();
// 		cout << "trajCtrl[" << i << "] : (" << x << "," << y << ",,,," << t << ")" << endl;
	      }
	    }
	}
      else
	{
	  // a control is defined here
	  double x0[6]; x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
	  x0[3] = cur->s() * tan(cur->ph()/rb_->getLength()); x0[4] = cur->s(); x0[5] = cur->t();
	      
	  double constraints[6]; 
	  constraints[0] = 0.5;
	  constraints[1] = rb_->getAMax();
	  constraints[2] = rb_->getVMax();
	  constraints[3] = rb_->getLength();
	  constraints[4] = rb_->getMaxSteering();
	  constraints[5] = rb_->getMaxSteerSpd();
	  
	  double parameters[7];
	  parameters[0] = (**it)[0];
	  parameters[1] = (**it)[1];
	  parameters[2] = (**it)[2];
	  parameters[3] = (**it)[3];
	  parameters[4] = (**it)[4];
	  parameters[5] = (**it)[5];
	  parameters[6] = (**it)[6];
	  
	  CLTrajGen carModel;
	  carModel.ComputeFromIG(x0,constraints,parameters);
	  
	  for(int k=0; k<=nb; k++)
	    {
	      double t = cur->t() + (((double)k)/((double)nb))*(nnext->t() - cur->t());
	      double rstatet[6];

	      carModel.getStateTReachedAt(rstatet,t-cur->t());

	      double x = rstatet[0];
	      double y = rstatet[1];
	      double th = rstatet[2];
	      double v = rstatet[4];
	      double om = rstatet[3];
	      double ph = atan(rb_->getLength() * om / v);

	      glVertex3f(x,y,t);
	      
	      if(k == nb) 
		{
		  glEnd();
// 		  cout << "ctrl : (" << (**it)[0] << "," << (**it)[1] << "," << (**it)[2] << "," << (**it)[3] 
// 		       << "," << (**it)[4] << "," << (**it)[5] << "," << (**it)[6] << ") -> ";
// 		  cout << "trajCtrl[" << i << "] : (" << x << "," << y << "," << th << "," << v << ","<< ph << "," << t << ")" << endl;
		  if(i%10 == 0)
		    {
		      glMaterialfv(GL_FRONT, GL_AMBIENT, yellowMaterial);
		      glMaterialfv(GL_FRONT, GL_DIFFUSE, yellowMaterial);
		    }
		  else
		    {
		      glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
		      glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);
		    }
		  glPushMatrix();
		  glTranslated(x,y,t);
		  gluSphere(gluNewQuadric(), 1.0, 5, 5);
		  glPopMatrix();
		}
	    }
	}
    }
//       int nb = 20;
//       for(int k=0; k<=nb; k++)
// 	{
// 	  double t = cur->t() + (((double)k)/((double)nb))*(nnext->t() - cur->t());
// 	  //cout << "  t : " << t;

// 	  if(*it == NULL)
// 	    {
// 	      // no control is defined here
// 	      double coef = (t - cur->t()) / (nnext->t() - cur->t());
	      
// 	      double x = cur->x()*(1.0-coef)   + nnext->x()*coef;
// 	      double y = cur->y()*(1.0-coef)   + nnext->y()*coef;

// 	      glVertex3f(x,y,t);
	      
// 	      if(k == nb) {
// 		glEnd();
// 		cout << "trajCtrl[" << i << "] : (" << x << "," << y << ",,,," << t << ")" << endl;
// 	      }
// 	    }
// 	  else
// 	    {
// 	      // initialize the forward method
// // 	      Vec_DP x0(6); x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
// // 	      x0[3] = cur->ph(); x0[4] = cur->s(); x0[5] = 0.0;
	      
// // 	      Vec_DP xf(6); xf[0] = nnext->x(); xf[1] = nnext->y(); xf[2] = nnext->th(); 
// // 	      xf[3] = nnext->ph(); xf[4] = nnext->s(); xf[5] = nnext->t() - cur->t();
	      
// // 	      Vec_DP constraints(6); 
// // 	      constraints[0] = 0.5;
// // 	      constraints[1] = rb_->getAMax();
// // 	      constraints[2] = rb_->getVMax();
// // 	      constraints[3] = rb_->getLength();
// // 	      constraints[4] = rb_->getMaxSteering();
// // 	      constraints[5] = rb_->getMaxSteerSpd();

// // // 	      cout << "rb_->getLength() : "<< rb_->getLength() << endl;
	      
// // 	      CLTrajGen carModel;
// // 	      carModel.setInitialGuess((**it)[0],(**it)[1],(**it)[2],(**it)[3],(**it)[4],(**it)[5],(**it)[6]);
// // 	      carModel.setInitialParam(x0,xf,constraints);
// // 	      carModel.ForwardSolution();
	      
// // 	      double x = carModel.funcX(t - cur->t()); //carModel.vecState_xsim[0];
// // 	      double y = carModel.funcY(t - cur->t()); //carModel.vecState_xsim[1];
	      
// 	      double x0[6]; x0[0] = cur->x(); x0[1] = cur->y(); x0[2] = cur->th(); 
// 	      x0[3] = cur->ph(); x0[4] = cur->s(); x0[5] = 0.0;

// 	      double constraints[6]; 
// 	      constraints[0] = 0.5;
// 	      constraints[1] = rb_->getAMax();
// 	      constraints[2] = rb_->getVMax();
// 	      constraints[3] = rb_->getLength();
// 	      constraints[4] = rb_->getMaxSteering();
// 	      constraints[5] = rb_->getMaxSteerSpd();

// 	      double parameters[7];
// 	      parameters[0] = (**it)[0];
// 	      parameters[1] = (**it)[1];
// 	      parameters[2] = (**it)[2];
// 	      parameters[3] = (**it)[3];
// 	      parameters[4] = (**it)[4];
// 	      parameters[5] = (**it)[5];
// 	      parameters[6] = (**it)[6];

// 	      CLTrajGen carModel;
// 	      carModel.ComputeFromIG(x0,constraints,parameters);

// 	      glVertex3f(x,y,t);

// 	      if(k == nb) 
// 		{
// 		  glEnd();
// 		  cout << "ctrl : (" << (**it)[0] << "," << (**it)[1] << "," << (**it)[2] << "," << (**it)[3] 
// 		       << "," << (**it)[4] << "," << (**it)[5] << "," << (**it)[6] << ") -> ";
// 		  cout << "trajCtrl[" << i << "] : (" << x << "," << y << ",,,," << t << ")" << endl;
// 		  if(i%10 == 0)
// 		    {
// 		      glMaterialfv(GL_FRONT, GL_AMBIENT, yellowMaterial);
// 		      glMaterialfv(GL_FRONT, GL_DIFFUSE, yellowMaterial);
// 		    }
// 		  else
// 		    {
// 		      glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
// 		      glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);
// 		    }
// 		  glPushMatrix();
// 		  glTranslated(x,y,t);
// 		  gluSphere(gluNewQuadric(), 1.0, 5, 5);
// 		  glPopMatrix();
// 		}
// 	    }
// 	}
//     }
}

// locate the loops along the trajectory
// return the number of loops found
// and fill the vector of loops
int STNHDeformation::locateLoop(int curStep,std::vector<int*> &vecLoop) const
{
  int nbLoop = 0;
  
  for(int i=curStep; i<traj_->getSize()-4; i++)
    {
      for(int j=i+2; j<traj_->getSize()-2; j++)
	{
	  // get back the four states s_{i}, s_{i+1}, s_{j}, s_{j+1}
	  StateS6D Si = *(traj_->at(i));
	  StateS6D Si1 = *(traj_->at(i+1));
	  StateS6D Sj = *(traj_->at(j));
	  StateS6D Sj1 = *(traj_->at(j+1));

	  // test intersection between vectors \overrightarrow{p_{i} p_{i+1}} and \overrightarrow{p_{j} p_{j+1}}
	  Vec2D Pi(Si.x(),Si.y());
	  Vec2D Pi1(Si1.x(),Si1.y());
	  Vec2D Pj(Sj.x(),Sj.y());
	  Vec2D Pj1(Sj1.x(),Sj1.y());
	  Vec2D Pres(0.0,0.0);

	  bool itsc = intersecSeg2D(Pi,Pi1,Pj,Pj1,Pres);
	  
	  if(itsc)
	    {
	      // if an intersection exists, save the initial and final states
	      int* posStates = new int[2];

	      posStates[0] = i;
	      posStates[1] = j+1;

	      // add these positions to the vector of loops
	      vecLoop.push_back(posStates);
	      
	      // update the number of loops found
	      nbLoop++;
	    }
	}
    }

  return nbLoop;
}

// remove the loops from the trajectory
void STNHDeformation::removeLoop(int curStep)
{
  // locate the loops along the trajectory
  vector<int*> vecLoop;
  int nbLoop = this->locateLoop(curStep,vecLoop);

//   cout << "nbLoop : " << nbLoop << endl;

//   for(int i=0; i<nbLoop; i++)
//     {

  if(nbLoop >= 1)
    {
      // get back the initial and final states
      int ind0 = vecLoop[0][0] - 2;
      int indf = vecLoop[0][1] + 2;

      if(ind0 <= curStep)
	{
	  // cannot remove the loop because the car-like is already travelling on it
	  cout << "A loop has been let unchanged (car-like already travelling on it...)" << endl;
	  return;
	}

      if(indf >= this->traj_->getSize()-1)
	{
	  // cannot remove a loop on the final state of the trajectory
	  cout << "A loop has been let unchanged (Include the final state...)" << endl;
	  return;
	}

      StateS6D s0 = *(this->traj_->at(ind0));
      StateS6D sf = *(this->traj_->at(indf));

//       cout << "///////////////////////////////" << endl;
//       cout << "replacement traj computation : " << endl;
//       cout << "s0 : "; s0.display(); cout << endl;
//       cout << "sf : "; sf.display(); cout << endl;
      
      // try to generate a trajectory between both states
      StateS6D sres;
      Vec_DP* param = NULL;
      int res = this->computeItdStateT_Strict(s0,sf,sres,&param);
      
      // check if a trajectory has been found
      if(res == 0)
	{
	  // a trajectory has been found : 
	  // remove the loop and add the new intermediate state
	  this->deleteStateFromTraj(ind0 + 1, indf);
	  this->insertStateInTraj(sres,ind0 + 1,&param,T_FEA);

	  cout << "A loop has been removed" << endl;
	}
      else
	{
	  // no trajectory has been found
	  // try to compute an alternative trajectory

	  // reinit the parameters
	  param = NULL;

	  // compute the distance between p0 and pf
	  double dx = sf.x() - s0.x();
	  double dy = sf.y() - s0.y();
	  double dist = sqrt(dx*dx + dy*dy);

	  // compute an average velocity between v0 and vf
	  double vmoy = 0.5 * (s0.s() + sf.s());

	  // compute the time to travel dist with the velocity vmoy
	  double tmoy;
	  if(vmoy <= 0.0)
	    {
	      // the average time cannot be computed
	      cout << "STNHDeformation::removeLoop : Warning. The average time cannot be computed" << endl;
	      return;
	    }
	  else
	    {
	      tmoy = dist / vmoy;
	    }
	  
	  double oldDT = sf.t() - s0.t();
	  sf.setT(s0.t() + 1.2*tmoy);

// 	  cout << "///////////////////////////////" << endl;
// 	  cout << "replacement traj computation : " << endl;
// 	  cout << "s0 : "; s0.display(); cout << endl;
// 	  cout << "sf : "; sf.display(); cout << endl;

	  int res2 = this->computeItdStateT_Strict(s0,sf,sres,&param);

// 	  cout << "///////////////////////////////" << endl;

	  if(res2 == 0)
	    {
	      // a trajectory has been found : 
	      // remove the loop and add the new intermediate state
	      this->deleteStateFromTraj(ind0 + 1, indf);
	      this->insertStateInTraj(sres,ind0 + 1,&param,T_FEA);
	      
	      // shift the end of the trajectory
	      traj_->shift(ind0 + 2, oldDT - 1.2*tmoy);

	      cout << "A loop has been removed (alternative solution)" << endl;
	    }
	  else
	    {
	      cout << "A loop has been let unchanged (no solution found to remove it...)" << endl;
	    }
	}
    }
}

// insert a state in the deformed trajectory
void STNHDeformation::insertStateInTraj(const StateS6D& s,int pos,Vec_DP** param,int fea)
{
  // copy the input state
  StateR* itdS = new StateS6D(s);
  // insert it in the trajectory
  traj_->insert(pos, itdS);
  
  // copy th input control parameters
  Vec_DP* newParam = new Vec_DP(**param);
  // insert it in the list of controls
  list <Vec_DP*>::iterator it = listCtrl_->begin();
  for(int i=0; i<pos && it != listCtrl_->end(); i++) it++;
  listCtrl_->insert(it,newParam);

  // insert the feasibility status in the corresponding list
  list <int>::iterator it2 = feaCtrl_->begin();
  for(int i=0; i<pos && it2 != feaCtrl_->end(); i++) it2++;
  feaCtrl_->insert(it2,fea);
}

// delete one or several states from the trajectory
// including start but not including end
void STNHDeformation::deleteStateFromTraj(int start,int end)
{
  // check if start and end indices are consistant
  if((start < 0) || (end > traj_->getSize()) || (end <= start))
    {
      cout << "STNHDeformation::deleteStateFromTraj : Warning. start and end indices are not consistent" << endl;
      cout << "start : " << start << "   end : " << end << endl;
      exit(1);
    }

  // remove states from the trajectory
  traj_->remove(start,end);

  // remove control parameters from the list of controls
  list <Vec_DP*>::iterator itC1 = listCtrl_->begin();  
  list <Vec_DP*>::iterator itC2 = listCtrl_->begin();
  for(int i=0; i<start; i++) itC1++;
  for(int i=0; i<end; i++) itC2++;
  listCtrl_->erase(itC1,itC2);

  // remove feasibility status from the corresponding list
  list <int>::iterator itF1 = feaCtrl_->begin();  
  list <int>::iterator itF2 = feaCtrl_->begin();
  for(int i=0; i<start; i++) itF1++;
  for(int i=0; i<end; i++) itF2++;
  feaCtrl_->erase(itF1,itF2);
}


/******************* Utility functions *******************/

// initialize the vector of constraints
void STNHDeformation::initConstraints()
{
  if(this->rb_ != NULL)
    {
      constraints_ = Vec_DP(6); 
      constraints_[0] = 0.5;
      constraints_[1] = rb_->getAMax();
      constraints_[2] = rb_->getVMax();
      constraints_[3] = rb_->getLength();
      constraints_[4] = rb_->getMaxSteering();
      constraints_[5] = rb_->getMaxSteerSpd();
    }
  else
    {
      cout << "STNHDeformation::initConstraints : Warning. The input robotic system has not been initialized" << endl;
      exit(1);
    }
}
