#include "TrajDeform.hpp"
using namespace mpc;

#include "gnuplot/inc/gnuplot_i.hpp"

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /**************** Class TrajDeform ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
TrajDeform::TrajDeform()
{
  this->traj_           = NULL;
  this->env_            = NULL;
  this->sys_            = NULL;

  this->curStateT_      = NULL;   // is it really used?

  this->K_INT           = 1.0;
  this->K_EXT           = 2.0;
  this->K_STEXT         = 2.0;
  this->K_SPT_TMP       = 3.0;
  
  this->alphaMin        = 0.5;
  this->alphaMax        = 2.0;
  this->betaMin         = 1.0;
  this->betaMax         = 10.0;
  
  this->spMed           = 3.0;
  this->tmpMed          = 5.0;
  this->maxStEffectDist = 3.0;

  this->reliabilityTime = 12.0;
  this->epsilonRT       = 0.05;
  this->predHorizon_    = 3.0;

  this->oldTime_        = 0;
  this->timeStep_       = 0.5;
  this->posStep_        = 1.5;
  this->modeDeform_     = SPATIOTEMP_DEFORM;
  this->lastDeformDuration_ = 0.0;
  this->useResampling_  = true;

  this->typeState_      = NULL;
  this->tableNearState_ = NULL;
  this->tableNearStState_ = NULL;

  this->setDebug(0);

  this->mutexTraj2_     = NULL;
}

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param typeS : pointer to the table of state types (for simulation display)
 *  \param tns : pointer to the table of nearest states (for simulation display)
 *  \param tfi : pointer to the table of internal forces (for simulation display)
 *  \param tfe : pointer to the table of external forces (for simulation display)
 */
TrajDeform::TrajDeform(STraj* traj,Environment* env,System* sys)
{
  this->traj_           = traj;
  this->env_            = env;
  this->sys_            = sys;

  this->curStateT_      = NULL;   // is it really used?

  this->K_INT           = 1.0;
  this->K_EXT           = 2.0;
  this->K_STEXT         = 2.0;
  this->K_SPT_TMP       = 3.0;
  
  this->alphaMin        = 0.5;
  this->alphaMax        = 2.0;
  this->betaMin         = 1.0;
  this->betaMax         = 10.0;
  
  this->spMed           = 3.0;
  this->tmpMed          = 5.0;
  this->maxStEffectDist = 3.0;

  this->reliabilityTime = 12.0;
  this->epsilonRT       = 0.05;
  this->predHorizon_    = 3.0;

  this->oldTime_        = 0;
  this->timeStep_       = 0.5;
  this->posStep_        = 1.5;
  this->modeDeform_     = SPATIOTEMP_DEFORM;
  this->lastDeformDuration_ = 0.0;
  this->useResampling_  = true;

  this->typeState_      = NULL;
  this->tableNearState_ = NULL;
  this->tableNearStState_ = NULL;

  this->setDebug(0);

  this->mutexTraj2_     = NULL;
}

// /**
//  *  \brief init constructor
//  *  \param traj : the initial trajectory to deform
//  *  \param env : the environment in which the system evolve
//  *  \param sys : the features of the robot system used
//  *  \param typeS : pointer to the table of state types (for simulation display)
//  *  \param tns : pointer to the table of nearest states (for simulation display)
//  *  \param tfi : pointer to the table of internal forces (for simulation display)
//  *  \param tfe : pointer to the table of external forces (for simulation display)
//  */
// TrajDeform::TrajDeform(STraj* traj,Environment* env,System* sys,int** typeS,PositionXYT*** tns,StateTime** tfi,StateTime** tfe)
// {
//   this->traj_           = traj;
//   this->env_            = env;
//   this->sys_            = sys;

//   this->curStateT_      = NULL;   // is it really used?

//   this->K_INT           = 1.0;
//   this->K_EXT           = 2.0;
//   this->K_STEXT         = 2.0;
//   this->K_SPT_TMP       = 3.0;
  
//   this->alphaMin        = 0.5;
//   this->alphaMax        = 2.0;
//   this->betaMin         = 1.0;
//   this->betaMax         = 10.0;
  
//   this->spMed           = 3.0;
//   this->tmpMed          = 5.0;
//   this->maxStEffectDist = 3.0;

//   this->reliabilityTime = 12.0;
//   this->epsilonRT       = 0.05;
//   this->predHorizon_    = 3.0;


//   this->oldTime_        = 0;
//   this->timeStep_       = 0.5;
//   this->posStep_        = 1.5;
//   this->modeDeform_     = SPATIOTEMP_DEFORM;
//   this->lastDeformDuration_ = 0.0;
//   this->useResampling_  = true;

//   this->typeState       = typeS;

//   this->setDebug(1);

// //   this->tableNearState  = (PositionXYT***)tns;
// //   this->tableFInt       = (StateTime**)tfi;
// //   this->tableFExt       = (StateTime**)tfe;

// //   this->debug_          = 1;
// }

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param dbg : the debugging mode used
 */
TrajDeform::TrajDeform(STraj* traj,Environment* env,System* sys,int dbg)
{
  this->traj_             = traj;
  this->env_              = env;
  this->sys_              = sys;

  this->curStateT_        = NULL;   // is it really used?

  this->K_INT             = 1.0;
  this->K_EXT             = 2.0;
  this->K_STEXT           = 2.0;
  this->K_SPT_TMP         = 3.0;
  
  this->alphaMin          = 0.5;
  this->alphaMax          = 2.0;
  this->betaMin           = 1.0;
  this->betaMax           = 10.0;
  
  this->spMed             = 3.0;
  this->tmpMed            = 5.0;
  this->maxStEffectDist   = 3.0;

  this->reliabilityTime   = 12.0;
  this->epsilonRT         = 0.05;
  this->predHorizon_      = 3.0;

  this->oldTime_          = 0;
  this->timeStep_         = 0.5;
  this->posStep_          = 1.5;
  this->modeDeform_       = SPATIOTEMP_DEFORM;
  this->lastDeformDuration_ = 0.0;
  this->useResampling_    = true;

  this->typeState_        = NULL;
  this->tableNearState_   = NULL;
  this->tableNearStState_ = NULL;

  this->setDebug(dbg);

  this->mutexTraj2_       = NULL;
}

/**
 *  \brief destructor
 */
TrajDeform::~TrajDeform()
{
//   if(tableNearState_ != NULL) delete [] tableNearState_;
//   if(tableNearStState_ != NULL) delete [] tableNearStState_;
}


/*************** Getters and Setters **************/

/**
 *  \brief set the coefficients of internal and external forces and the balance between external spatial and temporal forces
 *  \param kint : pointer to the internal forces coefficient
 *  \param kext : pointer to the external forces coefficient for dynamic obstacles
 *  \param kstext : pointer to the external forces coefficient for static obstacles
 *  \param amin : pointer to the minimal spatial coefficient
 *  \param amax : pointer to the maximal spatial coefficient
 *  \param bmin : pointer to the minimal temporal coefficient
 *  \param bmax : pointer to the maximal temporal coefficient
 */
void TrajDeform::setForcesCoef(double kint, double kext, double kstext, double amin, double amax, double bmin, double bmax)
{
  this->K_INT     = kint;
  this->K_EXT     = kext;
  this->K_STEXT   = kstext;
  this->alphaMin  = amin;
  this->alphaMax  = amax;
  this->betaMin   = bmin;
  this->betaMax   = bmax;
}
       
/**
 *  \brief set the maximal effective distances
 *  \param spMed : pointer to the maximal effective spatial distance from dynamic obstacles
 *  \param tmpMed : pointer to the maximal effective temporal distance from dynamic obstacles
 *  \param staticMed : pointer to the maximal effective spatial distance from static obstacles
 */
void TrajDeform::setMaxEffectDist(double spMed,double tmpMed, double staticMed)
{
  this->spMed           = spMed;
  this->tmpMed          = tmpMed;
  this->maxStEffectDist = staticMed;
}

/**
 *  \brief set the prediction horizon, the reliability time and the associated epsilon value
 *  \param predH : pointer to the prediction horizon time (time until which we consider the forecast model to be completely reliable)
 *  \param rtime : pointer to the reliability time (time until which we consider the forecast model to be partialy reliable)
 *  \param eps : pointer to the standard deviation of the gaussian model
 */
void TrajDeform::setReliability(double predH, double rtime, double eps)
{
  this->predHorizon_ = predH;
  this->reliabilityTime = rtime;
  this->epsilonRT = eps;
}


/*************** Forces computation ***************/

/**
 *  \brief compute the total external forces
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 *  \param timeEnv : the current time of the environment
 */
PositionXYT TrajDeform::computeFExt(const StateTime& cur, Environment* env, double time, bool last,double timeEnv)
{
  PositionXYT dynFExt = this->computeDynFExt2(cur,env,last,time,timeEnv);
  PositionXYT stFExt = this->computeStFExt2(cur,env,last);
  PositionXYT FExt = (dynFExt * (K_EXT)) + (stFExt * (K_STEXT));

  cout << "dynFExt : "; dynFExt.display();
  cout << "  stFExt : "; stFExt.display();
  cout << "  FExt : "; FExt.display();
//   cout << "  K_EXT : " << K_EXT << "  K_STEXT : " << K_STEXT;
//   cout << "  spMed : " << spMed;
//   cout << "  tmpMed : " << tmpMed;
//   cout << "  maxStEffectDist : " << maxStEffectDist;
  cout << endl;

  return FExt;
}

/**
 *  \brief compute the total external forces (due to dynamic obstacles)
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 *  \param timeEnv : the current time of the environment
 *
 *  Compute the resulting force from the nearest obstacle
 */
PositionXYT TrajDeform::computeDynFExt(const StateTime& cur, Environment* env, double time, bool last,double timeEnv)
{
  cout << "computeDynFExt  cur:"; cur.display(); cout << "  time:" << time << "  timeEnv:" << timeEnv << endl;

//   cout << "timeEnv : " << timeEnv << endl;
  // get a position in the workspace from the input state-time
  PositionXYT curPos = cur.posXYT();

  // check if the current state is not too far from the current time to be under external forces
  if(curPos.t() - time > (reliabilityTime))
    {
      PositionXYT force;
      if(fabs(force.t()) > 10.0) 
	{ cout << "force0 : "; force.display(); cout << endl; }
      return force;
    }

  // init needed variables
  double Dmax = spMed;
  double Tmax = tmpMed;
//   double MDmax = Dmax * (alphaMax);
//   double MTmax = Tmax * (betaMax);
  
  // compute spatial coefficient Alpha (for the metric)
  double Alpha = alphaMin + ((reliabilityTime) - curPos.t() + time) * ((alphaMax - alphaMin) / (reliabilityTime));
  
  // compute spatial coefficient Alpha (for the metric)
  double Beta = betaMax - (curPos.t() - time) * ((betaMax - betaMin) / (reliabilityTime));

//   cout << "  Dmax : " << Dmax << "  Tmax : " << Tmax << "  Alpha : " << Alpha << "  Beta : " << Beta; 

  // search for the nearest obstacle
  int nearestObs = -1;
  double sqMminDist = 0;
  PositionXYT nearestState; 
  double nearestRad = 0;
  
  if(this->modeDeform_ == SPATONLY_DEFORM)
    {
      cout << "TrajDeform::computeDynFExt : SPATONLY_DEFORM. Not implemented yet!" << endl;
      exit(1);

//       int N = this->env_->getNbObs2D();

//       for(int i=0; i<N; i++)
// 	{
// 	  // get back the current obstacle
// 	  DiskObs2D* curObs = (DiskObs2D*) this->env_->getObs2D(i);
	  
// 	  double radObs = curObs->getRadius();
	  
// 	  // get back the last position of the obstacle
// 	  PositionXYThT posObs;
// 	  bool defined = curObs->getLastPos(posObs);
	  
// 	  if(defined)
// 	    {
// 	      PositionXYT curRbNearObs(posObs.x() - curPos.x(), posObs.y() - curPos.y(), 0);

// 	      double sqMminDistObs = (Alpha*Alpha*curRbNearObs.x()*curRbNearObs.x() + Alpha*Alpha*curRbNearObs.y()*curRbNearObs.y());
	      
// 	      // check if this obstacle is the nearest obstacle from the current robot's control point
// 	      if(nearestObs == -1 || sqMminDistObs < sqMminDist)
// 		{
// 		  sqMminDist = sqMminDistObs;
// 		  nearestObs = i;
// 		  nearestState = PositionXYT(posObs.x(),posObs.y(),curPos.t());
// 		}
// 	    }
// 	}
    }
  else
    {  
//       cout << "env_->getNbObs2D() : " << env_->getNbObs2D() << "  ";

      for(int i=0; i<this->env_->getNbObs2D(); i++)
	{
	  // get back the current obstacle
// 	  DiskObs2D curObs = DiskObs2D(*((DiskObs2D*) &(this->env_->getObs2D(i))));
	  DiskObs2D* curObs = ((DiskObs2D*)(this->env_->getObs2D(i)));
	  
	  double radObs = curObs->radius();  
	  double sqMDTmax = (Alpha*Alpha*(Dmax + radObs + sys_->getRadius())*(Dmax + radObs + sys_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  
	  // get back the last position of the obstacle
	  PositionXYThT posObs(curObs->lastPosT());
	  if(timeEnv != -1)
	    {
// 	      cout << "  oldPosObs.t():" << posObs.t();
	      posObs.t() = posObs.t() - timeEnv + time;
// 	      cout << "  newPosObs.t():" << posObs.t() << "  timeEnv:" << timeEnv << "  time:" << time;
	    }
	  
	  // compute the coordinates of the obstacle's last control point in the metric frame
	  PositionXYT MconfigObs(Alpha * posObs.x(), Alpha * posObs.y(), Beta * posObs.t());
	  
	  // compute the coordinates of the robot's control point in the metric frame
	  PositionXYT MconfigRb(Alpha * curPos.x(), Alpha * curPos.y(), Beta * curPos.t());
	  
	  // compute the robot's speed vector in the metric frame
	  PositionXYT MspeedObs(Alpha * curObs->lastSpd().x(), Alpha * curObs->lastSpd().y(), Beta);
	  
	  // compute the vector configRobot - configObs with the metric
	  PositionXYT MORi (MconfigRb.x() - MconfigObs.x(), MconfigRb.y() - MconfigObs.y(), MconfigRb.t() - MconfigObs.t());
	  
	  // compute the norm of MspeedRb
	  double MnormSObs = sqrt(MspeedObs.x()*MspeedObs.x() + MspeedObs.y()*MspeedObs.y() + MspeedObs.t()*MspeedObs.t());
	  
	  // compute the normalized vector NMspeedRb
	  PositionXYT NMspeedObs;
	  if(MnormSObs != 0)
	    {
	      NMspeedObs.x() = MspeedObs.x()/MnormSObs;
	      NMspeedObs.y() = MspeedObs.y()/MnormSObs;
	      NMspeedObs.t() = MspeedObs.t()/MnormSObs;
	    }
	  
	  // compute the scalar product between MORi and nSObs
	  double pdScal = MORi.x() * NMspeedObs.x() + MORi.y() * NMspeedObs.y() + MORi.t() * NMspeedObs.t();
	  
	  //compute the vector OP (obstacle's nearest control point - obstacle's init control point) in the metric frame
	  PositionXYT NOP(pdScal * NMspeedObs.x(), pdScal * NMspeedObs.y(), pdScal * NMspeedObs.t());
	  
	  // compute the vector OP in the normal frame
	  PositionXYT OP(NOP.x()/Alpha, NOP.y()/Alpha, NOP.t()/Beta);
	  
	  // compute the obstacle's nearest position
	  PositionXYT nearestStateCurObs(posObs.x() + OP.x(), posObs.y() + OP.y(), posObs.t() + OP.t());
	  
	  // compute the vector nearestStateCurObs - curRb
	  PositionXYT curRbNearObs(nearestStateCurObs.x() - curPos.x(), nearestStateCurObs.y() - curPos.y(), nearestStateCurObs.t() - curPos.t());
	  
	  // compute the norm of curRbNearObs ie. the minimal distance from the current robot's control point to the current obs
	  double sqMminDistObs = (Alpha*Alpha*curRbNearObs.x()*curRbNearObs.x() + Alpha*Alpha*curRbNearObs.y()*curRbNearObs.y() + Beta*Beta*curRbNearObs.t()*curRbNearObs.t());
	  
 	  cout << "  obs" << i << " : "; posObs.display(); cout << "->"; nearestStateCurObs.display(); 
	  cout << " : (minDist:" << sqMminDistObs << ",MDTmax:" << sqMDTmax << ")" << endl;

	  // check if this obstacle is the nearest obstacle from the current robot's control point
	  if((sqMminDistObs < sqMDTmax) && (nearestObs == -1 || sqMminDistObs < sqMminDist))
	    {
	      sqMminDist = sqMminDistObs;
	      nearestObs = i;
	      nearestState = nearestStateCurObs;
	      nearestRad = radObs;
	    }

	  delete curObs;
	}
    }
  
  // compute the force to apply to the current state
  if(nearestObs != -1)
    {
      cout << "  nearest : "; nearestState.display(); cout << endl;

      {
	// block the access to the displaying tables
// 	cout << "TrajDeform::computeDynFExt : waiting for mutex Tables" << endl;
	boost::mutex::scoped_lock lock(mutexTables_);
// 	cout << "TrajDeform::computeDynFExt : mutex Tables locked" << endl;
	
	if(debug_ >= 1 && tableNearState_ != NULL)
	  {
	    // add the nearest state to the corresponding display table entry
	    if(tableNearState_[I] != NULL) delete tableNearState_[I];
	    tableNearState_[I] = new PositionXYT(nearestState);
	  }
      }

      // compute the coefficient due to the reliability time
      double sigma = sqrt(-0.5 * (reliabilityTime) * (reliabilityTime) / log(epsilonRT));
      double coefRT = exp(-0.5 * (time - curPos.t())/sigma * (time - curPos.t())/sigma );

//       cout << "  coefRT : " << coefRT << endl;

      // compute the direction of the force
      double dx = (curPos.x() - nearestState.x());
      double dy = (curPos.y() - nearestState.y());
      double dt = (curPos.t() - nearestState.t());
      double sqdd = dx*dx + dy*dy;
      double length = sqrt(dx*dx + dy*dy + dt*dt);
      double Ndd = 0;

//       cout << "  dx : " << dx << "  dy : " << dy << "  dt : " << dt << "  length : " << length;

      PositionXYT dir;
      if(length != 0 && !last)
 	{
	  dir = PositionXYT(dx/length,dy/length,dt/length);
 	}
      else if(length != 0 && last)
 	{
 	  dir = PositionXYT(0,0,dt/length);
 	}
      else
 	dir = PositionXYT();
      
      // check if this state is in collision
      double radObs = nearestRad; // ((DiskObs2D*)&(this->env_->getObs2D(nearestObs)))->radius();
      double radRb = sys_->getRadius();
      PositionXYT forceNoCol;
      if(!last && sqdd < (radObs + radRb)*(radObs + radRb))
	{
	  double dd = sqrt(sqdd);
	  double ddx = (dx * (radObs +radRb) / dd) - dx;
	  double ddy = (dy * (radObs +radRb) / dd) - dy;
	  //forceNoCol = PositionXYT(ddx,ddy,0);
	  forceNoCol = PositionXYT();
	}

      // compute the base external force
      double val = (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));
      PositionXYT baseExtForce;
      if(val > 0) baseExtForce = dir * (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));

//       cout << "  val : " << val << "  baseExtForce : "; baseExtForce.display(); 

      if(this->modeDeform_ == SPATONLY_DEFORM)
	{
	  PositionXYT force(baseExtForce.x() + forceNoCol.x(),
			    baseExtForce.y() + forceNoCol.y(),
			    0);
	  force = force * coefRT;
	  return force;
	}
      else if(this->modeDeform_ == TEMPONLY_DEFORM)
	{
	  PositionXYT force(0,
			    0,
			    baseExtForce.t() + forceNoCol.t());
	  force = force * coefRT;
	  return force;
	}
      else
	{
	  // compute the force to apply
	  PositionXYT force(baseExtForce.x() + forceNoCol.x(),
			    baseExtForce.y() + forceNoCol.y(),
			    baseExtForce.t() + forceNoCol.t());
// 	  cout << "  force1 : "; force.display();

	  force = force * coefRT;
// 	  cout << "  force : "; force.display();
	  return force;
	}
    }
  else
    {
//       cout << "no close obs found..." << endl;

      {
	// block the access to the displaying tables
// 	cout << "TrajDeform::computeDynFExt : waiting for mutex Tables" << endl;
	boost::mutex::scoped_lock lock(mutexTables_);
// 	cout << "TrajDeform::computeDynFExt : mutex Tables locked" << endl;

	if(debug_ >= 1 && tableNearState_ != NULL)
	  {
	    if(tableNearState_[I] != NULL) delete tableNearState_[I];
	    tableNearState_[I] = NULL;
	  }
      }
      
      if(last && (this->modeDeform_ != SPATONLY_DEFORM))
 	{
 	  if(useTemporalConsistency_)
 	    {
	      PositionXYT force;
// 	      force.t() = -0.5;
	      return force;
	    }
 	  else
 	    {
	      PositionXYT force;
 	      return force;
 	    }
 	}
      else
	{
	  PositionXYT force;
	  return force;
	}
    }
}

/**
 *  \brief compute the total external forces (due to dynamic obstacles)
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param last : true if the current node is the last node of the trajectory
 *  \param timeTraj : the current time on the followed trajectory
 *  \param timeEnv : the current time registered by the environment perception process
 *
 *  Compute the resulting force as the sum of all forces from the obstacles
 *  surrounding the robot in a close radius
 */
PositionXYT TrajDeform::computeDynFExt2(const StateTime& cur, Environment* env, bool last, double timeTraj, double timeEnv)
{
//   cout << "computeDynFExt2  cur:"; cur.display(); cout << "  time:" << timeTraj << "  timeEnv:" << timeEnv;

  // get a position in the workspace from the input state-time
  PositionXYT curPos = cur.posXYT();
  // shift the associated time to obtain a relative time to the current one
  curPos.t() -= timeTraj;

//   cout << "curPos : "; curPos.display(); cout << endl;

  // check if the current state is not too far from the current time to be under external forces
  if(curPos.t() > reliabilityTime) return PositionXYT();

  // compute the metric coefficent between time and space
  double Dmax = spMed;
  double Tmax = tmpMed;
  
//   cout << "  Dmax:" << Dmax << "  Tmax:" << Tmax << endl;

//   alphaMin = 1;
//   alphaMax = 1;
//   betaMin = 10;
//   betaMax = 10;

  // compute spatial coefficient Alpha (for the metric)
  double Alpha = alphaMin + (reliabilityTime - curPos.t()) * ((alphaMax - alphaMin) / reliabilityTime);
  
  // compute temporal coefficient Beta (for the metric)
  double Beta = betaMax - curPos.t() * ((betaMax - betaMin) / reliabilityTime);

  // search for the nearest obstacle
  int nearestObs = -1;
  double sqMminDist = 0;
  PositionXYT nearestState; 
  double nearestRad = 0;

  for(int i=0; i<this->env_->getNbObs2D(); i++)
    {
      // get back the current obstacle
      DiskObs2D* curObs = ((DiskObs2D*)(this->env_->getObs2D(i)));
	  
//       cout << "i : " << i << "  "; curObs->display(); 

      double radObs = curObs->radius();  
      double sqMDTmax = (Alpha*Alpha*(Dmax + radObs + sys_->getRadius())*(Dmax + radObs + sys_->getRadius()) + Beta*Beta*Tmax*Tmax);
	  
      // get back the last position of the obstacle
      PositionXYThT posObs(curObs->lastPosT());
      if(timeEnv != -1) posObs.t() -= timeEnv;
	  
//       cout << "obs" << i << " -> posObs : "; posObs.display(); cout << endl;

      // compute the coordinates of the obstacle's last control point in the metric frame
      PositionXYT MconfigObs(Alpha * posObs.x(), Alpha * posObs.y(), Beta * posObs.t());
	  
      // compute the coordinates of the robot's control point in the metric frame
      PositionXYT MconfigRb(Alpha * curPos.x(), Alpha * curPos.y(), Beta * curPos.t());
	  
      // compute the robot's speed vector in the metric frame
      PositionXYT MspeedObs(Alpha * curObs->lastSpd().x(), Alpha * curObs->lastSpd().y(), Beta);
	  
      // compute the vector configRobot - configObs with the metric
      PositionXYT MORi (MconfigRb.x() - MconfigObs.x(), MconfigRb.y() - MconfigObs.y(), MconfigRb.t() - MconfigObs.t());
	  
      // compute the norm of MspeedRb
      double MnormSObs = sqrt(MspeedObs.x()*MspeedObs.x() + MspeedObs.y()*MspeedObs.y() + MspeedObs.t()*MspeedObs.t());
	  
      // compute the normalized vector NMspeedRb
      PositionXYT NMspeedObs;
      if(MnormSObs != 0)
	{
	  NMspeedObs.x() = MspeedObs.x()/MnormSObs;
	  NMspeedObs.y() = MspeedObs.y()/MnormSObs;
	  NMspeedObs.t() = MspeedObs.t()/MnormSObs;
	}
	  
      // compute the scalar product between MORi and nSObs
      double pdScal = MORi.x() * NMspeedObs.x() + MORi.y() * NMspeedObs.y() + MORi.t() * NMspeedObs.t();
	  
      //compute the vector OP (obstacle's nearest control point - obstacle's init control point) in the metric frame
      PositionXYT NOP(pdScal * NMspeedObs.x(), pdScal * NMspeedObs.y(), pdScal * NMspeedObs.t());
	  
      // compute the vector OP in the normal frame
      PositionXYT OP(NOP.x()/Alpha, NOP.y()/Alpha, NOP.t()/Beta);
	  
      // compute the obstacle's nearest position
      PositionXYT nearestStateCurObs; 
      if(OP.t() > 0) nearestStateCurObs = PositionXYT(posObs.x() + OP.x(), posObs.y() + OP.y(), posObs.t() + OP.t());
      else nearestStateCurObs = PositionXYT(posObs.x(), posObs.y(), posObs.t());

      // compute the vector nearestStateCurObs - curRb
      PositionXYT curRbNearObs(nearestStateCurObs.x() - curPos.x(), nearestStateCurObs.y() - curPos.y(), nearestStateCurObs.t() - curPos.t());
	  
      // compute the norm of curRbNearObs ie. the minimal distance from the current robot's control point to the current obs
      double sqMminDistObs = (Alpha*Alpha*curRbNearObs.x()*curRbNearObs.x() + Alpha*Alpha*curRbNearObs.y()*curRbNearObs.y() + Beta*Beta*curRbNearObs.t()*curRbNearObs.t());
	  
//       cout << "  obs" << i << " : "; posObs.display(); cout << "->"; nearestStateCurObs.display(); 
//       cout << " : (minDist:" << sqMminDistObs << ",MDTmax:" << sqMDTmax << ")" << endl;

      // check if this obstacle is the nearest obstacle from the current robot's control point
      if((sqMminDistObs < sqMDTmax) && (nearestObs == -1 || sqMminDistObs < sqMminDist))
	{
	  sqMminDist = sqMminDistObs;
	  nearestObs = i;
	  nearestState = nearestStateCurObs;
	  nearestRad = radObs;
	}

      delete curObs;
    }
  
  // compute the force to apply to the current state
  if(nearestObs != -1)
    {
//       cout << "  nearest : "; nearestState.display(); cout << endl;

      {
	// block the access to the displaying tables
// 	cout << "TrajDeform::computeDynFExt2 : waiting for mutex Tables" << endl;
	boost::mutex::scoped_lock lock(mutexTables_);
// 	cout << "TrajDeform::computeDynFExt2 : mutex Tables locked" << endl;

	if(debug_ >= 1 && tableNearState_ != NULL)
	  {
	    // add the nearest state to the corresponding display table entry
	    if(tableNearState_[I] != NULL) delete tableNearState_[I];
	    tableNearState_[I] = new PositionXYT(nearestState);
	    if(timeEnv != -1) tableNearState_[I]->t() += timeEnv;
	  }
      }

      // compute the coefficient due to the reliability time
//       double sigma = sqrt(-0.5 * (reliabilityTime) * (reliabilityTime) / log(epsilonRT));
//       double coefRT = exp(-0.5 * curPos.t()/sigma * curPos.t()/sigma );
      double coefRT = this->getCoefReliability(curPos.t());
      
//       cout << "  coefRT : " << coefRT << "  reliabilityTime : " << reliabilityTime << "  epsilonRT : " << epsilonRT << "  predHorizon_ : " << predHorizon_ << endl;

      // compute the direction of the force
      double dx = (curPos.x() - nearestState.x());
      double dy = (curPos.y() - nearestState.y());
      double dt = (curPos.t() - nearestState.t());
      double sqdd = dx*dx + dy*dy;
      double length = sqrt(dx*dx + dy*dy + dt*dt);

//       cout << "  dx : " << dx << "  dy : " << dy << "  dt : " << dt << "  length : " << length;

      PositionXYT dir;
      if(length != 0 && !last)
 	{
	  dir = PositionXYT(dx/length,dy/length,dt/length);
 	}
      else if(length != 0 && last)
 	{
 	  dir = PositionXYT(0,0,dt/length);
 	}
      else
 	dir = PositionXYT();
      
      // check if this state is in collision
      double radObs = nearestRad; // ((DiskObs2D*)&(this->env_->getObs2D(nearestObs)))->radius();
      double radRb = sys_->getRadius();
      PositionXYT forceNoCol;
      if(!last && sqdd < (radObs + radRb)*(radObs + radRb))
	{
	  double dd = sqrt(sqdd);
	  double ddx = (dx * (radObs +radRb) / dd) - dx;
	  double ddy = (dy * (radObs +radRb) / dd) - dy;
	  //forceNoCol = PositionXYT(ddx,ddy,0);
	  forceNoCol = PositionXYT();
	}

      // compute the base external force
      double val = (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));
      PositionXYT baseExtForce;
      if(val > 0) baseExtForce = dir * (1 - sqrt((dt*dt/(Tmax*Tmax)) + (sqdd/(Dmax*Dmax))));

      // compute the force to apply
      PositionXYT force(baseExtForce.x() + forceNoCol.x(),
			baseExtForce.y() + forceNoCol.y(),
			baseExtForce.t() + forceNoCol.t());
      
      force = force * coefRT;
      return force;
    }
  else
    {
//       cout << "no close obs found..." << endl;

      {
	// block the access to the displaying tables
// 	cout << "TrajDeform::computeDynFExt2 : waiting for mutex Tables" << endl;
	boost::mutex::scoped_lock lock(mutexTables_);
// 	cout << "TrajDeform::computeDynFExt2 : mutex Tables locked" << endl;

	if(debug_ >= 1 && tableNearState_ != NULL)
	  {
	    if(tableNearState_[I] != NULL) delete tableNearState_[I];
	    tableNearState_[I] = NULL;
	  }
      }
      
      PositionXYT force;
      return force;
    }
}

/**
 *  \brief compute the total external forces (due to static obstacles)
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 *
 *  Compute the resulting force from the nearest obstacle
 */
PositionXYT TrajDeform::computeStFExt(const StateTime& cur, Environment* env, bool last)
{
  cerr << "TrajDeform::computeStFExt : Not implemented anymore..." << endl;
  exit(1);

//   if(last)
//     {
//       return PositionXYT();
//     }
//   else
//     {
//       // get a position in the workspace from the input state-time
//       PositionXYT curPos = cur.posXYT();

//       // get back the maximal effective distance
//       //double Dmax = *spMed * 0.8 + this->rb_->getRadius();
//       double Dmax = (maxStEffectDist) + this->sys_->getRadius();
      
//       // check if the environment is not null
//       if(this->env_ == NULL || this->env_->getNbStObs() <= 0)
// 	{
// 	  return PositionXYT();
// 	}

//       int nbObs = this->env_->getNbStObs();
//       double minDist = -1;
//       PositionXYT nearest;
//       PositionXYThT centerNearest;
//       bool collision = false;

//       // search for the nearest static obstacle from cur
//       for(int i=0; i<nbObs; i++)
// 	{
// 	  // get back the current obstacle
// 	  Obstacle2D* obs = this->env_->getStObs(i);

// 	  // compute the nearest position of the current obstacle to cur
// 	  double dist;
// 	  PositionXYT near;
// 	  bool inside = obs->nearBdry(curPos, near, dist);

// 	  // check if the current obstacle is the nearest obstacle from cur
// 	  if(inside || (!collision && dist < minDist) || (minDist == -1 && dist < Dmax))
// 	    {
// 	      minDist = dist;
// 	      nearest = near;
// 	      bool defined = obs->getPos(curPos.t(),centerNearest);
// 	      if(!defined) { cout << "TrajDeform::computeStFExt : Waning. Undefined position of the obstacle." << endl; exit(1); }
// 	      if(inside) collision = true;
// 	    }
// 	}

//       // check if a force must be applied
//       if(minDist != -1)
// 	{
// 	  // compute the direction in which the force must be applied
// 	  PositionXY vecDir(0.0,0.0);
// 	  if(collision)
// 	    {
// 	      vecDir.x() = nearest.x() - curPos.x();
// 	      vecDir.y() = nearest.y() - curPos.y();
// 	    }
// 	  else
// 	    {
// 	      vecDir.x() = curPos.x() - nearest.x();
// 	      vecDir.y() = curPos.y() - nearest.y();
// 	    }
// 	  PositionXY dir = vecDir.normalized();

// 	  // compute a small force coming from the center of the obstacle
// 	  PositionXY vecForceCenter(curPos.x() - centerNearest.x(),
// 				    curPos.y() - centerNearest.y());
// 	  PositionXY vecNForceCenter = vecForceCenter.normalized();
// 	  PositionXYT forceCenter(0.3*vecNForceCenter.x(),
// 				  0.3*vecNForceCenter.y(),
// 				  0.0);

// 	  if(collision)
// 	    {
// 	      PositionXY forceXY = ((dir * ((this->sys_->getRadius() + fabs(minDist))/Dmax + 1.0)) + forceCenter);
// 	      PositionXYT force(forceXY.x(),forceXY.y(),0.0);
// 	      return force;
// 	    }
// 	  else
// 	    {
// 	      PositionXY forceXY = ((dir * (1.0 - minDist/Dmax)) + forceCenter);
// 	      PositionXYT force(forceXY.x(),forceXY.y(),0.0);
// 	      return force;
// 	    }
// 	}
//       else return PositionXYT();
//     }

//   cout << "TrajDeform::computeStFExt : The process should not arrive here" << endl;
//   exit(1);
}

/**
 *  \brief compute the total external forces due to static obstacles (2nd version -> force applied from each static obstacle)
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 *
 *  Compute the resulting force as the sum of all forces from the obstacles
 *  surrounding the robot in a close radius
 */
PositionXYT TrajDeform::computeStFExt2(const StateTime& cur, Environment* env, bool last)
{
//   cerr << "TrajDeform::computeStFExt2 : Not implemented anymore..." << endl;
//   exit(1);

  if(last)
    {
      return PositionXYT();
    }
  else
    {
      // get a position in the workspace from the input state-time
      PositionXYT curPos = cur.posXYT();

      // get back the maximal effective distance
      double Dmax = (maxStEffectDist) + this->sys_->getRadius();
      
      // check if the environment is not null
      if(this->env_ == NULL || env_->getNbFixedObs2D() <= 0)
	{
	  return PositionXYT();
	}

      int nbObs = env_->getNbFixedObs2D();

      // initialize the total force computed
      PositionXY force;

      // compute the force applied by each obstacle
      for(int i=0; i<nbObs; i++)
	{
	  // get back the current obstacle
	  const Obs2D& obs = env_->getFixedObs2D(i);

	  // compute the nearest position of the current obstacle to cur
	  double dist;
	  PositionXY near;
	  bool inside = obs.obstruct(curPos,curPos.t(),near,dist);

	  // check is there is a collision between the current node and the current obstacle
	  // or if the current obstacle is enough close to exert a force on the robotic system

	  if(inside || dist < Dmax)
	    {
	      // compute the direction of the force to apply
	      PositionXY dir;
	      if(inside)
		{
		  // compute the direction between the nearest point of the obstacle and the robot
		  PositionXY vecDir1(near.x() - curPos.x(),
				     near.y() - curPos.y());
		  PositionXY dir1 = vecDir1.normalized();

		  // get the center position of the obstacle
		  PositionXYThT centerObs = obs.lastPosT();

		  // compute the direction between the center of the obstacle and the robot
		  PositionXY vecDir2(centerObs.x() - curPos.x(),
				     centerObs.y() - curPos.y());
		  PositionXY dir2 = vecDir2.normalized();

		  PositionXY vecDir(3.0*dir1.x() + dir2.x(),
				     3.0*dir1.y() + dir2.y());
		  dir = vecDir.normalized();
		}
	      else
		{
		  PositionXY vecDir(curPos.x() - near.x(),
				    curPos.y() - near.y());
		  dir = vecDir.normalized();
		}
	      
	      // compute the force to apply from the current obstacle
	      PositionXY curForce;
	      if(inside)
		{
		  curForce = dir * (1.0 + fabs(dist)/Dmax);
		}
	      else
		{
		  curForce = dir * (1.0 - dist/Dmax);
		}
	      
	      // add this force to the total force to apply
	      force += curForce;
	    }
	}

      return PositionXYT(force.x(),force.y(),0.0);
    }
}

/**
 *  \brief compute the total external forces due to static obstacles (3rd version -> force applied from the closest obstacle)
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 *
 *  Compute the resulting force from the nearest obstacle
 */
PositionXYT TrajDeform::computeStFExt3(const StateTime& cur, Environment* env, bool last)
{
//   cout << "computeStFExt3 -> cur : "; cur.display(); cout << endl;
  if(last)
    {
      return PositionXYT();
    }
  else
    {
      // get a position in the workspace from the input state-time
      PositionXYT curPos = cur.posXYT();

      // compute the maximal effective distance
      double Dmax = (maxStEffectDist) + this->sys_->getRadius();

      // search for the closest obstacle of the 
      // static environment to the current position
      PositionXY near;
      double dist;
      //bool inside = env->obstruct(curPos,curPos.t(),near,dist);
      bool inside = env->staticObstruct(curPos,near,dist);

      if(dist < Dmax)
	{
	  // save the nearest obstacle position for the debugging mode
	  {
	    // block the access to the displaying tables
// 	    cout << "TrajDeform::computeStFExt3 : waiting for mutex Tables" << endl;
	    boost::mutex::scoped_lock lock(mutexTables_);
// 	    cout << "TrajDeform::computeStFExt3 : mutex Tables locked" << endl;
	    
	    if(debug_ >= 1 && tableNearStState_ != NULL)
	      {
		// add the nearest state to the corresponding display table entry
		if(tableNearStState_[I] != NULL) delete tableNearStState_[I];
		tableNearStState_[I] = new PositionXYT(near.x(),near.y(),0.0);
	      }
	  }
	  
	  // compute the force exerted by the closest obstacle
	  if(inside)
	    {
	      // compute the direction between the nearest point of the obstacle and the robot
	      PositionXY vecDir(near.x() - curPos.x(),
				 near.y() - curPos.y());
	      PositionXY dir = vecDir.normalized();

	      PositionXY forceXY = dir * (1.0 + fabs(dist)/Dmax);
	      return PositionXYT(forceXY.x(),forceXY.y(),0.0);
	    }
	  else
	    {
	      PositionXY vecDir(curPos.x() - near.x(),
				curPos.y() - near.y());
	      PositionXY dir = vecDir.normalized();

	      PositionXY forceXY = dir * (1.0 - dist/Dmax);
	      return PositionXYT(forceXY.x(),forceXY.y(),0.0);
	    }
	}
      else 
	{
	  // remove the old nearest obstacle position for the debugging mode
	  {
	    // block the access to the displaying tables
// 	    cout << "TrajDeform::computeStFExt3 : waiting for mutex Tables" << endl;
	    boost::mutex::scoped_lock lock(mutexTables_);
// 	    cout << "TrajDeform::computeStFExt3 : mutex Tables locked" << endl;
	    
	    if(debug_ >= 1 && tableNearStState_ != NULL)
	      {
		// add the nearest state to the corresponding display table entry
		if(tableNearStState_[I] != NULL) delete tableNearStState_[I];
		tableNearStState_[I] = NULL;
	      }
	  }

	  return PositionXYT();
	}
    }
}

/**
 *  \brief compute the reliability coefficient
 *  \return the reliability coefficient
 *
 *  return the reliability coefficient to apply on the forces 
 *  knowing the difference dt between the current time and the 
 *  time of the node on which we apply the force
 */
double TrajDeform::getCoefReliability(double dt) const
{
  double uncerHorizon = (reliabilityTime) - (predHorizon_);
  
  if(dt <= (predHorizon_))
    {
      return 1.0;
    }
  else if(dt <= (reliabilityTime))
    {
      double uncerTime = dt - (predHorizon_);
      return exp(uncerTime*uncerTime*log(epsilonRT)/((uncerHorizon)*(uncerHorizon)));
    }
  else return 0.0;
}


/******* Connectivity and Collision checking ******/

/**
 *  \brief check if the still-to-be executed part of the trajectory is in collision with the environment
 *  \param traj : the trajectory to test with
 *  \param env : the environment of the system
 *  \param curTime : the already elapsed time
 *  \return the first time from the current time where a collision occurs. return -1 if there is no collision
 */
double TrajDeform::checkCollision(STraj* traj,Environment* env,double curTime)
{
//   cerr << "TrajDeform::checkCollision : Not implemented anymore..." << endl;
//   exit(1);

  double res = -1;

  // check if the trajectory is not null
  if(traj == NULL)
    return -1;
  
  // check if the environment is not null
  if(env == NULL)
    return -1;
  
  bool Trouve = false;
  int i=1;
  int NbNode = traj->getSize();

  while(i<NbNode)
    {
      // get a position in WxT from the current state
      PositionXYT curPos = traj->at(i).posXYT();

      // get a position in WxT from the previous state
      PositionXYT prevPos = traj->at(i-1).posXYT();

      // get back the time of the current state
      double x = curPos.x();
      double y = curPos.y();
      double time = curPos.t();

      // test if the current state has not been already visited
      if(time > curTime)
	{
	  bool trouve = false;

	  int NbStObs = env->getNbFixedObs2D();
	  int NbObs = env->getNbObs2D();

// // 	  // test if a static obstacle obstructs the path between the previous and the current state
// // 	  if(NbStObs != 0)
// // 	    {
// // 	      int j=0;
// // 	      while(!trouve && j<NbStObs)
// // 		{
// // 		  // test if the current static obstacle obstructs the path
// // 		  if(env->getStObs(j)->obstructSgmt(prevPos,curPos))
// // 		    {
// // 		      trouve = true;
// // 		      if(res == -1) res = time - curTime;

// // 		      (*typeState)[i] = -1;
// // 		    }
// // 		  else j++;
// // 		}
// // 	    }
	  
// // 	  // test if a dynamic obstacle obstructs the path between the previous and the current state
// // 	  if(NbObs != 0)
// // 	    {
// // 	      int j=0;
// // 	      while(!trouve && j<NbObs)
// // 		{
// // 		  // get back the configuration of the obstacle at time "time"
// // 		  PositionXYThT posObs;
// // 		  bool defined = env->getObs(j)->getForecastPos(time,posObs);

// // 		  if(defined)
// // 		    {
// // 		      // compute the distance between the current state and the current obstacle
// // 		      double dx = x - posObs.x();
// // 		      double dy = y - posObs.y();
// // 		      double sqDist = dx*dx + dy*dy;
		      
// // 		      double minDist = (((DiskObs2D*)&(env->getObs2D(j)))->radius()) + this->sys_->getRadius();
		      
// // 		      if(sqDist < minDist*minDist)
// // 			{
// // 			  trouve = true;
// // 			  //res = time - curTime;
// // 			  if(res == -1) res = 2.0*(time - curTime); // hack to take in account collision with dynamic obstacle
// // 			                                            // as late as possible
			  
// // 			  (*typeState)[i] = -1;
// // 			}
// // 		    }
// // 		  j++;
// // 		}
// // 	    }
	}
      i++;
    }

  return res;
}


/*************** Deformation process **************/

// /**
//  *  \brief deform the current trajectory
//  *  \param time : the current time
//  *  \param curStep : index of the node of the trajectory that has been already reached
//  */
// int TrajDeform::deform(double time, int curStep)
// {
// }

/**
 *  \brief deform the current trajectory
 *  \param time : the current time
 *  \return 0 if the trajectory generation process succeed without any problem
 *  \return -1 if no input trajectory has been defined
 *  \return -2 if a collision may occur during the prediction horizon
 *  \return -3 if the connectivity of the trajectory has been lost
 *  \return -4 if the final time of the trajectory has been already past
 */
int TrajDeform::deform(double time)
{
  // check if the input trajectory hazs been defined
  if(traj_ == NULL) return -1;

  // search for the current step of the trajectory already reached
  bool trouve = false;
  int i=0;
  while(!trouve && i<traj_->size())
    {
      if(traj_->at(i).t() > time) trouve = true;
      else i++;
    }
  
  // check if the final time of the trajectory has been already reached
  if(!trouve)
    {
      return -4;
    }
  else 
    {
      return this->deform(time,i);
    }
}


/**************** Display functions ***************/

/**
 *  \brief display the predictional horizon profile on interval [0;dt]
 *  \param dt : the length of time interval to display
 */
void TrajDeform::displayPredHorizonProfile(double dt) const
{
  //double predHorizon = 3.0;
  double uncerHorizon = (reliabilityTime) - (predHorizon_);

  // initialize both time and reliability vectors
  vector <double> vecTime;
  vector <double> vecReliability;
  
  // compute the number of points to display
  int nbP = ((int)((1.2 * (reliabilityTime)) / dt)) + 1; 

  cout << "prediction horizon : " << (predHorizon_) << endl;
  cout << "reliabilityTime : " << (reliabilityTime) << endl;

  // compute all points
  for(int i=0; i<nbP; i++)
    {
      double time = i*dt;
      
      // compute the coefficient due to the reliability time
      double coefRT = this->getCoefReliability(time);
      
      // add the new values in respective vectors
      vecTime.push_back(time);
      vecReliability.push_back(coefRT);
    }

  // display the graph in a gnuplot window
  Gnuplot gCoefRT;
  gCoefRT.reset_plot();
  gCoefRT.set_style("lines").plot_xy(vecTime,vecReliability,"Reliability Coef of the Previsional Model");

  wait_for_key();
}

/**
 *  \brief display the nearest position-time of the obstacles from each state-time of the trajectory
 */
void TrajDeform::dispNearObs()
{
  const GLfloat blackMaterial[]={0.,0.,0.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);

  // block the access to the displaying tables
//   cout << "TrajDeform::dispNearObs : waiting for mutex Tables" << endl;
  boost::mutex::scoped_lock lock(mutexTables_);
//   cout << "TrajDeform::dispNearObs : mutex Tables locked" << endl;

  int N = traj()->size();

  for(int i=0; i<N; i++)
    {
      if(tableNearState_[i] != NULL)
	{
	  glBegin(GL_LINES);
	  glVertex3f(traj()->at(i).posXYT().x(), traj()->at(i).posXYT().y(), traj()->at(i).posXYT().t());
	  glVertex3f(tableNearState_[i]->x(), tableNearState_[i]->y(), tableNearState_[i]->t());
	  glEnd();
	}
    }
}

/**
 *  \brief display the nearest position-time of the static obstacles from each state-time of the trajectory
 */
void TrajDeform::dispNearStObs()
{
  const GLfloat greyMaterial[]={0.6,0.6,0.6,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, greyMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, greyMaterial);

  int N = traj()->size();

  // block the access to the displaying tables
//   cout << "TrajDeform::dispNearStObs : waiting for mutex Tables" << endl;
  boost::mutex::scoped_lock lock(mutexTables_);
//   cout << "TrajDeform::dispNearStObs : mutex Tables locked" << endl;

  for(int i=0; i<N; i++)
    {
      if(tableNearStState_[i] != NULL)
	{
	  glBegin(GL_LINES);
	  glVertex3f(traj()->at(i).posXYT().x(), traj()->at(i).posXYT().y(), traj()->at(i).posXYT().t());
	  glVertex3f(tableNearStState_[i]->x(), tableNearStState_[i]->y(), tableNearStState_[i]->t());
	  glEnd();
	}
    }
}


// /**************** Running functions ***************/

// /**
//  *  \brief start the deformation process
//  */
// void TrajDeform::start()
// {
//   // start the timer
//   timer_.start();

//   while(timer_.isRunning())
//     {
      
//     }
// }

// /**
//  *  \brief stop the deformation process
//  */
// void TrajDeform::stop()
// {
//   // stop the timer
//   timer_.stop();

// }

// /**
//  *  \brief reset the deformation process
//  */
// void TrajDeform::reset()
// {
//   // reinit the timer
//   timer_.reinit();
// }


// /************* Access/Update functions ************/

// /**
//  *  \brief update the environment informations
//  *  \param env : the new input informations about the environment
//  */
// void TrajDeform::updateEnv(Environment* env)
// {
//   this->env_ = env;
// }

// /**
//  *  \brief update the system localization and state informations
//  *  \param s : the new input informations about the system state
//  */
// void TrajDeform::updateSysState(StateTime* s)
// {
//   this->curStateT_ = s;
// }
