#include <fstream>
#include <iostream>
#include <string>

#include "../inc/Map.hpp"
#include "../inc/Utils.hpp"

#include <time.h>
#include <sys/time.h>
#include <cstring>

using namespace std;

void printErr(){
  cout << "mapBuider <input_file.pgm> <output_file.pgm> <opt>" << endl;
  cout << "<opt> : -c              -> simple copy of the map" << endl;
  cout << "        -s <threshold>  -> thresholding of the map" << endl;
  cout << "        -g <nbPixels>   -> obstacles growing" << endl;
  cout << "        -pf <x> <y>     -> adding potential fields" << endl;   
  exit(1);
}

int main(int argc,char** argv)
{
	if(argc < 4)
	{
	  printErr();
	}
	else
	{
		// Reads input map
		Map* map = new Map();
		map->load(argv[1]);
		//map->display();

		if(strcmp(argv[3],"-c") == 0)
		{
		  // Copies map
		  map->save(argv[2]);
		  return 0;
		}
		else if(strcmp(argv[3],"-s") == 0)
		  {
		    if(argc != 5)
		      {
			printErr();
		      }
		    // Threshold the map
		    int threshold = atoi(argv[4]);
		    if((threshold >= 0) && (threshold <= map->getDepth()))
		      {
			map->thresholdPGM(threshold);
			map->save(argv[2]);
			return 0;
		      }
		    else
		      {
			cout << "main : The threshold must be between 0 and " << map->getDepth() << endl;
				return 1;
		      }
		}
		else if(strcmp(argv[3],"-g") == 0)
		  {
		    if(argc != 5)
		      {
			printErr();
		      }
		    if(map->getType() == Map::PGM_T)
		      {
			int nbPix = atoi(argv[4]);
			map->increaseObs(nbPix);
			map->save(argv[2]);
		      }
		    else
		      {
			cout << "main : The input map must have been thresholded before growing obstacles" << endl;
			return 1;
		      }
		  }
		else if(strcmp(argv[3],"-pf") == 0)
		  {
		    if(argc != 6)
		      {
			printErr();
		      }
		    if(map->getType() == Map::PGM_T)
		      {
			int x = atoi(argv[4]);
			int y = atoi(argv[5]);
			Pos p = Pos(x,y);
			if(p.isReachable(map))
			  {
			    // adding potential field from target position
			    Map* map1 = map->computeWave(&p);
			    string nom1 = addSuffixe(string(argv[2]),string("_tp"));
			    map1->save(nom1.c_str());
			    // adding potential field from obstacles
			    Map* map2 = map->applyPFields3();
			    string nom2 = addSuffixe(string(argv[2]),string("_pfo"));
			    map2->save(nom2.c_str());
			    // adding both potential fields
			    //Map* mapRes = Map::mediumMap(map1,map2);
			    //Map* mapRes = Map::addingPFMap(map1,map2);
			    Map* mapRes = Map::addingPFMap2(map1,map2);
			    // saving maps
			    mapRes->save(argv[2]);
			  }
			else
			  {
			    cout << "main : given position is not reachable" << endl;
			    return 1;
			  }
		      }
		    else
		      {
			cout << "main : The input map must have been thresholded before adding potential fields" << endl;
			return 1;
		      }
		  }
		
		delete map;
	}
	return 0;
}
