#include "Map.hpp"
using namespace mpc;

#include <stdio.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <cstring>

#include "Pos.hpp"
#include "Utils.hpp"

using namespace std;


                               /*******************************************************/
                               /************* Constructors and destructors ************/
                               /*******************************************************/

// default constructor
Map::Map()
{
	typ = ND;
	w = 0;
	h = 0;
	d = 0;
	tabC = NULL;
}

// Map::Map(const Map& c)
// {
// 	typ = c.typ;
// 	w = c.w;
// 	h = c.h;
// 	d = c.d;
// 	tabC = new double*[w];

// 	for(int i = 0; i < w; i++)
// 	{
// 		tabC[i] = new double[h];
// 	}
// 	for(int i = 0; i < w; i++)
// 	{
// 		for(int j = 0; j < h; j++)
// 		{
// 			tabC[i][j] = c.tabC[i][j];
// 		}
// 	}
// }

Map::Map(const Map& c)
{
  typ = c.typ;
  w = c.w;
  h = c.h;
  d = c.d;
  tabC = new double[w*h];
  
  for(int i = 0; i < w; i++)
    {
      for(int j = 0; j < h; j++)
	{
	  this->setCol(i,j,c.getCol(i,j));
	}
    }
}

// Map::Map(const int wt,const int ht)
// {
// 	typ = ND;
// 	w = wt;
// 	h = ht;
// 	d = 0;
// 	tabC = new double* [w];

// 	for(int i = 0; i < w; i++)
// 	{
// 		tabC[i] = new double[h];
// 	}
// }

Map::Map(const int wt,const int ht)
{
  typ = ND;
  w = wt;
  h = ht;
  d = 0;
  tabC = new double[w*h];
}

Map::Map(const int wt,const int ht,const int dph)
{
  typ = PGM;
  w = wt;
  h = ht;
  d = dph;
  tabC = new double[w*h];
}

Map::~Map()
{
  if(tabC != NULL)
    {
      delete [] tabC;
    }
  tabC = NULL;
}

                               /*******************************************************/
                               /********************** Accessors **********************/
                               /*******************************************************/

// return map type
Map::typ_t Map::getType() const
{
	return typ;
}

// return map width
int Map::getWidth() const
{
	return w;
}

// return map height
int Map::getHeight() const
{
	return h;
}

// return depth of the map
int Map::getDepth() const
{
	return d;
}

// // return point color at coord (x,y)
// double Map::getCol(int x,int y) const
// {
//         return tabC[x][y];
// }

// return point color at coord (x,y)
double Map::getCol(int x,int y) const
{
  return tabC[x*h + y];
}

// // return point color at position (x,y)
// double Map::getCol(Pos p){
//         return tabC[p.getX()][p.getY()];
// }

// return point color at position (x,y)
double Map::getCol(Pos p){
  return tabC[p.getX()*h + p.getY()];
}

// modify map type
void Map::setType(typ_t f)
{
	typ = f;
}

// modify map width
void Map::setWidth(int wt)
{
	w = wt;
}

// modify map height            
void Map::setHeight(int ht)
{
	h = ht;
}

// // modify point color at coord (x,y)
// void Map::setCol(int x,int y,double col)
// {
//         tabC[x][y] = col;
// }

// modify point color at coord (x,y)
void Map::setCol(int x,int y,double col)
{
        tabC[x*h + y] = col;
}

// modify point color at position p
void Map::setCol(Pos p,double col)
{
        tabC[p.getX()*h + p.getY()] = col;
}


                               /*******************************************************/
                               /**************** Map utility functions ****************/
                               /*******************************************************/

// void Map::initTabC(int wt,int ht)
// {
// 	tabC = new double*[wt];

// 	for(int i = 0; i < wt; i++)
// 	{
// 		tabC[i] = new double[ht];
// 	}

// 	for(int i = 0; i < wt; i++)
// 	{
// 		for(int j = 0; j < ht; j++)
// 		{
// 			tabC[i][j] = 0;
// 		}
// 	}
// }

void Map::initTabC(int wt,int ht)
{
  tabC = new double[wt*ht];
  
  for(int i = 0; i < wt; i++)
    {
      for(int j = 0; j < ht; j++)
	{
	  this->setCol(i,j,0);
	}
    }
}

// void Map::display() const
// {
//   cout << endl << "Map : " << endl;
//   for(int j=0;j<h;j++)
//   {
//     for(int i=0;i<w;i++)
//     {
//       int a = (int)(this->tabC[i][j]);
//       //cout << this->tabC[i][j] << " ";
//       printf("%3d ",a);
//     }
//     cout << endl;
//   }
// }

void Map::display() const
{
  cout << endl << "Map : " << endl;
  for(int j=0;j<h;j++)
  {
    for(int i=0;i<w;i++)
    {
      int a = (int)(this->getCol(i,j));
      //cout << this->tabC[i][j] << " ";
      printf("%3d ",a);
    }
    cout << endl;
  }
}


                               /*******************************************************/
                               /************ Map loading/saving functions *************/
                               /*******************************************************/

/****************************************************************************/
/******************** Reading function for PGM type map *********************/
/****************************************************************************/

/* This function follows the principle of a simple automaton whence main stapes are next :*/ 
/* stape 1 : File's type reading (P5 -> image PGM) */
/* stape 2 : Image's weight reading */
/* stape 3 : Image's height reading */
/* stape 4 : Colour depth reading */
/* stape 5 : Grey level of a point reading */
/* Note that a boolean has been definied to take on consideration comments */

int Map::load(const char* nomfic)
{
  cout << "Map::load" << endl;

	bool comment = false;
	int etape = 1;
	FILE* fe;
	//fstream fe;
	char* bufE = new char[1];
	int* buf = new int[1];
	char* bufCh = new char[20];
	int tailleBufCh = 0;
	long int nbPts = 0;
	bool estSeuillee = true;

	fe = fopen(nomfic,"rb");

	if(fe == NULL)
	{
		cout << "Impossible to open the input file" << endl;
		delete [] buf;
		delete [] bufE;
		delete [] bufCh;
		return -1;
	}

#ifdef DEBUG
	cout << "Reading of the input file..." << endl;
	cout << "Reading of the input file's heading" << endl;
#endif

	/**********************************************************************/
	/***************** reading of the input file heading ******************/
	/**********************************************************************/
	while( (etape != 5) && ((*bufE = fgetc(fe)) != EOF) )
	{
		if(comment)
		{
			// Comment treatment
			// If the '\n' has been found, the comment is left
			// Else nothing is done
			if(*bufE == '\n')
			{
				comment = false;
				//cout << "end of comment" << endl;
			}
		}
		else
		{
			// Treatement of the character of the beginnig of the comment
			if(*bufE == '#')
			{
				comment = true;
				//cout << "beginning of comment" << endl;
			}
			else
			{
				switch(etape)
				{
					case 1 : 
						if(isAlphaNum(*bufE))
						{
							bufCh[tailleBufCh] = *bufE;
							tailleBufCh++;
						}
						else
						{
							if(tailleBufCh != 0)
							{
								bufCh[tailleBufCh] = '\0';
								tailleBufCh++;
								// Map's type verification
								if(strcmp(bufCh,"P5") == 0)
								{
									setType(PGM);
								}
								else if(strcmp(bufCh,"P2") == 0)
								{
									setType(PGM_N);
								}
								else
								{
									cout << "Type unknown : \"" << bufCh << "\""<< endl;
									setType(ERR);
								}
								// needed reinitialisation to go to stape 2
								tailleBufCh = 0;
								etape= 2;
							}
							else
							{
								// case where the type hasn't begun to be read : nothing is done
							}
						}
						break;

					case 2 : 
					case 3 : 
					case 4 : 
						if(isNum(*bufE))
						{
							bufCh[tailleBufCh] = *bufE;
							tailleBufCh++;
						}
						else
						{
							if(tailleBufCh != 0)
							{
								bufCh[tailleBufCh] = '\0';
								tailleBufCh++;
								// treatement of the input figure
								switch(etape)
								{
									case 2 :
										w = atoi(bufCh);
										etape = 3;
										break;
									case 3 : 
										h = atoi(bufCh);
										etape = 4;
										break;
									case 4 :
										d = atoi(bufCh);
										etape = 5;
										break;
								}
								tailleBufCh = 0;
							}
							else
							{
								// case where the type hasn't begun to be read : nothing is done
							}
						}
						break;

					default :
						cout << "Stape unknown : " << etape << endl;
						return -1;
						break;
				}
			}
		}
	}

	initTabC(w, h);

#ifdef DEBUG
	cout << "Reading of input file's body" << endl;
#endif

	/**********************************************************************/
	/****************** reading of the input file's body ******************/
	/**********************************************************************/

	if(typ == PGM)
	{
		while(((*buf = fgetc(fe)) != EOF) && (nbPts < w * h))
		{
			// Test of the thresholding of the map
			if((estSeuillee) && (*buf != d) && (*buf != 0))
			{
				estSeuillee = false;
				//cout << "this map is not already thresholded : " << *buf << endl;
			}
			// Filling of the table
			//tabC[nbPts % w][h-1 - (nbPts / w)] = *buf;
			this->setCol(nbPts % w,h-1 - (nbPts / w), *buf);
			nbPts++;
		}
		if(estSeuillee)
		{
			typ = PGM_T;
			//cout << "This map is already thresholded" << endl;
		}
	}
	else if(typ == PGM_N)
	{
		while(((*bufE = fgetc(fe)) != EOF) && (nbPts < w * h))
		{
			// Filling of the table
			if(isNum(*bufE))
			{
				bufCh[tailleBufCh] = *bufE;
				tailleBufCh++;
			}
			else
			{
				if(tailleBufCh != 0)
				{
					bufCh[tailleBufCh] = '\0';
					tailleBufCh++;
					if((estSeuillee) && (atoi(bufCh) != d) && (atoi(bufCh) != 0)){
						estSeuillee = false;
						//cout << "this map is not already thresholded" << endl;
					}
					//tabC[nbPts % w][h-1 - (nbPts / w)] = atoi(bufCh);
					this->setCol(nbPts % w,h-1 - (nbPts / w), atoi(bufCh));
					nbPts++;
					tailleBufCh = 0;
				}
				else
				{
					// case where the type hasn't begun to be read : nothing is done
				}	
			}
		}
		if(estSeuillee)
		{
			this->typ = PGM_T;
			//cout << "This map is already thresholded." << endl;
		}
	}
	delete [] buf;
	delete [] bufE;
	delete [] bufCh;

	fclose(fe);

#ifdef DEBUG
	cout << "End of reading." << endl;
#endif
	return 0;
}

/****************************************************************************/
/******************** Writting function for PGM type map ********************/
/****************************************************************************/

// write the current occupancy grid in a pgm file
// return 0 if it works
// return -1 else
int Map::save(const char* nomfic) const
{
	FILE* fs;
	// Verification of the map's type
	//cout << "Type verification" << endl;
	if((typ != PGM) && (typ != PGM_T) && (typ != PGM_TG) && (typ != PGM_N) && (typ != PGM_M) && (typ != PGM_PFO) && (typ != PGM_PFT) && (typ != PGM_PF))
	{
		cout << "save : Can't write the file. Maps's type not available." << endl;
		return -1;
	}
	// Opening of the output file
	fs = fopen(nomfic,"wb");
	if(fs == NULL)
	{
		cout << "save : Error while trying to create the ouput file." << endl;
		return -1;
	}

#ifdef DEBUG
	cout << "Writting of the output map..." << endl;
#endif

	/**********************************************************************/
	/*************** Writting of the output file's heading ****************/
	/**********************************************************************/
	// Writting of the map's type
	fputs("P5\n",fs);
	// Writting of the map's weight
	char wch[20];
	int2String(w,wch);
	fputs(wch,fs);
	fputs(" ",fs);
	// Writting of the map's height
	char hch[20];
	int2String(h,hch);
	fputs(hch,fs);
	fputs("\n",fs);
	// Writting of colour's depth of the map
	char dch[20];
	int2String(d,dch);
	fputs(dch,fs);
	fputs("\n",fs);

	/**********************************************************************/
	/**************** Writting of the output file's body ******************/
	/**********************************************************************/
	for(int j = 0; j < h; j++)
	{
		for(int i = 0; i < w; i++)
		{
		  //fputc((int)(tabC[i][h-1-j]), fs);
		  fputc((int)(this->getCol(i,h-1-j)), fs);
		}
	}

#ifdef DEBUG
	cout << "End of writting." << endl;
#endif

	fclose(fs);
	return 0;
}


                               /*******************************************************/
                               /*************** Map treatment functions ***************/
                               /*******************************************************/

/****************************************************************************/
/**************** Thresholding function of the PGM typed map ****************/
/****************************************************************************/

void Map::thresholdPGM(int seuil)
{
	cout << "Thresholding of the map..." << endl;
	
	for(int i = 0; i < w; i++)
	{
		for(int j = 0; j < h; j++)
		{
		  //if(tabC[i][j] >= seuil)
		  if(this->getCol(i,j) >= seuil)
		    //tabC[i][j] = d;
		    this->setCol(i,j,d);
		  else
		    //tabC[i][j] = 0;
		    this->setCol(i,j,0);
		}
	}
	
	typ = PGM_T;
}

/****************************************************************************/
/************ Obstacles increasing of a thresholded map function ************/
/****************************************************************************/

void Map::increaseObs(int nbPix)
{
	if(typ != PGM_T)
	{
		cout << "increaseObs : The map must be PGM typed" << endl;
		return;
	}

#ifdef DEBUG
	cout << "Increasing of the obstacles..." << endl;
#endif

	// defining an obstacle border around the map
	
 	for(int i = 0; i < w; i++)
 	{
 		for(int j = 0; j < h; j++)
 		{
 		  if(i == 0 || i == w-1 || j == 0 || j == h-1)
 		    //if(tabC[i][j] != 0)
		    if(this->getCol(i,j) != 0)
 		      //tabC[i][j] = 0;
		      this->setCol(i,j,0);
 		}
 	}

	// increasing of the obstacles

	for(int i = 0; i < w; i++)
	{
		for(int j = 0; j < h; j++)
		{
		  //if(tabC[i][j] != 0)
		  if(this->getCol(i,j) != 0)
			{
				// We're looking at the nearest pixels of the current pixel to see if this one must be covered alike
				bool trouve = false;
				int k,l;
				k = -nbPix;

				while((k <= nbPix) && (!trouve))
				{
					// We're verifying if the current pixel is in the frame of the map
					if((i + k >= 0) && (i + k < w))
					{
						l = -nbPix;
						while((l <= nbPix) && (!trouve))
						{
							// We're verifying if the current pixel is in the frame of the map
							if((j + l >= 0) && (j + l < h))
							  {
							    // We're testing if an obstacle is present
							    //if(tabC[i + k][j + l] == 0 && (hypot(k, l) < nbPix))
							    //if(tabC[i + k][j + l] == 0 && (k*k + l*l < nbPix*nbPix))
							    if(this->getCol(i + k,j + l) == 0 && (k*k + l*l < nbPix*nbPix))
							      {
								trouve = true;
							      }
							  }
							l++;
						}
					}
					k++;
				}
				if(trouve)
				{
				  //tabC[i][j] = -1;
				  this->setCol(i,j,-1);
				}	
			}
			else
			{
				// Case where an obstacle is already present : nothing is done
			}
		}
	}
	
	for(int i = 0; i < w; i++)
	{
	        for(int j = 0; j < h; j++)
		{
		  //if(tabC[i][j] == -1)
		  if(this->getCol(i,j) == -1)
			{
			  //cout << "pixel modified" << endl;
			  //tabC[i][j] = 0;
			  this->setCol(i,j,0);
			}
		}
	}
}

/****************************************************************************/
/***************************** Wave computing *******************************/
/****************************************************************************/
// compute a wave on the map from current position from white to black
// Allows to describe potential fields due to target position

Map* Map::computeWave(Pos* p)
{
        if(!p->isReachable(this))
	{
	        cout << "computeWave : position is not reachable." << endl;
		exit(1);
	}
	else
	{
	        // compute a copy of the current map
                Map* mapRes = new Map(*this);
		
		// searching for distance from current position to the further corner
		float distMax = sqrt(p->getX()*p->getX() + p->getY()*p->getY());
		if( sqrt( p->getX()*p->getX() + (h-p->getY())*(h-p->getY()) ) > distMax)
		{
	                distMax = sqrt( p->getX()*p->getX() + (h-p->getY())*(h-p->getY()) );
		}
		if( sqrt( p->getY()*p->getY() + (w-p->getX())*(w-p->getX()) ) > distMax)
		{
	                distMax = sqrt( p->getY()*p->getY() + (w-p->getX())*(w-p->getX()) );
		}
		if( sqrt( (h-p->getY())*(h-p->getY()) + (w-p->getX())*(w-p->getX()) ) > distMax)
		{
	                distMax = sqrt( (h-p->getY())*(h-p->getY()) + (w-p->getX())*(w-p->getX()) );
		}

		// computing the color (due to target potential field) of each reachable position of the map
		for(int i=0;i<w;i++)
		{
		        for(int j=0;j<h;j++)
			{
		                Pos p1 = Pos(i,j);
				if(p1.isReachable(this))
				{
			                double d = sqrt((p->getX()-i)*(p->getX()-i) + (p->getY()-j)*(p->getY()-j));
					double modif = (((double)d)/((double)distMax)*((double)(this->getDepth())));
					if(modif >= this->getDepth())
					{
				                modif = this->getDepth() - 1;
					}
					mapRes->setCol(p1.getX(),p1.getY(),(double)((this->getDepth()-modif) *(((double)(this->getDepth()-modif))/((double)(this->getDepth()))) ));
				}
			}
		}
		mapRes->typ = PGM_PFT;
		return mapRes;
	}
}

/****************************************************************************/
/**************** Potential fields applying from obstacles ******************/
/****************************************************************************/

// 3rd version : 
// search all reachable positions which define the border of the map
// and compute the distance from the border to all reachable position
// of the map by reducing these borders
Map* Map::applyPFields3()
{
	Map *mapRes = new Map(*this);
	// searching for all reachable positions
	vector<Pos> *v = new vector<Pos>();
	vector<Pos> *vtmp = new vector<Pos>();
	for(int i=0;i<mapRes->getWidth();i++)
	{
		for(int j=0;j<mapRes->getHeight();j++)
		{
			Pos p = Pos(i,j);
			if(p.isReachable(mapRes))
			{
				v->push_back(p);
			}
		}
	} 

	// searching for all reachable positions which touch unreachable positions
	int d = -1;
	unsigned int a=0;
	while(a < v->size())
	{
		bool trouve = false;
		unsigned int i=0;
		vector<dir_t> vDir;
		vDir = ((*v)[a]).availableDir(mapRes);
		while(!trouve && i<vDir.size())
		{
			Pos p = Pos((*v)[a],vDir[i]);
			if(mapRes->getCol(p) == 0)
				trouve = true;
			else 
				i++;
		}
		if(trouve)
		{
			vtmp->push_back((*v)[a]);
			mapRes->setCol((*v)[a],d);
		}
		a++;
	}
	delete(v);
	v = vtmp;
	vtmp = new vector<Pos>();
	
	// searching iteratively all reachable positions from current vector still not modified
	while( (!v->empty()) )
	{
		a = 0;
		d--;
		while(a < v->size())
		{
			vector<dir_t> vDir;
			if((-d)%2 == 1) vDir = ((*v)[a]).reachableDir1(mapRes);
			else vDir = ((*v)[a]).reachableDir2(mapRes);
			for(unsigned int i=0;i<vDir.size();i++)
			{
				Pos p = Pos((*v)[a],vDir[i]);
				if(mapRes->getCol(p) == 255)
				{
					vtmp->push_back(p);
					mapRes->setCol(p,d);
				}
			}
			a++;
		}
		delete(v);
		v = vtmp;
		vtmp = new vector<Pos>();
	}
	delete vtmp;
	if(v != NULL) delete v;
	
	// giving "good colour" to all points that have been modified
	for(int i=0;i<mapRes->getWidth();i++)
	{
		for(int j=0;j<mapRes->getHeight();j++)
		{
			Pos p = Pos(i,j);
			if(p.isReachable(mapRes))
			{
			  double modif = (double)(((double)(mapRes->getCol(p))) * ((double)(mapRes->getDepth())) / ((double)d) );
			  if(mapRes->getCol(p) != 255) mapRes->setCol(p,(double)( ((double)(modif)) * sqrt(((double)modif)/255) ));
			}
		}
	}
	mapRes->typ = PGM_PFO;
	return mapRes;
}

// Compute medium colour map between two maps
Map* Map::mediumMap(Map* m1,Map* m2)
{
	Map* mapRes;
	// verifying that m1 and m2 have got the same dimensions
	if((m1->h != m2->h) || (m1->w != m2->w))
	{
	  	cout << "mediumMap : both map haven't got same dimensions. Can't compute a medium map" << endl;
		exit(1);
	}
	mapRes = new Map(m1->w,m1->h);
	mapRes->typ = PGM_M;
	mapRes->d = m1->d;
	
	// computing the intermediate colour of each position
	for(int i=0;i<m1->w;i++)
	{
		for(int j=0;j<m1->h;j++)
		{
		  //mapRes->tabC[i][j] = (double)((m1->tabC[i][j] + m2->tabC[i][j])/2);
		  mapRes->setCol(i,j, (double)((m1->getCol(i,j) + m2->getCol(i,j))/2));
		}
	}
	return mapRes;
}

// Compute adding of both potential fields map (optimized)
Map* Map::addingPFMap2(Map* m1,Map* m2)
{
	Map* mapRes;
	// verifying that m1 and m2 have got the same dimensions
	if((m1->h != m2->h) || (m1->w != m2->w))
	{
		cout << "mediumMap : both map haven't got same dimensions. Can't compute a medium map" << endl;
		exit(1);
	}
	mapRes = new Map(m1->w,m1->h);
	mapRes->typ = PGM_PF;
	mapRes->d = m1->d;

	// add the both potential fields to obtain the result map
	for(int i=0;i<m1->w;i++)
	{
		for(int j=0;j<m1->h;j++)
		{
		  //mapRes->tabC[i][j] = (double)( ( pow( ((double)(m1->tabC[i][j])) / 255.0 , 4 ) + sqrt( ((double)(m2->tabC[i][j])) / 255.0 ) ) * (255.0/2.0) );
		  mapRes->setCol(i,j, (double)( ( pow( ((double)(m1->getCol(i,j))) / 255.0 , 4 ) + sqrt( ((double)(m2->getCol(i,j))) / 255.0 ) ) * (255.0/2.0) ));
		}
	}
	return mapRes;
}


// Compute adding of both potential fields map
Map* Map::addingPFMap(Map* mapTP,Map* mapPFO)
{
	Map* mapRes = new Map(*mapPFO);
	// verifying that m1 and m2 have got the same dimensions
	if((mapTP->h != mapPFO->h) || (mapTP->w != mapPFO->w))
	{
		cout << "mediumMap : both map haven't got same dimensions. Can't compute a medium map" << endl;
		exit(1);
	}
	mapRes->typ = PGM_PF;
	
	// add the both potential fields to obtain the result map
	for(int i=0;i<mapTP->w;i++)
	{
		for(int j=0;j<mapTP->h;j++)
		{
		  //mapRes->tabC[i][j] = mapRes->tabC[i][j] + mapTP->tabC[i][j];
		  mapRes->setCol(i,j, mapRes->getCol(i,j) + mapTP->getCol(i,j));
		  //if(mapRes->tabC[i][j] > mapRes->d) mapRes->tabC[i][j] = mapRes->d;
		  if(mapRes->getCol(i,j) > mapRes->d) mapRes->setCol(i,j,mapRes->d);
		}
	}
	return mapRes;
}


/****************************************************************************/
/************************** Map merging functions ***************************/
/****************************************************************************/

// merge the obstacles of the current map with map2
void Map::merge(Map* map2)
{
  // check if the dimensions of both maps are equals
  if((this->h != map2->h) || (this->w != map2->w))
    {
      cout << "Map::merge : Warning. Dimensions of both maps are not equals." << endl;
      return;
    }
  
  for(int i=0; i<w; i++)
    {
      for(int j=0; j<h; j++)
	{
	  if(map2->isFree(i,j) == 0.0)
	    {
	      this->setFreedom(i,j,0.0);
	    }
	}
    }

  this->typ = PGM;
}


/****************************************************************************/
/************************** Obstruction functions ***************************/
/****************************************************************************/

// return the freedom probability of the position (x,y)
double Map::isFree(int x,int y) const
{
  return (tabC[x*h + y] / ((double)d));
}

// set the freedom probability of the position (x,y)
void Map::setFreedom(int x,int y,double freed)
{
  tabC[x*h + y] = (int) (freed * d);
}

// set the freedom probability of the line defined between p1 and p2
void Map::setFreedom(Pos p1,Pos p2,double freed)
{
  vector<Pos>* line = p1.getLine(p2);
  int N = line->size();
  for(int i=0; i<N; i++) 
    {
      int x = line->at(i).getX();
      int y = line->at(i).getY();
      if(x >= 0 && x < w && y >= 0 && y < h)
	this->setFreedom(x,y,freed);
    }

  delete line;
}

// set the freedom probability of the line defined between (x1,y1) and (x2,y2)
void Map::setFreedom(int x1,int y1,int x2,int y2,double freed)
{
  this->setFreedom(Pos(x1,y1),Pos(x2,y2),freed);
}

// set the freedom probability of borders of the map
void Map::setBordersFreedom(double freed)
{
  for(int i=0; i<w; i++)
    {
      this->setFreedom(i,0,freed);
      this->setFreedom(i,h-1,freed);
    }
  for(int i=0; i<h; i++)
    {
      this->setFreedom(0,i,freed);
      this->setFreedom(w-1,i,freed);
    }
}

// set the map completely free
void Map::setCompletelyFree()
{
  for(int i=0; i<w; i++)
    for(int j=0; j<h; j++)
      this->setFreedom(i,j,1.0);
}

// set the map completely obstructed
void Map::setCompletelyObstructed()
{
  for(int i=0; i<w; i++)
    for(int j=0; j<h; j++)
      this->setFreedom(i,j,0.0);
}

/**
 *  \brief return true if the map if free at position p
 *  \return true if the map if free at position p
 */
bool Map::isFreeAt(const PositionXY& p) const
{
  int x = (int)(p.x());
  int y = (int)(p.y());

  if(x>w || y>h) return false;
  else return ((tabC[x*h + y] / ((double)d)) != 0);  
}

/**
 *  \brief return the color of the map at position p
 *  \return the color of the map at position p
 */
double Map::getColAt(const PositionXY& p) const
{
  int x = (int)(p.x());
  int y = (int)(p.y());

  if(x>w || y>h) return 0;
  else return tabC[x*h + y];
}


/**************** Path tracing functions ***************/

/**
 *  \brief trace a path on the map
 *  \param p : the input path to trace
 */
void Map::tracePath(const CPath& p)
{
  int pathCol = (int) (this->getDepth() / 2);

  int N = p.size();
  for(int i=0; i<N; i++)
    {
      // compute the position of the current configuration of the path
      PositionXY pos = p.at(i).posXY();

      if(this->isFreeAt(pos))
	{
	  this->setCol((int)(pos.x()),(int)(pos.y()),pathCol);
	}
      else
	{
	  cout << "Path::traceOn : Warning. The path is not free on the map." << endl;
	}
    }
}
