#include "environment/inc/Map2D.hpp"
using namespace mpc;

#include <stdlib.h>
#include <iostream>
#include <fstream>
using namespace std;

#include "Utils.hpp"

                                    /**************************************************/
                                    /******************* Class Map2D ******************/
                                    /**************************************************/

// int nbCols_;              /**< number of columns of the DEM table */
// int nbLines_;             /**< number of lines of the DEM table */
    
// float xOrigin_;           /**< X coordinate of the top left corner of the (0,0) pixel in the global frame */
// float yOrigin_;	      /**< Y coordinate of the top left corner of the (0,0) pixel in the global frame */
    
// float xScale_;            /**< X size of a cell in meters */
// float yScale_;            /**< Y size of a cell in meters */

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Map2D::Map2D()
{
  nbCols_ = 0;             // number of columns of the DEM table
  nbLines_ = 0;            // number of lines of the DEM table
  
  xOrigin_ = 0.0;          // X coordinate of the top left corner of the (0,0) pixel in the global frame
  yOrigin_ = 0.0;          // Y coordinate of the top left corner of the (0,0) pixel in the global frame
  
  xScale_ = 0.1;           // X size of a cell in meters
  yScale_ = 0.1;           // Y size of a cell in meters
}

/**
 *  \brief copy constructor
 *  \param map : the map to copy
 */
Map2D::Map2D(const Map2D& map)
{
  this->nbCols_ = map.nbCols_;
  this->nbLines_ = map.nbLines_;

  this->xOrigin_ = map.xOrigin_;
  this->yOrigin_ = map.yOrigin_;
  
  this->xScale_ = map.xScale_;
  this->yScale_ = map.yScale_;  

//   this->initTables(map.nbCols_,map.nbLines_);
}

/**
 *  \brief init constructor
 *  \param nbCols : number of columns of the map table
 *  \param nbLines : number of lines of the map table
 *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param xScale : X size of a cell in meters
 *  \param yScale : Z size of a cell in meters
 */
Map2D::Map2D(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale)
{
  this->nbCols_ = nbCols;
  this->nbLines_ = nbLines;

  this->xOrigin_ = xOrigin;
  this->yOrigin_ = yOrigin;
  
  this->xScale_ = xScale;
  this->yScale_ = yScale;

//   this->initTables(nbCols,nbLines);
}

/**
 *  \brief destructor
 */
Map2D::~Map2D()
{
//   this->clearTables();
}


/***************** Initialization *****************/

// /*
//  *  \brief init tables from their dimesions
//  *  \param number of columns of the table
//  *  \param number of lines of the table
//  */
// void Map2D::initTables(int nbCols,int nbLines)
// {
//   nbCols_ = nbCols;            // number of columns of the DEM table
//   nbLines_ = nbLines;          // number of lines of the DEM table
// }
    
// /*
//  *  \brief delete all the content of the DEM tables
//  */
// void Map2D::clearTables()
// {
// }


/******************** Accessors *******************/

/*
 *  \brief get back the index of the input coord in the grid
 *  \param x : input absica
 *  \param y : input ordinate
 *  \param i : output number of the column
 *  \param j : output number of the line
 *  return 0 if the input coordinates are inside of the DEM grid, -1 else
 */
int Map2D::getIndexPos(float x,float y,int& i,int& j) const
{
//   cout << "Map2D::getIndexPos -> xOrigin_ : " << xOrigin_ 
//        << "  yOrigin_ : " << yOrigin_ 
//        << "  xScale_ : " << xScale_ 
//        << "  yScale_ : " << yScale_ << endl;

  // compute the indexes
  i = (int)((x - xOrigin_) / xScale_);
  j = (int)((y - yOrigin_) / yScale_);
//   i = (int)round((x - xOrigin_) / xScale_);
//   j = (int)round((y - yOrigin_) / yScale_);

//    cout << "Map2D::getIndexPos -> xOrigin_ : " << xOrigin_ << "  yOrigin_ : " << yOrigin_ << "  xScale_ : " << xScale_ << "  yScale_ : " << yScale_ << "  (x=" << x << ", y=" << y << ") -> (i=" << i << ", j=" << j << ")" << endl;
  
  // check if the indexes are in or out of the grid
  if (i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_) 
    {   
      return -1; 
    }
  else 
    {   
      return 0; 
    }
}

/*
 *  \brief get back the index of the input coord if it belongs to the grid, or the closest one if it doesn't belong to it
 *  \param x : input absica
 *  \param y : input ordinate
 *  \param i : output number of the column
 *  \param j : output number of the line
 *  return 0 if the input coordinates are inside of the DEM grid, -1 else
 */
int Map2D::getClosestIndexToPos(float x,float y,int& i,int& j) const
{
  int resGIP = getIndexPos(x,y,i,j);
  if(resGIP == 0) return 0;
  else
    {
      if(i < 0) i=0;
      else if(i >= nbCols_) i=nbCols_-1;
      if(j < 0) j=0;
      else if(j >= nbLines_) j=nbLines_-1;
      return -1;
    }
}

/*
 *  \brief get back the index of the input coord in a grid of ppm image
 *  \param x : input absica
 *  \param y : input ordinate
 *  \param i : output number of the column
 *  \param j : output number of the line
 *  return 0 if the input coordinates are inside of the DEM grid, -1 else
 */
int Map2D::getIndexPPMImagePos(float x,float y,int& i,int& j) const
{
  int iGrid, jGrid;
  int res = this->getIndexPos(x, y, iGrid, jGrid);

//   i = jGrid;
//   j = nbCols_-1 - iGrid;
  i = nbLines_ -1 -jGrid;
  j = iGrid;

//   cout << "iGrid : " << iGrid << "  jGrid : " << jGrid << "  i : " << i << "  j : " << j << "  nbCols : " << nbCols_ << "  nbLines : " << nbLines_ << endl;
  
  return res;
}

/*
 *  \brief get back the list of indices of the grid corresponding to the line defined between the two input points
 *  \param x1 : first point absica
 *  \param y1 : first point ordinate
 *  \param x2 : second point absica
 *  \param y2 : second point ordinate
 *  \param res : the resulting list of indices
 */
void Map2D::getIndicesLine(float x1, float y1, float x2 ,float y2, std::vector<Map2D_index>& res) const
{
  // compute index of both input positions
  int i1=0, j1=0, i2=0, j2=0;
  this->getIndexPos(x1, y1, i1, j1);
  this->getIndexPos(x2, y2, i2, j2);

//   cout << "Map2D::getIndicesLine (" << x1 << "," << y1 << "," << x2 << "," << y2 << ") -> i1=" << i1 << "  j1=" << j1 << "  i2=" << i2 << "  j2=" << j2 << endl;

  // check on which axis positions have to be sampled
  int di = i2 - i1;
  int dj = j2 - j1;
  int max_dij;
  if(abs(di) >= abs(dj)) max_dij = abs(di);
  else max_dij = abs(dj);

  // search for each index in the grid covered by the input line
  for(int k=0; k<=max_dij; k++)
    {
      Map2D_index curIndex;
      curIndex.col_x = (int) (i1 + ((float)k) / ((float)max_dij) * di );
      curIndex.line_y = (int) (j1 + ((float)k) / ((float)max_dij) * dj );

      if (curIndex.col_x >= 0 && curIndex.col_x < nbCols_ && curIndex.line_y >= 0 && curIndex.line_y < nbLines_) res.push_back(curIndex);
    }
}

/*
 *  \brief get back the index of the input coord in the grid, and the weights of each surrounding nodes
 *  \param x : input absica
 *  \param y : input ordinate
 *  \param i : output number of the column
 *  \param j : output number of the line
 *  \param wUL : weight of the upper left corner
 *  \param wUR : weight of the upper right corner
 *  \param wBL : weight of the upper lower left corner
 *  \param wBR : weight of the upper lower right corner
 *  return 0 if the input coordinates are inside of the DEM grid, -1 else
 */
int Map2D::getIndexPosAndWeights(float x,float y,int& i,int& j,float& wUL,float& wUR,float& wBL,float& wBR) const
{
  // compute the indexes
  int resGIP = this->getIndexPos(x,y,i,j);

  if(resGIP == 0)
    {
      // compute the weights
      float coefX = ((x - xOrigin_) - (i*xScale_)) / xScale_;
      float coefY = ((y - yOrigin_) - (j*yScale_)) / yScale_;
      
      wUL = 0.5*((1.0-coefX) + (1.0-coefY));
      wUR = 0.5*((coefX) + (1.0-coefY));
      wBL = 0.5*((1.0-coefX) + (coefY));
      wBR = 0.5*((coefX) + (coefY));
    }

  return resGIP;
}

/*
 *  \brief get back the coord corresponding to the input index of the grid
 *  \param i : input number of the column
 *  \param j : input number of the line
 *  \param x : output absica
 *  \param y : output ordinate
 *  return 0 if the input coordinates are inside of the DEM grid, -1 else
 */
int Map2D::getPosIndex(int i,int j,float& x,float& y) const
{
  // check if the indexes are in or out of the grid
  if (i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_) return -1;

  // return the coord corresponding to the input index of the grid
  x = xOrigin_ + (i + 0.5)*xScale_;
  y = yOrigin_ + (j + 0.5)*yScale_;
  return 0;
}

/**
 *  \brief Return the number of columns of the DEM table
 *  \return the number of columns of the DEM table
 */
int Map2D::nbCols() const { return this->nbCols_; }

/**
 *  \brief Return the number of lines of the DEM table
 *  \return the number of lines of the DEM table
 */
int Map2D::nbLines() const { return this->nbLines_; }
    
/**
 *  \brief Return the X coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \return the X coordinate of the top left corner of the (0,0) pixel in the global frame
 */
float Map2D::xOrigin() const { return this->xOrigin_; }

/**
 *  \brief Return the Y coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \return the Y coordinate of the top left corner of the (0,0) pixel in the global frame
 */
float Map2D::yOrigin() const { return this->yOrigin_; }
    
/**
 *  \brief Return the X size of a cell in meters
 *  \return the X size of a cell in meters
 */
float Map2D::xScale() const { return this->xScale_; }

/**
 *  \brief Return the Y size of a cell in meters
 *  \return the Y size of a cell in meters
 */
float Map2D::yScale() const { return this->yScale_; }

// /**
//  *  \brief set the content of a cell as empty
//  *  \param i : the column number
//  *  \param j : the line number
//  */
// void Map2D::setEmptyCell(int i, int j)
// {
// }

// /**
//  *  \brief return true is the cell corresponding to theinput index is empty
//  *  \param i : the column number
//  *  \param j : the line number
//  *  \return true is the cell corresponding to theinput index is empty
//  *
//  *  The user must have checked if the input coord are valid
//  */
// bool Map2D::isEmptyCell(int i, int j) const
// {
// }
    

/************ Input / Output functions ************/

/**
 *  \brief create a rgb image from the 2D map
 *  \param minZ : minimal Z value used to convert the DEM into an image
 *  \param maxZ : maximal Z value used to convert the DEM into an image
 *  \return a new instance of rgb image
 */
RGBImage* Map2D::rgbImage(float minZ, float maxZ)
{
  // init the Z color parameters
  float minColZ, maxColZ, scaleColZ;
  this->initZ_ColorParams(minColZ, maxColZ, scaleColZ);

  RGBImage* image = new RGBImage(nbLines_,nbCols_);

  for(int i=0; i<nbLines_; i++)
    {
      for(int j=0; j<nbCols_; j++)
	{
	  int R,G,B;
	  rgbColPix(j,nbLines_-1 - i,minColZ,maxColZ,scaleColZ,R,G,B);

	  image->setCol(i, j, R, G, B);
	}
    }
  return image;
}

/**
 *  \brief save the DEM into an output PGM file
 *  \param filename : the output filename
 *  \param minZ : minimal Z value used to convert the DEM into an image
 *  \param maxZ : maximal Z value used to convert the DEM into an image
 *  \return 0 if the process succeed, -1 else
 */
int Map2D::saveIntoPGM(const char* filename, float minZ, float maxZ) const
{
  // open the output file
  ofstream fic(filename, ios::out | ios::trunc | ios::binary);

  // check if the opening succeeded
  if(fic)
    {
      // init the Z color parameters
      float minColZ, maxColZ, scaleColZ;
      this->initZ_ColorParams(minColZ, maxColZ, scaleColZ);

      // write the PGM magic number
      fic << "P5" << endl;
      
      // write a dump comment
      fic << "# Map 2D saved into a PGM file to display it" << endl;

      // write the dimensions of the map
      char colsString[20], linesString[20];
      int2String(nbCols_,colsString);
      int2String(nbLines_,linesString);
      fic << colsString << " " << linesString << endl;

      // write the depth of the map
      fic << "255" << endl;

      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      int greyCol = greyColPix(i,j,minColZ,maxColZ,scaleColZ);
	      fic << ((char)greyCol);
	    }
	}
      fic << endl;
      
      fic.close();

      return 0;
    }
  else
    {
      cerr << "Map2D::saveIntoPGM : cannot open the output file" << endl;
      return -1;
    }
}

/**
 *  \brief save the DEM into an output PPM file
 *  \param filename : the output filename
 *  \param minZ : minimal Z value used to convert the DEM into an image
 *  \param maxZ : maximal Z value used to convert the DEM into an image
 *  \return 0 if the process succeed, -1 else
 */
int Map2D::saveIntoPPM(const char* filename, float minZ, float maxZ) const
{
  // open the output file
  ofstream fic(filename, ios::out | ios::trunc | ios::binary);

  // check if the opening succeeded
  if(fic)
    {
      // init the Z color parameters
      float minColZ, maxColZ, scaleColZ;
      this->initZ_ColorParams(minColZ, maxColZ, scaleColZ);

      // write the PGM magic number
      fic << "P6" << endl;
      
      // write a dump comment
      fic << "# Map 2D saved into a PPM file to display it" << endl;

      // write the dimensions of the map
      char colsString[20], linesString[20];
      int2String(nbCols_,colsString);
      int2String(nbLines_,linesString);
      fic << colsString << " " << linesString << endl;

      // write the depth of the map
      fic << "255" << endl;

      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
// 	      cout << "i : " << i << "  j : " << j << " -> j : " << j << "  nbCols_-1 - i : " << nbCols_-1 - i << endl;
	      int R,G,B;
	      rgbColPix(j,nbCols_-1 - i,minColZ,maxColZ,scaleColZ,R,G,B);

	      fic << ((char)R) << ((char)G) << ((char)B);
	    }
	}
      fic << endl;
      
      fic.close();

      return 0;
    }
  else
    {
      cerr << "Map2D::saveIntoPPM : cannot open the output file : " << filename << endl;
      return -1;
    }
}


/**************** Utility functions ***************/

/**
 *  \brief extract a sub-map of the current map
 *  \param xMin : minimal bound over the X axis
 *  \param xMax : maximal bound over the X axis
 *  \param yMin : minimal bound over the Y axis
 *  \param yMax : maximal bound over the Y axis
 *  \param forceSquareMap : force the resulting map to be square
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of Map2D is created
 */
Map2D* Map2D::subMap(float xMin, float xMax, float yMin, float yMax, bool forceSquareMap) const
{
  int iMin, iMax, jMin, jMax;
  int resGIP1 = this->getClosestIndexToPos(xMin, yMin, iMin, jMin);
  int resGIP2 = this->getClosestIndexToPos(xMax, yMax, iMax, jMax);

  if(forceSquareMap)
    {
      if(iMax - iMin < jMax - jMin) jMax = jMin + (iMax - iMin);
      else if(iMax - iMin > jMax - jMin) iMax = iMin + (jMax - jMin);
    }

//   cout << "Map2D::subMap -> iMin : " << iMin << "  iMax : " << iMax << "  jMin : " << jMin << "  jMax : " << jMax << endl;

  return this->subMap(iMin, iMax, jMin, jMax);
}

/**
 *  \brief extract a sub-map of the current map
 *  \param (xCenter,yCenter) : coordinates of the center of the sub-map
 *  \param size : size of the sub-map
 *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
 *
 *  A new instance of Map2D is created
 */
Map2D* Map2D::subMap(float xCenter, float yCenter, float size) const
{
  float xMin = xCenter - 0.5*size;
  float xMax = xCenter + 0.5*size;
  float yMin = yCenter - 0.5*size;
  float yMax = yCenter + 0.5*size;

  return this->subMap(xMin, xMax, yMin, yMax, true);
}

/**
 *  \brief display the main features of the DEM
 *  \param dispGrid : display the grid if true
 */
void Map2D::display() const
{
  // display the main features
  cout << "nbLines : " << nbLines_ <<  "  nbCols : " << nbCols_ << endl;
  cout << "xOrigin : " << xOrigin_ <<  "  yOrigin : " << yOrigin_ << endl;
  cout << "xScale : " << xScale_ <<  "  yScale : " << yScale_  << endl;
  cout << endl;
}
