#include "../inc/Pos.hpp"
using namespace mpc;

#include <math.h>


                    /**************************************************************************/
                    /********************** constructors and destructors **********************/
                    /**************************************************************************/

// default constructor
Pos::Pos()
{
	this->x = 0;
	this->y = 0;
}

// init constructor (from coord (x,y))
Pos::Pos(int x1,int y1)
{
	this->x = x1;
	this->y = y1;
}

// init constructor (from position p and direction d)
// the new position constructed is position p moved by 1 pixel to direction d
Pos::Pos(const Pos &p,dir_t d)
{
	switch(d)
	  {
	  case NW : 
	  	this->x = p.getX()-1;
		this->y = p.getY()-1;
		break;
	  case N : 
	  	this->x = p.getX()-1;
		this->y = p.getY();
		break;
	  case NE :
	  	this->x = p.getX()-1;
		this->y = p.getY()+1;
		break;
	  case E : 
	  	this->x = p.getX();
		this->y = p.getY()+1;
		break;
	  case SE : 
	  	this->x = p.getX()+1;
		this->y = p.getY()+1;
		break;
	  case S : 
	  	this->x = p.getX()+1;
		this->y = p.getY();
		break;
	  case SW : 
	  	this->x = p.getX()+1;
		this->y = p.getY()-1;
		break;
	  case W : 
	  	this->x = p.getX();
		this->y = p.getY()-1;
		break;
	  default : 
	  	cout << "Pos : Unrecognized direction." << endl;
		exit(1);
		break;
	  }
}

// init constructor (from Point2D)
// Pos::Pos(Point2D p)
// {
// 	this->x = (int)(p.x);
// 	this->y = (int)(p.y);
// }

// copy constructor
Pos::Pos(const Pos &p)
{
	this->x = p.x;
	this->y = p.y;
}

// destructor
Pos::~Pos(){}


                    /**************************************************************************/
                    /********************************* getters ********************************/
                    /**************************************************************************/

// return position's absicca
int Pos::getX() const
{
	return this->x;
}

// return position's ordinate
int Pos::getY() const
{
	return this->y;
}

// return true if both positions have got the same coordinates
bool Pos::equals(const Pos &p)
{
	return((this->x == p.x) && (this->y == p.y));
}


                    /**************************************************************************/
                    /***************************** reachability *******************************/
                    /**************************************************************************/

// return true if this position is available on the map
// ie if the position is inside of the borders of the map
bool Pos::isAvailable(Map *map)
{
	if((this->x >= 0) && (this->x < map->getWidth()) && (this->y >= 0) && (this->y < map->getHeight()))
	{
		return true;
	}
	else return false;
}

// return true if this position is reachable on this map 
// (the position is available and there is no obstacle on this position)
bool Pos::isReachable(Map *map)
{
	// verifying that this position is available on the map
	if(this->isAvailable(map))
	{
		// verifying that there is no obstacle on this position
		if(map->getCol(this->x,this->y) != 0)
		{
			return true;
		}
		else return false;
	}
	else
	{
		return false;
	}
}

// return a vector of all positions which are reachable by the current position
vector<Pos> Pos::reachablePos(Map *map)
{
	vector<Pos> v = vector<Pos>();
	for(int i=-1;i<=1;i++)
	{
		for(int j=-1;j<=1;j++)
		{
			Pos p = Pos(this->x+i,this->y+j);
			// we dont have to considerate the same position like a reachable position
			if(!this->equals(p))
			{
				// were looking for all free position surrounding the current position
				if(p.isReachable(map)) v.push_back(p);
			}
		}
	}
	return v;
}

// return a vector of all reachable direction by the current position
vector<dir_t> Pos::reachableDir(Map *map)
{
	return this->reachableDir2(map);
}

// return a vector of all reachable direction by the current position (consider only N,E,S,W positions)
vector<dir_t> Pos::reachableDir1(Map *map)
{
	vector<dir_t> v = vector<dir_t>();
	Pos p;
	p = Pos(this->x-1,this->y);
	if(p.isReachable(map)) v.push_back(N);
	p = Pos(this->x,this->y+1);
	if(p.isReachable(map)) v.push_back(E);
	p = Pos(this->x+1,this->y);
	if(p.isReachable(map)) v.push_back(S);
	p = Pos(this->x,this->y-1);
	if(p.isReachable(map)) v.push_back(W);
	return v;
}

// return a vector of all reachable direction by the current position (consider NW,N,NE,E,SE,S,SW,W positions)
vector<dir_t> Pos::reachableDir2(Map *map)
{
	vector<dir_t> v = vector<dir_t>();
	Pos p;
	p = Pos(this->x-1,this->y-1);
	if(p.isReachable(map)) v.push_back(NW);
	p = Pos(this->x-1,this->y);
	if(p.isReachable(map)) v.push_back(N);
	p = Pos(this->x-1,this->y+1);
	if(p.isReachable(map)) v.push_back(NE);
	p = Pos(this->x,this->y+1);
	if(p.isReachable(map)) v.push_back(E);
	p = Pos(this->x+1,this->y+1);
	if(p.isReachable(map)) v.push_back(SE);
	p = Pos(this->x+1,this->y);
	if(p.isReachable(map)) v.push_back(S);
	p = Pos(this->x+1,this->y-1);
	if(p.isReachable(map)) v.push_back(SW);
	p = Pos(this->x,this->y-1);
	if(p.isReachable(map)) v.push_back(W);
	return v;
}

// return a vector of all available direction from the current position (consider only N,E,S,W positions)
vector<dir_t> Pos::availableDir(Map *map)
{
	vector<dir_t> v = vector<dir_t>();
	Pos p;
	p = Pos(this->x-1,this->y);
	if(p.isAvailable(map)) v.push_back(N);
	p = Pos(this->x,this->y+1);
	if(p.isAvailable(map)) v.push_back(E);
	p = Pos(this->x+1,this->y);
	if(p.isAvailable(map)) v.push_back(S);
	p = Pos(this->x,this->y-1);
	if(p.isAvailable(map)) v.push_back(W);
	return v;
}


                    /**************************************************************************/
                    /******************************* visibility *******************************/
                    /**************************************************************************/

// get a line path between the current position and a target position
vector<Pos>* Pos::getLine(Pos p)
{
	vector<Pos>* v = new vector<Pos>();
	int dx = p.getX() - this->getX();
	int dy = p.getY() - this->getY();

	// adding of current position into the vector
	v->push_back(*this);

	if(abs(dx) >= abs(dy))
	{
	        // for each absicca value between this.x and p.x computing the corresponding ordinate 
		for(int i=1;i<abs(dx);i++)
		{
			int j;
			if(dx > 0) j=i;
			else j=-i;
			int x1 = this->getX() + j;
			int y1 = this->getY() + (int)(((float)j)*(((float)dy)/((float)dx)));
			Pos p1 = Pos(x1,y1);
			v->push_back(p1);
		}
	}
	else
	{
	        // for each ordinate value between this.y and p.y computing the corresponding absicca 
		for(int i=1;i<abs(dy);i++)
		{
			int j;
			if(dy > 0) j=i;
			else j=-i;
			int x1 = this->getX() + (int)(((float)j)*(((float)dx)/((float)dy)));
			int y1 = this->getY() + j;
			Pos p1 = Pos(x1,y1);
			v->push_back(p1);
		}
	}
	// adding of target into the vector
	v->push_back(p);
	return v;
}

// return true if position p is visible by current position
bool Pos::isVisible(Pos p,Map* map)
{
	vector<Pos>* v = this->getLine(p);
	bool trouve = false;
	unsigned int i=0;
	
	// testing if each position of the line defined between 
	// the current position and p is free on the map
	while((!trouve) && (i<v->size()))
	{
		if(((*v)[i]).isReachable(map))
		{
			i++;
		}
		else
		{
			trouve = true;
		}
	}
	delete v;

	if(trouve) return false;
	else return true;
}


                    /**************************************************************************/
                    /******************************* distances ********************************/
                    /**************************************************************************/

// search the distance to the nearest obstacle on the map
int Pos::searchNearObs(Map* map)
{
	bool trouve = false;
	int d = 1;

	// search iteratively if there is an obstacle 
	// to the distance d of the current position
	while(!trouve)
	{
		int i=-d;
		while(i<=d && !trouve)
		{
			if( (this->x + i >= 0) && (this->x + i < map->getHeight()) )
			{
				int j = (int) sqrt(d*d - i*i);
				if( (this->y - j >= 0) && (this->y - j < map->getWidth()) )
				{
					Pos p1 = Pos(this->x + i,this->y - j);
					if(!(p1.isReachable(map)))
					{
						trouve = true;
					}
				}
				else
				{
					trouve = true;
				}
				if( (this->y + j >= 0) && (this->y + j < map->getWidth()) )
				{
					Pos p2 = Pos(this->x + i,this->y + j);
					if(!(p2.isReachable(map)))
					{
						trouve = true;
					}
				}
				else
				{
					trouve = true;
				}
			}
			else
			{
				trouve = true;
			}
			if(!trouve) i++;
		}
		if(!trouve) d++;
	}
	return d;
}

// return the distance between this position and an other one
double Pos::getDist(Pos p)
{
	return sqrt((this->x - p.x)*(this->x - p.x) + (this->y - p.y)*(this->y - p.y));
}


                    /**************************************************************************/
                    /********************** static method for directions **********************/
                    /**************************************************************************/

// return opposite direction to direction dir
dir_t Pos::oppDir(dir_t dir)
{
	switch(dir)
	{
	case NW : return SE; break;
	case N  : return S;  break;
	case NE : return SW; break;
	case E  : return W;  break;
	case SE : return NW; break;
	case S  : return N;  break;
	case SW : return NE; break;
	case W  : return E;  break;
	default : 
		cout << "oppDir : Undefined or impossible direction. Can't define opposite direction" << endl;
		return UND;
		break;
	}
}

// return privileged direction from p1 to p2
// ie the direction (NW,N,NE,E,SE,S,SW,W) from which the vector p1p2 is the nearest
dir_t Pos::privDir(Pos* p1,Pos* p2)
{
	int dx = p2->getX() -p1->getX();
	int dy = p2->getY() -p1->getY();
	if(dx > 0)
	{
		if(dy > 0)
		{
			// E,SE,S
			if(abs(dx) > abs(dy))
			{
				// SE,S
				if(abs(dy) < (abs(dx)-abs(dy))/2*sqrt(2)) return S;
				else return SE;
			}
			else if(abs(dx) < abs(dy))
			{
				// E,SE
				if(abs(dx) < (abs(dy)-abs(dx))/2*sqrt(2)) return E;
				else return SE;
			}
			else return SE;
		}
		else if(dy < 0)
		{
			// S,SW,W
			if(abs(dx) > abs(dy))
			{
			  	// S,SW
				if(abs(dy) < (abs(dx)-abs(dy))/2*sqrt(2)) return S;
				else return SW;
			}
			else if(abs(dx) < abs(dy))
			{
				// SW,W
				if(abs(dx) < (abs(dy)-abs(dx))/2*sqrt(2)) return W;
				else return SW;
			}
			else return SW;
		}
		else return S;
	}
	else if(dx < 0)
	{
		if(dy > 0)
		{
			// N,NE,E
			if(abs(dx) > abs(dy))
			{
				// N,NE
				if(abs(dy) < (abs(dx)-abs(dy))/2*sqrt(2)) return N;
				else return NE;
			}
			else if(abs(dx) < abs(dy))
			{
				// NE,E
				if(abs(dx) < (abs(dy)-abs(dx))/2*sqrt(2)) return E;
				else return NE;
			}
			else return NE;
		}
		else if(dy < 0)
		{
			// W,NW,N
			if(abs(dx) > abs(dy))
			{
				// NW,N
				if(abs(dy) < (abs(dx)-abs(dy))/2*sqrt(2)) return N;
				else return NW;
			}
			else if(abs(dx) < abs(dy))
			{
				// W,NW
				if(abs(dx) < (abs(dy)-abs(dx))/2*sqrt(2)) return W;
				else return NW;
			}
			else return NW;
		}
		else return N;
	}
	else
	{
		if(dy > 0) return E;
		else if(dy < 0) return W;
		else return UND;
	}
}

// return surrounded direction of dir
vector<dir_t>* Pos::srdDir(dir_t dir)
{
	vector<dir_t>* v = new vector<dir_t>();
	switch(dir)
	{
	case NW :
		v->push_back(W); v->push_back(NW); v->push_back(N);
		break;
	case N : 
		v->push_back(NW); v->push_back(N); v->push_back(NE);
		break;
	case NE : 
		v->push_back(N); v->push_back(NE); v->push_back(E);
		break;
	case E : 
		v->push_back(NE); v->push_back(E); v->push_back(SE);
		break;
	case SE : 
		v->push_back(E); v->push_back(SE); v->push_back(S);
		break;
	case S : 
		v->push_back(SE); v->push_back(S); v->push_back(SW);
		break;
	case SW: 
		v->push_back(S); v->push_back(SW); v->push_back(W);
		break;
	case W : 
		v->push_back(SW); v->push_back(W); v->push_back(NW);
		break;
	default : 
		delete v;
		return NULL;
		break;
	}
	return v;
}


                    /**************************************************************************/
                    /**************************** drawing functions ***************************/
                    /**************************************************************************/

// draw the current point on the map
void Pos::drawOn(Map* map,int col)
{
        // verifying if the position is reachable
	if(this->isReachable(map))
	{
	        // setting the color of the current position
	        map->setCol(*this,col);
		vector<dir_t> v = this->reachableDir2(map);
		
		// setting the color of all positions surrounding the current position
		for(unsigned int i=0;i<v.size();i++)
		{
			Pos p = Pos(*this,v[i]);
			map->setCol(p,col);
		}
	}
}
