#ifndef MPC_PASCALTRIANGLE_HPP
#define MPC_PASCALTRIANGLE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /************** Class PascalTriangle **************/
                                    /**************************************************/

  /** \class PascalTriangle
   *  \brief Pascal's Triangle structures and building class
   */
  class PascalTriangle
  {
  protected : 
    static int ** tab_;                                    /**< table containing the pascal triangle values */
    static int maxOrder_;                                  /**< the maximal order up to which the Pascal's triangle has been already built */

  private : 
    /**
     *  \brief Build the Pascal's triangle n-th order
     *  \param n : the order of the Pascal's triangle that must be computed
     */
    static void buildOrder(int n);

  public : 
    /**
     *  \brief Build the Pascal's triangle up to order n
     *  \param n : the order up to which the Pascal's triangle must be computed
     */
    static void buildUpToOrder(int n);

    /**
     *  \brief return the value of the Pascal's triangle at order i, j-th value
     *  \param i : number of the Pascal's triangle order
     *  \param j : number of the value to compute
     *  \return the value of the Pascal's triangle at order i, j-th value
     */
    static int getVal(int i, int j);

  private : 
    /*
     *  \brief check if the input index is valid
     *  \return true if the input index is valid, false otherwise
     */
    static bool indexIsValid(int i, int j);

    /*
     *  \brief check if the input order has already been computed
     *  \return true if the input order has already been computed, false otherwise
     */
    static bool orderAlreadyComputed(int i);

    /*
     *  \brief clean the Pascal's triangle tables
     */
    static void cleanTables();

    /*
     *  \brief allocate the memory required to build the Pascal's triangle up to order n
     */
    static void allocTablesUpToOrder(int n);

  public : 
    /**
     *  \brief display the current Pascal's triangle table
     */
    static void display();
  };

};

#endif // PASCALTRIANGLE
