#include "nrtypes.h"     //Definition of matrix and vector
#include "nr.h"
#include "nrutil.h"

#include <cmath>

#include <iostream>
using namespace std;

using namespace NR;


Mat_DP prodMatDP(Mat_DP& A, Mat_DP& B)
{
  int M = A.nrows();
  int N = B.ncols();

  int M2 = A.ncols();
  int N1 = B.nrows();

  if(M2 != N1)
    {
      std::cerr << "The second dimension of A and the first dimension of B do not match. Cannot compute A*B." << std::endl;
      return Mat_DP(0,0);
    }

  Mat_DP res(M,N);

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  double sum = 0;
	  for(int k=0; k<M2; k++)
	    {
	      sum += (A[i][k] * B[k][j]);
	    }
	  if(fabs(sum) < 1e-10)
	    res[i][j] = 0.0;
	  else
	    res[i][j] = sum;
	}
    }

  return res;
}

Mat_DP transpMatDP(Mat_DP& A)
{
  int M = A.nrows();
  int N = A.ncols();

  Mat_DP At(N,M);

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  At[j][i] = A[i][j];
	}
    }

  return At;
}

void displayMatDP(Mat_DP& A)
{
  int M = A.nrows();
  int N = A.ncols();

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  cout << A[i][j] << " ";
	}
      cout << endl;
    }
}

int main()
{
//   const int M=4;
//   const int N=3;
  const int M=7;
  const int N=1;


  Mat_DP U(M,N);          // matrice initiale et matrice unitaire apres decomp
  Mat_DP A(M,N);         // copie de la matrice initiale
  Vec_DP E(N);           // vecteur de valeurs singulieres
  Mat_DP V(N,N);        // matrice adjointe

  double d;
  int i,j,k;

//   // initialise la matrice
//   U[0][0] = 1; U[0][1] = 2; U[0][2] = 3; //U[0][3] = 4;
//   U[1][0] = 4; U[1][1] = 2; U[1][2] = 3; //U[1][3] = 5;
//   U[2][0] = 1; U[2][1] = 4; U[2][2] = 5; //U[2][3] = 9;
//   U[3][0] = 2; U[3][1] = 0; U[3][2] = 6; //U[3][3] = 3;
  
  U[0][0] = 1; U[0][1] = 0.5; U[0][2] = 0.333333; U[0][3] = 0.25; U[0][4] = 0.0; U[0][5] = 0.0; U[0][6] = 0.0; 

  // effectue une copie de la matrice initiale
  for(i=0; i<M; i++)
    {
      for(j=0; j<N; j++)
	{
	  A[i][j] = U[i][j];
	}
    }

  // affiche la matrice initiale
  cout << "Matrice A : " << endl; 
  displayMatDP(A);
  
  svdcmp(U, E, V);

  cout << "U : " << U.nrows() << " " << U.ncols() << endl;
  cout << "E : " << E.size() << endl;
  cout << "V : " << V.nrows() << " " << V.ncols() << endl;

  // affiche la matrice orthogonale U
  cout << "Matrice U : " << endl; 
  displayMatDP(U);

  // affiche la matrice des valeurs singulieres E
  Mat_DP matE(N,N);
  for(i=0; i<N; i++)
    {
      for(j=0; j<N; j++)
	{
	  if(i == j) matE[i][j] = E[i];
	  else matE[i][j] = 0;
	}
    }
  
  cout << "Matrice E : " << endl; 
  displayMatDP(matE);

  // affiche la matrice orthogonale V
  cout << "Matrice V : " << endl; 
  displayMatDP(V);

  // calcule les transposees de U et V
  Mat_DP Ut = transpMatDP(U);
  Mat_DP Vt = transpMatDP(V);

//   // affiche la transposee de U
//   cout << "Matrice Ut : " << endl; 
//   displayMatDP(Ut);

//   // affiche la transposee de V
//   cout << "Matrice Vt : " << endl; 
//   displayMatDP(Vt);

  // calcul et affiche le produit U Ut
  Mat_DP prodUUt = prodMatDP(Ut,U);
  cout << "Matrice Ut*U : " << endl;
  displayMatDP(prodUUt);

  // calcul et affiche le produit V Vt
  Mat_DP prodVVt = prodMatDP(V,Vt);
  cout << "Matrice V*Vt : " << endl;
  displayMatDP(prodVVt);  

  // calcul et affiche le produit U E Vt
  Mat_DP prodUE = prodMatDP(U,matE);
  Mat_DP prodUEVt = prodMatDP(prodUE,Vt);
  cout << "Matrice U*E*Vt : " << endl;
  displayMatDP(prodUEVt); 

  // cree la matrice contenant l'oppose de la matrice des valeurs singulieres
  Mat_DP Einv(N,N);
  for(i=0; i<N; i++)
    {
      for(j=0; j<N; j++)
	{
	  if(i == j) Einv[i][j] = 1.0/E[i];
	  else Einv[i][j] = 0;
	}
    }

  // affiche la matrice Einv
  cout << "Matrice Einv : " << endl; 
  displayMatDP(Einv);
  
  // calcule la matrice inverse de A
  Mat_DP prod1 = prodMatDP(V,Einv);
  Mat_DP Ainv = prodMatDP(prod1,Ut);

  // affiche la matrice Ainv inverse de A
  cout << "Matrice Ainv : " << endl; 
  displayMatDP(Ainv);

  cout << "A : " << A.nrows() << " " << A.ncols() << endl;
  cout << "Ainv : " << Ainv.nrows() << " " << Ainv.ncols() << endl;

  // calcule le produit A*Ainv
  Mat_DP prod2 = prodMatDP(A,Ainv);
  Mat_DP prod3 = prodMatDP(Ainv,A);

  // affiche le produit de A*Ainv
  cout << "Matrice A*Ainv : " << endl;
  displayMatDP(prod2);

  // affiche le produit de Ainv*A
  cout << "Matrice Ainv*A : " << endl;
  displayMatDP(prod3);

  return 0;
}

