#include "Polynomial.hpp"
#include "MultiItvFunction.hpp"

#include "Utils.hpp"
#include "gnuplot/inc/gnuplot_i.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

#define STEP 300
#define BINF -10.0
#define BSUP 15.0

int main()
{
  // compute a polynom, its derivation and integration
  Polynomial poly(-5,40,-8);
  Function* deriv = poly.derivation();
  Function* itg = poly.integration(-6);

  // search for the roots of the function
  std::vector<double> rpoly,rderiv,ritg;
  poly.realRoots(rpoly,0.0);
  deriv->realRoots(rderiv,0.0);
  itg->realRoots(ritg,300.0);

  cout << "Polynomial : " << endl;
  poly.display();
  cout << "roots : "; displayVectDble(&rpoly);
  for(int i=0; i<rpoly.size(); i++)
    {
      cout << "f(" << rpoly[i] << ") = " << poly.getVal(rpoly[i]) << endl;
    }
  cout << endl;

  cout << "Derivation : " << endl;
  deriv->display();
  cout << "roots : "; displayVectDble(&rderiv); cout << endl;

  cout << "Integration : " << endl;
  itg->display();
  cout << "roots : "; displayVectDble(&ritg);
  for(int i=0; i<ritg.size(); i++)
    {
      cout << "f(" << ritg[i] << ") = " << itg->getVal(ritg[i]) << endl;
    }
  cout << endl;

  // compute the ladder of time
  double dt = (BSUP - BINF) / STEP;
  vector<double> stime;
  for(int i=0; i<=STEP; i++)
    {
      stime.push_back(BINF + i*dt);
    }

  // sample the three functions
  vector<double> spoly,sderiv,sitg;
  poly.sample(BINF, BSUP, STEP, spoly);
  deriv->sample(BINF, BSUP, STEP, sderiv);
  itg->sample(BINF, BSUP, STEP, sitg);

  // display the three functions in gnuplot windows
  Gnuplot gPoly;
  gPoly.reset_plot();
  gPoly.set_style("points").plot_xy(stime,spoly,"2nd oder polynom");

  Gnuplot gDeriv;
  gDeriv.reset_plot();
  gDeriv.set_style("points").plot_xy(stime,sderiv,"Derivative (1st oder polynom)");

  Gnuplot gItg;
  gItg.reset_plot();
  gItg.set_style("points").plot_xy(stime,sitg,"Integration (3rd oder polynom)");  

  wait_for_key();

  // create a multiple interval function from the 3rd order polynomial
  MultiItvFunction mitvItg(itg, itg->binf_,itg->bsup_);

  mitvItg.addAbsoluteValue();

  // sample the absolute value function
  vector<double> sabsMitvItg;
  mitvItg.sample(BINF, BSUP, STEP, sabsMitvItg);  

  // display the new function in a gnuplot window
  Gnuplot gabsMitvItg;
  gabsMitvItg.reset_plot();
  gabsMitvItg.set_style("points").plot_xy(stime,sabsMitvItg,"Absolute value function");

//   // change a part of the interval
//   Polynomial cste(200.0);
//   mitvItg.changeItv(3.0,5.5,&cste);
  
//   // sample the new function
//   vector<double> sMitvItg;
//   mitvItg.sample(BINF, BSUP, STEP, sMitvItg);

//   // display the new function in a gnuplot window
//   Gnuplot gMitvItg;
//   gMitvItg.reset_plot();
//   gMitvItg.set_style("points").plot_xy(stime,sMitvItg,"3rd oder polynom with change");

//   wait_for_key();

//   Polynomial linear(200.0,-400.0);
//   mitvItg.changeItv(2.0,4.0,&linear);

//   // sample the new function
//   vector<double> sMitvItg2;
//   mitvItg.sample(BINF, BSUP, STEP, sMitvItg2);

//   // display the new function in a gnuplot window
//   Gnuplot gMitvItg2;
//   gMitvItg2.reset_plot();
//   gMitvItg2.set_style("points").plot_xy(stime,sMitvItg2,"3rd oder polynom with change");

//   wait_for_key();

  cout << "//////////////////////////////////////////////////////////////////" << endl;
  
  std::vector<double> roots1,roots2;

  mitvItg.realRoots(roots1,0.0);
  cout << "roots1 : "; displayVectDble(&roots1);

  cout << "Fonction sans bornes : " << endl;
  mitvItg.display();

  // add a lower bound
  mitvItg.addMinBound(-1000.0);

  cout << "Fonction avec borne inf : " << endl;
  mitvItg.display();

  cout << "//////////////////////////////////////////////////////////////////" << endl;

  mitvItg.realRoots(roots2,0.0);
  cout << "roots2 : "; displayVectDble(&roots2);

  mitvItg.addMaxBound(320.0);
  
  cout << "Fonction bornee : " << endl;
  mitvItg.display();

  cout << "//////////////////////////////////////////////////////////////////" << endl;

  // sample the new function
  vector<double> sMitvItg;
  mitvItg.sample(BINF, BSUP, STEP, sMitvItg);  

  // display the new function in a gnuplot window
  Gnuplot gMitvItg;
  gMitvItg.reset_plot();
  gMitvItg.set_style("points").plot_xy(stime,sMitvItg,"3rd oder polynom with change");

  // compute the derivation and integration of the new function
  Function* derivMitvItg = mitvItg.derivation();
  Function* itgMitvItg = mitvItg.integration(0.0,0.0);

  cout << "Integrale fonction bornee : " << endl;
  itgMitvItg->display();

  // sample it
  vector<double> sDerivMitvItg,sItgMitvItg ;
  derivMitvItg->sample(BINF, BSUP, STEP, sDerivMitvItg);
  itgMitvItg->sample(BINF, BSUP, STEP, sItgMitvItg);

  //display it
  Gnuplot gDerivMitvItg;
  gDerivMitvItg.reset_plot();
  gDerivMitvItg.set_style("points").plot_xy(stime,sDerivMitvItg,"2nd order polynom with change");

  Gnuplot gItgMitvItg;
  gItgMitvItg.reset_plot();
  gItgMitvItg.set_style("points").plot_xy(stime,sItgMitvItg,"4th order polynom with change");

  wait_for_key();  

  return 0;
}
