#ifndef MPC_SHIFT_HPP
#define MPC_SHIFT_HPP

#include "common/inc/Vector.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class Shift ******************/
                                    /**************************************************/

  /** \class Shift
   *  \brief Shift between two paths
   */
  class Shift : public Vector
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Shift();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    Shift(const Shift& s);

    /**
     *  \brief destructor
     */
    virtual ~Shift();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    virtual Shift* copy() const = 0;

  protected : 
    /**
     *  \brief init constructor
     *  \param n : dimension of the state
     */
    Shift(int n);
  };
};

#endif // MPC_SHIFT_HPP


#ifndef MPC_YTHSHIFT_HPP
#define MPC_YTHSHIFT_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class YThShift *****************/
                                    /**************************************************/

  /** \class YThShift
   *  \brief Lateral and angular shift between two paths
   */
  class YThShift : public Shift
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    YThShift();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    YThShift(const YThShift& s);

    /**
     *  \brief init constructor
     *  \param y : the lateral shift
     *  \param th : the angular shift
     */
    YThShift(double y, double th);

    /**
     *  \brief destructor
     */
    virtual ~YThShift();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    virtual YThShift* copy() const  { return new YThShift(*this); }

  protected : 
    /**
     *  \brief init constructor
     *  \param n : dimension of the state
     */
    YThShift(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : lateral shift Y
     *  \return lateral shift Y
     */
    inline double& y() { return (*this)[0]; }

    /**
     *  \brief getter : angular shift Th
     *  \return angular shift Th
     */
    inline double& th() { return (*this)[1]; }

    /**
     *  \brief getter : lateral shift Y
     *  \return lateral shift Y
     */
    inline double y() const { return this->at(0); }

    /**
     *  \brief getter : angular shift Th
     *  \return angular shift Th
     */
    inline double th() const { return this->at(1); }
  };
};

#endif // MPC_YTHSHIFT_HPP


#ifndef MPC_YTHSSHIFT_HPP
#define MPC_YTHSSHIFT_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class YThSShift ****************/
                                    /**************************************************/

  /** \class YThSShift
   *  \brief Lateral and angular shift between two paths given at curve abscica
   */
  class YThSShift : public YThShift
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    YThSShift();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    YThSShift(const YThSShift& s);

    /**
     *  \brief init constructor
     *  \param y : the lateral shift
     *  \param th : the angular shift
     *  \param s : the curve abscica
     */
    YThSShift(double y, double th, double s);

    /**
     *  \brief destructor
     */
    virtual ~YThSShift();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    virtual YThSShift* copy() const  { return new YThSShift(*this); }

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : curve abscica S
     *  \return curve abscica S
     */
    inline double& s() { return (*this)[2]; }

    /**
     *  \brief getter : curve abscica S
     *  \return curve abscica S
     */
    inline double s() const { return this->at(2); }
  };
};

#endif // MPC_YTHSHIFT_HPP
