#include "motion/inc/ShiftRecord.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class ShiftRecord ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
ShiftRecord::ShiftRecord()
{
}

/**
 *  \brief destructor
 */
ShiftRecord::~ShiftRecord()
{
}


                                    /**************************************************/
                                    /**************** Class DShiftRecord **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DShiftRecord::DShiftRecord()
{
}

/**
 *  \brief copy constructor
 *  \param t : the trajectory to copy
 */
DShiftRecord::DShiftRecord(const DShiftRecord& r)
{
}

/**
 *  \brief destructor
 */
DShiftRecord::~DShiftRecord()
{
  this->clear();
}

/************** Getters and setters ***************/

/**
 *  \brief return true if the record is empty
 *  \return true if the record is empty
 */
bool DShiftRecord::empty() const
{
  return listShift_.empty();
}

/**************** list modification ***************/

/**
 *  \brief pop out the last element of the record
 */
void DShiftRecord::pop_back()
{
  this->listShift_.pop_back();
}

/**
 *  \brief remove the i-th element of the record
 *  \param i : index of the value to remove
 */
void DShiftRecord::remove(int i)
{
  if(i >= 0 && i < this->size())
    {
      vector <Shift*>::iterator it = this->listShift_.begin();
      for(int j=0; j<i; j++) it++;
      this->listShift_.erase(it);
    }
}

/**
 *  \brief remove the all elements of the record between start and end
 *  \param i : index of the value to remove
 *
 *  including start but not including end
 */
void DShiftRecord::remove(int start,int end)
{
  if(start >= 0 && start <= end && end < this->size())
    {
      vector <Shift*>::iterator itDeb = this->listShift_.begin();
      vector <Shift*>::iterator itFin = this->listShift_.begin();
      for(int j=0; j<start; j++) itDeb++;
      for(int j=0; j<end; j++) itFin++;
      this->listShift_.erase(itDeb,itFin);
    }
}

/**
 *  \brief clear the content of the trajectory
 */
void DShiftRecord::clear()
{
  for(int i=0; i<this->size(); i++)
    {
      delete this->listShift_.at(i);
    }
  this->listShift_.clear();
}


                                    /**************************************************/
                                    /************* Class YThS_DShiftRecord ************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
YThS_DShiftRecord::YThS_DShiftRecord()
{
}

/**
 *  \brief copy constructor
 *  \param r : the record to copy
 */
YThS_DShiftRecord::YThS_DShiftRecord(const YThS_DShiftRecord& r)
{
  for(int i=0; i<r.size(); i++)
    {
      this->append(r.at(i));
    }
}

/**
 *  \brief destructor
 */
YThS_DShiftRecord::~YThS_DShiftRecord()
{
}


/************** Getters and setters ***************/

/**
 *  \brief getter : middle value of the record
 *  \return the middle value of the record
 */
const YThSShift& YThS_DShiftRecord::middle() const
{
  if(this->size() != 0)
    return this->at( ((int) (0.5*this->size())) );
  else
    throw string("Empty record");
}

/**
 *  \brief getter : first value of the record
 *  \return the first value of the record
 */
const YThSShift& YThS_DShiftRecord::first() const
{
  if(this->size() != 0)
    return this->at(0);
  else
    throw string("Empty record");
}

/**
 *  \brief getter : last value of the record
 *  \return the last value of the record
 */
const YThSShift& YThS_DShiftRecord::last() const
{
  if(this->size() != 0)
    return this->at(this->size()-1);
  else
    throw string("Empty record");
}

/**
 *  \brief getter : i-th element of the record
 *  \return return the i-th element of the record
 */
const YThSShift& YThS_DShiftRecord::at(int i) const
{
  if(i>=0 && i<this->size())
    return *((YThSShift*)(this->listShift_.at(i)));
  else
    throw string("Index out of bounds of the record");
}


/**************** List modification ***************/

/**
 *  \brief append a new value at the end of the record
 *  \param s : the new value to append
 */
void YThS_DShiftRecord::append(const YThSShift& s)
{
  YThSShift* newS = new YThSShift(s);
  this->listShift_.push_back(newS);
}

/**
 *  \brief concatenate two trajectories
 *  \param t : the trajectory to concatenate to the current one
 */
void YThS_DShiftRecord::concat(const YThS_DShiftRecord& r)
{
  for(int i=0; i<r.size(); i++)
    {
      this->append(r.at(i));
    }
}

/**
 *  \brief insert the value s at index position i in the list
 *  \param i : index to insert the new value
 *  \param s : the new value to insert
 */
void YThS_DShiftRecord::insert(int i, const YThSShift& s)
{
  vector <Shift*>::iterator it = this->listShift_.begin();
  for(int j=0; j<i; j++) it++;
  YThSShift* newS = new YThSShift(s);
  this->listShift_.insert(it,newS);
}


/******************** Operators *******************/

/**
 *  \brief getter : i-th element of the record
 *  \return return the i-th element of the record
 */
YThSShift& YThS_DShiftRecord::operator[](int i)
{
  if(i>=0 && i<this->size())
    return *((YThSShift*)(this->listShift_[i]));
  else
    throw string("Index out of bounds of the record");
}
