#include "network/inc/TCP_server.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

boost::mutex mutex;

/**
 *  \brief default constructor
 */
TCP_server::TCP_server()
{
  port_ = 0;

  length_ = 0;
  message_ = NULL;

  waiting_time_ms_ = 10;
}

/**
 *  \brief constructor
 */
TCP_server::TCP_server(int length, int port, int wtime)
{
  port_ = port;

  length_ = length;
  message_ = new double[length];
  for(int i=0; i<length; i++) message_[i] = 0;

  waiting_time_ms_ = wtime;
}

/**
 *  \brief destructor
 */
TCP_server::~TCP_server()
{
  if(message_ != NULL) delete message_;
}

/*------------------------------------------------------------------------
 *  TCP server session
 */
void TCP_server::session()
{
  while(true)
    {
      boost::this_thread::sleep(boost::posix_time::milliseconds(waiting_time_ms_));
//       cout << "socket vector size : " << sock.size() << endl;

      // write the state vector into the socket buffer
      boost::mutex::scoped_lock lock(mutex);
      for(std::vector<socket_ptr>::iterator it = sock.begin(); it != sock.end(); ++it)
	{
	  try 
	    {
	      boost::mutex::scoped_lock lock(mutexMessage_);
	      boost::asio::write(*(*it), boost::asio::buffer(message_, length_*sizeof(double)));
	    }
	  catch(std::exception e)
	    {
	      std::cout << "client disconnected" << std::endl;
	      sock.erase(it);
	      break;				
	    }
	}
    }
}

/*------------------------------------------------------------------------
 *  TCP connection server
 */
void TCP_server::server(boost::asio::io_service& io_service)
{
  boost::thread t(boost::bind(&TCP_server::session,this));
  
  tcp::acceptor a(io_service, tcp::endpoint(tcp::v4(), port_));
  for (;;)
    {
      socket_ptr mysock(new tcp::socket(io_service));
      a.accept(*mysock);
      
      cout << "client connected" << endl;
//       mutex.lock();
      boost::mutex::scoped_lock lock(mutex);
      sock.push_back(mysock);
//       mutex.unlock();
    }
}

/*------------------------------------------------------------------------
 *  Start the server thread
 */
void TCP_server::startThreadServer()
{
  try
    {
      boost::asio::io_service io_service;
      server(io_service);
    }
  catch (std::exception& e)
    {
      std::cerr << "Exception: " << e.what() << "\n";
    }
}

/*------------------------------------------------------------------------
 *  Start the TCP Communication
 */
void TCP_server::start()
{
  boost::thread t(boost::bind(&TCP_server::startThreadServer,this));
}

/*------------------------------------------------------------------------
 *  Stop the TCP Communication
 */
void TCP_server::stop()
{
}

/*------------------------------------------------------------------------
 *  Set the sent message
 */
void TCP_server::setMessage(const double* msg, int length)
{
  boost::mutex::scoped_lock lock(mutexMessage_);
  for(int i=0; i<length; i++)
    message_[i] = msg[i];
}
