#include "PlannerIfc.hpp"
#include "Utils.hpp"

#include "STrajectory.hpp"
#include "PVDTrajPlanner.hpp"
#include "Environment.hpp"

#include "PointMassSys.hpp"

#include <fstream>
#include <iostream>
#include <string>

#include <time.h>
#include <sys/time.h>

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// default constructor (only used by subclasses)
PlannerIfc::PlannerIfc(){}

// init constructor (creation of the window)
PlannerIfc::PlannerIfc(FXApp* a):FXMainWindow(a,"Trajectory Planning",NULL,NULL,DECOR_ALL,0,0,950,650)
{
  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  map = NULL;

  maxSpd = 0;
  maxAcc = 0;

  thresMap = 0;
  incObs = 0;

  targetPosX = 850;
  targetPosY = 500;

  initPosX = 150;
  initPosY = 250;
  goalPosX = 850;
  goalPosY = 500;

  copyMapName = FXString("../map/ResultMap_c.pgm");
  thresMapName = FXString("../map/ResultMap_s.pgm");
  incObsName = FXString("../map/ResultMap_g.pgm");
  pfMapName = FXString("../map/ResultMap_pf.pgm");
  pfObsMapName = FXString("../map/ResultMap_pfObs.pgm");
  pfTgtMapName = FXString("../map/ResultMap_pfTgt.pgm");
  pfMapPPName = FXString("../map/ResultMap_pf.pgm");
  resMapPPName = FXString("../map/ResultMap_path.pgm");
  resPathPPName = FXString("../path/path.pth");

  // connecting system configuration values to interface components
  maxSpd_target.connect(maxSpd);
  maxAcc_target.connect(maxAcc);

  // connecting map's parameters values to interface components
  thresMap_target.connect(thresMap);
  incObs_target.connect(incObs);
  targetPosX_target.connect(targetPosX);
  targetPosY_target.connect(targetPosY);
  initPosX_target.connect(initPosX);
  initPosY_target.connect(initPosY);
  goalPosX_target.connect(goalPosX);
  goalPosY_target.connect(goalPosY);

  copyMapName_target.connect(copyMapName);
  thresMapName_target.connect(thresMapName);
  incObsName_target.connect(incObsName);
  pfMapName_target.connect(pfMapName);
  pfObsMapName_target.connect(pfObsMapName);
  pfTgtMapName_target.connect(pfTgtMapName);
  pfMapPPName_target.connect(pfMapPPName);
  resMapPPName_target.connect(resMapPPName);
  resPathPPName_target.connect(resPathPPName);

  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/

  contents=new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  
  /***************************************************************************************************************
   *********************************************** viewer frame **************************************************
   ***************************************************************************************************************/

  // LEFT pane to contain the canvas
  imageFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the canvas
  new FXLabel(imageFrame,"Viewer",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(imageFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

  // Make image widget
  imageview=new FXImageView(imageFrame,NULL,NULL,0,LAYOUT_FILL_X|LAYOUT_FILL_Y);

//   glvisual=new FXGLVisual(getApp(),VISUAL_DOUBLEBUFFER);
//   viewer=new FXGLViewer(glMainFrame,glvisual,this,ID_GLVIEWER,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT);
//   viewer->setBackgroundColor (FXVec4f(0.0f,0.0f,0.0f,0.0f));

  msgBox = new FXLabel(imageFrame,"No map loaded. Click on \"Load map\" to load a new map.",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  /***************************************************************************************************************
   ********************************************** commands frame *************************************************
   ***************************************************************************************************************/

  // RIGHT pane for the buttons
  buttonFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the buttons
  new FXLabel(buttonFrame,"Commands",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Tab book with switchable panels
  FXTabBook* panels=new FXTabBook(buttonFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /***************************************************************************************************************
   ***************************************** system configuration tab ********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Sys config");

  FXVerticalFrame *tabSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // System Configuration
  new FXLabel(tabSysCfg,"System configuration");

  FXMatrix *matSysCfg=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSysCfg,"Max speed :");
  new FXTextField(matSysCfg,6,&maxSpd_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max acceleration :");
  new FXTextField(matSysCfg,6,&maxAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  /***************************************************************************************************************
   ********************************************* map treatment tab ***********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Map Treatment");

  FXVerticalFrame *tabMapTtt=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

    FXMatrix *matCopyMap=new FXMatrix(tabMapTtt,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
    new FXLabel(matCopyMap,"Copy map filename :");
  new FXButton(matCopyMap,"...",NULL,this,ID_COPY_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matCopyMap,25,&copyMapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
 
  // Button to copy the map
  new FXButton(tabMapTtt,"&Copy map",NULL,this,ID_COPY_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

    FXMatrix *matThresMap=new FXMatrix(tabMapTtt,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matThresMap,"Threshold :");
  new FXTextField(matThresMap,6,&thresMap_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matThresMap,"Threshold map filename :");
  new FXButton(matThresMap,"...",NULL,this,ID_THRES_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matThresMap,25,&thresMapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to threshold the map
  new FXButton(tabMapTtt,"&Threshold map",NULL,this,ID_THRES_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matIncObs=new FXMatrix(tabMapTtt,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matIncObs,"Obstacles increasing :");
  new FXTextField(matIncObs,6,&incObs_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matIncObs,"Obstacles increasing filename :");
  new FXButton(matIncObs,"...",NULL,this,ID_INC_OBS_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matIncObs,25,&incObsName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to increase obstacles
  new FXButton(tabMapTtt,"&Increase Obstacles",NULL,this,ID_INC_OBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  /***************************************************************************************************************
   ******************************************** potential fields tab *********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Potential fields");

  FXVerticalFrame *tabPF=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matTgtPosPFMap=new FXMatrix(tabPF,3,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTgtPosPFMap,"Target position :");
  new FXTextField(matTgtPosPFMap,6,&targetPosX_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXTextField(matTgtPosPFMap,6,&targetPosY_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matPfMapObs=new FXMatrix(tabPF,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPfMapObs,"Obstacles PF map filename :");
  new FXButton(matPfMapObs,"...",NULL,this,ID_PFOBS_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matPfMapObs,25,&pfObsMapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to add obstacles potential field
  new FXButton(tabPF,"Add &Obstacles potential field",NULL,this,ID_PFOBS_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matPfMapTgt=new FXMatrix(tabPF,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPfMapTgt,"Target position PF map filename :");
  new FXButton(matPfMapTgt,"...",NULL,this,ID_PFTGT_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matPfMapTgt,25,&pfTgtMapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to add target position potential field
  new FXButton(tabPF,"Add &Target position potential fields",NULL,this,ID_PFTGT_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matPfMap=new FXMatrix(tabPF,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPfMap,"Potential field map filename :");
  new FXButton(matPfMap,"...",NULL,this,ID_PF_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matPfMap,25,&pfMapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to add potential fields
  new FXButton(tabPF,"Add &Potential fields",NULL,this,ID_PF_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  /***************************************************************************************************************
   ********************************************* path planning tab ***********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Path planning");

  FXVerticalFrame *tabPP=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matPosPathPl=new FXMatrix(tabPP,3,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPosPathPl,"Init position :");
  new FXTextField(matPosPathPl,6,&initPosX_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXTextField(matPosPathPl,6,&initPosY_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matPosPathPl,"Goal position :");
  new FXTextField(matPosPathPl,6,&goalPosX_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXTextField(matPosPathPl,6,&goalPosY_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matPfMapPathPl=new FXMatrix(tabPP,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPfMapPathPl,"Potential field map filename :");
  new FXButton(matPfMapPathPl,"...",NULL,this,ID_PF_MAP_PP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matPfMapPathPl,25,&pfMapPPName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matResMapPathPl=new FXMatrix(tabPP,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matResMapPathPl,"Result map filename :");
  new FXButton(matResMapPathPl,"...",NULL,this,ID_RES_MAP_PP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matResMapPathPl,25,&resMapPPName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matResPathPathPl=new FXMatrix(tabPP,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matResPathPathPl,"Result path filename :");
  new FXButton(matResPathPathPl,"...",NULL,this,ID_RES_PATH_PP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matResPathPathPl,25,&resPathPPName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to compute the path
  new FXButton(tabPP,"&Compute the path",NULL,this,ID_PATH_PL,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  /***************************************************************************************************************
   ********************************************* general commands ************************************************
   ***************************************************************************************************************/

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Button to load a map from a file
  new FXButton(buttonFrame,"&Load Map from File",NULL,this,ID_LOAD_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

#ifdef HUGR
  // Button to load a map from hugr
  new FXButton(buttonFrame,"&Load Map from Hugr",NULL,this,ID_LOAD_HUGRMAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
#endif

  // Exit button
  new FXButton(buttonFrame,"&Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   rb_ = new Robot();
//   this->loadSystemCfg(QString("configRobot.rcfg"));

  // this->initScene();
}

// void PlannerIfc::initScene()
// { 
  // initiating of the system configuration
//   rb_ = new Robot();
//   rb_->setVMax(maxSpd);
//   rb_->setAMax(maxAcc);
// }

// void PlannerIfc::loadSystemCfg(const QString& filename)
// {
//   QDomDocument doc( "configSystem" );
//   QFile file( filename );
  
//   if ( !file.open( QIODevice::ReadOnly ) )
//     {
//       cout << "Can't open the file : " << filename.toStdString() << endl;
//       return;
//     }
  
//   if ( !doc.setContent( &file ) ) {
//     cout << "Can't set the content of the document" << endl;
//     file.close();
//     return;
//   }
//   file.close();
  
//   QDomElement docElem = doc.documentElement();
  
//   QDomNode n = docElem.firstChild();
//   while( !n.isNull() ) 
//     {
//       QDomElement e = n.toElement(); // try to convert the node to an element.
//       if( !e.isNull() ) 
// 	{
// 	  // initialization of the current element ofthe system
// 	  if(e.tagName() == QString("Robot"))
// 	    {
// 	      this->rb_->initFromDOMElement(e);
// 	      maxSpd = rb_->getVMax();
// 	      maxAcc = rb_->getAMax();
// 	      cout << "init robot" << endl;
// 	    }
	  
// 	  else if(e.tagName() == QString("State"))
// 	    {
// 	      QString name;
// 	      if(e.hasAttribute(QString("name")))
// 		{
// 		  name = e.attribute(QString("name"),QString::null);
// 		}
// 	      else name = QString("none");
	      
// 	      if(name == QString("s1"))
// 		{
// 		  s1_.initFromDOMElement(e);
// 		  stbPos = s1_.getCfg().getPos()[0];
// 		  stbTim = s1_.getTime();
// 		  stbSpd = s1_.getSpd();
// 		  cout << "init s1" << endl;
// 		}
// 	      else if(name == QString("s2"))
// 		{
// 		  s2_.initFromDOMElement(e);
// 		  stfPos = s2_.getCfg().getPos()[0];
// 		  stfTim = s2_.getTime();
// 		  stfSpd = s2_.getSpd();
// 		  cout << "init s2" << endl;
// 		}
// 	    }
// 	}
//       n = n.nextSibling();
//     }
// }

// destructor
PlannerIfc::~PlannerIfc()
{
  if(this->map != NULL)
    delete map;
}

// Create and initialize
void PlannerIfc::create(){

  // Create the windows
  FXMainWindow::create();

  // Make the main window appear
  show(PLACEMENT_SCREEN);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long PlannerIfc::loadMapFromFile(FXObject*,FXSelector,void*)
{
  cout << "Load Map" << endl;
  FXFileDialog open(this,"Load map (PGM Image)");
  open.setFilename(filename);
  if(open.execute()){
    filename=open.getFilename();
    loadimage(filename);
  }
  return 1;
}

#ifdef HUGR
long PlannerIfc::loadHugrMap(FXObject*,FXSelector,void*)
{
  cout << "loadHugrMap" << endl;

  FXImage *img=NULL;
  FXImage *old;

  // delete the old map
  if(map != NULL) delete map;
  map = new MapHugr();

  cout << "loading the map" << endl;

  // Load the map
  map->load("mapImage");

  img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);

  cout << "map loaded" << endl;

  // Save the map in a file
  if(map->save("/tmp/MapHugr.pgm") != 0)
    {
      cout << "PlannerIfc::loadHugrMap : Error while trying to save the map. Check the output path. " << endl;
      FXMessageBox::error(this,MBOX_OK,"Error Loading Hugr Map","Error while trying to save the temporary map. Check the output path.");      
      return -1;
    }
  else
    {
      cout << "map saved" << endl;
    }

  FXString msg("A map has been loaded from Hugr");
  msgBox->setText(msg);

  cout << "da" << endl;

  // Load the image
  FXFileStream stream;
  if(stream.open("/tmp/MapHugr.pgm",FXStreamLoad)){

    cout << "stream opened" << endl;
    
    getApp()->beginWaitCursor();
    img->loadPixels(stream);
    stream.close();

    cout << "stream closed" << endl;

    img->create();

    // rescaling the image
    int wtF = imageview->getWidth();
    int htF = imageview->getHeight();

    int wtI = img->getWidth();
    int htI = img->getHeight();

    double ratioW = ((double) wtI) / ((double) wtF);
    double ratioH = ((double) htI) / ((double) htF);

    if(ratioW > ratioH)
      {
	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
      }
    else
      {
	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)(((double)img->getHeight())/ratioH));
      }

    old=imageview->getImage();
    imageview->setImage(img);
    delete old;
    getApp()->endWaitCursor();
    }
  return 1;
}
#endif

long PlannerIfc::closeMap(FXObject*,FXSelector,void*)
{
  return 1;
}

long PlannerIfc::copyMap(FXObject*,FXSelector,void*)
{
  if(map != NULL)
    {
      if(map->save(copyMapName.text()) != 0)
	{
	  FXMessageBox::error(this,MBOX_OK,"Error Copying the Map","Error while trying to save the map. Check the output path.");
	  return -3;
	}
      this->loadimage(copyMapName);
      return 1;
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error Copying Image","An image must be already loaded");
      return -2;
    }
}

long PlannerIfc::thresholdMap(FXObject*,FXSelector,void*)
{
  if(map != NULL)
    {
      if((thresMap >= 0) && (thresMap <= map->getDepth()))
	{
	  map->thresholdPGM(thresMap);
	  if(map->save(thresMapName.text()) != 0)
	    {
	      cout << "PlannerIfc::thresholdMap : Error while trying to save the map. Check the output path. " << endl;
	      FXMessageBox::error(this,MBOX_OK,"Error Thresholding Image","Error while trying to save the map. Check the output path.");
	      return -3;
	    }
	  this->loadimage(thresMapName);
	  return 1;
	}
      else
	{
	  cout << "PlannerIfc::thresholdMap : The threshold must be between 0 and " << map->getDepth() << endl;
	  FXMessageBox::error(this,MBOX_OK,"Error Thresholding Image","The threshold must be between 0 and %d",map->getDepth());
	  return -1;
	}
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error Thresholding Image","An image must be already loaded");
      return -2;
    }
}

long PlannerIfc::increaseObs(FXObject*,FXSelector,void*)
{
  if(map != NULL)
    {
      if(map->getType() == Map::PGM_T)
	{
	  if(incObs >= 0)
	    {
	      // initialize the timer
	      struct timeval t1,t2;
	      gettimeofday(&t1,NULL);

	      map->increaseObs(incObs);
	      if(map->save(incObsName.text()) != 0)
		{
		  cout << "PlannerIfc::increaseObs : Error while trying to save the map. Check the output path. " << endl;
		  FXMessageBox::error(this,MBOX_OK,"Error Increasing Obstacles","Error while trying to save the map. Check the output path.");
	      return -4;
		}
	      this->loadimage(incObsName);

	      // stop the timer
	      gettimeofday(&t2,NULL);
	      // compute the time past
	      int elapsedTs = t2.tv_sec - t1.tv_sec;
	      int elapsedTus = t2.tv_usec - t1.tv_usec;
	      double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
	      FXString msg("Obstacles have been increased in ");
	      msg += FXStringVal(elapsedT,2,FALSE);
	      msg += FXString(" sec.");
	      msgBox->setText(msg);
	      return 1;
	    }
	  else
	    {
	      FXMessageBox::error(this,MBOX_OK,"Error Loading Image","The obstacles increasing value must be greater than 0");
	      return -3;
	    }
	}
      else
	{
	  cout << "PlannerIfc::increaseObs : The input map must have been thresholded before growing obstacles" << endl;
	  FXMessageBox::error(this,MBOX_OK,"Error Loading Image","The input map must have been thresholded before growing obstacles");
	  return -1;
	}
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error Increasing Obstacles","An image must be already loaded");
      return -2;
    }
}

long PlannerIfc::pfMap(FXObject*,FXSelector,void*)
{
  if(map != NULL)
    {
      if(map->getType() == Map::PGM_T)
	{
	  Pos p = Pos(targetPosX,targetPosY);
	  if(p.isReachable(map))
	    {
	      // adding potential field from target position
	      Map* map1 = map->computeWave(&p);
	      string nom1 = addSuffixe(string(pfMapName.text()),string("_tp"));
	      if(map1->save(nom1.c_str()) != 0)
		{
		  FXMessageBox::error(this,MBOX_OK,"Error Computing Potential fields","Error while trying to save the map. Check the output path.");
		  return -4;
		}
	      // adding potential field from obstacles
	      Map* map2 = map->applyPFields3();
	      string nom2 = addSuffixe(string(pfMapName.text()),string("_pfo"));
	      if(map2->save(nom2.c_str()) != 0)
		{
		  FXMessageBox::error(this,MBOX_OK,"Error Computing Potential fields","Error while trying to save the map. Check the output path.");
		  return -5;		  
		}
	      // adding both potential fields
	      //Map* mapRes = Map::mediumMap(map1,map2);
	      //Map* mapRes = Map::addingPFMap(map1,map2);
	      Map* mapRes = Map::addingPFMap2(map1,map2);
	      // saving maps
	      if(mapRes->save(pfMapName.text()) != 0)
		{
		  FXMessageBox::error(this,MBOX_OK,"Error Computing Potential fields","Error while trying to save the map. Check the output path.");
		  return -6;		  
		}
	      
	      this->loadimage(pfMapName);
	      return 1;
	    }
	  else
	    {
	      cout << "PlannerIfc::pfMap : given position is not reachable" << endl;
	      FXMessageBox::error(this,MBOX_OK,"Error Adding Potential fields","Given position is not reachable");
	      return -3;
	    }
	}
      else
	{
	  cout << "PlannerIfc::pfMap : The input map must have been thresholded before adding potential fields" << endl;
	  FXMessageBox::error(this,MBOX_OK,"Error Adding Potential fields","The input map must have been thresholded before adding potential fields");
	  return -1;
	}
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error Adding Potential fields","An image must be already loaded");
      return -2;
    }
}

long PlannerIfc::pfObsMap(FXObject*,FXSelector,void*)
{
  if(map != NULL)
    {
      if(map->getType() == Map::PGM_T)
	{
	  // adding potential field from obstacles
	  Map* map2 = map->applyPFields3();
	  if(map2->save(pfObsMapName.text()) != 0)
	    {
	      FXMessageBox::error(this,MBOX_OK,"Error Computing Potential fields","Error while trying to save the map. Check the output path.");
	      return -3;
	    }
	  this->loadimage(pfObsMapName);
	  return 1;
	}
      else
	{
	  cout << "PlannerIfc::pfMap : The input map must have been thresholded before adding potential fields" << endl;
	  FXMessageBox::error(this,MBOX_OK,"Error Obstacles Adding Potential Field","The input map must have been thresholded before adding potential fields");
	  return -1;
	}
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error Adding Obstacles Potential Field","An image must be already loaded");
      return -2;
    }
}

long PlannerIfc::pfTgtMap(FXObject*,FXSelector,void*)
{
  if(map != NULL)
    {
      if(map->getType() == Map::PGM_T)
	{
	  Pos p = Pos(targetPosX,targetPosY);
	  if(p.isReachable(map))
	    {
	      // adding potential field from target position
	      Map* map1 = map->computeWave(&p);
	      if(map1->save(pfTgtMapName.text()) != 0)
		{
		  FXMessageBox::error(this,MBOX_OK,"Error Computing Potential fields","Error while trying to save the map. Check the output path.");
		  return -4;
		}
	      this->loadimage(pfTgtMapName);
	      return 1;
	    }
	  else
	    {
	      cout << "PlannerIfc::pfTgtMap : given position is not reachable" << endl;
	      FXMessageBox::error(this,MBOX_OK,"Error Adding Target Position Potential Field","Given position is not reachable");
	      return -3;
	    }
	}
      else
	{
	  cout << "PlannerIfc::pfMap : The input map must have been thresholded before adding potential fields" << endl;
	  FXMessageBox::error(this,MBOX_OK,"Error Adding Target Position Potential Field","The input map must have been thresholded before adding potential fields");
	  return -1;
	}
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error Adding Potential fields","An image must be already loaded");
      return -2;
    }
}

long PlannerIfc::pathPlanning(FXObject*,FXSelector,void*)
{
  // create both configuations for init and goal positions
  PMConfig cfg1(initPosX,initPosY);
  PMConfig cfg2(goalPosX,goalPosY);

  if(map == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Planning the Path","A map must be already loaded");
      return -2;
    }

  Map* pfMap = new Map();
  int resLoad = pfMap->load(pfMapPPName.text());

  if(resLoad != 0 || pfMap == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Planning the Path","Error while loading the potential fields map");
      return -3;
    }

  /************** plan the path *************/

  PFPathPlanner pp(map,pfMap);
  int resPP = pp.computePath(cfg1,cfg2);

  /**** display the result of the search ****/
  
  if(resPP == PATH_CPTD)
    {
      // trace the path on the standard map
      cout << "A path has been found." << endl;

      pp.getPath()->save(((char*)resPathPPName.text()));
      //pp.getPath()->traceOn(map);
      map->tracePath(*(pp.getPath()));
      pp.getPath()->display();
      
      // save the result map in the output file
      int resSave = map->save(resMapPPName.text());
      if(resSave != 0)
	{
	  FXMessageBox::error(this,MBOX_OK,"Error Planning the Path","Error while saving the result map");
	  return -4;
	}
      else 
	{
	  this->loadimage(resMapPPName);
	  
	  PointMassSys *sys = new PointMassSys;
	  sys->setVXMax(5);
	  sys->setVYMax(5);
	  sys->setAXMax(1);
	  sys->setAYMax(1);
	  PM_PVDTrajPlanner tp(sys,NULL);
	  int resTP = tp.computeTraj(PMState(cfg1.x(),cfg1.y(),0,0),PMState(cfg2.x(),cfg2.y(),0,0),pp.getPath());
	  
	  //if(resTP == TRAJ_CPTD) tp.getTraj()->display();
	  //else FXMessageBox::error(this,MBOX_OK,"Error Planning the Trajectory","Error while planning the trajectory");

	  return 1;
	}
    }
  else
    {
      cout << "No path has been found." << endl;

      switch(resPP)
	{
	case ERR_CONFIGS : 
	  msgBox->setText("No path computed : Wrong init or goal configurations");
	  break;
	case UNDEF_GPT_TYP : 
	  msgBox->setText("No path computed : Undefined geom path type");
	  break;
	case ERR_PF_MAP : 
	  msgBox->setText("No path computed : Wrong potential field map");
	  break;
	case ERR_STD_MAP : 
	  msgBox->setText("No path computed : Wrong standard map");
	  break;
	case ERR_GOAL_POS : 
	  msgBox->setText("No path computed : The goal configuration is not available or free ");
	  break;
	case ERR_INIT_POS : 
	  msgBox->setText("No path computed : The init configuration is not available or free ");
	  break;
	case NO_PATH : 
	  msgBox->setText("No path computed : No path has been found");
	  break;
	default : 
	  break;
	}

      FXMessageBox::error(this,MBOX_OK,"Path Planning","No path has been found");
      return 0;
    }
}

long PlannerIfc::changeCopyMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Copy map (PGM Image)");
  open.setFilename(copyMapName);
  if(open.execute()){
    copyMapName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changeThresMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Threshold map (PGM Image)");
  open.setFilename(thresMapName);
  if(open.execute()){
    thresMapName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changeIncObsName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Increase obstacles (PGM Image)");
  open.setFilename(incObsName);
  if(open.execute()){
    incObsName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changePfMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Potential fields map (PGM Image)");
  open.setFilename(pfMapName);
  if(open.execute()){
    pfMapName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changePfObsMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Obstacles potential field map (PGM Image)");
  open.setFilename(pfObsMapName);
  if(open.execute()){
    pfObsMapName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changePfTgtMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Target position potential field map (PGM Image)");
  open.setFilename(pfTgtMapName);
  if(open.execute()){
    pfTgtMapName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changePfMapPPName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Potential fields map (PGM Image)");
  open.setFilename(pfMapPPName);
  if(open.execute()){
    pfMapPPName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changeResMapPPName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Result path planning map (PGM Image)");
  open.setFilename(resMapPPName);
  if(open.execute()){
    resMapPPName = open.getFilename();
  }
  return 1;
}

long PlannerIfc::changeResPathPPName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Result path planning path");
  open.setFilename(resPathPPName);
  if(open.execute()){
    resPathPPName = open.getFilename();
  }
  return 1;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Load image
FXbool PlannerIfc::loadimage(const FXString& file){
  FXString ext=FXPath::extension(file);
  FXImage *img=NULL;
  FXImage *old;

  // delete the old map
  if(map != NULL) delete map;
  map = new Map();

  if(comparecase(ext,"pgm")==0){
    img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
  }

  if(img==NULL){
    FXMessageBox::error(this,MBOX_OK,"Error Loading Image","Unsupported type: %s",ext.text());
    return FALSE;
  }

  // Load the map
  map->load(file.text());
  
  FXString msg("A map has been loaded from the file ");
  msg += file;
  msgBox->setText(msg);

  // Load the image
  FXFileStream stream;
  if(stream.open(file,FXStreamLoad)){
    getApp()->beginWaitCursor();
    img->loadPixels(stream);
    stream.close();

    img->create();

    // rescaling the image
    int wtF = imageview->getWidth();
    int htF = imageview->getHeight();

    int wtI = img->getWidth();
    int htI = img->getHeight();

    double ratioW = ((double) wtI) / ((double) wtF);
    double ratioH = ((double) htI) / ((double) htF);

    if(ratioW > ratioH)
      {
	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
      }
    else
      {
	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)(((double)img->getHeight())/ratioH));
      }

    old=imageview->getImage();
    imageview->setImage(img);
    delete old;
    getApp()->endWaitCursor();
    }
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[]){
  
  // Make application
  FXApp application("Interface","TrajectoryPlanning");

  // Open the display
  application.init(argc,argv);
  
  // Make window
  new PlannerIfc(&application);
  
  // Create the application's windows
  application.create();
  
  // Run the application
  return application.run();
}
