#include "TrajGenIfc.hpp"

#include <time.h>
#include <sys/time.h>

#include "gnuplot_i.hpp"

#define THRESHOLD 200
#define INCREASING 20


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// default constructor (only used by subclasses)
TrajGenIfc::TrajGenIfc(){}

// init constructor (creation of the window)
TrajGenIfc::TrajGenIfc(FXApp* a):FXMainWindow(a,"Trajectory Generation",NULL,NULL,DECOR_ALL,0,0,950,650)
{
  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  // the map used
  map = NULL;
  wtI = 0;
  htI = 0;

  // init and goal configurations
  x1 = 0;
  y1 = 0;
  theta1 = 0;
  kp1 = 0;
  v1 = 0.01;
  t1 = 0;
  x2 = 0;
  y2 = 0;
  theta2 = 0;
  kp2 = 0;
  v2 = 5;
  t2 = 25;

  x1_target.connect(x1);
  y1_target.connect(y1);
  theta1_target.connect(theta1);
  kp1_target.connect(kp1);
  v1_target.connect(v1);
  t1_target.connect(t1);
  x2_target.connect(x2);
  y2_target.connect(y2);
  theta2_target.connect(theta2);
  kp2_target.connect(kp2);
  v2_target.connect(v2);
  t2_target.connect(t2);

  rb_Scale = 100.0;
  rb_L = 10.0;
  rb_VMAX = 32.0;
  rb_AMAX = 4.9;
  rb_KMAX = 22.92;
  rb_OMAX = 5.5;

  rb_Scale_target.connect(rb_Scale);
  rb_L_target.connect(rb_L);
  rb_VMAX_target.connect(rb_VMAX);
  rb_AMAX_target.connect(rb_AMAX);
  rb_KMAX_target.connect(rb_KMAX);
  rb_OMAX_target.connect(rb_OMAX);

  debugLevel = 1;
  maxNbStep = 20;
  maxResidual = 0.5;

  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;

  debugLevel_target.connect(debugLevel);
  maxNbStep_target.connect(maxNbStep);
  maxResidual_target.connect(maxResidual);

  nbBranches = 9;
  minAngRS = -105.0;
  maxAngRS = 105.0;
  minDistRS = 80.0;
  maxDistRS = 360.0;
  nbItvDistRS = 6;
  trajByBranch = 5;
  diffAngle = 50.0;

  nbBranches_target.connect(nbBranches);
  minAngRS_target.connect(minAngRS);
  maxAngRS_target.connect(maxAngRS);
  minDistRS_target.connect(minDistRS);
  maxDistRS_target.connect(maxDistRS);
  nbItvDistRS_target.connect(nbItvDistRS);
  trajByBranch_target.connect(trajByBranch);
  diffAngle_target.connect(diffAngle);

  drawUncompleteTraj = false;
  drawTargetPos = false;
  
  initColor = FXRGB(255,0,0);
  goalColor = FXRGB(0,0,255);

  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/

  contents=new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  
  /***************************************************************************************************************
   *********************************************** viewer frame **************************************************
   ***************************************************************************************************************/

  // LEFT pane to contain the canvas
  imageFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the canvas
  new FXLabel(imageFrame,"Viewer",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(imageFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

  // Make image widget
  imageview=new FXImageView(imageFrame,NULL,this,ID_IMAGE,LAYOUT_FILL_X|LAYOUT_FILL_Y);

  msgBox = new FXLabel(imageFrame,"No map loaded. Click on \"Load map\" to load a new map.",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  /***************************************************************************************************************
   ********************************************** commands frame *************************************************
   ***************************************************************************************************************/

  // RIGHT pane for the buttons
  buttonFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the buttons
  new FXLabel(buttonFrame,"Commands",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Tab book with switchable panels
  FXTabBook* panels=new FXTabBook(buttonFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);


  /***************************************************************************************************************
   ***************************************** system configuration tab ********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Sys config");

  FXVerticalFrame *tabSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabSysCfg,"System configuration :");

  FXMatrix *matSysCfg=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSysCfg,"Scale :");
  new FXTextField(matSysCfg,6,&rb_Scale_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCfg,"Robot Length:");
  new FXTextField(matSysCfg,6,&rb_L_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCfg,"Max velocity:");
  new FXTextField(matSysCfg,6,&rb_VMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCfg,"Max acceleration:");
  new FXTextField(matSysCfg,6,&rb_AMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCfg,"Max steering angle:");
  new FXTextField(matSysCfg,6,&rb_KMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCfg,"Max steering speed:");
  new FXTextField(matSysCfg,6,&rb_OMAX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Horizontal divider line
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSysCfg,"Init configuration :");
  FXMatrix *matInitCfg=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitCfg,"X1 :");
  new FXTextField(matInitCfg,12,&x1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"Y1 :");
  new FXTextField(matInitCfg,12,&y1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"Th1 :");
  new FXTextField(matInitCfg,12,&theta1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"Kp1 :");
  new FXTextField(matInitCfg,12,&kp1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"V1 :");
  new FXTextField(matInitCfg,12,&v1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitCfg,"T1 :");
  new FXTextField(matInitCfg,12,&t1_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Horizontal divider line
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSysCfg,"Goal configuration :");
  FXMatrix *matGoalCfg=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matGoalCfg,"X2 :");
  new FXTextField(matGoalCfg,12,&x2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"Y2 :");
  new FXTextField(matGoalCfg,12,&y2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"Th2 :");
  new FXTextField(matGoalCfg,12,&theta2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"Kp2 :");
  new FXTextField(matGoalCfg,12,&kp2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"V2 :");
  new FXTextField(matGoalCfg,12,&v2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matGoalCfg,"T2 :");
  new FXTextField(matGoalCfg,12,&t2_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   *********************************************** options tab ***************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Options");

  FXVerticalFrame *tabOptions=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matOptCst=new FXMatrix(tabOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptCst,"Constraints use :");
  FXCheckButton *velCstBtn = new FXCheckButton(matOptCst,"lin velocity cst",this,ID_VEL_CST);
  velCstBtn->setCheck(true,false);
  FXCheckButton *accCstBtn = new FXCheckButton(matOptCst,"lin acceleration cst",this,ID_ACC_CST);
  accCstBtn->setCheck(true,false);
  FXCheckButton *avelCstBtn = new FXCheckButton(matOptCst,"ang velocity cst",this,ID_AVEL_CST);
  avelCstBtn->setCheck(true,false);
  FXCheckButton *aaccCstBtn = new FXCheckButton(matOptCst,"ang acceleration cst",this,ID_AACC_CST);

  FXMatrix *matOptPrec=new FXMatrix(tabOptions,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptPrec,"Speed / Precision :");
  new FXLabel(matOptPrec,"");
  new FXLabel(matOptPrec,"Max number of steps:");
  new FXTextField(matOptPrec,6,&maxNbStep_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptPrec,"Max residual:");
  new FXTextField(matOptPrec,6,&maxResidual_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  FXMatrix *matOptDbg=new FXMatrix(tabOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptDbg,"Debugging mode :");
  new FXRadioButton(matOptDbg,"0",&debugLevel_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"1",&debugLevel_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"2",&debugLevel_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"3",&debugLevel_target,FXDataTarget::ID_OPTION+3,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);


  /***************************************************************************************************************
   ******************************************** Reachable Star tab ***********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Reachable Star");

  FXVerticalFrame *tabRStar=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabRStar,"RStar Parameters:");

  new FXHorizontalSeparator(tabRStar,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matParamRStar=new FXMatrix(tabRStar,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matParamRStar,"Nb branches:");
  new FXTextField(matParamRStar,6,&nbBranches_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Min angle:");
  new FXTextField(matParamRStar,6,&minAngRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Max angle:");
  new FXTextField(matParamRStar,6,&maxAngRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Min distance:");
  new FXTextField(matParamRStar,6,&minDistRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Max distance:");
  new FXTextField(matParamRStar,6,&maxDistRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Nb itv dist:");
  new FXTextField(matParamRStar,6,&nbItvDistRS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Traj by branch:");
  new FXTextField(matParamRStar,6,&trajByBranch_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matParamRStar,"Diff angle:");
  new FXTextField(matParamRStar,6,&diffAngle_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);  

  FXCheckButton *drawUncompBtn = new FXCheckButton(tabRStar,"draw uncomplete traj.",this,ID_DRAW_UNCOMP);
  FXCheckButton *drawTgtPosBtn = new FXCheckButton(tabRStar,"draw target position",this,ID_DRAW_TGTPOS);

  // Button to compute the reachable star
  new FXButton(tabRStar,"&Draw Reachable Star",NULL,this,ID_DRAW_RSTAR,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to display the reachable star in a gnuplot window
  new FXButton(tabRStar,"&Draw Gnuplot2D Star",NULL,this,ID_DRAW_GP2D,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  /***************************************************************************************************************
   ********************************************* general commands ************************************************
   ***************************************************************************************************************/

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Button to plan the path
  new FXButton(buttonFrame,"&Generate the Trajectory",NULL,this,ID_GEN_TRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load a map from a file
  new FXButton(buttonFrame,"&Load Map from file",NULL,this,ID_LOAD_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Exit button
  new FXButton(buttonFrame,"&Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
}

// destructor
TrajGenIfc::~TrajGenIfc()
{
}

// Create and initialize
void TrajGenIfc::create(){

  // Create the windows
  FXMainWindow::create();

  // Make the main window appear
  show(PLACEMENT_SCREEN);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long TrajGenIfc::loadMapFromFile(FXObject*,FXSelector,void*)
{
  cout << "Load Map from file" << endl;
  FXFileDialog open(this,"Load map (PGM Image)");
  open.setFilename(filename);
  if(open.execute()){
    filename=open.getFilename();
    loadimage(filename);

    msgBox->setText("A map has been Loaded. Click left to select init configuration and click right to select target configuration");
  }
  return 1;
}

long TrajGenIfc::genTraj(FXObject*,FXSelector,void*)
{
  cout << "///////////////////////////////////////////////////////////////////////////// " << endl;
  cout << "///////////////////////////////////////////////////////////////////////////// " << endl;
  cout << "///////////////////////////////////////////////////////////////////////////// " << endl;

  if(map == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","No map has been loaded");
      return -1;
    }

  try 
    {
      // init the timer
      struct timeval tv1,tv2;
      gettimeofday(&tv1,NULL);
      
      carModel = ModelFInt();

      Vec_DP vecConstraints(carModel.dLength);
      Vec_DP vecState_x0(carModel.xLength + 1);
      Vec_DP vecState_xf(carModel.xLength + 1);    	    	        
      
      // init the initial state vector
      vecState_x0[0] = x1;
      vecState_x0[1] = y1;
      vecState_x0[2] = degToRad(theta1);
      vecState_x0[3] = kp1;
      vecState_x0[4] = v1;
      vecState_x0[5] = t1;
      
      // init the final state vector
      vecState_xf[0] = x2;
      vecState_xf[1] = y2;
      vecState_xf[2] = degToRad(theta2);
      vecState_xf[3] = kp2;
      vecState_xf[4] = v2;
      vecState_xf[5] = t2;
      
      // init the constraints vector
      vecConstraints[0] = 0.0;
      vecConstraints[1] = rb_AMAX;
      vecConstraints[2] = rb_VMAX;
      vecConstraints[3] = rb_L;
      vecConstraints[4] = degToRad(rb_KMAX);
      vecConstraints[5] = degToRad(rb_OMAX);
      
      // init options of the trajectory generation
      carModel.setMaxNbStep(maxNbStep);
      carModel.setMaxResidual(maxResidual);
      
      carModel.setUseVelCst(useVelCst);
      carModel.setUseAccCst(useAccCst);
      carModel.setUseAVelCst(useAVelCst);
      carModel.setUseAAccCst(useAAccCst);
      
      carModel.setDBG(debugLevel);
      
      //Compute trajectory
      int res = carModel.ComputeTrajectory(vecState_x0,vecState_xf,vecConstraints,NULL);
      
      // stop the timer
      gettimeofday(&tv2,NULL);
      
      // compute the time passed
      int elapsedTs = tv2.tv_sec - tv1.tv_sec;
      int elapsedTus = tv2.tv_usec - tv1.tv_usec;
      double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
      cout << "Elapsed time : " << elapsedT << endl;

      if(res == 0)
	{
	  // case where the trajectory generation succeed
	  
	  // trace the trajectory on the map
	  carModel.traceOn(map);

 	  // saving the map
 	  map->save("/tmp/ResultMap_traj.pgm");

	  // load the map in the canvas
 	  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));
	  
	  FXString str("Trajectory generation succeed in ");
	  str += FXStringVal(elapsedT,2,FALSE);
	  str += FXString(" sec after ");
	  str += FXStringVal((FXdouble)carModel.finalNbSteps_,0,FALSE);
	  str += FXString(" steps.");
	  msgBox->setText(str);
	}
      else if(res == -1)
	{
	  // case where the trajectory generation didn't reach the goal : 
	  // trace the better trajectory found during the generation

	  // load the better parameters found during generation
	  // ...

	  // trace the trajectory on the map
	  carModel.traceOn(map);

 	  // saving the map
 	  map->save("/tmp/ResultMap_traj.pgm");

	  // load the map in the canvas
 	  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));
	  
	  FXString str("The goal has not been reach after ");
	  str += FXStringVal((FXdouble)ModelFInt::getMaxNbStep(),0,FALSE);
	  str += FXString(" steps. Elapsed time : ");
	  str += FXStringVal(elapsedT,2,FALSE);
	  str += FXString(" sec.");
	  msgBox->setText(str);
	}
      else // (res == -2)
	{
 	  FXMessageBox::error(this,MBOX_OK,"Error during Trajectory Generation","Trajectory generation failed.");
	  FXString str("Trajectory generation failed after ");
	  str += FXStringVal((FXdouble)carModel.finalNbSteps_,0,FALSE);
	  str += FXString(" steps.");
	  msgBox->setText(str);
	}
    }
  catch(exception& e)
    {
      cout << e.what() << "\n";
      return 1;
    } 
  return 0;
}

// long TrajGenIfc::planPath(FXObject*,FXSelector,void*)
// {
//   if(map == NULL)
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","No map has been loaded");
//       return -1;
//     }

//   struct timeval t1,t2;
//   gettimeofday(&t1,NULL);

//   // save the old map
//   Map* oldMap = new Map(*map);

//   Map* mapPF = NULL;
  
//   msgBox->setText("Treatment of the input map for path planning.");

//   // create threshoded map and increased map
//   map->thresholdPGM(THRESHOLD);
//   map->increaseObs(INCREASING);

//   mapThInc = new Map(*map);
//   //mapThInc->save("/tmp/mapThInc.pgm");
  
//   // loading robot's configuration
//   Robot* rb = new Robot(rb_L, rb_WF, rb_WR, rb_R, rb_KMAX, rb_OMAX, rb_Scale);

//   Point2D p1 = Point2D(x1,y1);
//   Point2D p2 = Point2D(x2,y2);
  
//   Angle Phi0A1 = Angle(theta1,Angle::DEG);
//   Angle Phi0A2 = Phi0A1.deg2rad();
//   Vector2D Phi0V = Phi0A2.getVect();
//   Angle Phi1A1 = Angle(theta2,Angle::DEG);
//   Angle Phi1A2 = Phi1A1.deg2rad();
//   Vector2D Phi1V = Phi1A2.getVect();
  
//   ConfigR beginCfg = ConfigR(rb,p1,Phi0V.x,Phi0V.y,0);
//   ConfigR targetCfg = ConfigR(rb,p2,Phi1V.x,Phi1V.y,0);
  
//   // getting back beginning and target positions
//   Position beginPos,targetPos;
//   beginPos = beginCfg.getPos();
//   targetPos = targetCfg.getPos();
  
//   if(!(beginPos.isReachable(map)))
//     {
//       // reinit the old map
//       delete map;
//       map = oldMap;

//       // display an error
//       FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","The init position is not reachable");
//       cout << "(" << beginPos.getX() << "," << beginPos.getY() << ")" << endl;

//       return -2;
//     }

//   if(!(targetPos.isReachable(map)))
//     {
//       // reinit the old map
//       delete map;
//       map = oldMap;

//       // display an error
//       FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","The goal position is not reachable");
//       cout << "(" << targetPos.getX() << "," << targetPos.getY() << ")" << endl;

//       return -2;
//     }

//   // computing pppf path between beginning position and target position
//   Path path;
  
//   msgBox->setText("Searching an holonomic path between the init and target positions.");
  
//   path.buildPathPPPF(map,&beginPos,&targetPos,1,7,mapPF);

//   // treatment of search result
//   switch(path.getType())
//     {
//     case Path::UNKWN : 
//       // reinit the old map
//       delete map;
//       map = oldMap;

//       // display an error
//       FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","Unknown path type");

//       return 3;
//       break;

//     case Path::CPLT : 
//       {
// 	// displaying positions of the path
// 	//path.displayPath();
	
// 	Map* map2 = new Map(*oldMap);

// 	path.smoothPath();

// 	// tracing the path on the map
// 	path.traceOn(map2);
	
// 	// saving the map
// 	//map2->save("$DEFORMATION_DIR/map/ResultMap_path.pgm");
// 	map2->save("/tmp/ResultMap_path.pgm");

// 	// load the new map
// 	//this->loadimage(FXString("map/ResultMap_path.pgm"));
// 	this->loadimage(FXString("/tmp/ResultMap_path.pgm"));

// 	// delete map2
// 	delete map2;

// 	break;
//       }

//     case Path::UNCPLT : 
//       // reinit the old map
//       delete map;
//       map = oldMap;
      
//       // display an error
//       FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","No path has been found to reach the target position");
//       msgBox->setText("No path has been found to reach the goal position.");

//       return 3;
//       break;

//     default :
//       // reinit the old map
//       delete map;
//       map = oldMap;

//       // display an error
//       FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","Unknown path type");

//       return 3;
//       break;
//     }

//   // building a trajectory from the path computed
//   Trajectory traj = Trajectory(rb);

//   msgBox->setText("Planning a non honolomic path between the init and target configurations.");

//   traj.buildNHTrajectory2(&path,&beginCfg,&targetCfg,mapThInc,mapPF);

//   cout << "J0" << endl;
  
//   // draw the trajectory if it's feasable
//       if(traj.getType() == Trajectory::FEA)
// 	{
// 	  Map* map3 = new Map(*oldMap);

// 	  cout << "J1" << endl;

// 	  Path pathNH;
// 	  traj.getPath(pathNH);

// 	  cout << "J2" << endl;
	  
// 	  // displaying positions of the path
// 	  pathNH.displayPath();

// 	  cout << "J3" << endl;

// 	  // tracing the path on the map
// 	  pathNH.traceOn(map3);

// 	  cout << "J4" << endl;

// 	  // saving the map
// 	  //map3->save("$DEFORMATION_DIR/map/ResultMap_traj.pgm");
// 	  map3->save("/tmp/ResultMap_traj.pgm");
	  
// 	  cout << "J5" << endl;

// 	  vector<vector<double> > tabInfo1,tabInfo2;
// 	  //traj.writeTrajGNUPfile("$DEFORMATION_DIR/NHPathBIN.nhp",tabInfo1,1,1);
// 	  //traj.writeTrajGNUPfile("$DEFORMATION_DIR/NHPath.nhp",tabInfo1,1,3);
// 	  //traj.writePathGNUPfile("$DEFORMATION_DIR/path4.dat",tabInfo2,15);
// 	  //traj.writeTrajGNUPfile("/tmp/NHPathBIN.nhp",tabInfo1,1,1);
// 	  //traj.writeTrajGNUPfile("/tmp/NHPath.nhp",tabInfo1,1,3);
// 	  //traj.writePathGNUPfile("/tmp/path4.dat",tabInfo2,15);

// 	  cout << "J6" << endl;

// 	  // Load the new map
// 	  //THIS->LOADIMAGE(FXSTRING("$DEFORMATION_DIR/map/ResultMap_traj.pgm"));
// 	  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));

// 	  cout << "J7" << endl;

// 	  delete map3;
// 	}
//       else
// 	{
// 	  // display an error
// 	  FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","No feasable trajectory has been found");
// 	  msgBox->setText("No feasable trajectory has been found.");

// 	  return 4;
// 	}

//       gettimeofday(&t2,NULL);
      
//       int elapsedTs = t2.tv_sec - t1.tv_sec;
//       int elapsedTus = t2.tv_usec - t1.tv_usec;

//       double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
//       cout << "Elapsed time : " << elapsedT << endl;

//       FXString str("A feasable trajectory has been found in ");
//       str += FXStringVal(elapsedT,2,FALSE);
//       str += FXString(" sec.");
//       msgBox->setText(str);

//       return 1;
// }

long TrajGenIfc::onLeftMouseDown(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // grab the image window
  imageview->grab();

  // the left button is pressed
  mlbFlag = true;

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init position on the map
  dc.setForeground(initColor);
  dc.drawEllipse(ev->win_x - 2, ev->win_y - 2, 4, 4);

  // set the coordinates of init position
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  x1 = xM;
  y1 = yM;
  return 1;
}

long TrajGenIfc::onLeftMouseUp(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // ungrab the image window
  imageview->ungrab();

  // the left button is unpressed
  mlbFlag = false;
  
  // compute the init direction
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  // compute the vector between init position and the new position
  int dx = (int)(xM - x1);
  int dy = (int)(yM - y1);
  
  Vec2D vect(dx,dy);
  Vec vectN = vect.normalized();
  Vec2D vectN20(vectN.getVal(0)*20.0f,vectN.getVal(1)*20.0f);

  int xTgtMap = (int)(x1 + vectN20.getVal(0));
  int yTgtMap = (int)(y1 + vectN20.getVal(1));

  // compute the direction from the vector
  double ang = atan2(dy,dx);
  theta1 = ang / M_PI * 180.0f;

  // draw the direction on the map
  int x1Frame=0,y1Frame=0;
  coordMapToCoordFrame((int)x1, (int)y1, x1Frame, y1Frame);

  int xTgtFrame=0,yTgtFrame=0;
  coordMapToCoordFrame(xTgtMap, yTgtMap, xTgtFrame, yTgtFrame);

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init direction on the map
  dc.setForeground(initColor);
  dc.drawLine(x1Frame, y1Frame, xTgtFrame, yTgtFrame);

//   this->loadimage(filename);

  return 1;
}

long TrajGenIfc::onRightMouseDown(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // grab the image window
  imageview->grab();

  // the right button is pressed
  mrbFlag = true;

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init position on the map
  dc.setForeground(goalColor);
  dc.drawEllipse(ev->win_x - 2, ev->win_y - 2, 4, 4);

  // set the coordinates of init position
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  x2 = xM;
  y2 = yM;
  return 1;
}

long TrajGenIfc::onRightMouseUp(FXObject*,FXSelector,void* ptr)
{
  // get back the event of the signal
  FXEvent *ev=(FXEvent*)ptr;

  // ungrab the image window
  imageview->ungrab();

  // the right button is unpressed
  mrbFlag = false;

  // compute the goal direction
  int xM=0,yM=0;
  this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

  // compute the vector between init position and the new position
  int dx = (int)(xM - x2);
  int dy = (int)(yM - y2);
  
  Vec2D vect(dx,dy);
  Vec vectN = vect.normalized();
  Vec2D vectN20(vectN.getVal(0)*20.0f,vectN.getVal(1)*20.0f);

  int xTgtMap = (int)(x2 + vectN20.getVal(0));
  int yTgtMap = (int)(y2 + vectN20.getVal(1));

  // compute the direction from the vector
  double ang = atan2(dy,dx);
  theta2 = ang / M_PI * 180.0f;

  // draw the direction on the map
  int x2Frame=0,y2Frame=0;
  coordMapToCoordFrame((int)x2, (int)y2, x2Frame, y2Frame);

  int xTgtFrame=0,yTgtFrame=0;
  coordMapToCoordFrame(xTgtMap, yTgtMap, xTgtFrame, yTgtFrame);

  // create the drawing window
  FXDCWindow dc(imageview);

  // draw the init direction on the map
  dc.setForeground(goalColor);
  dc.drawLine(x2Frame, y2Frame, xTgtFrame, yTgtFrame);

  return 1;
}

long TrajGenIfc::onMouseMove(FXObject*,FXSelector,void*)
{
  return 1;
}

long TrajGenIfc::displayBProfiles(FXObject*,FXSelector,void*)
{
}

long TrajGenIfc::displayWBProfiles(FXObject*,FXSelector,void*)
{
}

long TrajGenIfc::displayAllProfiles(FXObject*,FXSelector,void*)
{
}

long TrajGenIfc::updateUseVelCst(FXObject*,FXSelector,void*)
{
  // update the use of linear velocity constraints
  if(useVelCst == false)
    useVelCst = true;
  else
    useVelCst = false;
  return 1;
}

long TrajGenIfc::updateUseAccCst(FXObject*,FXSelector,void*)
{
  // update the use of linear acceleration constraints
  if(useAccCst == false)
    useAccCst = true;
  else
    useAccCst = false;
  return 1;
}

long TrajGenIfc::updateUseAVelCst(FXObject*,FXSelector,void*)
{
  // update the use of angular velocity constraints
  if(useAVelCst == false)
    useAVelCst = true;
  else
    useAVelCst = false;
  return 1;
}

long TrajGenIfc::updateUseAAccCst(FXObject*,FXSelector,void*)
{
  // update the use of angular acceleration constraints
  if(useAAccCst == false)
    useAAccCst = true;
  else
    useAAccCst = false;
  return 1;
}

long TrajGenIfc::drawRStar(FXObject*,FXSelector,void*)
{
  if(map == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while Planning the Path","No map has been loaded");
      return -1;
    }

  // init the timer
  struct timeval tv1,tv2;
  gettimeofday(&tv1,NULL);

  if(debugLevel >= 1)
    {
      tabPosTraj.clear();
      tabPosReachedX.clear();
      tabPosReachedY.clear();
      tabPosReachedT.clear();
      tabPosReachedV.clear();
      tabPosReachedSt.clear();
    }

  double totFinalNbSteps = 0;

  for(int i=0; i<nbBranches; i++)
    {
      vector<double>* subTabPosTrajX;
      vector<double>* subTabPosTrajY;
      vector<double>* subTabPosTrajT;
      vector<double>* subTabPosTrajV;

      if(debugLevel >= 1)
	{
	  subTabPosTrajX = new vector<double>;
	  subTabPosTrajY = new vector<double>;
	  subTabPosTrajT = new vector<double>;
	  subTabPosTrajV = new vector<double>;	  
	}

      for(int j=0; j<trajByBranch; j++)
	{
	  for(int k=0; k<nbItvDistRS; k++)
	    {
	      // create the new trajectory generator
	      carModel = ModelFInt();

	      Vec_DP vecConstraints(carModel.dLength);
	      Vec_DP vecState_x0(carModel.xLength + 1);
	      Vec_DP vecState_xf(carModel.xLength + 1);    	    	        
	      
	      // init the initial state vector
	      vecState_x0[0] = x1;
	      vecState_x0[1] = y1;
	      vecState_x0[2] = degToRad(theta1);
	      vecState_x0[3] = kp1;
	      vecState_x0[4] = v1;
	      vecState_x0[5] = t1;
	      
	      // init the final state vector
	      vecState_xf[0] = x1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * cos(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
	      vecState_xf[1] = y1 + (minDistRS + (((double)k)/(nbItvDistRS-1))*(maxDistRS-minDistRS)) * sin(degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS)));
	      if(trajByBranch == 1) 
		vecState_xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS));
	      else 
		vecState_xf[2] = degToRad(theta1 + minAngRS + (((double)i)/(nbBranches-1))*(maxAngRS-minAngRS) 
					  - diffAngle + (((double)j)/(trajByBranch-1)*2.0*diffAngle));

	      vecState_xf[3] = kp2;
	      vecState_xf[4] = v2;
	      vecState_xf[5] = t2;

	      if(j == 0) 
		{
		  tabPosReachedX.push_back(vecState_xf[0]);
		  tabPosReachedY.push_back(vecState_xf[1]);
		  tabPosReachedV.push_back(vecState_xf[4]);
		  tabPosReachedSt.push_back(vecState_xf[3]);
		  tabPosReachedT.push_back(vecState_xf[5]);
		}
	      
	      // init the constraints vector
	      vecConstraints[0] = 0.0;
	      vecConstraints[1] = rb_AMAX;
	      vecConstraints[2] = rb_VMAX;
	      vecConstraints[3] = rb_L;
	      vecConstraints[4] = degToRad(rb_KMAX);
	      vecConstraints[5] = degToRad(rb_OMAX);

	      // init options of the trajectory generation
	      carModel.setMaxNbStep(maxNbStep);
	      carModel.setMaxResidual(maxResidual);
	      
	      carModel.setUseVelCst(useVelCst);
	      carModel.setUseAccCst(useAccCst);
	      carModel.setUseAVelCst(useAVelCst);
	      carModel.setUseAAccCst(useAAccCst);
	      
	      carModel.setDBG(debugLevel);
	      
	      //Compute trajectory
	      int res = carModel.ComputeTrajectory(vecState_x0,vecState_xf,vecConstraints,NULL);

	      totFinalNbSteps += carModel.finalNbSteps_;
	      
	      if(res == 0)
		{
		  // case where the trajectory generation succeed
		  
		  if(debugLevel >= 1)
		    {
		      // trace the trajectory on the map
		      carModel.traceOn(map);

		      // compute a sample trajectory to display in a gnuplot viewer
		      carModel.SaveTrajectory();

		      for(int j=0; j<carModel.posX->size(); j++) subTabPosTrajX->push_back((*(carModel.posX))[j]);
		      for(int j=0; j<carModel.posX->size(); j++) subTabPosTrajX->push_back((*(carModel.posX))[carModel.posX->size() - j -1]);
		      for(int j=0; j<carModel.posY->size(); j++) subTabPosTrajY->push_back((*(carModel.posY))[j]);
		      for(int j=0; j<carModel.posY->size(); j++) subTabPosTrajY->push_back((*(carModel.posY))[carModel.posY->size() - j -1]);
		      for(int j=0; j<carModel.timeT->size(); j++) subTabPosTrajT->push_back((*(carModel.timeT))[j]);
		      for(int j=0; j<carModel.timeT->size(); j++) subTabPosTrajT->push_back((*(carModel.timeT))[carModel.timeT->size() - j -1]);
		      for(int j=0; j<carModel.velV->size(); j++) subTabPosTrajV->push_back((*(carModel.velV))[j]);
		      for(int j=0; j<carModel.velV->size(); j++) subTabPosTrajV->push_back((*(carModel.velV))[carModel.velV->size() - j -1]);
		    }
		}
	      else if(res == -1)
		{
		  // case where the trajectory generation didn't reach the goal : 
		  // trace the better trajectory found during the generation

		  // trace the trajectory on the map
		  if(drawUncompleteTraj) 
		    {
		      if(debugLevel >= 1)
			{
			  carModel.traceOn(map);

			  carModel.SaveTrajectory();
			  for(int j=0; j<carModel.posX->size(); j++) subTabPosTrajX->push_back((*(carModel.posX))[j]);
			  for(int j=0; j<carModel.posX->size(); j++) subTabPosTrajX->push_back((*(carModel.posX))[carModel.posX->size() - j -1]);
			  for(int j=0; j<carModel.posY->size(); j++) subTabPosTrajY->push_back((*(carModel.posY))[j]);
			  for(int j=0; j<carModel.posY->size(); j++) subTabPosTrajY->push_back((*(carModel.posY))[carModel.posY->size() - j -1]);
			  for(int j=0; j<carModel.timeT->size(); j++) subTabPosTrajT->push_back((*(carModel.timeT))[j]);
			  for(int j=0; j<carModel.timeT->size(); j++) subTabPosTrajT->push_back((*(carModel.timeT))[carModel.timeT->size() - j -1]);
			  for(int j=0; j<carModel.velV->size(); j++) subTabPosTrajV->push_back((*(carModel.velV))[j]);
			  for(int j=0; j<carModel.velV->size(); j++) subTabPosTrajV->push_back((*(carModel.velV))[carModel.velV->size() - j -1]);
			}
		    }
		}
	      else // (res == -2)
		{
		  // nothing to do
		}
	    }
	}
      if(debugLevel >= 1)
	{
	  tabPosTraj.push_back(subTabPosTrajX);
	  tabPosTraj.push_back(subTabPosTrajY);
	  tabPosTraj.push_back(subTabPosTrajT);
	  tabPosTraj.push_back(subTabPosTrajV);
	}
    }
  
  // stop the timer
  gettimeofday(&tv2,NULL);
  
  // compute the time passed
  int elapsedTs = tv2.tv_sec - tv1.tv_sec;
  int elapsedTus = tv2.tv_usec - tv1.tv_usec;
  double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  cout << "Elapsed time : " << elapsedT << endl; 

  // compute the average number of steps
  double nbTraj = (nbBranches*trajByBranch*nbItvDistRS);
  double avgNbSteps = totFinalNbSteps / nbTraj;

  // compute the average time needed by trajectory
  double avgTime = elapsedT / nbTraj;
  double avgTPS = 1.0 / avgTime;

  // saving the map
  map->save("/tmp/ResultMap_traj.pgm");
  
  // load the map in the canvas
  this->loadimage(FXString("/tmp/ResultMap_traj.pgm"));
  
  FXString str("Reachable Star computed in ");
  str += FXStringVal(elapsedT,2,FALSE);
  str += FXString(" sec.  Avg nb steps : ");
  str += FXStringVal(avgNbSteps,2,FALSE);
  str += FXString("  Avg time by traj : ");
  str += FXStringVal(avgTime,4,FALSE);
  str += FXString("  Avg Traj/sec : ");
  str += FXStringVal(avgTPS,2,FALSE);
  msgBox->setText(str);

  return 1;
}

long TrajGenIfc::updateDrawUncompleteTraj(FXObject*,FXSelector,void*)
{
  // update the drawing of uncomplete trajectories
  if(drawUncompleteTraj == false)
    drawUncompleteTraj = true;
  else
    drawUncompleteTraj = false;
  return 1;
}

long TrajGenIfc::updateDrawTargetPos(FXObject*,FXSelector,void*)
{
  // update the drawing of uncomplete trajectories
  if(drawTargetPos == false)
    drawTargetPos = true;
  else
    drawTargetPos = false;
  return 1;  
}

long TrajGenIfc::drawGnuplot2D(FXObject*,FXSelector,void*)
{
  Gnuplot gPosition;
  gPosition.reset_plot();

  if(drawTargetPos) gPosition.set_style("points").plot_xy(tabPosReachedX,tabPosReachedY,"");
  
  for(int i=0; i<tabPosTraj.size(); i+=4)
    {      
      if(tabPosTraj[i] != NULL && tabPosTraj[i+1] != NULL && tabPosTraj[i]->size() != 0 && tabPosTraj[i+1]->size() != 0)
 	{
	  gPosition.cmd("set linestyle 1");
	  //gPosition.cmd("set linestyle 1 lt 1 lw 6");
	  gPosition.set_style("lines").plot_xy(*(tabPosTraj[i]),*(tabPosTraj[i+1]),"");
 	}
    }

  Gnuplot gPosition2;
  gPosition2.reset_plot();
  
  if(drawTargetPos) gPosition2.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedT,"");
  
  for(int i=0; i<tabPosTraj.size(); i+=4)
    {      
      if(tabPosTraj[i] != NULL && tabPosTraj[i+1] != NULL && tabPosTraj[i]->size() != 0 && tabPosTraj[i+1]->size() != 0)
 	{
	  gPosition2.cmd("set linestyle 1");
	  //gPosition.cmd("set linestyle 1 lt 1 lw 6");
	  gPosition2.set_style("lines").plot_xyz(*(tabPosTraj[i]),*(tabPosTraj[i+1]),*(tabPosTraj[i+2]),"");
 	}
    }

  Gnuplot gPosition3;
  gPosition3.reset_plot();
  
  if(drawTargetPos) gPosition3.set_style("points").plot_xyz(tabPosReachedX,tabPosReachedY,tabPosReachedV,"");
  
  for(int i=0; i<tabPosTraj.size(); i+=4)
    {      
      if(tabPosTraj[i] != NULL && tabPosTraj[i+1] != NULL && tabPosTraj[i]->size() != 0 && tabPosTraj[i+1]->size() != 0)
 	{
	  gPosition3.cmd("set linestyle 1");
	  //gPosition.cmd("set linestyle 1 lt 1 lw 6");
	  gPosition3.set_style("lines").plot_xyz(*(tabPosTraj[i]),*(tabPosTraj[i+1]),*(tabPosTraj[i+3]),"");
 	}
    }
  
  wait_for_key();
  return 1;
}

long TrajGenIfc::drawGnuplot3D(FXObject*,FXSelector,void*)
{

  
  wait_for_key();
  return 1;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void TrajGenIfc::drawInitPos()
{
  // verify that the init position is available
  if(map != NULL && x1 >= 0 && x1 < map->getWidth() && x1 >= 0 && x1 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int xF=0, yF=0;
      coordMapToCoordFrame((int)x1, (int)y1, xF, yF);
      
      // draw the init position on the map
      dc.setForeground(initColor);
      dc.drawEllipse(xF - 2, yF - 2, 4, 4);

      cout << "init position : (" << xF << "," << yF << ")" << endl;
    }
  else
    {
      cout << "Can't draw init position" << endl;
    }
}

void TrajGenIfc::drawInitDir()
{
  // verify that the init position is available
  if(map != NULL && x1 >= 0 && x1 < map->getWidth() && x1 >= 0 && x1 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int x1F=0, y1F=0;
      coordMapToCoordFrame((int)x1, (int)y1, x1F, y1F);
      
      int xTgtM = (int)(x1 + 20.0f * cos(theta1));
      int yTgtM = (int)(y1 + 20.0f * sin(theta1));
      
      // compute the vector of the init position direction in the image frame
      int xTgtF=0, yTgtF=0;
      coordMapToCoordFrame(xTgtM, yTgtM, xTgtF, yTgtF);
      
      // draw the init direction on the map
      dc.setForeground(initColor);
      dc.drawLine(x1F, y1F, xTgtF, yTgtF);

      cout << "init direction : (" << x1F << "," << y1F << ") -> (" << xTgtF << "," << yTgtF << ")" << endl;
    }
  else
    {
      cout << "Can't draw init direction" << endl;
    }
}

void TrajGenIfc::drawInitCfg()
{
  drawInitPos();
  drawInitDir();
}

void TrajGenIfc::drawGoalPos()
{
  // verify that the goal position is available
  if(map != NULL && x2 >= 0 && x2 < map->getWidth() && x2 >= 0 && x2 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int xF=0, yF=0;
      coordMapToCoordFrame((int)x2, (int)y2, xF, yF);
      
      // draw the init position on the map
      dc.setForeground(goalColor);
      dc.drawEllipse(xF - 2, yF - 2, 4, 4);

      cout << "goal position : (" << xF << "," << yF << ")" << endl;
    }
  else
    {
      cout << "Can't draw goal position" << endl;
    }
}

void TrajGenIfc::drawGoalDir()
{
  // verify that the goal position is available
  if(map != NULL && x2 >= 0 && x2 < map->getWidth() && x2 >= 0 && x2 < map->getHeight())
    {
      // create the drawing window
      FXDCWindow dc(imageview);
      
      // compute the init position coord in the image frame
      int x2F=0, y2F=0;
      coordMapToCoordFrame((int)x2, (int)y2, x2F, y2F);
      
      int xTgtM = (int)(x2 + 20.0f * cos(theta2));
      int yTgtM = (int)(y2 + 20.0f * sin(theta2));
      
      // compute the vector of the init position direction in the image frame
      int xTgtF=0, yTgtF=0;
      coordMapToCoordFrame(xTgtM, yTgtM, xTgtF, yTgtF);
      
      // draw the init direction on the map
      dc.setForeground(goalColor);
      dc.drawLine(x2F, y2F, xTgtF, yTgtF);

      cout << "goal direction : (" << x2F << "," << y2F << ") -> (" << xTgtF << "," << yTgtF << ")" << endl;
    }
  else
    {
      cout << "Can't draw goal direction" << endl;
    }
}

void TrajGenIfc::drawGoalCfg()
{
  drawGoalPos();
  drawGoalDir();
}

// Load image
FXbool TrajGenIfc::loadimage(const FXString& file){
  FXString ext=FXPath::extension(file);
  FXImage *img=NULL;
  FXImage *old;

  // delete the old map
  if(map != NULL) delete map;
  map = new Map();

  if(comparecase(ext,"pgm")==0){
    img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
  }

  if(img==NULL){
    FXMessageBox::error(this,MBOX_OK,"Error Loading Image","Unsupported type: %s",ext.text());
    return FALSE;
  }

  // Load the map
  map->load(file.text());

  // Load the image
  FXFileStream stream;
  if(stream.open(file,FXStreamLoad)){
    getApp()->beginWaitCursor();
    img->loadPixels(stream);
    stream.close();

    img->create();

    // rescaling the image
    int wtF = imageview->getWidth();
    int htF = imageview->getHeight();

    wtI = img->getWidth();
    htI = img->getHeight();

    double ratioW = ((double) wtI) / ((double) wtF);
    double ratioH = ((double) htI) / ((double) htF);

    if(ratioW > ratioH)
      {
	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
      }
    else
      {
	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)(((double)img->getHeight())/ratioH));
      }

    old=imageview->getImage();
    imageview->setImage(img);
    delete old;
    getApp()->endWaitCursor();

    // display both configuration if they are defined
    drawInitCfg();
    drawGoalCfg();
    }
  return TRUE;
}

// convert frame coordinates to map coordinates
void TrajGenIfc::coordFrameToCoordMap(const int& xF, const int& yF, int& xM, int& yM)
{
  // get back the dimensions of the frame
  int wtF = imageview->getWidth();
  int htF = imageview->getHeight();

  // compute the ratio between frame and map
  double ratioW = ((double) wtI) / ((double) wtF);
  double ratioH = ((double) htI) / ((double) htF);

  if(ratioW > ratioH)
    {
      // compute the shift on y
      int shift = (int)((htF - ((double)htI)/ratioW)/2.0f);

      // compute the coordinates in the map
      xM = (int)(xF * ratioW);
      yM = htI -1 - (int)((yF-shift) * ratioW);
      return;
    }
  else
    {
      // compute the shift on x
      int shift = (int)((wtF - ((double)wtI)/ratioH)/2.0f);

      // compute the coordinates in the map
      xM = (int)((xF-shift) * ratioH);
      yM = htI -1 - (int)(yF * ratioH);
      return;
    }
}

// convert map coordinates to frame coordinates
void TrajGenIfc::coordMapToCoordFrame(const int& xM, const int& yM, int& xF, int& yF)
{
  // get back the dimensions of the frame
  int wtF = imageview->getWidth();
  int htF = imageview->getHeight();

  // compute the ratio between frame and map
  double ratioW = ((double) wtI) / ((double) wtF);
  double ratioH = ((double) htI) / ((double) htF);

  if(ratioW > ratioH)
    {
      // compute the shift on y
      int shift = (int)((htF - ((double)htI)/ratioW)/2.0f);

      // compute the coordinates in the frame
      xF = 1 + (int)(xM / ratioW);
      yF = 1 + (int)(((double)(htI -1 - yM) / ratioW) + shift);
      return;
    }
  else
    {
      // compute the shift on x
      int shift = (int)((wtF - ((double)wtI)/ratioH)/2.0f);

      // compute the coordinates in the frame
      xF = 1 + (int)(xM / ratioH) + shift;
      yF = 1 + (int)(((double)(htI -1 - yM)) / ratioH);
      return;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[]){
  
  // Make application
  FXApp application("Interface","TrajectoryGeneration");

  // Open the display
  application.init(argc,argv);
  
  // Make window
  new TrajGenIfc(&application);
  
  // Create the application's windows
  application.create();
  
  // Run the application
  return application.run();
}
