#include "PathPlanner.hpp"
#include "PFPathPlanner.hpp"
#include "Map.hpp"
#include "Config.hpp"

#include <time.h>
#include <sys/time.h>
#include <iostream>

int main(int argc,char** argv)
{
    if(argc != 3 && argc != 4 && argc != 7 && argc != 8)
    {
        cout << argv[0] << "<input_stdMap.pgm> [<input_pfMap.pgm>] <output_map.pgm> [<X1> <Y1> <X2> <Y2>]" << endl;
        return 0;
    }
    else
    {
      /************** Maps loading **************/

      // Reads input standard map
      Map* map = new Map();
      cout << "Std Map loading... ";
      map->load(argv[1]);
      cout << "done" << endl;
      
      Map* pfm = NULL;
      if(argc == 4 || argc == 8)
	{
	  // Reads input potential field map
	  pfm = new Map();
	  cout << "PF Map loading... ";
	  pfm->load(argv[2]);
	  cout << "done" << endl;	
	}
      
      /********* Configurations loading *********/

      int x1,x2,y1,y2;
      if(argc == 3 || argc == 4)
	{
	  // Ask configurations to the user
	  cout << "init position : " << endl;
	  cout << "\t X : ";
	  cin >> x1;
	  cout << "\t Y : ";
	  cin >> y1;
	  cout << "goal position : " << endl;
	  cout << "\t X : ";
	  cin >> x2;
	  cout << "\t Y : ";
	  cin >> y2;
	}
      else
	{
	  // Create configurations from last parameters
	  if(argc == 7)
	    {
	      x1 = atoi(argv[3]);
	      y1 = atoi(argv[4]);
	      x2 = atoi(argv[5]);
	      y2 = atoi(argv[6]);
	    }
	  else
	    {
	      x1 = atoi(argv[4]);
	      y1 = atoi(argv[5]);
	      x2 = atoi(argv[6]);
	      y2 = atoi(argv[7]);
	    }
	}
            
      cout << "Path planning..." << endl;
      ConfigR2D cfg1(x1,y1);
      ConfigR2D cfg2(x2,y2);
      
//       cout << "map->getWidth() : " << map->getWidth() << endl;
//       cout << "map->getHeight() : " << map->getHeight() << endl;
//       cout << "pfm->getWidth() : " << pfm->getWidth() << endl;
//       cout << "pfm->getHeight() : " << pfm->getHeight() << endl;

//       cout << "cfg1 : "; cfg1.display(2); cout << endl;
//       cout << "cfg2 : "; cfg2.display(2); cout << endl;

//       if(cfg1.isAvailableOn(pfm))
// 	{
// 	  cout << "cfg1 is available on map" << endl;
// 	}
//       else
// 	{
// 	  cout << "cfg1 is not available on map" << endl;
// 	}
//       if(cfg2.isAvailableOn(pfm))
// 	{
// 	  cout << "cfg2 is available on map" << endl;
// 	}
//       else
// 	{
// 	  cout << "cfg2 is not available on map" << endl;
// 	}
      
//       if(cfg1.isFreeOn(pfm))
// 	{
// 	  cout << "cfg1 is free on map" << endl;
// 	}
//       else
// 	{
// 	  cout << "cfg1 is not free on map" << endl;
// 	}
//       if(cfg2.isFreeOn(pfm))
// 	{
// 	  cout << "cfg2 is free on map" << endl;
// 	}
//       else
// 	{
// 	  cout << "cfg2 is not free on map" << endl;
// 	}

      /********* initialize chronometer *********/

      struct timeval t1,t2;
      gettimeofday(&t1,NULL);
      
      /************** plan the path *************/

      PFPathPlanner pp(map,pfm,cfg1,cfg2);
      bool trouve = pp.searchPathV2();
      
      /************ stop chronometer ************/

      gettimeofday(&t2,NULL);

      /**** display the result of the search ****/
            
      if(trouve)
	{
	  // trace the path on the standard map
	  cout << argv[0] << " : A path has been found." << endl;
	  pp.getPath()->traceOn(map);

	  // save the result map in the output file
	  if(argc == 3 || argc == 7)
	    map->save(argv[2]);
	  else
	    map->save(argv[3]);
	}
      else
	{
	  cout << argv[0] << " : No path has been found." << endl;
	  return 0;
	}
      
      /********* compute the time passed ********/

      int elapsedTs = t2.tv_sec - t1.tv_sec;
      int elapsedTus = t2.tv_usec - t1.tv_usec;
      double elapsedT = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
      cout << "Elapsed time : " << elapsedT << endl;
           
      if(map != NULL) delete map;
      if(pfm != NULL) delete pfm;
    }
}

