#include "TrajPlanner.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class TrajPlanner ***************/
                                    /**************************************************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
TrajPlanner::TrajPlanner()
{
  this->traj_   = NULL;
  this->env_    = NULL;
  this->sys_    = NULL;
  this->initS_  = NULL;
  this->goalS_  = NULL;
  this->status_ = PF_NOT_STARTED;
  this->type_   = ND_TP;
}

/**
 *  \brief init constructor
 *  \param e : environment in which the system evolve
 */
TrajPlanner::TrajPlanner(System* sys,Environment* e)
{
  this->traj_   = NULL;
  this->env_    = e;
  this->sys_    = sys;
  this->initS_  = NULL;
  this->goalS_  = NULL;
  this->status_ = PF_NOT_STARTED;
  this->type_   = ND_TP;
}

/**
 *  \brief destructor
 */
TrajPlanner::~TrajPlanner()
{
  if(traj_ != NULL) delete traj_;
  if(initS_ != NULL) delete initS_;
  if(goalS_ != NULL) delete goalS_;
}

  
/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return the current trajectory
 */
const STraj* TrajPlanner::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return this->traj_;
  else return NULL;
}

/**
 *  \brief return the current trajectory
 *  \return the current trajectory
 */
STraj* TrajPlanner::traj()
{
  if(this->status_ == TRAJ_CPTD) return this->traj_;
  else return NULL;
}


/***************** Collision Checking ******************/

/**
 *  \brief check if a collision exists between a trajectory and a static environment
 *  \param traj : the input trajectory
 *  \param env : the input environment
 *  \return true if a collision exists between the trajectory and the static environment
 */
bool mpc::staticCollision(const STraj& traj,const Environment* env)
{
  // check if the environment is correctly defined
  if(env == NULL) return false;

  bool found = false;
  int N = traj.size();
  int i=0;
  
  while(!found && i<N)
    {
      // get back the current state-time of the trajectory
      const StateTime& st = traj.at(i);

      // check if the current state-time is in collision with the environment
      if(env->staticObstruct(st.posXYTh())) found = true;
      else i++;
    }

  return found;
}
