#ifndef MANASYS_HPP
#define MANASYS_HPP

#include "system/inc/WheelChairSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class ManaSys *****************/
                                    /**************************************************/

  /** \class ManaSys
   * \brief Mana system class
   */
  class ManaSys : public WheelChairSys
  {
  public : 

    /********** Constructors and destructors **********/
    
    /**
     *  \brief default constructor
     */
    ManaSys();

    /**
     *  \brief copy constructor
     *  \param sys : the Mana system constraints to copy
     */
    ManaSys(const ManaSys& sys);
    
    /**
     *  \brief init constructor
     *  \param wr : radius of the wheels (m)
     *  \param wb : distance between the 2 wheels
     *  \param v : maximal linear speed (m/s)
     *  \param a : maximal linear acceleration (m/s^2)
     *  \param av : maximal rotation speed (rad/s)
     *  \param aa : maximal rotation acceleration (rad/s^2)
     *  \param bd : braking distance from max speed (m)
     *  \param l : lenght size (m)
     *  \param w : width size (m)
     *  \param cmb : true if the Mana can move backward
     */
    ManaSys(double v,double a,double av,double aa,double l,double w,bool cmb=false);

    /**
     *  \brief destructor
     */
    virtual ~ManaSys();

    /*************** Display functions ****************/
    
    /**
     *  \brief display the features of the Mana system
     */
    virtual void display() const;

    /**
     *  \brief convert the system into a string
     *  \return the resulting string
     */
    virtual const char* toString() const;
  };

};

#endif // MANASYS_HPP
