#include "system/inc/ManaSys.hpp"
using namespace mpc;

#include <stdlib.h>
#include <iostream>
using namespace std;

#include "common/inc/Utils.hpp"

                                    /**************************************************/
                                    /****************** Class ManaSys *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/
    
/**
 *  \brief default constructor
 */
ManaSys::ManaSys() : WheelChairSys(1.0,1.0,1.0,1.0,1.0,1.0,0.0,1.0,1.0,false)
{
  this->type_              = MANA_SYS;
}

/**
 *  \brief copy constructor
 *  \param sys : the Mana system constraints to copy
 */
ManaSys::ManaSys(const ManaSys& sys) : WheelChairSys(sys)
{
  this->type_              = MANA_SYS;
}

/**
 *  \brief init constructor
 *  \param wr : radius of the wheels (m)
 *  \param wb : distance between the 2 wheels
 *  \param v : maximal linear speed (m/s)
 *  \param a : maximal linear acceleration (m/s^2)
 *  \param av : maximal rotation speed (rad/s)
 *  \param aa : maximal rotation acceleration (rad/s^2)
 *  \param bd : braking distance from max speed (m)
 *  \param l : lenght size (m)
 *  \param w : width size (m)
 *  \param cmb : true if the Mana can move backward
 */
ManaSys::ManaSys(double v,double a,double av,double aa,double l,double w,bool cmb) : WheelChairSys(1.0,w,v,a,av,aa,0.0,l,w,cmb)
{
  this->type_              = MANA_SYS;
}

/**
 *  \brief destructor
 */
ManaSys::~ManaSys()
{
}

/*************** Display functions ****************/

/**
 *  \brief display the features of the Mana system
 */
void ManaSys::display() const
{
  cout << "MANA Robotic System : " << endl;
  cout << "\t max_speed : " << max_speed << endl;
  cout << "\t max_acc : " << max_acc << endl;
  cout << "\t max_rotation_speed : " << max_rotation_speed << endl;
  cout << "\t max_rotation_acc : " << max_rotation_acc << endl;
  cout << "\t length : " << length << endl;
  cout << "\t width : " << width << endl;  
  cout << endl;
}

/**
 *  \brief convert the system into a string
 *  \return the resulting string
 */
const char* ManaSys::toString() const
{
  char c3[50]; double2String(max_speed,c3,2);
  char c4[50]; double2String(max_acc,c4,2);
  char c5[50]; double2String(max_rotation_speed,c5,2);
  char c6[50]; double2String(max_rotation_acc,c6,2);
  char c8[50]; double2String(length,c8,2);
  char c9[50]; double2String(width,c9,2);

  std::string str("WheelChairSys :");
  str += "\n";
  str += "max_speed : ";
  str += c3;
  str += "\n";
  str += "max_acc : ";
  str += c4;
  str += "\n";
  str += "max_rotation_speed : ";
  str += c5;
  str += "\n";
  str += "max_rotation_acc : ";
  str += c6;
  str += "\n";
  str += "length : ";
  str += c8;
  str += "\n";
  str += "width : ";
  str += c9;
  str += "\n";

  return str.c_str();
}
