#include "PointMass3DSys.hpp"
using namespace mpc;

#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /************** Class PointMass3DSys **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PointMass3DSys::PointMass3DSys()
{
  this->radius_ = 0;
  this->vxMax_  = 0;
  this->vyMax_  = 0;
  this->vzMax_  = 0;
  this->axMax_  = 0;
  this->ayMax_  = 0;
  this->azMax_  = 0;
}

/**
 *  \brief destructor
 */
PointMass3DSys::~PointMass3DSys()
{
}


/************** Getters and setters ***************/

/**
 *  \brief getter : maximal speed
 *  return the maximal speed wrt. its environment
 *  \return maximal speed
 */
double PointMass3DSys::getVMax() const 
{ 
  return sqrt(vxMax_*vxMax_ + vyMax_*vyMax_ + vzMax_*vzMax_); 
}

/**
 *  \brief getter : maximal acceleration
 *  return the maximal acceleration wrt. its environment
 *  \return acceleration speed
 */
double PointMass3DSys::getAMax() const 
{ 
  return sqrt(axMax_*axMax_ + ayMax_*ayMax_ + azMax_*azMax_);
}


/*************** Display functions ****************/

/**
 *  \brief display the features of the car-like system
 */
void PointMass3DSys::display() const
{
  cout << "Point Mass System : " << endl;
  cout << "\t X Speed Max : " << vxMax_ << endl;
  cout << "\t Y Speed Max : " << vyMax_ << endl;
  cout << "\t Z Speed Max : " << vzMax_ << endl;
  cout << "\t X Acc Max : " << axMax_ << endl;
  cout << "\t Y Acc Max : " << ayMax_ << endl;
  cout << "\t Z Acc Max : " << azMax_ << endl;
  cout<< endl;
}

/**
 *  \brief draw the car-like system in an openGL viewer
 */
void PointMass3DSys::toGL() const
{
  cout << "PointMassSys::toGL : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief convert the system into a string
 *  \return the resulting string
 */
const char* PointMass3DSys::toString() const
{
  cout << "PointMassSys::toString : Not implemented yet!" << endl;
  exit(1);
}
