#include "PointMassSys.hpp"
using namespace mpc;

#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /*************** Class PointMassSys ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PointMassSys::PointMassSys()
{
  this->radius_ = 0;
  this->vxMax_  = 0;
  this->vyMax_  = 0;
  this->axMax_  = 0;
  this->ayMax_  = 0;
}

/**
 *  \brief destructor
 */
PointMassSys::~PointMassSys()
{
}


/************** Getters and setters ***************/

/**
 *  \brief getter : maximal speed
 *  return the maximal speed wrt. its environment
 *  \return maximal speed
 */
double PointMassSys::getVMax() const 
{ 
  return sqrt(vxMax_*vxMax_ + vyMax_*vyMax_); 
}

/**
 *  \brief getter : maximal acceleration
 *  return the maximal acceleration wrt. its environment
 *  \return acceleration speed
 */
double PointMassSys::getAMax() const 
{ 
  return sqrt(axMax_*axMax_ + ayMax_*ayMax_); 
}


/*************** Display functions ****************/

/**
 *  \brief display the features of the car-like system
 */
void PointMassSys::display() const
{
  cout << "Point Mass System : " << endl;
  cout << "\t X Speed Max : " << vxMax_ << endl;
  cout << "\t Y Speed Max : " << vyMax_ << endl;
  cout << "\t X Acc Max : " << axMax_ << endl;
  cout << "\t Y Acc Max : " << ayMax_ << endl;
  cout<< endl;
}

/**
 *  \brief draw the car-like system in an openGL viewer
 */
void PointMassSys::toGL() const
{
  cout << "PointMassSys::toGL : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief convert the system into a string
 *  \return the resulting string
 */
const char* PointMassSys::toString() const
{
  cout << "PointMassSys::toString : Not implemented yet!" << endl;
  exit(1);
}
