#include "SysComponent.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /*************** Class SysComponent ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SysComponent::SysComponent()
{
}

/**
 *  \brief copy constructor
 *  \param c : the control point to copy
 */
SysComponent::SysComponent(const SysComponent& sc)
{
  this->setPos(sc.pos());
  this->setBShape(sc.bshape());
  for(int i=0; i<sc.getNbCP(); i++)
    {
      this->addCP(sc.getCP(i));
    }
}

/**
 *  \brief init constructor
 *  \param pos : position of the component in the workspace (W)
 *  \param bbox : bounding box defined overthe system
 */
SysComponent::SysComponent(const PositionXYTh& pos,const BoundingShape& bshape)
{
  this->setPos(pos);
  this->setBShape(bshape);  
}

/**
 *  \brief destructor
 */
SysComponent::~SysComponent()
{
}


/************** Getters and setters ***************/
  
/**
 *  \brief Position accessor
 *  \return the position of the object in its workspace (W)
 */
PositionXYTh& SysComponent::pos() const
{
  cout << "SysComponent::pos() : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief setter : set the current position of the system component
 *  \param p : the new position of the system component
 */
void SysComponent::setPos(const PositionXYTh& p)
{
  cout << "SysComponent::setPos() : Not implemented yet!" << endl;
  exit(1);
}


/**************** Motion functions ****************/

/**
 *  \brief translate the system component
 *  \param p : vector of the translation
 */
void SysComponent::translate(const PositionXY& p)
{
  this->pos().translate(p);
}

/**
 *  \brief rotate the system component around p about an angle a
 *  \param p : center of the rotation
 *  \param a : angle of the rotation
 */
void SysComponent::rotate(const PositionXY& p,const double& a)
{
  this->pos().rotate(p,a);
}

/*************** Display functions ****************/

/**
 *  \brief display the system component's features
 */
void SysComponent::display() const
{
  cout << "Position : "; this->pos().display(); cout << endl;
  cout << "Bounding shape : "; this->bshape().display(); cout << endl;
}

/**
 *  \brief draw the system component in an openGL viewer
 */
void SysComponent::toGL() const
{
}
