#include "trajGen/inc/ManaTrajGen.hpp"
using namespace mpc;

// #include "gnuplot/inc/TrajPlot.hpp"

#include <stdio.h>
#include <math.h>
#include <vector>
#include "utilsLuis.h"
#include <complex>

#include "math/inc/InvMatrix.hpp"
#include "gnuplot/inc/gnuplot_i.hpp"

#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

// Parameters
#define B      (*vecParameters_p)[0]
#define C      (*vecParameters_p)[1]
#define D      (*vecParameters_p)[2]
#define B2     (*vecParameters_p)[3]
#define C2     (*vecParameters_p)[4]
#define D2     (*vecParameters_p)[5]
#define T      (*vecParameters_p)[6]

// States 
#define X0     vecState_x0[0]
#define Y0     vecState_x0[1]
#define TH0    vecState_x0[2]
#define V0     vecState_x0[3]
//#define OM0    vecState_x0[4]
#define W0     vecState_x0[4]
#define T0     vecState_x0[5]

#define Xf     vecState_xf[0]
#define Yf     vecState_xf[1]
#define THf    vecState_xf[2]
#define Vf     vecState_xf[3]
//#define OMf    vecState_xf[4]
#define Wf    vecState_xf[4]
#define Tf     vecState_xf[5]

#define Xsim   vecState_xsim[0]
#define Ysim   vecState_xsim[1]
#define THsim  vecState_xsim[2]
#define Vsim   vecState_xsim[3]
//#define OMsim  vecState_xsim[4]
#define Wsim  vecState_xsim[4]
#define Tsim   vecState_xsim[5]

#define Xind   matStateInd_[0]
#define Yind   matStateInd_[1]
#define THind  matStateInd_[2]
#define Vind   matStateInd_[3]
//#define OMind  matStateInd_[4]
#define Wind  matStateInd_[4]
#define Tind   matStateInd_[5]

#define Xerr   vecConstraints_g[Xind]
#define Yerr   vecConstraints_g[Yind]
#define THerr  vecConstraints_g[THind]
#define Verr   vecConstraints_g[Vind]
//#define OMerr  vecConstraints_g[OMind]
#define Werr  vecConstraints_g[Wind]

#define DFLT_DERIV_VAL -999999999.99

                                    /**************************************************/
                                    /**************** Class ManaTrajGen ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 */  
ManaTrajGen::ManaTrajGen() : TrajGen()
{
  xLength = 6;       // State length (x, y, th, v, w, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, v(t)=v_f, w(t)=w_f)
  dLength = 4;       // Dynamic Constraints length (maxAcc, maxAAcc, maxVel, maxAVel)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  omFixed_ = true;

  USE_VEL_CST = true;
  USE_ACC_CST = true;
  USE_AVEL_CST = true;
  USE_AACC_CST = true;

  matStateInd_ = new int[xLength];

  ManaWeights_ = ManaState(1,1,1,1,1);
}

/**
 *  \brief init constructor
 *  \param sys : the constraints of the wheelchair system
 */
ManaTrajGen::ManaTrajGen(const ManaSys* sys) : TrajGen(sys)
{
  xLength = 6;       // State length (x, y, th, v, om, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, v(t)=v_f, om(t)=om_f)
  dLength = 4;       // Dynamic Constraints length (maxAcc, maxAAcc, maxVel, maxAVel)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  omFixed_ = true;

  USE_VEL_CST = true;
  USE_ACC_CST = true;
  USE_AVEL_CST = true;
  USE_AACC_CST = true;

  matStateInd_ = new int[xLength];

  ManaWeights_ = ManaState(1,1,1,1,1);
}

/**
 *  \brief destructor
 */  
ManaTrajGen::~ManaTrajGen()
{
  if(matStateInd_ != NULL) delete [] matStateInd_;
}


/*************** Getters and Setters **************/

// return true if linear velocity constraints are used
bool ManaTrajGen::getUseVelCst()
{
  return this->USE_VEL_CST;
}

// set true if linear velocity constraints are used
void ManaTrajGen::setUseVelCst(bool u)
{
  this->USE_VEL_CST = u;
}

// return true if linear acceleration constraints are used
bool ManaTrajGen::getUseAccCst()
{
  return this->USE_ACC_CST;
}

// set true if linear acceleration constraints are used
void ManaTrajGen::setUseAccCst(bool u)
{
  this->USE_ACC_CST = u;
}

// return true if angular velocity constraints are used
bool ManaTrajGen::getUseAVelCst()
{
  return this->USE_AVEL_CST;
}

// set true if angular velocity constraints are used
void ManaTrajGen::setUseAVelCst(bool u)
{
  this->USE_AVEL_CST = u;
}

// return true if angular acceleration constraints are used
bool ManaTrajGen::getUseAAccCst()
{
  return this->USE_AACC_CST;
}

// set true if angular acceleration constraints are used
void ManaTrajGen::setUseAAccCst(bool u)
{
  this->USE_AACC_CST = u;
}

/**
 *  \brief fix or free the final position
 *  \param fixed : true if the final position must be fixed
 */
void ManaTrajGen::fixFinalXY(bool fixed)
{
  xyFixed_ = fixed;
}

/**
 *  \brief fix or free the final heading
 *  \param fixed : true if the final heading must be fixed
 */
void ManaTrajGen::fixFinalTH(bool fixed)
{
  thFixed_ = fixed;
}

/**
 *  \brief fix or free the final linear velocity
 *  \param fixed : true if the final linear velocity must be fixed
 */
void ManaTrajGen::fixFinalV(bool fixed)
{
  vFixed_ = fixed;
}

/**
 *  \brief fix or free the final angular velocity
 *  \param fixed : true if the final angular velocity must be fixed
 */
void ManaTrajGen::fixFinalOM(bool fixed)
{
  omFixed_ = fixed;
}


/********* Optimization related routines **********/

/**
 *  \brief compute a default initial guess of parameters
 */
void ManaTrajGen::ComputeDfltIG()
{
  //Inital guess of parameters
  double dist = sqrt(pow(Xf-X0,2) + pow(Yf-Y0,2));
  double dir = atan2(Yf-Y0,Xf-X0);
  double difDir = computeMinDiffAngle(normalizedAngle(dir),normalizedAngle(TH0));

  if(DBG_ >= 1) {
    cout << "Tf : " << Tf << endl;
    cout << "T0 : " << T0 << endl;
  }

  T = Tf - T0;

  if(T <= 0)
    {
      B = 0;
      C = 0;
      D = 0;
      B2 = 0;
      C2 = 0;
      D2 = 0;
    }
  else
    {
      double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
      double vmoy = (dist*coefDist)/T;
	
//       B = (-2.0*W0 + 0.2*difDir) / T;
      B = (-1.0*W0 + 0.1*difDir) / T;
      C = 0.5*W0 / (T*T);
//       C = 0.25*W0 / (T*T);
//       B = 0.0;
//       C = 0.0;
      D = 0;
	
      if(DBG_ >= 1) {
	printf ("Initial Guess Parameters K = a=%f,b=%f,c=%f,d=%f,t=%f)\n", 
		W0,B,C,D,T);
      }
	
//       B2 = 2.0*(vmoy - V0) / T;
      B2 = 1.0*(vmoy - V0) / T;
      C2 = 0.0;
      D2 = 0.0;
	
      if(DBG_ >= 1) {    
	printf ("Initial Guess Parameters V = a=%f,b=%f,c=%f,d=%f,t=%f)\n", 
		V0,B2,C2,D2,T);
      }
    }
}

/**
 *  \brief select an initial guess of parameters from an IG database
 *  \return 0 if a set of parameters has been found, -1 else
 */
int ManaTrajGen::ComputeIGfromDB()
{
  cerr << "ManaTrajGen::ComputeIGfromDB : Not implemented yet!" << endl;
  return -1;
}

/**
 *  \brief compute the error between the state reached and the state we want to reach
 */
void ManaTrajGen::ComputeG()
{
  if(xyFixed_) Xerr  = Xsim - Xf;
  if(xyFixed_) Yerr  = Ysim - Yf;
  if(thFixed_) THerr = computeMinDiffAngle(normalizedAngle(THsim),normalizedAngle(THf));
  if(vFixed_) Verr  = Vsim - Vf;
  if(omFixed_) Werr = Wsim - Wf;
}

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
 */
void ManaTrajGen::ComputeJacobianG_FD()
{
  cerr << "ManaTrajGen::ComputeJacobianG_FD : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters
 */
void ManaTrajGen::ComputeJacobianS()
{
  if(vFixed_)
    {
      // computation of dV/dP
      matJacobianG[Vind][0] = 0.0; // dV/dB
      matJacobianG[Vind][1] = 0.0; // dV/dC
      matJacobianG[Vind][2] = 0.0; // dV/dD
      matJacobianG[Vind][3] = T; // dV/dB2
      matJacobianG[Vind][4] = T*T; // dV/dC2
      matJacobianG[Vind][5] = T*T*T; // dV/dD2
//       if(tcStatus() != TIME_FIXED) matJacobianG[Vind][6] = funcAcc(T); // dV/T
      if(tcStatus() != TIME_FIXED) matJacobianG[Vind][6] = B2 + 2.0*C2*T + 3.0*D2*T*T; // dV/T
    }

  if(omFixed_)
    {
      // computation of dOM/dP
      matJacobianG[Wind][0] = T; // dOM/dB
      matJacobianG[Wind][1] = T*T; // dOM/dC
      matJacobianG[Wind][2] = T*T*T; // dOM/dD
      matJacobianG[Wind][3] = 0.0; // dOM/dB2
      matJacobianG[Wind][4] = 0.0; // dOM/dC2
      matJacobianG[Wind][5] = 0.0; // dOM/dD2
//       if(tcStatus() != TIME_FIXED) matJacobianG[Wind][6] = funcAAcc(T); // dOM/dT
      if(tcStatus() != TIME_FIXED) matJacobianG[Wind][6] = B + 2.0*C*T + 3.0*D*T*T; // dV/T
    }

  if(thFixed_)
    {
      // computation of dTH/dP
      matJacobianG[THind][0] = 0.5*T*T; // dTH/dB
      matJacobianG[THind][1] = 0.333333333*T*T*T; // dTH/dC
      matJacobianG[THind][2] = 0.25*T*T*T*T; // dTH/dD
      matJacobianG[THind][3] = 0.0; // dTH/dB2
      matJacobianG[THind][4] = 0.0; // dTH/dC2
      matJacobianG[THind][5] = 0.0; // dTH/dD2
//       if(tcStatus() != TIME_FIXED) matJacobianG[THind][6] = funcAVel(T); // dTH/dT
      if(tcStatus() != TIME_FIXED) matJacobianG[THind][6] = straj()->w(T); // dTH/dT
    }
  
  if(xyFixed_)
    {
      matJacobianG[Xind][0] = qgaus(&ManaTrajGen::itgdXdB,0,T); // dX/dB
      matJacobianG[Xind][1] = qgaus(&ManaTrajGen::itgdXdC,0,T); // dX/dC
      matJacobianG[Xind][2] = qgaus(&ManaTrajGen::itgdXdD,0,T); // dX/dD
      matJacobianG[Xind][3] = qgaus(&ManaTrajGen::itgdXdB2,0,T); // dX/dB2
      matJacobianG[Xind][4] = qgaus(&ManaTrajGen::itgdXdC2,0,T); // dX/dC2
      matJacobianG[Xind][5] = qgaus(&ManaTrajGen::itgdXdD2,0,T); // dX/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Xind][6] = dXdT(T); // dX/dT                                        -------------------------------------- a verifier!!!! --------------------------------------
      
      matJacobianG[Yind][0] = qgaus(&ManaTrajGen::itgdYdB,0,T);  // dY/dB
      matJacobianG[Yind][1] = qgaus(&ManaTrajGen::itgdYdC,0,T);  // dY/dC
      matJacobianG[Yind][2] = qgaus(&ManaTrajGen::itgdYdD,0,T);  // dY/dD
      matJacobianG[Yind][3] = qgaus(&ManaTrajGen::itgdYdB2,0,T); // dY/dB2
      matJacobianG[Yind][4] = qgaus(&ManaTrajGen::itgdYdC2,0,T); // dY/dC2
      matJacobianG[Yind][5] = qgaus(&ManaTrajGen::itgdYdD2,0,T); // dY/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Yind][6] = dYdT(T); // dY/dT                                        -------------------------------------- a verifier!!!! --------------------------------------
    }
}

/**
 *  \brief compute the second order partial derivatives of the state wrt. the parameters
 */
void ManaTrajGen::ComputeHessianS()
{
  cerr << "ManaTrajGen::ComputeHessianS : Not implemented yet!" << endl;
  exit(1);
}

double ManaTrajGen::itgdXdB(const double t) const
{
  return -0.5*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t;
}

double ManaTrajGen::itgdXdC(const double t) const
{
  return -0.333333333*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t;
}

double ManaTrajGen::itgdXdD(const double t) const
{
  return -0.25*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t*t;
}

double ManaTrajGen::itgdXdB2(const double t) const
{
  return t * cos(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double ManaTrajGen::itgdXdC2(const double t) const
{
  return t*t * cos(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double ManaTrajGen::itgdXdD2(const double t) const
{
  return t*t*t * cos(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double ManaTrajGen::dXdT(const double t) const
{
//   funcVel(t) * cos(normalizedAngle(funcTheta(t)));
  straj()->v(t) * cos(normalizedAngle(straj()->th(t)));
}

double ManaTrajGen::itgdYdB(const double t) const
{
  return 0.5*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t;
}

double ManaTrajGen::itgdYdC(const double t) const
{
  return 0.333333333*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t;
}

double ManaTrajGen::itgdYdD(const double t) const
{
  return 0.25*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t*t;
}

double ManaTrajGen::itgdYdB2(const double t) const
{
  return t * sin(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double ManaTrajGen::itgdYdC2(const double t) const
{
  return t*t * sin(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double ManaTrajGen::itgdYdD2(const double t) const
{
  return t*t*t * sin(TH0 + W0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double ManaTrajGen::dYdT(const double t) const
{
//   funcVel(t) * sin(normalizedAngle(funcTheta(t)));
  straj()->v(t) * sin(normalizedAngle(straj()->th(t)));
}

/**
 *  \brief compute the residual between the goal state and the stae reached
 *  \return the residual between the goal state and the stae reached
 */
double ManaTrajGen::ComputeResidual() const
{
  double sqRes = 0;

  if(xyFixed_)
    {
      sqRes += (Xerr * Xerr);
      sqRes += (Yerr * Yerr);
    }
  if(thFixed_) sqRes += (50.0 * (THerr * THerr));
  if(vFixed_) sqRes += (10.0 * (Werr * Werr));
  if(omFixed_) sqRes += (10.0 * (Verr * Verr));
  
  return sqrt(sqRes);
}


/**************** Quadrature tools ****************/

/**
 *  \brief Quadrature tools...
 */
double ManaTrajGen::qgaus(double (ManaTrajGen::*func)(const double) const, const double a, const double b) const
{
  static const int n = 20;
  double x[n];
  double w[n];
  int j;
  double s;
  
  gauleg(a, b, x, w, n);      
  s=0;
  for (j=0;j<n;j++) {        
    s += w[j]*((this->*func)(x[j]));
    }
  return s;
}


/*********** State profiles computations **********/

/**
 *  \brief compute all profiles
 */
void ManaTrajGen::computeAllFct()
{
  // delete previous control and state trajectories
  if(ctraj_ != NULL) delete ctraj_;
  if(straj_ != NULL) delete straj_;

  // compute the control trajectory from the vector of parameters 
  Polynomial polyV(D2,C2,B2,V0);
  Polynomial polyW(D,C,B,W0);

  ctraj_ = new ManaPCTraj(polyV,polyW,T);

  // add bounds over the control if needed
  if(USE_VEL_CST)
    {
      if(sys()->canMoveBack()) ctraj()->addVelBounds(-sys()->getVMax(),sys()->getVMax());
      else ctraj()->addVelBounds(0.0,sys()->getVMax());
    }
  if(USE_AVEL_CST)
    {
      ctraj()->addAVelBounds(-sys()->getAVMax(),sys()->getAVMax());
    }

  // compute the state trajectory from the control trajectory
  straj_ = new ManaPSTraj(ManaState(X0,Y0,TH0,V0,W0), *ctraj());

  // add bounds over the state if needed
  // no need... already bounded
}


/*************** Forward Solution *****************/

/**
 *  \brief compute the state reached by the current set of parameters
 */
void ManaTrajGen::ForwardSolution()
{
  for(int i=0; i<gLength; i++)
    for(int j=0; j<pLengthFree(); j++)
      matJacobianG[i][j] = DFLT_DERIV_VAL;
  
  // compute the different functions' profiles
  computeAllFct();

  // compute the simulated state
  Xsim  = straj()->x(T); 
  Ysim  = straj()->y(T); 
  THsim  = straj()->th(T); 
  Vsim  = straj()->v(T); 
  Wsim  = straj()->w(T); 
  if(tcStatus() != TIME_FIXED) Tsim  = T;
}


/***************** Initialization *****************/

/**
 *  \brief set the initial guess of parameters
 *  \param b,c,d : coefficients of the polynomial right wheel acceleration profile
 *  \param b2,c2,d2 : coefficients of the polynomial left wheel acceleration profile
 *  \param t : final time to compute the trajectory (time of the final state)
 */
void ManaTrajGen::setInitialGuess(double b1,double c1,double d1,double b2,double c2,double d2,double t)
{
  vecParameters_p = new Vec_DP(pLength);

  B = b1;
  C = c1;
  D = d1;
  B2 = b2;
  C2 = c2;
  D2 = d2;
  T = t;
}

/**
 *  \brief set the initial guess with a vector of parameters                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param parameters : the vector of parameters
 *
 *  x0 = (X0,Y0,TH0,V0,OM0,T0)
 */
void ManaTrajGen::setInitialConfig(const double* x0)
{
  cerr << "ManaTrajGen::setInitialConfig : Deprecated function (used by computeFromIG). Should not be used." << endl;
  exit(1);
}

/**
 *  \brief set the system parameters
 *  \param constraints : constraints defined over the system
 */
void ManaTrajGen::setSysParams(const System* constraints)
{
  // nothing to do : this function should be removed
  // direct acces to this->sys() should be used
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void ManaTrajGen::setInitialGuess(const double* p)
{
  this->setInitialGuess(p[0],p[1],p[2],p[3],p[4],p[5],p[6]);
}


/************* Trajectory Generation **************/

/**
 *  \brief init the fixed state parameters
 *  \return the size of the vector of free parameters
 */
int ManaTrajGen::initFixedSParam()
{
  int cpt = 0;

  if(xyFixed_)
    {
      Xind = 0;
      Yind = 1;
      cpt += 2;
    }
  if(thFixed_)
    {
      THind = cpt;
      cpt++;
    }
  if(vFixed_)
    {
      Vind = cpt;
      cpt++;
    }
  if(omFixed_)
    {
      Wind = cpt;
      cpt++;
    }

  return cpt;
}

/**
 *  \brief generate a trajectory by constraints optimiation                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param parameters : initial guess of parameters
 *  \return 0 if a solution has been found, -1 else
 */
int ManaTrajGen::ComputeTrajectory(const ManaStateT& s0,const ManaStateT& sf,double** parameters)
{
  cerr << "ManaTrajGen::ComputeTrajectory : Deprecated function. Use :    int ManaTrajGen::genTraj(const ManaStateT& s0,const ManaStateT& sf,const double* p)" << endl;
  exit(1);
}

/**
 *  \brief generate a trajectory by constraints optimiation                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param parameters : initial guess of parameters
 *  \return 0 if a solution has been found, -1 else
 */
int ManaTrajGen::ComputeTrajectory(const ManaStateT& s0,const ManaStateT& sf,Vec_DP** parameters)
{
  cerr << "ManaTrajGen::ComputeTrajectory : Deprecated function. Use :    int ManaTrajGen::genTraj(const ManaStateT& s0,const ManaStateT& sf,const double* p)" << endl;
  exit(1);
}

/**
 *  \brief generate a trajectory by constraints optimiation                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param s0 : initial state
 *  \param sys : constraints defined over the system
 *  \param param : initial guess of parameters
 */
void ManaTrajGen::ComputeFromIG(const StateTime& s0,const double* param)
{
  cerr << "ManaTrajGen::ComputeTrajectory : Deprecated function. Use ManaPCTraj or ManaPSTraj instead." << endl;
  exit(1);
}

/**
 *  \brief generate a trajectory by constraints optimiation                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param statet : initial state-time
 *  \param ctrlt : control-time applied on the system
 */
void ManaTrajGen::ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt)
{
  cerr << "ManaTrajGen::ComputeTrajectory : Deprecated function. Use ManaPCTraj or ManaPSTraj instead." << endl;
  exit(1);
}


/****** Trajectory Generation (2nd version) ******/

/**
 *  \brief init the cot function weights
 */
void ManaTrajGen::initWeights()
{
  if(xyFixed_)
    {
      weightsCF_[Xind] = ManaWeights_.x();
      weightsCF_[Yind] = ManaWeights_.y();
    }
  if(thFixed_)
    {
      weightsCF_[THind] = ManaWeights_.th();
    }
  if(omFixed_)
    {
      weightsCF_[Wind] = ManaWeights_.om();
    }
  if(vFixed_)
    {
      weightsCF_[Vind] = ManaWeights_.s();
    }
}

/**
 *  \brief generate a trajectory between two states
 *  \param s0 : initial state-time
 *  \param sf : initial state-time
 *  \param p : initial guess of parameters
 *  \return 0 if a solution has been found, -1 else
 */
int ManaTrajGen::genTraj(const ManaStateT& s0,const ManaStateT& sf,const double* p)
{
  double x0[6];
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.v();
  x0[4] = s0.w();
  x0[5] = 0;
  
  double xf[6];
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.v();
  xf[4] = sf.w();
  xf[5] = sf.t() - s0.t();

  return this->TrajGen::genTraj(x0,xf,p);
}

/**************** Output Functions ****************/

/**
 *  \brief return the piecewise polynomial state trajectory computed
 *
 *  A new instance of ManaPSTraj is returned
 */
ManaPCTraj* ManaTrajGen::getPCTraj() const
{
  return ctraj()->copy();
}

/**
 *  \brief return the piecewise polynomial state trajectory computed
 *
 *  A new instance of ManaPSTraj is returned
 */
ManaPSTraj* ManaTrajGen::getPSTraj() const
{
  return straj()->copy();
}

/**
 *  \brief return the piecewise polynomial state trajectory computed
 *
 *  A new instance of ManaSTraj is returned
 */
ManaSTraj* ManaTrajGen::getSTraj(double dtSampling) const
{
  return straj()->sample(dtSampling);
}

/**
 *  \brief return the piecewise polynomial state trajectory computed
 *  \param sampling : the number of samples of the resulting trajectory
 *
 *  A new instance of ManaSTraj is returned
 */
ManaSTraj* ManaTrajGen::getSTraj(int sampling) const
{
  return straj()->sample(sampling);
}

/**
 *  \brief get back the state reached at time t                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,v,w,t)
 */
void ManaTrajGen::getStateTReachedAt(double* rstatet,double t) const
{
  rstatet[0] = straj()->x(t);
  rstatet[1] = straj()->y(t);
  rstatet[2] = straj()->th(t);
  rstatet[3] = straj()->v(t);
  rstatet[4] = straj()->w(t);
  rstatet[5] = t;
}

/**
 *  \brief get back the state reached at time t                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,v,w,t)
 */
void ManaTrajGen::getFinalStateTReached(double* rstatet) const
{
  rstatet[0] = straj()->x(T);
  rstatet[1] = straj()->y(T);
  rstatet[2] = straj()->th(T);
  rstatet[3] = straj()->v(T);
  rstatet[4] = straj()->w(T);
  rstatet[5] = T;
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 */
void ManaTrajGen::traceOn(Map *map,int nbp)
{
  if(straj_ == NULL)
    {
      cerr << "ManaTrajGen:traceOn : Warning. No state trajectory has been generated yet. Can't trace the path." << endl;
      return;
    }
  if(map == NULL)
    {
      cerr << "ManaTrajGen:traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }

  int pathCol = (int) (map->getDepth() / 2);
  for(int i=0; i<=nbp; i++)
    {
      double x = straj()->x(i*(1.0/((double)nbp)*T));
      double y = straj()->y(i*(1.0/((double)nbp)*T));
      
	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 *  \param initPos : initial position
 */  
void ManaTrajGen::traceOn(Map *map,int nbp,const PositionXYTh& initPos)
{
  if(straj_ == NULL)
    {
      cerr << "ManaTrajGen:traceOn : Warning. No state trajectory has been generated yet. Can't trace the path." << endl;
      return;
    }
  if(map == NULL)
    {
      cerr << "ManaTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }

  int pathCol = (int) (map->getDepth() / 2);
  for(int i=0; i<=nbp; i++)
    {
      double x1 = straj()->x(i*(1.0/((double)nbp)*T));
      double y1 = straj()->y(i*(1.0/((double)nbp)*T));
      double th1 = straj()->th(i*(1.0/((double)nbp)*T));
      PositionXYTh pos1(x1,y1,th1);
      
      PositionXYTh pos = coordInFrame(pos1,initPos);
      double x = pos.x();
      double y = pos.y();
      
      if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	map->setCol(((int)x),((int)y),pathCol);
    }
}

/**
 *  \brief return a sample trajectory from the parameterized one                     -------------------------------------------- Deprecated --------------------------------------------
 *  \param ts : the sampling time step
 *  \return a sample trajectory from the parameterized one
 *
 *  A new instance of sample trajectory is returned here
 */
ManaSTraj* ManaTrajGen::sampledTraj(double ts) const
{
  if(straj_ != NULL) return this->getSTraj(ts);
  else return NULL;
}


/*************** Utility Functions ****************/

/**
 *  \brief convert the initial and final input state-times into tables of double
 *  \param initST : the input initial state-time
 *  \param goalST : the input goal state-time
 *  \param initTD : the resulting initial table
 *  \param goalTD : the resulting goal table
 *
 *  Tables must be allocated first
 */
void ManaTrajGen::InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const
{
  ManaStateT s0(*((ManaStateT*)(&initST)));
  ManaStateT sf(*((ManaStateT*)(&goalST)));

  initTD[0] = s0.x();
  initTD[1] = s0.y();
  initTD[2] = s0.th();
  initTD[3] = s0.v();
  initTD[4] = s0.w();
  initTD[5] = 0;
  
  goalTD[0] = sf.x();
  goalTD[1] = sf.y();
  goalTD[2] = sf.th();
  goalTD[3] = sf.v();
  goalTD[4] = sf.w();
  goalTD[5] = sf.t() - s0.t();
}

/**
 *  \brief display both wheels velocity, linear and angular velocity, heading and position profiles in a GNUPlot window                     -------------------------------------------- Deprecated --------------------------------------------
 */
void ManaTrajGen::displayProfiles()
{
//   cerr << "ManaTrajGen::displayProfiles : Deprecated function. Use : TrajPlot (gui/gnuplot/inc/TrajPlot.hpp) instead" << endl;
//   exit(1);

  ManaTrajGen::plotManaPCTraj(*ctraj(), sys(), true);
  ManaTrajGen::plotManaPSTraj(*straj(), sys(), true);
}

/**
 *  \brief display all functions
 */
void ManaTrajGen::displayAllFct()
{
  if(ctraj_ != NULL)
    {
      cout << "Control trajectory : " << endl;
      ctraj()->display();
    }
  else cout << "Control trajectory nul." << endl;

  if(straj_ != NULL)
    {
      cout << "State trajectory : " << endl;
      straj()->display();
    }
  else cout << "State trajectory nul." << endl;
}


/***************** Plot functions *****************/

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 */
void ManaTrajGen::plotManaCTraj(const ManaCTraj& ctraj, const ManaSys* sys, bool dispBounds)
{
  // initialize the control vectors
  vector<double> tim;

  vector<double> vel;
  vector<double> velBinf;
  vector<double> velBsup;

  vector<double> avel;
  vector<double> avelBinf;
  vector<double> avelBsup;

  // fill the vectors with the control trajectory values
  int N = ctraj.getSize();
  for(int i=0; i<N; i++)
    {
      ManaControlT ctrlt = ctraj.at(i);

      vel.push_back(ctrlt.v());
      avel.push_back(ctrlt.w());
      tim.push_back(ctrlt.t());

      if(sys != NULL && dispBounds)
	{
	  if(sys->canMoveBack()) velBinf.push_back(-sys->getVMax());
	  else velBinf.push_back(0.0);
	  velBsup.push_back(sys->getVMax());
	  avelBinf.push_back(-sys->getAVMax());
	  avelBsup.push_back(sys->getAVMax());
	}
    }

  // display the gnuplot windows
  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(tim,vel,"Linear Velocity V(t)");
  if(sys != NULL && dispBounds)
    {
      gVel.set_style("points").plot_xy(tim,velBinf,"Min Bound");
      gVel.set_style("points").plot_xy(tim,velBsup,"Max Bound");
    }

  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(tim,avel,"Angular Velocity W(t)");
  if(sys != NULL && dispBounds)
    {
      gAVel.set_style("points").plot_xy(tim,avelBinf,"Min Bound");
      gAVel.set_style("points").plot_xy(tim,avelBsup,"Max Bound");
    }

  wait_for_key();
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 */
void ManaTrajGen::plotManaSTraj(const ManaSTraj& straj, const ManaSys* sys, bool dispBounds)
{
  // initialize the control vectors
  vector<double> tim;

  vector<double> posX;
  vector<double> posY;
  vector<double> headTH;

  vector<double> posItgX;
  vector<double> posItgY;

  vector<double> vel;
  vector<double> velBinf;
  vector<double> velBsup;

  vector<double> avel;
  vector<double> avelBinf;
  vector<double> avelBsup;

  // fill the vectors with the state trajectory values
  int N = straj.getSize();
  for(int i=0; i<N; i++)
    {
      ManaStateT statet = straj.at(i);

      posX.push_back(statet.x());
      posY.push_back(statet.y());
      posItgX.push_back(statet.v() * cos(statet.th()));
      posItgY.push_back(statet.v() * sin(statet.th()));
      headTH.push_back(statet.th());
      vel.push_back(statet.v());
      avel.push_back(statet.w());
      tim.push_back(statet.t());

      if(sys != NULL && dispBounds)
	{
	  if(sys->canMoveBack()) velBinf.push_back(-sys->getVMax());
	  else velBinf.push_back(0.0);
	  velBsup.push_back(sys->getVMax());
	  avelBinf.push_back(-sys->getAVMax());
	  avelBsup.push_back(sys->getAVMax());
	}
    }

  // display the gnuplot windows
  Gnuplot gXY;
  gXY.reset_plot();
  gXY.set_style("lines").plot_xy(posX,posY,"Trajectory XxY");

  Gnuplot gTimXY;
  gTimXY.reset_plot();
  gTimXY.set_style("lines").plot_xy(tim,posX,"X(t)");
  gTimXY.set_style("lines").plot_xy(tim,posY,"Y(t)");

  Gnuplot gItgXY;
  gItgXY.reset_plot();
  gItgXY.set_style("lines").plot_xy(tim,posItgX,"Itg X");
  gItgXY.set_style("lines").plot_xy(tim,posItgY,"Itg Y");

  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(tim,headTH,"Heading Th(t)");

  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(tim,vel,"Linear Velocity V(t)");
  if(sys != NULL && dispBounds)
    {
      gVel.set_style("points").plot_xy(tim,velBinf,"Min Bound");
      gVel.set_style("points").plot_xy(tim,velBsup,"Max Bound");
    }

  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(tim,avel,"Angular Velocity W(t)");
  if(sys != NULL && dispBounds)
    {
      gAVel.set_style("points").plot_xy(tim,avelBinf,"Min Bound");
      gAVel.set_style("points").plot_xy(tim,avelBsup,"Max Bound");
    }

  wait_for_key();
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 *  \param sampling : the number of sampling poitns of the trajectory
 */
void ManaTrajGen::plotManaPCTraj(const ManaPCTraj& ctraj, const ManaSys* sys, bool dispBounds, int sampling)
{
  // sample the control trajectory
  ManaCTraj* sampledTraj = ctraj.sample(sampling);

  if(sampledTraj != NULL)
    {
      ManaTrajGen::plotManaCTraj(*sampledTraj, sys, dispBounds);
      
      // delete the sampled trajectory
      delete sampledTraj;
    }
  else cerr << "TrajPlot::plotManaPCTraj : Unable to sample the control trajectory." << endl;
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 *  \param sampling : the number of sampling poitns of the trajectory
 */
void ManaTrajGen::plotManaPSTraj(const ManaPSTraj& straj, const ManaSys* sys, bool dispBounds, int sampling)
{
  // sample the control trajectory
  ManaSTraj* sampledTraj = straj.sample(sampling);

  if(sampledTraj != NULL)
    {
      ManaTrajGen::plotManaSTraj(*sampledTraj, sys, dispBounds);
      
      // delete the sampled trajectory
      delete sampledTraj;
    }
  else cerr << "TrajPlot::plotManaPSTraj : Unable to sample the state trajectory." << endl;
}
