#include "SSTrajGen.hpp"
using namespace mpc;

#include "Utils.hpp"

#include "gnuplot/inc/gnuplot_i.hpp"

#include <iostream>
using namespace std;

// Parameters
#define B      (*vecParameters_p)[0]
#define C      (*vecParameters_p)[1]
#define D      (*vecParameters_p)[2]
#define B2     (*vecParameters_p)[3]
#define C2     (*vecParameters_p)[4]
#define D2     (*vecParameters_p)[5]
#define T      (*vecParameters_p)[6]

// States 
#define X0     vecState_x0[0]
#define Y0     vecState_x0[1]
#define TH0    vecState_x0[2]
#define VX0    vecState_x0[3]
#define VY0    vecState_x0[4]
#define OM0    vecState_x0[5]
#define T0     vecState_x0[6]

#define Xf     vecState_xf[0]
#define Yf     vecState_xf[1]
#define THf    vecState_xf[2]
#define VXf    vecState_xf[3]
#define VYf    vecState_xf[4]
#define OMf    vecState_xf[5]
#define Tf     vecState_xf[6]

#define Xsim   vecState_xsim[0]
#define Ysim   vecState_xsim[1]
#define THsim  vecState_xsim[2]
#define VXsim  vecState_xsim[3]
#define VYsim  vecState_xsim[4]
#define OMsim  vecState_xsim[5]
#define Tsim   vecState_xsim[6]

#define Xerr   vecConstraints_g[Xind]
#define Yerr   vecConstraints_g[Yind]
#define THerr  vecConstraints_g[THind]
#define VXerr  vecConstraints_g[VXind]
#define VYerr  vecConstraints_g[VYind]
#define OMerr  vecConstraints_g[OMind]

#define Xind   matStateInd_[0]
#define Yind   matStateInd_[1]
#define THind  matStateInd_[2]
#define VXind  matStateInd_[3]
#define VYind  matStateInd_[4]
#define OMind  matStateInd_[5]
#define Tind   matStateInd_[6]

#define TIME_FIXED true

#define COEF_COR 0.2
#define VARIABLE_COEF_COR true

#define DFLT_DERIV_VAL -999999999.99

#define FAST_PART_DERIV true

#define COEF_WHENBDD 0.2

#define MAXABS_POS_ERR    0.1
#define MAXREL_POS_ERR    0.01


                                    /**************************************************/
                                    /***************** Class SSTrajGen ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
SSTrajGen::SSTrajGen() : TrajGen()
{
  xLength = 7;       // State length (x, y, th, vx, vy, om, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 6;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, vx(t)=vx_f, vy(t)=vy_f, om(t)=om_f)
  dLength = 4;       // Dynamic Constraints length (maxVel, maxAcc, maxAVel, maxAAcc)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  posX   = NULL;
  posY   = NULL;
  headTH = NULL;
  velVX  = NULL;
  velVY  = NULL;
  avelOM = NULL;
  timeT  = NULL;

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  fctAcc_ = NULL;
  fctAAcc_ = NULL;
  fctAVel_ = NULL;
  fctTH_ = NULL;

  fctAccWithoutBounds_ = NULL;
  fctAAccWithoutBounds_ = NULL;
  fctAVelWithoutBounds_ = NULL;
  fctTHWithoutBounds_ = NULL;

  this->setFDpartDeriv(true);

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  omFixed_ = true;

  matStateInd_ = new int[xLength];

  SSweights_ = SSState(1,1,1,1,1,1);
}

/**
 *  \brief init constructor
 *  \param sys : the constraints of the spaceship system
 */
SSTrajGen::SSTrajGen(const SpaceShipSys* sys) : TrajGen(sys)
{
  xLength = 7;       // State length (x, y, th, vx, vy, om, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 6;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, vx(t)=vx_f, vy(t)=vy_f, om(t)=om_f)
  dLength = 4;       // Dynamic Constraints length (maxVel, maxAcc, maxAVel, maxAAcc)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  posX   = NULL;
  posY   = NULL;
  headTH = NULL;
  velVX  = NULL;
  velVY  = NULL;
  avelOM = NULL;
  timeT  = NULL;

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  fctAcc_ = NULL;
  fctAAcc_ = NULL;
  fctAVel_ = NULL;
  fctTH_ = NULL;

  fctAccWithoutBounds_ = NULL;
  fctAAccWithoutBounds_ = NULL;
  fctAVelWithoutBounds_ = NULL;
  fctTHWithoutBounds_ = NULL;

  this->setFDpartDeriv(true);

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  omFixed_ = true;

  matStateInd_ = new int[xLength];

  SSweights_ = SSState(1,1,1,1,1,1);
}

/**
 *  \brief destructor
 */  
SSTrajGen::~SSTrajGen()
{
  if(posX != NULL) delete posX;
  if(posY != NULL) delete posY;
  if(headTH != NULL) delete headTH;
  if(velVX != NULL) delete velVX;
  if(velVY != NULL) delete velVY;
  if(avelOM != NULL) delete avelOM;
  if(timeT != NULL) delete timeT;

  if(fctAcc_ != NULL) delete fctAcc_;
  if(fctAAcc_ != NULL)delete fctAAcc_;
  if(fctAVel_ != NULL)delete fctAVel_;
  if(fctTH_ != NULL)delete fctTH_;
  
  if(fctAccWithoutBounds_ != NULL)delete fctAccWithoutBounds_;
  if(fctAAccWithoutBounds_ != NULL)delete fctAAccWithoutBounds_;
  if(fctAVelWithoutBounds_ != NULL)delete fctAVelWithoutBounds_;
  if(fctTHWithoutBounds_ != NULL)delete fctTHWithoutBounds_;

  if(matStateInd_ != NULL) delete matStateInd_;
}


/*************** Getters and Setters **************/

/**
 *  \brief fix or free the final position
 *  \param fixed : true if the final position must be fixed
 */
void SSTrajGen::fixFinalXY(bool fixed)
{
  xyFixed_ = fixed;
}

/**
 *  \brief fix or free the final heading
 *  \param fixed : true if the final heading must be fixed
 */
void SSTrajGen::fixFinalTH(bool fixed)
{
  thFixed_ = fixed;
}

/**
 *  \brief fix or free the final linear velocity
 *  \param fixed : true if the final linear velocity must be fixed
 */
void SSTrajGen::fixFinalV(bool fixed)
{
  vFixed_ = fixed;
}

/**
 *  \brief fix or free the final angular velocity
 *  \param fixed : true if the final angular velocity must be fixed
 */
void SSTrajGen::fixFinalOM(bool fixed)
{
  omFixed_ = fixed;
}


/********* Optimization related routines **********/

/**
 *  \brief compute a default initial guess of parameters
 */
void SSTrajGen::ComputeDfltIG()
{
  //Inital guess of parameters
  double dist = sqrt(pow(Xf-X0,2) + pow(Yf-Y0,2));
  double dir = atan2(Yf-Y0,Xf-X0);
  double difDir = computeMinDiffAngle(normalizedAngle(dir),normalizedAngle(TH0));
  
  T = Tf - T0;
  
  double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
  double vmoy = (dist*coefDist)/T;
  
  // create two positions pfrom the initial and final states
  PositionXYTh P0(X0,Y0,TH0);
  PositionXYTh Pf(Xf,Yf,THf);
  PositionXYTh PfInP0 = coordInFrame(Pf,P0);

  B = 0.2*difDir / T;
  C = 0; // c
  D = 0; // d
  
  double V0 = sqrt(VX0*VX0 + VY0*VY0);
  
  if(PfInP0.x() >= 0)
    {
      B2 = 0.0001 + 2.0*(vmoy - V0) / T;
      C2 = 0.0;
      D2 = 0.0;
    }
  else 
    {
      B2 = -2.0*(vmoy - V0) / T;
      C2 = 0.0;
      D2 = 0.0;
    }
}

/**
 *  \brief select an initial guess of parameters from an IG database
 *  \return 0 if a set of parameters has been found, -1 else
 */
int SSTrajGen::ComputeIGfromDB()
{
  cerr << "SSTrajGen::ComputeIGfromDB : No spaceship ig database yet!" << endl;
  return -1;
}

/**
 *  \brief compute the error between the state reached and the state we want to reach
 */
void SSTrajGen::ComputeG()
{
  if(xyFixed_) Xerr = Xsim - Xf;
  if(xyFixed_) Yerr = Ysim - Yf;
  if(thFixed_) THerr = computeMinDiffAngle(normalizedAngle(THsim),normalizedAngle(THf));
  if(vFixed_) VXerr = VXsim - VXf;
  if(vFixed_) VYerr = VYsim - VYf;
  if(omFixed_) OMerr = OMsim - OMf;
}

// /**
//  *  \brief compute the partial derivatives of the state wrt. the parameters
//  */
// void SSTrajGen::ComputeJacobianG_FD()
// {
//   this->computeAllFct();
//   double oldX  = this->funcX(T);
//   double oldY  = this->funcY(T);
//   double oldTH = this->funcTheta(T);
//   double oldVX = this->funcVelX(T);
//   double oldVY = this->funcVelY(T);
//   double oldOM = this->funcAVel(T);

//   B += E;
//   this->computeAllFct();
//   double X_B1  = this->funcX(T);
//   double Y_B1  = this->funcY(T);
//   double TH_B1 = this->funcTheta(T);
//   double VX_B1 = this->funcVelX(T);
//   double VY_B1 = this->funcVelY(T);
//   double OM_B1 = this->funcAVel(T);  
//   B -= E;

//   // compute dS/dB1
//   matJacobianG[0][0] = (X_B1 - oldX)*UnSurE;            // dX/dB1
//   matJacobianG[1][0] = (Y_B1 - oldY)*UnSurE;            // dY/dB1
//   matJacobianG[2][0] = (TH_B1 - oldTH)*UnSurE;          // dTH/dB1
//   matJacobianG[3][0] = (VX_B1 - oldVX)*UnSurE;          // dVX/dB1
//   matJacobianG[4][0] = (VY_B1 - oldVY)*UnSurE;          // dVY/dB1
//   matJacobianG[5][0] = (OM_B1 - oldOM)*UnSurE;          // dOM/dB1

//   C += E;
//   this->computeAllFct();
//   double X_C1  = this->funcX(T);
//   double Y_C1  = this->funcY(T);
//   double TH_C1 = this->funcTheta(T);
//   double VX_C1 = this->funcVelX(T);
//   double VY_C1 = this->funcVelY(T);
//   double OM_C1 = this->funcAVel(T);  
//   C -= E;

//   // compute dS/dC1
//   matJacobianG[0][1] = (X_C1 - oldX)*UnSurE;            // dX/dC1
//   matJacobianG[1][1] = (Y_C1 - oldY)*UnSurE;            // dY/dC1
//   matJacobianG[2][1] = (TH_C1 - oldTH)*UnSurE;          // dTH/dC1
//   matJacobianG[3][1] = (VX_C1 - oldVX)*UnSurE;          // dVX/dC1
//   matJacobianG[4][1] = (VY_C1 - oldVY)*UnSurE;          // dVY/dC1
//   matJacobianG[5][1] = (OM_C1 - oldOM)*UnSurE;          // dOM/dC1

//   D += E;
//   this->computeAllFct();
//   double X_D1  = this->funcX(T);
//   double Y_D1  = this->funcY(T);
//   double TH_D1 = this->funcTheta(T);
//   double VX_D1 = this->funcVelX(T);
//   double VY_D1 = this->funcVelY(T);
//   double OM_D1 = this->funcAVel(T);  
//   D -= E;

//   // compute dS/dD1
//   matJacobianG[0][2] = (X_D1 - oldX)*UnSurE;            // dX/dD1
//   matJacobianG[1][2] = (Y_D1 - oldY)*UnSurE;            // dY/dD1
//   matJacobianG[2][2] = (TH_D1 - oldTH)*UnSurE;          // dTH/dD1
//   matJacobianG[3][2] = (VX_D1 - oldVX)*UnSurE;          // dVX/dD1
//   matJacobianG[4][2] = (VY_D1 - oldVY)*UnSurE;          // dVY/dD1
//   matJacobianG[5][2] = (OM_D1 - oldOM)*UnSurE;          // dOM/dD1

//   B2 += E;
//   this->computeAllFct();
//   double X_B2  = this->funcX(T);
//   double Y_B2  = this->funcY(T);
//   double TH_B2 = this->funcTheta(T);
//   double VX_B2 = this->funcVelX(T);
//   double VY_B2 = this->funcVelY(T);
//   double OM_B2 = this->funcAVel(T);  
//   B2 -= E;

//   // compute dS/dB2
//   matJacobianG[0][3] = (X_B2 - oldX)*UnSurE;            // dX/dB2
//   matJacobianG[1][3] = (Y_B2 - oldY)*UnSurE;            // dY/dB2
//   matJacobianG[2][3] = (TH_B2 - oldTH)*UnSurE;          // dTH/dB2
//   matJacobianG[3][3] = (VX_B2 - oldVX)*UnSurE;          // dVX/dB2
//   matJacobianG[4][3] = (VY_B2 - oldVY)*UnSurE;          // dVY/dB2
//   matJacobianG[5][3] = (OM_B2 - oldOM)*UnSurE;          // dOM/dB2

//   C2 += E;
//   this->computeAllFct();
//   double X_C2  = this->funcX(T);
//   double Y_C2  = this->funcY(T);
//   double TH_C2 = this->funcTheta(T);
//   double VX_C2 = this->funcVelX(T);
//   double VY_C2 = this->funcVelY(T);
//   double OM_C2 = this->funcAVel(T);  
//   C2 -= E;

//   // compute dS/dC2
//   matJacobianG[0][4] = (X_C2 - oldX)*UnSurE;            // dX/dC2
//   matJacobianG[1][4] = (Y_C2 - oldY)*UnSurE;            // dY/dC2
//   matJacobianG[2][4] = (TH_C2 - oldTH)*UnSurE;          // dTH/dC2
//   matJacobianG[3][4] = (VX_C2 - oldVX)*UnSurE;          // dVX/dC2
//   matJacobianG[4][4] = (VY_C2 - oldVY)*UnSurE;          // dVY/dC2
//   matJacobianG[5][4] = (OM_C2 - oldOM)*UnSurE;          // dOM/dC2

//   D2 += E;
//   this->computeAllFct();
//   double X_D2  = this->funcX(T);
//   double Y_D2  = this->funcY(T);
//   double TH_D2 = this->funcTheta(T);
//   double VX_D2 = this->funcVelX(T);
//   double VY_D2 = this->funcVelY(T);
//   double OM_D2 = this->funcAVel(T);  
//   D2 -= E;

//   // compute dS/dD2
//   matJacobianG[0][5] = (X_D2 - oldX)*UnSurE;            // dX/dD2
//   matJacobianG[1][5] = (Y_D2 - oldY)*UnSurE;            // dY/dD2
//   matJacobianG[2][5] = (TH_D2 - oldTH)*UnSurE;          // dTH/dD2
//   matJacobianG[3][5] = (VX_D2 - oldVX)*UnSurE;          // dVX/dD2
//   matJacobianG[4][5] = (VY_D2 - oldVY)*UnSurE;          // dVY/dD2
//   matJacobianG[5][5] = (OM_D2 - oldOM)*UnSurE;          // dOM/dD2
// }

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters
 */
void SSTrajGen::ComputeJacobianS()
{
  if(omFixed_)
    {
      // computation of dOM/dP
      matJacobianG[OMind][0] = T;      // dOM/dB
      matJacobianG[OMind][1] = T*T;    // dOM/dC
      matJacobianG[OMind][2] = T*T*T;  // dOM/dD
      matJacobianG[OMind][3] = 0.0;    // dOM/dB2
      matJacobianG[OMind][4] = 0.0;    // dOM/dC2
      matJacobianG[OMind][5] = 0.0;    // dOM/dD2
    }

  if(thFixed_)
    {
      // computation of dTH/dP
      matJacobianG[THind][0] = 0.5*T*T;           // dTH/dB
      matJacobianG[THind][1] = 0.333333333*T*T*T; // dTH/dC
      matJacobianG[THind][2] = 0.25*T*T*T*T;      // dTH/dD
      matJacobianG[THind][3] = 0.0;               // dTH/dB2
      matJacobianG[THind][4] = 0.0;               // dTH/dC2
      matJacobianG[THind][5] = 0.0;               // dTH/dD2
    }
  
  if(vFixed_)
    {
      // computation of dVX/dP
      matJacobianG[VXind][0] = dVXdB(T);  // dVX/dB
      matJacobianG[VXind][1] = dVXdC(T);  // dVX/dC
      matJacobianG[VXind][2] = dVXdD(T);  // dVX/dD
      matJacobianG[VXind][3] = dVXdB2(T); // dVX/dB2
      matJacobianG[VXind][4] = dVXdC2(T); // dVX/dC2
      matJacobianG[VXind][5] = dVXdD2(T); // dVX/dD2
      
      // computation of dVY/dP
      matJacobianG[VYind][0] = dVYdB(T);  // dVY/dB
      matJacobianG[VYind][1] = dVYdC(T);  // dVY/dC
      matJacobianG[VYind][2] = dVYdD(T);  // dVY/dD
      matJacobianG[VYind][3] = dVYdB2(T); // dVY/dB2
      matJacobianG[VYind][4] = dVYdC2(T); // dVY/dC2
      matJacobianG[VYind][5] = dVYdD2(T); // dVY/dD2
    }

  if(xyFixed_)
    {
      // computation of dX/dP
      matJacobianG[Xind][0] = (double) qgaus(&SSTrajGen::dVXdB,(long double) 0,(long double) T);  // dX/dB
      matJacobianG[Xind][1] = (double) qgaus(&SSTrajGen::dVXdC,(long double) 0,(long double) T);  // dX/dC
      matJacobianG[Xind][2] = (double) qgaus(&SSTrajGen::dVXdD,(long double) 0,(long double) T);  // dX/dD
      matJacobianG[Xind][3] = (double) qgaus(&SSTrajGen::dVXdB2,(long double) 0,(long double) T); // dX/dB2
      matJacobianG[Xind][4] = (double) qgaus(&SSTrajGen::dVXdC2,(long double) 0,(long double) T); // dX/dC2
      matJacobianG[Xind][5] = (double) qgaus(&SSTrajGen::dVXdD2,(long double) 0,(long double) T); // dX/dD2
      
      // computation of dY/dP
      matJacobianG[Yind][0] = (double) qgaus(&SSTrajGen::dVYdB,(long double) 0,(long double) T);  // dY/dB
      matJacobianG[Yind][1] = (double) qgaus(&SSTrajGen::dVYdC,(long double) 0,(long double) T);  // dY/dC
      matJacobianG[Yind][2] = (double) qgaus(&SSTrajGen::dVYdD,(long double) 0,(long double) T);  // dY/dD
      matJacobianG[Yind][3] = (double) qgaus(&SSTrajGen::dVYdB2,(long double) 0,(long double) T); // dY/dB2
      matJacobianG[Yind][4] = (double) qgaus(&SSTrajGen::dVYdC2,(long double) 0,(long double) T); // dY/dC2
      matJacobianG[Yind][5] = (double) qgaus(&SSTrajGen::dVYdD2,(long double) 0,(long double) T); // dY/dD2
    }
}

/**
 *  \brief compute the second order partial derivatives of the state wrt. the parameters
 */
void SSTrajGen::ComputeHessianS()
{
  cerr << "SSTrajGen::ComputeHessianS : Not implemented yet!" << endl;
  exit(1);
}

long double SSTrajGen::dVXdB(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVXdB,0,T);
}

long double SSTrajGen::dVXdC(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVXdC,0,T);
}

long double SSTrajGen::dVXdD(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVXdD,0,T);
}

long double SSTrajGen::dVXdB2(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVXdB2,0,T);
}

long double SSTrajGen::dVXdC2(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVXdC2,0,T);
}

long double SSTrajGen::dVXdD2(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVXdD2,0,T);
}

long double SSTrajGen::dVYdB(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVYdB,0,T);
}

long double SSTrajGen::dVYdC(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVYdC,0,T);
}

long double SSTrajGen::dVYdD(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVYdD,0,T);
}

long double SSTrajGen::dVYdB2(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVYdB2,0,T);
}

long double SSTrajGen::dVYdC2(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVYdC2,0,T);
}

long double SSTrajGen::dVYdD2(const long double t) const
{
  return qgaus(&SSTrajGen::itgdVYdD2,0,T);
}

long double SSTrajGen::itgdVXdB(const long double t) const
{
  return -0.5*t*t * (B2 + 2.0*C2*t + 3.0*D2*t*t) * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVXdC(const long double t) const
{
  return -0.33333333*t*t*t * (B2 + 2.0*C2*t + 3.0*D2*t*t) * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVXdD(const long double t) const
{
  return -0.25*t*t*t*t * (B2 + 2.0*C2*t + 3.0*D2*t*t) * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVXdB2(const long double t) const
{
  return cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVXdC2(const long double t) const
{
  return 2.0*t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVXdD2(const long double t) const
{
  return 3.0*t*t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVYdB(const long double t) const
{
  return 0.5*t*t * (B2 + 2.0*C2*t + 3.0*D2*t*t) * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVYdC(const long double t) const
{
  return 0.33333333*t*t*t * (B2 + 2.0*C2*t + 3.0*D2*t*t) * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVYdD(const long double t) const
{
  return 0.25*t*t*t*t * (B2 + 2.0*C2*t + 3.0*D2*t*t) * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVYdB2(const long double t) const
{
  return sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVYdC2(const long double t) const
{
  return 2.0*t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

long double SSTrajGen::itgdVYdD2(const long double t) const
{
  return 3.0*t*t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}


/**************** Quadrature tools ****************/

/**
 *  \brief Quadrature tools...
 */
long double SSTrajGen::qgaus(long double (SSTrajGen::*func)(const long double) const, const long double a, const long double b) const
{
  static const int n = 20;
  long double x[n];
  long double w[n];
  int j;
  long double s;
  
  gauleg2(a, b, x, w, n);      
  s=0;
  for (j=0;j<n;j++) {        
    s += w[j]*((this->*func)(x[j]));
  }
  return s;
}


/*********** State profiles computations **********/

/**
 *  \brief compute the linear acceleration profile
 */
void SSTrajGen::computeFctAcc()
{
  // delete the previous acceleration function
  if(fctAcc_ != NULL) 
    {
      fctAcc_->display();
      delete fctAcc_;
    }

  // create the polynomial acceleration function
  fctAccWithoutBounds_ = new Polynomial(3.0*D2,2.0*C2,B2);
  fctAcc_ = new MultiItvFunction(fctAccWithoutBounds_,0,INFINITY);

  // add the bounds over the acceleration
  list<ItvFunction> lMinBItv;
  fctAcc_->addMinBound(-maxAcc);
  list<ItvFunction> lMaxBItv;
  fctAcc_->addMaxBound(maxAcc);
}

/**
 *  \brief compute the angular acceleration profile
 */
void SSTrajGen::computeFctAAcc()
{
  // delete the previous angular acceleration function
  if(fctAAcc_ != NULL) delete fctAAcc_;

  // create the polynomial angular acceleration function
  fctAAccWithoutBounds_ = new Polynomial(3.0*D,2.0*C,B);
  fctAAcc_ = new MultiItvFunction(fctAAccWithoutBounds_,0,INFINITY);

  // add the bounds over the acceleration
  list<ItvFunction> lMinBItv;
  fctAAcc_->addMinBound(-maxAAcc);
  list<ItvFunction> lMaxBItv;
  fctAAcc_->addMaxBound(maxAAcc);
}

/**
 *  \brief compute the angular velocity profile
 */
void SSTrajGen::computeFctAVel()
{
  fctAVel_ = (MultiItvFunction*) fctAAcc_->integration(0.0,OM0);
  
//   list<ItvFunction>* lMinBItv = 
  fctAVel_->addMinBound(-maxAVel);
//   list<ItvFunction>* lMaxBItv = 
  fctAVel_->addMaxBound(maxAVel);
  
//   if(ItvFunction::isIncludeIn(T,lMinBItv)) lastItvAVelBdd_ = -1;
//   else if(ItvFunction::isIncludeIn(T,lMaxBItv)) lastItvAVelBdd_ = 1;
//   else lastItvAVelBdd_ = 0;
  
//   if(DBG_ >= 1) {
//     // recompute the angular acceleration if the angular velocity changed
//     delete fctAAcc_;
//     fctAAcc_ = (MultiItvFunction*) fctAVel_->derivation();
//   }
}

/**
 *  \brief compute the heading profile
 */
void SSTrajGen::computeFctTH()
{
  fctTH_ = (MultiItvFunction*) fctAVel_->integration(0.0,TH0);
}

// /**
//  *  \brief compute the linear velocity on X profile
//  */
// void SSTrajGen::computeFctVelX()
// {
// }

// /**
//  *  \brief compute the linear velocity on Y profile
//  */
// void SSTrajGen::computeFctVelY()
// {
// }

/**
 *  \brief compute all profiles
 */
void SSTrajGen::computeAllFct()
{
  computeFctAcc();
  computeFctAAcc();
  computeFctAVel();
  computeFctTH();
//   computeFctVelX();
//   computeFctVelY();
}


/********** State Parameters Computation **********/  

/**
 *  \brief get the value of the linear acceleration profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the linear acceleration profile at time t
 */
double SSTrajGen::funcAcc(const double t) const
{
  return this->fctAcc_->getVal(t);
}

/**
 *  \brief get the value of the angular acceleration profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular acceleration profile at time t
 */
double SSTrajGen::funcAAcc(const double t) const
{
  return this->fctAAcc_->getVal(t);
}

/**
 *  \brief integration function for the velocity VX profile computation
 *  \param t : the time at which the function is evaluated
 *  \return the value at time t of the integration function
 */  
long double SSTrajGen::funcItgVelX(const long double t) const
{
  return funcAcc(t) * cos(normalizedAngle(funcTheta(t)));
}

/**
 *  \brief integration function for the velovity VY profile computation
 *  \param t : the time at which the function is evaluated
 *  \return the value at time t of the integration function
 */  
long double SSTrajGen::funcItgVelY(const long double t) const
{
  return funcAcc(t) * sin(normalizedAngle(funcTheta(t)));
}

/**
 *  \brief get the value of the linear velocity on X profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the linear velocity on X profile at time t
 */
long double SSTrajGen::funcVelX(const long double t) const
{
  long double res = VX0 + qgaus(&SSTrajGen::funcItgVelX,0,t);
  if(res > maxVel) return maxVel;
  else if(res < -maxVel) return -maxVel;
  else return res;
}

/**
 *  \brief get the value of the linear velocity on Y profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the linear velocity on Y profile at time t
 */
long double SSTrajGen::funcVelY(const long double t) const
{
  long double res = VY0 + qgaus(&SSTrajGen::funcItgVelY,0,t);
  if(res > maxVel) return maxVel;
  else if(res < -maxVel) return -maxVel;
  else return res;
}

/**
 *  \brief get the value of the angular velocity profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular velocity profile at time t
 */
double SSTrajGen::funcAVel(const double t) const
{
  return this->fctAVel_->getVal(t);
}

/**
 *  \brief get the value of the heading profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the heading profile at time t
 */
double SSTrajGen::funcTheta(const double t) const
{
  return normalizedAngle(this->fctTH_->getVal(t));
}

/**
 *  \brief get the value of the position X profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the position X profile at time t
 */
double SSTrajGen::funcX(const double t) const
{
  return X0 + qgaus(&SSTrajGen::funcVelX,0,t);
}

/**
 *  \brief get the value of the position Y profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the position Y profile at time t
 */
double SSTrajGen::funcY(const double t) const
{
  return Y0 + qgaus(&SSTrajGen::funcVelY,0,t);
}


/*************** Forward Solution *****************/

/**
 *  \brief compute the state reached by the current set of parameters
 */
void SSTrajGen::ForwardSolution()
{
  // initialize the partial derivatives
  for(int i=0; i<gLength; i++)
    for(int j=0; j<pLength-1; j++)
      matJacobianG[i][j] = DFLT_DERIV_VAL;
  
//   lastItvAccBdd_ = 0;
//   lastItvVelBdd_ = 0;
//   lastItvAVelBdd_ = 0;

  // compute the different functions' profiles
  computeAllFct();

  // compute the simulated state
  Xsim = funcX(T);
  Ysim = funcY(T);
  THsim = funcTheta(T);
  VXsim = funcVelX(T);
  VYsim = funcVelY(T);
  OMsim = funcAVel(T);
}


/***************** Initialization *****************/

// /**
//  *  \brief set the initial state, final state and constraints vectors
//  *  \param x0 : initial state
//  *  \param xf : final state
//  *  \param constraints : constraints defined over the system
//  *
//  *  constraints[0] : max angular acceleration
//  *  constraints[1] : max linear acceleration
//  *  constraints[2] : max angular velocity
//  *  constraints[3] : max linear velocity
//  */
// void SSTrajGen::setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints)
// {
//  //Set initial, final states and curvature constraint
//   vecState_x0 = x0;
//   vecState_xf = xf;

//   maxAAcc = constraints[0];
//   maxAcc = constraints[1];
//   maxAVel = constraints[2];
//   maxVel = constraints[3];

//   fctAcc_ = NULL;
//   fctAAcc_ = NULL;
//   fctAVel_ = NULL;
//   fctTH_ = NULL;

//   fctAccWithoutBounds_ = NULL;
//   fctAAccWithoutBounds_ = NULL;
//   fctAVelWithoutBounds_ = NULL;
//   fctTHWithoutBounds_ = NULL;
// }

/**
 *  \brief set the system parameters
 *  \param constraints : constraints defined over the system
 */
void SSTrajGen::setSysParams(const System* constraints)
{
  if(constraints->getType() != SPACESHIP_SYS)
    {
      throw string("SSTrajGen::setSysParams : Wrong system type!");
    }

  maxAcc = ((SpaceShipSys*)constraints)->getAMax();
  maxVel = ((SpaceShipSys*)constraints)->getVMax();
  maxAAcc = ((SpaceShipSys*)constraints)->getAThMax();
  maxAVel = ((SpaceShipSys*)constraints)->getVThMax();
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void SSTrajGen::setInitialGuess(const double* parameters)
{
  if(vecParameters_p != NULL) delete vecParameters_p;
  vecParameters_p = new Vec_DP(xLength);

  B  = parameters[0];
  C  = parameters[1];
  D  = parameters[2];
  B2 = parameters[3];
  C2 = parameters[4];
  D2 = parameters[5];
  T  = parameters[6];
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void SSTrajGen::setInitialConfig(const double* x0)
{
  // Set initial and final states
  vecState_x0 = Vec_DP(xLength);
  vecState_x0[0] = x0[0];
  vecState_x0[1] = x0[1];
  vecState_x0[2] = x0[2];
  vecState_x0[3] = x0[3];
  vecState_x0[4] = x0[4];
  vecState_x0[5] = x0[5];
  vecState_x0[6] = x0[6];
  
  vecState_xf = vecState_x0;
}

/************* Trajectory Generation **************/

/**
 *  \brief init a vector of the constraints defined over the system
 *  \param sys : the robotic system features
 *  \param vcstt : the resulting vector of constraints
 */
void SSTrajGen::initCstt(System* sys,Vec_DP& vcstt)
{
  vcstt[0] = ((SpaceShipSys*)sys)->getAThMax();
  vcstt[1] = ((SpaceShipSys*)sys)->getAMax();
  vcstt[2] = ((SpaceShipSys*)sys)->getVThMax();
  vcstt[3] = ((SpaceShipSys*)sys)->getVMax();
}

/**
 *  \brief init the fixed state parameters
 *  \return the size of the vector of free parameters
 */
int SSTrajGen::initFixedSParam()
{
  int cpt = 0;

  if(xyFixed_)
    {
      Xind = 0;
      Yind = 1;
      cpt += 2;
    }
  if(thFixed_)
    {
      THind = cpt;
      cpt++;
    }
  if(vFixed_)
    {
      VXind = cpt;
      cpt++;
      VYind = cpt;
      cpt++;
    }
  if(omFixed_)
    {
      OMind = cpt;
      cpt++;
    }

  return cpt;
}

// /**
//  *  \brief generate a trajectory by constraints optimiation
//  *  \param x0 : initial state
//  *  \param xf : final state
//  *  \param constraints : constraints defined over the system
//  *  \param parameters : initial guess of parameters
//  *
//  *  x0 = (X0,Y0,TH0,VX0,VY0,OM0,T0)
//  *  xf = (Xf,Yf,THf,VXf,VYf,OMf,Tf)
//  *  constraints = (maxAAcc,maxAcc,maxAVel,maxVel)
//  */
// int SSTrajGen::ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters)
// {
//   // initialize the correction coefficient
//   double coef_cor = COEF_COR;
//   int nbDeacreaseStep = 0;
//   double oldResidual = 0;

//   // Init the initial best parameters and residual
//   double bestB=0, bestC=0, bestD=0, bestB2=0, bestC2=0, bestD2=0;
//   double bestResidual = -1;
//   int bestCNT = -1;

//   //Set initial state, final state and system constraints
//   setInitialParam(x0,xf,constraints);

//   if(DBG_ >= 1) {
//     cout << "trajGen : (X0=" << X0 << ",Y0=" << Y0 << ",TH0=" << TH0 << ",VX0=" << VX0 <<
//       ",VY0=" << VY0 << ",OM0=" << OM0 << ",T0=" << T0 << ") -> ";
//     cout << "(Xf=" << Xf << ",Yf=" << Yf << ",THf=" << THf << ",VXf=" << VXf <<
//       ",VYf=" << VYf << ",OMf=" << OMf << ",Tf=" << Tf << ")" << endl;
//   }

//   //Heuristic guess
//   if(parameters == NULL || *parameters == NULL)
//     InitialGuess(parameters);
//   else
//     {
//       vecParameters_p = *parameters;
//       T = vecState_xf[6] - vecState_x0[6];
//     }

//   if(DBG_ >= 1) {
//     cout << "Initial Guess :  ";
//     cout << "B:" << B << "  C:" << C << "  D:" << D << "  T:" << T << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << "  T:" << T << endl;
//   }

//   // Init the number of the algorithm's step and the residual
//   double residual = 1000000;  
//   int cnt=0;
//   CNT_ = 0;
  
//   vecCorrParameters_p = 0;  

//   while(residual > maxResidual_)
//     {
//       // save the best parameters
//       if(CNT_ == 0 || residual < bestResidual)
// 	{
// 	  bestB = B;
// 	  bestC = C;
// 	  bestD = D;
// 	  bestB2 = B2;
// 	  bestC2 = C2;
// 	  bestD2 = D2;
// 	  bestResidual = residual;
// 	  bestCNT = CNT_;
// 	}

// #if VARIABLE_COEF_COR
//       // update the correction coefficient
//       if(residual <= oldResidual)
// 	{
// 	  nbDeacreaseStep++;
	  
// 	  if(nbDeacreaseStep >= 10) coef_cor = 1.0;
// 	  else if(nbDeacreaseStep >= 9) coef_cor = 0.8;
// 	  else if(nbDeacreaseStep >= 8) coef_cor = 0.7;
// 	  else if(nbDeacreaseStep >= 7) coef_cor = 0.6;
// 	  else if(nbDeacreaseStep >= 6) coef_cor = 0.5;
// 	  else if(nbDeacreaseStep >= 5) coef_cor = 0.4;
// 	  else if(nbDeacreaseStep >= 3) coef_cor = 0.3;	
// 	}
//       else
// 	{
// 	  coef_cor = COEF_COR;
// 	  nbDeacreaseStep = 0;
// 	}
// //       coef_cor = 0.1;
      
//       oldResidual = residual;
// #endif

//       //Apply correction
// #if TIME_FIXED
//       for(int i=0; i<pLength-1;i++)
// #else
//       for(int i=0; i<pLength;i++)
// #endif
// 	(*vecParameters_p)[i] += coef_cor*vecCorrParameters_p[i];            
      
//       //MUST_DISPLAY = false;
//       ForwardSolution();
//       //MUST_DISPLAY = false;
//       CNT_++;

//       if(DBG_ >= 1) {
// 	printf ("Simulated state = (Xsim=%f,Ysim=%f,VXsim=%f,VYsim=%f,OMsim=%f)\n", 
// 		Xsim,Ysim,THsim,VXsim,VYsim,OMsim);
//       }

//       ComputeG();

//       if(DBG_ >= 1) {
// 	printf ("Errors = (%f,%f,%f,%f,%f,%f)\n", 
// 		Xerr,Yerr,THerr,VXerr,VYerr,OMerr);
// 	printf ("Parameters K = (OM0=%f,B=%f,C=%f,D=%f,T=%f)\n", 
// 		OM0,B,C,D,T);
// 	printf ("Parameters A = (B2=%f,C2=%f,D2=%f,T=%f)\n", 
// 		B2,C2,D2,T);
//       }

//       //ComputeJacobianG();
//       ComputeJacobianG_FD();
//       ComputeInvJacobianG();

//       //Compute the correction
// #if TIME_FIXED
//       for(int i=0; i<pLength-1;i++)
// #else
// 	for(int i=0; i<pLength;i++)
// #endif
// 	  {
// 	    vecCorrParameters_p[i] = 0;
	    
// 	    for(int j=0;j<gLength;j++)
// 	      {
// 		vecCorrParameters_p[i] += (-1.0)*vecConstraints_g[j]*matInvJacobianG[i][j]; 
// 	      }	    
// 	  }
      
//       if(DBG_ >= 1) {
// 	printf ("Correction = (B=%f,C=%f,D=%f,B2=%f,C2=%f,D2=%f)\n", 
// 		vecCorrParameters_p[0],vecCorrParameters_p[1],vecCorrParameters_p[2],
// 		vecCorrParameters_p[3],vecCorrParameters_p[4],vecCorrParameters_p[5]);
//       }

//       residual = sqrt( pow(vecConstraints_g[0],2) + 
// 		       pow(vecConstraints_g[1],2) +
// 		       pow(vecConstraints_g[2],2) +
// 		       pow(vecConstraints_g[3],2) +
// 		       pow(vecConstraints_g[4],2) + 
// 		       pow(vecConstraints_g[5],2));
      
//       if(DBG_ >= 1) {        
// 	printf ("Residual = %f\t CNT : %d\t COEF_COR : %f\n", 
// 		residual,CNT_,coef_cor); 
//       } 

//       if(DBG_ >= 2) {
// 	displayProfiles();
//       }     

//       cnt++;
//       if(cnt >= maxNbStep_)
// 	{
// 	  if(DBG_ >= 1) {
// 	    cout << "  No traj found in " << maxNbStep_ << "iterations" << endl;
// 	  }
	  
// 	  // get back the best parameters found
// 	  B = bestB;
// 	  C = bestC;
// 	  D = bestD;
// 	  B2 = bestB2;
// 	  C2 = bestC2;
// 	  D2 = bestD2;
	  
// 	  // compute the forward method with these parameters
// 	  ForwardSolution();
// 	  // compute the associate error
// 	  ComputeG();
	  
// 	  // get back errors on X, Y and TH
// 	  finalResidual_ = bestResidual;
// 	  finalNbSteps_ = CNT_;
// 	  finalErrX_ = vecConstraints_g[0];
// 	  finalErrY_ = vecConstraints_g[1];
// 	  finalErrTH_ = vecConstraints_g[2];
	  
// 	  return -1;
// 	}
//     }
//   finalResidual_ = residual;
//   finalNbSteps_ = CNT_;
//   finalErrX_ = vecConstraints_g[0];
//   finalErrY_ = vecConstraints_g[1];
//   finalErrTH_ = vecConstraints_g[2];
//   return 0;
// }

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param statet : initial state-time
 *  \param ctrlt : control-time applied on the system
 */
void SSTrajGen::ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt)
{
  SSStateT ssST(*((SSStateT*)&statet));
  double x0[7];
  x0[0] = ssST.x();
  x0[1] = ssST.y();
  x0[2] = ssST.th();
  x0[3] = ssST.sx();
  x0[4] = ssST.sy();
  x0[5] = ssST.om();
  x0[6] = 0;

//   double constraints[4];
//   constraints[0] = ((SpaceShipSys*)sys)->getAThMax();
//   constraints[1] = ((SpaceShipSys*)sys)->getAMax();
//   constraints[2] = ((SpaceShipSys*)sys)->getVThMax();
//   constraints[3] = ((SpaceShipSys*)sys)->getVMax();
  
  SSControlT ssCT(*((SSControlT*)(&ctrlt)));
        
  double param[7];
  param[0] = ssCT.aacc();
  param[1] = 0.0;
  param[2] = 0.0;
  param[3] = ssCT.acc();
  param[4] = 0.0;
  param[5] = 0.0;
  param[6] = ssCT.t();

  this->TrajGen::ComputeFromIG(x0,param);
}

/****** Trajectory Generation (2nd version) ******/

/**
 *  \brief init the cot function weights
 */
void SSTrajGen::initWeights()
{
  if(xyFixed_)
    {
      weightsCF_[Xind] = SSweights_.x();
      weightsCF_[Yind] = SSweights_.y();
    }
  if(thFixed_)
    {
      weightsCF_[THind] = SSweights_.th();
    }
  if(omFixed_)
    {
      weightsCF_[OMind] = SSweights_.om();
    }
  if(vFixed_)
    {
      weightsCF_[VXind] = SSweights_.sx();
      weightsCF_[VYind] = SSweights_.sy();
    }
}

/**
 *  \brief generate a trajectory between two states
 *  \param s0 : initial state-time
 *  \param sf : initial state-time
 *  \param p : initial guess of parameters
 */
int SSTrajGen::genTraj(const SSStateT& s0,const SSStateT& sf,const double* p)
{
  double x0[7];
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.sx();
  x0[4] = s0.sy();
  x0[5] = s0.om();
  x0[6] = 0;

  double xf[7];
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.sx();
  xf[4] = sf.sy();
  xf[5] = sf.om();
  xf[6] = sf.t() - s0.t();

  return this->TrajGen::genTraj(x0,xf,p);
}


/**************** Output Functions ****************/

/**
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,vx,vy,om,t)
 */
void SSTrajGen::getStateTReachedAt(double* rstatet,double t) const
{
  rstatet[0] = funcX(t);
  rstatet[1] = funcY(t);
  rstatet[2] = funcTheta(t);
  rstatet[3] = funcVelX(t);
  rstatet[4] = funcVelY(t);
  rstatet[5] = funcAVel(t);
  rstatet[6] = t;
}

/**
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
SSStateT SSTrajGen::getSSStateTReachedAt(double t) const
{
  return SSStateT(funcX(t),
		  funcY(t),
		  funcTheta(t),
		  funcVelX(t),
		  funcVelY(t),
		  funcAVel(t),
		  t);
}

/**
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,vx,vy,om,t)
 */
void SSTrajGen::getFinalStateTReached(double* rstatet) const
{
  rstatet[0] = funcX(T);
  rstatet[1] = funcY(T);
  rstatet[2] = funcTheta(T);
  rstatet[3] = funcVelX(T);
  rstatet[4] = funcVelY(T);
  rstatet[5] = funcAVel(T);
  rstatet[6] = T;
}

/**
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
SSStateT SSTrajGen::getFinalSSStateTReached() const
{
  return SSStateT(funcX(T),
		  funcY(T),
		  funcTheta(T),
		  funcVelX(T),
		  funcVelY(T),
		  funcAVel(T),
		  T);
}

/**
 *  \brief save a sampling of the trajectory profile
 *  \param nbp : number of sampling points
 */
void SSTrajGen::SaveTrajectory(int nbp)
{
  if(posX != NULL) delete posX;
  if(posY != NULL) delete posY;
  if(headTH != NULL) delete headTH;
  if(timeT != NULL) delete timeT;
  if(velVX != NULL) delete velVX;
  if(velVY != NULL) delete velVY;
  if(avelOM != NULL) delete avelOM;

  posX = new vector<double>;
  posY = new vector<double>;
  headTH = new vector<double>;
  timeT = new vector<double>;
  velVX = new vector<double>;
  velVY = new vector<double>;
  avelOM = new vector<double>;

  posX->clear();
  posY->clear();
  headTH->clear();
  timeT->clear();
  velVX->clear();
  velVY->clear();
  avelOM->clear();

  //Computing trajectory
  double step = 1.0 / (((double)nbp)-1.0);
  for(int j=0;j<nbp;j++)
    {
      posX->push_back(funcX(j*(step*T)));
      posY->push_back(funcY(j*(step*T)));
      headTH->push_back(funcTheta(j*(step*T)));
      timeT->push_back(j*step*T);
      velVX->push_back(funcVelX(j*step*T));
      velVY->push_back(funcVelY(j*step*T));
      avelOM->push_back(funcAVel(j*step*T));
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 */  
void SSTrajGen::traceOn(Map *map,int nbp)
{
  if(map == NULL)
    {
      cout << "SSTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x = funcX(i*(1.0/((double)nbp)*T));
	  double y = funcY(i*(1.0/((double)nbp)*T));

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 *  \param initPos : initial position
 */  
void SSTrajGen::traceOn(Map *map,int nbp,const PositionXYTh& initPos)
{
  if(map == NULL)
    {
      cout << "DDTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x1 = funcX(i*(1.0/((double)nbp)*T));
	  double y1 = funcY(i*(1.0/((double)nbp)*T));
	  double th1 = funcTheta(i*(1.0/((double)nbp)*T));
	  PositionXYTh pos1(x1,y1,th1);

	  PositionXYTh pos = coordInFrame(pos1,initPos);
	  double x = pos.x();
	  double y = pos.y();

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/**
 *  \brief return a sample trajectory from the parameterized one
 *  \param ts : the sampling time step
 *  \return a sample trajectory from the parameterized one
 *
 *  A new instance of sample trajectory is returned here
 */
SSSTraj* SSTrajGen::sampledTraj(double ts) const
{
  cout << "SSTrajGen::sampledTraj : Not implemented yet!" << endl;
  exit(1);
}


/*************** Utility Functions ****************/

/**
 *  \brief convert the initial and final input state-times into tables of double
 *  \param initST : the input initial state-time
 *  \param goalST : the input goal state-time
 *  \param initTD : the resulting initial table
 *  \param goalTD : the resulting goal table
 *
 *  Tables must be allocated first
 */
void SSTrajGen::InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const
{
  SSStateT s0(*((SSStateT*)(&initST)));
  SSStateT sf(*((SSStateT*)(&goalST)));

  initTD[0] = s0.x();
  initTD[1] = s0.y();
  initTD[2] = s0.th();
  initTD[3] = s0.sx();
  initTD[4] = s0.sy();
  initTD[5] = s0.om();
  initTD[6] = 0;
  
  goalTD[0] = sf.x();
  goalTD[1] = sf.y();
  goalTD[2] = sf.th();
  goalTD[3] = sf.sx();
  goalTD[4] = sf.sy();
  goalTD[5] = sf.om();
  goalTD[6] = sf.t() - s0.t();
}

/**
 *  \brief display linear and angular velocity, heading and position profiles
 */
void SSTrajGen::displayProfiles()
{
  // compute the forward solution with the new parameters
  this->ForwardSolution();
  
  //Computing Position
  this->SaveTrajectory(100);
  vector<double> xIGpos, yIGpos;
  xIGpos.push_back(X0);
  yIGpos.push_back(Y0);
  xIGpos.push_back(Xf);
  yIGpos.push_back(Yf);

  Gnuplot gPosition;
  gPosition.reset_plot();
  gPosition.set_style("lines").plot_xy(*(this->posX),*(this->posY),"Position");
  gPosition.set_style("points").plot_xy(xIGpos,yIGpos,"Initial and Final Pos");


  //Computing X Velocity
  vector<double> xvel;
  vector<double> distancevx;
  for(int j=0;j<=50;j++){
    xvel.push_back((double)this->funcVelX(j*(T/50)));
    distancevx.push_back((j*(T/50)));
  }    
  Gnuplot gVelX;
  gVelX.reset_plot();
  gVelX.set_style("lines").plot_xy(distancevx,xvel,"X Velocity");

  //Computing Y Velocity
  vector<double> yvel;
  vector<double> distancevy;
  for(int j=0;j<=50;j++){
    yvel.push_back((double)this->funcVelY(j*(T/50)));
    distancevy.push_back((j*(T/50)));
  }    
  Gnuplot gVelY;
  gVelY.reset_plot();
  gVelY.set_style("lines").plot_xy(distancevy,yvel,"Y Velocity");

  //Computing Velocity
  vector<double> acc;
  vector<double> distanceA;
  for(int j=0;j<=50;j++){
    acc.push_back((double)this->funcAcc(j*(T/50)));
    distanceA.push_back((j*(T/50)));
  }    
  Gnuplot gAcc;
  gAcc.reset_plot();
  gAcc.set_style("lines").plot_xy(distanceA,acc,"Acceleration");
  
  //Computing Angular Velocity
  vector<double> avel;
  vector<double> distanceAV;
  for(int j=0;j<=50;j++){
    avel.push_back((double)this->funcAVel(j*(T/50)));
    distanceAV.push_back((j*(T/50)));
  }    
  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Velocity");
  
  //Computing Heading
  vector<double> dir2;
  vector<double> dir2n;
  vector<double> cosdir2n;
  vector<double> sindir2n;
  vector<double> distance2;
  for(int j=0;j<=50;j++){
    double thtmp = (double)this->funcTheta(j*(T/50.0));
    double thtmpn = normalizedAngle(thtmp);
    double disttmp = (j*(T/50.0));
    dir2.push_back(thtmp);
    dir2n.push_back(thtmpn);
    cosdir2n.push_back(cos(thtmpn));
    sindir2n.push_back(sin(thtmpn));
    distance2.push_back(disttmp);
  } 
  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(distance2,dir2,"Heading");
  

  wait_for_key();
}

/**
 *  \brief display all functions
 */
void SSTrajGen::displayAllFct()
{
  cout << "Linear Acceleration : " << endl;
  fctAcc_->display();
  cout << "Angular Acceleration : " << endl;
  fctAAcc_->display();
  cout << "Angular Velocity : " << endl;
  fctAVel_->display();
  cout << "Heading : " << endl;
  fctTH_->display();
}
