#include "WCInitialGuess.hpp"
using namespace mpc;

#include "Timer.hpp"

#include <iostream>
#include <string>
#include <fstream>

#include "gnuplot/inc/gnuplot_i.hpp"
#include "Utils.hpp"

#define SPG_T   sampling_[0]
#define SPG_VI  sampling_[1]
#define SPG_VG  sampling_[2]
#define SPG_DX  sampling_[3]
#define SPG_DY  sampling_[4]
#define SPG_DTH sampling_[5]

#define MIN_SPG_T   minSampS_[0]
#define MIN_SPG_VI  minSampS_[1]
#define MIN_SPG_VG  minSampS_[2]
#define MIN_SPG_DX  minSampS_[3]
#define MIN_SPG_DY  minSampS_[4]
#define MIN_SPG_DTH minSampS_[5]

#define MAX_SPG_T   maxSampS_[0]
#define MAX_SPG_VI  maxSampS_[1]
#define MAX_SPG_VG  maxSampS_[2]
#define MAX_SPG_DX  maxSampS_[3]
#define MAX_SPG_DY  maxSampS_[4]
#define MAX_SPG_DTH maxSampS_[5]

#define USE_ACC_CST  false
#define USE_VEL_CST  true
#define USE_AACC_CST false
#define USE_AVEL_CST true

#define FIXED_XY true
#define FIXED_TH true
#define FIXED_V  false
#define FIXED_AV false
#define FIXED_T  true

                                    /**************************************************/
                                    /******************* Class WC_N2R *****************/
                                    /**************************************************/

double WC_N2R::wX_ = 1.0;                            /**< weight used to compute the distance between two nodes (on X) */
double WC_N2R::wY_ = 1.0;                            /**< weight used to compute the distance between two nodes (on Y) */
double WC_N2R::wTH_ = 10.0;                          /**< weight used to compute the distance between two nodes (on TH) */
double WC_N2R::wVI_ = 10.0;                          /**< weight used to compute the distance between two nodes (on VI) */
double WC_N2R::wVG_ = 10.0;                          /**< weight used to compute the distance between two nodes (on VG) */
double WC_N2R::wT_ = 1000.0;                         /**< weight used to compute the distance between two nodes (on T) */

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
WC_N2R::WC_N2R() : N2R()
{
}

/**
 *  \brief init constructor (of an ig node)
 *  \param initST : the initial state of the trajectory
 *  \param goalST : the final state to reached
 *  \param param : the parametric representation of the control
 */
WC_N2R::WC_N2R(const WCStateT& initST,const WCStateT& goalST,double* param) : N2R()
{
  dimP_ = 7;
  param_ = new double[dimP_];
  for(int i=0; i<dimP_; i++) param_[i] = param[i];

  initST_ = new WCStateT(initST);
  goalST_ = new WCStateT(goalST);

  pos_ = NULL;
  
//   lnear_ = NULL;

  status_ = IG_N2R;
}

/**
 *  \brief init constructor (of an explored node)
 *  \param initST : the initial state of the trajectory
 *  \param goalST : the final state to reached
 *  \param i,j,k,l,m,n : position of the node in the grid
 *  \param node : the node from which the current one may be derivated
 */
WC_N2R::WC_N2R(const WCStateT& initST,const WCStateT& goalST,
       int i,int j,int k,int l,int m,int n,N2R* node) : N2R()
{
  dimP_ = 7;
  param_ = NULL;

  initST_ = new WCStateT(initST);
  goalST_ = new WCStateT(goalST);

  pos_ = new int[6];
  pos_[0] = i;
  pos_[1] = j;
  pos_[2] = k;
  pos_[3] = l;
  pos_[4] = m;
  pos_[5] = n;
  
//   lnear_ = new std::list<N2R*>;
//   this->addNearNode(node);

  status_ = EXPLORED_N2R;
}

/**
 *  \brief init constructor (of a reached node)
 *  \param initST : the initial state of the trajectory
 *  \param goalST : the final state to reached
 *  \param i,j,k,l,m,n : position of the node in the grid
 *  \param param : the parametric representation of the control
 */
WC_N2R::WC_N2R(const WCStateT& initST,const WCStateT& goalST,
       int i,int j,int k,int l,int m,int n,double* param) : N2R()
{
  dimP_ = 7;
  param_ = new double[dimP_];
  for(int p=0; p<dimP_; p++) param_[p] = param[p];

  initST_ = new WCStateT(initST);
  goalST_ = new WCStateT(goalST);

  pos_ = new int[6];
  pos_[0] = i;
  pos_[1] = j;
  pos_[2] = k;
  pos_[3] = l;
  pos_[4] = m;
  pos_[5] = n;
  
//   lnear_ = NULL;

  status_ = REACHED_N2R;
}

/**
 *  \brief destructor
 */
WC_N2R::~WC_N2R()
{
//   this->clearNearNodes();
}


/*************** Getters and Setters **************/

// /**
//  *  \brief return the nearest neighbour of the current node
//  *  \return the nearest neighbour of the current node if one exists, NULL else
//  */
// WC_N2R* WC_N2R::getNearNeighbour() const
// {
//   // check if the list of neighbours has been defined and is not empty
//   if(lnear_ != NULL && (!(lnear_->empty()))) 
//     {
//       WC_N2R* bestNode = NULL;
//       double bestDist = 999999999;

//       std::list<N2R*>::iterator it;
//       for(it = lnear_->begin(); i != lnear_->end(); it++)
// 	{
// 	  N2R* node = *it;
// 	  double dist = this->computeWDistTo(node);

// 	  // check if the current node if the closest node
// 	  if(bestNode == NULL || dist < bestDist)
// 	    {
// 	      bestNode = node;
// 	      bestDist = dist;
// 	    }
// 	}

//       // return the closest node found
//       return bestNode;
//     }
//   else 
//     {
//       return NULL;
//     }
// }


/**************** Utility functions ***************/

/**
 *  \brief return a weighted distance between the current node and the input node
 *  \param n2 : the second node
 *  \return a weighted distance between both nodes
 */
double WC_N2R::computeWDistTo(WC_N2R* n2) const
{
  const WCStateT& ist = this->initST();
  const WCStateT& gst = this->goalST();
  const WCStateT& ist2 = n2->initST();
  const WCStateT& gst2 = n2->goalST();

  double dx = gst2.x() - gst.x();
  double dy = gst2.y() - gst.y();
  double dth = computeMinDiffAngle(gst2.th(),gst.th());
  double dvi = ist2.s() - ist.s();
  double dvg = gst2.s() - gst.s();
  double dt = gst2.t() - gst.t();

  return ((WC_N2R::wX_*fabs(dx)) + 
	  (WC_N2R::wY_*fabs(dy)) + 
	  (WC_N2R::wTH_*fabs(dth)) + 
	  (WC_N2R::wVI_*fabs(dvi)) + 
	  (WC_N2R::wVG_*fabs(dvg)) + 
	  (WC_N2R::wT_*fabs(dt)));
}


                                    /**************************************************/
                                    /***************** Class WC_N2RGrid ***************/
                                    /**************************************************/


/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
WC_N2RGrid::WC_N2RGrid() : N2RGrid()
{
}

/**
 *  \brief init constructor
 */
WC_N2RGrid::WC_N2RGrid(int* sampling,double* minSampS,double* maxSampS) 
  : N2RGrid(6,sampling,minSampS,maxSampS)
{
}


/********* Access/Modification of the grid ********/

/**
 *  \brief return an element of the grid at pos = (i,j,k,l,m,n)
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \return the element of the grid at pos = (i,j,k,l,m,n), NULL if the input coordinates are outside of the grid
 */
N2R* WC_N2RGrid::gridAt(int i,int j,int k,int l,int m,int n) const
{
  int pos[6] = {i,j,k,l,m,n};
  return this->N2RGrid::gridAt(pos);
}

/**
 *  \brief return an element of the grid at pos = (i,j,k,l,m,n)
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \return the element of the grid at pos = (i,j,k,l,m,n), NULL if the input coordinates are outside of the grid
 */
N2R*& WC_N2RGrid::gridAt(int i,int j,int k,int l,int m,int n)
{
  int pos[6] = {i,j,k,l,m,n};
  return this->N2RGrid::gridAt(pos);
}

/**
 *  \brief compute the initial state-time of a cell of the grid
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \param initST : resulting initial state-time
 */
void WC_N2RGrid::computeInitST(int i,int j,int k,int l,int m,int n,
		   WCStateT& initST) const
{
  initST.x() = 0;
  initST.y() = 0;
  initST.th() = 0;
  initST.s() = MIN_SPG_VI + j*(MAX_SPG_VI-MIN_SPG_VI)/((double)(SPG_VI-1));
  initST.om() = 0;
  initST.t() = 0;
}

/**
 *  \brief compute the goal state-times of a cell of the grid
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \param goalST : resulting goal state-time
 */
void WC_N2RGrid::computeGoalST(int i,int j,int k,int l,int m,int n,
		   WCStateT& goalST) const
{
  goalST.x() = MIN_SPG_DX + l*(MAX_SPG_DX-MIN_SPG_DX)/((double)(SPG_DX-1));
  goalST.y() = MIN_SPG_DY + m*(MAX_SPG_DY-MIN_SPG_DY)/((double)(SPG_DY-1));
  goalST.th() =MIN_SPG_DTH + n*(MAX_SPG_DTH-MIN_SPG_DTH)/((double)(SPG_DTH-1));
  goalST.s() = MIN_SPG_VG + k*(MAX_SPG_VG-MIN_SPG_VG)/((double)(SPG_VG-1));
  goalST.om() = 0;
  goalST.t() = MIN_SPG_T + i*(MAX_SPG_T-MIN_SPG_T)/((double)(SPG_T-1));
}

/**
 *  \brief search for reachable nodes on the grid around a reached node (and mark them as reachable on the grid)
 *  \param node : the input already reached node
 *  \return a list containing the reachable nodes on the grid around the input node
 *
 *  An new instance of list is allocated and returned here
 */
std::list<N2R*>* WC_N2RGrid::reachableNodesAround(N2R* node)
{
  std::list<N2R*>* lres = new std::list<N2R*>;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

  int* pos = node->pos();
  
  if(pos != NULL)
    {
      if(pos[0] == 0) indMinT = 0; 
      else indMinT = pos[0]-1;
      if(pos[1] == 0) indMinVI = 0; 
      else indMinVI = pos[1]-1;  
      if(pos[2] == 0) indMinVG = 0; 
      else indMinVG = pos[2]-1;  
      if(pos[3] == 0) indMinDX = 0; 
      else indMinDX = pos[3]-1;  
      if(pos[4] == 0) indMinDY = 0; 
      else indMinDY = pos[4]-1;  
      if(pos[5] == 0) indMinDTH = 0; 
      else indMinDTH = pos[5]-1;  
      
      if(pos[0] == SPG_T-1) indMaxT = SPG_T-1;
      else indMaxT = pos[0]+1;
      if(pos[1] == SPG_VI-1) indMaxVI = SPG_VI-1;
      else indMaxVI = pos[1]+1;  
      if(pos[2] == SPG_VG-1) indMaxVG = SPG_VG-1;
      else indMaxVG = pos[2]+1;  
      if(pos[3] == SPG_DX-1) indMaxDX = SPG_DX-1;
      else indMaxDX = pos[3]+1;  
      if(pos[4] == SPG_DY-1) indMaxDY = SPG_DY-1; 
      else indMaxDY = pos[4]+1;  
      if(pos[5] == SPG_DTH-1) indMaxDTH = SPG_DTH-1; 
      else indMaxDTH = pos[5]+1;
    }
  else
    {
      // search for the nearest node of the grid from the 
      // position of the final state of the current one

      // get back the final state-time of the node
      WCStateT initST = ((WC_N2R*)node)->initST();
      WCStateT goalST = ((WC_N2R*)node)->goalST();
      
      searchForItvSampling(goalST.t(),MIN_SPG_T,MAX_SPG_T,SPG_T,indMinT,indMaxT);
      searchForItvSampling(initST.s(),MIN_SPG_VI,MAX_SPG_VI,SPG_VI,indMinVI,indMaxVI);
      searchForItvSampling(goalST.s(),MIN_SPG_VG,MAX_SPG_VG,SPG_VG,indMinVG,indMaxVG);
      searchForItvSampling(goalST.x(),MIN_SPG_DX,MAX_SPG_DX,SPG_DX,indMinDX,indMaxDX);
      searchForItvSampling(goalST.y(),MIN_SPG_DY,MAX_SPG_DY,SPG_DY,indMinDY,indMaxDY);
      searchForItvSampling(goalST.th(),MIN_SPG_DTH,MAX_SPG_DTH,SPG_DTH,indMinDTH,indMaxDTH);
    }

  for(int i=indMinT; i<=indMaxT; i++)
    for(int j=indMinVI; j<=indMaxVI; j++)
      for(int k=indMinVG; k<=indMaxVG; k++)
	for(int l=indMinDX; l<=indMaxDX; l++)
	  for(int m=indMinDY; m<=indMaxDY; m++)
	    for(int n=indMinDTH; n<=indMaxDTH; n++)
	      {
		// get back the current neighbour node
		N2R* n2r = this->gridAt(i,j,k,l,m,n);
		
		// check if the node at position (i,j,k,l,m,n) has already been created
		if(n2r == NULL)
		  {
		    // create a new node at position (i,j,k,l,m,n)
		    int pos[6] = {i,j,k,l,m,n};
		    n2r = this->addN2Rfrom(pos,node);
		  }

		// add it to the resulting list of neighbours
		lres->push_back(n2r);
	      }
  
  return lres;
}

/**
 *  \brief search for already reached nodes on the grid around a node to reach
 *  \param node : the input node to reach
 *  \return a list containing the already reached nodes on the grid around the input node
 *
 *  An new instance of list is allocated and returned here
 */
std::list<N2R*>* WC_N2RGrid::reachedNodesAround(N2R* node) const
{
//   cout << "reachedNodesAround 1" << endl;
  std::list<N2R*>* lres = new std::list<N2R*>;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

//   cout << "reachedNodesAround 2" << endl;
  int* pos = node->pos();
//   cout << "reachedNodesAround 3" << endl;
  
  if(pos != NULL)   
    {
//       cout << "reachedNodesAround cas 11" << endl;
      if(pos[0] == 0) indMinT = 0; 
      else indMinT = pos[0]-1;
      if(pos[1] == 0) indMinVI = 0; 
      else indMinVI = pos[1]-1;  
      if(pos[2] == 0) indMinVG = 0; 
      else indMinVG = pos[2]-1;  
      if(pos[3] == 0) indMinDX = 0; 
      else indMinDX = pos[3]-1;  
      if(pos[4] == 0) indMinDY = 0; 
      else indMinDY = pos[4]-1;  
      if(pos[5] == 0) indMinDTH = 0; 
      else indMinDTH = pos[5]-1;  

//       cout << "reachedNodesAround cas 12" << endl;      
      if(pos[0] == SPG_T-1) indMaxT = SPG_T-1;
      else indMaxT = pos[0]+1;
      if(pos[1] == SPG_VI-1) indMaxVI = SPG_VI-1;
      else indMaxVI = pos[1]+1;  
      if(pos[2] == SPG_VG-1) indMaxVG = SPG_VG-1;
      else indMaxVG = pos[2]+1;  
      if(pos[3] == SPG_DX-1) indMaxDX = SPG_DX-1;
      else indMaxDX = pos[3]+1;  
      if(pos[4] == SPG_DY-1) indMaxDY = SPG_DY-1; 
      else indMaxDY = pos[4]+1;  
      if(pos[5] == SPG_DTH-1) indMaxDTH = SPG_DTH-1; 
      else indMaxDTH = pos[5]+1;
//       cout << "reachedNodesAround cas 13" << endl;
    }
  else
    {
      cerr << "WC_N2RGrid::reachedNodesAround : Should not arrive here!" << endl;
      exit(1);
    }

//   cout << "reachedNodesAround 4" << endl;
  for(int i=indMinT; i<=indMaxT; i++)
    for(int j=indMinVI; j<=indMaxVI; j++)
      for(int k=indMinVG; k<=indMaxVG; k++)
	for(int l=indMinDX; l<=indMaxDX; l++)
	  for(int m=indMinDY; m<=indMaxDY; m++)
	    for(int n=indMinDTH; n<=indMaxDTH; n++)
	      {
// 		cout << "reachedNodesAround 41" << endl;
		// get back the current neighbour node
		N2R* n2r = this->gridAt(i,j,k,l,m,n);
// 		cout << "reachedNodesAround 42" << endl;
		
		// check if the node at position (i,j,k,l,m,n) has already been created
		if(n2r != NULL && n2r->status() == REACHED_N2R) 
		  {
// 		    cout << "reachedNodesAround 43" << endl;
		    // add it to the resulting list of neighbours
		    lres->push_back(n2r);
// 		    cout << "reachedNodesAround 44" << endl;
		  }
	      }
  
//   cout << "reachedNodesAround 5" << endl;
  return lres;
}

/**
 *  \brief create a N2R from its coordinates and a a node from which it can be derivated
 *  \param coord : coordinates of the N2R
 *  \param node : node from which the N2R may be reachable
 *
 *  An new instance of N2R is allocated here
 *  Put directly the new node in the grid
 */
N2R* WC_N2RGrid::createExploredN2R(int* coord,N2R* node)
{
  // compute the init and goal state-times of the node
  WCStateT initST;
  WCStateT goalST;
  computeInitST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],initST);
  computeGoalST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],goalST);
  
  // create the new node
  WC_N2R* n2r = new WC_N2R(initST,goalST,coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],node);

  // put the new node in the grid
  this->gridAt(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5]) = n2r;

  return n2r;
}

/**
 *  \brief create a N2R from its coordinates and a a node from which it can be derivated
 *  \param coord : coordinates of the N2R
 *  \param node : node from which the N2R may be reachable
 *
 *  An new instance of N2R is allocated here
 *  Both state-times and the control parameterization are copied here
 */
N2R* WC_N2RGrid::createIGN2R(const StateTime* initST,const StateTime* goalST,double* param)
{
  WC_N2R* n2r = new WC_N2R(*((WCStateT*)initST),*((WCStateT*)goalST),param);
  return n2r;
}

/**
 *  \brief create a reached node from an input file thanks to its position and parameters
 *  \param pos : the position of the node in the grid
 *  \param params : the parameterized control defining the trajectory
 *  \return a pointer to the created node
 */
N2R* WC_N2RGrid::createReachedNode(int* coord,double* params)
{
  // compute the initial and goal state-times from their positions in the grid
  WCStateT initST;
  WCStateT goalST;
  this->computeInitST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],initST);
  this->computeGoalST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],goalST);

  // create the node
  N2R* node = new WC_N2R(initST,goalST,coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],params);

  // put the new node in the grid
  this->gridAt(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5]) = node;

  // add the node to the list of reached nodes
  this->listNReached_->push_back(node);

  return node;
}

/**
 *  \brief compute the number of nodes noted as "nodes to reach" on the grid
 *  \return the number of nodes noted as "nodes to reach" on the grid
 */
int WC_N2RGrid::N2RonGrid() const
{
  int res = 0;

  for(int i=0; i<SPG_T; i++)
      for(int k=0; k<SPG_VG; k++)
    for(int j=0; j<SPG_VI; j++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL) res++;
	      }

  return res;
}

/**
 *  \brief compute the number of nodes noted as reached on the grid
 *  \return the number of nodes noted as reached on the grid
 */
int WC_N2RGrid::NReachedonGrid() const
{
  int res = 0;

  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL && node->status() == REACHED_N2R) res++;
	      }

  return res;
}

/**
 *  \brief compute the number of nodes noted as used on the grid
 *  \return the number of nodes noted as used on the grid
 */
int WC_N2RGrid::NUsedonGrid() const
{
  int res = 0;

  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL && node->status() == USED_N2R) res++;
	      }

  return res;
}

/**
 *  \brief display the nodes set as reached on the grid
 */
void WC_N2RGrid::displayReachedOnGrid() const
{
  cout << "Reached nodes on grid : " << endl;
  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL && node->status() == REACHED_N2R)
		  {
		    cout << "(" << i << "," << j << "," << k << "," << l << "," << m << "," << n << ") -> ";
		    displayTab(node->params(),7); cout << endl;
		  }
	      }
}

/**
 *  brief display the sampling of the grid
 */
void WC_N2RGrid::displayGridSampling() const
{
  cout << "T : " << SPG_T << "  [";
  for(int i=0; i<SPG_T-1; i++) cout << (MIN_SPG_T + (double)i * (MAX_SPG_T - MIN_SPG_T) / (SPG_T-1)) << ";";
  cout << MAX_SPG_T << "]" << endl;

  cout << "VI : " << SPG_VI << "  [";
  for(int i=0; i<SPG_VI-1; i++) cout << (MIN_SPG_VI + (double)i * (MAX_SPG_VI - MIN_SPG_VI) / (SPG_VI-1)) << ";";
  cout << MAX_SPG_VI << "]" << endl;

  cout << "VG : " << SPG_VG << "  [";
  for(int i=0; i<SPG_VG-1; i++) cout << (MIN_SPG_VG + (double)i * (MAX_SPG_VG - MIN_SPG_VG) / (SPG_VG-1)) << ";";
  cout << MAX_SPG_VG << "]" << endl;

  cout << "DX : " << SPG_DX << "  [";
  for(int i=0; i<SPG_DX-1; i++) cout << (MIN_SPG_DX + (double)i * (MAX_SPG_DX - MIN_SPG_DX) / (SPG_DX-1)) << ";";
  cout << MAX_SPG_DX << "]" << endl;

  cout << "DY : " << SPG_DY << "  [";
  for(int i=0; i<SPG_DY-1; i++) cout << (MIN_SPG_DY + (double)i * (MAX_SPG_DY - MIN_SPG_DY) / (SPG_DY-1)) << ";";
  cout << MAX_SPG_DY << "]" << endl;

  cout << "DTH : " << SPG_DTH << "  [";
  for(int i=0; i<SPG_DTH-1; i++) cout << (MIN_SPG_DTH + (double)i * (MAX_SPG_DTH - MIN_SPG_DTH) / (SPG_DTH-1)) << ";";
  cout << MAX_SPG_DTH << "]" << endl;
}

// /**
//  *  \brief search for the nearest defined node of the grid to compute a trajectory between both input state-times
//  *  \param is : the input init state-time
//  *  \param gs : the input goal state-time
//  *  \return the nearest defined node of the grid to compute a trajectory between both input state-times
//  */
// N2R* WC_N2RGrid::nearestNodeToComputeTrajBetween(const WCStateT& is,const WCStateT& gs) const
// {
// }

/**
 *  \brief search for the nearest defined node of the grid to compute a trajectory between both input state-times (2nd version)
 *  \param is : the input init state-time
 *  \param gs : the input goal state-time
 *  \return the nearest defined node of the grid to compute a trajectory between both input state-times
 */
N2R* WC_N2RGrid::nearestNodeToComputeTrajBetweenV2(const WCStateT& is,const WCStateT& gs) const
{
  N2R* selectedNode = NULL;

  double dxWRot = gs.x() - is.x();
  double dyWRot = gs.y() - is.y();
  PositionXY posXY(dxWRot,dyWRot);
  posXY.rotate(PositionXY(0,0),-is.th());
  double dx = posXY.x();
  double dy = posXY.y();
  double ddist = sqrt(dx*dx + dy*dy);

  double dth = computeMinDiffAngle(gs.th(),is.th());
  double vi = is.s();
  double vg = gs.s();
  double t = gs.t() - is.t();

//   cout << "dx: " << dx << "  dy:" << dy << "  dth:" << dth << "  vi:" << vi << "  vg:" << vg << "  t:" << t << endl;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

  int* indMin = new int[n_];
  int* indMax = new int[n_];

  searchForItvSampling(t,MIN_SPG_T,MAX_SPG_T,SPG_T,indMin[0],indMax[0]);
  //compute the closest available time to t
  double tIMin = (MIN_SPG_T + ((double)(indMin[0])) * (MAX_SPG_T - MIN_SPG_T) / (SPG_T-1));
  double tIMax = (MIN_SPG_T + ((double)(indMax[0])) * (MAX_SPG_T - MIN_SPG_T) / (SPG_T-1));
  double dtIMin = fabs(t - tIMin);
  double dtIMax = fabs(t - tIMax);
  if(dtIMin < dtIMax) indMax[0] = indMin[0];
  else indMin[0] = indMax[0];

  searchForItvSampling(vi,MIN_SPG_VI,MAX_SPG_VI,SPG_VI,indMin[1],indMax[1]);
  //compute the closest available speed to vi
  double viIMin = (MIN_SPG_VI + ((double)(indMin[1])) * (MAX_SPG_VI - MIN_SPG_VI) / (SPG_VI-1));
  double viIMax = (MIN_SPG_VI + ((double)(indMax[1])) * (MAX_SPG_VI - MIN_SPG_VI) / (SPG_VI-1));
  double dviIMin = fabs(vi - viIMin);
  double dviIMax = fabs(vi - viIMax);
  if(dviIMin < dviIMax) indMax[1] = indMin[1];
  else indMin[1] = indMax[1];

  searchForItvSampling(vg,MIN_SPG_VG,MAX_SPG_VG,SPG_VG,indMin[2],indMax[2]);
  searchForItvSampling(dx,MIN_SPG_DX,MAX_SPG_DX,SPG_DX,indMin[3],indMax[3]);
  searchForItvSampling(dy,MIN_SPG_DY,MAX_SPG_DY,SPG_DY,indMin[4],indMax[4]);
  searchForItvSampling(dth,MIN_SPG_DTH,MAX_SPG_DTH,SPG_DTH,indMin[5],indMax[5]);

//   cout << "indT:[" << indMin[0] << ";" << indMax[0] << "]  ";
//   cout << "indVI:[" << indMin[1] << ";" << indMax[1] << "]  ";
//   cout << "indVG:[" << indMin[2] << ";" << indMax[2] << "]  ";
//   cout << "indDX:[" << indMin[3] << ";" << indMax[3] << "]  ";
//   cout << "indDY:[" << indMin[4] << ";" << indMax[4] << "]  ";
//   cout << "indDTH:[" << indMin[5] << ";" << indMax[5] << "]  ";
//   cout << endl;

  bool trouve = false;
  int maxDist = 0;
  while(!trouve && (maxDist < SPG_T && maxDist < SPG_VI && maxDist < SPG_VG && maxDist < SPG_DX && maxDist < SPG_DY && maxDist < SPG_DTH))
    {
//       cout << "cur maxDist : " << maxDist << endl;

      // compute the lower index, upper index and step of each dimension
      int* minInd = new int[n_];
      int* maxInd = new int[n_];

      for(int i=0; i<n_; i++)
	{
	  if(indMin[i] - maxDist >= 0) minInd[i] = indMin[i] - maxDist;
	  else minInd[i] = 0;
	  if(indMax[i] + maxDist < sampling_[i]) maxInd[i] = indMax[i] + maxDist;
	  else maxInd[i] = sampling_[i]-1;
	}

//       cout << "indT2:[" << minInd[0] << ";" << maxInd[0] << "]  ";
//       cout << "indVI2:[" << minInd[1] << ";" << maxInd[1] << "]  ";
//       cout << "indVG2:[" << minInd[2] << ";" << maxInd[2] << "]  ";
//       cout << "indDX2:[" << minInd[3] << ";" << maxInd[3] << "]  ";
//       cout << "indDY2:[" << minInd[4] << ";" << maxInd[4] << "]  ";
//       cout << "indDTH2:[" << minInd[5] << ";" << maxInd[5] << "]  ";
//       cout << endl;
      
      // compute a list of all reached nodes at a distance maxDist from 
      // the nodes surronding the position of the node to reach in the grid
      std::list<N2R*> lDefinedNodes;
      N2R* bestNode = NULL;
      double bestDist = -1;

      for(int j=minInd[0]; j<maxInd[0]; j++)
	{
	  for(int k=minInd[1]; k<maxInd[1]; k++)
	    {
	      for(int l=minInd[2]; l<maxInd[2]; l++)
		{
		  for(int m=minInd[3]; m<maxInd[3]; m++)
		    {
		      for(int n=minInd[4]; n<maxInd[4]; n++)
			{
			  for(int p=minInd[5]; p<maxInd[5]; p++)
			    {
			      // get back the node defined at the current position if it exists
			      WC_N2R* node = (WC_N2R*)(this->gridAt(j,k,l,m,n,p));
			      
			      // add it to the list of candidates nodes if it has been defined
			      if(node != NULL && node->status() == REACHED_N2R) 
				{
// 				  cout << "candidate node: ";
// 				  cout << "(" << j << "," << k << "," << l << "," << m << "," << n << "," << p << ") ";
// 				  displayTab(node->pos(),6); 
//  				  cout << " -> initST : ";
// 				  node->initST().display();
//  				  cout << " -> goalST : ";
// 				  node->goalST().display();
//  				  cout << endl;

				  double dxN = node->goalST().x() - node->initST().x() - dx;
				  double dyN = node->goalST().y() - node->initST().y() - dy;
				  double ddistN = sqrt(dxN*dxN + dyN*dyN);
				  double dthN = computeMinDiffAngle(gs.th(),is.th()) - dth;
				  double dviN = node->initST().s() - vi;
				  double dvgN = node->goalST().s() - vg;
				  double dtN = node->goalST().t() - t;
				  
				  double dist = WC_N2R::wX_*fabs(dxN) 
				    + WC_N2R::wY_*fabs(dyN) 
				    + WC_N2R::wTH_*fabs(dthN) 
				    + WC_N2R::wVI_*fabs(dviN) 
				    + WC_N2R::wVG_*fabs(dvgN) 
				    + WC_N2R::wT_*fabs(dtN);

// 				  double dist = ddistN + fabs(dthN) + fabs(dviN*t) + fabs(dvgN*t) + fabs(0.5*dtN*(vi+vg));

// 				  cout << "dxN:" << dxN << "  dyN:" << dyN << "  ddistN:" << ddistN << "  dthN:" << dthN 
// 				       << "  dviN:" << dviN << "  dvgN:" << dvgN << "  dtN:" << dtN;
// 				  cout << "  dist : " << dist << endl;

				  if(bestNode == NULL || dist < bestDist)
				    {
				      bestNode = node;
				      bestDist = dist;
				    }
				}
			    }
			}
		    }
		}
	    }
	}      
      
      delete minInd;
      delete maxInd;

      // check if at least one defined node has been found
      if(bestNode != NULL)
	{
// 	  cout << "maxDist : " << maxDist << endl;
// 	  cout << "bestDist:" << bestDist << "\t";

	  selectedNode = bestNode;
	  trouve = true;
	}
      else
	{
	  maxDist++;
	}
    }

  delete [] indMin;
  delete [] indMax;

//   if(selectedNode != NULL) 
//     { 
//       cout << "selected node : "; 
//       displayTab(selectedNode->pos(),6); 
//       cout << " -> initST : ";
//       selectedNode->initST().display();
//       cout << " -> goalST : ";
//       selectedNode->goalST().display();
//       cout << endl;
//     }
//   else cout << "No node found in the grid..." << endl;
  
  return selectedNode;
}



                                    /**************************************************/
                                    /*************** Class WCInitialGuess *************/
                                    /**************************************************/


/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
WCInitialGuess::WCInitialGuess()
{
  firstIGMode_ = SCP_FIRSTIG; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  this->dimS_ = 6;
  this->dimP_ = 7;

  this->dimSampS_ = 6;

  this->sampling_ = new int [dimSampS_];
  this->minSampS_ = new double [dimSampS_];
  this->maxSampS_ = new double [dimSampS_];
  for(int i=0; i<dimSampS_; i++) sampling_[i] = 0;
  for(int i=0; i<dimSampS_; i++) minSampS_[i] = 0;
  for(int i=0; i<dimSampS_; i++) maxSampS_[i] = 0;

//   this->spg_ = NULL;
  this->grid_ = NULL;

  this->SC_AccSp_ = 8;
  this->SC_AAccSp_ = 8;
}

/**
 *  \brief init constructor
 */  
WCInitialGuess::WCInitialGuess(WheelChairSys* sys) : InitialGuess(sys)
{
  firstIGMode_ = SCP_FIRSTIG; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  this->dimS_ = 6;
  this->dimP_ = 7;

  this->dimSampS_ = 6;

  this->sampling_ = new int [dimSampS_];
  this->minSampS_ = new double [dimSampS_];
  this->maxSampS_ = new double [dimSampS_];
  this->setDfltSampS();
  this->setDfltMinMaxSampS();

  this->trajGen_ = NULL; // ... Should be changed once the trajgen class modified ...  = new CLTrajGen(sys)

//   this->spg_ = NULL;
  this->initGrid();

  this->SC_AccSp_ = 8;
  this->SC_AAccSp_ = 8;
}

/**
 *  \brief init constructor
 *  \param sys : the wheelchair features
 *  \param tmin : the minimal time
 *  \param tmax : the maximal time
 *  \param spg_t : the time sampling
 *  \param spg_vi : the initial speed sampling
 *  \param spg_vg : the goal speed sampling
 *  \param spg_dx : the position sampling
 *  \param spg_dy : the position sampling
 *  \param spg_dth : the heading sampling
 */
WCInitialGuess::WCInitialGuess(WheelChairSys* sys,
			       int tmin,int tmax,
			       int spg_t,int spg_vi,int spg_vg,int spg_dx,int spg_dy,int spg_dth) : InitialGuess(sys)
{
  firstIGMode_ = SCP_FIRSTIG; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  this->dimS_ = 6;
  this->dimP_ = 7;

  this->dimSampS_ = 6;

  this->sampling_ = new int [dimSampS_];
  this->minSampS_ = new double [dimSampS_];
  this->maxSampS_ = new double [dimSampS_];
  this->initSampS(spg_t,spg_vi,spg_vg,spg_dx,spg_dy,spg_dth);
  this->setDfltMinMaxSampS();
  MIN_SPG_T = tmin;
  MAX_SPG_T = tmax;

  this->trajGen_ = NULL; // ... Should be changed once the trajgen class modified ...  = new CLTrajGen(sys)

//   this->spg_ = NULL;
  this->initGrid();

  this->SC_AccSp_ = 8;
  this->SC_AAccSp_ = 8;
}

/**
 *  \brief destructor
 */  
WCInitialGuess::~WCInitialGuess()
{
  this->clearGrid();
}


/*************** Getters and Setters **************/

/**
 *  \brief use the sampled control mode to determine the first initial guess 
 *  \param sampAcc : number of samples of the acceleration control
 *  \param sampAAcc : number of samples of the angular acceleration control
 */
void WCInitialGuess::use_SC_FIGMode(double sampAcc,double sampAAcc)
{
  this->firstIGMode_ = SCP_FIRSTIG;

  this->SC_AccSp_ = sampAcc;
  this->SC_AAccSp_ = sampAAcc;
}


/***************** Initialization *****************/

/**
 *  \brief initialize the sampling of the state space
 *  \param spg_t : sampling over the time
 *  \param spg_vi : sampling over the initial speed
 *  \param spg_vg : sampling over the goal speed
 *  \param spg_dx : sampling over the position x
 *  \param spg_dy : sampling over the position y
 *  \param spg_dth : sampling over the heading
 */
void WCInitialGuess::initSampS(int spg_t,int spg_vi,int spg_vg,int spg_dx,int spg_dy,int spg_dth)
{
  cout << "WCInitialGuess::initSampS" << endl;

  SPG_T = spg_t;
  SPG_VI = spg_vi;
  SPG_VG = spg_vg;
  SPG_DX = spg_dy;
  SPG_DY = spg_dx;
  SPG_DTH = spg_dth;

  cout << "SPG_T : " << SPG_T << endl;
  cout << "SPG_VI : " << SPG_VI << endl;
  cout << "SPG_VG : " << SPG_VG << endl;
  cout << "SPG_DX : " << SPG_DX << endl;
  cout << "SPG_DY : " << SPG_DY << endl;
  cout << "SPG_DTH : " << SPG_DTH << endl;
}

/**
 *  \brief set the default state sampling parameters
 */
void WCInitialGuess::setDfltSampS()
{
  SPG_T = 25;
  SPG_VI = 9;
  SPG_VG = 9;
  SPG_DX = 16;
  SPG_DY = 16;
  SPG_DTH = 8;

//   SPG_T = 9;
//   SPG_VI = 9;
//   SPG_VG = 9;
//   SPG_DX = 20;
//   SPG_DY = 20;
//   SPG_DTH = 12;
}

/**
 *  \brief set the default state sampling min/max values
 */
void WCInitialGuess::setDfltMinMaxSampS()
{
  MIN_SPG_T = 0.1;
  if(sys()->canMoveBack())
    {
      MIN_SPG_VI = -sys()->getVMax();
      MIN_SPG_VG = -sys()->getVMax();
    }
  else 
    {
      MIN_SPG_VI = 0;
      MIN_SPG_VG = 0;
    }

  MAX_SPG_T = 5.0;
  MAX_SPG_VI = sys()->getVMax();
  MAX_SPG_VG = sys()->getVMax();

  MIN_SPG_DX = -MAX_SPG_T * sys()->getVMax();
  MIN_SPG_DY = -MAX_SPG_T * sys()->getVMax();
  MIN_SPG_DTH = -M_PI;

  MAX_SPG_DX = MAX_SPG_T * sys()->getVMax();
  MAX_SPG_DY = MAX_SPG_T * sys()->getVMax();
  MAX_SPG_DTH = M_PI;
}

/**
 *  \brief initialize the sampling grid
 */
void WCInitialGuess::initGrid()
{
  // delete the old grid if needed
  this->clearGrid();

  // create the new grid with the actual sampling parameters
  this->grid_ = new WC_N2RGrid(sampling_,minSampS_,maxSampS_);
}


/******************** Learning ********************/

/**
 *  \brief compute a random initial guess for a sample of the state space
 *  \param ig : the resulting initial guess (must be allocated)
 *  \param t : time the control generated from the set of parameters will last
 */
void WCInitialGuess::computeRandFirstIG(double* ig,double t)
{
  cerr << "WCInitialGuess::computeRandFirstIG : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute random initial guesses for all samples of the state space
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* WCInitialGuess::computeRandFirstIG()
{
  cerr << "WCInitialGuess::computeRandFirstIG : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute a regular sampled control to determine the first initial guess
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* WCInitialGuess::computeSCFirstIG()
{
  cerr << "WCInitialGuess::computeSCFirstIG : Cannot work!" << endl;
  exit(1);

  std::list<N2R*>* lfig = new std::list<N2R*>;

  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minAAcc = -sys()->getAAMax();
  double maxAAcc = sys()->getAAMax();

  cout << "minAcc : " << minAcc << endl;
  cout << "maxAcc : " << maxAcc << endl;
  cout << "minAAcc : " << minAAcc << endl;
  cout << "maxAAcc : " << maxAAcc << endl;

  cout << "SPG_T : " << SPG_T << endl;
  cout << "SPG_VI : " << SPG_VI << endl;
  cout << "SC_AccSp_ : " << SC_AccSp_ << endl;
  cout << "SC_AAccSp_ : " << SC_AAccSp_ << endl;


  // compute the set of final times
  double* timeSp = new double[SPG_T];
  for(int i=0; i<SPG_T; i++) timeSp[i] = MIN_SPG_T + (((double)i) * (MAX_SPG_T - MIN_SPG_T)/((double)(SPG_T-1)));

  // compute the set of initial velocity
  double* ivelSp = new double[SPG_VI];
  for(int j=0; j<SPG_VI; j++) ivelSp[j] = MIN_SPG_VI + (((double)j) * (MAX_SPG_VI - MIN_SPG_VI)/((double)(SPG_VI-1)));
  
  // compute the set of available acceleration controls
  double* accSp = new double[SC_AccSp_];
  for(int k=0; k<SC_AccSp_; k++) accSp[k] = minAcc + (((double)k) * (maxAcc - minAcc)/((double)(SC_AccSp_-1)));

  // compute the set of available steering speed controls
  double* aaccSp = new double[SC_AAccSp_];
  for(int l=0; l<SC_AAccSp_; l++) aaccSp[l] = minAAcc + (((double)l) * (maxAAcc - minAAcc)/((double)(SC_AAccSp_-1)));

  // add an IG for each 4-uplet of (initial velocity, acceleration, steering speed control, time)
  for(int j=0; j<SPG_VI; j++)
    {
      double ivel = ivelSp[j];
	  
      for(int k=0; k<SC_AccSp_; k++)
	{
	  double acc = accSp[k];
	      
	  for(int l=0; l<SC_AAccSp_; l++)
	    {
	      double aacc = aaccSp[l];

	      // compute the initial state-time
	      WCStateT initST(0,0,0,ivel,0,0);
		  
	      // compute the control to apply
	      WCControlT ctrlt(acc,aacc,MAX_SPG_T);

	      cout << "ctrlt : "; ctrlt.display(); 

	      // generate a trajectory from initST by applying the control ctrl
	      WCTrajGen trajgen(sys());

	      trajgen.setUseVelCst(USE_VEL_CST);
	      trajgen.setUseAccCst(USE_ACC_CST);
	      trajgen.setUseAVelCst(USE_AVEL_CST);
	      trajgen.setUseAAccCst(USE_AACC_CST);

	      trajgen.setDBG(0);
	      trajgen.ComputeFromCtrl(initST,ctrlt);
	      
	      // get the parameters of the trajectory generator corresponding to the control
	      double params[7];
	      trajgen.finalParams(params);

	      cout << "  params : "; displayTab(params,7); cout << endl;

	      // get back the state reached at each time sample 
	      for(int i=0; i<SPG_T; i++)
		{
		  double t = timeSp[i];

		  WCStateT finalST = trajgen.getWCStateTReachedAt(t);

		  // create the node corresponding to this sample
		  WC_N2R* newNodeIG = new WC_N2R(initST,finalST,params);

		  // append it to the list of nodes
		  lfig->push_back(newNodeIG);
		}
	    }
	}
    }

  cout << "//////////////////////////////////////////////////////////////////" << endl;

  // delete the sampling vectors
  delete [] timeSp;
  delete [] ivelSp;
  delete [] accSp;
  delete [] aaccSp;

  return lfig;
}

// /**
//  *  \brief fake initialization
//  *  \return the resulting list of first initial guesses
//  */
// std::list<N2R*>* WCInitialGuess::computeSCPFirstIG()
// {
//  std::list<N2R*>* lfig = new std::list<N2R*>;

//  WCStateT initST(0,0,0,0,0,0);
//  WCStateT goalST(1,1,0,0,0,5);
//  double curParams[7] = {0,0,0,1,0,0,5};

//  // create the node corresponding to this sample
//  WC_N2R* newNodeIG = new WC_N2R(initST,goalST,curParams);
 
//  // append it to the list of nodes
//  lfig->push_back(newNodeIG);

//  return lfig;
// }

/**
 *  \brief compute a regular sampled parameterization of the control to determine the first initial guess
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* WCInitialGuess::computeSCPFirstIG()
{
  std::list<N2R*>* lfig = new std::list<N2R*>;

  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minAAcc = -sys()->getAAMax();
  double maxAAcc = sys()->getAAMax();
  double minJ = (-sys()->getAMax())/((double)MAX_SPG_T);
  double maxJ = (sys()->getAMax())/((double)MAX_SPG_T);

  cout << "minAcc : " << minAcc << endl;
  cout << "maxAcc : " << maxAcc << endl;
  cout << "minAAcc : " << minAAcc << endl;
  cout << "maxAAcc : " << maxAAcc << endl;

  cout << "SPG_T : " << SPG_T << endl;
  cout << "SPG_VI : " << SPG_VI << endl;
  cout << "SC_AccSp_ : " << SC_AccSp_ << endl;
  cout << "SC_AAccSp_ : " << SC_AAccSp_ << endl;

  // compute the set of final times
  double* timeSp = new double[SPG_T];
  for(int i=0; i<SPG_T; i++) timeSp[i] = MIN_SPG_T + (((double)i) * (MAX_SPG_T - MIN_SPG_T)/((double)(SPG_T-1)));

  // compute the set of initial velocity
  double* ivelSp = new double[SPG_VI];
  for(int j=0; j<SPG_VI; j++) ivelSp[j] = MIN_SPG_VI + (((double)j) * (MAX_SPG_VI - MIN_SPG_VI)/((double)(SPG_VI-1)));
  
  // compute the set of available acceleration controls
  double* accSp = new double[SC_AccSp_];
  for(int k=0; k<SC_AccSp_; k++) accSp[k] = minAcc + (((double)k) * (maxAcc - minAcc)/((double)(SC_AccSp_-1)));

  // compute the set of available steering speed controls
  double* aaccSp = new double[SC_AAccSp_];
  for(int l=0; l<SC_AAccSp_; l++) aaccSp[l] = minAAcc + (((double)l) * (maxAAcc - minAAcc)/((double)(SC_AAccSp_-1)));

  // add an IG for each 4-uplet of (initial velocity, acceleration, steering speed control, time)
  for(int j=0; j<SPG_VI; j++)
    {
      double ivel = ivelSp[j];
	  
      for(int k=0; k<SC_AccSp_; k++)
	{
	  double acc = accSp[k];
	      
	  double jMin = 2.0*((double)k)/((double)(SC_AccSp_-1))*(minAcc/((double)MAX_SPG_T));
	  double jMax = 2.0*((double)(SC_AccSp_-1-k))/((double)(SC_AccSp_-1))*(maxAcc/((double)MAX_SPG_T));

// 	  cout << "acc : " << acc << "  jMin : " << jMin << "  jMax : " << jMax << endl;

	  for(int m=0; m<SC_AccSp_; m++)
	    {
	      double jerk = jMin + ((double)m)/((double)(SC_AccSp_-1)) * (jMax-jMin);

	      for(int l=0; l<SC_AAccSp_; l++)
		{
		  double aacc = aaccSp[l];

		  double ajMin = 3.0*((double)l)/((double)(SC_AAccSp_-1))*(minAAcc/((double)MAX_SPG_T));
		  double ajMax = 3.0*((double)(SC_AAccSp_-1-l))/((double)(SC_AAccSp_-1))*(maxAAcc/((double)MAX_SPG_T));
	      
// 		  cout << "aacc : " << aacc << "  ajMin : " << ajMin << "  ajMax : " << ajMax << endl;
	      
		  for(int n=0; n<SC_AAccSp_; n++)
		    {
		      double ajerk = ajMin + ((double)n)/((double)(SC_AAccSp_-1)) * (ajMax-ajMin);

// 		      cout << "ajerk : " << ajerk << endl;

		      // compute the initial state-time
		      WCStateT initST(0,0,0,ivel,0,0);
		  
		      // compute the parameterized control to apply
		      double params[7];
		      params[0] = aacc;
		      params[1] = ajerk;
		      params[2] = 0;
		      params[3] = acc;
		      params[4] = jerk;
		      params[5] = 0;
		      params[6] = MAX_SPG_T;

		      //WCControlT ctrlt(acc,aacc,MAX_SPG_T);

		      // 	      cout << "initST : "; initST.display();
		      // 	      cout << "  params : "; displayTab(params,7);

		      // generate a trajectory from initST by applying the control ctrl
		      WCTrajGen trajgen(sys());
		      trajgen.setDBG(0);

		      trajgen.setUseVelCst(USE_VEL_CST);
		      trajgen.setUseAccCst(USE_ACC_CST);
		      trajgen.setUseAVelCst(USE_AVEL_CST);
		      trajgen.setUseAAccCst(USE_AACC_CST);

		      //trajgen.ComputeFromCtrl(initST,ctrlt);
		      trajgen.ComputeFromIG(initST,params);

		      // get back the state reached at each time sample 
		      for(int i=0; i<SPG_T; i++)
			{
			  double t = timeSp[i];

			  if(fabs(trajgen.funcTheta(t)) < 1.25*M_PI)
			    {
			      WCStateT finalST = trajgen.getWCStateTReachedAt(t);
			      
// 		  if(i == SPG_T-1) { cout << "  finalST : "; finalST.display(); cout << endl; }
			      
			      double curParams[7];
// 			      curParams[0] = params[0];
// 			      curParams[1] = params[1];
// 			      curParams[2] = params[2];
// 			      curParams[3] = params[3];
// 			      curParams[4] = params[4];
// 			      curParams[5] = params[5];
			      curParams[0] = aacc;
			      curParams[1] = ajerk;
			      curParams[2] = 0;
			      curParams[3] = acc;
			      curParams[4] = jerk;
			      curParams[5] = 0;
			      curParams[6] = t;
			      
			      // create the node corresponding to this sample
			      WC_N2R* newNodeIG = new WC_N2R(initST,finalST,curParams);
			      
			      // append it to the list of nodes
			      lfig->push_back(newNodeIG);
			    }
			}
		    }
		}
	    }
	}
    }

//   cout << "//////////////////////////////////////////////////////////////////" << endl;

  // delete the sampling vectors
  delete [] timeSp;
  delete [] ivelSp;
  delete [] accSp;
  delete [] aaccSp;

  return lfig;
}

/**
 *  \brief compute a  to determine the first initial guess
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* WCInitialGuess::computeRStarFirstIG()
{
  cerr << "WCInitialGuess::computeRStarFirstIG : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief load the first initial guess from a file
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* WCInitialGuess::loadFirstIGfromFile()
{
  cerr << "WCInitialGuess::loadFirstIGfromFile : Implementation in progress!" << endl;
  exit(1);
}

/**
 *  \brief try to generate a trajectory to the input node n2r thanks to the parameters of the node near
 *  \param n2r : the input node to reach
 *  \param near : the nearest node chosen as initial guess to converge to n2r
 *  \param param : the resulting parametrized control used to reach the goal defined if the trajectory generation succeed
 *  \return 0 if the node n2r has been reached, -1 else
 *
 *  The node n2r if updated if it has been reached
 */
int WCInitialGuess::genTrajTo(N2R* n2r,N2R* near,double* param,bool argh) const
{
  // get back the initial and final states between which a trajectory must be compted
  WCStateT initST = ((WC_N2R*)n2r)->initST();
  WCStateT goalST = ((WC_N2R*)n2r)->goalST();
  double t_f = goalST.t();

//   cout << "initST : "; initST.display(); cout << "  goalST : "; goalST.display(); cout << endl;
//   cout << endl;
  
  // initialize the table of parameters
  const double* paramsNear = near->params();
  double* params = new double[7];
  params[0] = paramsNear[0];
  params[1] = paramsNear[1];
  params[2] = paramsNear[2];
  params[3] = paramsNear[3];
  params[4] = paramsNear[4];
  params[5] = paramsNear[5];
  params[6] = t_f;
//   params[6] = paramsNear[6];
  
//   cout << "params : "; displayTab(params,7); cout << endl;

  WCTrajGen trajgen(sys());

  trajgen.setUseVelCst(USE_VEL_CST);
  trajgen.setUseAccCst(USE_ACC_CST);
  trajgen.setUseAVelCst(USE_AVEL_CST);
  trajgen.setUseAAccCst(USE_AACC_CST);
  
  trajgen.fixFinalXY(FIXED_XY);
  trajgen.fixFinalTH(FIXED_TH);
  trajgen.fixFinalV(FIXED_V);
  trajgen.fixFinalOM(FIXED_AV);

  if(FIXED_T) trajgen.setTcStatus(TIME_FIXED,0,999999999);
  else trajgen.setTcStatus(TIME_FREE,0,999999999);

  if(!argh) trajgen.setDBG(0);
  else trajgen.setDBG(1);
  trajgen.setMaxNbStep(maxTGSteps_);
  trajgen.setMaxResidual(maxTGResidual_);
//   trajgen.setTcStatus(TIME_FIXED,0,999999999);

  int res = trajgen.ComputeTrajectory(initST,goalST,&params);
//   int res = 0;
  if(fabs(trajgen.funcTheta(t_f)) >= 2.0*M_PI) res = -1;

  if(res == 0)
    {
      double finalParams[7];
      trajgen.finalParams(finalParams);
      
      param[0] = finalParams[0];
      param[1] = finalParams[1];
      param[2] = finalParams[2];
      param[3] = finalParams[3];
      param[4] = finalParams[4];
      param[5] = finalParams[5];
      param[6] = finalParams[6];
//       param[0] = 0;
//       param[1] = 0;
//       param[2] = 0;
//       param[3] = 0;
//       param[4] = 0;
//       param[5] = 0;
//       param[6] = 0;
    }

  delete [] params;
  return res;
}


/********************* Query **********************/

/**
 *  \brief query the initial guess base
 *  \param is : the initial state-time
 *  \param is : the goal state-time
 *  \return the best set of parameters found
 */
const double* WCInitialGuess::queryIG(const WCStateT& is,const WCStateT& gs) const
{
  // search for the nearest defined node of the grid to compute a trajectory between both input state-times
  N2R* node = ((WC_N2RGrid*)grid_)->nearestNodeToComputeTrajBetweenV2(is,gs);
  
  // check if a node has been defined at the selected position
  if(node != NULL)
    {
      cout << "node found" << endl;
      cout << "selected params : "; displayTab(node->params(),7); 
      cout << "  initST: "; node->initST().display();
      cout << "  goalST: "; node->goalST().display();
      cout << endl;
      return node->params();
    }
  else 
    {
      cout << "no node found" << endl;
      return NULL;
    }
}


/*************** Utility Functions ****************/
  
/**
 *  \brief sample the trajectory defined on the node into tables of position and time
 *  \param node : the input node containing a trajectory
 *  \param tabX : the resulting table containing the position X values
 *  \param tabY : the resulting table containing the position Y values
 *  \param tabT : the resulting table containing the time T values
 *  \param n : the number of elements defined into tables
 *
 *  Tables must be allocated first
 */
void WCInitialGuess::sampleTrajNode(N2R* node,double* tabX,double* tabY,double *tabT,int n) const
{
  // get back the initial and final state-times of the trajectory
  WCStateT s0 = ((WC_N2R*)node)->initST();
  WCStateT sf = ((WC_N2R*)node)->goalST();

  // get back the parameterized control of the trajectory
  const double* param = node->params();

  // generate the trajectory from the initial state and the parameterized control
  WCTrajGen trajgen(sys());

  trajgen.setUseVelCst(USE_VEL_CST);
  trajgen.setUseAccCst(USE_ACC_CST);
  trajgen.setUseAVelCst(USE_AVEL_CST);
  trajgen.setUseAAccCst(USE_AACC_CST);

  trajgen.setDBG(0);
  trajgen.ComputeFromIG(s0,param);

//   WCStateT finalST = trajgen.getFinalWCStateTReached();
//   cout << "WCInitialGuess::sampleTrajNode : final state of the traj : "; finalST.display(); cout << endl;
  
  // sample the trajectory generated
  trajgen.SaveTrajectory(n);

  // copy the sampled trajectory into the resulting tables
  for(int i=0; i<n; i++)
    {
      tabX[i] = trajgen.posX->at(i);
      tabY[i] = trajgen.posY->at(i);
      tabT[i] = trajgen.timeT->at(i);
    }
}

/**
 *  \brief list all nodes of the grid
 *  \return a pointer to a vector containing all nodes of the grid
 */
std::vector<N2R*>* WCInitialGuess::listAllNodes() const
{
  int N = computeNbTargetST();
  std::vector<N2R*>* nodes = new std::vector<N2R*>(N);

//   int p=0;

//   for(int i=0; i<SPG_T; i++)
//     for(int j=0; j<SPG_VI; j++)
//       for(int k=0; k<SPG_VG; k++)
// 	for(int l=0; l<SPG_DX; l++)
// 	  for(int m=0; m<SPG_DY; m++)
// 	    for(int n=0; n<SPG_DTH; n++)
// 	      {
// 		(*nodes)[p] = spg_[i][j][k][l][m][n];
// 		p++;
// 	      }

  cerr << "WCInitialGuess::listAllNodes : Not correctly implemented!" << endl;
  exit(1);

  return nodes;
}

/**
 *  \brief list all reached nodes of the grid
 *  \return a pointer to a vector containing all reached nodes of the grid
 */
std::vector<N2R*>* WCInitialGuess::listAllReachedNodes() const
{
  int N = computeNbTargetST();
  std::vector<N2R*>* nodes = new std::vector<N2R*>(N);

  int p=0;

//   for(int i=0; i<SPG_T; i++)
//     for(int j=0; j<SPG_VI; j++)
//       for(int k=0; k<SPG_VG; k++)
// 	for(int l=0; l<SPG_DX; l++)
// 	  for(int m=0; m<SPG_DY; m++)
// 	    for(int n=0; n<SPG_DTH; n++)
// 	      {
// 		// add the node to the vector only if it has been reached
// 		if(spg_[i][j][k][l][m][n]->isReached())
// 		  {
// 		    (*nodes)[p] = spg_[i][j][k][l][m][n];
// 		    p++;
// 		  }
// 	      }

  cerr << "WCInitialGuess::listAllReachedNodes : Not correctly implemented!" << endl;
  exit(1);

  nodes->resize(p);
  return nodes;
}

// /*
//  *  \brief display in a gnuplot window the status of each node of the grid
//  */
// void WCInitialGuess::drawGPGrid()
// {
//   // get back the lists of explored, reached and used nodes
//   std::list<N2R*> lN2R = grid_->listN2R();
//   std::list<N2R*> lNReached = grid_->listNReached();
//   std::list<N2R*> lNUsed = grid_->listNUsed();

// //   cout << "Nodes reached : " << endl;
// //   grid_->displayLReached();

//   int N_N2R = lN2R.size();
//   std::vector<double> vecN2RX(N_N2R);
//   std::vector<double> vecN2RY(N_N2R);
//   std::vector<double> vecN2RT(N_N2R);

//   std::list<N2R*>::iterator it1;
//   int i = 0;
//   for(it1=lN2R.begin(); it1 != lN2R.end(); it1++)
//     {
//       PositionXYT pos = (*it1)->goalST().posXYT();
//       vecN2RX[i] = pos.x();
//       vecN2RY[i] = pos.y();
//       vecN2RT[i] = pos.t();
//       i++;
//     }

//   int N_Reached = lNReached.size();
//   std::vector<double> vecNReachedX(N_Reached);
//   std::vector<double> vecNReachedY(N_Reached);
//   std::vector<double> vecNReachedT(N_Reached);

//   std::list<N2R*>::iterator it2;
//   int j = 0;
//   for(it2=lNReached.begin(); it2 != lNReached.end(); it2++)
//     {
//       PositionXYT pos = (*it2)->goalST().posXYT();
//       vecNReachedX[j] = pos.x();
//       vecNReachedY[j] = pos.y();
//       vecNReachedT[j] = pos.t();
//       j++;
//     }

//   int N_Used = lNUsed.size();
//   std::vector<double> vecNUsedX(N_Used);
//   std::vector<double> vecNUsedY(N_Used);
//   std::vector<double> vecNUsedT(N_Used);

//   std::list<N2R*>::iterator it3;
//   int k = 0;
//   for(it3=lNUsed.begin(); it3 != lNUsed.end(); it3++)
//     {
//       PositionXYT pos = (*it3)->goalST().posXYT();
//       vecNReachedX[k] = pos.x();
//       vecNReachedY[k] = pos.y();
//       vecNReachedT[k] = pos.t();
//       k++;
//     }


//   int NbP = 40;

//   // reinit the tables containing the coordinates of trajectories
//   if(tabX_ != NULL) delete tabX_;
//   if(tabY_ != NULL) delete tabY_;
//   if(tabV_ != NULL) delete tabV_;
//   if(tabT_ != NULL) delete tabT_;
//   tabX_ = new vector<double>;
//   tabY_ = new vector<double>;
//   tabV_ = new vector<double>;
//   tabT_ = new vector<double>;
  
//   // get back the list of reached nodes
//   std::list<N2R*> lReached = grid_->listNReached();

//   // sample all generated trajectories
//   int N = lReached.size();
//   std::list<N2R*>::iterator it = lReached.begin();

//   for(int i=0; i<N; i++)
//     {
//       // check if a trajectory has been found between the state-times of this sample
//       if(*it != NULL && (*it)->params() != NULL)
// 	{
// 	  // hack to display only a part of the trajectories
// 	  WCStateT initST = ((WC_N2R*)(*it))->initST();
// 	  WCStateT goalST = ((WC_N2R*)(*it))->goalST();
// 	  double dx = goalST.x() - initST.x();
// 	  double dy = goalST.y() - initST.y();
// 	  double dist = sqrt(dx*dx+dy*dy);

// 	  if(initST.s() == 0 && initST.om() == 0 && goalST.x() >= -15 && goalST.s() == 0 && goalST.om() == 0 && goalST.t() == 5 && dist >= 20 && i%4 == 0)
// 	    {
// 	      // sample the trajectory defined by the parameterized control
// 	      double tabX[NbP];
// 	      double tabY[NbP];
// 	      double tabT[NbP];
// 	      this->sampleTrajNode(*it,tabX,tabY,tabT,NbP);

// 	      // 	  cout << "last node of the traj : (" << tabX[NbP-1] << "," << tabY[NbP-1] << "," << tabT[NbP-1] << ")" << endl;

// 	      // append these tables to the list of all points of the trajectories
// 	      for(int j=0; j<NbP; j++)
// 		{
// 		  tabX_->push_back(tabX[j]);
// 		  tabY_->push_back(tabY[j]);
// 		  tabT_->push_back(tabT[j]);
// 		}
// 	      for(int j=NbP-1; j>=0; j--)
// 		{
// 		  tabX_->push_back(tabX[j]);
// 		  tabY_->push_back(tabY[j]);
// 		  tabT_->push_back(tabT[j]);
// 		}
// 	    }
// 	}

//       it++;
//     }

//   Gnuplot gPosXYT;
//   gPosXYT.reset_plot();
//   gPosXYT.cmd("set linestyle 1");
//   gPosXYT.set_style("lines").plot_xyz(*tabX_,*tabY_,*tabT_,"Trajectories (XxYxT)");
//   if(N_N2R != 0) gPosXYT.set_style("points").plot_xyz(vecN2RX,vecN2RY,vecN2RT,"Nodes to reach");
//   if(N_Reached != 0) gPosXYT.set_style("points").plot_xyz(vecNReachedX,vecNReachedY,vecNReachedT,"Reached nodes");
// //   if(N_Used != 0) gPosXYT.set_style("points").plot_xyz(vecNUsedX,vecNUsedY,vecNUsedT,"Used nodes");

//   wait_for_key();
// }
