#ifndef FXGLDYNCANVAS_HPP
#define FXGLDYNCANVAS_HPP

#include "fx.h"
#include "fx3d.h"

#include <GL/gl.h>
#include <GL/glu.h>

namespace FX {

class FXAPI FXGLDynCanvas : public FXGLCanvas
{
  FXDECLARE(FXGLDynCanvas)
  friend class FXGLObject;

protected : 
  bool leftButton_,rightButton_;
  FXVec3d center_, eye_, up_;
  FXVec2d oldPos_;

  FXObject* mainWin_;
  FXSelector mainWinSel_;

protected : 
  /********** Constructors and destructors **********/
  FXGLDynCanvas();                                    // default constructor

public : 
  /********** Constructors and destructors **********/
  FXGLDynCanvas(FXComposite* p,FXGLVisual *vis,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,FXint x=0,FXint y=0,FXint w=0,FXint h=0);                                        // init constructor
  FXGLDynCanvas(FXComposite* p,FXGLVisual *vis,FXGLCanvas* sharegroup,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,FXint x=0,FXint y=0,FXint w=0,FXint h=0);                 // init constructor
  virtual ~FXGLDynCanvas();                           // destructor

  void init(double e_x,double e_y,double e_z,double c_x,double c_y,double c_z,double u_x,double u_y,double u_z);

  /******************* Accessors ********************/
  double getEyeX() const;
  double getEyeY() const;
  double getEyeZ() const;
  double getCenterX() const;
  double getCenterY() const;
  double getCenterZ() const;
  double getUpX() const;
  double getUpY() const;
  double getUpZ() const;

  void setMainWin(FXObject* mw,FXSelector sel);

  /**************** Utility functions ***************/
  double distEyeCenter() const;
  FXVec3d toWorldVec(double x,double y) const;

  /**************** Signal functions ****************/
  long onMotion(FXObject*,FXSelector,void*);
  long onMouseWheel(FXObject*,FXSelector,void*);
  long onLeftBtnPress(FXObject*,FXSelector,void*);
  long onLeftBtnRelease(FXObject*,FXSelector,void*);
  long onRightBtnPress(FXObject*,FXSelector,void*);
  long onRightBtnRelease(FXObject*,FXSelector,void*);

};

}
#endif
