#include "FXGLDynCanvas.hpp"

#include <iostream>
using namespace std;

using namespace FX;

namespace FX {

FXDEFMAP(FXGLDynCanvas) FXGLDynCanvasMap[]={
  //________Message_Type_____________________ID_______________Message_Handler_______
  FXMAPFUNC(SEL_MOTION,              0,                        FXGLDynCanvas::onMotion),
  FXMAPFUNC(SEL_MOUSEWHEEL,          0,                        FXGLDynCanvas::onMouseWheel),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,     0,                        FXGLDynCanvas::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,   0,                        FXGLDynCanvas::onLeftBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,    0,                        FXGLDynCanvas::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,  0,                        FXGLDynCanvas::onRightBtnRelease),
};

FXIMPLEMENT(FXGLDynCanvas,FXGLCanvas,FXGLDynCanvasMap,ARRAYNUMBER(FXGLDynCanvasMap))

/********** Constructors and destructors **********/

// default constructor
FXGLDynCanvas::FXGLDynCanvas()
{
}

// init constructor
FXGLDynCanvas::FXGLDynCanvas(FXComposite* p,FXGLVisual *vis,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h) : FXGLCanvas(p,vis,tgt,sel,opts,x,y,w,h)
{
  eye_.x = 5.;
  eye_.y = 10.;
  eye_.z = 15.;
  center_.x = 0.;
  center_.y = 0.;
  center_.z = 0.;
  up_.x = 0.;
  up_.y = 1.;
  up_.z = 0.;
  leftButton_ = false;
  rightButton_ = false;
}

// init constructor
FXGLDynCanvas::FXGLDynCanvas(FXComposite* p,FXGLVisual *vis,FXGLCanvas* sharegroup,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h) : FXGLCanvas(p,vis,sharegroup,tgt,sel,opts,x,y,w,h)
{
  eye_.x = 5.;
  eye_.y = 10.;
  eye_.z = 15.;
  center_.x = 0.;
  center_.y = 0.;
  center_.z = 0.;
  up_.x = 0.;
  up_.y = 1.;
  up_.z = 0.;
  leftButton_ = false;
  rightButton_ = false;
}

// destructor
FXGLDynCanvas::~FXGLDynCanvas()
{
}

void FXGLDynCanvas::init(double e_x,double e_y,double e_z,double c_x,double c_y,double c_z,double u_x,double u_y,double u_z)
{
  eye_.x = e_x;
  eye_.y = e_y;
  eye_.z = e_z;
  center_.x = c_x;
  center_.y = c_y;
  center_.z = c_z;
  up_.x = u_x;
  up_.y = u_y;
  up_.z = u_z;
}
    
/******************* Accessors ********************/
double FXGLDynCanvas::getEyeX() const
{
  return this->eye_.x;
}
  
double FXGLDynCanvas::getEyeY() const
{
  return this->eye_.y;
}
  
double FXGLDynCanvas::getEyeZ() const
{
  return this->eye_.z;
}
  
double FXGLDynCanvas::getCenterX() const
{
  return this->center_.x;
}
  
double FXGLDynCanvas::getCenterY() const
{
  return this->center_.y;
}
  
double FXGLDynCanvas::getCenterZ() const
{
  return this->center_.z;
}
  
double FXGLDynCanvas::getUpX() const
{
  return this->up_.x;
}
  
double FXGLDynCanvas::getUpY() const
{
  return this->up_.y;
}
  
double FXGLDynCanvas::getUpZ() const
{
  return this->up_.z;
}

void FXGLDynCanvas::setMainWin(FXObject* mw,FXSelector sel)
{
  this->mainWin_ = mw;
  this->mainWinSel_ = sel;
}


/**************** Utility functions ***************/

double FXGLDynCanvas::distEyeCenter() const
{
  double dx = center_.x - eye_.x;
  double dy = center_.y - eye_.y;
  double dz = center_.z - eye_.z;
  return sqrt(dx * dx + dy * dy + dz * dz);
}

FXVec3d FXGLDynCanvas::toWorldVec(double x,double y) const
{
  FXVec3d vecEyeCenter(center_.x - eye_.x,center_.y - eye_.y,center_.z - eye_.z);
  FXVec3d vecNEyeCenter = normalize(vecEyeCenter);
  FXVec3d vecNUp = up_;
  FXVec3d vecNRight = vecNUp^vecNEyeCenter;

  double dist = this->distEyeCenter();
  FXVec3d res = x*vecNRight + y*vecNUp;
  return res;
}

/**************** Signal functions ****************/

long FXGLDynCanvas::onMotion(FXObject*,FXSelector,void* ptr)
{
  // cout << "Motion" << endl;
  FXEvent *ev=(FXEvent*)ptr;

  if(leftButton_)
    {
      // cout << "(" << ev->last_x << "," << ev->last_y << ") -> (" << ev->win_x << "," << ev->win_y << ")" << endl;
      double x = ev->win_x - ev->last_x;
      double y = ev->win_y - ev->last_y;
      FXVec3d vec = this->toWorldVec(x,y);
      
      FXVec3d oldVecCE = FXVec3d(eye_.x - center_.x,eye_.y - center_.y,eye_.z - center_.z);
      double oldDist = oldVecCE.length();
      this->eye_ += vec;
      FXVec3d newVecCE(eye_.x - center_.x,eye_.y - center_.y,eye_.z - center_.z);
      double newDist = newVecCE.length();
      FXVec3d newVecCE2 = newVecCE * (oldDist/newDist);

      this->eye_ = this->center_ + newVecCE2;

      if(this->mainWin_ != NULL)
	{
	  this->mainWin_->tryHandle(this,FXSEL(SEL_COMMAND,this->mainWinSel_),NULL);
	}
    }
  else if(rightButton_)
    {
      // cout << "(" << ev->last_x << "," << ev->last_y << ") <- (" << ev->win_x << "," << ev->win_y << ")" << endl;
      double x = ev->win_x - ev->last_x;
      double y = ev->win_y - ev->last_y;
      FXVec3d vec = this->toWorldVec(x,y);
      this->center_ += vec;
      this->eye_ += vec;

      if(this->mainWin_ != NULL)
	{
	  this->mainWin_->tryHandle(this,FXSEL(SEL_COMMAND,this->mainWinSel_),NULL);
	}
    }

  return 1;
}

long FXGLDynCanvas::onMouseWheel(FXObject*,FXSelector,void* ptr)
{
  // cout << "MouseWheel" << endl;
  FXEvent* event=(FXEvent*)ptr;
  FXVec3d vec = eye_ - center_;
  vec = vec*pow(2.0,-0.1*event->code/120.0);
  eye_ = center_ + vec;

  if(this->mainWin_ != NULL)
    {
      this->mainWin_->tryHandle(this,FXSEL(SEL_COMMAND,this->mainWinSel_),NULL);
    }
  return 1;
}

long FXGLDynCanvas::onLeftBtnPress(FXObject*,FXSelector,void*)
{
  // cout << "LeftBtnPress" << endl;
  leftButton_ = true;
  return 1;
}

long FXGLDynCanvas::onLeftBtnRelease(FXObject*,FXSelector,void*)
{
  // cout << "LeftBtnRelease" << endl;
  leftButton_ = false;
  return 1;
}

long FXGLDynCanvas::onRightBtnPress(FXObject*,FXSelector,void*)
{
  // cout << "RightBtnPress" << endl;
  rightButton_ = true;
  return 1;
}

long FXGLDynCanvas::onRightBtnRelease(FXObject*,FXSelector,void*)
{
  // cout << "RightBtnRelease" << endl;
  rightButton_ = false;
  return 1;
}

}
