#ifndef MPC_TEDDY_IFC_HPP
#define MPC_TEDDY_IFC_HPP

#include "motion/inc/STrajectory.hpp"
#include "motion/inc/Path.hpp"

#include "deformation/inc/PMTrajDeform.hpp"
#include "deformation/inc/CLTrajDeform3.hpp"
#include "deformation/inc/DDTrajDeform.hpp"
#include "deformation/inc/WCTrajDeform3.hpp"

#include "environment/inc/RoadMap.hpp"

#include "trajFollow/inc/TrajFollow.hpp"

#include "system/inc/PointMassSys.hpp"
#include "system/inc/CarLikeSys.hpp"
#include "system/inc/DiffDriveSys.hpp"

#include "common/inc/Timer.hpp"
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#include "foxExtension/inc/FXGLDynCanvas.hpp"
#include "fx.h"
#include "fx3d.h"
#include <stdio.h>
#include <iostream>

using namespace std;
using namespace mpc;

#define TEDDY_PROC 0
#define TF_PROC 1
#define PATH_PLNG_PROC 2
#define TRAJ_PLNG_PROC 3

#define CL_TF_MODE 0
#define CC_TF_MODE 1
#define WC_TF_MODE 2

const FXuint TIMER_INTERVAL = 1;

class MPC_TeddyIfc : public FXMainWindow {
  FXDECLARE(MPC_TeddyIfc)

private : 
  boost::thread*     threadTeddy_;                /**< thread used for the trajectory deformer Teddy */
  boost::thread*     threadCtrlr_;                /**< thread used for the controller */
  boost::thread*     threadEnv_;                  /**< thread used for the environment update */
  boost::thread*     threadDisplay_;              /**< thread used for the display */
  
//   Timer              timerTeddy_;                 /**< timer used for the trajectory deformer Teddy */
//   Timer              timerCtrlr_;                 /**< timer used for the controller */
//   Timer              timerEnv_;                   /**< timer used for the environment update */
//   Timer              timerDisplay_;               /**< timer used for the display */

  Environment        *env_;                       /**< the environment used */
  Path               *path_;                      /**< the path followed */
  STraj         *traj_;                     /**< the trajectory followed */
  STraj         *initTraj_;                 /**< copy of the initial trajectory */
//   Robot              *rb_;

//   PointMassSys       *pointmass_;                 /**< point mass system configuration */
//   CarLikeSys         *carlike_;                   /**< car-like system configuration */
//   DiffDriveSys       *diffdrive_;                 /**< differential drive system configuration */
  System             *sys_;                       /**< the system configuration */

  TrajDeform         *deformer_;
  TrajFollow         *follower_;
  //  MultiTrajFollow    *mfollower_;

  DRoadMap           *roadmap_;
  Map                *map_;

  int                **typeState;                 /**< type of each state of the trajectory (in collision, connec lost, ok) */
//   PositionXYT        ***tableNearState;           /**< table of the nearest states from each state of the trajectory */
//   StateTime          **tableFInt;                 /**< table of the internal forces applied on the trajectory */
//   StateTime          **tableFExt;                 /**< table of the external forces applied on the trajectory */

  StateTime          *robotStateTF_;          // state of the robot (only used for the trajectory following process)
  StateTime          *stateToReachTF_;        // state the robot try to reach (for the trajectory following process)

  FXGLDynCanvas      *glcanvas;               // GL Canvas to draw into
  int                spinning;                // Is box spinning
  double             angle;                   // Rotation angle of box
  FXGLVisual         *glvisual;               // OpenGL visual

  FXList*            list[9];
  FXHeader*          listHeader;

  FXRealSlider*      slider;
  bool		     running;

  int                processType_;            // type of the processus simulated (trajectory deformation or trajectory following)

  vector<FXVec3d>    lVec;

  vector<double>*    TF_X;                    // position X of the trajectory followed
  vector<double>*    TF_Y;                    // position Y of the trajectory followed
  vector<double>*    TF_T;                    // time T of the trajectory followed
  vector<double>*    TF_locX;                 // position X of the localization percieved during the traj following process
  vector<double>*    TF_locY;                 // position Y of the localization percieved during the traj following process
  vector<double>*    TF_rbX;                  // position X of the robot during the traj following process
  vector<double>*    TF_rbY;                  // position Y of the robot during the traj following process
  vector<double>*    TF_DerivLoc;             // derivation from the localization to the trajectory followed
  vector<double>*    TF_ErrorLoc;             // error of the localization
  vector<double>*    TF_DerivRb;              // derivation from the robot position to the trajectory followed
  vector<double>*    TF_rbCtrl1;              // first control applied on the robot
  vector<double>*    TF_rbCtrl2;              // second control applied on the robot
  vector<double>*    TF_noisedCtrl1;          // first control applied on the robot
  vector<double>*    TF_noisedCtrl2;          // second control applied on the robot

  // viewer components
  FXVec3d            eye_;
  FXVec3d            center_;
  FXVec3d            up_;

  FXLabel           *viewModeBox;             // label to say wich view mode is active
  FXLabel           *msgBox;                  // message box of the application

  // input components
  FXdouble           maxSpd;                  // maximal speed
  FXdouble           maxAcc;                  // maximal acceleration
  FXdouble           maxWAcc;                 // maximal wheel acceleration (for the DiffDrive)
  FXdouble           maxWVel;                 // maximal wheel velocity (for the DiffDrive)

  FXdouble           maxSteer;                // maximal steering
  FXdouble           maxSteerS;               // maximal steering speed
  FXdouble           maxSpdCoef;              // maximal speed coef for traj planning
  FXdouble           maxAccCoef;              // maximal acceleration coef for traj planning
  FXdouble           robotRadius;             // radius of the robot
  FXdouble           robotLength;             // length of the robot

  FXint              seqNum;                  // number of the random sequence
  FXint              optSeqNum;               // choice to compute a defined or random sequence

  FXString           idObs;                   // new obstacle ID
  FXdouble           xObs;                    // new obstacle X position
  FXdouble           yObs;                    // new obstacle Y position
  FXdouble           sxObs;                   // new obstacle X speed
  FXdouble           syObs;                   // new obstacle Y speed
  FXdouble           rObs;                    // new obstacle radius
  FXdouble           initT;                   // new obstacle init time
  FXdouble           lifeT;                   // new obstacle life time
  FXString           randObs;                 // new obstacle random mode

//   path_t             pathType;                // type of the path (holonomic, non-holonomic, nh with continuous curavature)
  FXint              genTrajType;             // type of the trajectory (holonomic or non holonomic trajectory)
  FXint              trajType;                // type of the trajectory (for trajectory tracking)
  FXdouble           trajTimeStep;            // time step for the trajectory computation
  FXdouble           trajPosStep;             // position step for the trajectory computation
  FXint  	     curStep;                 // current step of the trajectory
  FXdouble	     trajStep;                // current percentage step of the trajectory
  FXdouble           maxTrajStep;             // number of steps of the trajectory
  FXdouble           time;                    // current time of the trajectory
  FXdouble           maxTime;                 // maximal time of the trajectory
  struct timeval     t1;
  struct timeval     t2;

  bool               dispFuture;              // boolean to display the future model
  FXdouble	     reliabilityTime;         // time until which we consider the previsional model to be enough reliable
  FXdouble	     epsilonRT;               // epsilon for reliability time
  FXdouble	     predHorizon;             // time during which the previsional model is considered to be fully reliable

  FXdouble	     FPS;                     // medium number of frames per second
  FXdouble	     lastFPS;                 // last number of frames per second
  FXdouble	     deformDur;               // last deformation duration
  FXint              nbST;                    // number of states
  FXint              nbSTtot;                 // total number of states

  FXint              nbTimeout;               // number of frames computed
  FXint              totalNbTimeout;          // total number of frames computed
  bool               dispForces;              // boolean to display forces on the trajectory
  bool               dispAllForces;           // boolean to display all forces on the trajectory
  int                indDispForces;           // index of the config of the trajectory on which displaying forces
  bool               dispNearStates;          // boolean to display the nearest states from the dynamic obstacles to the trajectory
  bool               dispNearStStates;        // boolean to display the nearest states from the static obstacles to the trajectory

  double             alreadyElapsedTime;      // the time which has been already elapsed
  bool               oneStep;                 // compute just one step of the trajectory deformation

  FXString           inputPathName;           // input path filename
  FXString           mapName;                 // input map name
  FXint              sampleXOccGrid;          // sampling of the occupancy grid on the X dimension
  FXint              sampleYOccGrid;          // sampling of the occupancy grid on the Y dimension
  FXString           occgridName;             // input occupancy grid name

  FXint              nbRandObs;               // number of random obstacles

  FXint              optModeDeform;           // deformation mode

  FXdouble           K_INT;                   // internal forces coefficient
  FXdouble           K_EXT_SPT;               // external spatial forces coefficient
  FXdouble           K_EXT_TMP;               // external temporal forces coefficient
  FXdouble           K_SPT_TMP;               // balance between spatial and temporal forces
  FXdouble           K_EXT;                   // external forces coefficient (dynamic obstacles)
  FXdouble           K_STEXT;                 // external forces coefficient (static obstacles)
  FXdouble           maxEffectDist;           // maximum distance until which external forces take effects (dynamic obstacles)
  FXdouble           maxStEffectDist;         // maximum distance until which external forces take effects (static obstacles)
  FXdouble           alphaMin;                // minimal spatial coefficient (for the metric)
  FXdouble           alphaMax;                // maximal spatial coefficient (for the metric)
  FXdouble           betaMin;                 // minimal temporal coefficient (for the metric)
  FXdouble           betaMax;                 // maximal temporal coefficient (for the metric)
  FXdouble           maxSpDist;               // maximal spatial distance distance until which external forces take effects
  FXdouble           maxTmpDist;              // maximal temporal distance distance until which external forces take effects

  bool               useConnectivityChecker;  // true if the connection checker is used
  bool               useTemporalConsistency;  // true if the temporal consistency is used
  bool               pauseColLoc;             // true if it pause when a collision or a lose of connectivity is detected
  bool               dynamicView;             // true if the view move along the time
  bool               roadmapView;             // true if the roadmap is displayed

  FXdouble           seuilChgt;               // obstacle motion change threshold
  FXdouble           maxAngleChgt;            // obstacle max angle change
  FXdouble           maxNormChgt;             // obstacle max norm change

  FXint              typeMotion;              // type of motion used (traj tracking or positions tracking)

  FXint              initXRb;                 // init x position of the robot in position tracking mode
  FXint              initYRb;                 // init y position of the robot in position tracking mode
  FXdouble           initThRb;                // init th direction of the robot in position tracking mode
  FXdouble           initVRb;                 // init v velocity of the robot in position tracking mode
  FXdouble           initPhRb;                // init ph steering angle of the robot in position tracking mode

  FXint              goalXRb;                 // goal x position of the robot in position tracking mode
  FXint              goalYRb;                 // goal y position of the robot in position tracking mode
  FXdouble           goalThRb;                // goal th direction of the robot in position tracking mode

  FXint              dfltXMap;                // default x dimension of the map (used if no map has been loaded)
  FXint              dfltYMap;                // default y dimension of the map (used if no map has been loaded)
  bool               drawMapBoundaries;       // draw the boundaries of the map

  FXint              goalPosX;                // goal x position in position tracking mode
  FXint              goalPosY;                // goal y position in position tracking mode

  bool               realTimeAlgo;            // use real time algorithm if true

  bool               orthoView;               // Orthogonal / Perspective view

  FXint              debugLevel;              // debugging level
  FXint              maxNbStep;               // maximal number of steps of the algorithm
  FXdouble           maxResidual;             // maximal residual accepted
  
  FXbool             useVelCst;               // use linear velocity constraints
  FXbool             useAccCst;               // use linear acceleration constraints
  FXbool             useAVelCst;              // use angular velocity constraints
  FXbool             useAAccCst;              // use angular acceleration constraints
  
  FXbool             drawTrajGen;             // draw the generated trajectories for internal forces computation
  FXbool             drawTrajGenTF;           // draw the generated trajectories for the trajectory following process

  FXbool             parallelize;             // true if the trajectory generation process is parallelized
  FXint              nbThread;                // number of threads used to parallelize the trajectory generation process

  FXdouble           noiseX;                  // noise over the position X
  FXdouble           noiseY;                  // noise over the position Y
  FXdouble           noiseTH;                 // noise over the heading TH
  FXdouble           noiseV;                  // noise over the velocity V
  FXdouble           noisePH;                 // noise over the steering angle PH

  FXdouble           noiseAcc;                // noise over the acceleration control
  FXdouble           noiseSS;                 // noise over the steering speed control

  FXint              TF_mode;                 // mode of the trajectory following process (simple or multi traj following)

  FXdouble           TF_dt;                   // dt param for the trajectory following process
  FXdouble           TF_timeMeeting;          // timeMeeting param for the trajectory following process

  FXdouble           MTF_itdTime;             // intermediate time between two trajectory generation trial

  FXint              initXTraj;               // initial position of the trajectory (X)
  FXint              initYTraj;               // initial position of the trajectory (Y)
  FXdouble           initThTraj;              // initial heading of the trajectory (Th)
  FXdouble           initTTraj;               // initial time of the trajectory (T)
  
  FXdouble           trueLastElapsedT;        // true elapsed time since the last following (or deformation?) process
  FXdouble           oldTime;                 // value used to compute the true elasped time
  time_t             oldTimeTs;
  suseconds_t        oldTimeTus;

  FXbool             fixRbGoalPos;            // true if the goal position for the motion planning if fixed (else random pos)

  int                II_;                     // trick to display the trajectory followed
  double             medDerivLoc_;
  double             medDerivRb_;

  // link components
  FXDataTarget       maxSpd_target;
  FXDataTarget       maxAcc_target;
  FXDataTarget       maxWAcc_target;
  FXDataTarget       maxWVel_target;

  FXDataTarget       maxSteer_target;
  FXDataTarget       maxSteerS_target;
  FXDataTarget       maxSpdCoef_target;
  FXDataTarget       maxAccCoef_target;
  FXDataTarget       robotRadius_target;
  FXDataTarget       robotLength_target;
  
  FXDataTarget       seqNum_target;
  FXDataTarget       optSeqNum_target;
  
  FXDataTarget       idObs_target;
  FXDataTarget       xObs_target;
  FXDataTarget       yObs_target;
  FXDataTarget       sxObs_target;
  FXDataTarget       syObs_target;
  FXDataTarget       rObs_target;
  FXDataTarget       initT_target;
  FXDataTarget       lifeT_target;
  FXDataTarget       randObs_target;
  
  FXDataTarget       trajType_target;
  FXDataTarget       trajTimeStep_target;
  FXDataTarget       trajPosStep_target;
  FXDataTarget       trajStep_target;
  FXDataTarget       maxTrajStep_target;
  FXDataTarget       time_target;
  FXDataTarget       maxTime_target;
  
  FXDataTarget       inputPathName_target;
  FXDataTarget       mapName_target;
  FXDataTarget       sampleXOccGrid_target;
  FXDataTarget       sampleYOccGrid_target;
  FXDataTarget       occgridName_target;
  
  FXDataTarget       nbRandObs_target;
  
  FXDataTarget       reliabilityTime_target;
  FXDataTarget       epsilonRT_target;
  FXDataTarget       predHorizon_target;
  FXDataTarget       FPS_target;
  FXDataTarget       lastFPS_target;
  FXDataTarget       deformDur_target;
  FXDataTarget       nbST_target;
  FXDataTarget       nbSTtot_target;
  
  FXDataTarget       indDispForces_target;
  
  FXDataTarget       optModeDeform_target;
  
  FXDataTarget       K_INT_target;
  FXDataTarget       K_EXT_SPT_target;
  FXDataTarget       K_EXT_TMP_target;
  FXDataTarget       K_SPT_TMP_target;
  FXDataTarget       K_EXT_target;
  FXDataTarget       K_STEXT_target;
  FXDataTarget       maxEffectDist_target;
  FXDataTarget       maxStEffectDist_target;
  
  FXDataTarget       alphaMin_target;
  FXDataTarget       alphaMax_target;
  FXDataTarget       betaMin_target;
  FXDataTarget       betaMax_target;
  FXDataTarget       maxSpDist_target;
  FXDataTarget       maxTmpDist_target;
  
  FXDataTarget       seuilChgt_target;
  FXDataTarget       maxAngleChgt_target;
  FXDataTarget       maxNormChgt_target;

  FXDataTarget       typeMotion_target;

  FXDataTarget       initXRb_target;
  FXDataTarget       initYRb_target;
  FXDataTarget       initThRb_target;
  FXDataTarget       initVRb_target;
  FXDataTarget       initPhRb_target;

  FXDataTarget       goalXRb_target;
  FXDataTarget       goalYRb_target;
  FXDataTarget       goalThRb_target;

  FXDataTarget       dfltXMap_target;
  FXDataTarget       dfltYMap_target;

  FXDataTarget       goalPosX_target;
  FXDataTarget       goalPosY_target;

  FXDataTarget       debugLevel_target;
  FXDataTarget       maxNbStep_target;
  FXDataTarget       maxResidual_target;

  FXDataTarget       nbThread_target;

  FXDataTarget       noiseX_target;
  FXDataTarget       noiseY_target;
  FXDataTarget       noiseTH_target;
  FXDataTarget       noiseV_target;
  FXDataTarget       noisePH_target;

  FXDataTarget       noiseAcc_target;
  FXDataTarget       noiseSS_target;

  FXDataTarget       TF_mode_target;

  FXDataTarget       TF_dt_target;
  FXDataTarget       TF_timeMeeting_target;

  FXDataTarget       MTF_itdTime_target;

  FXDataTarget       initXTraj_target;
  FXDataTarget       initYTraj_target;
  FXDataTarget       initThTraj_target;
  FXDataTarget       initTTraj_target;


protected : 
  MPC_TeddyIfc();


public : 
  MPC_TeddyIfc(FXApp* a);
  virtual ~MPC_TeddyIfc();
  virtual void create();

  void initObs();
  void initObs2();
  void initObs3();
  void initObs4();
  void initObs5();
  void initObs6();
  void initObsRand(int nb);

  // signal functions
  long onConfigure(FXObject*,FXSelector,void*);
  long onExpose(FXObject*,FXSelector,void*);
  long loadPath(FXObject*,FXSelector,void*);
  long loadPrecompPath(FXObject*,FXSelector,void*);
  long loadPrecompTraj(FXObject*,FXSelector,void*);
  long loadMap(FXObject*,FXSelector,void*);
  long computeTrajectory(FXObject*,FXSelector,void*);
  long computeRandRKTraj(FXObject*,FXSelector,void*);
  long addObstacle(FXObject*,FXSelector,void*);
//   long addObstacle(FXString id, ObsDisk2D* obs, bool rand);
  long addObstacle(FXString id, const DiskObs2D& obs, bool rand);
//   long addStObstacle(FXString id, Obstacle2D* obs);
  long addStObstacle(FXString id, const BBoxObs2D& obs);
  long removeObstacle(FXObject*,FXSelector,void*);
  long changeInputPathName(FXObject*,FXSelector,void*);
  long changeMapName(FXObject*,FXSelector,void*);
  long onCmdHeaderButton(FXObject*,FXSelector,void*);
  long updateScene(FXObject*,FXSelector,void*);
  long startTraj(FXObject*,FXSelector,void*);
  long oneStepTraj(FXObject*,FXSelector,void*);
  long stopTraj(FXObject*,FXSelector,void*);
  long restartTraj(FXObject*,FXSelector,void*);
  long repaint(FXObject*,FXSelector,void*);
  long timeView(FXObject*,FXSelector,void*);
  long spaceView(FXObject*,FXSelector,void*);
  long perspView(FXObject*,FXSelector,void*);
  long updateDispFuture(FXObject*,FXSelector,void*);
  long updateDispForces(FXObject*,FXSelector,void*);
  long updateDispAllForces(FXObject*,FXSelector,void*);
  long updateDispNearSt(FXObject*,FXSelector,void*);
  long updateDispNearStSt(FXObject*,FXSelector,void*);
  long init(FXObject*,FXSelector,void*);
  long init2(FXObject*,FXSelector,void*);
  long init3(FXObject*,FXSelector,void*);
  long init4(FXObject*,FXSelector,void*);
  long init5(FXObject*,FXSelector,void*);
  long init6(FXObject*,FXSelector,void*);
  long initRandObs(FXObject*,FXSelector,void*);
  long clearEnv(FXObject*,FXSelector,void*);
  long useConnecChk(FXObject*,FXSelector,void*);
  long useTempCstcy(FXObject*,FXSelector,void*);
  long usePauseColLoc(FXObject*,FXSelector,void*);
  long initPositionTracking(FXObject*,FXSelector,void*);
  long updateDrawMapBoundaries(FXObject*,FXSelector,void*);
  long updateUseRTAlgo(FXObject*,FXSelector,void*);
  long updateDynamicView(FXObject*,FXSelector,void*);
  long updateRoadMapView(FXObject*,FXSelector,void*);
  long adjustParameters(FXObject*,FXSelector,void*);
  long perspectiveMode(FXObject*,FXSelector,void*);
  long orthogonalMode(FXObject*,FXSelector,void*);
  long displayRTGraph(FXObject*,FXSelector,void*);
  long updateUseVelCst(FXObject*,FXSelector,void*);
  long updateUseAccCst(FXObject*,FXSelector,void*);
  long updateUseAVelCst(FXObject*,FXSelector,void*);
  long updateUseAAccCst(FXObject*,FXSelector,void*);
  long updateDrawTrajGen(FXObject*,FXSelector,void*);
  long convertEnvOccGrid(FXObject*,FXSelector,void*);
  long changeOccGridName(FXObject*,FXSelector,void*);
  long updateParallel(FXObject*,FXSelector,void*);
  long computeDiffDriveTraj(FXObject*,FXSelector,void*);
  long computeItscTraj(FXObject*,FXSelector,void*);
  long computeMtwyTraj(FXObject*,FXSelector,void*);

  long initTrajFollowing(FXObject*,FXSelector,void*);
  long startTF(FXObject*,FXSelector,void*);
  long oneStepTF(FXObject*,FXSelector,void*);
  long stopTF(FXObject*,FXSelector,void*);
  long restartTF(FXObject*,FXSelector,void*);
  long updateTF(FXObject*,FXSelector,void*);
  long updateDrawTrajGenTF(FXObject*,FXSelector,void*);
  long updateRbGoalPos(FXObject*,FXSelector,void*);
  long dispTFTraj(FXObject*,FXSelector,void*);
  long dispJoinTraj(FXObject*,FXSelector,void*);

  // other functions
  void initDeform();                                  // initialize the deformation process
  int deformTraj();                                   // execute a step of the deformation algorithm

  void initTF();                                      // initialize the traj following process params that can change at each step
  int followTraj();                                   // execute a step of the following algorithm

  void displayTFTraj();                               // display the trajectories computed during the trajectory following process
  void displayJoinTraj();

  void initRobot();                                   // initialize the robot system constraints
  void initCarLikeSys();                              // initialize the car-like system constraints
  void initPointMassSys();                            // initialize the point mass system constraints
  void initWheelChairSys();                           // initialize the wheelchair system constraints

  void initTabForces();                               // init the tables of internal forces, external forces and nearest points
  void initViewer_PathLdd();                          // init the viewer after loading of a path

  void loadCarLikeParam();                            // load the system parameters for the car-like system
  void loadDiffDriveParam();                          // load the system parameters for the differential drive system
  void loadItscParam();                               // load the system parameters for the intersection scenario
  void loadMtwyParam();                               // load the system parameters for the motorway scenario

  PositionXYTh computeRandTargetPos(StateTime& initSt); // compute a new random target position
  int computeNewRandTrajFrom(StateTime& initCfg);     // compute a new random trajectory from initCfg
  int computeNewCarLikeTraj(StateTime& initCfg);      // compute a new Trajectory for the Car-Like Sys from initCfg
  int computeIntersectionTraj();                      // compute a trajectory for the intersection scenario
  int computeMotorwayTraj();                          // compute a trajectory for the motorway scenario
  int computeNewDiffDriveTraj(StateTime& initCfg);    // compute a new Trajectory for the DiffDrive Sys from initCfg
  void loadMapInViewer();                             // load a map in the viewer
  void drawMap();                                     // draw the map
  void verifyObsCfg(PMConfig* cfgObs,int i);          // verify and correct if needed the configuration of the obstacle
  void drawEnv();                                     // draw the environment components
//   void drawObstacles();                               // draw circular obstacles
//   void drawPolygonObs();                              // draw polygonal obstacles
  void drawPath();                                    // draw the path
  void drawTraj();                                    // draw the trajectory
  void drawTrajTF();                                  // draw the trajectory (trajectory following mode)
  void drawCLRobot(double heading);                   // draw the car-like robotic system
  void drawWCRobot();                                 // draw the wheelchair robotic system
  void drawRobot();                                   // draw the robotic system
  void drawRobotTF();                                 // draw the robotic system (trajectory following mode)
  void drawRoadMap();                                 // draw the roadmap
//   void drawGenTraj();                                 // draw the trajectories generated for the internal force computation
  void drawGenTrajTF();                               // draw the trajectories generated for the TF process
  void drawItdTraj();                                 // draw the intermediate trajectory
  void updDynamicView();                              // update the dynamic view
  void setLookAt(Environment* env,Path* path,STraj* traj); // change the point of view of the camera
  void drawScene();                                   //  draw all elements of the scene
  int elemListExists(const FXString& s);
  void initRandomDeform();                            // initialize the random deformation


  // Messages for our class
  enum{
    ID_CANVAS=FXMainWindow::ID_LAST,
    ID_LOAD_PATH,
    ID_LOAD_PCPPATH,
    ID_LOAD_PCPTRAJ,
    ID_LOAD_MAP,
    ID_COMPUTE_TRAJ,
    ID_CPT_RANDRKTRAJ,
    ID_ADD_OBS,
    ID_REM_OBS,
    ID_PATH_NAME,
    ID_MAP_NAME,
    ID_HEADER,
    ID_TIMEOUT,
    ID_START,
    ID_ONESTEP,
    ID_STOP,
    ID_RESTART,
    ID_GLVIEWER,
    ID_REPAINT,
    ID_TIME_VIEW,
    ID_SPACE_VIEW,
    ID_PERSP_VIEW,
    ID_DISP_FUTURE,
    ID_DISP_FORCES,
    ID_DISP_ALLFORCES,
    ID_DISP_NEAR_ST,
    ID_DISP_NEAR_STST,
    ID_INIT,
    ID_INIT2,
    ID_INIT3,
    ID_INIT4,
    ID_INIT5,
    ID_INIT6,
    ID_INITRANDOBS,
    ID_CLEAR_ENV,
    ID_CONNEC_CHK,
    ID_TEMP_CSTCY,
    ID_PAUSE_COL_LOC,
    ID_INIT_POSTRACK,
    ID_DRAW_MAPBOUND,
    ID_USE_RTALGO,
    ID_DYNVIEW,
    ID_ROADMAPVIEW,
    ID_ADJ_PARAM,
    ID_PERSP_MODE,
    ID_ORTH_MODE,
    ID_RT_GRAPH,
    ID_VEL_CST,
    ID_ACC_CST,
    ID_AVEL_CST,
    ID_AACC_CST,
    ID_DRAW_TRAJGEN,
    ID_CONVERT_OCCGRID,
    ID_OCCGRID_NAME,
    ID_PARALLEL,
    ID_CPT_DIFFDRIVETRAJ,
    ID_CPT_ITSCTRAJ,
    ID_CPT_MTWYTRAJ,
    ID_INIT_TFOLLOW,
    ID_START_TF,
    ID_ONESTEP_TF,
    ID_STOP_TF,
    ID_RESTART_TF,
    ID_TIMEOUT_TF,
    ID_DRAW_TRAJGEN_TF,
    ID_RB_GOALPOS,
    ID_DISP_TFTRAJ,
    ID_DISP_JOINTRAJ,
    ID_LAST
  };

};

FXDEFMAP(MPC_TeddyIfc) MPC_TeddyIfcMap[]={
  //________Message_Type_____________________ID_______________Message_Handler_______
  FXMAPFUNC(SEL_PAINT,           MPC_TeddyIfc::ID_CANVAS,            MPC_TeddyIfc::onExpose),
  FXMAPFUNC(SEL_CONFIGURE,       MPC_TeddyIfc::ID_CANVAS,            MPC_TeddyIfc::onConfigure),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_LOAD_PATH,         MPC_TeddyIfc::loadPath),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_LOAD_PCPPATH,      MPC_TeddyIfc::loadPrecompPath),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_LOAD_PCPTRAJ,      MPC_TeddyIfc::loadPrecompTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_LOAD_MAP,          MPC_TeddyIfc::loadMap),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_COMPUTE_TRAJ,      MPC_TeddyIfc::computeTrajectory),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CPT_RANDRKTRAJ,    MPC_TeddyIfc::computeRandRKTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ADD_OBS,           MPC_TeddyIfc::addObstacle),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_REM_OBS,           MPC_TeddyIfc::removeObstacle),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_PATH_NAME,         MPC_TeddyIfc::changeInputPathName),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_MAP_NAME,          MPC_TeddyIfc::changeMapName),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_HEADER,            MPC_TeddyIfc::onCmdHeaderButton),
  FXMAPFUNC(SEL_TIMEOUT,         MPC_TeddyIfc::ID_TIMEOUT,           MPC_TeddyIfc::updateScene),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_START,             MPC_TeddyIfc::startTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ONESTEP,           MPC_TeddyIfc::oneStepTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_STOP,              MPC_TeddyIfc::stopTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_RESTART,           MPC_TeddyIfc::restartTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_REPAINT,           MPC_TeddyIfc::repaint),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_TIME_VIEW,         MPC_TeddyIfc::timeView),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_SPACE_VIEW,        MPC_TeddyIfc::spaceView),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_PERSP_VIEW,        MPC_TeddyIfc::perspView),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_FUTURE,       MPC_TeddyIfc::updateDispFuture),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_FORCES,       MPC_TeddyIfc::updateDispForces),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_ALLFORCES,    MPC_TeddyIfc::updateDispAllForces),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_NEAR_ST,      MPC_TeddyIfc::updateDispNearSt),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_NEAR_STST,    MPC_TeddyIfc::updateDispNearStSt),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT,              MPC_TeddyIfc::init),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT2,             MPC_TeddyIfc::init2),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT3,             MPC_TeddyIfc::init3),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT4,             MPC_TeddyIfc::init4),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT5,             MPC_TeddyIfc::init5),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT6,             MPC_TeddyIfc::init6),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INITRANDOBS,       MPC_TeddyIfc::initRandObs),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CLEAR_ENV,         MPC_TeddyIfc::clearEnv),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CONNEC_CHK,        MPC_TeddyIfc::useConnecChk),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_TEMP_CSTCY,        MPC_TeddyIfc::useTempCstcy),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_PAUSE_COL_LOC,     MPC_TeddyIfc::usePauseColLoc),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT_POSTRACK,     MPC_TeddyIfc::initPositionTracking),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DRAW_MAPBOUND,     MPC_TeddyIfc::updateDrawMapBoundaries),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_USE_RTALGO,        MPC_TeddyIfc::updateUseRTAlgo),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DYNVIEW,           MPC_TeddyIfc::updateDynamicView),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ROADMAPVIEW,       MPC_TeddyIfc::updateRoadMapView),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ADJ_PARAM,         MPC_TeddyIfc::adjustParameters),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_PERSP_MODE,        MPC_TeddyIfc::perspectiveMode),  
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ORTH_MODE,         MPC_TeddyIfc::orthogonalMode),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_RT_GRAPH,          MPC_TeddyIfc::displayRTGraph),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_VEL_CST,           MPC_TeddyIfc::updateUseVelCst),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ACC_CST,           MPC_TeddyIfc::updateUseAccCst),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_AVEL_CST,          MPC_TeddyIfc::updateUseAVelCst),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_AACC_CST,          MPC_TeddyIfc::updateUseAAccCst),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DRAW_TRAJGEN,      MPC_TeddyIfc::updateDrawTrajGen),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CONVERT_OCCGRID,   MPC_TeddyIfc::convertEnvOccGrid),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_OCCGRID_NAME,      MPC_TeddyIfc::changeOccGridName),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_PARALLEL,          MPC_TeddyIfc::updateParallel),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CPT_DIFFDRIVETRAJ, MPC_TeddyIfc::computeDiffDriveTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CPT_ITSCTRAJ,      MPC_TeddyIfc::computeItscTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_CPT_MTWYTRAJ,      MPC_TeddyIfc::computeMtwyTraj),

  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_INIT_TFOLLOW,      MPC_TeddyIfc::initTrajFollowing),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_START_TF,          MPC_TeddyIfc::startTF),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_ONESTEP_TF,        MPC_TeddyIfc::oneStepTF),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_STOP_TF,           MPC_TeddyIfc::stopTF),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_RESTART_TF,        MPC_TeddyIfc::restartTF),
  FXMAPFUNC(SEL_TIMEOUT,         MPC_TeddyIfc::ID_TIMEOUT_TF,        MPC_TeddyIfc::updateTF),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DRAW_TRAJGEN_TF,   MPC_TeddyIfc::updateDrawTrajGenTF),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_RB_GOALPOS,        MPC_TeddyIfc::updateRbGoalPos),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_TFTRAJ,       MPC_TeddyIfc::dispTFTraj),
  FXMAPFUNC(SEL_COMMAND,         MPC_TeddyIfc::ID_DISP_JOINTRAJ,     MPC_TeddyIfc::dispJoinTraj),
};

FXIMPLEMENT(MPC_TeddyIfc,FXMainWindow,MPC_TeddyIfcMap,ARRAYNUMBER(MPC_TeddyIfcMap))


#endif // MPC_TEDDY_IFC_HPP
