#ifndef TRAJGENIFC_HPP
#define TRAJGENIFC_HPP

#include "environment/inc/Map.hpp"
#include "system/inc/System.hpp"

#include "trajGen/inc/TrajGen.hpp"
#include "trajGen/inc/CLTrajGen.hpp"
#include "trajGen/inc/SSTrajGen.hpp"
#include "trajGen/inc/WCTrajGen.hpp"
#include "trajGen/inc/ManaTrajGen.hpp"

#include "sslattice/inc/SSLattice.hpp"

#include "trajGen/inc/InitialGuess.hpp"

#include <fx.h>
#include <fx3d.h>
#include <stdio.h>
#include <iostream>

using namespace mpc;
using namespace std;

/*  \class TrajGenIfc
 *  \brief FOX-1.6 interface for the trajectory generator TIJI
 */
class TrajGenIfc : public FXMainWindow {
  FXDECLARE(TrajGenIfc)

private : 
  // application components
  System*            sys_;                     // the robotic system handled

  TrajGen*           trajGenerator;            // Trajectory Generation Model
  SSLattice*         lattice_;                 // state space lattice
  InitialGuess*      ig_;                      // initial guess of parameters database and learning module

  StateTime*         initST_;                  // initial state-time
  StateTime*         goalST_;                  // goal state-time

  Map*               map;                      // current map
  int                wtI;                      // width of the image
  int                htI;                      // height of the image

  FXint              genTrajMethod;            // method used for the trajectory generation process (ComputeTraj or genTraj)

  FXString           filename;                 // File being viewed

  FXdouble           x1;                       // x init position
  FXdouble           y1;                       // y init position
  FXdouble           theta1;                   // init direction
  FXdouble           kp1;                      // init steering angle
  FXdouble           v1;                       // init speed
  FXdouble           t1;                       // init time

  FXdouble           om1;                      // init angular velocity
  FXdouble           vx1;                      // init speed on X
  FXdouble           vy1;                      // init speed on Y
  
  FXdouble           x2;                       // x goal position
  FXdouble           y2;                       // y goal position
  FXdouble           theta2;                   // goal direction
  FXdouble           kp2;                      // goal steering angle
  FXdouble           v2;                       // goal speed
  FXdouble           t2;                       // goal time

  FXdouble           om2;                      // init angular velocity
  FXdouble           vx2;                      // init speed on X
  FXdouble           vy2;                      // init speed on Y

  FXdouble           MA_TH1;                   // Config of the first rotation link of the manipulator arm
  FXdouble           MA_TH2;                   // Config of the second rotation link of the manipulator arm
  FXdouble           MA_TH3;                   // Config of the third rotation link of the manipulator arm
  FXdouble           MA_OM1;                   // Speed of the first rotation link of the manipulator arm
  FXdouble           MA_OM2;                   // Speed of the first rotation link of the manipulator arm
  FXdouble           MA_OM3;                   // Speed of the first rotation link of the manipulator arm  

  FXdouble           x1_3D;                    // x init position (3D)
  FXdouble           y1_3D;                    // y init position (3D)
  FXdouble           z1_3D;                    // z init position (3D)
  FXdouble           vx1_3D;                   // vx init speed (3D)
  FXdouble           vy1_3D;                   // vy init speed (3D)
  FXdouble           vz1_3D;                   // vz init speed (3D)
  FXdouble           t1_3D;                    // init time (3D)

  FXdouble           x2_3D;                    // x goal position (3D)
  FXdouble           y2_3D;                    // y goal position (3D)
  FXdouble           z2_3D;                    // z goal position (3D)
  FXdouble           vx2_3D;                   // vx goal speed (3D)
  FXdouble           vy2_3D;                   // vy goal speed (3D)
  FXdouble           vz2_3D;                   // vz goal speed (3D)
  FXdouble           t2_3D;                    // goal time (3D)
  
  FXdouble           rb_Scale;                 // scale of the robot on the map
  FXdouble           rb_L;                     // length of the robot
  FXdouble           rb_VMAX;                  // maximal speed
  FXdouble           rb_AMAX;                  // maximal acceleration
  FXdouble           rb_KMAX;                  // maximal steering of the robot
  FXdouble           rb_OMAX;                  // maximal steering derivative of the robot
//   FXdouble           rb_AVMAX;                 // maximal angular velocity
//   FXdouble           rb_AAMAX;                 // maximal angular acceleration

  FXdouble           WC_AccMax;                // maximal linear acceleration of the wheelchair
  FXdouble           WC_AAccMax;               // maximal angular acceleration of the wheelchair
  FXdouble           WC_VelMax;                // maximal linear velocity of the wheelchair
  FXdouble           WC_AVelMax;               // maximal angular velocity of the wheelchair

  FXdouble           SS_AccMax;                // maximal linear acceleration of the spaceship
  FXdouble           SS_AAccMax;               // maximal angular acceleration of the spaceship
  FXdouble           SS_VelMax;                // maximal linear velocity of the spaceship
  FXdouble           SS_AVelMax;               // maximal angular velocity of the spaceship

  FXdouble           MA_AAcc1;                 // maximal angular acceleration around TH1 of the manipulator arm
  FXdouble           MA_AAcc2;                 // maximal angular acceleration around TH2 of the manipulator arm
  FXdouble           MA_AAcc3;                 // maximal angular acceleration around TH3 of the manipulator arm
  FXdouble           MA_AVel1;                 // maximal angular velocity around TH1 of the manipulator arm
  FXdouble           MA_AVel2;                 // maximal angular velocity around TH2 of the manipulator arm
  FXdouble           MA_AVel3;                 // maximal angular velocity around TH3 of the manipulator arm

  FXdouble           PM3D_AX;                  // maximal acceleration on X of the 3D point mass
  FXdouble           PM3D_AY;                  // maximal acceleration on Y of the 3D point mass
  FXdouble           PM3D_AZ;                  // maximal acceleration on Z of the 3D point mass
  FXdouble           PM3D_VX;                  // maximal speed on X of the 3D point mass
  FXdouble           PM3D_VY;                  // maximal speed on Y of the 3D point mass
  FXdouble           PM3D_VZ;                  // maximal speed on Z of the 3D point mass

  FXint              debugLevel;               // debugging level
  FXint              maxNbStep;                // maximal number of steps of the algorithm
  FXdouble           maxResidual;              // maximal residual accepted

  FXbool             useVelCst;                // use linear velocity constraints (of the car-like system)
  FXbool             useAccCst;                // use linear acceleration constraints (of the car-like system)
  FXbool             useAVelCst;               // use angular velocity constraints (of the car-like system)
  FXbool             useAAccCst;               // use angular acceleration constraints (of the car-like system)

  FXbool             canMoveBackCL;            // true if the car-like can move backward

  FXbool             dispCurvCirc;             // true if the car-like minimal curvature circles must be displayed

  FXbool             useVelCstWC;              // use linear velocity constraints (of the wheelchair system)
  FXbool             useAccCstWC;              // use linear acceleration constraints (of the wheelchair system)
  FXbool             useAVelCstWC;             // use angular velocity constraints (of the wheelchair system)
  FXbool             useAAccCstWC;             // use angular acceleration constraints (of the wheelchair system)

  FXbool             canMoveBackWC;            // true if the wheelchair can move backward

//   FXbool             useVelCstSS;              // use linear velocity constraints (of the spaceship system)
//   FXbool             useAccCstSS;              // use linear acceleration constraints (of the spaceship system)
//   FXbool             useAVelCstSS;             // use angular velocity constraints (of the spaceship system)
//   FXbool             useAAccCstSS;             // use angular acceleration constraints (of the spaceship system)

  FXint              nbBranches;               // number of branches of the reachable star
  FXdouble           minAngRS;                 // minimal angle from which the rstar is computed
  FXdouble           maxAngRS;                 // maximal angle from which the rstar is computed
  FXdouble           minDistRS;                // distance from which the rstar is computed
  FXdouble           maxDistRS;                // distance up to which the rstar is computed
  FXint              nbItvDistRS;              // number of intervals of distance computed
  FXint              trajByBranch;             // number of trajectories computed by branches
  FXdouble           diffAngle;                // diff angle for the different trajectories of a branch

  FXbool             drawUncompleteTraj;       // draw the uncomplete trajectories of the reachable star if true
  FXbool             drawTargetPos;            // draw the target positions

  FXbool             CL_XYf_fixed;             // true if the final position of the car-like is fixed
  FXbool             CL_THf_fixed;             // true if the final heading of the car-like is fixed
  FXbool             CL_PHf_fixed;             // true if the final steering angle of the car-like is fixed
  FXbool             CL_Vf_fixed;              // true if the final velocity of the car-like is fixed

  FXbool             WC_XYf_fixed;             // true if the final position of the wheelchair is fixed
  FXbool             WC_THf_fixed;             // true if the final heading of the wheelchair is fixed
  FXbool             WC_OMf_fixed;             // true if the final angular velocity of the wheelchair is fixed
  FXbool             WC_Vf_fixed;              // true if the final linear velocity of the wheelchair is fixed

  FXbool             SS_XYf_fixed;             // true if the final position of the spaceship is fixed
  FXbool             SS_THf_fixed;             // true if the final heading of the spaceship is fixed
  FXbool             SS_OMf_fixed;             // true if the final angular velocity of the spaceship is fixed
  FXbool             SS_Vf_fixed;              // true if the final linear velocity of the spaceship is fixed


  //vector<vector<double>*> tabPosTraj;          // table to stock the trajectories computed
  vector<vector<double>*> tabPosTrajX;         // table to stock the X coord of the trajectories computed
  vector<vector<double>*> tabPosTrajY;         // table to stock the Y coord of the trajectories computed
  vector<vector<double>*> tabPosTrajZ;         // table to stock the Z coord of the trajectories computed
  vector<vector<double>*> tabPosTrajT;         // table to stock the T coord of the trajectories computed
  vector<vector<double>*> tabPosTrajV;         // table to stock the V coord of the trajectories computed
  vector<vector<double>*> tabPosTrajVX;        // table to stock the VX coord of the trajectories computed
  vector<vector<double>*> tabPosTrajVY;        // table to stock the VY coord of the trajectories computed
  vector<vector<double>*> tabPosTrajVZ;        // table to stock the VZ coord of the trajectories computed
  vector<double>* tabPosTrajTmpX;              // table to stock the temporary X coord of the trajectories computed
  vector<double>* tabPosTrajTmpY;              // table to stock the temporary Y coord of the trajectories computed
  vector<double>* tabPosTrajTmpZ;              // table to stock the temporary Z coord of the trajectories computed
  vector<double>* tabPosTrajTmpT;              // table to stock the temporary T coord of the trajectories computed
  vector<double>* tabPosTrajTmpV;              // table to stock the temporary V coord of the trajectories computed
  vector<double>* tabPosTrajTmpVX;             // table to stock the temporary VX coord of the trajectories computed
  vector<double>* tabPosTrajTmpVY;             // table to stock the temporary VY coord of the trajectories computed
  vector<double>* tabPosTrajTmpVZ;             // table to stock the temporary VZ coord of the trajectories computed

  vector<double> tabPosReachedX;               // table to stock the positions X we try to reach
  vector<double> tabPosReachedY;               // table to stock the positions Y we try to reach
  vector<double> tabPosReachedZ;               // table to stock the positions Y we try to reach
  vector<double> tabPosReachedT;               // table to stock the times T we try to reach
  vector<double> tabPosReachedV;               // table to stock the velocities V we try to reach
  vector<double> tabPosReachedVX;              // table to stock the velocities VX we try to reach
  vector<double> tabPosReachedVY;              // table to stock the velocities VY we try to reach
  vector<double> tabPosReachedVZ;              // table to stock the velocities VZ we try to reach
  vector<double> tabPosReachedSt;              // table to stock the steering angles St we try to reach

  FXint              optSys;                   // option of the system used

  FXdouble dd_WVMAX;                           // differential drive maximal wheel velocity
  FXdouble dd_WAMAX;                           // differential drive maximal wheel acceleration
  FXdouble dd_RAD;                             // differential drive radius

  FXint              timeCsttStatus;           // time constraint status
  FXdouble           minTime;                  // minimal time
  FXdouble           maxTime;                  // maximal time

  FXString           IGfilename;               // filename of the IG database
  FXbool             useIG;                    // true if the ig database is used
  FXdouble           IG_tmin;                  // minimal time for the IG database learning
  FXdouble           IG_tmax;                  // maximal time for the IG database learning

  FXbool             saveRStar;                // true if the RStar trajectories have to be saved into an output file

  FXbool             useFDPD;                  // true if the finite difference partial derivative must be used

  // state space lattice components
  FXint              SSL_height;               // height of the tree defining the state space lattice
  FXint              SSL_nbCS;                 // number of control sampling used to compute the state space lattice
  FXbool             SSL_dispTraj;             // display the trajectories of the sslattice if true
  FXint              SSL_trajSp;               // sampling of the trajectories of the sslattice

  FXint              testReachability;         // true if the reachability of the final state must be tested

  // initial guess learning components
  FXint              IG_debugMode;             // debugging mode of the initial guess learning process
  FXint              IG_minNNTSR;              // minimal NTSR number for the IG learning process

  FXint              IG_1stParams;             // initialization mode of the IG learning process 
  FXString           IG1stParamsName;          // name of the file containing a list of trajectory parameterization for IG learning

  std::list<N2R*>    listParamTraj;            // list of parameterized trajectories to save
  
  FXint              CL_samp_t;                // sampling of the time for the CL initial guess learning
  FXint              CL_samp_vi;               // sampling of the initial speed for the CL initial guess learning
  FXint              CL_samp_vg;               // sampling of the final speed for the CL initial guess learning
  FXint              CL_samp_dx;               // sampling of the position for the CL initial guess learning
  FXint              CL_samp_dy;               // sampling of the position for the CL initial guess learning
  FXint              CL_samp_dth;              // sampling of the heading for the CL initial guess learning

  FXint              WC_samp_t;                // sampling of the time for the WC initial guess learning
  FXint              WC_samp_vi;               // sampling of the initial speed for the WC initial guess learning
  FXint              WC_samp_vg;               // sampling of the final speed for the WC initial guess learning
  FXint              WC_samp_dx;               // sampling of the position for the WC initial guess learning
  FXint              WC_samp_dy;               // sampling of the position for the WC initial guess learning
  FXint              WC_samp_dth;              // sampling of the heading for the WC initial guess learning

  FXint              CL_sampAcc;               // sampling of the acceleration control of a car-like for IG learning
  FXint              CL_sampStSpd;             // sampling of the steering speed control of a car-like for IG learning
  FXint              WC_sampAcc;               // sampling of the linear acceleration control of a wheelchair for IG learning
  FXint              WC_sampAAcc;              // sampling of the angular acceleration control of a wheelchair for IG learning
  FXint              SS_sampAcc;               // sampling of the linear acceleration control of a spaceship for IG learning
  FXint              SS_sampAAcc;              // sampling of the angular acceleration control of a spaceship for IG learning

  // interface components
  FXHorizontalFrame  *contents;                // Content frame
  FXVerticalFrame    *imageFrame;              // Image frame
  FXVerticalFrame    *buttonFrame;             // Button frame
  
  FXImageView        *imageview;               // Image viewer 
  
  bool               mlbFlag;                  // true if the left button of the mouse is pressed
  bool               mrbFlag;                  // true if the right button of the mouse is pressed
  
  FXColor            initColor;                // Color for the init configuration
  FXColor            goalColor;                // Color for the goal configuration
  
  FXLabel            *msgBox;                  // message box of the application


  
  // link components
  FXDataTarget       genTrajMethod_target;
  FXDataTarget       filename_target;

  FXDataTarget       x1_target;
  FXDataTarget       y1_target;
  FXDataTarget       theta1_target;
  FXDataTarget       kp1_target;
  FXDataTarget       v1_target;
  FXDataTarget       t1_target;

  FXDataTarget       om1_target;
  FXDataTarget       vx1_target;
  FXDataTarget       vy1_target;

  FXDataTarget       x2_target;
  FXDataTarget       y2_target;
  FXDataTarget       theta2_target;
  FXDataTarget       kp2_target;
  FXDataTarget       v2_target;
  FXDataTarget       t2_target;

  FXDataTarget       om2_target;
  FXDataTarget       vx2_target;
  FXDataTarget       vy2_target;

  FXDataTarget       rb_Scale_target;
  FXDataTarget       rb_L_target;
  FXDataTarget       rb_VMAX_target;
  FXDataTarget       rb_AMAX_target;
  FXDataTarget       rb_KMAX_target;
  FXDataTarget       rb_OMAX_target;

//   FXDataTarget       rb_AVMAX_target;
//   FXDataTarget       rb_AAMAX_target;

  FXDataTarget       WC_AccMax_target;
  FXDataTarget       WC_AAccMax_target;
  FXDataTarget       WC_VelMax_target;
  FXDataTarget       WC_AVelMax_target;

  FXDataTarget       SS_AccMax_target;
  FXDataTarget       SS_AAccMax_target;
  FXDataTarget       SS_VelMax_target;
  FXDataTarget       SS_AVelMax_target;

  FXDataTarget       debugLevel_target;
  FXDataTarget       maxNbStep_target;
  FXDataTarget       maxResidual_target;

  FXDataTarget       nbBranches_target;
  FXDataTarget       minAngRS_target;
  FXDataTarget       maxAngRS_target;
  FXDataTarget       minDistRS_target;
  FXDataTarget       maxDistRS_target;
  FXDataTarget       nbItvDistRS_target;
  FXDataTarget       trajByBranch_target;
  FXDataTarget       diffAngle_target;

  FXDataTarget       optSys_target;

  FXDataTarget       dd_WVMAX_target;
  FXDataTarget       dd_WAMAX_target;
  FXDataTarget       dd_RAD_target;

  FXDataTarget       timeCsttStatus_target;
  FXDataTarget       minTime_target;
  FXDataTarget       maxTime_target;

  FXDataTarget       SSL_height_target;
  FXDataTarget       SSL_nbCS_target;
  FXDataTarget       SSL_trajSp_target;

  FXDataTarget       testReachability_target;

  FXDataTarget       x1_3D_target;
  FXDataTarget       y1_3D_target;
  FXDataTarget       z1_3D_target;
  FXDataTarget       vx1_3D_target;
  FXDataTarget       vy1_3D_target;
  FXDataTarget       vz1_3D_target;
  FXDataTarget       t1_3D_target;

  FXDataTarget       x2_3D_target;
  FXDataTarget       y2_3D_target;
  FXDataTarget       z2_3D_target;
  FXDataTarget       vx2_3D_target;
  FXDataTarget       vy2_3D_target;
  FXDataTarget       vz2_3D_target;
  FXDataTarget       t2_3D_target;

  FXDataTarget       PM3D_AX_target;
  FXDataTarget       PM3D_AY_target;
  FXDataTarget       PM3D_AZ_target;
  FXDataTarget       PM3D_VX_target;
  FXDataTarget       PM3D_VY_target;
  FXDataTarget       PM3D_VZ_target;

  FXDataTarget       MA_TH1_target;
  FXDataTarget       MA_TH2_target;
  FXDataTarget       MA_TH3_target;
  FXDataTarget       MA_OM1_target;
  FXDataTarget       MA_OM2_target;
  FXDataTarget       MA_OM3_target;

  FXDataTarget       MA_AAcc1_target;
  FXDataTarget       MA_AAcc2_target;
  FXDataTarget       MA_AAcc3_target;
  FXDataTarget       MA_AVel1_target;
  FXDataTarget       MA_AVel2_target;
  FXDataTarget       MA_AVel3_target;

  FXDataTarget       IG_debugMode_target;
  FXDataTarget       IG_minNNTSR_target;

  FXDataTarget       IGfilename_target;
  FXDataTarget       IG_tmin_target;
  FXDataTarget       IG_tmax_target;
  FXDataTarget       IG_1stParams_target;
  FXDataTarget       IG1stParamsName_target;

  FXDataTarget       CL_sampAcc_target;
  FXDataTarget       CL_sampStSpd_target;
  FXDataTarget       WC_sampAcc_target;
  FXDataTarget       WC_sampAAcc_target;
  FXDataTarget       SS_sampAcc_target;
  FXDataTarget       SS_sampAAcc_target;

  FXDataTarget       CL_samp_t_target;
  FXDataTarget       CL_samp_vi_target;
  FXDataTarget       CL_samp_vg_target;
  FXDataTarget       CL_samp_dx_target;
  FXDataTarget       CL_samp_dy_target;
  FXDataTarget       CL_samp_dth_target;

  FXDataTarget       WC_samp_t_target;
  FXDataTarget       WC_samp_vi_target;
  FXDataTarget       WC_samp_vg_target;
  FXDataTarget       WC_samp_dx_target;
  FXDataTarget       WC_samp_dy_target;
  FXDataTarget       WC_samp_dth_target;


  /********** Constructors and destructors **********/

protected : 
  /**
   *  \brief default constructor (never used)
   */
  TrajGenIfc();

public : 
  /**
   *  \brief default constructor
   */
  TrajGenIfc(FXApp* a);

  /**
   *  \brief destructor
   */
  virtual ~TrajGenIfc();
  virtual void create();


  // signal functions
  long loadMapFromFile(FXObject*,FXSelector,void*);
  long changeMapName(FXObject*,FXSelector,void*);
  long genTraj(FXObject*,FXSelector,void*);
  long onLeftMouseDown(FXObject*,FXSelector,void*);
  long onLeftMouseUp(FXObject*,FXSelector,void*);
  long onRightMouseDown(FXObject*,FXSelector,void*);
  long onRightMouseUp(FXObject*,FXSelector,void*);
  long onMouseMove(FXObject*,FXSelector,void*);
  long displayBProfiles(FXObject*,FXSelector,void*);
  long displayWBProfiles(FXObject*,FXSelector,void*);
  long displayAllProfiles(FXObject*,FXSelector,void*);
  long updateUseVelCst(FXObject*,FXSelector,void*);
  long updateUseAccCst(FXObject*,FXSelector,void*);
  long updateUseAVelCst(FXObject*,FXSelector,void*);
  long updateUseAAccCst(FXObject*,FXSelector,void*);
  long updateUseVelCstWC(FXObject*,FXSelector,void*);
  long updateUseAccCstWC(FXObject*,FXSelector,void*);
  long updateUseAVelCstWC(FXObject*,FXSelector,void*);
  long updateUseAAccCstWC(FXObject*,FXSelector,void*);
  long drawRStar(FXObject*,FXSelector,void*);
  long updateDrawUncompleteTraj(FXObject*,FXSelector,void*);
  long updateDrawTargetPos(FXObject*,FXSelector,void*);
  long drawGnuplot2D(FXObject*,FXSelector,void*);
  long drawGnuplot3D(FXObject*,FXSelector,void*);
  long drawGnuplot(FXObject*,FXSelector,void*);
  long emptyVecTraj(FXObject*,FXSelector,void*);
  long drawSSLattice(FXObject*,FXSelector,void*);
  long updateDispLattice(FXObject*,FXSelector,void*);
  long drawGPLattice(FXObject*,FXSelector,void*);
  long genTrajSSLNodes(FXObject*,FXSelector,void*);
  long learnIGCmd(FXObject*,FXSelector,void*);
  long changeIGName(FXObject*,FXSelector,void*);
  long loadIG(FXObject*,FXSelector,void*);
  long displayIG(FXObject*,FXSelector,void*);
  long updateUseIG(FXObject*,FXSelector,void*);
  long updateUseFDPD(FXObject*,FXSelector,void*);
  long updateCL_XYf_fixed(FXObject*,FXSelector,void*);
  long updateCL_THf_fixed(FXObject*,FXSelector,void*);
  long updateCL_PHf_fixed(FXObject*,FXSelector,void*);
  long updateCL_Vf_fixed(FXObject*,FXSelector,void*);
  long updateDD_XYf_fixed(FXObject*,FXSelector,void*);
  long updateDD_THf_fixed(FXObject*,FXSelector,void*);
  long updateDD_OMf_fixed(FXObject*,FXSelector,void*);
  long updateDD_Vf_fixed(FXObject*,FXSelector,void*);
  long updateSS_XYf_fixed(FXObject*,FXSelector,void*);
  long updateSS_THf_fixed(FXObject*,FXSelector,void*);
  long updateSS_OMf_fixed(FXObject*,FXSelector,void*);
  long updateSS_Vf_fixed(FXObject*,FXSelector,void*);
  long changeIG1stParamsName(FXObject*,FXSelector,void*);
  long updateSaveRStar(FXObject*,FXSelector,void*);
  long changeSaveRStarName(FXObject*,FXSelector,void*);
  long updateCMBackCL(FXObject*,FXSelector,void*);
  long updateCMBackWC(FXObject*,FXSelector,void*);
  long computeCLManeuver(FXObject*,FXSelector,void*);
  long updateDispCurvCirc(FXObject*,FXSelector,void*);

  // other functions

  /**
   *  \brief init the time constraint of the trajectory generation process
   *  \return 0 if the process succeed
   *  \return -1 else
   */
  int initTimeCstt();

  /**
   *  \brief init the state space lattice
   *  \return 0 if the process succeed
   *  \return -1 else
   */
  int initSSLattice();

  /**
   *  \brief initialize the trajectory generator for a car-like system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   */
  void initCarLikeSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt);

  /**
   *  \brief initialize the trajectory generator for a car-like system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   */
  void initCarLikeSys2(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt);

  /**
   *  \brief initialize the trajectory generator for a spaceship system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   */
  void initSpaceShipSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt);

  /**
   *  \brief initialize the trajectory generator for a differential drive system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   */
  void initDiffDriveSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt);

  /**
   *  \brief initialize the trajectory generator for a differential drive system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   */
  void initWheelChairSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt);

  /**
   *  \brief initialize the trajectory generator for a Mana system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   */
  void initManaSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt);

  /**
   *  \brief initialize the trajectory generator for a car-like system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initCarLikeSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the trajectory generator for a car-like system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initCarLikeSys2(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the trajectory generator for a spaceship system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initSpaceShipSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the trajectory generator for a differential drive system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initDiffDriveSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the trajectory generator for a differential drive system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initWheelChairSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the trajectory generator for a Mana system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initManaSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the trajectory generator for a differential drive system
   *  \param x0 : init state
   *  \param xf : final state
   *  \param cstt : constraints defined over the system
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initPM3DSys(Vec_DP& x0,Vec_DP& xf,Vec_DP& cstt,int i,int j,int k);

  /**
   *  \brief initialize the robotic system constraints from the gui
   */
  void initSys();

  /**
   *  \brief initialize a car-like system from the gui
   */
  void initCarLikeSys();

  /**
   *  \brief initialize a wheelchair system from the gui
   */
  void initWheelChairSys();

  /**
   *  \brief initialize a Mana system from the gui
   */
  void initManaSys();

//   /**
//    *  \brief initialize a 3D point mass system from the gui
//    */
//   void initPM3DSys();

  /**
   *  \brief initialize a spaceship system from the gui
   */
  void initSpaceShipSys();

  /**
   *  \brief init the trajectory generation process from the gui
   */
  void initTrajGen();

  /**
   *  \brief init the trajectory generation process for a car-like system from the gui
   */
  void initCLTrajGen();

  /**
   *  \brief init the trajectory generation process for a wheelchair system from the gui
   */
  void initWCTrajGen();

  /**
   *  \brief init the trajectory generation process for a Mana system from the gui
   */
  void initManaTrajGen();

  /**
   *  \brief init the trajectory generation process for a spaceship system from the gui
   */
  void initSSTrajGen();

  /**
   *  \brief init the initial state-time from the gui
   */
  void initIST();

  /**
   *  \brief init the initial car-like state-time from the gui
   */
  void initICLST();

  /**
   *  \brief init the initial spaceship state-time from the gui
   */
  void initISSST();

  /**
   *  \brief init the initial wheelchair state-time from the gui
   */
  void initIWCST();

  /**
   *  \brief init the initial Mana state-time from the gui
   */
  void initIManaST();

//   /**
//    *  \brief init the initial 3D point mass state-time from the gui
//    */
//   void initIPM3DST();

  /**
   *  \brief save the current goal state-time to be able to display it
   *  \param gst : the goal state-time to save
   */
  void saveGST(const StateTime* gst);

  /**
   *  \brief init the goal state-time from the gui
   */
  void initGST();

  /**
   *  \brief init the goal car-like state-time from the gui
   */
  void initGCLST();

  /**
   *  \brief init the goal spaceship state-time from the gui
   */
  void initGSSST();

  /**
   *  \brief init the goal wheelchair state-time from the gui
   */
  void initGWCST();

  /**
   *  \brief init the goal Mana state-time from the gui
   */
  void initGManaST();

//   /**
//    *  \brief init the initial 3D point mass state-time from the gui
//    */
//   void initGPM3DST();

  /**
   *  \brief init the goal state-time from the gui
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initGST(int i,int j,int k);

  /**
   *  \brief init the goal car-like state-time from the gui
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initGCLST(int i,int j,int k);

  /**
   *  \brief init the goal spaceship state-time from the gui
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initGSSST(int i,int j,int k);

  /**
   *  \brief init the goal wheelchair state-time from the gui
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initGWCST(int i,int j,int k);

  /**
   *  \brief init the goal Mana state-time from the gui
   *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
   */
  void initGManaST(int i,int j,int k);

//   /**
//    *  \brief init the initial 3D point mass state-time from the gui
//    *  \param i,j,k : parameters defining the position of the final state computed over a regular sampling
//    */
//   void initGPM3DST(int i,int j,int k);

  /**
   *  \brief sample the trajectory generated to display it in a gnuplot window
   *  \param tg : the trajectory generator used
   */
  void sampleTraj(TrajGen* tg);

  /**
   *  \brief initialize the trajectory sampling
   */
  void initSampleTraj();

  /**
   *  \brief stock the trajectories sampled
   */
  void stockSampleTraj();

  /**
   *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
   *  \param tg : the trajectory generator used
   *  \param tabX : the table to stock X values
   *  \param tabY : the table to stock Y values
   *  \param tabT : the table to stock T values
   *  \param tabV : the table to stock V values
   */
  void sampleTrajCarLikeSys(CLTrajGen* tg,
			    vector<double>* tabX,
			    vector<double>* tabY,
			    vector<double>* tabT,
			    vector<double>* tabV );
  
//   /**
//    *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
//    *  \param tg : the trajectory generator used
//    *  \param tabX : the table to stock X values
//    *  \param tabY : the table to stock Y values
//    *  \param tabT : the table to stock T values
//    *  \param tabV : the table to stock V values
//    */
//   void sampleTrajCarLikeSys2(CLTrajGen2* tg,
// 			     vector<double>* tabX,
// 			     vector<double>* tabY,
// 			     vector<double>* tabT,
// 			     vector<double>* tabV );
  
  /**
   *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
   *  \param tg : the trajectory generator used
   *  \param tabX : the table to stock X values
   *  \param tabY : the table to stock Y values
   *  \param tabT : the table to stock T values
   *  \param tabV : the table to stock V values
   */
  void sampleTrajSpaceShipSys(SSTrajGen* tg,
			      vector<double>* tabX,
			      vector<double>* tabY,
			      vector<double>* tabT,
			      vector<double>* tabV );

//   /**
//    *  \brief sample the trajectory generated (for a car-like system) to display it in a gnuplot window
//    *  \param tg : the trajectory generator used
//    *  \param tabX : the table to stock X values
//    *  \param tabY : the table to stock Y values
//    *  \param tabT : the table to stock T values
//    *  \param tabV : the table to stock V values
//    */
//   void sampleTrajSpaceShipSys2(SSTrajGen2* tg,
// 			       vector<double>* tabX,
// 			       vector<double>* tabY,
// 			       vector<double>* tabT,
// 			       vector<double>* tabV );
  
//   /**
//    *  \brief sample the trajectory generated (for a differential drive system) to display it in a gnuplot window
//    *  \param tg : the trajectory generator used
//    *  \param tabX : the table to stock X values
//    *  \param tabY : the table to stock Y values
//    *  \param tabT : the table to stock T values
//    *  \param tabV : the table to stock V values
//    */
//   void sampleTrajDiffDriveSys(DDTrajGen* tg,
// 			      vector<double>* tabX,
// 			      vector<double>* tabY,
// 			      vector<double>* tabT,
// 			      vector<double>* tabV );

  /**
   *  \brief sample the trajectory generated (for a differential drive system) to display it in a gnuplot window
   *  \param tg : the trajectory generator used
   *  \param tabX : the table to stock X values
   *  \param tabY : the table to stock Y values
   *  \param tabT : the table to stock T values
   *  \param tabV : the table to stock V values
   */
  void sampleTrajWheelChairSys(WCTrajGen* tg,
			       vector<double>* tabX,
			       vector<double>* tabY,
			       vector<double>* tabT,
			       vector<double>* tabV );

  /**
   *  \brief sample the trajectory generated (for a mana system) to display it in a gnuplot window
   *  \param tg : the trajectory generator used
   *  \param tabX : the table to stock X values
   *  \param tabY : the table to stock Y values
   *  \param tabT : the table to stock T values
   *  \param tabV : the table to stock V values
   */
  void sampleTrajManaSys(ManaTrajGen* tg,
			 vector<double>* tabX,
			 vector<double>* tabY,
			 vector<double>* tabT,
			 vector<double>* tabV );

//   /**
//    *  \brief sample the trajectory generated (for a differential drive system) to display it in a gnuplot window
//    *  \param tg : the trajectory generator used
//    *  \param tabX : the table to stock X values
//    *  \param tabY : the table to stock Y values
//    *  \param tabT : the table to stock T values
//    *  \param tabV : the table to stock V values
//    */
//   void sampleTrajWheelChairSys2(WCTrajGen2* tg,
// 				vector<double>* tabX,
// 				vector<double>* tabY,
// 				vector<double>* tabT,
// 				vector<double>* tabV );

//   /**
//    *  \brief sample the trajectory generated (for a 3D point mass system) to display it in a gnuplot window
//    *  \param tg : the trajectory generator used
//    *  \param tabX : the table to stock X values
//    *  \param tabY : the table to stock Y values
//    *  \param tabZ : the table to stock Z values
//    *  \param tabT : the table to stock T values
//    *  \param tabVX : the table to stock VX values
//    *  \param tabVY : the table to stock VY values
//    *  \param tabVZ : the table to stock VZ values
//    */
//   void sampleTrajPM3DSys(PM3DTrajGen* tg,
// 			 vector<double>* tabX,
// 			 vector<double>* tabY,
// 			 vector<double>* tabZ,
// 			 vector<double>* tabT,
// 			 vector<double>* tabVX,
// 			 vector<double>* tabVY,
// 			 vector<double>* tabVZ );

  /**
   *  \brief empty the trajectory vectors
   */
  void emptyTrajVec();

  /**
   *  \brief test the reachability of the final state
   *  \return true if the input state is reachable, false else
   */
  bool testReachty(const StateTime& s);

  /**
   *  \brief test the reachability of the final car-like state
   *  \return true if the input state is reachable, false else
   */
  bool testCLReachty(const StateTime& s);

  /**
   *  \brief test the reachability of the final wheelchair state
   *  \return true if the input state is reachable, false else
   */
  bool testWCReachty(const StateTime& s);

  /**
   *  \brief test the reachability of the final Mana state
   *  \return true if the input state is reachable, false else
   */
  bool testManaReachty(const StateTime& s);

  /**
   *  \brief draw a 2D reachable star
   */
  void drawRStar2D();

  /**
   *  \brief draw a 3D reachable star
   */
  void drawRStar3D();

  /**
   *  \brief draw the set of trajectories computed in a gnuplot viewer
   */
  void drawGP();

  /**
   *  \brief draw the set of trajectories computed for a 2D directed mobile system in a gnuplot viewer
   */
  void drawGP_DM2D();

  /**
   *  \brief draw the set of trajectories computed for a 2D point mass system in a gnuplot viewer
   */
  void drawGP_PM2D();

  /**
   *  \brief draw the set of trajectories computed for a 3D point mass system in a gnuplot viewer
   */
  void drawGP_PM3D();

  /**
   *  \brief draw the set of trajectories computed in a gnuplot viewer
   */
  void drawGP_MA();

  /**
   *  \brief initialize the IG learning process from the gui
   *  \return 0 if the process succeed, -1 else
   */
  int initIG();

  /**
   *  \brief learn the initial guess of parameters
   */
  void learnIG();

  /**
   *  \brief clear the list of parameterized trajectories to save
   */
  void clearListParamTraj();

  /**
   *  \brief save the list of parameterized trajectories
   */
  void saveListParamTraj(const char* filename);

  /**
   *  \brief add a new parametric trajectory to save to the current list
   */
  void addParamTrajToSave();

  /**
   *  \brief init the car-like system from an input configuration file
   *  \param filename : the input filename
   */
  void initCLSysFromFile(const char* filename);

  /**
   *  \brief init the wheelchair system from an input configuration file
   *  \param filename : the input filename
   */
  void initWCSysFromFile(const char* filename);

  /**
   *  \brief init the spaceship system from an input configuration file
   *  \param filename : the input filename
   */
  void initSSSysFromFile(const char* filename);

  /**
   *  \brief init the gui from an input configuration file
   *  \param filename : the input filename
   */
  void initTrajGenIfcFromFile(const char* filename);


  // ifc drawing functions

  void drawInitPos();                                                         // draw the init position on the image
  void drawInitDir();                                                         // draw the init direction on the image
  void drawInitCfg();                                                         // draw the init configuration on the image
  void drawGoalPos();                                                         // draw the goal position on the image
  void drawGoalDir();                                                         // draw the goal direction on the image
  void drawGoalCfg();                                                         // draw the goal configuration on the image  

  FXbool loadimage(const FXString& file);                                     // load the image of the map
  void coordFrameToCoordMap(const int& xF, const int& yF, int& xM, int& yM);  // convert frame coordinates to map coordinates
  void coordMapToCoordFrame(const int& xM, const int& yM, int& xF, int& yF);  // convert map coordinates to frame coordinates

  // Messages for our class
  enum{
    ID_LOAD_MAP=FXMainWindow::ID_LAST,
    ID_MAP_NAME,
    ID_GEN_TRAJ,
    ID_IMAGE,
    ID_DISP_BPF,
    ID_DISP_WBPF,
    ID_DISP_ALLPF,
    ID_VEL_CST,
    ID_ACC_CST,
    ID_AVEL_CST,
    ID_AACC_CST,
    ID_VEL_WC_CST,
    ID_ACC_WC_CST,
    ID_AVEL_WC_CST,
    ID_AACC_WC_CST,
    ID_DRAW_RSTAR,
    ID_DRAW_UNCOMP,
    ID_DRAW_TGTPOS,
    ID_DRAW_GP2D,
    ID_DRAW_GP3D,
    ID_EMPTY_VECTRAJ,
    ID_DRAW_SSLATTICE,
    ID_DISP_LATTICE_CB,
    ID_DRAW_GPLATTICE,
    ID_GENTRAJ_SSLNODES,
    ID_LEARN_IG,
    ID_IG_NAME,
    ID_LOAD_IG,
    ID_DISP_IG,
    ID_USE_IG,
    ID_USE_FDPD,
    ID_CL_XYF_FIXED,
    ID_CL_THF_FIXED,
    ID_CL_PHF_FIXED,
    ID_CL_VF_FIXED,
    ID_DD_XYF_FIXED,
    ID_DD_THF_FIXED,
    ID_DD_OMF_FIXED,
    ID_DD_VF_FIXED,
    ID_SS_XYF_FIXED,
    ID_SS_THF_FIXED,
    ID_SS_OMF_FIXED,
    ID_SS_VF_FIXED,
    ID_IG_1STPARAM,
    ID_SAVE_RSTAR,
    ID_SAVE_RSTAR_FILENAME,
    ID_CMBACK_CL,
    ID_CMBACK_WC,
    ID_CPT_MNVER,
    ID_DISP_CURVCIRC,
    ID_LAST
  };

};

FXDEFMAP(TrajGenIfc) TrajGenIfcMap[]={
  //________Message_Type_______________________ID________________________Message_Handler_______
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_LOAD_MAP,            TrajGenIfc::loadMapFromFile),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_MAP_NAME,            TrajGenIfc::changeMapName),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_GEN_TRAJ,            TrajGenIfc::genTraj),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,      TrajGenIfc::ID_IMAGE,               TrajGenIfc::onLeftMouseDown),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,    TrajGenIfc::ID_IMAGE,               TrajGenIfc::onLeftMouseUp),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,     TrajGenIfc::ID_IMAGE,               TrajGenIfc::onRightMouseDown),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,   TrajGenIfc::ID_IMAGE,               TrajGenIfc::onRightMouseUp),
  FXMAPFUNC(SEL_MOTION,               TrajGenIfc::ID_IMAGE,               TrajGenIfc::onMouseMove),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DISP_BPF,            TrajGenIfc::displayBProfiles),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DISP_WBPF,           TrajGenIfc::displayWBProfiles),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DISP_ALLPF,          TrajGenIfc::displayAllProfiles),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_VEL_CST,             TrajGenIfc::updateUseVelCst),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_ACC_CST,             TrajGenIfc::updateUseAccCst),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_AVEL_CST,            TrajGenIfc::updateUseAVelCst),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_AACC_CST,            TrajGenIfc::updateUseAAccCst),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_VEL_WC_CST,          TrajGenIfc::updateUseVelCstWC),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_ACC_WC_CST,          TrajGenIfc::updateUseAccCstWC),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_AVEL_WC_CST,         TrajGenIfc::updateUseAVelCstWC),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_AACC_WC_CST,         TrajGenIfc::updateUseAAccCstWC),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_RSTAR,          TrajGenIfc::drawRStar),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_UNCOMP,         TrajGenIfc::updateDrawUncompleteTraj),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_TGTPOS,         TrajGenIfc::updateDrawTargetPos),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_GP2D,           TrajGenIfc::drawGnuplot),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_GP3D,           TrajGenIfc::drawGnuplot3D),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_EMPTY_VECTRAJ,       TrajGenIfc::emptyVecTraj),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_SSLATTICE,      TrajGenIfc::drawSSLattice),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DISP_LATTICE_CB,     TrajGenIfc::updateDispLattice),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DRAW_GPLATTICE,      TrajGenIfc::drawGPLattice),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_GENTRAJ_SSLNODES,    TrajGenIfc::genTrajSSLNodes),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_LEARN_IG,            TrajGenIfc::learnIGCmd),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_IG_NAME,             TrajGenIfc::changeIGName),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_LOAD_IG,             TrajGenIfc::loadIG),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DISP_IG,             TrajGenIfc::displayIG),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_USE_IG,              TrajGenIfc::updateUseIG),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_USE_FDPD,            TrajGenIfc::updateUseFDPD),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CL_XYF_FIXED,        TrajGenIfc::updateCL_XYf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CL_THF_FIXED,        TrajGenIfc::updateCL_THf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CL_PHF_FIXED,        TrajGenIfc::updateCL_PHf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CL_VF_FIXED,         TrajGenIfc::updateCL_Vf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DD_XYF_FIXED,        TrajGenIfc::updateDD_XYf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DD_THF_FIXED,        TrajGenIfc::updateDD_THf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DD_OMF_FIXED,        TrajGenIfc::updateDD_OMf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DD_VF_FIXED,         TrajGenIfc::updateDD_Vf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_SS_XYF_FIXED,        TrajGenIfc::updateSS_XYf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_SS_THF_FIXED,        TrajGenIfc::updateSS_THf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_SS_OMF_FIXED,        TrajGenIfc::updateSS_OMf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_SS_VF_FIXED,         TrajGenIfc::updateSS_Vf_fixed),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_IG_1STPARAM,         TrajGenIfc::changeIG1stParamsName),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_SAVE_RSTAR ,         TrajGenIfc::updateSaveRStar),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_SAVE_RSTAR_FILENAME, TrajGenIfc::changeSaveRStarName),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CMBACK_CL,           TrajGenIfc::updateCMBackCL),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CMBACK_WC,           TrajGenIfc::updateCMBackWC),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_CPT_MNVER,           TrajGenIfc::computeCLManeuver),
  FXMAPFUNC(SEL_COMMAND,              TrajGenIfc::ID_DISP_CURVCIRC,       TrajGenIfc::updateDispCurvCirc),
};

FXIMPLEMENT(TrajGenIfc,FXMainWindow,TrajGenIfcMap,ARRAYNUMBER(TrajGenIfcMap))

#endif // TRAJGENIFC_HPP
