#include "STDeformIfc.hpp"

#include "Utils.hpp"

#include "PVDTrajPlanner.hpp"
#include "GeomPathPlanner.hpp"
#include "RungeKuttaTrajGen.hpp"
#include "AccASTrajGen.hpp"
#include "WAccTrajGen.hpp"
#include "RoadMapPP.hpp"

#include "CLTrajGen.hpp"
#include "CL2TrajGen.hpp"
#include "DDTrajGen.hpp"
#include "UtilsTG.hpp"

#include "CLTrajFollow.hpp"
#include "CLMultiTrajFollow.hpp"

#include <time.h>
#include <sys/time.h>

#include <GL/glut.h>
#include <list>

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// default constructor (only used by subclasses)
STDeformIfc::STDeformIfc(){}

// init constructor
STDeformIfc::STDeformIfc(FXApp* a):FXMainWindow(a,"Trajectory Deformation",NULL,NULL,DECOR_ALL,0,0,950,650)
{
  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  FXHorizontalFrame *mainFrame;
  FXVerticalFrame *glcanvasFrame;
  FXVerticalFrame *buttonFrame;
  FXComposite *glpanel;

  path_     = NULL;
  traj_     = NULL;
  initTraj_ = NULL;

  rb_       = new Robot;
  carlike_  = NULL;
  //diffdrive_= new DiffDriveSys;


  deformer_ = NULL;
  follower_ = NULL;
  mfollower_ = NULL;

  roadmap_  = NULL;
  map_      = NULL;

  env_ = new Environment;

  typeState = NULL;
  tableFInt = NULL;
  tableFExt = NULL;
  
  glcanvas = NULL;
  glvisual = NULL;

  robotStateTF_ = NULL;
  stateToReachTF_ = NULL;

  processType_ = TEDDY_PROC;

  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 2.0*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  //robotRadius = 15.0f;
  
  seqNum    = 0;
  optSeqNum = 0;
  
  idObs   = FXString("obs1");
  xObs    = 0;
  yObs    = 0;
  sxObs   = 0;
  syObs   = 0;
  rObs    = 0;
  initT   = 0;
  lifeT   = 0;
  randObs = FXString("no");
  
  genTrajType  = HOLONOM_TRAJ;
  trajType     = HOLONOM_TRAJ;
  trajTimeStep = 1.0;
  trajPosStep  = 12.0;
  trajStep     = 0;
  maxTrajStep  = 1;
  time         = 0.0;
  maxTime      = 1.0;
  running      = false;
  
  inputPathName = FXString("../../path/path2.pth");
  mapName       = FXString("../../map/OccGrid.pgm");
  occgridName   = FXString("../../map/OccGrid.pgm");
  
  sampleXOccGrid = 400;
  sampleYOccGrid = 400;

  nbRandObs = 10;
  
  dispFuture         = false;
  alreadyElapsedTime = 0;
  oneStep            = false;
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  nbTimeout          = 0;
  totalNbTimeout     = 0;
  
  FPS                = 0;
  lastFPS            = 0;
  deformDur          = 0;
  nbST               = 0;
  nbSTtot            = 0;
  
  dispForces     = false;
  dispAllForces  = false;
  indDispForces  = 0;
  dispNearStates = false;
  
  optModeDeform = 0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;
//   maxEffectDist   = 80.0;
  maxStEffectDist = 80.0;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.0;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 80.0;
  maxTmpDist      = 5.0;
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  pauseColLoc = false;
  dynamicView = false;
  roadmapView = true;
  
  seuilChgt = 0.7;
  maxAngleChgt = 2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
  initXRb  = 50;
  initYRb  = 50;
  initThRb = 0.0;
  initVRb  = 0.0;
  initPhRb = 0.0;  

  dfltXMap          = 400;
  dfltYMap          = 400;
  drawMapBoundaries = true;
  
  goalPosX = -1;
  goalPosY = -1;
  
  realTimeAlgo = true;
  
  orthoView = false;
  
  debugLevel = 0;
  maxNbStep = 20;
  maxResidual = 0.5;
  
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;
  drawTrajGenTF = true;

  parallelize = false;
  nbThread = 5;

  noiseX = 250.0;
  //noiseY = 2.0;
  noiseTH = 0.4*M_PI;
  noiseV = 12.0;
  noisePH = 0.05;

  noiseAcc = 2.0;
  noiseSS = 0.02*M_PI;

  TF_mode = SIMPLE_TF_MODE;
  TF_dt = 0.5;
  TF_timeMeeting = 10.0;

  initXTraj  = 100;
  initYTraj  = 100;
  initThTraj = 0.25*M_PI;
  initTTraj  = 0.0;
  
  trueLastElapsedT = 0.0;

  
  maxSpd_target.connect(maxSpd);
  maxAcc_target.connect(maxAcc);
  maxWVel_target.connect(maxWVel);
  maxWAcc_target.connect(maxWAcc);

  maxSteer_target.connect(maxSteer);
  maxSteerS_target.connect(maxSteerS);
  maxSpdCoef_target.connect(maxSpdCoef);
  maxAccCoef_target.connect(maxAccCoef);
  robotRadius_target.connect(robotRadius);
  robotLength_target.connect(robotLength);
  
  seqNum_target.connect(seqNum);
  optSeqNum_target.connect(optSeqNum);
  
  idObs_target.connect(idObs);
  xObs_target.connect(xObs);
  yObs_target.connect(yObs);
  sxObs_target.connect(sxObs);
  syObs_target.connect(syObs);
  rObs_target.connect(rObs);
  initT_target.connect(initT);
  lifeT_target.connect(lifeT);
  randObs_target.connect(randObs);
  
  trajType_target.connect(trajType);
  trajTimeStep_target.connect(trajTimeStep);
  trajPosStep_target.connect(trajPosStep);

  trajStep_target.connect(trajStep);
  maxTrajStep_target.connect(maxTrajStep);
  time_target.connect(time);
  maxTime_target.connect(maxTime);
  
  inputPathName_target.connect(inputPathName);
  mapName_target.connect(mapName);
  occgridName_target.connect(occgridName);

  sampleXOccGrid_target.connect(sampleXOccGrid);
  sampleYOccGrid_target.connect(sampleYOccGrid);
  
  nbRandObs_target.connect(nbRandObs);
  
  reliabilityTime_target.connect(reliabilityTime);
  epsilonRT_target.connect(epsilonRT);
  predHorizon_target.connect(predHorizon);
  
  FPS_target.connect(FPS);
  lastFPS_target.connect(lastFPS);
  deformDur_target.connect(deformDur);
  nbST_target.connect(nbST);
  nbSTtot_target.connect(nbSTtot);

  indDispForces_target.connect(indDispForces);

  optModeDeform_target.connect(optModeDeform);

  K_INT_target.connect(K_INT);
  K_EXT_SPT_target.connect(K_EXT_SPT);
  K_EXT_TMP_target.connect(K_EXT_TMP);
  K_SPT_TMP_target.connect(K_SPT_TMP);
  K_EXT_target.connect(K_EXT);
  K_STEXT_target.connect(K_STEXT);
  maxStEffectDist_target.connect(maxStEffectDist);

  alphaMin_target.connect(alphaMin);
  alphaMax_target.connect(alphaMax);
  betaMin_target.connect(betaMin);
  betaMax_target.connect(betaMax);
  maxSpDist_target.connect(maxSpDist);
  maxTmpDist_target.connect(maxTmpDist);

  seuilChgt_target.connect(seuilChgt);
  maxAngleChgt_target.connect(maxAngleChgt);
  maxNormChgt_target.connect(maxNormChgt);

  typeMotion_target.connect(typeMotion);

  initXRb_target.connect(initXRb);
  initYRb_target.connect(initYRb);
  initThRb_target.connect(initThRb);
  initVRb_target.connect(initVRb);
  initPhRb_target.connect(initPhRb);

  dfltXMap_target.connect(dfltXMap);
  dfltYMap_target.connect(dfltYMap);

  goalPosX_target.connect(goalPosX);
  goalPosY_target.connect(goalPosY);

  debugLevel_target.connect(debugLevel);
  maxNbStep_target.connect(maxNbStep);
  maxResidual_target.connect(maxResidual);

  nbThread_target.connect(nbThread);

  noiseX_target.connect(noiseX);
  noiseY_target.connect(noiseY);
  noiseTH_target.connect(noiseTH);
  noiseV_target.connect(noiseV);
  noisePH_target.connect(noisePH);

  noiseAcc_target.connect(noiseAcc);
  noiseSS_target.connect(noiseSS);

  TF_mode_target.connect(TF_mode);
  TF_dt_target.connect(TF_dt);
  TF_timeMeeting_target.connect(TF_timeMeeting);

  initXTraj_target.connect(initXTraj);
  initYTraj_target.connect(initYTraj);
  initThTraj_target.connect(initThTraj);
  initTTraj_target.connect(initTTraj);


  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/
  
  mainFrame = new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /***************************************************************************************************************
   *********************************************** canvas frame **************************************************
   ***************************************************************************************************************/

  glcanvasFrame = new FXVerticalFrame(mainFrame,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Label above the buttons
  new FXLabel(glcanvasFrame,"Viewer",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  // Horizontal divider line
  new FXHorizontalSeparator(glcanvasFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Drawing glcanvas
  glpanel=new FXVerticalFrame(glcanvasFrame,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0, 0,0,0,0);

  // A Visual to drag OpenGL
  glvisual=new FXGLVisual(getApp(),VISUAL_DOUBLEBUFFER|VISUAL_STEREO);

  // Drawing glcanvas
  glcanvas=new FXGLDynCanvas(glpanel,glvisual,this,(FXSelector)ID_CANVAS,(FXuint)(LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT));
  glcanvas->setMainWin(this,ID_REPAINT);
  //glcanvas=new FXGLViewer(glpanel,glvisual,this,ID_GLVIEWER,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT);

  msgBox = new FXLabel(glpanel,"Select an environment and a motion behaviour",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);


  /***************************************************************************************************************
   *********************************************** button frame **************************************************
   ***************************************************************************************************************/
  
  buttonFrame = new FXVerticalFrame(mainFrame,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Label above the buttons
  new FXLabel(buttonFrame,"Commands",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  
  // Tab book with switchable panels
  FXTabBook* panels=new FXTabBook(buttonFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);


  /***************************************************************************************************************
   ***************************************** system configuration tab ********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Sys config");

  FXVerticalFrame *tabSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabSysCfg,"System configuration");

  FXMatrix *matSysCfg=new FXMatrix(tabSysCfg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSysCfg,"Max speed :");
  new FXTextField(matSysCfg,6,&maxSpd_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max acceleration :");
  new FXTextField(matSysCfg,6,&maxAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max wvel :");
  new FXTextField(matSysCfg,6,&maxWVel_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max wacc :");
  new FXTextField(matSysCfg,6,&maxWAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max steering :");
  new FXTextField(matSysCfg,6,&maxSteer_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max steer speed :");
  new FXTextField(matSysCfg,6,&maxSteerS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Robot radius :");
  new FXTextField(matSysCfg,6,&robotRadius_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Robot length :");
  new FXTextField(matSysCfg,6,&robotLength_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matCoef=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
new FXLabel(matCoef,"Max speed coef :");
  new FXTextField(matCoef,6,&maxSpdCoef_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matCoef,"Max acc coef :");
  new FXTextField(matCoef,6,&maxAccCoef_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  //FXVerticalFrame *paramFrame = new FXVerticalFrame(tabSysCfg,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabSysCfg,"Deformation parameters");

  FXHorizontalFrame* deformParamFrame = new FXHorizontalFrame(tabSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matOptionTraj=new FXMatrix(deformParamFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptionTraj,"reliability time :");
  new FXTextField(matOptionTraj,6,&reliabilityTime_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptionTraj,"epsilon RT :");
  new FXTextField(matOptionTraj,6,&epsilonRT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptionTraj,"prediction horizon :");
  new FXTextField(matOptionTraj,6,&predHorizon_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptionTraj,"max effective st dist :");
  new FXTextField(matOptionTraj,6,&maxEffectDist_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
//   FXCheckButton *connecChkBtn = new FXCheckButton(matOptionTraj,"Use connectivity checker",this,ID_CONNEC_CHK);
//   connecChkBtn->setCheck(true,false);
//   new FXLabel(matOptionTraj," ");
  FXCheckButton *tempCstcyBtn = new FXCheckButton(matOptionTraj,"Use temporal consistency",this,ID_TEMP_CSTCY);
  tempCstcyBtn->setCheck(true,false);
  new FXLabel(matOptionTraj," ");

  FXMatrix *matModeDeform=new FXMatrix(deformParamFrame,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matModeDeform,"Deformation Mode :");
  new FXRadioButton(matModeDeform,"Spatio-temporal deform",&optModeDeform_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matModeDeform,"Spatial deform",&optModeDeform_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matModeDeform,"Temporal deform",&optModeDeform_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  // Button to adjust parameters in function of the mode used
  new FXButton(matModeDeform,"Adjust parameters",NULL,this,ID_ADJ_PARAM,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matCoefForces =new FXMatrix(tabSysCfg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matCoefForces,"internal forces :");
  new FXTextField(matCoefForces,6,&K_INT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"spt-tmp balance :");
  new FXTextField(matCoefForces,6,&K_SPT_TMP_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"external forces :");
  new FXTextField(matCoefForces,6,&K_EXT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"static ext forces :");
  new FXTextField(matCoefForces,6,&K_STEXT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"min spatial coef :");
  new FXTextField(matCoefForces,6,&alphaMin_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"max spatial coef :");
  new FXTextField(matCoefForces,6,&alphaMax_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"min temporal coef :");
  new FXTextField(matCoefForces,6,&betaMin_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"max temporal coef :");
  new FXTextField(matCoefForces,6,&betaMax_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"spatial max dist :");
  new FXTextField(matCoefForces,6,&maxSpDist_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"temporal max dist :");
  new FXTextField(matCoefForces,6,&maxTmpDist_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);  

  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSysCfg,"Other options");

  FXMatrix *matRandMode=new FXMatrix(tabSysCfg,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  FXMatrix *matRandSeq=new FXMatrix(matRandMode,2,MATRIX_BY_COLUMNS|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);
  new FXRadioButton(matRandSeq,"Random sequence",&optSeqNum_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  FXMatrix *matSeqNum=new FXMatrix(matRandMode,2,MATRIX_BY_COLUMNS|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);
  new FXRadioButton(matSeqNum,"Sequence num : ",&optSeqNum_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXTextField(matSeqNum,10,&seqNum_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matRTAlgo=new FXMatrix(tabSysCfg,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *useRTAbtn = new FXCheckButton(matRTAlgo,"Use real-time algorithm",this,ID_USE_RTALGO);
  useRTAbtn->setCheck(TRUE);


  /***************************************************************************************************************
   ********************************************* Environment tab *************************************************
   ***************************************************************************************************************/
  
  new FXTabItem(panels,"Environment");

  FXVerticalFrame *tabEnv=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   FXMatrix *matMap=new FXMatrix(tabEnv,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
//   new FXLabel(matMap,"Map filename :");
//   new FXButton(matMap,"...",NULL,this,ID_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
//   new FXTextField(matMap,25,&mapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
 
//   // Button to load the map
//   new FXButton(tabEnv,"Load &Map",NULL,this,ID_LOAD_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matDfltMapDim=new FXMatrix(tabEnv,3,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matDfltMapDim,"Default map dimensions :");
  new FXTextField(matDfltMapDim,6,&dfltXMap_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXTextField(matDfltMapDim,6,&dfltYMap_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabEnv,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matObsList=new FXMatrix(tabEnv,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matObsList,"Obstacles list :");
  
  listHeader = new FXHeader(matObsList,this,ID_HEADER,HEADER_BUTTON|HEADER_RESIZE|FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X);

  listHeader->appendItem("ID Obs",NULL,45);
  listHeader->appendItem("X",NULL,45);
  listHeader->appendItem("Y",NULL,45);
  listHeader->appendItem("SX",NULL,45);
  listHeader->appendItem("SY",NULL,45);
  listHeader->appendItem("radius",NULL,45);
  listHeader->appendItem("init t",NULL,45);
  listHeader->appendItem("life t",NULL,45);
  listHeader->appendItem("rand",NULL,45);
  

  FXHorizontalFrame* listObsFame = new FXHorizontalFrame(matObsList,FRAME_SUNKEN|FRAME_THICK|LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0, 0,0);
  
  list[0]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[1]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[2]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[3]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[4]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[5]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[6]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[7]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[8]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[0]->setBackColor(FXRGB(255,240,240));
  list[1]->setBackColor(FXRGB(240,255,240));
  list[2]->setBackColor(FXRGB(240,240,255));
  list[3]->setBackColor(FXRGB(255,255,240));
  list[4]->setBackColor(FXRGB(255,240,255));
  list[5]->setBackColor(FXRGB(240,255,255));
  list[6]->setBackColor(FXRGB(255,255,255));
  list[7]->setBackColor(FXRGB(240,240,240));
  list[8]->setBackColor(FXRGB(255,240,240));

  new FXHorizontalSeparator(tabEnv,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabEnv,"New obstacle :");
  FXMatrix *matNewObs=new FXMatrix(tabEnv,5,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXTextField(matNewObs,6,&idObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Rand mode : ");
  new FXTextField(matNewObs,6,&randObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Rad :");
  new FXTextField(matNewObs,6,&rObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Position :");
  new FXLabel(matNewObs,"X :");
  new FXTextField(matNewObs,6,&xObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Y :");
  new FXTextField(matNewObs,6,&yObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Speed :");
  new FXLabel(matNewObs,"SX :");
  new FXTextField(matNewObs,6,&sxObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"SY :");
  new FXTextField(matNewObs,6,&syObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Time :");
  new FXLabel(matNewObs,"init t :");
  new FXTextField(matNewObs,6,&initT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"life t :");
  new FXTextField(matNewObs,6,&lifeT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);


  // Button to copy the map
  new FXButton(tabEnv,"&Add/Change Obstacle",NULL,this,ID_ADD_OBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(tabEnv,"&Remove Obstacle",NULL,this,ID_REM_OBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabEnv,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame *optObsFrame = new FXHorizontalFrame(tabEnv,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXVerticalFrame *randObsFrame = new FXVerticalFrame(optObsFrame,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(randObsFrame,"Obstacle random motion :");
  FXMatrix *matRandMotion=new FXMatrix(randObsFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matRandMotion,"Random threshold :");
  new FXTextField(matRandMotion,8,&seuilChgt_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matRandMotion,"Max angle change :");
  new FXTextField(matRandMotion,8,&maxAngleChgt_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matRandMotion,"Max norm change :");
  new FXTextField(matRandMotion,8,&maxNormChgt_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  FXVerticalFrame *initEnvFrame = new FXVerticalFrame(optObsFrame,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matInitEnv=new FXMatrix(initEnvFrame,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);

  new FXLabel(matInitEnv,"Environment \n initialisation :");

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Dyn Obs",NULL,this,ID_INIT2,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Labyrinth",NULL,this,ID_INIT3,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   // Button to load the preconfigured obstacles
//   new FXButton(matInitEnv,"Init4",NULL,this,ID_INIT4,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Motorway",NULL,this,ID_INIT5,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Labth (RM)",NULL,this,ID_INIT6,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Intersection",NULL,this,ID_INIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matInitRandEnv=new FXMatrix(initEnvFrame,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);
  
  // Button to load random obstacles
  new FXButton(matInitEnv,"InitRandObs",NULL,this,ID_INITRANDOBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5); 
  new FXTextField(matInitEnv,6,&nbRandObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matMap=new FXMatrix(tabEnv,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matMap,"Map filename :");
  new FXButton(matMap,"...",NULL,this,ID_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matMap,25,&mapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matMap,"");
  new FXLabel(matMap,"sample X : ");
  new FXTextField(matMap,8,&sampleXOccGrid_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matMap,"sample Y : ");
  new FXTextField(matMap,8,&sampleYOccGrid_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  

  // Button to load the map
  new FXButton(tabEnv,"Convert into Occupancy Grid",NULL,this,ID_CONVERT_OCCGRID,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ******************************************** motion config tab ************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Motion config");

  FXVerticalFrame *tabMotionCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matTrajConfig=new FXMatrix(tabMotionCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTrajConfig,"Trajectory computing time step :");
  new FXTextField(matTrajConfig,6,&trajTimeStep_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTrajConfig,"Trajectory computing pos step :");
  new FXTextField(matTrajConfig,6,&trajPosStep_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  new FXRadioButton(tabMotionCfg,"Trajectory tracking",&typeMotion_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXVerticalFrame *trajPlngFrame=new FXVerticalFrame(tabMotionCfg,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matInputPath=new FXMatrix(trajPlngFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInputPath,"Input path filename :");
  new FXButton(matInputPath,"...",NULL,this,ID_PATH_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matInputPath,25,&inputPathName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to load the init path
  new FXButton(trajPlngFrame,"Load Path",NULL,this,ID_LOAD_PATH,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to compute the trajectory
  new FXButton(trajPlngFrame,"Compute Trajectory",NULL,this,ID_COMPUTE_TRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to compute the trajectory
  new FXButton(trajPlngFrame,"Load Precomputed Trajectory",NULL,this,ID_LOAD_PCPPATH,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  // Button to compute a random trajectory by Runge Kutta
  new FXButton(trajPlngFrame,"Compute CarLike Trajectory",NULL,this,ID_CPT_RANDRKTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to compute a trajectory for the intersection scenario
  new FXButton(trajPlngFrame,"Compute Intersection Traj",NULL,this,ID_CPT_ITSCTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to compute a trajectory for the motorway scenario
  new FXButton(trajPlngFrame,"Compute Motorway Traj",NULL,this,ID_CPT_MTWYTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to compute the diffdrive trajectory
  new FXButton(trajPlngFrame,"Compute DiffDrive Trajectory",NULL,this,ID_CPT_DIFFDRIVETRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXLabel(trajPlngFrame,"Init trajectory cfg :");

  FXMatrix *matInitTrajCfg=new FXMatrix(trajPlngFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitTrajCfg,"X :");
  new FXTextField(matInitTrajCfg,6,&initXTraj_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitTrajCfg,"Y :");
  new FXTextField(matInitTrajCfg,6,&initYTraj_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitTrajCfg,"Th :");
  new FXTextField(matInitTrajCfg,6,&initThTraj_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitTrajCfg,"T :");
  new FXTextField(matInitTrajCfg,6,&initTTraj_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabMotionCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXRadioButton(tabMotionCfg,"Position tracking",&typeMotion_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXVerticalFrame *posTrckFrame=new FXVerticalFrame(tabMotionCfg,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(posTrckFrame,"Init robot position :");

  FXMatrix *matInitRbPos=new FXMatrix(posTrckFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitRbPos,"X :");
  new FXTextField(matInitRbPos,6,&initXRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matInitRbPos,"Y :");
  new FXTextField(matInitRbPos,6,&initYRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  new FXLabel(matInitRbPos,"Th :");
  new FXTextField(matInitRbPos,6,&initThRb_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXRadioButton(posTrckFrame,"Holonomic trajectories",&trajType_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(posTrckFrame,"NonHolonomic trajectories",&trajType_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  // Button to init the position tracking mode
  new FXButton(posTrckFrame,"&Init Position Tracking",NULL,this,ID_INIT_POSTRACK,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  // Button to display the reliabilty coefficient graph
  new FXButton(tabMotionCfg,"Reliability Coef Graph",NULL,this,ID_RT_GRAPH,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ************************************ Trajectory Generation options tab ****************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Options");

  /***************************************** NH Path Planning options ********************************************/

  FXVerticalFrame *tabNHOptions=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabNHOptions,"NH Path Planning Options");

  FXMatrix *matOccGrid=new FXMatrix(tabNHOptions,2,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOccGrid,"Occ Grid filename :");
  new FXButton(matOccGrid,"...",NULL,this,ID_OCCGRID_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matOccGrid,25,&occgridName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);  

  /************************************** Trajectory Generation options ******************************************/

  new FXHorizontalSeparator(tabNHOptions,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabNHOptions,"Trajectory Generation Options");
  
  FXVerticalFrame *tabTGOptions=new FXVerticalFrame(tabNHOptions,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matOptCst=new FXMatrix(tabTGOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptCst,"Constraints used :");
  FXCheckButton *velCstBtn = new FXCheckButton(matOptCst,"lin velocity cst",this,ID_VEL_CST);
  velCstBtn->setCheck(true,false);
  FXCheckButton *accCstBtn = new FXCheckButton(matOptCst,"lin acceleration cst",this,ID_ACC_CST);
  accCstBtn->setCheck(true,false);
  FXCheckButton *avelCstBtn = new FXCheckButton(matOptCst,"ang velocity cst",this,ID_AVEL_CST);
  avelCstBtn->setCheck(true,false);
  FXCheckButton *aaccCstBtn = new FXCheckButton(matOptCst,"ang acceleration cst",this,ID_AACC_CST);

  FXMatrix *matOptPrec=new FXMatrix(tabTGOptions,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptPrec,"Speed / Precision :");
  new FXLabel(matOptPrec,"");
  new FXLabel(matOptPrec,"Max number of steps:");
  new FXTextField(matOptPrec,6,&maxNbStep_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptPrec,"Max residual:");
  new FXTextField(matOptPrec,6,&maxResidual_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  FXMatrix *matOptDbg=new FXMatrix(tabTGOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptDbg,"Debugging mode :");
  new FXRadioButton(matOptDbg,"0",&debugLevel_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"1",&debugLevel_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"2",&debugLevel_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"3",&debugLevel_target,FXDataTarget::ID_OPTION+3,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);


  /************************************ Parallelization Generation options ***************************************/

  new FXHorizontalSeparator(tabNHOptions,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabNHOptions,"Parallelization Options");
  
  FXMatrix *matParallelOpt=new FXMatrix(tabNHOptions,2,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  FXCheckButton *parallelBtn = new FXCheckButton(matParallelOpt,"Parallelization",this,ID_PARALLEL);
  //parallelBtn->setCheck(true,false);
  new FXLabel(matParallelOpt,"");
  new FXLabel(matParallelOpt,"Nb threads : ");
  new FXTextField(matParallelOpt,6,&nbThread_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  

  /***************************************************************************************************************
   **************************************** trajectory deformation tab *******************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Traj deformation");

  FXVerticalFrame *tabTrajPl=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matCmdTraj=new FXMatrix(tabTrajPl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  // slider of the time
  slider = new FXRealSlider(matCmdTraj,&time_target,FXDataTarget::ID_VALUE,LAYOUT_CENTER_Y|LAYOUT_FILL_ROW|LAYOUT_FIX_WIDTH,0,0,250,0,0,0,0,0);
  slider->setRange(0.0,1.0);
  new FXTextField(matCmdTraj,6,&time_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  // Button to start the trajectory
  new FXButton(matCmdTraj,"Start",NULL,this,ID_START,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to compute just one step of the trajectory deformation
  new FXButton(matCmdTraj,"One step",NULL,this,ID_ONESTEP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to stop the trajectory
  new FXButton(matCmdTraj,"Stop",NULL,this,ID_STOP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to restart the trajectory
  new FXButton(matCmdTraj,"Restart",NULL,this,ID_RESTART,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabTrajPl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame* viewFrame = new FXHorizontalFrame(tabTrajPl,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matViewTraj=new FXMatrix(viewFrame,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matViewTraj, "Point of view : ");
  new FXButton(matViewTraj,"Time view",NULL,this,ID_TIME_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matViewTraj,"Space view",NULL,this,ID_SPACE_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matViewTraj,"Space-Time view",NULL,this,ID_PERSP_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matPerspOrth=new FXMatrix(viewFrame,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPerspOrth, "View Mode : ");
  viewModeBox = new FXLabel(matPerspOrth, "Perspective View");
  new FXButton(matPerspOrth,"Perspective view",NULL,this,ID_PERSP_MODE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matPerspOrth,"Orthogonal view",NULL,this,ID_ORTH_MODE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matOptViewTraj=new FXMatrix(tabTrajPl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptViewTraj, "Display Options : ");
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display future model",this,ID_DISP_FUTURE);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display forces",this,ID_DISP_FORCES);
  new FXTextField(matOptViewTraj,6,&indDispForces_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXCheckButton(matOptViewTraj,"Display all forces",this,ID_DISP_ALLFORCES);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display nearest obstacles states",this,ID_DISP_NEAR_ST);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Pause when collision/lose of connectivity",this,ID_PAUSE_COL_LOC);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton *drawMBbtn = new FXCheckButton(matOptViewTraj,"Draw map boundaries",this,ID_DRAW_MAPBOUND);
  drawMBbtn->setCheck(TRUE);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Dynamic view",this,ID_DYNVIEW);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton* rmviewBtn = new FXCheckButton(matOptViewTraj,"RoadMap view",this,ID_ROADMAPVIEW);
  rmviewBtn->setCheck(TRUE);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton* drawGenTrajBtn = new FXCheckButton(matOptViewTraj,"Draw generated trajectories",this,ID_DRAW_TRAJGEN);
  new FXLabel(matOptViewTraj, "");

  new FXHorizontalSeparator(tabTrajPl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matFPS=new FXMatrix(tabTrajPl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matFPS,"Medium FPS :");
  new FXTextField(matFPS,6,&FPS_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Last FPS :");
  new FXTextField(matFPS,6,&lastFPS_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Deformation duration :");
  new FXTextField(matFPS,6,&deformDur_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Nb of state-times :");
  new FXTextField(matFPS,6,&nbST_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Tot nb of state-times :");
  new FXTextField(matFPS,6,&nbSTtot_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   ***************************************** trajectory following tab ********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Traj following");
  
  FXVerticalFrame *tabTrajFlwg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabTrajFlwg,"Traj following commands:");

  FXMatrix *matCmdTF=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  // slider of the time
  slider = new FXRealSlider(matCmdTF,&time_target,FXDataTarget::ID_VALUE,LAYOUT_CENTER_Y|LAYOUT_FILL_ROW|LAYOUT_FIX_WIDTH,0,0,250,0,0,0,0,0);
  slider->setRange(0.0,1.0);
  new FXTextField(matCmdTF,6,&time_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  // Button to start the trajectory
  new FXButton(matCmdTF,"Start",NULL,this,ID_START_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to compute just one step of the trajectory deformation
  new FXButton(matCmdTF,"One step",NULL,this,ID_ONESTEP_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to stop the trajectory
  new FXButton(matCmdTF,"Stop",NULL,this,ID_STOP_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to restart the trajectory
  new FXButton(matCmdTF,"Restart",NULL,this,ID_RESTART_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Traj following parameters:");

  FXMatrix *matTFparam=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTFparam,"DT :");
  new FXTextField(matTFparam,6,&TF_dt_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTFparam,"T meet :");
  new FXTextField(matTFparam,6,&TF_timeMeeting_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  FXCheckButton* dispTG_TF_Btn = new FXCheckButton(matTFparam,"Disp trajgen",this,ID_DRAW_TRAJGEN_TF);  
  dispTG_TF_Btn->setCheck(TRUE);
  new FXLabel(matTFparam,"");

  new FXLabel(tabTrajFlwg,"Traj Following Mode:");
  FXMatrix *matTFmode=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXRadioButton(matTFmode,"Simple Traj Following",&TF_mode_target,FXDataTarget::ID_OPTION+SIMPLE_TF_MODE,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTFmode,"Multi Traj Following",&TF_mode_target,FXDataTarget::ID_OPTION+MULTI_TF_MODE,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  
  new FXButton(tabTrajFlwg,"Init Traj Following",NULL,this,ID_INIT_TFOLLOW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Init Robot State:");

  FXMatrix *matInitRbState=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitRbState,"X :");
  new FXTextField(matInitRbState,6,&initXRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"Y :");
  new FXTextField(matInitRbState,6,&initYRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"Th :");
  new FXTextField(matInitRbState,6,&initThRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"V :");
  new FXTextField(matInitRbState,6,&initVRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"Ph :");
  new FXTextField(matInitRbState,6,&initPhRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Max State Noise:");

  FXMatrix *matStateNoise=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matStateNoise,"XY noise:");
  new FXTextField(matStateNoise,6,&noiseX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matStateNoise,"Y noise:");
//   new FXTextField(matStateNoise,6,&noiseY_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"Th noise:");
  new FXTextField(matStateNoise,6,&noiseTH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"V noise:");
  new FXTextField(matStateNoise,6,&noiseV_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"Ph noise:");
  new FXTextField(matStateNoise,6,&noisePH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Max Control Noise:");

  FXMatrix *matControlNoise=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matControlNoise,"Acc noise:");
  new FXTextField(matControlNoise,6,&noiseAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matControlNoise,"SS noise:");
  new FXTextField(matControlNoise,6,&noiseSS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   ********************************************* general commands ************************************************
   ***************************************************************************************************************/

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Exit button
  new FXButton(buttonFrame,"&Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
}

// destructor
STDeformIfc::~STDeformIfc()
{
  if(env_ != NULL) delete env_;
  if(path_ != NULL) delete path_;
  if(traj_ != NULL) delete traj_;

  if(glcanvas != NULL) delete glcanvas;
  if(glvisual != NULL) delete glvisual; 
}

// Create and initialize
void STDeformIfc::create(){
  FXMainWindow::create();
  show(PLACEMENT_SCREEN);
}

// intersections environment
void STDeformIfc::initObs()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(1370);
  this->env_->setDfltY(900);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());
  
  // create the static obstacles

  ConfigR2D pos1(400,200);
  Obstacle2D* stObs1 = new ObsRect2D(pos1,400, 30.0, 0.5*M_PI);
  ConfigR2D pos2(500,200);
  Obstacle2D* stObs2 = new ObsRect2D(pos2,400, 30.0, 0.5*M_PI);
  ConfigR2D pos3(400,700);
  Obstacle2D* stObs3 = new ObsRect2D(pos3,400, 30.0, 0.5*M_PI);
  ConfigR2D pos4(500,700);
  Obstacle2D* stObs4 = new ObsRect2D(pos4,400, 30.0, 0.5*M_PI);

  ConfigR2D pos5(215,400);
  Obstacle2D* stObs5 = new ObsRect2D(pos5,400, 30.0, 0.0);
  ConfigR2D pos6(685,400);
  Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.0);
  ConfigR2D pos7(215,500);
  Obstacle2D* stObs7 = new ObsRect2D(pos7,400, 30.0, 0.0);
  ConfigR2D pos8(685,500);
  Obstacle2D* stObs8 = new ObsRect2D(pos8,400, 30.0, 0.0);

  ConfigR2D pos9(870,200);
  Obstacle2D* stObs9 = new ObsRect2D(pos9,400, 30.0, 0.5*M_PI);
  ConfigR2D pos10(970,200);
  Obstacle2D* stObs10 = new ObsRect2D(pos10,400, 30.0, 0.5*M_PI);
  ConfigR2D pos11(870,700);
  Obstacle2D* stObs11 = new ObsRect2D(pos11,400, 30.0, 0.5*M_PI);
  ConfigR2D pos12(970,700);
  Obstacle2D* stObs12 = new ObsRect2D(pos12,400, 30.0, 0.5*M_PI);

  ConfigR2D pos13(1155,400);
  Obstacle2D* stObs13 = new ObsRect2D(pos13,400, 30.0, 0.0);
  ConfigR2D pos14(1155,500);
  Obstacle2D* stObs14 = new ObsRect2D(pos14,400, 30.0, 0.0);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);

  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  this->addStObstacle(FXString("stObs14"),stObs14);

  // create the dynamic obstacles
  DynamicObs* obs1  = new ObsDisk2D(ConfigR2D(450,375), Speed2D(0,20),  0,  500, 20);
  DynamicObs* obs2  = new ObsDisk2D(ConfigR2D(450,250), Speed2D(0,20),  0,  500, 20);
  DynamicObs* obs3  = new ObsDisk2D(ConfigR2D(450,125),  Speed2D(0,20),  0,  500, 20);
  DynamicObs* obs4  = new ObsDisk2D(ConfigR2D(450,0),   Speed2D(0,20),  3,  500, 20);
  DynamicObs* obs5  = new ObsDisk2D(ConfigR2D(920,500), Speed2D(0,-20), 0,  500, 20);
  DynamicObs* obs6  = new ObsDisk2D(ConfigR2D(920,650), Speed2D(0,-20), 0,  500, 20);
  DynamicObs* obs7  = new ObsDisk2D(ConfigR2D(920,800), Speed2D(0,-20), 0,  500, 20);
  DynamicObs* obs8  = new ObsDisk2D(ConfigR2D(920,800), Speed2D(0,-20), 6,  500, 20);
  DynamicObs* obs9  = new ObsDisk2D(ConfigR2D(920,800), Speed2D(0,-20), 12, 500, 20);
  DynamicObs* obs10 = new ObsDisk2D(ConfigR2D(920,800), Speed2D(0,-20), 18, 500, 20);

  // append them to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  this->addObstacle(FXString("obs4"),obs4,true);
  this->addObstacle(FXString("obs5"),obs5,true);
  this->addObstacle(FXString("obs6"),obs6,true);
  this->addObstacle(FXString("obs7"),obs7,true);
  this->addObstacle(FXString("obs8"),obs8,true);
  this->addObstacle(FXString("obs9"),obs9,true);
  this->addObstacle(FXString("obs10"),obs10,true);

  // refresh the viewer
  this->drawScene();
  
  // init the random
  // initRandomDeform();

  msgBox->setText("Environment \"Intersection\" loaded.");

  // compute the intersection trajectory
  orthoView = true;
  dynamicView = true;
  this->computeIntersectionTraj();
}

// dynamic obstacles environment
void STDeformIfc::initObs2()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  // create the dynamic obstacles
  DynamicObs* obs1 = new ObsDisk2D(ConfigR2D(150,187), Speed2D(12,-10), 0, 500, 10);
  DynamicObs* obs2 = new ObsDisk2D(ConfigR2D(125,0), Speed2D(6,20), 0, 500, 13);
  DynamicObs* obs3 = new ObsDisk2D(ConfigR2D(263,38), Speed2D(-16,10), 0, 500, 17);
  DynamicObs* obs4 = new ObsDisk2D(ConfigR2D(38,315), Speed2D(20,-8), 0, 500, 10);
  DynamicObs* obs5 = new ObsDisk2D(ConfigR2D(170,263), Speed2D(10,16), 0, 500, 13);
  DynamicObs* obs6 = new ObsDisk2D(ConfigR2D(300,150), Speed2D(-20,-4), 0, 500, 17);
  DynamicObs* obs7 = new ObsDisk2D(ConfigR2D(38,150), Speed2D(16,10), 0, 500, 10);
  DynamicObs* obs8 = new ObsDisk2D(ConfigR2D(187,105), Speed2D(0,24), 0, 500, 13);
  DynamicObs* obs9 = new ObsDisk2D(ConfigR2D(285,338), Speed2D(-14,-14), 0, 500, 17);
  DynamicObs* obs10 = new ObsDisk2D(ConfigR2D(338,225), Speed2D(12,-12), 0, 500, 10);

  DynamicObs* obs11 = new ObsDisk2D(ConfigR2D(550,287), Speed2D(-12,10), 0, 500, 10);
  DynamicObs* obs12 = new ObsDisk2D(ConfigR2D(75,600), Speed2D(-6,-20), 0, 500, 13);
  DynamicObs* obs13 = new ObsDisk2D(ConfigR2D(363,538), Speed2D(16,-10), 0, 500, 17);
  DynamicObs* obs14 = new ObsDisk2D(ConfigR2D(238,615), Speed2D(-20,8), 0, 500, 10);
  DynamicObs* obs15 = new ObsDisk2D(ConfigR2D(570,563), Speed2D(-10,-16), 0, 500, 13);
  DynamicObs* obs16 = new ObsDisk2D(ConfigR2D(650,150), Speed2D(20,4), 0, 500, 17);
  DynamicObs* obs17 = new ObsDisk2D(ConfigR2D(438,450), Speed2D(-16,-10), 0, 500, 10);
  DynamicObs* obs18 = new ObsDisk2D(ConfigR2D(287,405), Speed2D(0,-24), 0, 500, 13);
  DynamicObs* obs19 = new ObsDisk2D(ConfigR2D(485,598), Speed2D(14,14), 0, 500, 17);
  DynamicObs* obs20 = new ObsDisk2D(ConfigR2D(638,325), Speed2D(-12,12), 0, 500, 10);

  // append them to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  this->addObstacle(FXString("obs4"),obs4,true);
  this->addObstacle(FXString("obs5"),obs5,true);
  this->addObstacle(FXString("obs6"),obs6,true);
  this->addObstacle(FXString("obs7"),obs7,true);
  this->addObstacle(FXString("obs8"),obs8,true);
  this->addObstacle(FXString("obs9"),obs9,true);
  this->addObstacle(FXString("obs10"),obs10,true);

  this->addObstacle(FXString("obs11"),obs11,true);
  this->addObstacle(FXString("obs12"),obs12,true);
  this->addObstacle(FXString("obs13"),obs13,true);
  this->addObstacle(FXString("obs14"),obs14,true);
  this->addObstacle(FXString("obs15"),obs15,true);
  this->addObstacle(FXString("obs16"),obs16,true);
  this->addObstacle(FXString("obs17"),obs17,true);
  this->addObstacle(FXString("obs18"),obs18,true);
  this->addObstacle(FXString("obs19"),obs19,true);
  this->addObstacle(FXString("obs20"),obs20,true);

  // create the static obstacles
  // borders obstacles
  ConfigR2D pos1(-5.0,0.5*dfltYMap);
  Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
  ConfigR2D pos2((double)dfltXMap + 5.0,0.5*dfltYMap);
  Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
  ConfigR2D pos3(0.5*dfltXMap,-5.0);
  Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 10.0, 0.0);
  ConfigR2D pos4(0.5*dfltXMap,(double)dfltYMap + 5.0);
  Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 10.0, 0.0);

  // inside obstacles
  ConfigR2D pos5(100,175);
  Obstacle2D* stObs5 = new ObsRect2D(pos5,200, 30.0, 0.0);
  ConfigR2D pos6(350,200);
  Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.5*M_PI);
  ConfigR2D pos7(250,350);
  Obstacle2D* stObs7 = new ObsRect2D(pos7,200, 30.0, 0.0);
  ConfigR2D pos8(450,200);
  Obstacle2D* stObs8 = new ObsRect2D(pos8,200, 30.0, 0.0);
  ConfigR2D pos9(225,600);
  Obstacle2D* stObs9 = new ObsRect2D(pos9,200, 30.0, 0.5*M_PI);
  ConfigR2D pos10(175,515);
  Obstacle2D* stObs10 = new ObsRect2D(pos10,100, 30.0, 0.0);
  ConfigR2D pos11(305,585);
  Obstacle2D* stObs11 = new ObsRect2D(pos11,150, 30.0, 0.0);
  ConfigR2D pos12(625,400);
  Obstacle2D* stObs12 = new ObsRect2D(pos12,150, 30.0, 0.0);
  ConfigR2D pos13(565,515);
  Obstacle2D* stObs13 = new ObsRect2D(pos13,200, 30.0, 0.5*M_PI);

//   this->addStObstacle(FXString("stObs1"),stObs1);
//   this->addStObstacle(FXString("stObs2"),stObs2);
//   this->addStObstacle(FXString("stObs3"),stObs3);
//   this->addStObstacle(FXString("stObs4"),stObs4);
//   this->addStObstacle(FXString("stObs5"),stObs5);
//   this->addStObstacle(FXString("stObs6"),stObs6);
//   this->addStObstacle(FXString("stObs7"),stObs7);
//   this->addStObstacle(FXString("stObs8"),stObs8);
//   this->addStObstacle(FXString("stObs9"),stObs9);
//   this->addStObstacle(FXString("stObs10"),stObs10);
//   this->addStObstacle(FXString("stObs11"),stObs11);
//   this->addStObstacle(FXString("stObs12"),stObs12);
//   this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

  // refresh the viewer
  this->drawScene();

  msgBox->setText("Environment 3 loaded (car-like system environment).");
}

// labyrinth with dynamic obstacles environment
void STDeformIfc::initObs3()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  // create the dynamic obstacles
  DynamicObs* obs1 = new ObsDisk2D(ConfigR2D(150,187), Speed2D(12,-10), 0, 500, 10);
  DynamicObs* obs2 = new ObsDisk2D(ConfigR2D(125,0), Speed2D(6,20), 0, 500, 13);
  DynamicObs* obs3 = new ObsDisk2D(ConfigR2D(263,38), Speed2D(-16,10), 0, 500, 17);
  DynamicObs* obs4 = new ObsDisk2D(ConfigR2D(38,315), Speed2D(20,-8), 0, 500, 10);
  DynamicObs* obs5 = new ObsDisk2D(ConfigR2D(170,263), Speed2D(10,16), 0, 500, 13);
  DynamicObs* obs6 = new ObsDisk2D(ConfigR2D(300,150), Speed2D(-20,-4), 0, 500, 17);
  DynamicObs* obs7 = new ObsDisk2D(ConfigR2D(38,150), Speed2D(16,10), 0, 500, 10);
  DynamicObs* obs8 = new ObsDisk2D(ConfigR2D(187,105), Speed2D(0,24), 0, 500, 13);
  DynamicObs* obs9 = new ObsDisk2D(ConfigR2D(285,338), Speed2D(-14,-14), 0, 500, 17);
  DynamicObs* obs10 = new ObsDisk2D(ConfigR2D(338,225), Speed2D(12,-12), 0, 500, 10);

  DynamicObs* obs11 = new ObsDisk2D(ConfigR2D(550,287), Speed2D(-12,10), 0, 500, 10);
  DynamicObs* obs12 = new ObsDisk2D(ConfigR2D(75,600), Speed2D(-6,-20), 0, 500, 13);
  DynamicObs* obs13 = new ObsDisk2D(ConfigR2D(363,538), Speed2D(16,-10), 0, 500, 17);
  DynamicObs* obs14 = new ObsDisk2D(ConfigR2D(238,615), Speed2D(-20,8), 0, 500, 10);
  DynamicObs* obs15 = new ObsDisk2D(ConfigR2D(570,563), Speed2D(-10,-16), 0, 500, 13);
  DynamicObs* obs16 = new ObsDisk2D(ConfigR2D(650,150), Speed2D(20,4), 0, 500, 17);
  DynamicObs* obs17 = new ObsDisk2D(ConfigR2D(438,450), Speed2D(-16,-10), 0, 500, 10);
  DynamicObs* obs18 = new ObsDisk2D(ConfigR2D(287,405), Speed2D(0,-24), 0, 500, 13);
  DynamicObs* obs19 = new ObsDisk2D(ConfigR2D(485,598), Speed2D(14,14), 0, 500, 17);
  DynamicObs* obs20 = new ObsDisk2D(ConfigR2D(638,325), Speed2D(-12,12), 0, 500, 10);

  // append it to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  this->addObstacle(FXString("obs4"),obs4,true);
  this->addObstacle(FXString("obs5"),obs5,true);
  this->addObstacle(FXString("obs6"),obs6,true);
  this->addObstacle(FXString("obs7"),obs7,true);
  this->addObstacle(FXString("obs8"),obs8,true);
  this->addObstacle(FXString("obs9"),obs9,true);
  this->addObstacle(FXString("obs10"),obs10,true);

  this->addObstacle(FXString("obs11"),obs11,true);
  this->addObstacle(FXString("obs12"),obs12,true);
  this->addObstacle(FXString("obs13"),obs13,true);
  this->addObstacle(FXString("obs14"),obs14,true);
  this->addObstacle(FXString("obs15"),obs15,true);
  this->addObstacle(FXString("obs16"),obs16,true);
  this->addObstacle(FXString("obs17"),obs17,true);
  this->addObstacle(FXString("obs18"),obs18,true);
  this->addObstacle(FXString("obs19"),obs19,true);
  this->addObstacle(FXString("obs20"),obs20,true);

  // create the static obstacles
  // borders obstacles
//   ConfigR2D pos1(-5.0,0.5*dfltYMap);
//   Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
//   ConfigR2D pos2((double)dfltXMap + 5.0,0.5*dfltYMap);
//   Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
//   ConfigR2D pos3(0.5*dfltXMap,-5.0);
//   Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 10.0, 0.0);
//   ConfigR2D pos4(0.5*dfltXMap,(double)dfltYMap + 5.0);
//   Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 10.0, 0.0);

  ConfigR2D pos1(-50.0,0.5*dfltYMap);
  Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 200.0, 100.0, 0.5*M_PI);
  ConfigR2D pos2((double)dfltXMap + 50.0,0.5*dfltYMap);
  Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 200.0, 100.0, 0.5*M_PI);
  ConfigR2D pos3(0.5*dfltXMap,-50.0);
  Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 100.0, 0.0);
  ConfigR2D pos4(0.5*dfltXMap,(double)dfltYMap + 50.0);
  Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 100.0, 0.0);

  // inside obstacles
  ConfigR2D pos5(100,175);
  Obstacle2D* stObs5 = new ObsRect2D(pos5,200, 30.0, 0.0);
  ConfigR2D pos6(350,200);
  Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.5*M_PI);
  ConfigR2D pos7(250,350);
  Obstacle2D* stObs7 = new ObsRect2D(pos7,200, 30.0, 0.0);
  ConfigR2D pos8(450,200);
  Obstacle2D* stObs8 = new ObsRect2D(pos8,200, 30.0, 0.0);
  ConfigR2D pos9(225,600);
  Obstacle2D* stObs9 = new ObsRect2D(pos9,200, 30.0, 0.5*M_PI);
  ConfigR2D pos10(175,515);
  Obstacle2D* stObs10 = new ObsRect2D(pos10,100, 30.0, 0.0);
  ConfigR2D pos11(305,585);
  Obstacle2D* stObs11 = new ObsRect2D(pos11,150, 30.0, 0.0);
  ConfigR2D pos12(625,400);
  Obstacle2D* stObs12 = new ObsRect2D(pos12,150, 30.0, 0.0);
  ConfigR2D pos13(565,515);
  Obstacle2D* stObs13 = new ObsRect2D(pos13,200, 30.0, 0.5*M_PI);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);
  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

  // refresh the viewer
  this->drawScene();

  msgBox->setText("Environment 3 loaded (car-like system environment).");
}

// not used...
void STDeformIfc::initObs4()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  for(int i=0; i<40; i++)
    {
      double x = getIntRandom(0,dfltXMap);
      double y = getIntRandom(0,dfltYMap);
      double sx = getIntRandom(-30,30);
      double sy = getIntRandom(-30,30);
      double initTime = getIntRandom(0,15);
      double lifeTime = getIntRandom(50,200);
      double sizeObs = getIntRandom(10,20);
      
      // create the obstacles
      DynamicObs* newObs = new ObsDisk2D(ConfigR2D(x,y), Speed2D(sx,sy), initTime, lifeTime, sizeObs);
      // append it to the list of obstacles
      this->addObstacle(FXString("obs"),newObs,true);
    }

  // create the static obstacles
  // borders obstacles
  ConfigR2D pos1(-5.0,0.5*dfltYMap);
  Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
  ConfigR2D pos2((double)dfltXMap + 5.0,0.5*dfltYMap);
  Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
  ConfigR2D pos3(0.5*dfltXMap,-5.0);
  Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 10.0, 0.0);
  ConfigR2D pos4(0.5*dfltXMap,(double)dfltYMap + 5.0);
  Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 10.0, 0.0);

  // inside obstacles
  ConfigR2D pos5(100,175);
  Obstacle2D* stObs5 = new ObsRect2D(pos5,200, 30.0, 0.0);
  ConfigR2D pos6(350,200);
  Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.5*M_PI);
  ConfigR2D pos7(250,350);
  Obstacle2D* stObs7 = new ObsRect2D(pos7,200, 30.0, 0.0);
  ConfigR2D pos8(450,200);
  Obstacle2D* stObs8 = new ObsRect2D(pos8,200, 30.0, 0.0);
  ConfigR2D pos9(225,600);
  Obstacle2D* stObs9 = new ObsRect2D(pos9,200, 30.0, 0.5*M_PI);
  ConfigR2D pos10(175,515);
  Obstacle2D* stObs10 = new ObsRect2D(pos10,100, 30.0, 0.0);
  ConfigR2D pos11(305,585);
  Obstacle2D* stObs11 = new ObsRect2D(pos11,150, 30.0, 0.0);
  ConfigR2D pos12(625,400);
  Obstacle2D* stObs12 = new ObsRect2D(pos12,150, 30.0, 0.0);
  ConfigR2D pos13(565,515);
  Obstacle2D* stObs13 = new ObsRect2D(pos13,200, 30.0, 0.5*M_PI);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);
  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

  // compute the roadmap on the environment
  roadmap_ = new RoadMap(this->env_);
  roadmap_->setNbNodesNeeded(3000);
  roadmap_->setMaxDistNgh(25.0);
  roadmap_->construct();

  // refresh the viewer
  this->drawScene();

  msgBox->setText("Environment 6 loaded.");
}

// motorway environment
void STDeformIfc::initObs5()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(1420);
  this->env_->setDfltY(680);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());
  
  // create the static obstacles

  ConfigR2D pos1(610,515);
  //ConfigR2D pos1(610,465);
  Obstacle2D* stObs1 = new ObsRect2D(pos1,1220, 30.0, 0.0);
  ConfigR2D pos2(920,400);
  Obstacle2D* stObs2 = new ObsRect2D(pos2,600, 30.0, 0.01*M_PI);
  ConfigR2D pos3(245,400);
  Obstacle2D* stObs3 = new ObsRect2D(pos3,490, 30.0, 0.0);

  ConfigR2D pos4(282,297);
  Obstacle2D* stObs4 = new ObsRect2D(pos4,450, 30.0, 0.15*M_PI);
  ConfigR2D pos5(360,258);
  Obstacle2D* stObs5 = new ObsRect2D(pos5,600, 30.0, 0.15*M_PI);


  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);

  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);

  // create the dynamic obstacles
  DynamicObs* obs1  = new ObsDisk2D(ConfigR2D(330,436), Speed2D(32,0),  0,    27.5, 20);
  DynamicObs* obs2  = new ObsDisk2D(ConfigR2D(180,436),  Speed2D(32,0),  0,    32,   20);
  DynamicObs* obs3  = new ObsDisk2D(ConfigR2D(30,436),   Speed2D(32,0),  0,    37.5, 20);
  DynamicObs* obs4  = new ObsDisk2D(ConfigR2D(0,  436), Speed2D(32,0),  3.5,  37.5,   20);
  DynamicObs* obs5  = new ObsDisk2D(ConfigR2D(0,  436), Speed2D(32,0),  8,   37.5, 20);
  DynamicObs* obs6  = new ObsDisk2D(ConfigR2D(350,479), Speed2D(40,0),  0,    28,   20);
  DynamicObs* obs7  = new ObsDisk2D(ConfigR2D(200,479), Speed2D(40,0),  0,    32,   20);
  DynamicObs* obs8  = new ObsDisk2D(ConfigR2D(50, 479), Speed2D(40,0),  0,    36,   20);
  DynamicObs* obs9  = new ObsDisk2D(ConfigR2D(0,  479), Speed2D(40,0),  2.5,  37.5,   20);
  DynamicObs* obs10 = new ObsDisk2D(ConfigR2D(0,  479), Speed2D(40,0),  6.5,  37.5,   20);

  // append them to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  this->addObstacle(FXString("obs4"),obs4,true);
  this->addObstacle(FXString("obs5"),obs5,true);
  this->addObstacle(FXString("obs6"),obs6,true);
  this->addObstacle(FXString("obs7"),obs7,true);
  this->addObstacle(FXString("obs8"),obs8,true);
  this->addObstacle(FXString("obs9"),obs9,true);
  this->addObstacle(FXString("obs10"),obs10,true);

  // refresh the viewer
  this->drawScene();
  
  // init the random
  // initRandomDeform();

  msgBox->setText("Environment \"Motorway\" loaded.");

  // compute the intersection trajectory
  orthoView = true;
  dynamicView = true;
  this->computeMotorwayTraj();
}

// labyrinth with static obstacles and roadmap environment
void STDeformIfc::initObs6()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  // create the dynamic obstacles
  DynamicObs* obs1 = new ObsDisk2D(ConfigR2D(150,187), Speed2D(12,-10), 0, 500, 10);
  DynamicObs* obs2 = new ObsDisk2D(ConfigR2D(75,0), Speed2D(6,20), 0, 500, 13);
  DynamicObs* obs3 = new ObsDisk2D(ConfigR2D(263,38), Speed2D(-16,10), 0, 500, 17);
  DynamicObs* obs4 = new ObsDisk2D(ConfigR2D(38,315), Speed2D(20,-8), 0, 500, 10);
  DynamicObs* obs5 = new ObsDisk2D(ConfigR2D(170,263), Speed2D(10,16), 0, 500, 13);
  DynamicObs* obs6 = new ObsDisk2D(ConfigR2D(300,150), Speed2D(-20,-4), 0, 500, 17);
  DynamicObs* obs7 = new ObsDisk2D(ConfigR2D(38,150), Speed2D(16,10), 0, 500, 10);
  DynamicObs* obs8 = new ObsDisk2D(ConfigR2D(187,105), Speed2D(0,24), 0, 500, 13);
  DynamicObs* obs9 = new ObsDisk2D(ConfigR2D(285,338), Speed2D(-14,-14), 0, 500, 17);
  DynamicObs* obs10 = new ObsDisk2D(ConfigR2D(338,225), Speed2D(12,-12), 0, 500, 10);

  DynamicObs* obs11 = new ObsDisk2D(ConfigR2D(550,287), Speed2D(-12,10), 0, 500, 10);
  DynamicObs* obs12 = new ObsDisk2D(ConfigR2D(75,600), Speed2D(-6,-20), 0, 500, 13);
  DynamicObs* obs13 = new ObsDisk2D(ConfigR2D(363,538), Speed2D(16,-10), 0, 500, 17);
  DynamicObs* obs14 = new ObsDisk2D(ConfigR2D(238,615), Speed2D(-20,8), 0, 500, 10);
  DynamicObs* obs15 = new ObsDisk2D(ConfigR2D(570,563), Speed2D(-10,-16), 0, 500, 13);
  DynamicObs* obs16 = new ObsDisk2D(ConfigR2D(650,150), Speed2D(20,4), 0, 500, 17);
  DynamicObs* obs17 = new ObsDisk2D(ConfigR2D(438,450), Speed2D(-16,-10), 0, 500, 10);
  DynamicObs* obs18 = new ObsDisk2D(ConfigR2D(287,405), Speed2D(0,-24), 0, 500, 13);
  DynamicObs* obs19 = new ObsDisk2D(ConfigR2D(485,598), Speed2D(14,14), 0, 500, 17);
  DynamicObs* obs20 = new ObsDisk2D(ConfigR2D(638,325), Speed2D(-12,12), 0, 500, 10);


  // append it to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  this->addObstacle(FXString("obs4"),obs4,true);
  this->addObstacle(FXString("obs5"),obs5,true);
  this->addObstacle(FXString("obs6"),obs6,true);
  this->addObstacle(FXString("obs7"),obs7,true);
  this->addObstacle(FXString("obs8"),obs8,true);
  this->addObstacle(FXString("obs9"),obs9,true);
  this->addObstacle(FXString("obs10"),obs10,true);

  this->addObstacle(FXString("obs11"),obs11,true);
  this->addObstacle(FXString("obs12"),obs12,true);
  this->addObstacle(FXString("obs13"),obs13,true);
  this->addObstacle(FXString("obs14"),obs14,true);
  this->addObstacle(FXString("obs15"),obs15,true);
  this->addObstacle(FXString("obs16"),obs16,true);
  this->addObstacle(FXString("obs17"),obs17,true);
  this->addObstacle(FXString("obs18"),obs18,true);
  this->addObstacle(FXString("obs19"),obs19,true);
  this->addObstacle(FXString("obs20"),obs20,true);
  
//   // create the static obstacles
//   ConfigR2D pos1(-5.0,0.5*dfltYMap);
//   Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
//   ConfigR2D pos2((double)dfltXMap + 5.0,0.5*dfltYMap);
//   Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
//   ConfigR2D pos3(0.5*dfltXMap,-5.0);
//   Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 10.0, 0.0);
//   ConfigR2D pos4(0.5*dfltXMap,(double)dfltYMap + 5.0);
//   Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 10.0, 0.0);
//   ConfigR2D pos7(0.25*dfltXMap,0.3*dfltYMap);
//   Obstacle2D* stObs7 = new ObsRect2D(pos7,0.65*dfltXMap, 30.0, 0.5*M_PI);
//   ConfigR2D pos8(0.5*dfltXMap,0.7*dfltYMap);
//   Obstacle2D* stObs8 = new ObsRect2D(pos8,0.65*dfltXMap, 30.0, 0.5*M_PI);
//   ConfigR2D pos9(0.75*dfltXMap,0.3*dfltYMap);
//   Obstacle2D* stObs9 = new ObsRect2D(pos9,0.65*dfltXMap, 30.0, 0.5*M_PI);


  // create the static obstacles
  // borders obstacles
  ConfigR2D pos1(-5.0,0.5*dfltYMap);
  Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
  ConfigR2D pos2((double)dfltXMap + 5.0,0.5*dfltYMap);
  Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
  ConfigR2D pos3(0.5*dfltXMap,-5.0);
  Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 10.0, 0.0);
  ConfigR2D pos4(0.5*dfltXMap,(double)dfltYMap + 5.0);
  Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 10.0, 0.0);

  // inside obstacles
  ConfigR2D pos5(100,175);
  Obstacle2D* stObs5 = new ObsRect2D(pos5,200, 30.0, 0.0);
  ConfigR2D pos6(350,200);
  Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.5*M_PI);
  ConfigR2D pos7(250,350);
  Obstacle2D* stObs7 = new ObsRect2D(pos7,200, 30.0, 0.0);
  ConfigR2D pos8(450,200);
  Obstacle2D* stObs8 = new ObsRect2D(pos8,200, 30.0, 0.0);
  ConfigR2D pos9(225,600);
  Obstacle2D* stObs9 = new ObsRect2D(pos9,200, 30.0, 0.5*M_PI);
  ConfigR2D pos10(175,515);
  Obstacle2D* stObs10 = new ObsRect2D(pos10,100, 30.0, 0.0);
  ConfigR2D pos11(305,585);
  Obstacle2D* stObs11 = new ObsRect2D(pos11,150, 30.0, 0.0);
  ConfigR2D pos12(625,400);
  Obstacle2D* stObs12 = new ObsRect2D(pos12,150, 30.0, 0.0);
  ConfigR2D pos13(565,515);
  Obstacle2D* stObs13 = new ObsRect2D(pos13,200, 30.0, 0.5*M_PI);



  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);
  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

//   // test of copy of the roadmap
//   RoadMap* copyRM = new RoadMap(this->env_);
//   copyRM->construct();
//   cout << "copyRM->getNbNodes() : " << copyRM->getNbNodes() << endl;
//   roadmap_ = new RoadMap(*copyRM);
//   cout << "roadmap_->getNbNodes() : " << roadmap_->getNbNodes() << endl;

  // compute the roadmap on the environment
  roadmap_ = new RoadMap(this->env_);
  roadmap_->setNbNodesNeeded(3000);
  roadmap_->setMaxDistNgh(25.0);
  roadmap_->construct();

  // refresh the viewer
  this->drawScene();

  msgBox->setText("Environment 6 loaded (double integrator system environment) and road map constructed.");
}

void STDeformIfc::initObsRand(int nb)
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs();
    }

  for(int i=0; i<nb; i++)
    {
      double x = getIntRandom(0,dfltXMap);
      double y = getIntRandom(0,dfltYMap);
      double sx = getIntRandom(-15,15);
      double sy = getIntRandom(-15,15);
      double initTime = getIntRandom(0,15);
      double lifeTime = getIntRandom(20,200);
      double sizeObs = getIntRandom(10,30);
      
      // create the obstacles
      DynamicObs* newObs = new ObsDisk2D(ConfigR2D(x,y), Speed2D(sx,sy), initTime, lifeTime, sizeObs);
      // append it to the list of obstacles
      this->addObstacle(FXString("obs"),newObs,true);
    }

  // refresh the viewer
  this->drawScene();

  // init the random
  initRandomDeform();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Widget has been resized
long STDeformIfc::onConfigure(FXObject*,FXSelector,void*)
{
//   if(glcanvas->makeCurrent())
//     {
//       glViewport(0,0,glcanvas->getWidth(),glcanvas->getHeight());
//       glcanvas->makeNonCurrent();
//     }
  return 1;
}

long STDeformIfc::onExpose(FXObject*,FXSelector,void*)
{
  drawScene();
  return 1;
}

long STDeformIfc::loadPath(FXObject*,FXSelector,void*)
{
  // destroy the old path
  if(this->path_ != NULL)
    {
      delete this->path_;
    }

  // reinit the time already past
  alreadyElapsedTime = 0;
  FPS = 0;
  nbTimeout = 0;
  totalNbTimeout = 0;

//   // destroy the old trajectory
//   if(this->traj_ != NULL)
//     {
//       delete this->traj_;
//     }

  // check the type of the path
  if(isExtFilename((char*)(inputPathName.text()), (char*)("pth")))
    {
      // the path is a holonomic path : create a new holonomic path
      this->path_ = new Path;
      genTrajType = HOLONOM_TRAJ;
      trajType = HOLONOM_TRAJ;
    }
  else if(isExtFilename((char*)(inputPathName.text()), (char*)("nhp")))
    {
      // the path is a non holonomic path : create a new non holonomic path
      this->path_ = new NHPath;
      genTrajType = NH_TRAJ;
      trajType = NH_TRAJ;
    }
  else
    {
      // unknown path type
      FXMessageBox::error(this,MBOX_OK,"Error while loading the path","The input path filename extension is not recognized");
      return -1;
    }
  

  //load the path
  bool resLoad = this->path_->load((char*)(inputPathName.text()));

  if(!resLoad)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while loading the path","The input path filename isn't available");
      return -1;
    }

  // init the viewer
  ConfigR* center = path_->center();
  double radius = path_->radius();
  double depth = 5*radius;
  double dt = 0;

  this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, 0, 1, 0);
//   eye_ = FXVec3d(center->x(), center->y() + dt, depth);
//   center_ = FXVec3d(center->x(), center->y(), 0);
//   up_ = FXVec3d(0, 1, 0);
  
  delete center;

  // draw the scene
  this->drawScene();
  return 1;
}

long STDeformIfc::loadPrecompPath(FXObject*,FXSelector,void*)
{
//   // compute a straight line path until this position
//   ConfigR3D cfg1(50,100,0.5*M_PI);
//   ConfigR3D cfg2(50,300,0.5*M_PI);
//   ConfigR3D cfg3(50,300,0);
//   ConfigR3D cfg4(150,300,0);
//   ConfigR3D cfg5(150,300,-0.5*M_PI);
//   ConfigR3D cfg6(150,100,-0.5*M_PI);
//   ConfigR3D cfg7(150,100,0);
//   ConfigR3D cfg8(250,100,0);
//   ConfigR3D cfg9(250,100,0.5*M_PI);
//   ConfigR3D cfg10(250,300,0.5*M_PI);
//   ConfigR3D cfg11(250,300,0);
//   ConfigR3D cfg12(350,300,0);
//   ConfigR3D cfg13(350,300,-0.5*M_PI);
//   ConfigR3D cfg14(350,100,-0.5*M_PI);
  
//   this->path_ = new Path();
  
//   GeomPathPlanner gpp1(NULL, &cfg1, &cfg2);
//   gpp1.initStrLineCptg();
//   int resPP1 = gpp1.computePath();
//   Path* path1 = gpp1.getPath()->getPath2D();
//   path_->concat(*path1);

//   GeomPathPlanner gpp2(NULL, &cfg3, &cfg4);
//   gpp2.initStrLineCptg();
//   int resPP2 = gpp2.computePath();
//   Path* path2 = gpp2.getPath()->getPath2D();
//   path_->concat(*path2);

//   GeomPathPlanner gpp3(NULL, &cfg5, &cfg6);
//   gpp3.initStrLineCptg();
//   int resPP3 = gpp3.computePath();
//   Path* path3 = gpp3.getPath()->getPath2D();
//   path_->concat(*path3);

//   GeomPathPlanner gpp4(NULL, &cfg7, &cfg8);
//   gpp4.initStrLineCptg();
//   int resPP4 = gpp4.computePath();
//   Path* path4 = gpp4.getPath()->getPath2D();
//   path_->concat(*path4);

//   GeomPathPlanner gpp5(NULL, &cfg9, &cfg10);
//   gpp5.initStrLineCptg();
//   int resPP5 = gpp5.computePath();
//   Path* path5 = gpp5.getPath()->getPath2D();
//   path_->concat(*path5);

//   GeomPathPlanner gpp6(NULL, &cfg11, &cfg12);
//   gpp6.initStrLineCptg();
//   int resPP6 = gpp6.computePath();
//   Path* path6 = gpp6.getPath()->getPath2D();
//   path_->concat(*path6);

//   GeomPathPlanner gpp7(NULL, &cfg13, &cfg14);
//   gpp7.initStrLineCptg();
//   int resPP7 = gpp7.computePath();
//   Path* path7 = gpp7.getPath()->getPath2D();
//   path_->concat(*path7);

  if(roadmap_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Path","No roadmap has been defined");
      return -1;
    }

  // compute a path on the road map
  this->path_ = new Path();
  ConfigR2D cfg1(50,100);
  ConfigR2D cfg2(350,100);
  RoadMapPP rmpp(roadmap_,&cfg1,&cfg2);
  rmpp.computePath();
  Path* path = rmpp.getPath()->getPath2D();
  path_->concat(*path);

  // compute the trajectory until this position
  int res = this->computeTrajectory(NULL, 0, NULL);
}

long STDeformIfc::loadMap(FXObject*,FXSelector,void*)
{
  // Load the map
  bool resLoad = env_->loadMap((char*)(mapName.text()));

  if(resLoad != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Loading the Map","Map's filename doesn't exist");
      return -1;
    }

  this->loadMapInViewer();
  return 1;
}

long STDeformIfc::computeTrajectory(FXObject*,FXSelector,void*)
{
  // init the robot configuration
  this->rb_->setVMax(maxSpd);
  this->rb_->setAMax(maxAcc);
  this->rb_->setMaxSteering(maxSteer);
  this->rb_->setMaxSteerSpd(maxSteerS);
  this->rb_->setLength(robotLength);

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  if(this->path_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","No init path has been defined");
      return -1;
    }

  // init the general trajectory type
  if(typeMotion == 1)
    {
      if(trajType == HOLONOM_TRAJ) genTrajType = HOLONOM_TRAJ;
      else genTrajType = NH_TRAJ;
    }

  PVDTrajPlanner tp;

  if(genTrajType == HOLONOM_TRAJ)
    {
      tp = PVDTrajPlanner(rb_,env_,StateR5D(((*(path_))[0])->x(),((*(path_))[0])->y(),0,0,0),StateR5D(((*(path_))[path_->getSize()-1])->x(),((*(path_))[path_->getSize()-1])->y(),0,0,0),path_);
    }
  else
    {
      tp = PVDTrajPlanner(rb_,env_,StateS6D(((*(path_))[0])->x(),((*(path_))[0])->y(),((*(path_))[0])->th(),0,0,0),StateS6D(((*(path_))[path_->getSize()-1])->x(),((*(path_))[path_->getSize()-1])->y(),((*(path_))[path_->getSize()-1])->th(),0,0,0),path_);
    }

  tp.setTimeStep(trajTimeStep);
  int resTP = tp.computeTraj();

  if(resTP != TRAJ_CPTD)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
      return -2;
    }
  else
    {
      if(this->traj_ != NULL) delete traj_;

      this->traj_ = tp.getTraj();
      this->traj_->shift(time);

      if(this->traj_->getSize() != 0) slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
      curStep = 0;
      //time = 0;
      //this->traj_->display();

      // init the table of state types
      //cout << "this->traj_->getSize() : " << this->traj_->getSize() << endl;
      typeState = new int*;
      (*typeState) = new int[this->traj_->getSize()];
      for(int i=0; i < this->traj_->getSize(); i++)
	{
	  (*typeState)[i] = 0;
	}

      //init the table of nearest states
      tableNearState = new StateR**;
      (*tableNearState) = new StateR* [this->traj_->getSize()];
      for(int i=0; i < this->traj_->getSize(); i++)
	{
	  (*tableNearState)[i] = NULL;
	}

      // init the tables of internal and external forces
      tableFInt = new StateR*;
      *tableFInt = NULL;
      tableFExt = new StateR*;
      *tableFExt = NULL;

      // init the trajectory deformer
      if(deformer_ != NULL) delete deformer_;


      if(genTrajType == HOLONOM_TRAJ)
	{
	  deformer_ = new STDeformation(this->traj_, this->env_, this->rb_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
	}
      else
	{
// 	  cout << "input trajectory 111 : " << endl;
// 	  this->traj_->display();

	  deformer_ = new STNHDeformation(this->traj_, this->env_, this->rb_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
	  ((STNHDeformation*)deformer_)->setParallelExec(parallelize);
	  ((STNHDeformation*)deformer_)->setNbThread(nbThread);
	}

      if(optModeDeform == 0) deformer_->setModeDeform(SPATIOTEMP_DEFORM);
      else if(optModeDeform == 1) deformer_->setModeDeform(SPATONLY_DEFORM);
      else if(optModeDeform == 2) deformer_->setModeDeform(TEMPONLY_DEFORM);
      deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &K_SPT_TMP);
      deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &alphaMin, &alphaMax, &betaMin, &betaMax);
//       deformer_->setMaxEffectDist(&maxEffectDist);
//       deformer_->setMaxStEffectDist(&maxStEffectDist);
      deformer_->setMaxEffectDist(&maxSpDist,&maxTmpDist,&maxStEffectDist);
      deformer_->setReliability(&predHorizon, &reliabilityTime, &epsilonRT);
      deformer_->setTemporalConst(&useTemporalConsistency);
      deformer_->setTimeStep(trajTimeStep);
      //deformer_->setPosStep(trajTimeStep * maxSpd);
      deformer_->setPosStep(trajPosStep);

      // init the trajectory generation parameters
      DDTrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
      DDTrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
      DDTrajGen::setDBG(debugLevel);                     // set the debugging level
      CLTrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
      CLTrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
      CLTrajGen::setDBG(debugLevel);                     // set the debugging level
      CLTrajGen::setUseVelCst(useVelCst);                // set true if linear velocity constraints are used
      CLTrajGen::setUseAccCst(useAccCst);                // set true if linear acceleration constraints are used
      CLTrajGen::setUseAVelCst(useAVelCst);              // set true if angular velocity constraints are used
      CLTrajGen::setUseAAccCst(useAAccCst);              // set true if angular acceleration constraints are used

      if(typeMotion == 0) // trajectory tracking
	{
	  // init the viewer
	  ConfigR* center = path_->center();
	  double radius = path_->radius();
	  double depth = 5*radius;
	  double dt = 3*(*traj_)[traj_->getSize()-1]->t();
	  double upx,upy,upz;
	  
	  Vec v1(3);
	  v1.setVal(0,0);
	  v1.setVal(1,-dt);
	  v1.setVal(2,-depth);
	  double v1Norm = v1.norm();
	  Vec v2(3);
	  v2.setVal(0,0);
	  v2.setVal(1,-dt/v1Norm);
	  v2.setVal(2,-depth/v1Norm);
	  upx = 0;
	  upy = -v2[2];
	  upz = v2[1];

	  this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
	  //       eye_ = FXVec3d(center->x(), center->y() + dt, depth);
	  //       center_ = FXVec3d(center->x(), center->y(), 0);
	  //       up_ = FXVec3d(upx, upy, upz);
	  
	  delete center;

	  // draw the scene
	  this->drawScene();
	}
//       cout << "deform:ctraj41" << endl;
//       cout << "traj[0] : "; (*(this->traj_))[0]->display(); cout << endl << endl; 
//       cout << "deform:ctraj42" << endl;

      return 1;
    }
}

long STDeformIfc::computeRandRKTraj(FXObject*,FXSelector,void*)
{
  StateS6D initCfg((double)initXRb,(double)initYRb,0,10.0,0,0);
  return this->computeNewCarLikeTraj(initCfg);
}

long STDeformIfc::addObstacle(FXObject*,FXSelector,void*)
{
  int id;
  
  if(lifeT < 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while adding obstacle","The life of time must be positive");
      return -2;
    }

  if(randObs != FXString("yes") && randObs != FXString("no"))
    {
      FXMessageBox::error(this,MBOX_OK,"Error while adding obstacle","The random mode must be 'yes' or 'no'");
      return -3;
    }

  if((id = this->elemListExists(idObs)) != -1)
    {
      // change the display list parameters
      list[3]->setItem(id,FXStringVal(sxObs,2,FALSE));
      list[4]->setItem(id,FXStringVal(syObs,2,FALSE));
      list[8]->setItem(id,randObs);

      // change the obstacle parameters
      time = time + 0.0001;
      this->env_->getObs(id)->setLastSpd(new Speed2D(sxObs,syObs));
      
      if(randObs == FXString("yes"))
	{
	  this->env_->getObs(id)->setRdmMvt(true);
	}
      else
	{
	  this->env_->getObs(id)->setRdmMvt(false);
	}

      return 1;
    }

  // create the obstacle list element
  list[0]->appendItem(idObs);
  list[1]->appendItem(FXStringVal(xObs,2,FALSE));
  list[2]->appendItem(FXStringVal(yObs,2,FALSE));
  list[3]->appendItem(FXStringVal(sxObs,2,FALSE));
  list[4]->appendItem(FXStringVal(syObs,2,FALSE));
  list[5]->appendItem(FXStringVal(rObs,2,FALSE));
  list[6]->appendItem(FXStringVal(initT,2,FALSE));
  list[7]->appendItem(FXStringVal(lifeT,2,FALSE));
  list[8]->appendItem(randObs);

  // create the obstacle
  ConfigR2D cfgO(xObs,yObs);
  Speed2D spdO(sxObs,syObs);
  DynamicObs* obs = new ObsDisk2D(ConfigR2D(xObs,yObs), Speed2D(sxObs,syObs), initT, lifeT, rObs);

  if(randObs == FXString("yes"))
	{
	  obs->setRdmMvt(true);
	}
      else
	{
	  obs->setRdmMvt(false);
	}

  // append it to the list of obstacles
  this->env_->appendObs(obs);
  
  // repaint the viewer
  this->drawScene();

  return 1;
}

long STDeformIfc::addObstacle(FXString id, DynamicObs* obs, bool rand)
{
  // create the obstacle list element
  list[0]->appendItem(id);
  list[1]->appendItem(FXStringVal(obs->getInitCfg()->x(),2,FALSE));
  list[2]->appendItem(FXStringVal(obs->getInitCfg()->y(),2,FALSE));
  list[3]->appendItem(FXStringVal(obs->getInitSpd()->sx(),2,FALSE));
  list[4]->appendItem(FXStringVal(obs->getInitSpd()->sy(),2,FALSE));
  list[5]->appendItem(FXStringVal(((ObsDisk2D*)obs)->getRadius(),2,FALSE));
  list[6]->appendItem(FXStringVal(obs->getInitTime(),2,FALSE));
  list[7]->appendItem(FXStringVal(obs->getLifeTime(),2,FALSE));
  if(rand) list[8]->appendItem(FXString("yes"));
  else list[8]->appendItem(FXString("no"));

  if(rand) obs->setRdmMvt(true);
  else obs->setRdmMvt(false);

  // append it to the list of obstacles
  this->env_->appendObs(obs);
  
  // repaint the viewer
  this->drawScene();

  return 1;
}

long STDeformIfc::addStObstacle(FXString id, Obstacle2D* obs)
{
  // append it to the list of obstacles
  this->env_->appendStObs(obs);
  
  // repaint the viewer
  this->drawScene();
  
  return 1;
}

long STDeformIfc::removeObstacle(FXObject*,FXSelector,void*)
{
  int id;
  if((id = this->elemListExists(idObs)) != -1)
    {
      for(int i=0; i<8; i++)
	{
	  list[i]->removeItem(id);
	}
      this->env_->removeObs(id);

      // repaint the viewer
      this->drawScene();
    
      return 1;
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Error while removing obstacle","The obstacle ID doesn't exists");
      return -1;
    }
}

long STDeformIfc::changeInputPathName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Input path");
  open.setFilename(inputPathName);
  if(open.execute()){
    inputPathName = open.getFilename();
  }
  return 1;
}

long STDeformIfc::changeMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Environment map");
  open.setFilename(mapName);
  if(open.execute()){
    mapName = open.getFilename();
  }
  return 1;
}

// Clicked a header button:- we highlight all in the list
long STDeformIfc::onCmdHeaderButton(FXObject*,FXSelector,void* /* ptr */){
//   FXint which=(FXint)(long)ptr;
//   FXint i;
//   FXASSERT(0<=which && which<5);
//   for(i=0; i<list[which]->getNumItems(); i++){
//     list[which]->selectItem(i);
//   }
  return 1;
}

long STDeformIfc::updateScene(FXObject*,FXSelector,void*)
{
  // update obstacles random motion parameters
  DynamicObs::setSeuilChgt(seuilChgt);
  DynamicObs::setMaxAngleChgt(maxAngleChgt);
  DynamicObs::setMaxNormChgt(maxNormChgt);

  // compute the time before the deformation
  struct timeval t3;
  gettimeofday(&t3,NULL);

  // deform the trajectory
  int status = this->deformTraj();

//   if(this->traj_ != NULL)
//     {
//       cout << "last state-time : ";
//       (*this->traj_)[this->traj_->getSize()-1]->display();
//       cout << endl;
//     }

  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;

  double elapsedT = alreadyElapsedTime + ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);

  // cout << "timeout : \t Ts : "<< elapsedTs << "\t Tus : " << elapsedTus << " \t time : " << elapsedT << endl;

  // set the current time
  if(realTimeAlgo) time = elapsedT;
  else{
    if(elapsedT - time < 0.5)
      {
	cout << "inf 0.5sec" << endl;
	sleep((int)((0.5 - elapsedT + time)*1000.0));
      }
    time = time + 0.5;
  }
  // compute the deformation duration
  time_t deformTs = t2.tv_sec - t3.tv_sec;
  suseconds_t deformTus = t2.tv_usec - t3.tv_usec;
  deformDur = ((double)deformTs) + 0.000001 * ((double)deformTus);
  if(deformDur != 0) lastFPS = 1.0/deformDur;
  else lastFPS = 0.0;
  if(this->traj_ != NULL) nbST = this->traj_->getSize() - curStep - 1;
  else nbST = 0;
  if(this->traj_ != NULL) nbSTtot = this->traj_->getSize();
  else nbSTtot = 0;

  bool stop = false;

  switch(status)
    {
    case 0 : 
      // search the step of the trajectory reached at this time
      if(this->traj_ != NULL)
	{
	  if(curStep >= this->traj_->getSize()-1)
	    {
	      cout << "fin traj" << endl;
	      cout << "typeMotion : " << typeMotion << endl;
	      
	      if(typeMotion == 0) // trajectory tracking
		{
		  msgBox->setText("The trajectory deformation process achieved sucessfully.");
		  this->stopTraj(NULL,0,NULL);
		  stop = true;
		}
	      else // positions tracking
		{
		  if(trajType == HOLONOM_TRAJ)
		    {
		      cout << "new holonomic trajectory" << endl;
		      StateR5D newInitCfg(*((*(this->traj_))[this->traj_->getSize()-1]));
		      this->computeNewRandTrajFrom(newInitCfg);
		    }
		  else
		    {
		      FXMessageBox::error(this,MBOX_OK,"Car-like position tracking error","This function has not been implemented yet :-(");
		      msgBox->setText("This function has not been implemented yet :-(");
		      return -1;

// 		      StateS6D newInitCfg(*((*(this->traj_))[this->traj_->getSize()-1]));
// 		      this->computeNewCarLikeTraj(newInitCfg);
		    }
		}
	    }
	  else
	    {
	      bool trouve = false;
	      double oldCurStep = curStep;
	      curStep = 0;
	      while(!trouve)
		{
		  if(curStep >= 0 && curStep < this->traj_->getSize()-1 && time >= (*(this->traj_))[(int)(curStep+1)]->t())
		    {
		      curStep = curStep+1;
		      if((curStep > oldCurStep && pauseColLoc) && ((*typeState)[(int)curStep] == -1 || ((*typeState)[(int)curStep] == -2 && !useTemporalConsistency)))
			{
			  this->stopTraj(NULL,0,NULL);
			}
		    }
		  else
		    trouve = true;
		}
	    }
	}
      
      break;
    case -1 : 
      msgBox->setText("Warning. The input trajectory is null.");
      cout << "STDeformIfc::deformTraj() : Warning. The input trajectory is null." << endl;
      this->stopTraj(NULL, 0, NULL);
      break;
    case -2 : 
      msgBox->setText("Warning. A collision will occur during the next prediction horizon.");
      cout << "STDeformIfc::deformTraj() : Warning. A collision will occur during the next prediction horizon." << endl;
      this->stopTraj(NULL, 0, NULL);
      break;
    case -3 : 
      msgBox->setText("Warning. The connectivity has been lost during the next prediction horizon.");
      cout << "STDeformIfc::deformTraj() : Warning. The connectivity has been lost during the next prediction horizon." << endl;
      this->stopTraj(NULL, 0, NULL);
      break;    
    }
  
  // repaint the viewer
  this->drawScene();
  
  // rerun the timeout
  if(oneStep)
    {
      alreadyElapsedTime = time;
      oneStep = false;
    }
  else if(running)
    {
      getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);
    }

  // update the FPS number
  nbTimeout++;
  FPS = ((double)(totalNbTimeout + nbTimeout))/time;

  return 1;
}

long STDeformIfc::startTraj(FXObject*,FXSelector,void*)
{
  // verifying that the trajectory is not NULL
  if(this->traj_ == NULL && typeMotion == 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Deformation","The initial trajectory hasn't been computed");
      return -1;
    }

  // saving the initial trajectory
  if(time == 0)
    {
      if(this->initTraj_ != NULL) delete this->initTraj_;
      if(this->path_->getType() == HOLONOMIC_PATH)
	this->initTraj_ = new Trajectory5D(*((Trajectory5D*)(this->traj_)));
      else if(this->path_->getType() == NONHOLONOMIC_PATH)
	this->initTraj_ = new Trajectory6D(*((Trajectory6D*)(this->traj_)));
    }

  cout << "starting timeout" << endl;
  // init the timer
  gettimeofday(&t1,NULL);
  running = true;
  getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);
  return 1;
}

long STDeformIfc::oneStepTraj(FXObject* ,FXSelector ,void*)
{
  // verifying that the trajectory is not NULL
  if(this->traj_ == NULL && typeMotion == 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Deformation","The initial trajectory hasn't been computed");
      return -1;
    }
  
  // init the timer
  gettimeofday(&t1,NULL);
  oneStep = true;
  getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);
  return 1;
}

long STDeformIfc::stopTraj(FXObject*,FXSelector,void*)
{
  cout << "stoping timeout" << endl;
  running = false;
  getApp()->removeTimeout(this,ID_TIMEOUT);
  alreadyElapsedTime = time;
  return 1;
}

long STDeformIfc::restartTraj(FXObject*,FXSelector,void*)
{
  // reinit the trajectory
  if(this->initTraj_ != NULL)
    {
      if(this->traj_ != NULL) delete this->traj_;
      if(this->path_->getType() == HOLONOMIC_PATH)
	this->traj_ = new Trajectory5D(*((Trajectory5D*)(this->initTraj_)));
      else if(this->path_->getType() == NONHOLONOMIC_PATH)
	this->traj_ = new Trajectory6D(*((Trajectory6D*)(this->initTraj_)));
    }

  // reinit the environment
  if(this->env_ != NULL && this->env_->getNbObs() != -1)
    {
      for(int i=0; i<this->env_->getNbObs(); i++)
	{
	  this->env_->getObs(i)->reinit();
	}
    }

  // reinit the robot
  

  // reinit the time already past
  time = 0;
  alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout = 0;
  nbTimeout = 0;
  
  // repaint the viewer
  this->drawScene();

  cout << "nbObs : " << this->env_->getNbObs() << endl;
  cout << "stackSize : " << this->env_->getObs(0)->stackSize() << endl;
  cout << "initTime : " << this->env_->getObs(0)->getInitTime() << endl;
  cout << "stateStack : "; this->env_->getObs(0)->getStateStack(0)->display(); cout << endl;
  cout << "stateStack : "; this->env_->getObs(0)->getStateStack(1)->display(); cout << endl;

  return 1;
}

long STDeformIfc::repaint(FXObject*,FXSelector,void*)
{
  this->drawScene();
  return 1;
}

long STDeformIfc::timeView(FXObject*,FXSelector,void*)
{
  if(this->traj_ != NULL)
    {
      // init the viewer
      ConfigR* center = path_->center();
      double radius = path_->radius();
      double depth = 5*radius;
      double dt = (*traj_)[traj_->getSize()-1]->t();
      
      // change the point of view
      this->glcanvas->init(center->x(), -depth, dt/2, center->x(), center->y(), dt/2, 0, 0, 1);
      
      // repaint the viewer
      this->drawScene();

      return 1;
    }
  else return 0;
}

long STDeformIfc::spaceView(FXObject*,FXSelector,void*)
{
  if(this->traj_ != NULL)
    {
      // init the viewer
      ConfigR* center = path_->center();
      double radius = path_->radius();
      double depth = 5*radius;
      
      // change the point of view
      this->glcanvas->init(center->x(), center->y(), depth, center->x(), center->y(), 0, 0, 1, 0);
      
      // repaint the viewer
      this->drawScene();
      
      return 1;
    }
  else return 0;
}

long STDeformIfc::perspView(FXObject*,FXSelector,void*)
{
  if(this->traj_ != NULL)
    {
      // init the viewer
      ConfigR* center = path_->center();
      double radius = path_->radius();
      double depth = 5*radius;
      double dt = 3*(*traj_)[traj_->getSize()-1]->t();
      double upx,upy,upz;
      
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];
      
      // change the point of view
      this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
      
      // repaint the viewer
      this->drawScene();
      
      return 1;
    }
  else return 0;
}

long STDeformIfc::updateDispFuture(FXObject*,FXSelector,void*)
{
  // update the disp future mode
  if(dispFuture == false)
    dispFuture = true;
  else
    dispFuture = false;

  // repaint the viewer
      this->drawScene();

  return 1;
}

long STDeformIfc::updateDispForces(FXObject*,FXSelector,void*)
{
  // update the disp forces mode
  if(dispForces == false)
    dispForces = true;
  else
    dispForces = false;

  // repaint the viewer
  //this->drawScene();

  return 1;
}

long STDeformIfc::updateDispAllForces(FXObject*,FXSelector,void*)
{
  // update the disp forces mode
  if(dispAllForces == false)
    dispAllForces = true;
  else
    dispAllForces = false;

  // repaint the viewer
  //this->drawScene();

  return 1;
}

long STDeformIfc::updateDispNearSt(FXObject*,FXSelector,void*)
{
  // update the disp nearest states mode
  if(dispNearStates == false)
    dispNearStates = true;
  else
    dispNearStates = false;
  return 1;
}

long STDeformIfc::init(FXObject*,FXSelector,void*)
{
  initObs();
  return 1;
}

long STDeformIfc::init2(FXObject*,FXSelector,void*)
{
  initObs2();
  return 1;
}

long STDeformIfc::init3(FXObject*,FXSelector,void*)
{
  initObs3();
  return 1;
}

long STDeformIfc::init4(FXObject*,FXSelector,void*)
{
  initObs4();
  return 1;
}

long STDeformIfc::init5(FXObject*,FXSelector,void*)
{
  initObs5();
  return 1;
}

long STDeformIfc::init6(FXObject*,FXSelector,void*)
{
  initObs6();
  return 1;
}

long STDeformIfc::initRandObs(FXObject*,FXSelector,void*)
{
  initObsRand(nbRandObs);
  return 1;
}

long STDeformIfc::useConnecChk(FXObject*,FXSelector,void*)
{
  if(useConnectivityChecker == true) useConnectivityChecker = false;
  else useConnectivityChecker = true;
  return 1;
}

long STDeformIfc::useTempCstcy(FXObject*,FXSelector,void*)
{
  if(useTemporalConsistency == true) useTemporalConsistency = false;
  else useTemporalConsistency = true;
  return 1;
}

long STDeformIfc::usePauseColLoc(FXObject*,FXSelector,void*)
{
  if(pauseColLoc == true) pauseColLoc = false;
  else pauseColLoc = true;
  return 1;
}

long STDeformIfc::initPositionTracking(FXObject*,FXSelector,void*)
{
  // init the random
  initRandomDeform();

  // compute the first trajectory to follow
  if(trajType == HOLONOM_TRAJ)
    {
      StateR5D initSt(initXRb,initYRb,0,0,0);
      this->computeNewRandTrajFrom(initSt);
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Car-like position tracking error","This function has not been implemented yet :-(");
      msgBox->setText("This function has not been implemented yet :-(");
      return -1;

//       StateS6D initSt(initXRb,initYRb,initThRb,0,0,0);
//       this->computeNewRandRKTrajFrom(initSt);
    }

  this->traj_->display();

  // get back the dimensions of the map
  double dimMapX;
  double dimMapY;

  if(this->env_ != NULL && this->env_->getMap() != NULL)
    {
      dimMapX = this->env_->getMap()->getWidth();
      dimMapY = this->env_->getMap()->getHeight();
    }
  else if(this->env_ != NULL)
    {
      dimMapX = this->env_->getDfltX();
      dimMapY = this->env_->getDfltY();
    }
  else
    {
      dimMapX = dfltXMap;
      dimMapY = dfltYMap;
    }

  // init the viewer
  ConfigR* center = new ConfigR2D(dimMapX/2, dimMapY/2);
  //double radius = path_->radius();
  double radius = sqrt(dimMapX*dimMapX/4 * dimMapY*dimMapY/4);
  double depth = 0.022 * radius;
  double dt = 3*(*traj_)[traj_->getSize()-1]->t();
  double upx,upy,upz;

  Vec v1(3);
  v1.setVal(0,0);
  v1.setVal(1,-dt);
  v1.setVal(2,-depth);
  double v1Norm = v1.norm();
  Vec v2(3);
  v2.setVal(0,0);
  v2.setVal(1,-dt/v1Norm);
  v2.setVal(2,-depth/v1Norm);
  upx = 0;
  upy = -v2[2];
  upz = v2[1];

  this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
  //       eye_ = FXVec3d(center->x(), center->y() + dt, depth);
  //       center_ = FXVec3d(center->x(), center->y(), 0);
  //       up_ = FXVec3d(upx, upy, upz);
  
  delete center;

  // draw the scene
  this->drawScene();

  return 1;
}

long STDeformIfc::updateDrawMapBoundaries(FXObject*,FXSelector,void*)
{
  // update the drawing of the boundaries of the map
  if(drawMapBoundaries == false)
    drawMapBoundaries = true;
  else
    drawMapBoundaries = false;
  return 1;
}

long STDeformIfc::updateUseRTAlgo(FXObject*,FXSelector,void*)
{
  // update the drawing of the boundaries of the map
  if(realTimeAlgo == false)
    realTimeAlgo = true;
  else
    realTimeAlgo = false;
  return 1;
}

long STDeformIfc::updateDynamicView(FXObject*,FXSelector,void*)
{
  // update the dynamic view option
  if(dynamicView == false)
    dynamicView = true;
  else
    dynamicView = false;

  this->drawScene();
  return 1;
}

long STDeformIfc::updateRoadMapView(FXObject*,FXSelector,void*)
{
  // update the roadmap view option
  if(roadmapView == false)
    roadmapView = true;
  else
    roadmapView = false;

  this->drawScene();
  return 1;
}

long STDeformIfc::adjustParameters(FXObject*,FXSelector,void*)
{
  // adjust the parameters of the algorithm in function of the mode used
  if(optModeDeform == 0) // spatio-temporal deformation
    {
      K_INT         = 1.0;
      K_EXT         = 2.0;
      K_STEXT       = 2.0;
      K_SPT_TMP     = 3.0;

      alphaMin      = 0.5;
      alphaMax      = 2.0;
      betaMin       = 1.0;
      betaMax       = 10.0;

      maxSpDist     = 80.0;
      maxTmpDist    = 5.0;
    }
  else if(optModeDeform == 1) // spatial deformation
    {
      K_INT         = 1.0;
      K_EXT         = 2.0;
      K_SPT_TMP     = 3.0;

      alphaMin      = 0.1;
      alphaMax      = 1.0;
      betaMin       = 0.0;
      betaMax       = 20.0;

      maxSpDist     = 90.0;
      maxTmpDist    = 5.0;
    }
  else // optModeDeform == 2 // temporal deformation
    {
      K_INT         = 0.1;
      K_EXT         = 8.0;
      K_SPT_TMP     = 3.0;

      alphaMin      = 0.1;
      alphaMax      = 1.0;
      betaMin       = 0.0;
      betaMax       = 20.0;

      maxSpDist     = 100.0;
      maxTmpDist    = 8.0;
    }

  return 1;
}

long STDeformIfc::perspectiveMode(FXObject*,FXSelector,void*)
{
  FXString msg("Perspective View");
  viewModeBox->setText(msg);

  orthoView = false;  

  return 1;
}

long STDeformIfc::orthogonalMode(FXObject*,FXSelector,void*)
{
  FXString msg("Orthogonal View");
  viewModeBox->setText(msg);

  orthoView = true;

  return 1;
}

long STDeformIfc::displayRTGraph(FXObject*,FXSelector,void*)
{
  if(this->deformer_ != NULL)
    this->deformer_->displayPredHorizonProfile(0.01);
  
  return 1;
}

long STDeformIfc::updateUseVelCst(FXObject*,FXSelector,void*)
{
  // update the use of linear velocity constraints
  if(useVelCst == false)
    useVelCst = true;
  else
    useVelCst = false;
  return 1;
}

long STDeformIfc::updateUseAccCst(FXObject*,FXSelector,void*)
{
  // update the use of linear acceleration constraints
  if(useAccCst == false)
    useAccCst = true;
  else
    useAccCst = false;
  return 1;
}

long STDeformIfc::updateUseAVelCst(FXObject*,FXSelector,void*)
{
  // update the use of angular velocity constraints
  if(useAVelCst == false)
    useAVelCst = true;
  else
    useAVelCst = false;
  return 1;
}

long STDeformIfc::updateUseAAccCst(FXObject*,FXSelector,void*)
{
  // update the use of angular acceleration constraints
  if(useAAccCst == false)
    useAAccCst = true;
  else
    useAAccCst = false;
  return 1;
}

long STDeformIfc::updateDrawTrajGen(FXObject*,FXSelector,void*)
{
  // update the drawing of generated trajectories
  if(drawTrajGen == false)
    drawTrajGen = true;
  else
    drawTrajGen = false;
  return 1;
}

// convert the current environment into an occupancy grid
long STDeformIfc::convertEnvOccGrid(FXObject*,FXSelector,void*)
{
  // check if the current environment is not null
  if(this->env_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Converting the Environment","No environment has been defined yet.");      
      return -1;
    }
  
  // delete the current occupancy grid if it exists
  if(this->map_ != NULL)
    delete this->map_;
  
  struct timeval t1;
  struct timeval t2;
  
  // init the timer
  gettimeofday(&t1,NULL);
  
  // convert the current environment into an occupancy grid
  cout << "sampleXOccGrid : " << sampleXOccGrid << endl;
  cout << "sampleYOccGrid : " << sampleYOccGrid << endl;
  this->map_ = this->env_->sample(this->env_->getDfltX(), this->env_->getDfltY(), sampleXOccGrid, sampleYOccGrid);
  
  if(this->map_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Converting the Environment","The resulting occupancy grid is nul.");      
      return -2;
    }
  
  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;
  double elapsedTime = elapsedTs + 0.000001*elapsedTus;
  
  // save the occupancy grid in a pgm file
  if(this->map_->save(mapName.text()) != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Converting the Environment","The ouput pgm file cannot be saved");      
      return -3;
    }
  
  FXString msg("The environment has been sucessfully converted into an occupancy grid in ");
  msg += FXStringVal(elapsedTime,2,FALSE);
  msg += FXString(" sec.");
  msgBox->setText(msg);
  return 0;
}

long STDeformIfc::changeOccGridName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Occupancy Grid");
  open.setFilename(occgridName);
  if(open.execute()){
    occgridName = open.getFilename();
  }
  return 1;
}

long STDeformIfc::updateParallel(FXObject*,FXSelector,void*)
{
  // update the parallelization of trajectory generation
  if(parallelize == false)
    parallelize = true;
  else
    parallelize = false;
  return 1;
}

long STDeformIfc::computeDiffDriveTraj(FXObject*,FXSelector,void*)
{
  DDState initCfg((double)initXRb,(double)initYRb,0,0,0,0);
  return this->computeNewDiffDriveTraj(initCfg);
}

long STDeformIfc::computeItscTraj(FXObject*,FXSelector,void*)
{
  return this->computeIntersectionTraj();
}

long STDeformIfc::computeMtwyTraj(FXObject*,FXSelector,void*)
{
  return this->computeMotorwayTraj();
}

long STDeformIfc::initTrajFollowing(FXObject*,FXSelector,void*)
{
  // check if a carlike trajectory has been computed
  if(trajType != NH_TRAJ || this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while initializing the trajectory follower","No car-like trajectory has been loaded");
      return -1;
    }

  processType_ = TF_PROC;

  // init the car-like system constraints
  this->initCarLikeSys();

  // initialize the car-like state
  if(this->robotStateTF_ != NULL) delete this->robotStateTF_;
  this->robotStateTF_ = new StateS6D(initXRb,initYRb,initThRb,initVRb,initPhRb,0.0);

  // initialize the trajectory follower
  if(TF_mode == SIMPLE_TF_MODE)
    {
      this->follower_ = new CLTrajFollow(carlike_,(Trajectory6D*)traj_);
    }
  else if(TF_mode == MULTI_TF_MODE)
    {
      this->mfollower_ = new CLMultiTrajFollow(carlike_,(Trajectory6D*)traj_);
    }

  // draw the scene
  this->drawScene();
  return 1;
}

long STDeformIfc::startTF(FXObject*,FXSelector,void*)
{
  // verifying that the trajectory and the follower are not NULL
  if(this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The initial trajectory hasn't been computed");
      return -1;
    }
  else if(TF_mode == SIMPLE_TF_MODE && this->follower_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The trajectory follower must be initialized");
      return -2;
    }
  else if(TF_mode == MULTI_TF_MODE && this->mfollower_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The multi trajectory follower must be initialized");
      return -2;
    }

  // saving the initial trajectory
  if(time == 0)
    {
      if(this->initTraj_ != NULL) delete this->initTraj_;
      if(this->path_->getType() == HOLONOMIC_PATH)
	this->initTraj_ = new Trajectory5D(*((Trajectory5D*)(this->traj_)));
      else if(this->path_->getType() == NONHOLONOMIC_PATH)
	this->initTraj_ = new Trajectory6D(*((Trajectory6D*)(this->traj_)));
    }

  cout << "starting timeout" << endl;
  // init the timer
  gettimeofday(&t1,NULL);
  running = true;
  getApp()->addTimeout(this,ID_TIMEOUT_TF,TIMER_INTERVAL);

  oldTimeTs = t1.tv_sec;
  oldTimeTus = t1.tv_usec;
//   cout << "oldTime (start) : " << oldTimeTs << "," << oldTimeTus << endl;
  return 1;
}

long STDeformIfc::oneStepTF(FXObject*,FXSelector,void*)
{
  // verifying that the trajectory and the follower are not NULL
  if(this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following process","The initial trajectory hasn't been computed");
      return -1;
    }
  else if(TF_mode == SIMPLE_TF_MODE && this->follower_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The trajectory follower must be initialized");
      return -2;
    }
  else if(TF_mode == MULTI_TF_MODE && this->mfollower_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The multi trajectory follower must be initialized");
      return -2;
    }
  
  // init the timer
  gettimeofday(&t1,NULL);
  oneStep = true;
  getApp()->addTimeout(this,ID_TIMEOUT_TF,TIMER_INTERVAL);

  oldTimeTs = t1.tv_sec;
  oldTimeTus = t1.tv_usec;
//   cout << "oldTime (one step) : " << oldTimeTs << "," << oldTimeTus << endl;
  return 1;
}

long STDeformIfc::stopTF(FXObject*,FXSelector,void*)
{
  cout << "stoping timeout" << endl;
  running = false;
  getApp()->removeTimeout(this,ID_TIMEOUT_TF);
  alreadyElapsedTime = time;
  return 1;
}

long STDeformIfc::restartTF(FXObject*,FXSelector,void*)
{
  // TODO... 
  FXMessageBox::error(this,MBOX_OK,"Error","TODO!!!");
  return -1;
}

long STDeformIfc::updateTF(FXObject*,FXSelector,void*)
{
  // update obstacles random motion parameters
  DynamicObs::setSeuilChgt(seuilChgt);
  DynamicObs::setMaxAngleChgt(maxAngleChgt);
  DynamicObs::setMaxNormChgt(maxNormChgt);

  // compute the time before the following process
  struct timeval t3;
  gettimeofday(&t3,NULL);

  // compute the trajectory following process
  int status = this->followTraj();  

  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;

  double elapsedT = alreadyElapsedTime + ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);

  time_t timeTs = t2.tv_sec - oldTimeTs;
  suseconds_t timeTus = t2.tv_usec - oldTimeTus;
  trueLastElapsedT = ((double)(timeTs)) + 0.000001 * ((double)(timeTus));
//   cout << "trueLastElapsedT : " << trueLastElapsedT << endl;

  oldTimeTs = t2.tv_sec;
  oldTimeTus = t2.tv_usec;
//   cout << "oldTime (update) : " << oldTimeTs << "," << oldTimeTus << endl;

  // set the current time
  time = elapsedT;

  // compute the trajectory following duration
  time_t deformTs = t2.tv_sec - t3.tv_sec;
  suseconds_t deformTus = t2.tv_usec - t3.tv_usec;
  deformDur = ((double)deformTs) + 0.000001 * ((double)deformTus);
  if(deformDur != 0) lastFPS = 1.0/deformDur;
  else lastFPS = 0.0;

  switch(status)
    {
    case 0 : 
      
      break;
    case -1 : 
      cout << "STDeformIfc::deformTraj() : Warning. The trajectory generation process failed at time << " << time << "." << endl;
      break;
    }

  // check if the end time of the trajectory has been reached
  if(time >= this->traj_->at(traj_->getSize()-1)->t())
    {
      cout << "fin traj" << endl;
      
      // stop the following process
      this->stopTF(NULL,0,NULL);
    }
  
  // repaint the viewer
  this->drawScene();
  
  // rerun the timeout
  if(oneStep)
    {
      alreadyElapsedTime = time;
      oneStep = false;
    }
  else if(running)
    {
      getApp()->addTimeout(this,ID_TIMEOUT_TF,TIMER_INTERVAL);
    }

  // update the FPS number
  nbTimeout++;
  FPS = ((double)(totalNbTimeout + nbTimeout))/time;

  return 1;
}

long STDeformIfc::updateDrawTrajGenTF(FXObject*,FXSelector,void*)
{
  // update the drawing of generated trajectories
  if(drawTrajGenTF == false)
    drawTrajGenTF = true;
  else
    drawTrajGenTF = false;
  return 1;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// initialize the deformation process
void STDeformIfc::initDeform()
{
  deformer_->setLastDeformDuration(deformDur);

  if(optModeDeform == 0) deformer_->setModeDeform(SPATIOTEMP_DEFORM);
  else if(optModeDeform == 1) deformer_->setModeDeform(SPATONLY_DEFORM);
  else if(optModeDeform == 2) deformer_->setModeDeform(TEMPONLY_DEFORM);
  deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &K_SPT_TMP);
  deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &alphaMin, &alphaMax, &betaMin, &betaMax);
//   deformer_->setMaxEffectDist(&maxEffectDist);
//   deformer_->setMaxStEffectDist(&maxStEffectDist);
  deformer_->setMaxEffectDist(&maxSpDist,&maxTmpDist,&maxStEffectDist);
  deformer_->setReliability(&predHorizon, &reliabilityTime, &epsilonRT);
  deformer_->setTemporalConst(&useTemporalConsistency);
  deformer_->setTimeStep(trajTimeStep);
  //deformer_->setPosStep(trajTimeStep * maxSpd);
  deformer_->setPosStep(trajPosStep);

  switch(trajType)
    {
    case HOLONOM_TRAJ : 
      break;

    case NH_TRAJ : 
      ((STNHDeformation*)deformer_)->setParallelExec(parallelize);
      ((STNHDeformation*)deformer_)->setNbThread(nbThread);
      
      // init the trajectory generation parameters
      CLTrajGen::setMaxNbStep(maxNbStep);               // set the maximal number of steps
      CLTrajGen::setMaxResidual(maxResidual);           // set the maximal residual allowed
      CLTrajGen::setDBG(debugLevel);                    // set the debugging level
      CLTrajGen::setUseVelCst(useVelCst);               // set true if linear velocity constraints are used
      CLTrajGen::setUseAccCst(useAccCst);               // set true if linear acceleration constraints are used
      CLTrajGen::setUseAVelCst(useAVelCst);             // set true if angular velocity constraints are used
      CLTrajGen::setUseAAccCst(useAAccCst);             // set true if angular acceleration constraints are used
      break;

    case DD_TRAJ : 
      ((DDDeformation*)deformer_)->setParallelExec(parallelize);
      ((DDDeformation*)deformer_)->setNbThread(nbThread);
      
      // init the trajectory generation parameters
      DDTrajGen::setMaxNbStep(maxNbStep);               // set the maximal number of steps
      DDTrajGen::setMaxResidual(maxResidual);           // set the maximal residual allowed
      DDTrajGen::setDBG(debugLevel);                    // set the debugging level
      break;
    }
}

int STDeformIfc::deformTraj()
{
  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      this->stopTraj(NULL, 0, NULL);
      return -1;
    }
  else
    {
      // initialize the deformation process
      this->initDeform();

      // launching the deformation algorithm
      glcanvas->makeCurrent();
      int status = deformer_->deform(time, curStep);
      glcanvas->makeNonCurrent();

      if(this->path_ != NULL) delete this->path_;
      this->path_ = this->traj_->getPath();

      return status;
    }
}

// initialize the traj following process params that can change at each step
void STDeformIfc::initTF()
{
  if(this->TF_mode == SIMPLE_TF_MODE)
    {
      if(this->follower_ == NULL) return;

      // initialize the trajectory process parameters
      this->follower_->dt() = TF_dt;
      this->follower_->timeMeeting() = TF_timeMeeting;
    }
  else if(this->TF_mode == MULTI_TF_MODE)
    {
      if(this->mfollower_ == NULL) return;

      this->mfollower_->setItdTime(MTF_itdTime);
    }
  
  // initialize the noise parameters
  StateS6D::maxNoiseXY_ = noiseX;
  StateS6D::maxNoiseTH_ = noiseTH;
  StateS6D::maxNoiseS_ = noiseV;
  StateS6D::maxNoisePH_ = noisePH;

  CLControl::maxNoiseAcc_ = noiseAcc;
  CLControl::maxNoiseSS_ = noiseSS;
}

// execute a step of the following algorithm
int STDeformIfc::followTraj()
{
  // init the trajectory following process
  this->initTF();
  
  // check if the state of the robot is available
  if(robotStateTF_ == NULL)
    {
      cout << "STDeformIfc::followTraj : Warning. The robot state is nul" << endl;
      return -1;
    }

  // compute a noised localization from the robot state
  StateS6D noisedS = ((StateS6D*)robotStateTF_)->noisedState(trueLastElapsedT);

  // init the control that will be computed by the following process
  CLControl ctrl;

  int res;
  if(this->TF_mode == SIMPLE_TF_MODE)
    {
      // perform the trajectory following process
      res = ((CLTrajFollow*)follower_)->getControl(noisedS,            // localization of the robot (noised state)
						   noisedS.t() + trueLastElapsedT,  // use the previous TF process duration to evaluate the time at which will be applied the control
						   trueLastElapsedT,   // use the previous TF process duration as the control application duration
						   ctrl);              // control computed by the TF process
      
      // get back the state the trajectory follower try to reach (for displaying)
      if(stateToReachTF_ != NULL) delete stateToReachTF_;
      stateToReachTF_ = new StateS6D;
      ((CLTrajFollow*)follower_)->selectStateToReach(*((StateS6D*)stateToReachTF_));
    }
  else if(this->TF_mode == MULTI_TF_MODE)
    {
      // perform the trajectory following process
      res = ((CLMultiTrajFollow*)mfollower_)->getControl(noisedS,            // localization of the robot (noised state)
							 this->env_,         // the robot's environment
							 noisedS.t() + trueLastElapsedT,  // use the previous TF process duration to evaluate the time at which will be applied the control
							 trueLastElapsedT,   // use the previous TF process duration as the control application duration
							 ctrl);              // control computed by the TF process

    }

  // compute a noised control from the control computed by the TF process
  CLControl noisedC = ctrl.noisedCtrl(trueLastElapsedT);

  /// compute the state reached by the previous real state and the noised control
  
  // initialize the init state
  double* initTS = NULL;
  StateS6D_1_CL2tabD(*robotStateTF_,&initTS);

  // initialize the system constraints
  double* cstt = NULL;
  CLSys_2_CL2TabCstt(*carlike_,&cstt);

  // initialize the parameters IG
  double* param = NULL;
  CLctrlt_2_CL2IG(noisedC,trueLastElapsedT,&param);

  // compute the state reached
  CL2TrajGen trajgen;
  CL2TrajGen::setUseVelCst(true);
  CL2TrajGen::setUseAccCst(true);
  CL2TrajGen::setUseStACst(true);
  CL2TrajGen::setUseStSCst(false);
  trajgen.ComputeFromIG(initTS,cstt,param);

  // get back the state reached
  double oldT = robotStateTF_->t();
  double tabSReached[6];
  trajgen.getFinalStateTReached(tabSReached);
  
  // update the local state of the robot
  robotStateTF_->setX(tabSReached[0]);
  robotStateTF_->setY(tabSReached[1]);
  robotStateTF_->setTH(tabSReached[2]);
  robotStateTF_->setPH(tabSReached[3]);
  robotStateTF_->setS(tabSReached[4]);
  robotStateTF_->setT(oldT + tabSReached[5]);
}

// initialize the robot system constraints
void STDeformIfc::initRobot()
{
  this->rb_->setVMax(maxSpd);
  this->rb_->setAMax(maxAcc);
  this->rb_->setMaxSteering(maxSteer);
  this->rb_->setMaxSteerSpd(maxSteerS);
  this->rb_->setLength(robotLength);
}

// initialize the car-like system constraints
void STDeformIfc::initCarLikeSys()
{
  if(carlike_ != NULL) delete carlike_;
  carlike_ = new CarLikeSys();

  this->carlike_->setVMax(maxSpd);
  this->carlike_->setAMax(maxAcc);
  this->carlike_->setMaxSteering(maxSteer);
  this->carlike_->setMaxSteerSpd(maxSteerS);
  this->carlike_->setLength(robotLength);
}

// init the tables of internal forces, external forces and nearest points
void STDeformIfc::initTabForces()
{
  // init the table of state types
  typeState = new int*;
  (*typeState) = new int[this->traj_->getSize()];
  for(int i=0; i < this->traj_->getSize(); i++)
    {
      (*typeState)[i] = 0;
    }
  
  //init the table of nearest states
  tableNearState = new StateR**;
  (*tableNearState) = new StateR* [this->traj_->getSize()];
  for(int i=0; i < this->traj_->getSize(); i++)
    {
      (*tableNearState)[i] = NULL;
    }
  
  // init the tables of internal and external forces
  tableFInt = new StateR*;
  *tableFInt = NULL;
  tableFExt = new StateR*;
  *tableFExt = NULL;
}

// init the viewer after loading of a path
void STDeformIfc::initViewer_PathLdd()
{
  if(typeMotion == 0) // trajectory tracking
    {
      // get back the path of the trajectory
      path_ = traj_->getPath();
      
      // init the viewer
      ConfigR* center = path_->center();
      double radius = path_->radius();
      double depth = 5*radius;
      double dt = 3*(*traj_)[traj_->getSize()-1]->t();
      double upx,upy,upz;
      
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];
      
      this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
      
      delete center;
      
      // draw the scene
      this->drawScene();
    }
}

// load the system parameters for the car-like system
void STDeformIfc::loadCarLikeParam()
{
  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 0.10*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  
  genTrajType  = NH_TRAJ;
  trajType     = NH_TRAJ;
  
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.0;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 80.0;
  maxTmpDist      = 5.0;
  maxStEffectDist = 80.0;  
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  roadmapView = true;
  
  seuilChgt = 0.7;
  maxAngleChgt = 2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
  initXRb = 50;
  initYRb = 50;
  initThRb = 0.0;
  
//   dfltXMap          = 700;
//   dfltYMap          = 700;
  drawMapBoundaries = true;
    
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;

  parallelize = false;
  nbThread = 5;
}

// load the system parameters for the differential drive system
void STDeformIfc::loadDiffDriveParam()
{
}

// load the system parameters for the intersection scenario
void STDeformIfc::loadItscParam()
{
  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 0.10*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  
  genTrajType  = NH_TRAJ;
  trajType     = NH_TRAJ;
  
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.1;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 50.0;
  maxTmpDist      = 7.0;
  maxStEffectDist = 40.0;  
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  roadmapView = false;
  
  seuilChgt = 0.7;
  maxAngleChgt = 0.0; //2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
//   initXRb = 50;
//   initYRb = 50;
//   initThRb = 0.0;
  
//   dfltXMap          = 1370;
//   dfltYMap          = 900;
  drawMapBoundaries = true;
    
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;

  parallelize = false;
  nbThread = 5;
}

// load the system parameters for the motorway scenario
void STDeformIfc::loadMtwyParam()
{
  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 0.10*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  
  genTrajType  = NH_TRAJ;
  trajType     = NH_TRAJ;
  
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.1;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 50.0;
  maxTmpDist      = 7.0;
  maxStEffectDist = 30.0;  
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  roadmapView = false;
  
  seuilChgt = 0.7;
  maxAngleChgt = 0.0; //2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
//   initXRb = 50;
//   initYRb = 50;
//   initThRb = 0.0;
  
//   dfltXMap          = 1370;
//   dfltYMap          = 900;
  drawMapBoundaries = true;
    
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;

  parallelize = false;
  nbThread = 5;
}

// compute a new straight line trajectory from initCfg
int STDeformIfc::computeNewRandTrajFrom(StateR& initSt)
{
  if(trajType == HOLONOM_TRAJ)
    {
      // get a new random goal position
      goalPosX = -1;
      goalPosY = -1;
      
      while(goalPosX == -1 && goalPosY == -1)
	{
	  // choose a new goal position randomly
	  if(this->env_ != NULL && this->env_->getMap() != NULL)
	    {
	      goalPosX = getIntRandom(0, this->env_->getMap()->getWidth());
	      goalPosY = getIntRandom(0, this->env_->getMap()->getHeight());
	    }
	  else if(this->env_ != NULL)
	    {
	      goalPosX = getIntRandom(0, (int)(this->env_->getDfltX()));
	      goalPosY = getIntRandom(0, (int)(this->env_->getDfltY()));
	    }
	  else
	    {
	      goalPosX = getIntRandom(0, dfltXMap);
	      goalPosY = getIntRandom(0, dfltYMap);
	    }
	  
	  // verify that the position is not to close to the init position
	  double dxIG = goalPosX - initSt.x();
	  double dyIG = goalPosY - initSt.y();
	  double sqDistInitGoal = dxIG*dxIG + dyIG*dyIG;
	  if(sqDistInitGoal < 100*100)
	    {
	      goalPosX = -1;
	      goalPosY = -1;
	    }

	  // verify that the position is not in collision or too close from a static obstacle
	  if(this->env_ != NULL)
	    {
	      bool trouve = false;
	      ConfigR2D randCfg(goalPosX,goalPosY);
	      int i=0;
	      while(!trouve && i<this->env_->getNbStObs())
		{
		  ConfigR2D near;
		  double dist;
		  bool inside = this->env_->getStObs(i)->nearPosFrom(&randCfg, near, dist);
		  
		  if(inside || dist < this->rb_->getRadius() + 30.0)
		    {
		      trouve = true;
		      goalPosX = -1;
		      goalPosY = -1;
		    }
		  else i++;
		}
	    }
	}
      
      if(this->roadmap_ == NULL)
	{
	  // compute a straight line path up to this position
	  double dir = atan2(goalPosY - initSt.y(), goalPosX - initSt.x());
	  ConfigR3D cfg1(initSt.x(), initSt.y(), dir);
	  ConfigR3D cfg2(goalPosX, goalPosY, dir);
	  
	  GeomPathPlanner gpp(NULL, &cfg1, &cfg2);
	  gpp.initStrLineCptg();
	  int resPP = gpp.computePath();
	  if(resPP != PATH_CPTD)
	    {
	      // case where the path hasn't been computed correctly
	      return -1;
	    }
	  this->path_ = gpp.getPath()->getPath2D();
	}
      else
	{
	  // reinitialize the roadmap
	  this->roadmap_->reinit();

	  // compute a path on the roadmap
	  ConfigR2D cfg1(initSt.x(), initSt.y());
	  ConfigR2D cfg2(goalPosX, goalPosY);

	  this->path_ = new Path();
	  RoadMapPP rmpp(roadmap_,&cfg1,&cfg2);
	  rmpp.computePath();
	  Path* path = rmpp.getPath()->getPath2D();
	  path_->concat(*path);
	}

      // compute the trajectory until this position
      int res = this->computeTrajectory(NULL, 0, NULL);
            
      return res;
    }
  else if(trajType == NH_TRAJ)
    {
      // TODO...
    }
  else if(trajType == DD_TRAJ)
    {
      // TODO...
    }
}

// compute a new Trajectory for the Car-Like Sys from initCfg
int STDeformIfc::computeNewCarLikeTraj(StateR& initCfg)
{
  // load the car-like deformation and trajectory generation predefined parameters
  this->loadCarLikeParam();

  // init the robot configuration
  this->initRobot();

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  AccASTrajGen aastg(rb_,
		    env_,
		    new StateS6D(initCfg),
		    new StateS6D(200.0,200.0,0,0,0,0));
  int resTP = aastg.computeTraj();

  if(resTP != TRAJ_CPTD)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
      return -2;
    }
  else
    {
      if(this->traj_ != NULL) delete traj_;
      this->traj_ = new Trajectory6D(*((Trajectory6D*)aastg.getTraj()));
      this->traj_->shift(time);

      // set the type of the trajectory
      trajType = NH_TRAJ;

      // compute the list of initial feddback controls
      std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
      for(int i=0; i<aastg.vecCtrl_->size(); i++)
	{
 	  Vec_DP* nvVecParam = new Vec_DP(7); 
	  (*nvVecParam)[0] = aastg.vecCtrl_->at(i).getAS();
	  (*nvVecParam)[1] = 0.0;
	  (*nvVecParam)[2] = 0.0;
	  (*nvVecParam)[3] = aastg.vecCtrl_->at(i).getAcc();
	  (*nvVecParam)[4] = 0.0;
	  (*nvVecParam)[5] = 0.0;
	  (*nvVecParam)[6] = aastg.vecCtrl_->at(i).getTime();
	  listFBCtrl->push_back(nvVecParam);
	}

      slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
      curStep = 0;

      // init the tables of internal forces, external forces and nearest points
      this->initTabForces();

      // init the trajectory deformer
      if(deformer_ != NULL) delete deformer_;
      deformer_ = new STNHDeformation(this->traj_, this->env_, this->rb_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
      this->initDeform();

      // initialize the list of initial feedback controls
      //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

      // init the viewer once the path loaded
      this->initViewer_PathLdd();
    }  

  return 1;
}

// compute a trajectory for the intersection scenario
int STDeformIfc::computeIntersectionTraj()
{
  // load the car-like deformation and trajectory generation 
  // predefined parameters for the intersection scenario
  this->loadItscParam();

  // init the robot configuration
  this->initRobot();

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  AccASTrajGen aastg(rb_,
		     env_,
		     new StateS6D(28.0,445.0,0.00602402,10.0,0,0),
		     new StateS6D(0,0,0,0,0,0));
  int resTP = aastg.computeTraj2();

  if(resTP != TRAJ_CPTD)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
      return -2;
    }
  else
    {
      if(this->traj_ != NULL) delete traj_;
      this->traj_ = new Trajectory6D(*((Trajectory6D*)aastg.getTraj()));
      this->traj_->shift(time);

      // set the type of the trajectory
      trajType = NH_TRAJ;

      // compute the list of initial feddback controls
      std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
      for(int i=0; i<aastg.vecCtrl_->size(); i++)
	{
 	  Vec_DP* nvVecParam = new Vec_DP(7); 
	  (*nvVecParam)[0] = aastg.vecCtrl_->at(i).getAS();
	  (*nvVecParam)[1] = 0.0;
	  (*nvVecParam)[2] = 0.0;
	  (*nvVecParam)[3] = aastg.vecCtrl_->at(i).getAcc();
	  (*nvVecParam)[4] = 0.0;
	  (*nvVecParam)[5] = 0.0;
	  (*nvVecParam)[6] = aastg.vecCtrl_->at(i).getTime();
	  listFBCtrl->push_back(nvVecParam);
	}

      slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
      curStep = 0;

      // init the tables of internal forces, external forces and nearest points
      this->initTabForces();

      // init the trajectory deformer
      if(deformer_ != NULL) delete deformer_;
      deformer_ = new STNHDeformation(this->traj_, this->env_, this->rb_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
      this->initDeform();

      // dont use the presious parameters for the next trajectory generation
      ((STNHDeformation*)deformer_)->setUsePrevParam(false);
      ((STNHDeformation*)deformer_)->setUseResampling(false);

      // initialize the list of initial feedback controls
      //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

      // init the viewer once the path loaded
      this->initViewer_PathLdd();
    }  

  return 1;
}

// compute a trajectory for the motorway scenario
int STDeformIfc::computeMotorwayTraj()
{
  // load the car-like deformation and trajectory generation 
  // predefined parameters for the intersection scenario
  this->loadMtwyParam();

  // init the robot configuration
  this->initRobot();

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  AccASTrajGen aastg(rb_,
		     env_,
		     new StateS6D(96.0,161.0,0.47783105,10.0,0,0),
		     new StateS6D(0,0,0,0,0,0));
  int resTP = aastg.computeTraj3();

  if(resTP != TRAJ_CPTD)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
      return -2;
    }
  else
    {
      if(this->traj_ != NULL) delete traj_;
      this->traj_ = new Trajectory6D(*((Trajectory6D*)aastg.getTraj()));
      this->traj_->shift(time);

      // set the type of the trajectory
      trajType = NH_TRAJ;

      // compute the list of initial feddback controls
      std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
      for(int i=0; i<aastg.vecCtrl_->size(); i++)
	{
 	  Vec_DP* nvVecParam = new Vec_DP(7); 
	  (*nvVecParam)[0] = aastg.vecCtrl_->at(i).getAS();
	  (*nvVecParam)[1] = 0.0;
	  (*nvVecParam)[2] = 0.0;
	  (*nvVecParam)[3] = aastg.vecCtrl_->at(i).getAcc();
	  (*nvVecParam)[4] = 0.0;
	  (*nvVecParam)[5] = 0.0;
	  (*nvVecParam)[6] = aastg.vecCtrl_->at(i).getTime();
	  listFBCtrl->push_back(nvVecParam);
	}

      slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
      curStep = 0;

      // init the tables of internal forces, external forces and nearest points
      this->initTabForces();

      // init the trajectory deformer
      if(deformer_ != NULL) delete deformer_;
      deformer_ = new STNHDeformation(this->traj_, this->env_, this->rb_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
      this->initDeform();

      // dont use the presious parameters for the next trajectory generation
      ((STNHDeformation*)deformer_)->setUsePrevParam(false);
      ((STNHDeformation*)deformer_)->setUseResampling(false);

      // initialize the list of initial feedback controls
      //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

      // init the viewer once the path loaded
      this->initViewer_PathLdd();
    }  

  return 1;
}

// compute a new Trajectory for the DiffDrive Sys from initCfg
int STDeformIfc::computeNewDiffDriveTraj(StateR& initCfg)
{
  // load the car-like deformation and trajectory generation predefined parameters
  this->loadDiffDriveParam();

  // init the robot configuration
  this->rb_->setWVMax(maxWVel);
  this->rb_->setWAMax(maxWAcc);
  this->rb_->setLength(robotRadius);

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  WAccTrajGen wacctg(rb_,
		     env_,
// 		     new DDState(initCfg),
		     new DDState(50.0,50.0,0,10.0,10.0,0),
		     new DDState(200.0,200.0,0,0.0,0.0,0));
  int resTP = wacctg.computeTraj();

  if(resTP != TRAJ_CPTD)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
      return -2;
    }
  else
    {
      if(this->traj_ != NULL) delete traj_;
      this->traj_ = new DDTrajectory(*((DDTrajectory*)wacctg.getTraj()));
      this->traj_->shift(time);

      // set the type of the trajectory
      trajType = DD_TRAJ;

      // compute the list of initial feddback controls
      std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
      for(int i=0; i<wacctg.vecCtrl_->size(); i++)
	{
 	  Vec_DP* nvVecParam = new Vec_DP(7); 
	  (*nvVecParam)[0] = wacctg.vecCtrl_->at(i).getWRAcc();
	  (*nvVecParam)[1] = 0.0;
	  (*nvVecParam)[2] = 0.0;
	  (*nvVecParam)[3] = wacctg.vecCtrl_->at(i).getWLAcc();
	  (*nvVecParam)[4] = 0.0;
	  (*nvVecParam)[5] = 0.0;
	  (*nvVecParam)[6] = wacctg.vecCtrl_->at(i).getTime();
	  listFBCtrl->push_back(nvVecParam);
	}

      slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
      curStep = 0;

      typeState = new int*;
      (*typeState) = new int[this->traj_->getSize()];
      for(int i=0; i < this->traj_->getSize(); i++)
	{
	  (*typeState)[i] = 0;
	}

      //init the table of nearest states
      tableNearState = new StateR**;
      (*tableNearState) = new StateR* [this->traj_->getSize()];
      for(int i=0; i < this->traj_->getSize(); i++)
	{
	  (*tableNearState)[i] = NULL;
	}

      // init the tables of internal and external forces
      tableFInt = new StateR*;
      *tableFInt = NULL;
      tableFExt = new StateR*;
      *tableFExt = NULL;

      // init the trajectory deformer
      if(deformer_ != NULL) delete deformer_;
      
      deformer_ = new DDDeformation(this->traj_, this->env_, this->rb_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
      ((DDDeformation*)deformer_)->setParallelExec(parallelize);
      ((DDDeformation*)deformer_)->setNbThread(nbThread);

      if(optModeDeform == 0) deformer_->setModeDeform(SPATIOTEMP_DEFORM);
      else if(optModeDeform == 1) deformer_->setModeDeform(SPATONLY_DEFORM);
      else if(optModeDeform == 2) deformer_->setModeDeform(TEMPONLY_DEFORM);
      deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &K_SPT_TMP);
      deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &alphaMin, &alphaMax, &betaMin, &betaMax);
//       deformer_->setMaxEffectDist(&maxEffectDist);
//       deformer_->setMaxStEffectDist(&maxStEffectDist);
      deformer_->setMaxEffectDist(&maxSpDist,&maxTmpDist,&maxStEffectDist);
      deformer_->setReliability(&predHorizon, &reliabilityTime, &epsilonRT);
      deformer_->setTemporalConst(&useTemporalConsistency);
      deformer_->setTimeStep(trajTimeStep);
      //deformer_->setPosStep(trajTimeStep * maxSpd);
      deformer_->setPosStep(trajPosStep);
      
      // init the trajectory generation parameters
      DDTrajGen::setMaxNbStep(maxNbStep);               // set the maximal number of steps
      DDTrajGen::setMaxResidual(maxResidual);           // set the maximal residual allowed
      DDTrajGen::setDBG(debugLevel);                    // set the debugging level

      // display the list of initial feedback controls
      std::list <Vec_DP*>::iterator itLC;

      // initialize the list of initial feedback controls
      //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

      if(typeMotion == 0) // trajectory tracking
	{
	  // get back the path of the trajectory
	  path_ = traj_->getPath();

	  // init the viewer
	  ConfigR* center = path_->center();
	  double radius = path_->radius();
	  double depth = 5*radius;
	  double dt = 3*(*traj_)[traj_->getSize()-1]->t();
	  double upx,upy,upz;

	  Vec v1(3);
	  v1.setVal(0,0);
	  v1.setVal(1,-dt);
	  v1.setVal(2,-depth);
	  double v1Norm = v1.norm();
	  Vec v2(3);
	  v2.setVal(0,0);
	  v2.setVal(1,-dt/v1Norm);
	  v2.setVal(2,-depth/v1Norm);
	  upx = 0;
	  upy = -v2[2];
	  upz = v2[1];

	  this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
	  
	  delete center;

	  // draw the scene
	  this->drawScene();
	}
    }  

  return 1;
}

void STDeformIfc::loadMapInViewer()
{
}

void STDeformIfc::drawMap()
{
  if(this->env_ == NULL)
    {
      // nothing to draw
    }
  else if(this->env_->getMap() != NULL)
    {
      // draw the map of the environment
      
    }
  else
    {
      if(drawMapBoundaries)
	{
	  // draw the default bouding box of the map
	  glBegin(GL_LINE_LOOP);
	  glVertex3f(0.0, 0.0, time);
	  glVertex3f(dfltXMap, 0.0, time);
	  glVertex3f(dfltXMap, dfltYMap, time);
	  glVertex3f(0.0, dfltYMap, time);
	  glEnd();
	}
    }
}

// verify and correct if needed the configuration of the obstacle
void STDeformIfc::verifyObsCfg(ConfigR* cfgObs,int i)
{
  // test if the new configuration of the obstacle is in or out the environment
  double limEnvX;
  double limEnvY;
  if(this->env_->getMap() == NULL)
    {
      limEnvX = dfltXMap;
      limEnvY = dfltYMap;
    }
  else
    {
      limEnvX = this->env_->getMap()->getWidth();
      limEnvY = this->env_->getMap()->getHeight();
    }
  bool out = false;
  if(cfgObs->x() < 0)
    {
      cfgObs->setX(0);
      out = true;
    }
  else if(cfgObs->x() > limEnvX)
    {
      cfgObs->setX(limEnvX);
      out = true;
    }
  if(cfgObs->y() < 0)
    {
      cfgObs->setY(0);
      out = true;
    }
  else if(cfgObs->y() > limEnvY)
    {
      cfgObs->setY(limEnvY);
      out = true;
    }
  if(out)
    {
      // cout << "obs is out" << endl;
      this->env_->getObs(i)->remLastStateStack();
      ConfigR* lastCfg = this->env_->getObs(i)->getLastCfg();

      if(lastCfg != NULL)
	{
// 	  cout << "cfgObs : "; cfgObs->display();
// 	  cout << "lastCfg : "; lastCfg->display();
	  double dx = cfgObs->x() - lastCfg->x();
	  double dy = cfgObs->y() - lastCfg->y();
	  double dt = time - this->env_->getObs(i)->getLastTime();
	  double dsx = dx / dt;
	  double dsy = dy / dt;

	  double normSpd = sqrt(dsx*dsx + dsy*dsy);
	  if(normSpd == 0.0)
	    cout << "Warning : obstacle spd is null :-(" << endl;
	  else if(normSpd < 10.0)
	    {
	      dsx = dsx / normSpd * 10.0;
	      dsy = dsy / normSpd * 10.0;
	    }
	  
// 	  cout << "dx : " << dx << endl;
// 	  cout << "dy : " << dy << endl;
// 	  cout << "dt : " << dt << endl;
// 	  cout << "dsx : " << dsx << endl;
// 	  cout << "dsy : " << dsy << endl;
	  
	  StateR5D* newST = new StateR5D(cfgObs->x(),cfgObs->y(),time,-dsx,-dsy);
	  this->env_->getObs(i)->appendStateStack(newST);
	  delete lastCfg;
	}
    }
}

void STDeformIfc::drawObstacles()
{
  // verify that the environment is not null
  if(this->env_ != NULL)
    {
      for(int i=0; i < this->env_->getNbObs(); i++)
	{
	  const GLfloat redMaterial[]={1.,0.,0.,1.};
	  const GLfloat greenMaterial[]={0.,1.,0.,1.};
	  const GLfloat greenTsMaterial[]={0.,1.,0.,0.2};
	  
	  if(dispFuture)
	    {
	      // display the entiere future model
	      if(this->traj_ != NULL)
		{
		  int nbSide = 10;

		  double radObs = ((ObsDisk2D*)this->env_->getObs(i))->getRadius();
		  
		  for(int k=0; k<this->env_->getObs(i)->stackSize(); k++)
		    {
		      ConfigR* cfgObs = this->env_->getObs(i)->getCfg(time);
		      
		      if(cfgObs != NULL)
			{
			  this->verifyObsCfg(cfgObs,i);
			  
			  delete cfgObs;
			}

		      double initTime;
		      double lastTime;
		      ConfigR2D initCfgObs;
		      ConfigR2D goalCfgObs;

		      if(k != this->env_->getObs(i)->stackSize()-1)
			{
			  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
			  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);

			  StateR* s1 = this->env_->getObs(i)->getStateStack(k);
			  StateR* s2 = this->env_->getObs(i)->getStateStack(k+1);
			  initTime = s1->t();
			  lastTime = s2->t();
			  initCfgObs = ConfigR2D(s1->x(),s1->y());
			  goalCfgObs = ConfigR2D(s2->x(),s2->y());
			}
		      else
			{
			  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
			  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);

			  StateR* s1 = this->env_->getObs(i)->getStateStack(k);
			  initTime = s1->t();
			  lastTime = this->env_->getObs(i)->getInitTime() + this->env_->getObs(i)->getLifeTime();
			  ConfigR* cfg2 = this->env_->getObs(i)->getFutureCfg(lastTime);
			  initCfgObs = ConfigR2D(s1->x(),s1->y());
			  goalCfgObs = ConfigR2D(*cfg2);
			  delete cfg2;
			}
		      
		     //  cout << "draw : " << i << " " << k << endl;

		      glPushMatrix();
		      glTranslated(initCfgObs.x(),initCfgObs.y(),initTime);
		      glNormal3f(0,0,-1);
		      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
		      glPopMatrix();
		      
		      glPushMatrix();
		      glTranslated(goalCfgObs.x(),goalCfgObs.y(),lastTime);
		      glNormal3f(0,0,1);
		      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
		      glPopMatrix();
		      
		      for(int j=0; j<nbSide; j++)
			{
			  glBegin(GL_QUADS);
			  FXVec3d v1(initCfgObs.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
				     initCfgObs.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
				     initTime);
			  
			  FXVec3d v2(initCfgObs.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
				     initCfgObs.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
				     initTime);
			  
			  FXVec3d v3(goalCfgObs.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
				     goalCfgObs.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
				     lastTime);
			  
			  FXVec3d v4(goalCfgObs.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
				     goalCfgObs.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
				     lastTime);
			  
			  FXVec3d v1v2 = normalize(v2 - v1);
			  FXVec3d v1v4 = normalize(v4 - v1);
			  FXVec3d n = v1v2^v1v4;
		      
			  glNormal3f(n.x,n.y,n.z);
			  glVertex3f(v1.x, v1.y, v1.z);
			  glNormal3f(n.x,n.y,n.z);
			  glVertex3f(v2.x, v2.y, v2.z);
			  glNormal3f(n.x,n.y,n.z);
			  glVertex3f(v3.x, v3.y, v3.z);
			  glNormal3f(n.x,n.y,n.z);
			  glVertex3f(v4.x, v4.y, v4.z);
			  
			  glEnd();
			}
		    }
		}
	    }
	  else
	    {
	      // display obstacles only on the current time
	      obs_t typeObs = this->env_->getObs(i)->getType(time);
	      ConfigR* cfgObs = this->env_->getObs(i)->getCfg(time);
	      double radObs = ((ObsDisk2D*)this->env_->getObs(i))->getRadius();

	      if(cfgObs != NULL)
		{ 
		  if(drawMapBoundaries) this->verifyObsCfg(cfgObs,i);

		  if(typeObs == NOT_BORN)
		    {
		      glMaterialfv(GL_FRONT, GL_AMBIENT, greenTsMaterial);
		      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenTsMaterial);

		      glPushMatrix();
		      glTranslated(cfgObs->x(),cfgObs->y(),time);
		      gluSphere(gluNewQuadric(),radObs,8,8);
		      glPopMatrix();
		    }
		  else if(typeObs == ALIVE)
		    {
		      glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
		      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
		      
		      glPushMatrix();
		      glTranslated(cfgObs->x(),cfgObs->y(),time);
		      //gluSphere(gluNewQuadric(),radObs,8,8);
		      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
		      gluCylinder(gluNewQuadric(), radObs, radObs, robotRadius/100, 20, 20);
		      glTranslated(0,0,radObs/100);
		      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
		      glPopMatrix();
		    }
		  else // typeObs == DEAD
		    {
		    }

		  delete cfgObs;

		  //this->env_->getObs(i)->displayStack();
		}

	    }
	}
    }
}

// draw polygonal obstacles
void STDeformIfc::drawPolygonObs()
{
  // check if the environment is not null
  if(this->env_ != NULL)
    {
      for(int i=0; i < this->env_->getNbStObs(); i++)
	{
	  this->env_->getStObs(i)->toGL(time);
	}
    }
}

void STDeformIfc::drawPath()
{
  if(this->path_ != NULL)
    {
      this->path_->toGL();
    }
  else
    {
      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(0.,0.,-1.);
      glVertex3f(-1, -1, -1);
      glVertex3f(-1, 1, -1);
      glVertex3f(1, -1, -1);
      glVertex3f(1, 1, -1);
      glEnd();
    } 
}

void STDeformIfc::drawTraj()
{
  if(this->traj_ != NULL)
    {
      if(drawTrajGen)
	{
	  if(trajType == DD_TRAJ && time != 0)
	    {
	      ((DDDeformation*)deformer_)->displayTraj();
	    }
	  if(trajType == NH_TRAJ && time != 0)
	    {
	      ((STNHDeformation*)deformer_)->displayTraj();
	    }
	}
      
      //this->traj_->toGL();
      this->traj_->toGL(time, reliabilityTime, (*typeState));

      const GLfloat blackMaterial[]={0.,0.,0.,1.};
      glMaterialfv(GL_FRONT, GL_AMBIENT, blackMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, blackMaterial);

      if(dispNearStates && deformer_ != NULL)
	{
	  for(int i=0; i<this->traj_->getSize(); i++)
	    {
	      if((*tableNearState)[i] != NULL)
		{
		  glBegin(GL_LINES);
		  glVertex3f((*this->traj_)[i]->x(), (*this->traj_)[i]->y(), (*this->traj_)[i]->t());
		  glVertex3f((*tableNearState)[i]->x(), (*tableNearState)[i]->y(), (*tableNearState)[i]->t());
		  glEnd();
		}
	    }
	}

      const GLfloat purpleMaterial[]={1.,0.,1.,1.};
      glMaterialfv(GL_FRONT, GL_AMBIENT, purpleMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, purpleMaterial);

      if(dispAllForces && deformer_ != NULL && *tableFInt != NULL)
	{
	  for(int i=0; i<this->traj_->getSize(); i++)
	    {
	      glBegin(GL_LINES);
	      glVertex3f((*this->traj_)[i]->x(), (*this->traj_)[i]->y(), (*this->traj_)[i]->t());
	      glVertex3f((*this->traj_)[i]->x() + 30 * (*tableFInt)[i].x(),
			 (*this->traj_)[i]->y() + 30 * (*tableFInt)[i].y(),
			 (*this->traj_)[i]->t() + 30 * (*tableFInt)[i].t());
	      glEnd();
	    }
	}

      const GLfloat orangeMaterial[]={1.,1.,0.,1.};
      glMaterialfv(GL_FRONT, GL_AMBIENT, orangeMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, orangeMaterial);

      if(dispAllForces && deformer_ != NULL && *tableFExt != NULL)
	{
	  for(int i=0; i<this->traj_->getSize(); i++)
	    {
	      glBegin(GL_LINES);
	      glVertex3f((*this->traj_)[i]->x(), (*this->traj_)[i]->y(), (*this->traj_)[i]->t());
	      glVertex3f((*this->traj_)[i]->x() + 30 * (*tableFExt)[i].x(),
			 (*this->traj_)[i]->y() + 30 * (*tableFExt)[i].y(),
			 (*this->traj_)[i]->t() + 30 * (*tableFExt)[i].t());
	      glEnd();
	    }
	}
    }
  else
    {
      this->drawPath();
    }
}

// draw the trajectory (trajectory following mode)
void STDeformIfc::drawTrajTF()
{
  const GLfloat yellowMaterial[]={1.,1.,0.,1.};
  const GLfloat redMaterial[]={1.,0.,0.,1.};

  if(this->traj_ != NULL)
    {
      this->traj_->toGL();

      // display the state the trajectory follower try to reach
      if(stateToReachTF_ != NULL)
	{
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  
	  glPushMatrix();
	  glTranslated(stateToReachTF_->x(),
		       stateToReachTF_->y(),
		       stateToReachTF_->t());
	  gluSphere(gluNewQuadric(), 1.5, 5, 5);
	  glPopMatrix();
	}
    }
  else
    {
      this->drawPath();
    }
}

// draw the car-like robotic system
void STDeformIfc::drawCLRobot(double heading)
{
  const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
  const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};
  const GLfloat redMaterial[]    = {1.,0.,0.,1.};

  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);  

  glBegin(GL_QUADS);
  glVertex3f(1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading + 0.5*M_PI),
	     1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading + 0.5*M_PI),
	     0.0);
  glVertex3f(1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading - 0.5*M_PI),
	     1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading - 0.5*M_PI),
	     0.0);
  glVertex3f(-1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading - 0.5*M_PI),
	     -1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading - 0.5*M_PI),
	     0.0);
  glVertex3f(-1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading + 0.5*M_PI),
	     -1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading + 0.5*M_PI),
	     0.0);
  
  glEnd();
  
  // draw the wheels of the robot
  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading + 0.5*M_PI),
	       1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading - 0.5*M_PI),
	       1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(-1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading - 0.5*M_PI),
	       -1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(-1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading + 0.5*M_PI),
	       -1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  // draw the orientation of the robot
  glLineWidth(((GLfloat)(3.0)));
  glBegin(GL_LINES);
  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
  glVertex3f(0.0,0.0,0.0);
  glVertex3f(50.0*cos(heading), 
	     50.0*sin(heading), 
	     0.0);
  glEnd();
  glLineWidth(((GLfloat)(1.0)));
}

void STDeformIfc::drawRobot()
{
  if(this->traj_ != NULL)
    {
      const GLfloat redMaterial[]    = {1.,0.,0.,1.};
      const GLfloat greenMaterial[]  = {0.,1.,0.,1.};
      const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
      const GLfloat purpleMaterial[] = {1.,0.,1.,1.};
      const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};

      //draw the frame of the robot
      glLineWidth(((GLfloat)(3.0)));
      glBegin(GL_LINES);
      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
      glVertex3f((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
      glVertex3f((*(this->traj_))[(int)curStep]->x() + 100,(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
      glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
      glVertex3f((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
      glVertex3f((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y() + 100,(*(this->traj_))[(int)curStep]->t());
      glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
      glVertex3f((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
      glVertex3f((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t() + 10);
      glEnd();
      glLineWidth(((GLfloat)(1.0)));
      
      // draw the robot
      if(trajType == HOLONOM_TRAJ)
	{
	  
	  // mass-point robot drawing
	  glMaterialfv(GL_FRONT, GL_AMBIENT, purpleMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, purpleMaterial);
	  
	  glPushMatrix();
	  glTranslated((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
	  gluCylinder(gluNewQuadric(), robotRadius, robotRadius, robotRadius/100, 20, 20);
	  glTranslated(0,0,robotRadius/100);
	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
	  glPopMatrix();
	  
	}
      else if(trajType == DD_TRAJ)
	{
	  // compute the exact state of the differential drive system
	  DDState ddState;
	  if(time != 0) ddState = ((DDDeformation*)deformer_)->getStateTraj(time);

	  // differential drive robot drawing
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  
	  glPushMatrix();
	  if(time == 0) glTranslated((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
	  else glTranslated(ddState.x(),ddState.y(),time);
	  // draw the robot itself
	  //double heading = (*(this->traj_))[(int)curStep]->th();
	  double heading;
	  if(time == 0) heading = (*(this->traj_))[(int)curStep]->th();
	  else heading = ddState.th();

	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
	  gluCylinder(gluNewQuadric(), robotRadius, robotRadius, robotRadius/100, 20, 20);
	  glTranslated(0,0,robotRadius/100);
	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);

	  // draw the wheels of the robot
	  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);

	  glPushMatrix();
	  glTranslated(1.0*robotRadius*cos(heading + 0.5*M_PI),
		       1.0*robotRadius*sin(heading + 0.5*M_PI),
		       0.0);
	  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
	  glPopMatrix();

	  glPushMatrix();
	  glTranslated(1.0*robotRadius*cos(heading - 0.5*M_PI),
		       1.0*robotRadius*sin(heading - 0.5*M_PI),
		       0.0);
	  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
	  glPopMatrix();

	  // draw the orientation of the robot
	  glLineWidth(((GLfloat)(3.0)));
	  glBegin(GL_LINES);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  glVertex3f(0.0,0.0,0.0);
	  glVertex3f(50.0*cos(heading), 
		     50.0*sin(heading), 
		     0.0);
	  glEnd();
	  glLineWidth(((GLfloat)(1.0)));

	  glPopMatrix();	  
	}
      else 
	{
	  // compute the exact state of the differential drive system
	  StateS6D clState;
	  if(time != 0) clState = ((STNHDeformation*)deformer_)->getStateTraj(time);

	  // car-like robot drawing
	  glPushMatrix();
	  if(time == 0) glTranslated((*(this->traj_))[(int)curStep]->x(),(*(this->traj_))[(int)curStep]->y(),(*(this->traj_))[(int)curStep]->t());
	  else glTranslated(clState.x(),clState.y(),time);
	  // draw the robot itself
	  double heading;
	  if(time == 0) heading = (*(this->traj_))[(int)curStep]->th();
	  else heading = clState.th();

	  this->drawCLRobot(heading);

	  glPopMatrix();
	}
    }
}

// draw the robotic system (trajectory following mode)
void STDeformIfc::drawRobotTF()
{
  if(robotStateTF_ == NULL)
    {
      cout << "STDeformIfc::drawRobotTF : Warning. The robotic system state is not defined." << endl;
      return;
    }

  // car-like robot drawing
  glPushMatrix();
  glTranslated(robotStateTF_->x(),robotStateTF_->y(),robotStateTF_->t());
  // draw the robot itself
  double heading = robotStateTF_->th();
  
  this->drawCLRobot(heading);
  
  glPopMatrix();
}

// draw the roadmap
void STDeformIfc::drawRoadMap()
{
  if(this->roadmap_ != NULL)
    {
      this->roadmap_->toGL(time);
    }
}

// draw the trajectories generated for the internal force computation
void STDeformIfc::drawGenTraj()
{
  if(trajType == HOLONOM_TRAJ)
    {
      // nothing to do
    }
  else if(this->traj_ != NULL && deformer_ != NULL && ((STNHDeformation*)deformer_)->listCtrl_ != NULL)
    {
      std::list <Vec_DP*>::iterator it;
      int i=1;
      
      for(it = ((STNHDeformation*)deformer_)->listCtrl_->begin(); i < this->traj_->getSize()-1 && it != ((STNHDeformation*)deformer_)->listCtrl_->end(); it++)
	{
	  if(i != 0)
	    {
	      StateS6D prev = *((StateS6D*)((*(this->traj_))[i-1]));
	      StateS6D cur  = *((StateS6D*)((*(this->traj_))[i]));
	      StateS6D next = *((StateS6D*)((*(this->traj_))[i+1]));
	      
	      // get bach the i-th control
	      Vec_DP* ctrl = *it;
	      
	      if(ctrl != NULL)
		{
		  // compute the state reached by the random control
		  //ModelFInt carModel2;
		  CLTrajGen carModel2;
		  carModel2.L_ = rb_->getLength();
		  carModel2.setInitialGuess((*ctrl)[0],(*ctrl)[1],(*ctrl)[2],(*ctrl)[3],(*ctrl)[4],(*ctrl)[5],(*ctrl)[6]);
		  
		  Vec_DP x0(6); x0[0] = prev.x(); x0[1] = prev.y(); x0[2] = prev.th(); 
		  x0[3] = prev.s()*tan(prev.ph())/rb_->getLength(); x0[4] = prev.s(); x0[5] = 0.0;
		  Vec_DP xf(6); xf[0] = prev.x(); xf[1] = prev.y(); xf[2] = prev.th(); 
		  xf[3] = prev.s()*tan(prev.ph())/rb_->getLength(); xf[4] = prev.s(); xf[5] = 0.0;
		  Vec_DP constraints(5); constraints[0] = 0.5; constraints[1] = rb_->getAMax(); constraints[2] = rb_->getVMax(); 
		  constraints[3] = rb_->getLength(); constraints[4] = rb_->getMaxSteering();
		  
		  carModel2.setInitialParam(x0,xf,constraints);
		  carModel2.ForwardSolution();
		  
		  vector<vector<double> > vecPoints = carModel2.sampleXYT(prev.t());
		  
		  glColor3f(1.0,0.0,0.0);
		  glBegin(GL_LINE_STRIP);
		  for(int j=0; j<vecPoints.size(); j++)
		    {
		      glVertex3f(vecPoints[j][0],vecPoints[j][1],vecPoints[j][2]);
		    }
		  glEnd();
		}
	    }

	  i++;
	}
    }
}

// draw the trajectories generated for the TF process
void STDeformIfc::drawGenTrajTF()
{
  // check if the follower is available
  if(this->follower_ != NULL)
    {
      follower_->displayTrajGen();
    }
}

// draw the intermediate trajectory
void STDeformIfc::drawItdTraj()
{
  const GLfloat purpleMaterial[]={1.,0.,1.,1.};
  const GLfloat yellowMaterial[]={1.,1.,0.,1.};

  if(deformer_ == NULL) return;

  switch(trajType)
    {
    case HOLONOM_TRAJ : 
      break;
    case NH_TRAJ : 
      {
	int N = ((STNHDeformation*)deformer_)->itdTraj_.size();
	for(int i=0; i<N; i++)
	  {
	    if(i%10 == 0)
	      {
		glMaterialfv(GL_FRONT, GL_AMBIENT, yellowMaterial);
		glMaterialfv(GL_FRONT, GL_DIFFUSE, yellowMaterial);
	      }
	    else
	      {
		glMaterialfv(GL_FRONT, GL_AMBIENT, purpleMaterial);
		glMaterialfv(GL_FRONT, GL_DIFFUSE, purpleMaterial);
	      }

	    StateS6D cstate = ((STNHDeformation*)deformer_)->itdTraj_.at(i);
	    glPushMatrix();
	    glTranslated(cstate.x(),
			 cstate.y(),
			 cstate.t());
	    gluSphere(gluNewQuadric(), 1.0, 5, 5);
	    glPopMatrix();
	  }
      }
      break;
    case DD_TRAJ : 
      break;
    }

  const GLfloat cyanMaterial[]={0.,1.,1.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, cyanMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, cyanMaterial);
  
  switch(trajType)
    {
    case HOLONOM_TRAJ : 
      break;
    case NH_TRAJ : 
      {	
	int N = ((STNHDeformation*)deformer_)->itdTraj2_.size();
	for(int i=0; i<N; i++)
	  {
	    StateS6D cstate = ((STNHDeformation*)deformer_)->itdTraj2_.at(i);
	    glPushMatrix();
	    glTranslated(cstate.x(),
			 cstate.y(),
			 cstate.t());
	    gluSphere(gluNewQuadric(), 1.0, 5, 5);
	    glPopMatrix();
	  }
      }
      break;
    case DD_TRAJ : 
      break;
    }
}

void STDeformIfc::updDynamicView()
{
  if(this->traj_ != NULL && this->path_ != NULL)
    {
      // get back the dimensions of the map
      double dimMapX;
      double dimMapY;
      if(this->env_ == NULL || this->env_->getMap() == NULL)
	{
	  dimMapX = dfltXMap;
	  dimMapY = dfltYMap;
	}
      else
	{
	  dimMapX = this->env_->getMap()->getWidth();
	  dimMapY = this->env_->getMap()->getHeight();
	}
      
      // init the viewer
      ConfigR* center = new ConfigR2D(dimMapX/2, dimMapY/2);
      //double radius = path_->radius();
      double radius = sqrt(dimMapX*dimMapX/4 * dimMapY*dimMapY/4);
      double depth = 0.022 * radius;
      double dt = 3*(*traj_)[traj_->getSize()-1]->t();
      double upx,upy,upz;
      
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];
      
      this->glcanvas->init(center->x(), center->y() /*+ dt*/, depth + (time * 10.0), center->x(), center->y(), (time * 10.0), upx, upy, upz);
    }
}

void STDeformIfc::setLookAt(Environment* env,Path* path,Trajectory* traj)
{
  if(path != NULL)
    {
      ConfigR* center = path->center();
      double radius = path->radius();
      double depth = 5*radius;
      double dt;
      double upx,upy,upz;
      if(traj != NULL)
	{
	  dt = 3*(*traj)[traj->getSize()-1]->t();
	  
	  Vec v1(3);
	  v1.setVal(0,0);
	  v1.setVal(1,-dt);
	  v1.setVal(2,-depth);
	  double v1Norm = v1.norm();
	  Vec v2(3);
	  v2.setVal(0,0);
	  v2.setVal(1,-dt/v1Norm);
	  v2.setVal(2,-depth/v1Norm);
	  upx = 0;
	  upy = -v2[2];
	  upz = v2[1];
	}
      else
	{
	  dt = 0;
	  upx = 0;
	  upy = 1;
	  upz = 0;
	}
      gluLookAt(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
      delete center;
    }
  else
    {
      gluLookAt(5.,10.,15.,0.,0.,0.,0.,1.,0.);
    }
}

void STDeformIfc::drawScene()
{
  const GLfloat lightPosition[]={-20.,-100.,100.,1.};
  const GLfloat lightAmbient[]={.3f,.3f,.3f,1.};
  const GLfloat lightDiffuse[]={.9f,.9f,.9f,1.};
  const GLfloat redMaterial[]={1.,0.,0.,1.};
  const GLfloat blueMaterial[]={0.,0.,1.,1.};
  const GLfloat greenMaterial[]={0.,1.,0.,1.};
  const GLfloat otherMaterial[]={1.,0.,1.,1.};

  GLdouble width = glcanvas->getWidth();
  GLdouble height = glcanvas->getHeight();
  GLdouble aspect = height>0 ? width/height : 1.0;

  glcanvas->makeCurrent();

  glViewport(0,0,glcanvas->getWidth(),glcanvas->getHeight());

  glClearColor(1.0,1.0,1.0,1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable(GL_DEPTH_TEST);

  glDisable(GL_DITHER);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(this->path_ != NULL)
    {
      if(!orthoView) gluPerspective(30.,aspect,1.,-1);
      else glOrtho(-0.6*dfltXMap, 0.6*dfltXMap, -0.6*dfltYMap, 0.6*dfltYMap, -1000.0, 10000.0);
    }
  else
    {
      if(!orthoView) gluPerspective(30.,aspect,1.,100.);
      else glOrtho(-0.6*dfltXMap, 0.6*dfltXMap, -0.6*dfltYMap, 0.6*dfltYMap, -1000.0, 1000.0);
    }

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if(dynamicView)
    {
      this->updDynamicView();
    }

  gluLookAt(glcanvas->getEyeX(),glcanvas->getEyeY(),glcanvas->getEyeZ(),
	    glcanvas->getCenterX(),glcanvas->getCenterY(),glcanvas->getCenterZ(),
	    glcanvas->getUpX(),glcanvas->getUpY(),glcanvas->getUpZ());

//   cout << "Eye : (" << glcanvas->getEyeX() << "," << glcanvas->getEyeY() << "," << glcanvas->getEyeZ()<< ")  "
//        << "Center : (" << glcanvas->getCenterX() << "," << glcanvas->getCenterY() << "," << glcanvas->getCenterZ()<< ")  "
//        << "Up : (" << glcanvas->getUpX() << "," << glcanvas->getUpY() << "," << glcanvas->getUpZ()<< ")  "
//        << endl;


  glShadeModel(GL_SMOOTH);
  glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);

  glScalef(1.0,1.0,10.0);

  if(processType_ == TEDDY_PROC)
    {
      // draw the trajectory
      this->drawTraj();
      // draw the robot
      this->drawRobot();
      // draw the intermediate trajectory
      this->drawItdTraj();
//   // draw the trajectories generated for internal force computation
//   if(drawTrajGen)this->drawGenTraj();
    }
  else if(processType_ == TF_PROC)
    {
      // draw the trajectory
      this->drawTrajTF();
      // draw the robot
      this->drawRobotTF();      
      if(drawTrajGenTF) this->drawGenTrajTF();
    }

  // draw the obstacles
  this->drawObstacles();
  // draw the polygonal obstacles
  this->drawPolygonObs();
  // draw the map
  this->drawMap();
  // draw the roadmap
  if(roadmapView) this->drawRoadMap();

  // draw the forces 
  for(int i=0; i<lVec.size(); i = i+2)
    {
      if(i%6 == 0 || i%6 == 1)
	{
	  glColor3f(1.0f,0.0f,0.0f);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	}
      else if(i%6 == 2 || i%6 == 3)
	{
	  glColor3f(0.0f,1.0f,0.0f);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
	}
      else
	{
	  glColor3f(1.0f,1.0f,0.0f);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, otherMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, otherMaterial);
	}

      glBegin(GL_LINES);
      glVertex3f(lVec[i].x,lVec[i].y,lVec[i].z);
      glVertex3f(lVec[i+1].x,lVec[i+1].y,lVec[i+1].z);
      glEnd();
    }


  // Swap if it is double-buffered
  if(glvisual->isDoubleBuffer()){
    glcanvas->swapBuffers();
  }
  
  // Make context non-current
  glcanvas->makeNonCurrent();
}

int STDeformIfc::elemListExists(const FXString& s)
{
  int i=0;
  bool trouve = false;

  while(!trouve && i<list[0]->getNumItems())
    {
      if(list[0]->getItemText(i).text() == s)
	trouve = true;
      else i++;
    }
  
  if(trouve) return i;
  else return -1;
}

void STDeformIfc::initRandomDeform()
{
  int seq;

  if(optSeqNum == 0)
    {
      seq = initRandom();
    }
  else
    {
      seq = initRandom(seqNum);
    }
  seqNum = (FXint) seq;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[]){

  // Make application
  FXApp application("STTrajectoryDeformation","STTrajectoryDeformation");

  // Open the display
  application.init(argc,argv);

  // Make window
  new STDeformIfc(&application);

  // Create the application's windows
  application.create();

  // Run the application
  return application.run();
}
